/* Inline frame unwinder for GDB.

   Copyright (C) 2007 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "defs.h"
#include "addrmap.h"
#include "block.h"
#include "frame-unwind.h"
#include "symtab.h"

#include "gdb_assert.h"

static int inline_skip_frames;

/* Only valid if INLINE_SKIP_FRAMES is non-zero.  This is the PC used
   when calculating INLINE_SKIP_FRAMES; used to check whether we have
   moved to a new location by user request.  */
static CORE_ADDR inline_skip_pc;

/* Only valid if INLINE_SKIP_FRAMES is non-zero.  This is the symbol
   of the outermost skipped inline function.  It's used to find the
   call site of the current frame.  */
static struct symbol *inline_skip_symbol;

static void
inline_frame_this_id (struct frame_info *next_frame,
		      void **this_cache,
		      struct frame_id *this_id)
{
  struct symbol *func;

  /* In order to have a stable frame ID for a given inline function,
     we must get the stack / special addresses from the underlying
     real frame's this_id method.  So we must call get_prev_frame.
     Because we are inlined into some function, there must be previous
     frames, so this is safe - as long as we're careful not to
     create any cycles.  */
  *this_id = get_frame_id (get_prev_frame (get_prev_frame (next_frame)));

  /* We need a valid frame ID, so we need to be based on a valid
     frame.  FSF submission NOTE: this would be a good assertion to
     apply to all frames, all the time.  That would fix the ambiguity
     of null_frame_id (between "no/any frame" and "the outermost
     frame").  This will take work.  */
  gdb_assert (frame_id_p (*this_id));

  /* Future work NOTE: Alexandre Oliva applied a patch to GCC 4.3
     which generates DW_AT_entry_pc for inlined functions when
     possible.  If this attribute is available, we should use it
     in the frame ID (and eventually, to set breakpoints).  */
  func = get_frame_function (get_prev_frame (next_frame));
  gdb_assert (func != NULL);
  (*this_id).block_addr = BLOCK_START (SYMBOL_BLOCK_VALUE (func));
  (*this_id).inline_depth++;
}

static void
inline_frame_prev_register (struct frame_info *next_frame,
			    void **this_cache,
			    int prev_regnum,
			    int *optimizedp,
			    enum lval_type * lvalp,
			    CORE_ADDR *addrp,
			    int *realnump, gdb_byte *valuep)
{
  *optimizedp = 0;
  *lvalp = lval_register;
  *addrp = 0;
  *realnump = prev_regnum;
  if (valuep)
    frame_unwind_register (next_frame, *realnump, valuep);
}

/* Check whether we are at an inlining site that does not already
   have an associated frame.  */

static int
inline_frame_sniffer (const struct frame_unwind *self,
		      struct frame_info *next_frame,
		      void **this_cache)
{
  CORE_ADDR this_pc;
  struct block *frame_block, *cur_block;
  int depth;

  this_pc = frame_unwind_address_in_block (next_frame, INLINE_FRAME);
  frame_block = block_for_pc (this_pc);
  if (frame_block == NULL)
    return 0;

  /* Calculate DEPTH, the number of inlined functions at this
     location.  */
  depth = 0;
  cur_block = frame_block;
  while (BLOCK_SUPERBLOCK (cur_block))
    {
      if (block_inlined_p (cur_block))
	depth++;

      cur_block = BLOCK_SUPERBLOCK (cur_block);
    }

  /* Check whether we were requested to skip some frames, so they
     can be stepped into later.  */
  if (inline_skip_frames > 0 && frame_relative_level (next_frame) < 0)
    {
      if (this_pc != inline_skip_pc)
	inline_skip_frames = 0;
      else
	{
	  gdb_assert (depth >= inline_skip_frames);
	  depth -= inline_skip_frames;
	}
    }

  if (depth == 0)
    return 0;

  /* There are inlined functions here.  Check how many of them already
     have frames.  */
  while (next_frame && get_frame_type (next_frame) == INLINE_FRAME)
    {
      gdb_assert (depth > 0);
      depth--;
      next_frame = get_next_frame (next_frame);
    }

  /* If all the inlined functions already have frames, then pass to the
     normal unwinder for this PC.  */
  if (depth == 0)
    return 0;

  /* If NEXT_FRAME is an inlined function, but not the outermost, then
     we are the next outer.  If it is not an inlined function, then we
     are the innermost inlined function of a different real frame.  */
  return 1;
}

const struct frame_unwind inline_frame_unwinder = {
  INLINE_FRAME,
  inline_frame_this_id,
  inline_frame_prev_register,
  NULL,
  inline_frame_sniffer
};

const struct frame_unwind *const inline_frame_unwind = {
    &inline_frame_unwinder
};

/* Return non-zero if BLOCK, an inlined function block containing PC,
   has a group of contiguous instructions starting at PC (but not
   before it).  */

static int
block_starting_point_at (CORE_ADDR pc, struct block *block)
{
  struct blockvector *bv;
  struct block *new_block;

  bv = blockvector_for_pc (pc, NULL);
  if (BLOCKVECTOR_MAP (bv) == NULL)
    return 0;

  new_block = addrmap_find (BLOCKVECTOR_MAP (bv), pc - 1);
  if (new_block == NULL)
    return 1;

  if (new_block == block || contained_in (new_block, block))
    return 0;

  /* The immediately preceeding address belongs to a different block,
     which is not a child of this one.  Treat this as an entrance into
     BLOCK.  */
  return 1;
}

void
set_skipped_inline_frames (int skip_ok)
{
  CORE_ADDR this_pc;
  struct block *frame_block, *cur_block;
  struct symbol *last_sym = NULL;
  int skip_count = 0;

  if (!skip_ok)
    {
      if (inline_skip_frames != 0)
	{
	  inline_skip_frames = 0;
	  reinit_frame_cache ();
	}
      return;
    }

  /* This function is called right after reinitializing the frame
     cache.  We try not to do more unwinding than absolutely
     necessary, for performance.  */
  this_pc = get_frame_pc (get_current_user_frame ());
  frame_block = block_for_pc (this_pc);

  if (frame_block != NULL)
    {
      cur_block = frame_block;
      while (BLOCK_SUPERBLOCK (cur_block))
	{
	  if (block_inlined_p (cur_block))
	    {
	      /* See comments in inline_frame_this_id about this use
		 of BLOCK_START.  */
	      if (BLOCK_START (cur_block) == this_pc
		  || block_starting_point_at (this_pc, cur_block))
		{
		  skip_count++;
		  last_sym = BLOCK_FUNCTION (cur_block);    
		}
	      else
		break;
	    }
	  cur_block = BLOCK_SUPERBLOCK (cur_block);
	}
    }

  inline_skip_pc = this_pc;
  inline_skip_symbol = last_sym;

  if (inline_skip_frames != skip_count)
    {
      inline_skip_frames = skip_count;
      reinit_frame_cache ();
    }
}

void
step_into_inline_frame (void)
{
  gdb_assert (inline_skip_frames > 0);
  inline_skip_frames--;
  reinit_frame_cache ();
}

int
inline_skipped_frames (void)
{
  return inline_skip_frames;
}

struct symbol *
inline_skipped_symbol (void)
{
  return inline_skip_symbol;
}
