/*
 * Copyright (c) 2000-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * flashtest.c: Flash memory demonstration
 *
 * Test programming, erasing and reading Flash memory
 */


#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

/* Use /dev/flash interface */
#define POSIX_INTERFACE

#ifdef POSIX_INTERFACE
/* /dev/flash interface */
#include <sys/ioctl.h>
#include <sys/flashio.h>
int devfd;
#else
/* direct flash call */
#include <mips/cpu.h>
#include <sbd.h>
#include <../share/flashdev.h>
flashcookie_t cookie;
#endif

/* flash sector information */
struct flashinfo info;

/* sector buffers */
char *pbuf;
char *ebuf;
char *rbuf;

/* status flags */
int neednl;
int failures;


/* 
 * nl()
 * Output a newline if one is pending.
 */
void
nl (void)
{
    if (neednl) {
	printf ("\n"); 
	neednl = 0;
    }
}


/* 
 * start()
 * Display the name of the current operation.
 */
void
start (const char *msg, ...)
{
    va_list ap;

    printf ("     ");
    va_start (ap, msg);
    vprintf (msg, ap);
    va_end (ap);
    neednl = 1;
}


/* 
 * percent()
 * Display progress through operation
 */
void
percent (unsigned int offs, unsigned int min, unsigned int max)
{
    static unsigned int lastpercent = 101;
    unsigned int percent = (offs - min) * 100 / (max - min);
    if (percent != lastpercent) {
	printf ("\r%3d%%", percent);
	lastpercent = percent;
	neednl = 1;
    }
}

/* 
 * done()
 * Indicate that an operation is complete
 */
void
done ()
{
    printf ("\rDONE\n");
    neednl = 0;
}


#ifndef POSIX_INTERFACE
/* 
 * flasherror()
 * Return an error message for a flash API return value
 */
static const char *
flasherror (int result) 
{
    static char buf[40];

    switch (result) {
    case FLASHDEV_OK:
	return "completed ok";
    case FLASHDEV_FAIL:
	return "parameter error";
    case FLASHDEV_PROTECTED:
	return "can't program protected sector";
    case FLASHDEV_PARTIAL:
	return "can't erase partial sector";
    case FLASHDEV_NOMEM:
	return "no programming buffer space";
    case FLASHDEV_FATAL:
	return "fatal programming error";
    default:
	sprintf (buf, "unknown flash code %d\n", result);
	return buf;
    }
}
#endif


/* 
 * err()
 * Display an error message including sector address.
 */
void
err (const char *when, int code, unsigned int offs, unsigned int len)
{
    nl ();
#ifdef POSIX_INTERFACE
    printf ("%s (%s) for 0x%06x-0x%06x\n",
	    when, strerror (code), offs, offs + len);
#else
    printf ("%s (%s) for 0x%06x-0x%06x\n",
	    when, flasherror (code), offs, offs + len);
#endif
    failures++;
}


/* 
 * sectinfo()
 * Retrieve information (geometry) regarding sector at offset OFFS
 * into global structure INFO.
 */
int
sectinfo (unsigned int offs)
{
#ifdef POSIX_INTERFACE
    info.soffs = offs;
    return ioctl (devfd, FLASHIOINFO, &info);
#else
    return flash_info (cookie, offs, &info);
#endif
}


/* 
 * program()
 * Program LEN bytes from BUF to flash offset OFFS.
 */
int
program (const char *buf, unsigned int offs, unsigned int len)
{
    int code = 0;

#ifdef POSIX_INTERFACE
    if (pwrite (devfd, buf, len, offs) != len)
	code = errno;
#else
    code = flash_programbytes (cookie, offs, buf, len,
			       FLASHDEV_PROG_MERGE|FLASHDEV_PROG_STREAM);
#endif

    if (code != 0) {
	err ("PROGRAM FAILURE", code, offs, len);
	return 1;
    }
    return 0;
}


/* 
 * flush()
 * Flush any pending programmed bytes to Flash.
 */
int
flush (void)
{
#ifdef POSIX_INTERFACE
    return ioctl (devfd, FLASHIOFLUSH, 0);
#else
    return flash_flush (cookie);
#endif
}


/* 
 * getbytes()
 * Read LEN bytes from flash offset OFFS into global buffer RBUF.
 */
int
getbytes (unsigned int offs, unsigned int len)
{
    int code = 0;

    if (len > info.maxssize)
	abort ();

#ifdef POSIX_INTERFACE
    if (pread (devfd, rbuf, len, offs) != len)
	code = errno;
#else
    code = flash_readbytes (cookie, offs, rbuf, len, FLASHDEV_READ_STREAM);
#endif

    if (code != 0) {
	err ("READ ERROR", code, offs, len);
	return 1;
    }

    return 0;
}


/* 
 * compare()
 * Compare buffer PB with flash data at offset OFFS, length LEN.
 */
int
compare (char *pb, unsigned int offs, unsigned int len)
{
    int match, errs, toterrs, repeat, i;

    /* read the flash */
    if (getbytes (offs, len) != 0)
	return 1;

    /* quick test for exact match */
    if (memcmp (pb, rbuf, len) == 0)
	return 0;

    nl ();
    match = errs = toterrs = repeat = 0;
    for (i = 0; i < len; i++) {
	if (pb[i] == rbuf[i]) {
	    if (errs > 16 && ++match >= 16) {
		if (++repeat >= 4) {
		    if (repeat == 4)
			printf ("... and so on\n");
		}
		else {
		    printf ("... until 0x%06x\n", offs + i - 16);
		    errs = 0;
		}
	    }		    
	    continue;
	}
	else {
	    match = 0;
	    ++toterrs;
	    if (++errs <= 16) {
		nl ();
		printf ("0x%06x: EXPECT 0x%02x GOT 0x%02x XOR 0x%02x\n",
			offs + i, pb[i], rbuf[i], pb[i] ^ rbuf[i]);
	    }
	}
    }
    if (errs > 16)
	printf ("... until 0x%06x (%u/%u errs)\n", offs + i, toterrs, len);
    failures++;
    return 1;
}


/* 
 * erasesect()
 * Erase single Flash sector at OFFS.
 */
int
erasesect (unsigned int offs)
{
    int code = 0;

#ifdef POSIX_INTERFACE
    if (ioctl (devfd, FLASHIOERASESECT, &offs) != 0)
	code = errno;
#else
    code = flash_erasesector (cookie, offs);
#endif

    if (code != 0)
	err ("SECT ERASE ERROR", code, offs, 0);
    return code;
}


/* 
 * eraserange()
 * Erase a large region (possibly all) of the Flash
 */
void
eraserange (unsigned int minoffs, unsigned int maxoffs)
{
    unsigned int offs;
    int code = -1;

    if (minoffs == 0 && maxoffs == info.size) {
	start  ("Erasing whole flash"); 
	/* attempt to erase whole device quickly */
#ifdef POSIX_INTERFACE	
	/* ioctl always erases whole device */
	if (ioctl (devfd, FLASHIOERASEDEV, 0) != 0)
	    err ("DEV ERASE ERROR", errno, 0, maxoffs);
	return;
#else
	/* flash_erasedevice may return indicating that
	   sector by sector erasure is required */
	code = flash_erasedevice (cookie);
	if (code != 0 && code != FLASHDEV_FAIL) {
	    err ("DEV ERASE ERROR", code, 0, maxoffs);
	    return;
	}
#endif
    }
    else {
	start  ("Erasing test region"); 
    }

    /* erase device a sector at a time */
    if (code != 0) {
	for (offs = minoffs; offs < maxoffs; offs += info.ssize) {
	    percent (offs, minoffs, maxoffs);
	    sectinfo (offs);
	    erasesect (offs);
	}
    }
    done ();
	
    /* now check that the whole device is erased */
    start ("Blank check");
    for (offs = minoffs; offs < maxoffs; offs += info.ssize) {
	percent (offs, minoffs, maxoffs);
	sectinfo (offs);
	compare (ebuf, offs, info.ssize);
    }
    done ();
}



/* 
 * ftest()
 * Test one flash memory device.
 * If ERASE is non-zero force an erase of some of the flash (but
 * excluding any non-volatile environment store).
 * If READONLY is non-zero just display info about the device.
 * MIN and MAX are the region of flash to be erased.
 */
int
ftest (int erase, int readonly, long min, long max)
{
    struct flashparts parts;
    unsigned int offs, minoffs, maxoffs;
    unsigned int curmin, curmax;
    int n, i, ret;
    unsigned int mid, foffs, eoffs, a, b;
    int prevempty, prevprot, match;
    unsigned int prevsize;
    int anyprot = 0;

#ifdef POSIX_INTERFACE
    if (devfd < 0) {
	if (errno == ENXIO)
	    printf ("Unknown device\n");
	else if (errno == ENODEV)
	    printf ("Missing device\n");
	else
	    printf ("%s\n", strerror (errno));
	return 0;
    }
#else
    if (cookie == FLASH_UNKNOWN) {
	printf ("Unknown device\n");
	return 0;
    }
    if (cookie == FLASH_MISSING) {
	printf ("Missing device\n");
	return 0;
    }
#endif
	
    if (sectinfo (0)) {
	printf ("Can't get flash info\n");
	return 0;
    }

    /* Allocate sector buffers */
    pbuf = (char *) malloc (info.maxssize);
    ebuf = (char *) malloc (info.maxssize);
    rbuf = (char *) malloc (info.maxssize);

    /* Create simple address pattern (incrementing addr & ~addr) */
    for (i = 0; i < info.maxssize / 2; i += 4)
	*(unsigned int *)&pbuf[i] = i;
    for (;i < info.maxssize; i += 4)
	*(unsigned int *)&pbuf[i] = ~i;

    /* Create "empty" pattern */
    memset (ebuf, ~0, info.maxssize);

    printf ("Device:\t%s\n", info.name); 
    printf ("Base:\t0x%08x\n", info.base); 
    printf ("Size:\t0x%08x\n", info.size); 
    printf ("MapAdr:\t0x%08x\n", info.mapbase); 
    printf ("Unit:\t%d byte%s / %d bits\n", info.unit, 
	    info.unit != 1 ? "s" : "", info.unit * 8); 
    printf ("MaxSec:\t%d bytes\n", info.maxssize); 
    printf ("Sectors:\n");

    /* Display a list of sectors, and simultaneously find the
       largest blank region. */

    minoffs = maxoffs = -1;
    curmin = curmax = -1;
    prevempty = prevprot = -1;
    for (n = offs = 0; offs < info.size; offs += info.ssize, n++) {
	int empty;
	
	sectinfo (offs);
	if (offs != info.soffs) {
	    printf (" %2d: %06x/%06x: offset mismatch\n",
		    n, offs, info.soffs);
	    exit (1);
	}

	if (getbytes (offs, info.ssize))
	    empty = 0;
	else
	    empty = (memcmp (ebuf, rbuf, info.ssize) == 0);

	/* if sector matches previous one, and isn't the last sector
	   then don't print it out */
	if (empty == prevempty
	    && info.sprot == prevprot
	    && info.ssize == prevsize
	    && offs + info.ssize < info.size) {
	    ++match;
	}
	else {
	    prevempty = empty;
	    prevprot = info.sprot;
	    prevsize = info.ssize;
	    match = 0;
	}

	if (match <= 1) {
	    printf (" %4d: %07x: %6d bytes (%s, %s)\n", 
		    n, offs, info.ssize, empty ? "empty" : "USED",
		    info.sprot ? "PROTECTED" : "unprotected");
	}
	else if (match == 2) {
	    printf ("       ...\n");
	}

	anyprot |= info.sprot;

	/* find the largest contiguous section of blank, unprotected flash */
	if (!empty || info.sprot) {
	    curmin = curmax = -1;
	}
	else {
	    if (curmin == -1)
		curmin = offs;
	    curmax = offs + info.ssize;
	    if ((curmax - curmin) > (maxoffs - minoffs)) {
		minoffs = curmin;
		maxoffs = curmax;
	    }
	}
    }

    /* Display the "partition" table. */
    printf ("Partitions:\n");
#ifdef POSIX_INTERFACE
    ioctl (devfd, FLASHIOGPARTS, &parts);
#else
    flash_parts (cookie, &parts);
#endif
    for (n = 0; n < FLASHNPART; n++) {
	struct flashpart *p = &parts.part[n];
	if (p->size != 0) {
	    printf (" %3d: %07x-%07x (%s)\n", 
		    n, p->offs, p->offs + p->size - 1, 
		    p->type == FLASHPART_UNDEF ? "undef" :
		    p->type == FLASHPART_RAW ? "raw" :
		    p->type == FLASHPART_BOOT ? "boot" :
		    p->type == FLASHPART_POST ? "post" :
		    p->type == FLASHPART_ENV ? "env" :
		    p->type == FLASHPART_FFS ? "ffs" : "???");
	}
    }

    /* Select the whole flash (excluding environment store), if total
       erasure has benn requested. */
    if (erase) {
	int i;

	/* default to whole flash */
	minoffs = (min < 0) ? 0 : min;
	maxoffs = (max < 0) ? info.size : max;
	
	/* but don't touch environment */
	for (i = 0; i < FLASHNPART; i++) {
	    if (parts.part[i].type == FLASHPART_ENV) {
		unsigned int pstart = parts.part[i].offs;
		unsigned int pend = pstart + parts.part[i].size;
		if (pstart < maxoffs && pstart > minoffs)
		    maxoffs = pstart;
		else if (pend > minoffs && pend < maxoffs)
		    minoffs = pend;
	    }
	}
    }

    if (minoffs == maxoffs) {
	printf ("Cannot find blank, writable section\n");
	return 0;
    }

    printf ("\nTest region: 0x%07x - 0x%07x (%u bytes)\n",
	    minoffs, maxoffs, maxoffs - minoffs);

    if (readonly)
	return 1;

    /* Erase the flash, if requested */
    if (erase)
	eraserange (minoffs, maxoffs);

    /* Program the whole of the flash region, sector by sector,
       checking as we go. */
    start ("Program test region");
    failures = 0;
    for (offs = minoffs; offs < maxoffs; offs += info.ssize) {
	percent (offs, minoffs, maxoffs);
	sectinfo (offs);
	if (info.sprot) 
	    continue;
	if (program (pbuf, offs, info.ssize) == 0)
	    compare (pbuf, offs, info.ssize);
    }
    done ();

    /* Now check the data in the flash as a whole - to make sure 
       previous programming has not been disturbed. */
    start ("Checking test region");
    for (offs = minoffs; offs < maxoffs; offs += info.ssize) {
	percent (offs, minoffs, maxoffs);
	sectinfo (offs);
	if (info.sprot) 
	    continue;
	compare (pbuf, offs, info.ssize);
    }
    done ();

    /* Erase the first and last sectors */
    start ("Erasing first test sector");
    erasesect (minoffs);
    done ();

    start ("Erasing last test sector");
    erasesect (maxoffs - 1);
    done ();

    /* Check all the data a final time - to make sure that the sector
       erasures have not disturbed any other data. */
    start ("Checking test region");
    for (offs = minoffs; offs < maxoffs; offs += info.ssize) {
	percent (offs, minoffs, maxoffs);
	sectinfo (offs);
	if (info.sprot) 
	    continue;
	if (offs == minoffs || info.soffs + info.ssize == maxoffs)
	    compare (ebuf, offs, info.ssize);
	else
	    compare (pbuf, offs, info.ssize);
    }
    done ();

    /* Program middle of flash, unaligned, and straddling two sectors */
    if (failures == 0 && maxoffs - minoffs > info.maxssize * 3) {
	mid = (maxoffs + minoffs) / 2;	/* midpoint */
	mid &= ~(info.maxssize - 1);	/* align mid on sect boundary */
	foffs = mid - (info.maxssize / 2) - 2;
	eoffs = foffs + info.maxssize;

	start ("Program two partial sectors [%06x-%06x]", 
	       foffs, eoffs);
	
	if (program (pbuf, foffs, info.maxssize) == 0
	    && flush () == 0) {
	
	    done ();

	    /* before test region should be unchanged */
	    a = mid - info.maxssize;
	    b = foffs;
	    start ("Check before (%x-%x)", a, b);
	    compare (&pbuf[a % info.maxssize], a, b - a);
	    done ();

	    /* test region itself should be as programmed */
	    a = foffs;
	    b = eoffs;
	    start ("Check range (%x-%x)", a, b);
	    compare (pbuf, a, b - a);
	    done ();
	    
	    /* after test region should be unchanged */
	    a = eoffs;
	    b = mid + info.maxssize;
	    start ("Check after (%x-%x)", a, b);
	    compare (&pbuf[a % info.maxssize], a, b - a);
	    done ();
	}
    }

    /* Start with a clean sector */
    if (failures == 0) {
	start ("Erasing first sector again");
	erasesect (minoffs);
	sectinfo (minoffs);
	compare (ebuf, minoffs, info.ssize);
	done ();
    }

    /* Program sector one byte at a time (merging) */
    if (failures == 0) {
	start ("Byte program first sector");
	for (i = 0; i < info.ssize && failures == 0; i++) {
	    percent (i, 0, info.ssize);
	    program (&pbuf[i], minoffs + i, 1);
	}
	flush ();
	if (failures == 0)
	    compare (pbuf, minoffs, info.ssize);
	done ();
    }

    /* Simple bulk data pattern tests, just clearing bits */
    if (failures == 0) {
	start ("Write 0x55 to first sector");
	memset (pbuf, 0x55, info.ssize);
	if (program (pbuf, minoffs, info.ssize) == 0)
	    compare (pbuf, minoffs, info.ssize);
	done ();
    }

    if (failures == 0) {
	start ("Write 0x00 to first sector");
	memset (pbuf, 0x00, info.ssize);
	if (program (pbuf, minoffs, info.ssize) == 0)
	    compare (pbuf, minoffs, info.ssize);
	done ();
    }

    /* Now set bits */
    if (failures == 0) {
	start ("Write 0xaa to first sector");
	memset (pbuf, 0xaa, info.ssize);
	if (program (pbuf, minoffs, info.ssize) == 0)
	    compare (pbuf, minoffs, info.ssize);
	done ();
    }

    /* Final clean up */
    eraserange (minoffs, maxoffs);
    return failures == 0;
}


int main (int argc, char **argv, char **envp)
{
    char devname[32];
    int erase = 0, readonly = 0;
    int dev = -1, n;
    long min = -1, max = -1;
    int opti;

    opti = 1;
    while (opti < argc) {
	char *ep = "";
	if (isdigit (argv[opti][0]))
	    dev = strtol (argv[opti], &ep, 10);
	else if (strncmp (argv[opti], "/dev/flash",
			  sizeof "/dev/flash" - 1) == 0)
	    dev = strtol (argv[opti] + sizeof "/dev/flash" - 1, 
			  &ep, 10);
        else if (strcmp (argv[opti], "erase") == 0) {
	    if (opti >= argc - 2) {
		fprintf (stderr, "missing start & end for erase command\n");
		exit (1);
	    }
	    ++erase;
	    min = strtoul (argv[++opti], &ep, 16);
	    if (*ep == '\0')
		max = strtoul (argv[++opti], &ep, 16);
	}
        else if (strcmp (argv[opti], "eraseall") == 0) {
	    ++erase;
	}
	else if (strcmp (argv[opti], "read") == 0)
	    readonly = 1;
	else {
	    fprintf (stderr, "unknown argument: %s\n", argv[opti]);
	    exit (1);
	}
	if (*ep) {
	    fprintf (stderr, "bad numeric argument: %s\n", argv[opti]);
	    exit (1);
	}
	++opti;
    }

    if (erase > 1 || (erase && readonly)) {
	fprintf (stderr, "bad argument combination\n");
	exit (1);
    }

    setvbuf (stdout, NULL, _IONBF, 0);
    if (dev >= 0) {
#ifdef POSIX_INTERFACE
	sprintf (devname, "/dev/flash%d", dev);
	devfd = open (devname, readonly ? O_RDONLY : O_RDWR);
	printf ("Flash device %s\n", devname);
#else
	cookie = flash_open (dev);
	printf ("Flash device #%d\n", dev);
#endif
	ftest (erase, readonly, min, max);
    }
    else {
#ifdef POSIX_INTERFACE
	for (dev = n = 0; dev < 10; dev++) {
	    sprintf (devname, "/dev/flash%d", dev);
	    devfd = open (devname, readonly ? O_RDONLY : O_RDWR);
	    if (devfd >= 0) {
		printf ("Flash device %s\n", devname);
		ftest (erase, readonly, min, max);
		++n;
	    }
	}
#else
	for (dev = n = 0; (cookie = flash_open (dev)) != FLASH_UNKNOWN; dev++) {
	    if (cookie) {
		printf ("Flash device #%d\n", dev);
		ftest (erase, readonly, min, max);
		++n;
	    }
	}
#endif
	if (n == 0)
	    fprintf (stderr, "\n\n*** No flash devices found\n");
    }
    exit (0);
}
