/*
 * Copyright (c) 2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ex1/hello.c: MIPS SDE example #1 (hello world!)
 */


/*
 * Example #1
 * A very basic operational confidence test for low-level hardware
 * and platform specific i/o code.
 *
 * NOTE until you get past this first stage you cannot expect
 * to remotely debug, because the remote debugger interface
 * is written in C code and does I/O.  Until then you're stuck
 * with a logic analyser.
 *
 * - data and bss loaded in uncached memory
 * - check simple instructions and partial word memory access
 * - checks execution relatively complex C code (stdio)
 *   NOTE stderr and stdout are connected to the (rom) console device
 *
 * NOTE 
 * writing to stdout (a buffered stream) requires malloc to work,
 * since libc malloc is horrendously complicated, try compiling with
 * NODDYMALLOC defined if malloc crashes program.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <machine/cpu.h>
#include <machine/endian.h>

#define abort() __asm("break 1")

union memfrag {
    struct {
	unsigned int    abcd;
    } ui;
    struct {
#if BYTE_ORDER == BIG_ENDIAN
	unsigned short  ab, cd;
#else
	unsigned short  cd, ab;
#endif
    } us;
    struct {
#if BYTE_ORDER == BIG_ENDIAN
	unsigned char   a, b, c, d;
#else
	unsigned char   d, c, b, a;
#endif
    } uc;
};


static void
memchk (volatile union memfrag *up, unsigned v)
{
    if (up->ui.abcd != v)
	abort ();
    if (up->us.ab != (v >> 16))
	abort ();
    if (up->us.cd != (v & 0xffff))
	abort ();
    if (up->uc.a != ((v >> 24) & 0xff))
	abort ();
    if (up->uc.b != ((v >> 16) & 0xff))
	abort ();
    if (up->uc.c != ((v >> 8) & 0xff))
	abort ();
    if (up->uc.d != (v & 0xff))
	abort ();
}


int
main (int argc, char **argv, char **envp)
{
    extern void *sbrk (int);
    char *s;

    /* test various simple instruction sequences 
     * please dont OPTIMISE this away
     */
    {
	volatile int    a = 5;
	volatile int    b = 6;
	volatile int    d = 25;
	volatile int    e = 5;

	if (a != e)
	    abort ();
	if (a == b)
	    abort ();
	if (a >= b)
	    abort ();
	if (a > b)
	    abort ();
	if (b <= a)
	    abort ();
	if (b < a)
	    abort ();
	if (a < 0)
	    abort ();
	if (a == 0)
	    abort ();

	if ((unsigned)a == (unsigned)b)
	    abort ();
	if ((unsigned)a >= (unsigned)b)
	    abort ();
	if ((unsigned)a > (unsigned)b)
	    abort ();
	if ((unsigned)b <= (unsigned)a)
	    abort ();
	if ((unsigned)b < (unsigned)a)
	    abort ();
	if ((unsigned)a < 0)
	    abort ();
	if ((unsigned)a == 0)
	    abort ();

	/* check multiplier */
	if (a * a != d)
	    abort ();

	/* check divider */
	if (d / a != a)
	    abort ();
	if (a % d != a)
	    abort ();

	/* check constant divide, which may be done as a mult/mfhi */
	if (d / 5 != a)
	    abort ();
	if (d % 5 != 0)
	    abort ();
	if (a % 25 != 5)
	    abort ();
    }

    /* up until now we have probably only done word accesses
     * check half and byte read write operations, uncached
     */
    {
	union memfrag u;
	volatile union memfrag *up = (union memfrag *) KVA0_TO_KVA1 (&u);

	up->ui.abcd = 0xaabbccdd;
	memchk (up, 0xaabbccdd);

	up->us.ab = 0x0011;
	memchk (up, 0x0011ccdd);
	up->us.cd = 0x2233;
	memchk (up, 0x00112233);

	up->ui.abcd = 0xaabbccdd;
	memchk (up, 0xaabbccdd);

	up->uc.a = 0x44;
	memchk (up, 0x44bbccdd);
	up->uc.b = 0x55;
	memchk (up, 0x4455ccdd);
	up->uc.c = 0x66;
	memchk (up, 0x445566dd);
	up->uc.d = 0x77;
	memchk (up, 0x44556677);
    }

    /* also check tri-byte writes */
    {
	char dst[16], *dp;
	char src[16], *sp;
	int i;

	sp = (char *) KVA0_TO_KVA1 (src);
	dp = (char *) KVA0_TO_KVA1 (dst);

	/* fill arrays */
	for (i = 0; i < sizeof(src); i++)
	    dp[i] = sp[i] = i;

	/* do completely unaligned copy */
	memcpy (&dp[1], &sp[2], sizeof (src) - 2);

	/* check limits unmodified */
	if (dp[0] != 0)
	    abort ();
	if (dp[sizeof(dst) - 1] != sizeof(dst) - 1)
	    abort ();

	/* check copy */
	for (i = 0; i < sizeof(src) - 2; i++)
	    if (dp[i + 1] != i + 2)
		abort ();
    }

    /* straight to "monitor", character at a time */
#define _(x) _mon_putc(x);
    _('\r') _('\n') _('H') _('e') _('l') _('l') _('o')
#undef _

    /* straight to "monitor", string at a time */
    _mon_puts (" World! (monitor direct)\r\n");

    /* low-level POSIX I/O */
    s = "Hello World! (posix syscall)\n";
    write (STDERR_FILENO, s, strlen (s));

    /* stdio */
    fputs ("Hello World! (unbuffered stdio)\n", stderr);

    /* test sbrk as we need it for malloc */
    s = sbrk (16384);
    if (s == (void *)-1) {
	extern void *_minbrk, *_maxbrk;
	perror ("SBRK FAILED");
	fprintf (stderr, "[_minbrk=%p _maxbrk=%p]\n", _minbrk, _maxbrk);
	abort ();
    }
#ifdef DEBUG
    fprintf (stderr, "[sbrk(16384) returns %p]\n", s);
#endif
    sbrk (-16384);

    /* test malloc as we need it for stdio on buffered streams */
    s = malloc (8192);
    if (!s) {
	perror ("MALLOC FAILED");
	abort ();
    }
#ifdef DEBUG
    fprintf (stderr, "[malloc(8192) returns %p]\n", s);
#endif
    free (s);

    /* finally, fully buffered stdio */
    fputs ("Hello World! (buffered stdio)\n", stdout);

    return EXIT_SUCCESS;
}


#ifdef NODDYMALLOC
void           *
malloc (size_t n)
{
    void           *p = sbrk (n);

    return (p == (void *)-1) ? 0 : p;
}

void
free (void *p)
{
}

#endif
