/*
 * Copyright (c) 2003-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ex3/minimon.c: MIPS SDE example #3 (command-line monitor)
 */


/*
 * Example #3
 * A VERY simple command line monitor.
 */

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <setjmp.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <limits.h>
#include <fcntl.h>
#include <alloca.h>
#include <sys/clock.h>
#include <sys/time.h>

/* processor-specific stuff */
#include <mips/cpu.h>
#include <mips/xcpt.h>

extern char   **environ;

static int
Usage (const char *s)
{
    fprintf (stderr, "Usage: %s\n", s);
    return EXIT_FAILURE;
}


struct comtab {
    const char     *name;
    const char     *usage;
    const char     *help;
    int           (*handler) (int argc, char **argv, char **envp);
};

static const struct comtab comtab[];


static const char usg_help[] = "help [COMMAND]";
static const char hlp_help[] = "give brief help on commands";

static int
com_help (int argc, char **argv, char **envp)
{
    static int maxusg = 0;
    const struct comtab *ct;
    int i;
    
    if (argc == 2) {
	for (ct = comtab; ct->name; ct++)
	    if (strcmp (argv[1], ct->name) == 0)
		break;
	if (ct->name) 
	    printf ("%s\t%s\n", ct->usage, ct->help);
	else
	    fprintf (stderr, "%s: not found\n", argv[1]);
	return 0;
    }

    if (maxusg == 0) {
	for (ct = comtab; ct->name; ct++) {
	    int len = strlen (ct->usage);
	    if (len > maxusg) maxusg = len;
	}
    }
	    
    printf ("Available commands:\n");
    for (ct = comtab; ct->name; ct++)
	printf ("  %-*s\t%s\n", maxusg, ct->usage, ct->help);

    return 0;
}


static const char usg_exit[] = "exit [N]";
static const char hlp_exit[] = "terminate monitor (status N)";

static int
com_exit (int argc, char **argv, char **envp)
{
    if (argc > 2)
	return Usage (usg_exit);
    if (argc == 2)
	exit (atoi (argv[1]));
    else
	exit (0);
}


#define MB	0x100000
#define KB	0x400

static const char usg_stat[] = "stat";
static const char hlp_stat[] = "report system status";

static int
com_stat (int argc, char **argv, char **envp)
{
    static const int fraction[] = {1000000, 100000, 10000, 1000, 100, 10, 1};
    size_t  mem_size;
    unsigned long freq, frac;
    int ntlb;

    if (argc != 1)
	return Usage (usg_stat);

    printf ("PrID:\t\t%.8x\n", mips_getprid ());
    printf ("Status:\t\t%.8x\n", mips_getsr ());
    printf ("Cause:\t\t%.8x\n", mips_getcr ());

    ntlb = mips_tlb_size ();
    if (ntlb <= 0)
	printf ("TLB:\t\tnone\n");
    else
	printf ("TLB:\t\t%d entries\n", ntlb);

    printf ("FPU:\t\t%s\n", fpa_enable () ? "present" : "absent");

#define MEMPREC	1		/* 1 decimal point of precision */

    mem_size = sysconf (_SC_PHYS_PAGES) * sysconf (_SC_PAGESIZE);
    mem_size += fraction[MEMPREC] / 2;
    frac = (mem_size % MB) / fraction[MEMPREC];

    printf ("Memory Size:\t% 3lu.%0*lu MB\n", mem_size / MB, MEMPREC, frac);

#define ONEMHZ		(1000*1000)
#define FREQPREC	2	/* 2 decimal points of precision */

    freq = _sbd_pipefreq ();
    freq += fraction[FREQPREC] / 2;
    frac = (freq % ONEMHZ) / fraction[FREQPREC];

    if (frac == 0)
	printf ("PClock:\t\t% 3u MHz\n", freq / ONEMHZ);
    else
	printf ("PClock:\t\t% 3u.%0*u MHz\n", freq / ONEMHZ, FREQPREC, frac);

    freq = _sbd_cpufreq ();
    freq += fraction[FREQPREC] / 2;
    frac = (freq % ONEMHZ) / fraction[FREQPREC];

    if (frac == 0)
	printf ("Bus Clock:\t% 3u MHz\n", freq / ONEMHZ);
    else
	printf ("Bus Clock:\t% 3u.%0*u MHz\n", freq / ONEMHZ, FREQPREC, frac);

    return 0;
}

static const char usg_cache[] = "cache";
static const char hlp_cache[] = "display cache info";

static int
com_cache (int argc, char **argv, char **envp)
{
    extern char     _fdata[];
    extern char     _edata[];

    if (argc != 1)
	return Usage (usg_cache);

#if ! #cache(no)
    mips_size_cache ();

    if (mips_icache_size > 0)
	printf ("Icache: %5u bytes, %u bytes/line (%d way)\n",
		mips_icache_size, mips_icache_linesize,
		mips_icache_ways);
    else
#endif
	printf ("Icache: none detected!\n");

#if ! #cache(no)
    if (mips_dcache_size > 0)
	printf ("Dcache: %5u bytes, %u bytes/line (%d way)\n", 
		mips_dcache_size, mips_dcache_linesize,
		mips_dcache_ways);
    else
#endif
	printf ("Dcache: none detected!\n");

#if ! #cache(no)
    if (mips_scache_size > 0)
	printf ("Scache: %4uK bytes, %u bytes/line (%d way, %s)\n", 
		mips_scache_size / 1024, mips_scache_linesize,
		mips_scache_ways,
		mips_scache_split ? "split" 
		: mips_scache_discontig ? "discontiguous" 
		: "unified");

    if (mips_tcache_size > 0)
	printf ("Tcache: %4uK bytes, %u bytes/line (%d way)\n", 
		mips_tcache_size / 1024, mips_tcache_linesize,
		mips_tcache_ways);
	
    if (mips_dcache_size > 0) {
	/* a very very poor cache test !! */
	mips_flush_cache ();
	printf ("Trivial cache test: %s\n",
		memcmp (KVA0_TO_KVA1(_fdata), KVA1_TO_KVA0 (_fdata),
			_edata - _fdata) ? "failed" : "passed");
    }
#endif

    return 0;
}

static const char usg_time[] = "time";
static const char hlp_time[] = "display date and time";

static int
com_time (int argc, char **argv, char **envp)
{
    time_t          t;
    clock_t 	    c;

    if (argc != 1)
	return Usage (usg_time);

    if ((t = time (0)) == -1)
	perror ("time");
    else
	fputs (asctime (gmtime (&t)), stdout);

    if ((c = clock ()) == -1)
	perror ("clock");
    else
      	printf ("%u.%03u secs elapsed\n", c / CLOCKS_PER_SEC,
		(c % CLOCKS_PER_SEC) / 1000);
    
    return 0;
}


static const char usg_itimer[] = "itimer [NSECS]";
static const char hlp_itimer[] = "test interval timer for NSECS (default 120)";

static volatile int ticker;

static void
alrmhandler (int sig)
{
    ++ticker;
}


static int
com_itimer (int argc, char **argv, char **envp)
{
    struct itimerval it, oit;
    clock_t start, last, now;
    int limit, stuck, diff, secs, lastsecs;
    void (*oint)(int);
    void (*oalrm)(int);
    int errs = 0;

    if (argc > 2)
	return Usage (usg_time);
    if (argc == 2)
	limit = atoi (argv[1]);
    else
#if defined(R3KEMU) || defined(GNUSIM)
	limit = 5;
#else
	limit = 120;
#endif
	    
    fprintf (stderr, "% 3d ", limit);

    oint = signal (SIGINT, SIG_IGN);
    oalrm = signal (SIGALRM, alrmhandler);

    ticker = 0;
    it.it_interval.tv_sec = 1;
    it.it_interval.tv_usec = 0;
    it.it_value = it.it_interval;
    if (setitimer (ITIMER_REAL, &it, &oit) < 0) {
	fprintf (stderr, "timer not supported on this board\n");
	(void) signal (SIGALRM, oalrm);
	(void) signal (SIGINT, oint);
	return 0;
    }

    start = last = clock ();
    lastsecs = stuck = 0;

    do {
	now = clock ();
	if (now == last) {
	    if (++stuck > 100000) {
		fprintf (stderr, "clock not advancing, last=%u\n", last);
		errs++;
		break;
	    }
	    continue;
	}
	stuck = 0;

	diff = (now - last) / (CLOCKS_PER_SEC / 1000); /* in ms */
	if (diff > 50) {
	    fprintf (stderr, "clock error: last=%u now=%u\n", last, now);
	    errs++;
	    now = clock();
	}
	
	secs = (now - start) / CLOCKS_PER_SEC;
	if (ticker < secs || ticker > secs + 1) {
	    fprintf (stderr, "interval timer error: secs=%d ticker=%d\n",
		     secs, ticker);
	    errs++;
	    now = clock();
	}	    

	if (secs != lastsecs) {
	    fprintf (stderr, "\r% 3d ", limit - secs);
	    now = clock();
	}

	last = now;
	lastsecs = secs;
    } while (secs < limit);

    setitimer (ITIMER_REAL, &oit, 0);
    (void) signal (SIGALRM, oalrm);
    (void) signal (SIGINT, oint);

    fprintf (stderr, "%s\n", errs ? "failed" : "passed");
    return errs;
}


static jmp_buf  memjb;

static void
memhandler (int sig)
{
    longjmp (memjb, sig);
}


static const char usg_g[] = "g [-bhwd] ADDR";
static const char hlp_g[] = "get byte/halfword/word/dword at ADDR";

static int
com_g (int argc, char **argv, char **envp)
{
    void *	    addr;
    int             asz = 2;
    unsigned long long val;

    {
	char           *ap;

	while ((ap = *++argv) && ap[0] == '-') {
	    if (strcmp (ap, "-b") == 0)
		asz = 0;
	    else if (strcmp (ap, "-h") == 0)
		asz = 1;
	    else if (strcmp (ap, "-w") == 0)
		asz = 2;
	    else if (strcmp (ap, "-d") == 0)
		asz = 3;
	    else
		return Usage (usg_g);
	}
	if (!argv[0] || argv[1])
	    return Usage (usg_g);

	addr = (void *) strtoul (*argv, 0, 16);
    }

    printf ("%p: ", addr);
    fflush (stdout);
    {
	void            (*osigbus) (int) = signal (SIGBUS, memhandler);
	void            (*osigsegv) (int) = signal (SIGSEGV, memhandler);
	int		sig;
	int		err = 0;

	if (sig = setjmp (memjb)) {
	    psignal (sig, "");
	    signal (SIGBUS, osigbus);
	    signal (SIGSEGV, osigsegv);
	    return 0;
	}

	switch (asz) {
	case 0:
	    val = mips_get_byte (addr, &err);
	    break;
	case 1:
	    val = mips_get_half (addr, &err);
	    break;
	case 2:
	    val = mips_get_word (addr, &err);
	    break;
	case 3:
	    val = mips_get_dword (addr, &err);
	    break;
	}
	signal (SIGBUS, osigbus);
	signal (SIGSEGV, osigsegv);
	if (err)
	    printf ("memory/address error\n");
	else
	    printf ("%.*llx\n", 2 << asz, val);
    }
    return 0;
}


static const char usg_p[] = "p [-bhwd] ADDR VALUE";
static const char hlp_p[] = "put byte/halfword/word/dword VALUE at ADDR";

static int
com_p (int argc, char **argv, char **envp)
{
    int             asz = 2;
    void *          addr;
    int		    err;
    unsigned long long val;

    {
	char           *ap;

	while ((ap = *++argv) && ap[0] == '-') {
	    if (strcmp (ap, "-b") == 0)
		asz = 0;
	    else if (strcmp (ap, "-h") == 0)
		asz = 1;
	    else if (strcmp (ap, "-w") == 0)
		asz = 2;
	    else if (strcmp (ap, "-d") == 0)
		asz = 3;
	    else
		return Usage (usg_p);
	}
	if (!argv[0] || !argv[1] || argv[2])
	    return Usage (usg_p);

	addr = (void *) strtoul (argv[0], 0, 16);
	val = strtoull (argv[1], 0, 16);
    }

    printf ("%p: ", addr);
    fflush (stdout);
    {
	void            (*osigbus) (int) = signal (SIGBUS, memhandler);
	void            (*osigsegv) (int) = signal (SIGSEGV, memhandler);
	int 		sig;

	if (sig = setjmp (memjb)) {
	    psignal (sig, "");
	    signal (SIGBUS, osigbus);
	    signal (SIGSEGV, osigsegv);
	    return 0;
	}

	switch (asz) {
	case 0:
	    err = mips_put_byte (addr, val);
	    break;
	case 1:
	    err = mips_put_half (addr, val);
	    break;
	case 2:
	    err = mips_put_word (addr, val);
	    break;
	case 3:
	    err = mips_put_dword (addr, val);
	    break;
	}
	signal (SIGBUS, osigbus);
	signal (SIGSEGV, osigsegv);
    }
    if (err)
	printf ("memory/address error\n");
    else
	printf ("%.*llx\n", 2 << asz, val);
    return 0;
}


#if ! #cache(no) && defined(Index_Load_Tag_D) && !defined(__mips16)

#define stringify(x)	#x

#define mkcacheop(x,y) \
static inline void x (volatile void *va) \
{ \
    asm volatile ("%(nop;nop;cache " stringify(y) ",0(%0);nop;nop%)" : : "r" (va)); \
}

mkcacheop (index_load_tag_i, Index_Load_Tag_I)
mkcacheop (index_load_tag_d, Index_Load_Tag_D)
mkcacheop (index_store_tag_i, Index_Store_Tag_I)
mkcacheop (index_store_tag_d, Index_Store_Tag_D)

static const char usg_gt[] = "gt [-id] INDEX";
static const char hlp_gt[] = "get i/d cache tag at INDEX ";

static int
com_gt (int argc, char **argv, char **envp)
{
    unsigned int    offs;
    void *	    addr;
    int             ic = 0;
    unsigned long   val;

    {
	char           *ap;

	while ((ap = *++argv) && ap[0] == '-') {
	    if (strcmp (ap, "-i") == 0)
		ic = 1;
	    else if (strcmp (ap, "-d") == 0)
		ic = 0;
	    else
		return Usage (usg_gt);
	}
	if (!argv[0] || argv[1])
	    return Usage (usg_gt);

	offs = strtoul (*argv, 0, 16);
    }

    printf ("get %c-ctag[%05x] = ", ic ? 'i' : 'd', offs);
    fflush (stdout);
    addr = PA_TO_KVA0 (offs);
    if (ic)
	index_load_tag_i (addr);
    else
	index_load_tag_d (addr);
#if defined(mips_gettaghi)
    printf ("%08lx.%08lx\n", mips_gettaghi(), mips_gettaglo());
#else
    printf ("%08lx\n", mips_gettaglo());
#endif
    return 0;
}


static const char usg_pt[] = "pt [-id] INDEX VALUE";
static const char hlp_pt[] = "put VALUE in i/d cache tag INDEX";

static int
com_pt (int argc, char **argv, char **envp)
{
    unsigned int    offs;
    void *	    addr;
    int             ic = 0;
    unsigned long long  val;

    {
	char           *ap;

	while ((ap = *++argv) && ap[0] == '-') {
	    if (strcmp (ap, "-i") == 0)
		ic = 1;
	    else if (strcmp (ap, "-d") == 0)
		ic = 0;
	    else
		return Usage (usg_gt);
	}
	if (!argv[0] || !argv[1] || argv[2])
	    return Usage (usg_pt);

	offs = strtoul (argv[0], 0, 16);
	val = strtoull (argv[1], 0, 16);
    }

    printf ("put %c-tag[%04x] = ", ic ? 'i' : 'd', offs);
    fflush (stdout);

#if defined(mips_settaghi)
    mips_settaghi ((unsigned int)(val >> 32));
#endif
    mips_settaglo ((unsigned int)val);

    addr = PA_TO_KVA0 (offs);
    if (ic)
	index_store_tag_i (addr);
    else
	index_store_tag_d (addr);

    if (val >> 32)
	printf ("%08lx.%08lx\n", (unsigned int) (val >> 32),
		(unsigned int)val);
    else
	printf ("%08lx\n", (unsigned int)val);
    return 0;
}
#endif

static const char usg_setenv[] = "setenv [NAME [VALUE]]";
static const char hlp_setenv[] = "display/set environment variable";

static int
com_setenv (int argc, char **argv, char **envp)
{
    if (argc == 1) {
	getenv ("foo");		/* may be needed to init environ */
	for (envp = environ; *envp; envp++)
	    puts (*envp);
	return EXIT_SUCCESS;
    } else if (argc == 2) {
	char *s = getenv (argv[1]);
	if (s)
	    printf ("%s=%s\n", argv[1], s);
	else
	    printf ("%s undefined\n", argv[1]);
	return EXIT_SUCCESS;
    } else if (argc == 3) {
	/* set in permanent environment */
	if (setenv (argv[1], argv[2], 3) < 0)
	    perror ("setenv");
	return EXIT_SUCCESS;
    } else
	return Usage (usg_setenv);
}


static const char usg_unsetenv[] = "unsetenv NAME";
static const char hlp_unsetenv[] = "unset environment variable";

static int
com_unsetenv (int argc, char **argv, char **envp)
{
    if (argc == 2) {
	/* remove from permanent environment */
	deleteenv (argv[1]);
	return EXIT_SUCCESS;
    } else
	return Usage (usg_unsetenv);
}


static const char usg_type[] = "type FILE";
static const char hlp_type[] = "copy FILE to the terminal";

static int
com_type (int argc, char **argv, char **envp)
{
    FILE *fp;
    char buf[256];
    int n;
    
    if (argc != 2)
	return Usage (usg_type);

    fp = fopen (argv[1], "r");
    if (!fp) {
	perror (argv[1]);
	return EXIT_FAILURE;
    }

    while (fgets (buf, sizeof(buf), fp))
	fputs (buf, stdout);

    if (ferror (fp))
	perror (argv[1]);
    fclose (fp);
    return EXIT_SUCCESS;
}

static const char usg_dump[] = "dump FILE [OFFS [LEN]]";
static const char hlp_dump[] = "display FILE in hex";

static int
com_dump (int argc, char **argv, char **envp)
{
    char *fname;
    FILE *fp;
    unsigned long offs = 0;
    size_t len = INT_MAX;
    
    if (argc-- < 2)
	return Usage (usg_dump);

    fname = *++argv; 
    --argc;

    if (argc > 0) {
	offs = strtoul (*++argv, NULL, 0);
	--argc;
    }

    if (argc > 0) {
	len = strtoul (*++argv, NULL, 0);
	--argc;
    }
    
    if (argc > 0)
	return Usage (usg_dump);

    fp = fopen (fname, "rb");
    if (!fp) {
	perror (fname);
	return EXIT_FAILURE;
    }

    if (fseek (fp, offs, SEEK_SET)) {
	perror ("seek");
	return EXIT_FAILURE;
    }

    while (len > 0) {
	unsigned char buf[16];
	size_t nb, nr;
	unsigned int i;

	nb = len < 16 ? len : 16;
	nr = fread (buf, 1, nb, fp);
	if (nr == 0)
	    break;

	printf ("%08x:", offs);
	for (i = 0; i < 16; i++) {
	    if (i < nr)
		printf (" %02x", buf[i]);
	    else
		printf ("   ");
	}

	printf ("  *");
	for (i = 0; i < nr; i++)
	    putchar (isprint (buf[i]) ? buf[i] : '.');
	printf ("*\n");

	offs += nr;
	len -= nr;
    }

    if (ferror (fp))
	perror (fname);
    fclose (fp);

    return EXIT_SUCCESS;
}

static const char usg_echo[] = "echo [-n] ... FILE";
static const char hlp_echo[] = "echo arguments to FILE";

static int
com_echo (int argc, char **argv, char **envp)
{
    FILE *fp;
    size_t len;
    char *sp;
    int arg1, nolf, i;

    if (argc < 3)
	return Usage (usg_echo);

    nolf = 0;
    arg1 = 1;
    if (strcmp (argv[1], "-n") == 0) {
	nolf = 1;
	arg1 = 2;
    }
	
    fp = fopen (argv[argc - 1], "w");
    if (!fp) {
	perror (argv[argc -1]);
	return EXIT_FAILURE;
    }

    len = 0;
    for (i = arg1; i < argc - 1; i++)
	len += strlen (argv[i]) + 1;
    
    if (len > 0) {
	/* concatenate args into single string */
	sp = alloca (len); sp[0] = '\0';
	for (i = arg1; i < argc - 1; i++) {
	    if (i > arg1)
		strcat (sp, " ");
	    strcat (sp, argv[i]);
	}
	fputs (sp, fp);
    }

    if (!nolf)
	fputc ('\n', fp);
	
    if (ferror (fp))
	perror (argv[argc - 1]);
    fclose (fp);

    return EXIT_SUCCESS;
}


#include <dirent.h>
#include <limits.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>


static const char usg_ls[] = "ls [DIR]";
static const char hlp_ls[] = "list directory DIR";

static int
com_ls (int argc, char **argv, char **envp)
{
    char *dirname;
    struct dirent **namelist;
    struct passwd *pw = NULL;
    struct group *gr = NULL;
    uid_t last_uid = -1;
    gid_t last_gid = -1;
    time_t now;
    int n, i;
    
    if (argc == 2)
	dirname = argv[1];
    else if (argc == 1)
	dirname = ".";
    else
	return Usage (usg_ls);

    n = scandir (dirname, &namelist, NULL, alphasort);
    if (n < 0) {
	perror (dirname);
	return EXIT_FAILURE;
    }

    now = time (0);
    for (i = 0; i < n; i++) {
	struct dirent *dd = namelist[i];
	struct stat stb;
	struct tm tmb, *tm;
	char pathname[1024];
	char modes[11];
	char tstamp[16];
	char *p;

	strcpy (pathname, dirname);
	p = &pathname[strlen (pathname) - 1];
	if (*p != '/')
	    *++p = '/';
	strcpy (p + 1, dd->d_name);
	if (stat (pathname, &stb) != 0) {
	    perror (pathname);
	    free (dd);
	    continue;
	}
	
	switch (stb.st_mode & S_IFMT) {
	case S_IFCHR:	modes[0] = 'c'; break;
	case S_IFDIR:	modes[0] = 'd'; break;
	case S_IFBLK:	modes[0] = 'b'; break;
	case S_IFLNK:	modes[0] = 'l'; break;
	case S_IFSOCK:	modes[0] = 's'; break;
	case S_IFREG: 	modes[0] = '-'; break;
	default: 	modes[0] = '?'; break;
	}
	modes[1] = ((stb.st_mode & S_IRUSR) ? 'r' : '-');
	modes[2] = ((stb.st_mode & S_IWUSR) ? 'w' : '-');
	modes[3] = ((stb.st_mode & S_IXUSR) ? 'x' : '-');
	modes[4] = ((stb.st_mode & S_IRGRP) ? 'r' : '-');
	modes[5] = ((stb.st_mode & S_IWGRP) ? 'w' : '-');
	modes[6] = ((stb.st_mode & S_IXGRP) ? 'x' : '-');
	modes[7] = ((stb.st_mode & S_IROTH) ? 'r' : '-');
	modes[8] = ((stb.st_mode & S_IWOTH) ? 'w' : '-');
	modes[9] = ((stb.st_mode & S_IXOTH) ? 'x' : '-');
	modes[10] = '\0';

	tm = localtime_r (&stb.st_mtime, &tmb);
	if (stb.st_mtime < now - 6*30*24*60*60)
	    /* if more than 6 months, do "MMM DD YYYY" */
	    strftime (tstamp, sizeof(tstamp), "%b %e  %C%y", tm);
	else
	    /* recent file, include time "MMM DD HH:MM" */
	    strftime (tstamp, sizeof(tstamp), "%b %e %R", tm);

	/* get owner and group names */
	if (stb.st_uid != last_uid)
	    pw = getpwuid (last_uid = stb.st_uid);
	if (stb.st_gid != last_gid)
	    gr = getgrgid (last_gid = stb.st_gid);
	    
	printf ("%s %2d", modes, stb.st_nlink);
	if (pw)
	    printf (" %8s", pw->pw_name);
	else if (stb.st_uid == 0)
	    printf (" %8s", "root");
	else
	    printf (" %8d", stb.st_uid);
	if (gr)
	    printf (" %8s", gr->gr_name);
	else if (stb.st_gid == 0)
	    printf (" %8s", "staff");
	else
	    printf (" %8d", stb.st_gid);
	printf (" %8u", stb.st_size);
	printf (" %s %s\n", tstamp, dd->d_name);

	free (dd);
    }

    free (namelist);
    return EXIT_SUCCESS;
}


static const char usg_cd[] = "cd [DIR]";
static const char hlp_cd[] = "change directory to DIR";

static int
com_cd (int argc, char **argv, char **envp)
{
    char *wd;

    if (argc == 2) {
	if (chdir (argv[1]) != 0) {
	    perror (argv[1]);
	    return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
    }

    if (argc != 1)
	return Usage (usg_cd);

    /* print the current working directory */
    wd = getcwd (NULL, 0);
    if (!wd) {
	perror ("getcwd");
	return EXIT_FAILURE;
    }

    printf ("%s\n", wd);
    free (wd);
    return EXIT_SUCCESS;
}


#include <sys/elfmips.h>

/* Validate an ELF file header */
static const char *
ehdr_validate (const Elf32_Ehdr *ehdr)
{
  int i;

  if(ehdr->e_ident[0] != ELFMAG0 ||
     ehdr->e_ident[1] != ELFMAG1 ||
     ehdr->e_ident[2] != ELFMAG2 ||
     ehdr->e_ident[3] != ELFMAG3)
      return ("not an ELF image");
  
  if(ehdr->e_ident[4] != ELFCLASS32)
    return ("not 32-bit image");
    
#if #endian(big)
  if(ehdr->e_ident[5] != ELFDATA2MSB)
#else
  if(ehdr->e_ident[5] != ELFDATA2LSB)
#endif
    return ("wrong endianness");

  for (i = 7; i < EI_NIDENT; i++)
    if(ehdr->e_ident[i] != 0)
      return ("bad ELF ident");

  if (ehdr->e_type != ET_EXEC && ehdr->e_type != ET_DYN)
    return ("not executable or shared object");

  if (ehdr->e_phoff == 0 || ehdr->e_phnum == 0)
    return ("no ELF pheader");

  return NULL;
}


static inline int
overlap (as, ae, bs, be)
    unsigned long as, ae, bs, be;
{
    return (ae > bs && be > as);
}

/* Validate an ELF program segment header */
static const char *
phdr_validate (const Elf32_Phdr *ph, const Elf32_Addr base)
{
    extern char _ftext[], etext[], _fdata[], edata[], _fbss[], end[];
    extern void *_minbrk, *_maxbrk;
    extern paddr_t _mem_top;
    unsigned long s = KVA_TO_PA (ph->p_vaddr + base);
    unsigned long e = s + ph->p_memsz;

    if (overlap (s, e, 0x1fc00000, 0x20000000))
	return ("attempt to load into PROM space");
    if (overlap (s, e, KVA_TO_PA (_ftext), KVA_TO_PA (etext)))
	return ("attempt to overwrite monitor code");
    if (overlap (s, e, KVA_TO_PA (_fdata), KVA_TO_PA (end)))
	return ("attempt to overwrite monitor data");
    if (overlap (s, e, KVA_TO_PA (_minbrk), KVA_TO_PA (sbrk(0))))
	return ("attempt to overwrite monitor heap");
#if 0
    if (overlap (s, e, KVA_TO_PA (_maxbrk), _mem_top))
	return ("attempt to overwrite monitor stack");
#endif
    if (overlap (ph->p_vaddr + base, ph->p_vaddr + base + ph->p_memsz,
		 0x00000000, 0x80000000))
	return ("attempt to load into mapped space");
    
    return NULL;
}

static const char *
phtype (Elf32_Phdr *ph)
{
    static char buf[11];

    switch (ph->p_type) {
    case PT_NULL:
 	return "NULL    ";
    case PT_LOAD:
 	return "LOAD    ";
    case PT_DYNAMIC:
	return "DYNAMIC ";
    case PT_INTERP:
 	return "INTERP  ";
    case PT_NOTE:
	return "NOTE    ";
    case PT_SHLIB:
	return "SHLIB   ";
    case PT_PHDR:
	return "PHDR    ";
    case PT_MIPS_REGINFO:
	return "REGINFO ";
    default:		
	sprintf (buf, "%08x", ph->p_type);
	return buf;
   }
}


static const char *
phflags (Elf32_Phdr *ph)
{
    static char buf[4];
    buf[0] = (ph->p_flags & PF_R) ? 'r' : '-';
    buf[1] = (ph->p_flags & PF_W) ? 'w' : '-';
    buf[2] = (ph->p_flags & PF_X) ? 'x' : '-';
    buf[3] = '\0';
    return buf;
}


static const char usg_boot[] = "boot [-ln] FILE";
static const char hlp_boot[] = "load and run ELF executable";

static int __attribute__((nomips16,noinline))
call_dso_pic (int argc, char **argv, char **envp, Elf32_Addr entry)
{
	register int __a0 asm("$5") = argc;
	register void *__a1 asm("$4") = argv;
	register void *__a2 asm("$6") = envp;
	int status;

	/* jump to the loaded program's entrypoint */
	__asm__ __volatile__ (
	"	.set push		\n"
	"	.set nomips16		\n"
	"	addiu	$29, -24	\n"
	"	move	$25, %4		\n"
	"	sw	$31, 20($29)	\n"
	"	sw	$28, 16($29)	\n"
	"	jalr	$25		\n"
	"	lw	$28, 16($29)	\n"
	"	lw	$31, 20($29)	\n"
	"	move	%0, $2		\n"
	"	addiu	$29, 24		\n"
	"	.set pop		\n"
	: "=r" (status), "+r" (argc), "+r" (argv), "+r" (envp)
	: "r" (entry)
	: "$2", "$3", "$7",
	  "$8", "$9", "$10", "$11", "$12", "$13", "$14", "$15",
	  "$24", "$25",
	  "memory");

	return status;
}

static int
com_boot (int argc, char **argv, char **envp)
{
    int fd;
    Elf32_Ehdr ehdr;
    Elf32_Phdr *phtab, *ph;
    Elf32_Word align = 1;
    Elf32_Addr load_addr;
    Elf32_Addr base_addr = 0;
    Elf32_Addr vaddr;
    Elf32_Addr dyntmp;
    Elf32_Dyn *dynamic, *dn;
    Elf32_Word dynsize = 0;
    Elf32_Sym *dynsym, *d;
    Elf32_Rel *rel, *r;
    Elf32_Word relsize = 0;
    Elf32_Addr *got;
    Elf32_Word g;
    Elf32_Word gotsize = 0;
    Elf32_Word localgotno = 0, gotsym = 0;
    Elf32_Addr entry;
    int relocate = 0;
    int dynfl = 0, dynok = 0;
    int lnum = 0;
    int symnum = 0, symentnum = 0;
    int pltgotnum = 0, relnum = 0, relsznum = 0, relentnum = 0;
    int baddrnum = 0, locgotnum = 0, symtabnum = 0, gotsymnum = 0;
    int status;
    const char *err = NULL;
    int lflag = 0;
    int nflag = 0;
    char *ap;

    while ((ap = *++argv) && ap[0] == '-') {
	--argc;
	while (*++ap) {
	    switch (*ap) {
	    case 'l':
		++lflag;	/* just list program header */
		break;
	    case 'n':
		++nflag;	/* load but don't run */
		break;	
	    default:
		return Usage (usg_boot);
	    }
	}
    }
    
    if (!*argv)
	return Usage (usg_boot);

    fd = open (*argv, O_RDONLY | O_BINARY);
    if (fd < 0) {
	perror (*argv);
	return EXIT_FAILURE;
    }

    /* Read and validate ELF file header */
    if (read (fd, &ehdr, sizeof (ehdr)) != sizeof(ehdr)) {
	perror (*argv);
	err = "ELF file header read failed"; goto error;
    }

    if (err = ehdr_validate (&ehdr))
	goto error;

    /* Read ELF program headers */
    phtab = alloca (sizeof (Elf32_Phdr) * ehdr.e_phnum);
    if (lseek (fd, ehdr.e_phoff, SEEK_SET) == -1) {
	perror (*argv);
	err = "Seek to ELF program header failed"; goto error;
    }
    if (read (fd, phtab, sizeof (Elf32_Phdr) * ehdr.e_phnum)
	!= sizeof (Elf32_Phdr) * ehdr.e_phnum) {
	perror (*argv);
	err = "ELF program header read failed"; goto error;
    }
    
    if (lflag) {
	printf ("Segment List\n");
	printf ("Num Type       Vaddr    Paddr  MemSz FSize FOffs Align Flgs\n");
	for (ph = phtab; ph < &phtab[ehdr.e_phnum]; ph++) {
	    printf (" %1d  %8s %08x %08x %05x %05x %05x %05x %s\n",
		    ph-phtab, phtype(ph), ph->p_vaddr, ph->p_paddr,
		    ph->p_memsz, ph->p_filesz, ph->p_offset, ph->p_align,
		    phflags(ph));
	}
	close (fd);
	return EXIT_SUCCESS;
    }

    for (ph = phtab; ph < &phtab[ehdr.e_phnum]; ph++)
	switch (ph->p_type) {
	case PT_LOAD:
	    /* check for valid address */
	    err = phdr_validate (ph, 0);
	    if (err)
		relocate = 1;
	    if (ph->p_align > align)
		align = ph->p_align;
	    break;
	case PT_DYNAMIC:
	    if (dynfl) {
		err = "Duplicated dynamic segment";
		goto error;
	    }
	    dynsize = ph->p_filesz;
	    dyntmp = ph->p_vaddr;
	    dynfl = 1;
	    break;
	}
    load_addr = ((Elf32_Addr)sbrk(0) + align - 1) & ~(align - 1);

    /* Load each program segment in turn */
    printf ("Loading:"); fflush (stdout);
    for (ph = phtab; ph < &phtab[ehdr.e_phnum]; ph++) {
	switch (ph->p_type) {
	case PT_LOAD:
	    /* relocate */
	    if (relocate && lnum == 0 && ehdr.e_type == ET_DYN)
		base_addr = load_addr - ph->p_vaddr;
	    /* check for valid address */
	    err = phdr_validate (ph, base_addr);
	    vaddr = ph->p_vaddr + base_addr;
	    printf (" 0x%08lx-0x%08lx", vaddr, vaddr + ph->p_memsz);
	    fflush (stdout);
	    if (err)
		goto error;
	    if (ph->p_filesz > 0) {
		/* load code/data */
		if (lseek (fd, ph->p_offset, SEEK_SET) == -1) {
		    fprintf (stderr, "\n");
		    perror (*argv);
		    err = "Seek to segment data failed"; goto error;
		}
		if (read (fd, (void *)vaddr, ph->p_filesz) != ph->p_filesz) {
		    fprintf (stderr, "\n");
		    perror (*argv);
		    err = "Segment data read failed"; goto error;
		}
	    }
	    if (ph->p_memsz > ph->p_filesz) {
		/* clear uninitialised data */
		memset ((void *)(vaddr + ph->p_filesz), 0, 
			ph->p_memsz - ph->p_filesz);
	    }
	    /* Check if the dynamic segment is within this one */
	    if (dyntmp >= ph->p_vaddr &&
		dyntmp < ph->p_vaddr + ph->p_filesz &&
	        dyntmp + dynsize >= ph->p_vaddr &&
		dyntmp + dynsize < ph->p_vaddr + ph->p_filesz)
		dynok = 1;
	    lnum++;
	    break;
	case PT_NULL:
	case PT_DYNAMIC:
	case PT_MIPS_REGINFO:
	    break;
	default:
	    err = "Unsupported segment type"; goto error;
	    break;
	}
    }
    close (fd);

    /* Validate the dynamic segment and fetch GOT information */
    if (dynfl) {
	if (!dynok) {
	    err = "Invalid dynamic segment";
	    goto error;
	}
	dyntmp += base_addr;
	dynamic = (Elf32_Dyn *)dyntmp;
	printf (" DYN:0x%08lx-0x%08lx\n", dyntmp, dyntmp + dynsize);
	fflush (stdout);
	for (dn = dynamic; dn < dynamic + dynsize / sizeof (*dynamic) &&
	     dn->d_tag != DT_NULL; dn++) {
	    err = "Dynamic segment inconsistency";
	    switch (dn->d_tag) {
	    case DT_SYMTAB:
		if (symnum)
		    goto error;
		dynsym = (Elf32_Sym *)(dn->d_un.d_ptr + base_addr);
		symnum = 1;
		break;
	    case DT_SYMENT:
		if (symentnum)
		    goto error;
		if (dn->d_un.d_val != sizeof (Elf32_Sym)) {
		    err = "Dynamic SYM table inconsistency";
		    goto error;
		}
		symentnum = 1;
		break;
	    case DT_REL:
		if (relnum)
		    goto error;
		rel = (Elf32_Rel *)(dn->d_un.d_ptr + base_addr);
		relnum = 1;
		break;
	    case DT_RELSZ:
		if (relsznum)
		    goto error;
		relsize = dn->d_un.d_val;
		relsznum = 1;
		break;
	    case DT_RELENT:
		if (relentnum)
		    goto error;
		if (dn->d_un.d_val != sizeof (Elf32_Rel)) {
		    err = "Dynamic REL table inconsistency";
		    goto error;
		}
		relentnum = 1;
		break;
	    case DT_PLTGOT:
		if (pltgotnum)
		    goto error;
		got = (Elf32_Addr *)(dn->d_un.d_ptr + base_addr);
		pltgotnum = 1;
		break;
	    case DT_MIPS_BASE_ADDRESS:
		if (baddrnum)
		    goto error;
		if (!relocate)
		    load_addr = dn->d_un.d_ptr;
		else if (load_addr - dn->d_un.d_ptr != base_addr) {
		    err = "Base address inconsistency";
		    goto error;
		}
		baddrnum = 1;
		break;
	    case DT_MIPS_LOCAL_GOTNO:
		if (locgotnum)
		    goto error;
		localgotno = dn->d_un.d_val;
		gotsize += localgotno;
		locgotnum = 1;
		break;
	    case DT_MIPS_SYMTABNO:
		if (symtabnum)
		    goto error;
		gotsize += dn->d_un.d_val;
		symtabnum = 1;
		break;
	    case DT_MIPS_GOTSYM:
		if (gotsymnum)
		    goto error;
		gotsym = dn->d_un.d_val;
		gotsize -= gotsym;
		gotsymnum = 1;
		break;
	    case DT_NEEDED:
	    case DT_RELA:
	    case DT_RELASZ:
	    case DT_RELAENT:
		err = "Unsupported dynamic tag"; goto error;
		break;
	    default:
		break;
	    }
	}
    } else {
	printf ("\n");
	fflush (stdout);
    }
    if ((Elf32_Sword)gotsize < 0 || (gotsize > 0 && !pltgotnum)) {
	err = "Invalid GOT";
	goto error;
    }
    if ((Elf32_Sword)relsize < 0 || (relsize > 0 && !relnum)) {
	err = "Invalid dynamic REL table";
	goto error;
    }

    /* Relocate GOT */
    if (base_addr)
	for (g = 0; g < gotsize; g++) {
	    if (g >= localgotno) {
		d = dynsym + gotsym + g - localgotno;
		if (d->st_shndx == SHN_UNDEF) {
		    err = "Undefined symbol reference";
		    goto error;
		}
		if (ELF32_ST_TYPE(d->st_info) == STT_FUNC &&
		    d->st_value != got[g]) {
		    err = "Unsupported function stub";
		    goto error;
		}
	    }
	    got[g] += base_addr;
	}

    /* Calculate REL */
    for (r = rel; r < rel + relsize / sizeof (Elf32_Rel); r++) {
	Elf32_Word r_sym = ELF32_R_SYM(r->r_info);
	Elf32_Addr *addr = (Elf32_Addr *)(r->r_offset + base_addr);

	switch (ELF32_R_TYPE(r->r_info)) {
	case R_MIPS_REL32:
	    if (r_sym == STN_UNDEF)
		*addr += base_addr;
	    else {
		d = dynsym + r_sym;
		if (d->st_shndx == SHN_UNDEF) {
		    err = "Undefined symbol reference";
		    goto error;
		}
		if (r_sym < gotsym)
		    *addr += d->st_value + base_addr;
		else
		    *addr += got[localgotno + r_sym - gotsym];
	    }
	    break;
	case R_MIPS_NONE:
	    break;
	default:
	    err = "Unsupported reloc type"; goto error;
	    break;
	}
    }

    entry = ehdr.e_entry + base_addr;
    printf ("entrypoint 0x%08lx\n", entry);
    fflush (stdout);

#if ! #cache(no)
    /* make sure i & d caches are consistent */
    mips_flush_cache ();
#endif
	
    if (!nflag) {
	status = call_dso_pic(argc, argv, envp, entry);
	printf ("status: %i\n", status);
	fflush (stdout);
    }

    return EXIT_SUCCESS;

 error:
    close (fd);
    if (err)
	fprintf (stderr, "%s: %s\n", *argv, err);
    return EXIT_FAILURE;
}


#define CMD(x)	{#x, usg_##x, hlp_##x, com_##x}

static const struct comtab comtab[] =
{
    CMD(help),
    CMD(stat),
    CMD(cache),
    CMD(time),
    CMD(itimer),
    CMD(g),
    CMD(p),
#if ! #cache(no) && defined(Index_Load_Tag_D) && !defined(__mips16)
    CMD(gt),
    CMD(pt),
#endif
    CMD(setenv),
    CMD(unsetenv),
    CMD(cd),
    CMD(ls),
    CMD(echo),
    CMD(type),
    CMD(dump),
    CMD(boot),
    CMD(exit),
    {0},
};

#define IFS 	" \t\n"
#define MAXARG	32


static char    *
envrep (char *in)
{
    char           *envval;

    if (in && in[0] == '$' && (envval = getenv (in + 1))) {
	return envval;
    } else
	return in;
}


int
cmdloop (char *prompt)
{
    while (1) {
	char            combuf[1024];
	char           *xargv[MAXARG];
	int             xargc = 0;
	int             i;
	char           *cp;

	if (prompt) {
	    fputs (prompt, stdout);
	    fflush (stdout);
	}

#ifdef MINIMON_DEBUG_COMMAND
	strcpy (combuf, MINIMON_DEBUG_COMMAND "\n");
#else
	if (fgets (combuf, sizeof (combuf), stdin) == NULL)
	    return 0;
#endif

	/* break up to argc,argv */
	if ((cp = strtok (combuf, IFS)) == NULL)
	    continue;

	do {
	    xargv[xargc++] = envrep (cp);
	} while (cp = strtok (0, IFS));

	xargv[xargc] = 0;

	for (i = 0; comtab[i].name; i++)
	    if (strcmp (xargv[0], comtab[i].name) == 0) {
		break;
	    }
	if (comtab[i].name)
	    comtab[i].handler (xargc, xargv, environ);
	else
	    fprintf (stderr, "%s: not found\n", xargv[0]);

#ifdef MINIMON_DEBUG_COMMAND
	exit (0);
#endif
    }
}

static sigjmp_buf  intrjb;

static void
intrhandler (int sig)
{
    siglongjmp (intrjb, 1);
}


int
main ()
{
    struct termios tio;

    fprintf (stdout, "\nROM monitor built %s %s\n", __DATE__, __TIME__);

    /* little termios example */
    tcgetattr (STDOUT_FILENO, &tio);
    tio.c_cc[VERASE]	= 'h' & 037;	/* I prefer ^H for backspace */
    tio.c_cc[VKILL]	= 'x' & 037;	/* and ^X for line erase */
    if (tcsetattr (STDOUT_FILENO, TCSAFLUSH, &tio) < 0)
	perror ("tcsetattr");

    signal (SIGINT, intrhandler);
    if (sigsetjmp (intrjb, 1)) 
	fprintf (stderr, "Interrupt!\n");

    return cmdloop ("minimon> ");
}
