/*
 * Copyright (c) 2000-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * pcitest.c: PCI bus demonstration
 *
 * Displays the contents of a set of PCI buses
 */


#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if #pci(yes)

#include <pci/pcivar.h>
#include <pci/pcireg.h>

/* PCI Boot ROM header */
struct pciromhdr {
    unsigned char	romsig[2];
#define PCIHDR_SIG_0	0x55
#define PCIHDR_SIG_1	0xaa
    unsigned char	x86size;
    unsigned char	x86entry[21];
    unsigned short	dataoffs;
};

/* PCI Boot ROM data */
struct pciromdata {
    unsigned char	pcisig[4];
#define PCIDATA_SIG_0	'P'
#define PCIDATA_SIG_1	'C'
#define PCIDATA_SIG_2	'I'
#define PCIDATA_SIG_3	'R'
    unsigned short	vendor;
    unsigned short	device;
    unsigned short	vpdoffs;
    unsigned short	length;
    unsigned char	rev;
    unsigned int	class:24;
    unsigned short	imagelen;
    unsigned short	coderev;
    unsigned char	codetype;
#define PCIDATA_CODETYPE_X86	0
#define PCIDATA_CODETYPE_OFW	1
    unsigned char	indicator;
#define PCIDATA_INDICATOR_LAST	0x80
    unsigned char	reserved;
};

#if #endian(little)
void
pcidumprom (pcitag_t tag, vm_offset_t va)
{
    struct pciromhdr hdr;
    struct pciromdata dat;
    int image = 0;

    do {
	printf ("\n ROM image %d:\n", image++);

	memcpy (&hdr, (void *)va, sizeof(hdr));
	printf ("  ROM sig:    %02x %02x\n", 
		hdr.romsig[0], hdr.romsig[1]);
	if (hdr.romsig[0] != PCIHDR_SIG_0 ||
	    hdr.romsig[1] != PCIHDR_SIG_1) {
	    printf ("  Bad ROM header signature!\n");
	    return;
	}
	printf ("  PCI offs:   %04x\n\n", hdr.dataoffs);

	memcpy (&dat, (void *)va + hdr.dataoffs, sizeof(dat));
	printf ("  PCI sig:    %02x %02x %02x %02x\n",
		dat.pcisig[0], dat.pcisig[1],
		dat.pcisig[2], dat.pcisig[3]);
	
	if (dat.pcisig[0] != PCIDATA_SIG_0 ||
	    dat.pcisig[1] != PCIDATA_SIG_1 ||
	    dat.pcisig[2] != PCIDATA_SIG_2 ||
	    dat.pcisig[3] != PCIDATA_SIG_3) {
	    printf ("  Bad PCI data signature!\n");
	    return;
	}
	
	printf ("  Vendor:     %04x\n", dat.vendor);
	printf ("  Device:     %04x\n", dat.device);
	printf ("  VPDoffs:    %04x\n", dat.vpdoffs);
	printf ("  Length:     %04x\n", dat.length);
	printf ("  Class:      %06x\n", dat.class);
	printf ("  Rev:        %02x\n", dat.rev);
	printf ("  Image Len:  %04x\n", dat.imagelen);
	printf ("  Last Image: %s\n",   
		(dat.indicator & PCIDATA_INDICATOR_LAST) ? "yes" : "no");
	printf ("  Code Rev:   %04x\n", dat.coderev);
	printf ("  Code Type:  %s\n", 
		(dat.codetype == PCIDATA_CODETYPE_X86) ? "x86"
		: (dat.codetype == PCIDATA_CODETYPE_OFW) ? "ofw"
		: "???");

	if (dat.codetype == PCIDATA_CODETYPE_X86) {
	    unsigned int len = hdr.x86size * 512;
	    unsigned char *bp = (unsigned char *)va;
	    unsigned char csum = 0;
	    printf ("  x86 size:   %02x\n", hdr.x86size);
	    while (len-- != 0)
		csum += *bp++;
	    printf ("  x86 chksum: %02x (%s)\n", csum,
		    (csum == 0) ? "OK" : "FAIL");
	}

	va = va + dat.imagelen * 512;

    } while (!(dat.indicator & PCIDATA_INDICATOR_LAST));
}
#endif


void
pciprint (pcitag_t tag)
{
    pcireg_t class, id, stat, bparam, misc;
    pcireg_t base;
    vm_offset_t va, pa;
    int bus, device, function;
    unsigned int x;
    char devinfo[256];
    int supported;
    int reg;
    int line;

    _pci_break_tag (tag, &bus, &device, &function);
    printf ("\nPCI");
    if (_pci_nbus > 1)
	printf(" bus %d", bus);
    printf (" slot %d", device);
    if (function != 0)
	printf (" function %d", function & 0x07);
    fflush (stdout);

    id = _pci_conf_read32 (tag, PCI_ID_REG);
    class = _pci_conf_read32 (tag, PCI_CLASS_REG);
    _pci_devinfo (id, class, devinfo, &supported);
    printf (": %s\n", devinfo);

    stat = _pci_conf_read32 (tag, PCI_COMMAND_STATUS_REG);

    printf (" 66MHz support:      %s\n", 
	    (stat & PCI_STATUS_66MHZ_OKAY) ? "yes" : "no");

    printf (" UDF support:        %s\n", 
	    (stat & PCI_STATUS_UDF_SUPPORTED) ? "yes" : "no");

    x = (stat & PCI_STATUS_DEVSEL_MASK);
    printf (" Device select:      %s\n", 
	    (x == PCI_STATUS_DEVSEL_FAST) ? "fast" 
	    : (x == PCI_STATUS_DEVSEL_MEDIUM) ? "medium" 
	    : (x == PCI_STATUS_DEVSEL_SLOW) ? "slow"
	    : "UNKNOWN");

    printf (" Fast back-to-back:  %s/%s\n", 
	    (stat & PCI_STATUS_BACKTOBACK_OKAY) ? "yes" : "no",
	    (stat & PCI_COMMAND_BACKTOBACK_ENABLE) ? "on" : "off");

    printf (" IO space:           %s\n", 
	    (stat & PCI_COMMAND_IO_ENABLE) ? "on" : "off");

    printf (" Memory space:       %s\n", 
	    (stat & PCI_COMMAND_MEM_ENABLE) ? "on" : "off");

    printf (" Bus master:         %s\n", 
	    (stat & PCI_COMMAND_MASTER_ENABLE) ? "on" : "off");

    bparam = _pci_conf_read32 (tag, PCI_BPARAM_REG);

    printf (" Min Grant:          %d\n", 
	    PCI_BPARAM_MIN_GNT (bparam));

    printf (" Max Latency:        %d\n", 
	    PCI_BPARAM_MAX_LAT (bparam));

    misc = _pci_conf_read32 (tag, PCI_MISC_REG);

    printf (" Latency Timer:      %d\n", 
	    PCI_MISC_LTIM (misc));

    printf (" CacheLine Size:     %d\n", 
	    PCI_MISC_CLSZ (misc));

    /* Display the local interrupt number */
    line = _pci_map_int (tag);
    if (line < 0)
	printf (" Interrupt:          unknown\n");
    else if (line == 0)
	printf (" Interrupt:          none\n");
    else
	printf (" Interrupt:          %d\n", line);

    /* Scan the mapping registers */
    for (reg = PCI_MAP_REG_START; reg < PCI_MAP_REG_END; reg += 4) {

	base = _pci_conf_read32 (tag, reg);
	if (base == 0 || base == ~0)
	    continue;

	if (base & PCI_MAP_IO) {
	    printf (" IoAddress%d:         0x%08x/",
		    (reg - PCI_MAP_REG_START)/4,
		    base & PCI_MAP_IO_ADDRESS_MASK);
	    /* map i/o region to CPU virtual address */
	    if (_pci_map_io (tag, reg, &va, &pa) == 0)
		printf ("%p\n", va);
	    else
		printf ("unmappable\n");
	}
	else {
	    printf (" MemAddress%d:        0x%08x/",
		    (reg - PCI_MAP_REG_START)/4, 
		    base & PCI_MAP_MEMORY_ADDRESS_MASK);
	    /* map mem region to CPU virtual address */
	    if (_pci_map_mem (tag, reg, &va, &pa) == 0)
		printf ("%p\n", va);
	    else
		printf ("unmappable\n");
	}
    }

    /* Check for a valid Boot ROM mapping register */
    base = _pci_conf_read32 (tag, PCI_MAP_REG_ROM);
    if (base != ~0 && (base & PCI_MAP_ROM)) {
	printf (" ROMAddress:         0x%08x/", 
		base & PCI_MAP_ROM_ADDRESS_MASK);

	/* map ROM region to CPU virtual address */
	if (_pci_map_mem (tag, PCI_MAP_REG_ROM, &va, &pa) == 0)
	    printf ("%p\n", va);
	else
	    printf ("unmappable\n");

	/* dump the ROM contents */
	if (stat & PCI_COMMAND_MEM_ENABLE) {
#if #endian(little)
	    pcidumprom (tag, va);
#else
	    printf (" cannot dump ROM in big-endian mode\n");
#endif
	}
	else
	    printf (" access to ROM not enabled\n");
    }
}


main (int argc, char **argv)
{
    struct pci_match match;
    pcitag_t tag;
    unsigned int n;

    printf ("\nPCI BUS SCAN:\n");
    
    /* use _pci_find() to scan; match any device */
    match.idmask 	= 0x0;	/* id = don't care */
    match.id		= 0x0;
    match.classmask 	= 0x0;	/* class = don't care */
    match.class		= 0x0;	
    for (n = 0; (tag = _pci_find (&match, n)) != ~(pcitag_t)0; n++)
	pciprint (tag);

    exit (0);
}

#else /* !#pci(yes) */

main (int argc, char **argv)
{
    fprintf (stderr, "No PCI bus support on this target\n");
    return (1);
}

#endif /* #pci(yes) */
