/*
 * Copyright (c) 2003-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ex2/tlbxcpt.c: MIPS SDE example #2 (exception handling)
 */


/*
 * Example #2
 * Demonstrates use of xception handler;
 * and TLB functionality, uncached access only.
 * Tests basic operation of kuseg and kseg2 and TLB reload.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <mips/xcpt.h>
#include <mips/cpu.h>

/* NOT POSIX !! */
extern void    *sbrk (int);

static int	ntlb;
static int	nwired;

#define NITER	500

#if ! #mmu(no)

#if #mmu(type2)
#define TLBPGSIZE (VMPGSIZE * 2)	/* each tlb entry maps two pages */
#elif #mmu(type1) || #mmu(cw10)
#define TLBPGSIZE VMPGSIZE		/* each tlb entry maps one page */
#else
#error unknown mmu architecture
#endif

#define TLBPGMASK (TLBPGSIZE - 1)

static int      ntlbmiss;
static unsigned char *tpg;

/*
 * This function is called on a TLBmiss exception
 */
static int
mytlbhandler (int code, struct xcptcontext *xcp)
{
    unsigned        pfn = paToPn (KVA_TO_PA (tpg));
    reg_t           va = xcp->vaddr;
    static reg_t    lastva = ~0;

#ifdef VERBOSE
    fprintf (stderr, "\nTLB reload va = %08x\n", REG_TO_VA (va));
#endif

    if (va == lastva) {
	fprintf (stderr, "\nrepeated TLB miss va = %08x\n", REG_TO_VA (va));
	return SIGBUS;
    }
    lastva = va;

#if #mmu(type1)
    mips_tlbwr (mktlbhi (vaToVpn (va), 0),
		mktlblo (pfn, TLB_V | TLB_G | TLB_N));
#elif #mmu(type2)
    mips_tlbwr2 (mktlbhi (vaToVpn (va), 0),
		 mktlblo (pfn, TLB_V | TLB_G | TLB_UNCACHED),
		 mktlblo (pfn + 1, TLB_V | TLB_G | TLB_UNCACHED),
		 0);
#elif #mmu(cw10)
    cw401x_tlbwr (mktlbhi (vaToVpn (va), 0),
		  mktlblo (pfn, TLB_V | TLB_G | TLB_UNCACHED), 
		  0);
#else
#error unhandled mmu architecture
#endif
    ntlbmiss++;
    return 0;
}


static void
checkbrand0 (void *base, unsigned range, int niter)
{
    int             iter;

    for (iter = 0; iter < niter; iter++) {
	unsigned        exp, got;
	unsigned        r = rand () * rand ();

	r %= range;

	if ((iter % 64) == 0)
	    fprintf (stderr, "Pass %8d\r", iter);

        exp = tpg[r & TLBPGMASK];
	got = ((unsigned char *)base)[r];
        if (got != exp) {
	    fprintf (stderr, "Pass %d error expected %#x at %08x, but got %#x\n",
		     iter, exp, base + r, got);
	    abort ();
	}
    }
}
#endif /* !#mmu(no) */


static void
clear_tlb (void)
{
#if !#mmu(no)
    unsigned long pfn = (unsigned long)KSEG1_BASE >> VMPGSHIFT;
    unsigned int sr = _mips_intdisable ();
    int i;

#if #mmu(type1)
    for (i = nwired; i < ntlb; i++)
	mips_tlbwi (mktlbhi (pfn + (ntlb - 1 - i), 0), 0, i);
#elif #mmu(type2)
    for (i = nwired; i < ntlb; i++)
	mips_tlbwi2 (mktlbhi (pfn + (ntlb - 1 - i) * 2, 0), 0, 0, 0x0, i);
#elif #mmu(cw10)
    for (i = nwired; i < ntlb; i++)
	cw401x_tlbwi (mktlbhi (pfn + (ntlb - 1 - i), 0), 0, 0x0, i);
#else
#error unhandled mmu architecture
#endif
    _mips_intrestore (sr);
#endif
}


static void
done (int status)
{
    clear_tlb ();

    exit (status);
}


int
main (int argc, char **argv, char **envp)
{
    unsigned        pfn;
    unsigned 	    sr;
    int             niter = NITER;
    int		    i;

    ntlb = mips_tlb_size ();
    if (ntlb <= 0) {
	fprintf (stderr, "No TLB on this CPU, sorry!\n");
	exit (EXIT_FAILURE);
    }

#if #mmu(no)

    fprintf (stderr, "TLB size %d, but no TLB support!\n", ntlb);
    exit (EXIT_FAILURE);

#else

    /* use unwired tlb entries only */
    nwired = mips_getwired ();

    printf ("TLB size: %d entries, %d wired\n", ntlb, nwired);
    
    /* align current break to page size */
    {
	unsigned va = (unsigned)sbrk (0);
	if (va & TLBPGMASK)
	    (void)sbrk (TLBPGSIZE - (va & TLBPGMASK));
    }

    /* allocate a physical page and fill with 
     * arbitrary values (the code for main!)
     */
    tpg = sbrk (TLBPGSIZE);
    assert (((unsigned)tpg & TLBPGMASK) == 0);

    /* fill with a distinctive pattern */
    for (i = 0; i < TLBPGSIZE; i++)
	tpg[i] = i;

#if ! #cache(no)
    /* clean writeback cache */
    if (IS_KVA0 (tpg))
	mips_clean_dcache ((vaddr_t)tpg, TLBPGSIZE);
#endif

    /* reset TLB */
    clear_tlb ();

    /* tlb[0] := va=0,tlbid=0,global,readonly,uncached */
    pfn = KVA_TO_PA (tpg) >> VMPGSHIFT;		/* phys page number */

    /* set first unwired entry */
    sr = _mips_intdisable ();			/* disable interrupts */
#if #mmu(type1)
    mips_tlbwi (mktlbhi (0, 0),
		mktlblo (pfn, TLB_V | TLB_G | TLB_N),
		nwired);
#elif #mmu(type2)
    mips_tlbwi2 (mktlbhi (0, 0),
		mktlblo (pfn, TLB_V | TLB_G | TLB_UNCACHED),
		mktlblo (pfn + 1, TLB_V | TLB_G | TLB_UNCACHED),
		0x0, nwired);
#elif #mmu(cw10)
    cw401x_tlbwi (mktlbhi (0, 0),
		  mktlblo (pfn, TLB_V | TLB_G | TLB_UNCACHED),
		  0x0, nwired);
#else
#error unhandled mmu architecture
#endif
    _mips_intrestore (sr);			/* restore interrupts */

    /* check contents match */
    if (memcmp ((void *)0, tpg, TLBPGSIZE) != 0) {
	fprintf (stderr, "KUSEG failed virtual to physical equivalence\n");
	done (EXIT_FAILURE);
    }

    /* reset TLB again */
    clear_tlb ();

    /* trap read TLB faults with a handler which reloads to zpg */
    (void)xcption (XCPTTLBL, mytlbhandler);

    /* run over half KUSEG randomly accesing memory over the whole range */
    ntlbmiss = 0;
    {
	void           *base = KUSEG_BASE;
	unsigned        range = KUSEG_SIZE / 2;

	printf ("KUSEG tlb reload test: base %08x range %08x\n", base, range);
	checkbrand0 (base, range, niter);
	printf ("Done %8d iterations: %d tlbmisses\n", niter, ntlbmiss);
    }

    /* run over KUSEG but restrict to a small range 
     * an R3000-type TLB contains 64 entries each mapping 4K 
     * an R4000-type TLB contains 16-48 entries each mapping 8K
     * of these the first NWIRED are not written by tlbwr 
     */
    ntlbmiss = 0;
    {
	void           *base = KUSEG_BASE;
	unsigned        range = (ntlb - nwired) * TLBPGSIZE / 2; /* half range */

	printf ("KUSEG tlb reload test: base %08x range %08x\n", base, range);
	checkbrand0 (base, range, niter);
	printf ("Done %8d iterations: %d tlbmisses\n", niter, ntlbmiss);
    }

    ntlbmiss = 0;
    {
	void           *base = KSEG2_BASE;
	unsigned        range = KSEG2_SIZE;

	printf ("KSEG2 tlb reload test: base %08x range %08x\n", base, range);
	checkbrand0 (base, range, niter);
	printf ("Done %8d iterations: %d tlbmisses\n", niter, ntlbmiss);
    }

    done (EXIT_SUCCESS);
#endif /* !#mmu(no) */
}
