/*
 * Copyright (c) 2000-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * zload.c: decompressing bootstrap loader, and /dev/panel demo
 */

/*  
 * This program is a bootstrap which will decommpress a gzip
 * compressed ELF executable file from PROM and load the code and data
 * sections into RAM. Finally it jumps at the program's entrypoint.
 * 
 * Download zloadram and exec.gz.s3, then execute zloadram
 * 
 * To run on a simulator with host I/O (eg. MIPSsim), run zloadram with the
 * exec.gz as the first argument, ie.  
 * 	zloadram exec.gz
 * 	
 * Note that exec must be built to run in RAM, but without
 * expecting PMON or any other PROM monitor to be present. In other
 * words you may need to build a special kit with MONITOR=bare
 * defined.  See .../sde/kit/P4300LR/sbd.mk */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mips/cpu.h>
#include <mips/bt.h>

#include "zload.h"
#include "elf.h"

uch	*inbuf;
uch	*inptr;
uch	*inlim;
ulg	insize;

uch	slide[WSIZE];

ulg	orig_crc = 0;		/* original crc */
ulg	orig_len = 0;		/* original uncompressed length */
ulg 	bytes_out = 0;		/* number of bytes written */

ulg updcrc (uch *s, unsigned n);

#ifndef UNIXDEBUG
/* front panel display */
#include <sys/panelio.h>
struct panelinfo panelinfo;
FILE	*panel;
#endif

void fatal (char *m)
{
    fprintf(stderr, "\n%s\n", m);
#ifndef UNIXDEBUG
    if (panel) {
	int percent = -1;
	ioctl (fileno (panel), PANELIOPROGRESS, &percent);
	if (panelinfo.rows > 1)
	    fprintf (panel, "ERROR\n%s", m);
	else if (panelinfo.cols > 8)
	    fprintf (panel, "ERR %s", m);
	else
	    fprintf (panel, "%s", m);
	fflush (panel);
    }
#endif
    exit (2);
}


void *xmalloc (unsigned size)
{
    void * cp = malloc (size);
    if (cp == NULL) 
	fatal ("!mem");
    return cp;
}



Elf32_Ehdr	ehdr;
Elf32_Phdr	*phdr;
unsigned	phdrsize;
unsigned int	*loaded;
unsigned	bytes_loaded = 0;
unsigned	bytes_to_load = 0;

void ehdr_validate (void)
{
  int i;

  /* just filled the header */
  if(ehdr.e_ident[0] != ELFMAG0 ||
     ehdr.e_ident[1] != ELFMAG1 ||
     ehdr.e_ident[2] != ELFMAG2 ||
     ehdr.e_ident[3] != ELFMAG3)
    fatal ("!ELF");
  
  if(ehdr.e_ident[4] != ELFCLASS32)
    fatal ("!32-bit");
    
#ifdef MIPSEB
  if(ehdr.e_ident[5] != ELFDATA2MSB)
#else
  if(ehdr.e_ident[5] != ELFDATA2LSB)
#endif
    fatal ("endianness");

  for (i = 7; i < EI_NIDENT; i++)
    if(ehdr.e_ident[i] != 0)
      fatal ("bad id");

  if (ehdr.e_type != ET_EXEC)
    fatal ("!executable");

  if (ehdr.e_phoff == 0 || ehdr.e_phnum == 0)
    fatal ("!pheader");
}


dopush (void *obj, unsigned foffs, unsigned fsize, 
	int loadseg, 
#ifdef UNIXDEBUG
	char *objname,
#endif
	uch **bp, unsigned *bsize)
{
  unsigned fend = foffs + fsize;
  unsigned offs, nb;

  if (bytes_out < foffs || bytes_out >= fend)
      return 0;

  offs = bytes_out - foffs;
  nb = fend - bytes_out;
  if (nb > *bsize)
      nb = *bsize;

#ifdef UNIXDEBUG    
  printf ("push offs %6x-%6x to ", bytes_out, bytes_out + nb);
  if (loadseg < 0) {
      printf ("%s+0x%x\n", objname, offs);
  }
  else {
      printf ("seg[%d] 0x%08x+%06x\n", loadseg, obj, offs);
  }
#else
#ifndef QUIETROM
  if (bytes_out == foffs && loadseg >= 0)
      fprintf (stderr, " %08x/%06x", obj, fsize); /* start of a segment */
#endif /* QUIETROM */
  if (loadseg >= 0) {
      bytes_loaded += nb;
      if (panel && bytes_to_load != 0) {
	  int percent = bytes_loaded * 100UL / bytes_to_load;
	  ioctl (fileno (panel), PANELIOPROGRESS, &percent);
      }
  }
#endif /* !UNIXDEBUG */

  memcpy ((char *)obj + offs, *bp, nb);
  *bp += nb;
  *bsize -= nb;
  bytes_out += nb;
  return nb;
}


#ifdef UNIXDEBUG
#define push(obj, foffs, fsize, loadseg) \
dopush ((void *)obj, foffs, fsize, loadseg, #obj, &bp, &bsize)
#else
#define push(obj, foffs, fsize, loadseg) \
dopush ((void *)obj, foffs, fsize, loadseg, &bp, &bsize)
#endif

unsigned
write_data (uch *bp, unsigned bsize)
{
    unsigned n = 0;
    int i;

    if (bsize == 0)
	return 0;

    /* fill up ELF file header */
    n += push (&ehdr, 0, sizeof(ehdr), -1);
    
    if (bytes_out >= sizeof(ehdr)) {
	
	if (!phdr) {
	    /* got file header, validate it */
	    ehdr_validate ();
	    
	    /* now we can allocate the program header */
	    phdrsize = ehdr.e_phnum * sizeof(Elf32_Phdr);
	    phdr = xmalloc (phdrsize);
	    loaded = xmalloc (ehdr.e_phnum * sizeof (unsigned int));
	    memset (loaded, 0, ehdr.e_phnum * sizeof (unsigned int));
	}
	
	/* fill up program header */
	n += push (phdr, ehdr.e_phoff, phdrsize, -1);

	if (bytes_out == ehdr.e_phoff + phdrsize) {
	    /* validate program header */
	    bytes_to_load = 0;
	    for (i = 0; i < ehdr.e_phnum; i++)
		if (phdr[i].p_type == PT_LOAD)
		    bytes_to_load += phdr[i].p_filesz;
	}

	if (bytes_out >= ehdr.e_phoff + phdrsize) {
	    /* scan all program segments */
	    for (i = 0; i < ehdr.e_phnum; i++) {
		if (phdr[i].p_type == PT_LOAD) {
		    unsigned int nb;
		    /* fill one program segment */
		    nb = push (phdr[i].p_vaddr, phdr[i].p_offset, 
			       phdr[i].p_filesz, i);
		    loaded[i] += nb;
		    n += nb;
		}
	    }
	}
    }

    if (n == 0) {
	/* nobody wanted anything: junk one byte */
	bytes_out++;
	n++;
    }

    return n;
}


void check_program_loaded (void)
{
    int nph = ehdr.e_phnum;
    Elf32_Phdr *ph;
    
    if (!phdr) 
	fatal("!pheader");
    
    for (ph = phdr; ph < &phdr[nph]; ph++) {
	if (ph->p_type != PT_LOAD)
	    continue;
	if (ph->p_memsz > ph->p_filesz) {
	    char *bss_addr = (char *)ph->p_vaddr + ph->p_filesz;
	    unsigned bss_size = ph->p_memsz - ph->p_filesz;
#ifndef QUIETROM
	    fprintf (stderr, " %08x/%06x", bss_addr, bss_size);
#endif
#ifndef UNIXDEBUG
	    memset (bss_addr, 0, bss_size);
#endif
	}
    }
}



void flush_output (unsigned size)
{
    uch *buf = slide;
    unsigned n;

    if (size == 0) return;
    updcrc(slide, size);
    while ((n = write_data(buf, size)) != size) {
	size -= n;
	buf += n;
    }
}


int fill_buf (void)
{
    fatal ("exhausted input");
    return EOF;
}



int main (int argc, char **argv, char **envp)
{
    char magic[2]; /* magic header */
    int method;
    uch flags;
    long time_stamp;
    uch crc_buf[8];
    int n;
    int ret;
#ifdef UNIXDEBUG
    struct stat stb;
    int fd;
#elif 0
    struct btpackage *bt;
#endif

#ifndef UNIXDEBUG
    panel = fopen ("/dev/panel", "w");
    if (panel) {
	int percent = 0;
	ioctl (fileno (panel), PANELIOPROGRESS, &percent);
	ioctl (fileno (panel), PANELIOINFO, &panelinfo);
	fprintf (panel, "INITIALISING");
	fflush (panel);
    }
#endif

#ifdef UNIXDEBUG
    if (argc != 2)
	fatal ("Usage: zload name");
    fd = open (argv[1], O_RDONLY);
    if (fd < 0) {
	perror (argv[1]);
	exit (2);
    }
    fstat (fd, &stb);
    insize = stb.st_size;
    inbuf = malloc (insize);
    if (read (fd, inbuf, insize) != insize) {
	perror (argv[1]);
	exit (3);
    }
    close (fd);
#elif 0
    /* in the fullness of time sde-conv will prepare this for us */
    bt = &BT_PACKAGEBASE[BT_NPKG - 2];
    if (bt->bt_magic != BT_ZMAGIC)
	fatal ("missing package");
    inbuf = KVA_TO_KVA0 (bt->bt_start);
    insize = bt->bt_end - bt->bt_start;
#else

#ifndef LOAD_ADDRESS
    /*rom base + 64k */
#define LOAD_ADDRESS 0x9fc10000
#endif 

    inbuf = (void *)LOAD_ADDRESS;
    /* assume 2MB of PROM */
    insize = 0x200000 - 0x10000;
#endif

    inptr = inbuf;
    inlim = inbuf + insize;

    magic[0] = (char)get_byte();
    magic[1] = (char)get_byte();
    method = (int)get_byte();

    if (magic[0] == 0xff && magic[1] == 0xff && method == 0xff)
	fatal ("empty eprom?");

    if ((memcmp(magic, GZIP_MAGIC, 2) && memcmp(magic, OLD_GZIP_MAGIC, 2))
	|| method != DEFLATED) {
	fatal ("compression type");
    }

    flags  = (uch)get_byte();

    if ((flags & ASCII_FLAG) != 0) {
	fatal ("ASCII data");
    }
    if ((flags & ENCRYPTED) != 0) {
	fatal ("encrypted data");
    }
    if ((flags & CONTINUATION) != 0) {
	fatal ("continuation data");
    }
    if ((flags & RESERVED) != 0) {
	fatal ("badflags");
    }

    /* Discard time_stamp */
    time_stamp  = (ulg)get_byte();
    time_stamp |= ((ulg)get_byte()) << 8;
    time_stamp |= ((ulg)get_byte()) << 16;
    time_stamp |= ((ulg)get_byte()) << 24;
  
    (void)get_byte();  /* Ignore extra flags for the moment */
    (void)get_byte();  /* Ignore OS type for the moment */

    if ((flags & EXTRA_FIELD) != 0) {
	unsigned len = (unsigned)get_byte();
	len |= ((unsigned)get_byte())<<8;
	while (len--) (void)get_byte();
    }

    /* Discard original file name */
    if ((flags & ORIG_NAME) != 0) {
	while (get_byte() != 0) /* null */ ;
    }

    /* Discard file comment if any */
    if ((flags & COMMENT) != 0) {
	while (get_byte() != 0) /* null */ ;
    }

#ifndef QUIETROM
    fprintf (stderr, "Decompress ");
#endif

    switch (inflate()) {
    case 0:
	break;
    case 3:
	fatal ("!mem");
    default:
	fatal ("gzip error");
    }

    for (n = 0; n < 8; n++) {
	crc_buf[n] = (uch)get_byte();
    }
    orig_crc = LG(crc_buf);
    orig_len = LG(crc_buf+4);

    /* Validate decompression */
    if (orig_crc != updcrc(inbuf, 0)) {
	fatal("crc error");
    }
    if (orig_len != bytes_out) {
#ifndef QUIETROM
	fprintf (stderr, "\noriginal=%d uncompressed=%d\n",
		 orig_len, bytes_out);
#endif
	fatal("length error");
    }
  
    check_program_loaded ();

#if ! #cache(no)
    mips_flush_cache ();
#ifndef QUIETROM
    fprintf (stderr, "Icache size %u\n", mips_icache_size);
    fprintf (stderr, "Dcache size %u\n", mips_dcache_size);
#endif
#endif

#ifndef QUIETROM
    fprintf (stderr, " \nJumping to Entrypoint  %x\n", ehdr.e_entry);
#endif
    ret = (*(int (*)()) ehdr.e_entry) (argc, argv, envp);
    printf ("\n%s, return = %d\n",ret==42 ? "Success" : "Error", ret);

    exit (0);
}



/* ========================================================================
 * Table of CRC-32's of all single-byte values (made by makecrc.c)
 */
const ulg crc_32_tab[] = {
  0x00000000L, 0x77073096L, 0xee0e612cL, 0x990951baL, 0x076dc419L,
  0x706af48fL, 0xe963a535L, 0x9e6495a3L, 0x0edb8832L, 0x79dcb8a4L,
  0xe0d5e91eL, 0x97d2d988L, 0x09b64c2bL, 0x7eb17cbdL, 0xe7b82d07L,
  0x90bf1d91L, 0x1db71064L, 0x6ab020f2L, 0xf3b97148L, 0x84be41deL,
  0x1adad47dL, 0x6ddde4ebL, 0xf4d4b551L, 0x83d385c7L, 0x136c9856L,
  0x646ba8c0L, 0xfd62f97aL, 0x8a65c9ecL, 0x14015c4fL, 0x63066cd9L,
  0xfa0f3d63L, 0x8d080df5L, 0x3b6e20c8L, 0x4c69105eL, 0xd56041e4L,
  0xa2677172L, 0x3c03e4d1L, 0x4b04d447L, 0xd20d85fdL, 0xa50ab56bL,
  0x35b5a8faL, 0x42b2986cL, 0xdbbbc9d6L, 0xacbcf940L, 0x32d86ce3L,
  0x45df5c75L, 0xdcd60dcfL, 0xabd13d59L, 0x26d930acL, 0x51de003aL,
  0xc8d75180L, 0xbfd06116L, 0x21b4f4b5L, 0x56b3c423L, 0xcfba9599L,
  0xb8bda50fL, 0x2802b89eL, 0x5f058808L, 0xc60cd9b2L, 0xb10be924L,
  0x2f6f7c87L, 0x58684c11L, 0xc1611dabL, 0xb6662d3dL, 0x76dc4190L,
  0x01db7106L, 0x98d220bcL, 0xefd5102aL, 0x71b18589L, 0x06b6b51fL,
  0x9fbfe4a5L, 0xe8b8d433L, 0x7807c9a2L, 0x0f00f934L, 0x9609a88eL,
  0xe10e9818L, 0x7f6a0dbbL, 0x086d3d2dL, 0x91646c97L, 0xe6635c01L,
  0x6b6b51f4L, 0x1c6c6162L, 0x856530d8L, 0xf262004eL, 0x6c0695edL,
  0x1b01a57bL, 0x8208f4c1L, 0xf50fc457L, 0x65b0d9c6L, 0x12b7e950L,
  0x8bbeb8eaL, 0xfcb9887cL, 0x62dd1ddfL, 0x15da2d49L, 0x8cd37cf3L,
  0xfbd44c65L, 0x4db26158L, 0x3ab551ceL, 0xa3bc0074L, 0xd4bb30e2L,
  0x4adfa541L, 0x3dd895d7L, 0xa4d1c46dL, 0xd3d6f4fbL, 0x4369e96aL,
  0x346ed9fcL, 0xad678846L, 0xda60b8d0L, 0x44042d73L, 0x33031de5L,
  0xaa0a4c5fL, 0xdd0d7cc9L, 0x5005713cL, 0x270241aaL, 0xbe0b1010L,
  0xc90c2086L, 0x5768b525L, 0x206f85b3L, 0xb966d409L, 0xce61e49fL,
  0x5edef90eL, 0x29d9c998L, 0xb0d09822L, 0xc7d7a8b4L, 0x59b33d17L,
  0x2eb40d81L, 0xb7bd5c3bL, 0xc0ba6cadL, 0xedb88320L, 0x9abfb3b6L,
  0x03b6e20cL, 0x74b1d29aL, 0xead54739L, 0x9dd277afL, 0x04db2615L,
  0x73dc1683L, 0xe3630b12L, 0x94643b84L, 0x0d6d6a3eL, 0x7a6a5aa8L,
  0xe40ecf0bL, 0x9309ff9dL, 0x0a00ae27L, 0x7d079eb1L, 0xf00f9344L,
  0x8708a3d2L, 0x1e01f268L, 0x6906c2feL, 0xf762575dL, 0x806567cbL,
  0x196c3671L, 0x6e6b06e7L, 0xfed41b76L, 0x89d32be0L, 0x10da7a5aL,
  0x67dd4accL, 0xf9b9df6fL, 0x8ebeeff9L, 0x17b7be43L, 0x60b08ed5L,
  0xd6d6a3e8L, 0xa1d1937eL, 0x38d8c2c4L, 0x4fdff252L, 0xd1bb67f1L,
  0xa6bc5767L, 0x3fb506ddL, 0x48b2364bL, 0xd80d2bdaL, 0xaf0a1b4cL,
  0x36034af6L, 0x41047a60L, 0xdf60efc3L, 0xa867df55L, 0x316e8eefL,
  0x4669be79L, 0xcb61b38cL, 0xbc66831aL, 0x256fd2a0L, 0x5268e236L,
  0xcc0c7795L, 0xbb0b4703L, 0x220216b9L, 0x5505262fL, 0xc5ba3bbeL,
  0xb2bd0b28L, 0x2bb45a92L, 0x5cb36a04L, 0xc2d7ffa7L, 0xb5d0cf31L,
  0x2cd99e8bL, 0x5bdeae1dL, 0x9b64c2b0L, 0xec63f226L, 0x756aa39cL,
  0x026d930aL, 0x9c0906a9L, 0xeb0e363fL, 0x72076785L, 0x05005713L,
  0x95bf4a82L, 0xe2b87a14L, 0x7bb12baeL, 0x0cb61b38L, 0x92d28e9bL,
  0xe5d5be0dL, 0x7cdcefb7L, 0x0bdbdf21L, 0x86d3d2d4L, 0xf1d4e242L,
  0x68ddb3f8L, 0x1fda836eL, 0x81be16cdL, 0xf6b9265bL, 0x6fb077e1L,
  0x18b74777L, 0x88085ae6L, 0xff0f6a70L, 0x66063bcaL, 0x11010b5cL,
  0x8f659effL, 0xf862ae69L, 0x616bffd3L, 0x166ccf45L, 0xa00ae278L,
  0xd70dd2eeL, 0x4e048354L, 0x3903b3c2L, 0xa7672661L, 0xd06016f7L,
  0x4969474dL, 0x3e6e77dbL, 0xaed16a4aL, 0xd9d65adcL, 0x40df0b66L,
  0x37d83bf0L, 0xa9bcae53L, 0xdebb9ec5L, 0x47b2cf7fL, 0x30b5ffe9L,
  0xbdbdf21cL, 0xcabac28aL, 0x53b39330L, 0x24b4a3a6L, 0xbad03605L,
  0xcdd70693L, 0x54de5729L, 0x23d967bfL, 0xb3667a2eL, 0xc4614ab8L,
  0x5d681b02L, 0x2a6f2b94L, 0xb40bbe37L, 0xc30c8ea1L, 0x5a05df1bL,
  0x2d02ef8dL
};


/* ===========================================================================
 * Run a set of bytes through the crc shift register.  If s is a NULL
 * pointer, then initialize the crc shift register contents instead.
 * Return the current crc in either case.
 */
ulg updcrc(uch *s, unsigned n)
{
    ulg c;         /* temporary variable */

    static ulg crc = (ulg)0xffffffffL; /* shift register contents */

    if (s == NULL) {
	c = 0xffffffffL;
    } else {
	c = crc;
        if (n) do {
            c = crc_32_tab[((int)c ^ (*s++)) & 0xff] ^ (c >> 8);
        } while (--n);
    }
    crc = c;
    return c ^ 0xffffffffL;       /* (instead of ~c for 64-bit machines) */
}
