#ident "kit/ALGVME/sbdvic.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ALGVME/sbdvic.c: VIC068 interrupt handling for Algorithmics VME boards
 */


#include <stddef.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <machine/endian.h>
#include <sys/signal.h>
#include <assert.h>

#include "kit_impl.h"
#include "sbd.h"
#include "vic068.h"
#include "vac068.h"


#undef HANDLE_VME_INTERRUPTS
#undef HANDLE_DMA_INTERRUPT
#undef HANDLE_ICGS_INTERRUPTS
#undef HANDLE_ICMS_INTERRUPTS
#define HANDLE_ERROR_INTERRUPTS
#define HANDLE_PANIC_INTERRUPT


/* VIC interrupt vector bases */
#define VITABBASE	64
#define vic_errgrpvec	0+VITABBASE	/* 8 off error interrupts */
#define vic_locgrpvec	8+VITABBASE	/* 8 off local interrupts */
#define vic_icgsvec	16+VITABBASE	/* 4 off icgs interrupts */
#define vic_icmsvec	24+VITABBASE	/* 4 off icms interrupts*/


/*
 * VIC (second level) interrupt handler 
 *
 * The generic handler in ../share/intrupt.c does most of the work.
 */

static struct {
    xcpt_t	handler;
    int		arg;
} vitab[256];


static int
vistray (int vec, struct xcptcontext *xcp)
{
    /* should probably disable it: it won't go away by itself :-) */
    _mon_puts ("ERROR: stray VIC interrupt\r\n");
    return SIGKILL;
}


/*
 * Analogous to intrupt() function, but takes a VIC vector number
 * instead of a MIPS interrupt number.
 */
static xcpt_t
vic_intrupt (unsigned int ivec, xcpt_t handler, int arg)
{
    xcpt_t ohandler;
    unsigned long sr;

    if (ivec >= 256 || ivec < VITABBASE)
	return XCPT_ERR;

    sr = mips_bicsr (SR_IE);

    ohandler = vitab[ivec].handler;
    if (handler == XCPT_DFL) {
	vitab[ivec].handler = vistray;
	vitab[ivec].arg = ivec;
    } else {
	vitab[ivec].handler = handler;
	vitab[ivec].arg = arg;
    }
    
    mips_setsr (sr);

    return (ohandler == vistray) ? XCPT_DFL : ohandler;
}


static void *iackaddrs[8] = {
    0,
    PA_TO_KVA1(IACK_BASE2+2),
    PA_TO_KVA1(IACK_BASE2+4),
    PA_TO_KVA1(IACK_BASE2+6),
    PA_TO_KVA1(IACK_BASE2+8),
    PA_TO_KVA1(IACK_BASE2+10),
    PA_TO_KVA1(IACK_BASE2+12),
    PA_TO_KVA1(IACK_BASE2+14),
};


/*
 * Second level interrupt handler for VIC interrupt controller
 */
static int
vic_intrhandler (int intr, struct xcptcontext *xcp)
{
    unsigned vec;
    
    /* acknowledge VIC interrupt, and fetch vector */
#if BYTE_ORDER==BIG_ENDIAN
    if (intr & 1)
	vec = *(volatile unsigned short *)iackaddrs[intr];
    else
	vec = *(volatile unsigned int *)iackaddrs[intr];
#endif
#if BYTE_ORDER==LITTLE_ENDIAN
    /* bit order preserved */
    if (intr & 1)
	vec = *(volatile unsigned short *)iackaddrs[intr] >> 8;
    else
	vec = *(volatile unsigned int *)iackaddrs[intr] >> 24;
#endif

    vec &= 0xff;
    assert(vec != 0);
    return (*vitab[vec].handler) (vitab[vec].arg, xcp);
}


static void
vic_intrupt_init ()
{
    int i;

    for (i = 0; i < 256; i++) {
	vitab[i].handler = vistray;
	vitab[i].arg = i;
    }

    /* MIPS hardware interrupts come straight from the VIC */
    intrupt (2, vic_intrhandler, 2);
    intrupt (3, vic_intrhandler, 3);
    intrupt (4, vic_intrhandler, 4);
#if #cpu(r4000)
    intrupt (5, vic_intrhandler, 5); /* fpu interrupt on R3000 */
#endif
    intrupt (6, vic_intrhandler, 6);
    intrupt (7, vic_intrhandler, 7);
}



/*
 * VMEBus interrupts
 */
static volatile unsigned int *vmeicr[] = {
  VIC_VICR1,
  VIC_VICR2,
  VIC_VICR3,
  VIC_VICR4,
  VIC_VICR5,
  VIC_VICR6,
  VIC_VICR7
};


static void
vic_vmeinterrupt (int vmevec, int ipl)
{
  *vmeicr[(vmevec & 7) - 1] = VIC_INT_DIS | (ipl & 7);
  mips_wbflush ();
}

#ifdef HANDLE_VME_INTERRUPTS
void
_vic_vmeinterrupt_enable (int vmevec)
{
  *vmeicr[*vmevec & 7) - 1] &= ~VIC_INT_DIS;
  mips_wbflush ();
}

void
_vic_vmeinterrupt_disable (int vmevec)
{
  *vmeicr[(vmevec & 7) - 1] |= VIC_INT_DIS;
  mips_wbflush ();
}
#endif

static void
vic_setup_vme_interrupts ()
{
  int i;
  for (i = 1; i <= 7; i++)
      vic_vmeinterrupt (i, i);
}



/*
 * DMA status interrupt
 */
#ifdef HANDLE_DMA_INTERRUPT
static void
vic_dmainterrupt (int ipl)
{
  *VIC_DMAICR = VIC_INT_DIS | (ipl & 7);
  mips_wbflush ();
}

void
_vic_dmainterrupt_enable ()
{
  *VIC_DMAICR &= ~VIC_INT_DIS;
  mips_wbflush ();
}

void
_vic_dmainterrupt_disable ()
{
  *VIC_DMAICR |= VIC_INT_DIS;
  mips_wbflush ();
}
#endif



/*
 * Interprocessor communications global interrupt
 */
static void
vic_icgsinterrupt (int ipl)
{
  unsigned int val;
  val = VIC_ICGS3_DIS | VIC_ICGS2_DIS | VIC_ICGS1_DIS | VIC_ICGS0_DIS;
  *VIC_ICGSICR = val | (ipl & 7);
  mips_wbflush ();
}

static void
vic_icgsvector (int n)
{
  *VIC_ICGSVEC = n & VIC_ICGSUMASK;
  mips_wbflush ();
}

#ifdef HANDLE_ICGS_INTERRUPTS
void
_vic_icgsinterrupt_enable (int s3, int s2, int s1, int s0)
{
  unsigned int val;
  val = s3 ? VIC_ICGS3_DIS : 0;
  val |= s2 ? VIC_ICGS2_DIS : 0;
  val |= s1 ? VIC_ICGS1_DIS : 0;
  val |= s0 ? VIC_ICGS0_DIS : 0;
  *VIC_ICGSICR &= ~val;
  mips_wbflush ();
}

void
_vic_icgsinterrupt_disable (int s3, int s2, int s1, int s0)
{
  unsigned int val;
  val = s3 ? VIC_ICGS3_DIS : 0;
  val |= s2 ? VIC_ICGS2_DIS : 0;
  val |= s1 ? VIC_ICGS1_DIS : 0;
  val |= s0 ? VIC_ICGS0_DIS : 0;
  *VIC_ICGSICR |= val;
  mips_wbflush ();
}

static int
vic_icgsintr (int swtch, struct xcptcontext *xcp)
{
  *VIC_ICSR &= ~(0x10 << swtch);
  mips_wbflush ();
#if 0
  /* Do something more interesting here! */
  _mon_printf ("Note: ICGS%d set", swtch);
#endif
  return 0;
}
#endif

static void
vic_setup_icgs_interrupts ()
{
  assert ((vic_icgsvec & 3) == 0);
  vic_icgsvector (vic_icgsvec);
  vic_icgsinterrupt (4);	/* ICGS -> Intr4 */
#ifdef HANDLE_ICGS_INTERRUPTS
  {
      int i;
      for (i = 0; i < 4; i++)
	  vic_intrupt (vic_icgsvec + i, vic_icgsintr, i);
      _vic_icgsinterrupt_enable (1, 1, 1, 1);
  }
#endif
}




/*
 * Interprocessor communications module interrupt
 */
static void
vic_icmsinterrupt (int ipl)
{
  unsigned int val;
  val = VIC_ICMS3_DIS | VIC_ICMS2_DIS | VIC_ICMS1_DIS | VIC_ICMS0_DIS;
  *VIC_ICMSICR = val | (ipl & 7);
  mips_wbflush ();
}

static void
vic_icmsvector (int n)
{
  *VIC_ICMSVEC = n & VIC_ICMSUMASK;
  mips_wbflush ();
}

#ifdef HANDLE_ICMS_INTERRUPTS
void
_vic_icmsinterrupt_enable (int s3, int s2, int s1, int s0)
{
  unsigned int val;
  val = s3 ? VIC_ICMS3_DIS : 0;
  val |= s2 ? VIC_ICMS2_DIS : 0;
  val |= s1 ? VIC_ICMS1_DIS : 0;
  val |= s0 ? VIC_ICMS0_DIS : 0;
  *VIC_ICMSICR &= ~val;
  mips_wbflush ();
}

void
_vic_icmsinterrupt_disable (int s3, int s2, int s1, int s0)
{
  unsigned int val;
  val = s3 ? VIC_ICMS3_DIS : 0;
  val |= s2 ? VIC_ICMS2_DIS : 0;
  val |= s1 ? VIC_ICMS1_DIS : 0;
  val |= s0 ? VIC_ICMS0_DIS : 0;
  *VIC_ICMSICR |= val;
  mips_wbflush ();
}

static int
vic_icmsintr (int swtch, struct xcptcontext *xcp)
{
  *VIC_ICSR &= ~(0x01 << swtch); 
  mips_wbflush ();
#if 0
  /* Do something more interesting here! */
  _mon_printf ("Note: ICMS%d set", swtch);
#endif
  return 0;
}
#endif


static void
vic_setup_icms_interrupts ()
{
  assert ((vic_icmsvec & 3) == 0);
  vic_icmsvector (vic_icmsvec);
  vic_icmsinterrupt (4);	/* ICMS -> Intr4 */
#ifdef HANDLE_ICMS_INTERRUPTS
  {
      int i;
      for (i = 0; i < 4; i++)
	  vic_intrupt (vic_icmsvec + i, vic_icmsintr, i);
      _vic_icmsinterrupt_enable (1, 1, 1, 1);
  }
#endif
}



/*
 * Error group interrupt
 */
static void
vic_errorinterrupt (int ipl)
{
  unsigned int val;
  val = VIC_ACFAIL_DIS | VIC_WPOST_DIS | VIC_ARBF_DIS | VIC_SYSF_DIS;
  *VIC_EGRPICR = val | (ipl & 7);  
  mips_wbflush ();
}

static void
vic_errorvector (int n)
{
  *VIC_EGRPVEC = n & VIC_EGRPUMASK;
  mips_wbflush ();
}

#ifdef HANDLE_ERROR_INTERRUPTS
void
_vic_errorinterrupt_enable (int acfail, int wpost, int arbf, int sysf)
{
  unsigned int val;
  val = acfail ? VIC_ACFAIL_DIS : 0;
  val |= wpost ? VIC_WPOST_DIS : 0;
  val |= arbf ? VIC_ARBF_DIS : 0;
  val |= sysf ? VIC_SYSF_DIS : 0;
  *VIC_EGRPICR &= ~val;
  mips_wbflush ();
}

void
_vic_errorinterrupt_disable (int acfail, int wpost, int arbf, int sysf)
{
  unsigned int val;
  val = acfail ? VIC_ACFAIL_DIS : 0;
  val |= wpost ? VIC_WPOST_DIS : 0;
  val |= arbf ? VIC_ARBF_DIS : 0;
  val |= sysf ? VIC_SYSF_DIS : 0;
  *VIC_EGRPICR |= val;
  mips_wbflush ();
}

static int
vic_errorintr (int vec, struct xcptcontext *xcp)
{
    const char *msg;

    switch (vec) {
    case VIC_ACFAILVEC:
	msg = "ACFAIL*";
	break;
    case VIC_WPOSTVEC:
	msg = "failed writepost";
	break;
    case VIC_ARBFVEC:
	msg = "arbitration timeout";
	break;
    case VIC_SYSFVEC:
	msg = "SYSFAIL*";
	break;
    case VIC_VIIVEC:
	msg = "interrupter acknowledge";
	break;
    case VIC_DMAVEC:
	msg = "DMA completion";
	break;
    default:
	msg = "unknown";
	break;
    }

    _mon_printf ("VIC error: %s\r\n", msg);
    return SIGKILL;
}
#endif

static void
vic_setup_error_interrupts ()
{
  int i;

  assert ((vic_errgrpvec & 7) == 0);
  vic_errorvector (vic_errgrpvec);
  vic_errorinterrupt (7);
#ifdef HANDLE_ERROR_INTERRUPTS
  for (i = 0; i < 6; i++)
    vic_intrupt (vic_errgrpvec + i, vic_errorintr, i);
  _vic_errorinterrupt_enable (1, 1, 1, 0);
#endif
}



/*
 * Local board device interrupts
 */

static volatile unsigned int *locicr[7] = {
  VIC_LICR1,
  VIC_LICR2,
  VIC_LICR3,
  VIC_LICR4,
  VIC_LICR5,
  VIC_LICR6,
  VIC_LICR7,
};

enum lpolarity	{LO, HI};
enum lsense	{LEVEL, EDGE};

static void
vic_locvector (int n)
{
  *VIC_LOCVEC = n & VIC_LIRQUMASK;
  mips_wbflush ();
}

static void
vic_locinterrupt (unsigned local, enum lpolarity polarity,
		  enum lsense sense, unsigned vector, unsigned ipl)
{
  unsigned int val;
  
  val = ipl & 7;
  val |= (polarity == HI) ? VIC_LACTIVEHI : VIC_LACTIVELO;
  val |= (sense == EDGE) ? VIC_LEDGE : VIC_LLEVEL;
  if (vector)
      val |= VIC_LVEC;

  *locicr[(local & 7) - 1] = VIC_INT_DIS | val;
  mips_wbflush ();
}

void
_vic_locinterrupt_enable (int local)
{
  *locicr[(local & 7) - 1] &= ~VIC_INT_DIS;
  mips_wbflush ();
}

void
_vic_locinterrupt_disable (int local)
{
  *locicr[(local & 7) - 1] |= VIC_INT_DIS;
  mips_wbflush ();
}


xcpt_t
_vic_locintrupt (unsigned int lvec, xcpt_t handler, int arg)
{
    return vic_intrupt (vic_locgrpvec + lvec, handler, arg);
}

static void
vic_setup_local_interrupts ()
{
    /* Set up local device interrupt wiring */
    assert ((vic_locgrpvec & 7) == 0);
    vic_locvector (vic_locgrpvec);
    vic_locinterrupt (LIRQ_DEBUG,   LO, LEVEL, 1, 7);	/* Dbg/Perr -> Intr7 */
    vic_locinterrupt (LIRQ_MAILBOX, LO, LEVEL, 1, 6);	/* VAC mbox -> Intr6 */
    vic_locinterrupt (LIRQ_TIMER,   LO, EDGE,  1, 4);	/* Timer    -> Intr4 */
    vic_locinterrupt (LIRQ_SONIC,   LO, LEVEL, 1, 3);	/* Ethernet -> Intr3 */
    vic_locinterrupt (LIRQ_SCSI,    LO, LEVEL, 1, 3);	/* SCSI	    -> Intr3 */
    vic_locinterrupt (LIRQ_VACSER,  LO, LEVEL, 1, 2);	/* VAC ser  -> Intr2 */
    vic_locinterrupt (LIRQ_MPSC,    LO, LEVEL, 1, 2);	/* MPSC ser -> Intr2 */
}	    



/*
 * VAC068 interrupts.
 */

static void
vac_perrctl (on)
int on;
{
  unsigned int oval;

  oval = *VACICR & ~VAC_MAPPERR(VAC_MAP_MASK);
  if (on)
    *VACICR = oval | VAC_MAPPERR(VAC_ICR_PIO11);
  else
    *VACICR = oval | VAC_MAPPERR(VAC_ICR_NONE);
  mips_wbflush ();
}

void
_vac_acktimer ()
{
    unsigned int oval;
    oval = *VACICR;

    *VACICR = (oval & ~VAC_MAPTIMER(VAC_MAP_MASK)) | 
	 VAC_MAPTIMER(VAC_ICR_NONE);
    mips_wbflush ();
    
    *VACICR = oval;
    mips_wbflush ();
}

#ifdef HANDLE_PANIC_INTERRUPT

static void
vac_ackdebug ()
{
  unsigned int oval;

  oval = *VACICR;

  *VACICR = (oval & ~VAC_MAPDEBUG(VAC_MAP_MASK)) | VAC_MAPDEBUG(VAC_ICR_NONE);
  mips_wbflush ();

  *VACICR = oval;
  mips_wbflush ();
}

static void
vac_ackperr ()
{
  unsigned int oval;

  oval = *VACICR;

  *VACICR = (oval & ~VAC_MAPPERR(VAC_MAP_MASK)) | VAC_MAPPERR(VAC_ICR_NONE);
  mips_wbflush ();

  *VACICR = oval;
  mips_wbflush ();
}


static int
vac_panicintr (int arg, struct xcptcontext *xcp)
{
  unsigned int vacisr;
    
  vacisr = *VACISR;
    
  if (vacisr & VAC_ISR_DEBUG) {
    vac_ackdebug ();
    _debug_breakpoint ();
    /*
     * any pending parity error will get handled on next interrupt
     * and this saves confusion
     */
    return 0;
  }   

  if (vacisr & VAC_ISR_PERR) {
    int errstat;
    errstat = *(volatile unsigned *)PA_TO_KVA1(ERRREG);
    vac_ackperr ();
    _mon_puts ("VAC parity error\r\n");
    return SIGBUS;
  }

  _mon_puts ("Unexpected VAC panic\r\n");
  return SIGKILL;
}
#endif

static void
vac_setup_panic_interrupt ()
{
    vac_perrctl (0);
#ifdef HANDLE_PANIC_INTERRUPT
    _vic_locintrupt (LIRQ_DEBUG, vac_panicintr, 0);
    _vic_locinterrupt_enable (LIRQ_DEBUG);
    vac_perrctl (1);
#endif
}



void
_vic_initialise ()
{
    static int vic_initialised = 0;

    if (!vic_initialised) {
	vic_initialised = 1;
	vic_intrupt_init ();
	vic_setup_icms_interrupts ();
	vic_setup_icgs_interrupts ();
	vic_setup_local_interrupts ();
	vic_setup_vme_interrupts ();
	vic_setup_error_interrupts ();
	vac_setup_panic_interrupt ();
#if #cpu(r4000)
	mips_bissr (SR_IM2 | SR_IM3 | SR_IM4 | SR_IM5 | SR_IM6 | SR_IM7);
#else
	mips_bissr (SR_IM2 | SR_IM3 | SR_IM4 | SR_IM6 | SR_IM7);
#endif
    }
}
