#ident "kit/SKEL/sbdpci.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1997-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * SKEL/sbdpci.c: Machine-specific functions for PCI autoconfiguration.
 */


#if #pci(yes)

#include <sys/types.h>
#include <mips/xcpt.h>
#include <mips/cpu.h>
#include <pci/pcivar.h>
#include <pci/pcireg.h>

#include "kit_impl.h"
#include "sbd.h"

/* PCI regions in PCI space */
#define PCI_IO_SPACE_PCI_BASE	0x00000000
#define PCI_MEM_SPACE_PCI_BASE	0x10000000

/* advancing addresses for allocate_mem/io */
static pcireg_t nextpciioaddr = PCI_IO_SPACE_PCI_BASE;
static pcireg_t nextpcimemaddr = PCI_MEM_SPACE_PCI_BASE;

/* initial value of pci bus descriptor */
static const struct pci_bus skel_pci_bus = {
	0,		/* minimum grant */
	255,		/* maximum latency */
	0,		/* devsel time = fast */
	1,		/* we support fast back-to-back */
	1,		/* we support prefetch */
	0,		/* we don't support 66 MHz */
	0,		/* we don't support 64 bits */
	4000000,	/* bandwidth: in 0.25us cycles / sec */
	1		/* initially one device on bus (i.e. us) */
};

/* one such structure for each PCI bus (allowing for PCI/PCI bridges) */
#define MAXBUS	3
struct pci_bus _pci_bus[MAXBUS];

__attribute__ ((section (".sdata")))
int _pci_maxbus = MAXBUS;

extern int _pciverbose;

static unsigned long pci_mem_space_size;


/*
 * Called to initialise the bridge at the beginning of time
 */
int
_pci_hwinit (int initialise)
{
    /* Reset the bridge chip & PCI bus */

    /* Set up any PCI <-> memory apertures, etc */
    pci_mem_space_size = 256 * 1024 * 1024;

    if (initialise) {
	/* Unreset the PCI bus and program bridge chip */
    }

    /* Initialise global data */
    nextpciioaddr = PCI_IO_SPACE_PCI_BASE;
    nextpcimemaddr = PCI_MEM_SPACE_PCI_BASE;
    _pci_bus[0] = skel_pci_bus;

    /* Return 1 if PCI controller, and should initialise PCI devices.
       Return 0 if not PCI controller, and should just scan bus.
       Return -1 on error, don't scan bus */
    return initialise;
}


/*
 * Called to reinitialise the bridge after we've scanned each PCI device
 * and know what is possible.
 */
void
_pci_hwreinit (void)
{
    if (_pci_bus[0].fast_b2b) {
	/* fast back-to-back is supported by all devices, 
	   enable on bridge. */
    }

    /* program variable latency timer */
    XXX_CFG = _pci_bus[0].def_ltim;

    if (_pci_bus[0].prefetch) {
	/* we can safely prefetch from all pci devices, 
	   enable on bridge */
    }
}


/* Map the <BUS,DEVICE,FUNC> triplet into a single TAG */
pcitag_t
_pci_make_tag (int bus, int device, int function)
{
    pcitag_t tag;
    tag = (bus << 16) | (device << 8) | function;
    return tag;
}

/* Map the TAG back into the <BUS,DEVICE,FUNC> triplet */
void
_pci_break_tag (pcitag_t tag, int *busp, int *devicep, int *functionp)
{
    if (busp) *busp = (tag >> 16) & 0x0f;
    if (devicep) *devicep = (tag >> 8) & 0x1f;
    if (functionp) *functionp = tag & 0x87;
}


/* Read register REG from configuration space of device TAG; 
   return ~0 if not readable */
pcireg_t
_pci_conf_read (pcitag_t tag, int reg)
{
    u_int32_t addr;
    pcireg_t data;
    int bus, device, function;

    _pci_break_tag (tag, &bus, &device, &function); 

    if (reg & 3 || reg < 0 || reg >= 0x100) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: bad reg 0x%x\r\n", reg);
	return ~0;
    }

    /* Example encoding of <device/func/reg> number into 
       config space address. */
    if (bus != 0 || device > 20)
	return ~0;	/* device out of range */
    addr = (0x800 << device) | ((function & 0x07) << 8) | reg;

    /* high 12 bits of address go in map register, and set for conf space */
    XXX_LB_MAP0 = ((addr >> 16) & XXX_LB_MAPx_MAP_ADR) | XXX_LB_TYPE_CONF;

    /* low 20 bits of address are in the actual address */
    data = *(volatile pcireg_t *) PA_TO_KVA1(PCI_CONF_SPACE + (addr&0xfffff));

    if (XXX_PCI_STAT & XXX_PCI_STAT_M_ABORT) {
	XXX_PCI_STAT |= XXX_PCI_STAT_M_ABORT;
#if 0
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: master abort\r\n");
#endif
	return ~0;
    }

    if (XXX_PCI_STAT & XXX_PCI_STAT_T_ABORT) {
	XXX_PCI_STAT |= XXX_PCI_STAT_T_ABORT;
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: target abort\r\n");
	return ~0;
    }

    return data;
}


/* Write DATA to register REG in configuration space of device TAG */
void
_pci_conf_write (pcitag_t tag, int reg, pcireg_t data)
{
    u_int32_t addr;
    int bus, device, function;

    _pci_break_tag (tag, &bus, &device, &function);

    if (reg & 3 || reg < 0 || reg >= 0x100) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: bad reg %x\r\n", reg);
	return;
    }

    /* Example encoding of <device/func/reg> number into 
       config space address. */
    if (bus != 0 || device > 20)
	return ~0;	/* device out of range */
    addr = (0x800 << device) | ((function & 0x07) << 8) | reg;

    /* high 12 bits of address go in map register, and set for conf space */
    XXX_LB_MAP0 = ((addr >> 16) & XXX_LB_MAPx_MAP_ADR) | XXX_LB_TYPE_CONF;

    /* low 20 bits of address are in the actual address */
    *(volatile pcireg_t *) PA_TO_KVA1(PCI_CONF_SPACE + (addr&0xfffff)) = data;

    if (XXX_PCI_STAT & XXX_PCI_STAT_M_ABORT) {
	XXX_PCI_STAT |= XXX_PCI_STAT_M_ABORT;
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_write: master abort\r\n");
    }

    if (XXX_PCI_STAT & XXXX_PCI_STAT_T_ABORT) {
	XXX_PCI_STAT |= XXX_PCI_STAT_T_ABORT;
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_write: target abort\r\n");
    }
}



/* Return a virtual and physical address which maps to the area of pci
   i/o space described by mapping register REG on device TAG. */
int
_pci_map_io (pcitag_t tag, int reg, vm_offset_t *vap, vm_offset_t *pap)
{
    if (reg < PCI_MAP_REG_START || reg >= PCI_MAP_REG_END || (reg & 3)) {
	if (_pciverbose >= 1)
	    _pci_tagprintf(tag, "_pci_map_io: bad request\r\n");
	return -1;
    }

    address = _pci_conf_read(tag, reg);

    if ((address & PCI_MAP_IO) == 0) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_io: attempt to memory map a memory region\r\n");
	return -1;
    }

    pa = (address & PCI_MAP_IO_ADDRESS_MASK) - PCI_IO_SPACE_PCI_BASE;
    *pap = pa;

    *vap = (vm_offset_t) PA_TO_KVA1 (PCI_IO_SPACE + pa);

    if (_pciverbose >= 3)
	_pci_tagprintf (tag, "_pci_map_io: mapping io at virtual 0x%x, physical 0x%x\r\n", *vap, *pap);

    return 0;
}


/* Return a virtual and physical address which maps to the area of pci
   memory space described by mapping register REG on device TAG. */
int
_pci_map_mem (pcitag_t tag, int reg, vm_offset_t *vap, vm_offset_t *pap)
{
    pcireg_t address;
    vm_offset_t pa;

    if (reg < PCI_MAP_REG_START || reg >= PCI_MAP_REG_END || (reg & 3)) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_mem: bad request\r\n");
	return -1;
    }

    address = _pci_conf_read(tag, reg);

    if ((address & PCI_MAP_IO) != 0) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_mem: attempt to memory map an I/O region\r\n");
	return -1;
    }

    switch (address & PCI_MAP_MEMORY_TYPE_MASK) {
    case PCI_MAP_MEMORY_TYPE_32BIT:
    case PCI_MAP_MEMORY_TYPE_32BIT_1M:
	break;
    case PCI_MAP_MEMORY_TYPE_64BIT:
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_mem: attempt to map 64-bit region\r\n");
	return -1;
    default:
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_mem: reserved mapping type\r\n");
	return -1;
    }

    pa = (address & PCI_MAP_MEMORY_ADDRESS_MASK) - PCI_MEM_SPACE_PCI_BASE;
    *pap = pa;

    if (address & PCI_MAP_MEMORY_CACHABLE) 
	*vap = (vm_offset_t) PA_TO_KVA0 (PCI_MEM_SPACE + pa);
    else
	*vap = (vm_offset_t) PA_TO_KVA1 (PCI_MEM_SPACE + pa);

    if (_pciverbose >= 3)
	_pci_tagprintf (tag, "_pci_map_mem: mapping memory at virtual 0x%x, physical 0x%x\r\n", *vap, *pap);
    return 0;
}


/* Return interrupt number used by given device */
int
_pci_map_int (pcitag_t tag)
{
    static const int sbdintrmap[4] = {INTR_PCIA, INTR_PCIB,
				      INTR_PCIC, INTR_PCID};
    pcireg_t data;
    int pin, bus, device;

    data = _pci_conf_read(tag, PCI_INTERRUPT_REG);
    pin = PCI_INTERRUPT_PIN(data);
    if (pin == 0)
	/* No IRQ used. */
	return 0;

    if (--pin > 3) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_int: bad interrupt pin %d\r\n", pin);
	return NULL;
    }


    _pci_break_tag (tag, &bus, &device, NULL);

    if (bus != 0 || device > 20)
	return -1;

    /* work out which CPU interrupt signal (e.g. Intr0-5) corresponds
       to the given PCI interrupt line (INTA, INTB, INTC or INTD).
       This probably depends on the device number, because each slot
       usually rotates the interrupt wiring*/
    return sbdintrmap[pin % 3];
}

/*
 * Allocate SIZE bytes in pci memory space for pci device TAG
 */
pcireg_t
_pci_allocate_mem (pcitag_t tag, size_t size)
{
    pcireg_t address;

    address = (nextpcimemaddr + size - 1) & ~(size - 1);
    nextpcimemaddr = address + size;
    if (nextpcimemaddr > PCI_MEM_SPACE_PCI_BASE + pci_mem_space_size) {
	if (_pciverbose >= 1)
	    _mon_puts ("_pci_allocate_mem: not enough PCI memory space\r\n");
	return -1;
    }
    return address;
}


/*
 * Allocate SIZE bytes in pci i/o space for pci device TAG
 */
pcireg_t
_pci_allocate_io (pcitag_t tag, size_t size)
{
    pcireg_t address;

    address = (nextpciioaddr + size - 1) & ~(size - 1);
    nextpciioaddr = address + size;
    if (nextpciioaddr > PCI_IO_SPACE_PCI_BASE + PCI_IO_SPACE_SIZE) {
	if (_pciverbose >= 1)
	    _mon_puts ("_pci_allocate_io: not enough PCI i/o space\r\n");
	return -1;
    }
    return address;
}

#endif /* pci(yes) */
