#ident "kit/atlas/sbdfreq.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * atlas/sbdfreq.c: CPU frequency determination for MIPS Atlas board
 */


#include <stddef.h>
#include <time.h>
#include <errno.h>
#include <mips/cpu.h>
#include <mips/prid.h>

#include "kit_impl.h"

#include "sbd.h"
#include "ds1687.h"

/*
 * Dynamically determine CPU pipeline clock frequency by running the
 * on-chip timer against the fixed frequency periodic timer.
 */

#define RTC_HZ		4
#define RTC_RATE	RTC_RATE_4Hz

long
_sbd_pipefreq (void)
{
    static long pipefreq = 0;
    volatile unsigned long * const intreq = PA_TO_KVA1 (INTRAW_ADDR);
    unsigned int osta, ostb;
    unsigned long start, cnt, timeout;
    int i;

    if (pipefreq > 0)
	return pipefreq;

    /* set RTC periodic interrupt frequency */
    osta = _rtc_bic (RTC_STATUSA, RTC_RATE_MASK);
    (void) _rtc_bis (RTC_STATUSA, RTC_RATE);

    /* enable periodic interrupts */
    ostb = _rtc_bic (RTC_STATUSB, RTCSB_AIE | RTCSB_UIE);
    (void) _rtc_bis (RTC_STATUSB, RTCSB_PIE);

    /* run more than once to be sure that we're in the cache */
    for (i = 2, cnt = 0; i != 0; i--) {
	start = cnt;
	_rtc_get (RTC_INTR);		/* clear pending rtc i/u */
	/* wait until we see a timer i/u */
	timeout = 10000000 / RTC_HZ;
	while ((*intreq & INTREQ_RTC) == 0)
	    if (--timeout == 0) return 0;
	cnt = mips_getcount ();
    }

    /* restore old rate and interrupt enables */
    _rtc_set (RTC_STATUSA, osta);
    _rtc_set (RTC_STATUSB, ostb);

    /* work out number of counter ticks per sec */
    pipefreq = (cnt - start) * RTC_HZ;

    /* pipeline runs at 2x counter rate */
    pipefreq *= 2;


    return pipefreq;
}


struct ratio {char mult; char div;};
static const struct ratio ratios_norml[8] = {{2,1}, {3,1}, {4,1}, {5,1},
					     {6,1}, {7,1}, {8,1}, {1,1}};

/*
 * Return the CPU bus clock frequency (usually same input clock).
 */

long
_sbd_cpufreq (void)
{
    static long cpufreq;

    if (cpufreq == 0) {
	long freq = _sbd_pipefreq ();
#ifdef CFG_ECMASK
	/* pipeline runs at multiple of input clock, so divide down */
	const struct ratio *ratios;
	switch ((mips_getprid() >> 8) & 0xff) {
	default:
	    ratios = ratios_norml;
	    break;
	}
	ratios += (mips_getconfig() & CFG_ECMASK) >> CFG_ECSHIFT;
	cpufreq = freq * ratios->div / ratios->mult;
#else
	/* how can we detect the ratio on MIPS32 cpu */
	cpufreq = freq / 2;
#endif
    }

    return cpufreq;
}
