#ident "kit/atlas/sbdfrom.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * atlas/sbdfrom.c: Flash ROM support for MIPS Atlas board
 */


#ifdef IN_PMON
#include "pmon.h"
#include "sde-compat.h"
#include <assert.h>
#else
#include <sys/types.h>
#include <mips/cpu.h>
#include <stdlib.h>
#include <string.h>
#include "kit_impl.h"
#endif

#include "sbdfrom.h"
#include "flashdev.h"
#include "flashrom.h"
#include "sbd.h"

/* writable cookie */
static struct fromcookie sys_cookie;
static struct fromcookie mon_cookie;
static struct fromcookie boot_cookie;

/* board-specific code */
static int fromgetparts (flashcookie_t, struct flashparts *);
static int fromsetparts (flashcookie_t, const struct flashparts *);

static const struct flashboard fromboard = {
    fromgetparts,
    fromsetparts
};


flashcookie_t
_sbd_flashopen (paddr_t devaddr)
{
    struct fromcookie *fcp;
    paddr_t dbase, mbase;
    int banks;

    /* Referencing the BOOTPROM could access either the PROM
       socket or the Flash, depending on a board link setting. */

    if (devaddr == 0
	 || (devaddr >= SYSFLASH_BASE
	     && devaddr < SYSFLASH_BASE + SYSFLASH_SIZE)) {
	/* only safe to access via flash space when little endian */
	fcp = &sys_cookie;
	dbase = mbase = SYSFLASH_BASE;
	banks = 2;
    }
    else if (devaddr == 1
	     || (devaddr >= MONFLASH_BASE
		 && devaddr < MONFLASH_BASE + MONFLASH_SIZE)) {
	/* only safe to access via flash space when little endian */
	fcp = &mon_cookie;
	dbase = mbase = MONFLASH_BASE;
	banks = 1;
    }
    else if (devaddr >= BOOTPROM_BASE
	     && devaddr < BOOTPROM_BASE + BOOTPROM_SIZE) {
	fcp = &boot_cookie;
	if (*(unsigned int *) PA_TO_KVA1 (SWITCH_ADDR) & SWITCH_MONBOOT) {
	    dbase = MONFLASH_BASE;
	    mbase = BOOTPROM_BASE;
	    banks = 1;
	}
	else {
	    dbase = SYSFLASH_BASE;
	    mbase = BOOTPROM_BASE;
	    banks = 2;
	}
    }
    else
	return FLASH_UNKNOWN;

    if (fcp->size == 0) {
	/* always program via flash address */
	fcp->pbase = PA_TO_KVA1 (dbase);

	/* read from there too */
	if (IS_KVA0 (fcp))
	    fcp->rbase = PA_TO_KVA0 (dbase);
	else
	    fcp->rbase = PA_TO_KVA1 (dbase);
	
	/* probe for device */
	fcp->nbanks = banks;
	if (_flashrom_probe (fcp, 0) < 0)
	    return FLASH_MISSING;
	
	/* check upper bank */
	if (fcp->nbanks > 1 && 
	    _flashrom_probe (fcp, fcp->devsize * FROM_NCOLS) < 0)
	    fcp->nbanks = 1;

	fcp->mapbase = mbase;

	if (fcp == &boot_cookie && dbase == SYSFLASH_BASE) {
	    /* uses top 4MB of system flash only */
	    fcp->size = 0x400000;
	    fcp->virtoffs = (fcp->devsize * FROM_NCOLS * fcp->nbanks)
		- 0x400000;
	}
	else {
	    /* compute total size */
	    fcp->size = fcp->devsize * FROM_NCOLS * fcp->nbanks;
	}
    }


    /* plug in board-specific functions */
    fcp->common.board = &fromboard;

    return &fcp->common;
}


/*
 * Return flash partition table 
 */
static int
fromgetparts (flashcookie_t cookie, struct flashparts *parts)
{
    struct fromcookie *fcp = (struct fromcookie *)cookie;
    int monboot;

    monboot = *(unsigned int *) PA_TO_KVA1 (SWITCH_ADDR) & SWITCH_MONBOOT;

    memset (parts, 0, sizeof (*parts));

    _flashdev_setpart (parts, FLASHPART_RAW, FLASHPART_RAW, 0, fcp->size);

    if (fcp == &sys_cookie) {
	/* system flash */
	if (monboot) {
	    /* all available */
	    _flashdev_setpart (parts, FLASHPART_FFS, FLASHPART_UNDEF,
			       0, fcp->size);
	}
	else {
	    /* top 4MB reserved for bootstrap */
	    _flashdev_setpart (parts, FLASHPART_FFS, FLASHPART_UNDEF,
			       0, fcp->size - 0x400000);
	    _flashdev_setpart (parts, FLASHPART_BOOT, FLASHPART_BOOT,
			       fcp->size - 0x400000, 0x400000);
	}
    }
    else if (fcp == &mon_cookie || monboot) {
	/* monitor flash: top 128K reserved for environment */
	_flashdev_setpart (parts, FLASHPART_BOOT, FLASHPART_BOOT, 
			   0, fcp->size - 0x20000);
	_flashdev_setpart (parts, FLASHPART_ENV, FLASHPART_ENV, 
			   fcp->size - 0x20000, 0x20000);
    }
    else {
	/* boot flash: mapped to sysflash */
	_flashdev_setpart (parts, FLASHPART_BOOT, FLASHPART_BOOT, 
			   0, fcp->size);
    }

    return FLASHDEV_OK;
}


static int
fromsetparts (flashcookie_t cookie, const struct flashparts *parts)
{
    /* not supported yet, sorry */
    return FLASHDEV_FAIL;
}
