/*#ident "kit/bare/baretty.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."*/

/*
 * Copyright (c) 1996-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * baretty.c: Posix style terminal i/o on bare machine
 */


#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/init.h>
#include <string.h>
#include <stdlib.h>
#include <termios.h>
#include <assert.h>
#include "sbdser.h"
#include "file.h"
#include "tty.h"
#include "kit_impl.h"

#ifndef CONSPEED
#define CONSPEED 9600		/* default speed of console */
#endif

static int	ttyread (struct file *, struct uio *);
static int	ttywrite (struct file *, struct uio *);
static int	ttyioctl (struct file *, int, void *);
static int	ttyselect (struct file *, int);
static int	ttyopen (struct file *, dev_t, int);
static int	ttyclose (struct file *);

static const struct fileops _tty_ops = 
{
    ttyread, 
    ttywrite, 
    ttyioctl, 
    ttyselect, 
    ttyopen,
    ttyclose,
};


#define NTTY 4
static struct tty *ttys[NTTY];



/*
 * The lowest level device-dependent routines, called by the generic
 * tty code.
 */

static int
ttyparams (struct tty *tp, int set, struct termios *tio)
{
    struct sbdser *ss = (struct sbdser *) tp->t_addr;

    if (set) {
	/* CHANGE SETTINGS */
	if (tio->c_ispeed != tio->c_ospeed)
	    return EINVAL;
	if (tio->c_ospeed != 0) {
	    if ((*ss->ss_params) (ss->ss_data, tio->c_ospeed, 0)) 
		return EINVAL;
	}
	return 0;
    }

    /* GET SETTINGS (leave unchanged) */
    return 0;
}


static int
ttyrxpoll (struct tty *tp, int canblock)
{
    struct sbdser *ss = (struct sbdser *) tp->t_addr;
    return (*ss->ss_rxpoll) (ss->ss_data, canblock, 0);
}

static int
ttytxrdy (struct tty *tp)
{
    struct sbdser *ss = (struct sbdser *) tp->t_addr;
    return (*ss->ss_txrdy) (ss->ss_data, 0);
}

static void
ttytx (struct tty *tp, int c)
{
    struct sbdser *ss = (struct sbdser *) tp->t_addr;
    (*ss->ss_tx) (ss->ss_data, c, 0);
}



static int
ttyopen (struct file *fp, dev_t dev, int fflags)
{
    struct tty *tp;
    int err;

    if (dev >= NTTY || dev >= _sbdnser)
	return ENXIO;

    /* initialise our structure */
    if (!(tp = ttys[dev])) {
	err = _ttymalloc ("tty(bare)", &tp);
	if (err)
	    return err;
	ttys[dev] = tp;
	/* default initial speed and line setting */
	tp->t_ispeed = tp->t_ospeed = _sbdspeed[dev];
    }

    if (!(tp->t_state & TS_OPEN)) {
	const struct sbdser *ss = &_sbdsertab[dev];
	tp->t_dev	= dev;
	tp->t_params	= ttyparams;
	tp->t_rxpoll	= ttyrxpoll;
	tp->t_txrdy	= ttytxrdy;
	tp->t_tx	= ttytx;
	tp->t_addr	= (void *) ss;
	tp->t_cflag	= TTYDEF_CFLAG;
	(*ss->ss_init) (ss->ss_data, 0);
    }

    /* point file at our private data */
    fp->f_data = tp;

    /* generic open */
    return _ttyopen (tp);
}

static int
ttyclose (struct file *fp)
{
    int error = 0;
    /* generic close */
    if (fp->f_data)
	error = _ttyclose ((struct tty *) fp->f_data);
    return error;
}

static int
ttywrite (struct file *fp, struct uio *uio)
{
    /* generic write */
    return _ttywrite ((struct tty *)fp->f_data, uio, fp->f_flag);
}

static int
ttyread (struct file *fp, struct uio *uio)
{
    /* generic read */
    return _ttyread ((struct tty *)fp->f_data, uio, fp->f_flag);
}

static int
ttyioctl (struct file *fp, int cmd, void *data)
{
    /* generic ioctl */
    return _ttyioctl ((struct tty *)fp->f_data, cmd, data);
}

static int
ttyselect (struct file *fp, int rw)
{
    /* generic select */
    return _ttyselect ((struct tty *)fp->f_data, rw);
}



/* /dev/tty*:	crw-rw-r-- */
#define TTYMODE	(S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IWOTH)

static void
ttydev_mount (void)
{
    dir_t ttynode;
    int d;
    int err;

    memset (&ttynode, 0, sizeof (ttynode));
    ttynode.d_mode = TTYMODE;
    ttynode.d_ops = &_tty_ops;

#define ADDDESC(name, dev)				\
    ttynode.d_name = name;				\
    ttynode.d_dev = dev;				\
    err = _mknod ("/dev", &ttynode);			\
    if (err && err != EEXIST) {				\
      _mon_puts ("mknod(/dev/" name ") failed\n");	\
      abort ();						\
    }

    ADDDESC("console",	0);
    ADDDESC("tty",	0);
    ADDDESC("tty0",	0);
    for (d = 1; d < _sbdnser; d++) {
	static char name[] = "ttyx";
	name[3] = d + '0';
	ttynode.d_name = name;
	ttynode.d_dev = d;
	if (_mknod ("/dev", &ttynode))
	    break;
    }
}

DECL_INIT_FUNC(IF_MOUNT, ttydev_mount)

/* call this dummy function to load this module */
void _ttydev_load (void);
void _ttydev_load (void) {} 
