#ident "kit/idt33x/sbdpci.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * idt33x/pci_machdep.c:  PCI autoconfiguration for IDT 79S33x boards
 */


#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <mips/xcpt.h>
#include <mips/cpu.h>
#include <sdethread.h>
#include <pci/pcivar.h>
#include <pci/pcireg.h>

#include "kit_impl.h"
#define printf _mon_printf

#include <idt/rc3233x.h>
#include "sbd.h"
#include "sbdicu.h"

/* Define max size of CPU->PCI mem aperture, restricted by need
   to use TLB resources to access it: 128MB should be enough. */
#define PCI_MEM_APERTURE_SIZE	0x08000000

/* soft versions of PCI region bases */
static pcireg_t pci_mem_space_pci_base	= PCI_MEM_SPACE_PCI_BASE;
static pcireg_t pci_local_mem_pci_base	= PCI_LOCAL_MEM_PCI_BASE;
static pcireg_t pci_io_space_pci_base	= PCI_IO_SPACE_PCI_BASE;

/* PCI mem space allocation */
static pcireg_t minpcimemaddr;
static pcireg_t nextpcimemaddr;

/* PCI i/o space allocation (note - bottom 64KB reserved for ISA i/o space) */
static pcireg_t minpciioaddr;
static pcireg_t nextpciioaddr;

static int pcimaster;
static int pcireserved;

static const struct pci_bus rc32134_pci_bus = {
	0,		/* minimum grant */
	255,		/* maximum latency */
	1,		/* devsel time = medium */
	1,		/* we support fast back-to-back */
	1,		/* we support prefetch */
	0,		/* we *may* support 66 MHz */
	0,		/* we don't support 64 bits */
	4000000,	/* bandwidth: in 0.25us cycles / sec */
	1		/* initially one device on bus (i.e. us) */
};


/* one such structure for each PCI bus (allowing for PCI/PCI bridges) */
#define MAXBUS	3
struct pci_bus _pci_bus[MAXBUS];

__attribute__ ((section (".sdata")))
int _pci_maxbus = MAXBUS;

extern int _pciverbose;

static __sdethread_mutex_t	pci_cfg_mx = __SDETHREAD_MUTEX_INITIALIZER("pci_cfg");

#define pci_cfg_lock() \
  (void) __sdethread_mutex_lock (&pci_cfg_mx)

#define pci_cfg_unlock() \
  (void) __sdethread_mutex_unlock (&pci_cfg_mx)

static volatile pcireg_t * const config_addr = 
	(pcireg_t *) PA_TO_KVA1 (RC3233X_BASE + 0x2cf8);

static volatile pcireg_t * const config_data =
	(pcireg_t *) PA_TO_KVA1 (RC3233X_BASE + 0x2cfc);

#define CONFIG_SEL(reg) \
    do { \
      *config_addr = RC33X_PCI_CONF_ENABLE | (reg); mips_wbflush (); \
    } while (0)

#define CONFIG_RESET() \
    do { \
      *config_addr = 0; mips_wbflush (); \
    } while (0)

#define CONFIG_SET(reg, val) \
    do { \
      CONFIG_SEL(reg); *config_data = (val); mips_wbflush (); \
    } while (0)

#define CONFIG_BIS(reg, val) \
    do { \
      CONFIG_SEL(reg); *config_data |= (val); mips_wbflush (); \
    } while (0)

#define CONFIG_BIC(reg, val) \
    do { \
      CONFIG_SEL(reg); *config_data &= ~(val); mips_wbflush (); \
    } while (0)


static __inline__ pcireg_t 
CONFIG_GET (unsigned int reg)
{
    pcireg_t val;
    CONFIG_SEL (reg);
    val = *config_data;
    return val;
}


/* <bus, device, func> triples to access own config registers */
#define SELF_TAG	(0)

#if #endian(big)
#define PCI_SWAP	RC33X_PCI_MEM_BSWAP
#else	
#define PCI_SWAP	0
#endif

/*
 * Called to initialise the bridge at the beginning of time
 */
int
_pci_hwinit (int initialise)
{
    volatile unsigned char *intstat = PA_TO_KVA1 (INTSTAT_BASE);
    unsigned int idx;
    vaddr_t addr;
    char *s;

    /* slot 0 is self - never scan it */
    pcireserved = 1;

    /* parse the $pcireserved variable to find the devices to be skipped */
    if (s = _sbd_getenv ("pcireserved")) {
	char *sp = s;
	while (*sp) {
	    if (*sp == ',') {
		char *ep;
		long device = strtol (++sp, &ep, 10);
		if (sp == ep || device < 0 || device > 31) {
		    printf ("Invalid $pcireserved value %s\n", s);
		    return -1;
		}
		pcireserved |= 1 << device;
		sp = ep;
	    }
	}
    }

    pci_cfg_lock ();

    _pci_bus[0] = rc32134_pci_bus;

    /* only allow 66MHz operation if the jumper is IN */
    _pci_bus[0].freq66 = !!(*intstat & INTSTAT_PCIM66);

    /* are we master (host) or slave (satellite)? */
    pcimaster = !(*intstat & INTSTAT_PCISAT);

    /* Increase bus timeouts beyond that programmed by IDT/sim, to
       cope with slow PCI devices. Also increase number of PCI retrys,
       and decrease Trdy wait delay. */

#define SYSCLK_USEC	(SYSCLK_HZ / 1000000)

    /* set CPU bus timer (2ms) */
    RC33X_TMR_CTRL(RC33X_CPUBUS_TMR) = 0;
    RC33X_TMR_COMPARE(RC33X_CPUBUS_TMR) = 2000 * SYSCLK_USEC;
    RC33X_TMR_COUNT(RC33X_CPUBUS_TMR) = 0;
    RC33X_TMR_CTRL(RC33X_CPUBUS_TMR) = RC33X_TMR_CTRL_ENB;
	
    /* set IP bus timer (2ms + 1) */
    RC33X_TMR_CTRL(RC33X_IPBUS_TMR) = 0;
    RC33X_TMR_COMPARE(RC33X_IPBUS_TMR) = 2000 * SYSCLK_USEC + 1;
    RC33X_TMR_COUNT(RC33X_IPBUS_TMR) = 0;
    RC33X_TMR_CTRL(RC33X_IPBUS_TMR) = RC33X_TMR_CTRL_ENB;

    /* increase retry timeout */
    CONFIG_SET (PCI_TIMEOUT_REG,
		(0xff << PCI_RETRY_TIMEOUT_SHIFT) |
		(0x20 << PCI_TRDY_TIMEOUT_SHIFT));
	
    if (!initialise) {
	/* not initialising PCI for ourselves: probe the existing setup */
	pci_mem_space_pci_base = RC33X_PCI_MEM1_BASE & RC33X_PCI_MEM_MASK;
	pci_io_space_pci_base = RC33X_PCI_IO_BASE & RC33X_PCI_IO_MASK;

	pci_local_mem_pci_base = (CONFIG_GET (SELF_TAG + PCI_MAP_REG_START)
				  & PCI_MAP_MEMORY_ADDRESS_MASK);
	/* we assume that someone else has initialised the TLB! */
	pci_cfg_unlock ();
	return 0;
    }

    /* we are defining the address regions (see sbdreset.sx) */
    pci_mem_space_pci_base = PCI_MEM_SPACE_PCI_BASE;
    pci_local_mem_pci_base = PCI_LOCAL_MEM_PCI_BASE;
    pci_io_space_pci_base = PCI_IO_SPACE_PCI_BASE;

    /* Note: we have to access the PCI memory window through the TLB,
       so we restrict the space to 128MB (8 * max 16MB pagesz) to
       avoid using up too many TLB entries (4 out of 16). The 128MB
       should be enough to map one hires VGA card plus space for other
       devices. If we needed to handle more than that we'd aither need
       to grab more wired TLB entries, or we'd need to install a TLB
       handler to handle the mapping dynamically. */
    
    /* set usable PCI mem region (leaving bottom 1M for ISA) */
    minpcimemaddr  = pci_mem_space_pci_base + 0x100000;
    nextpcimemaddr = pci_mem_space_pci_base + PCI_MEM_APERTURE_SIZE;

    /* set usable PCI i/o region (leaving bottom 64K for ISA) */
    minpciioaddr  = pci_io_space_pci_base + 0x10000;
    nextpciioaddr = pci_io_space_pci_base + PCI_IO_SPACE_SIZE;

    /* setup the TLB entries to map usable PCI mem space into KUSEG */
    for (idx = 0, addr = PCI_MEM_SPACE;
	 addr < PCI_MEM_SPACE + PCI_MEM_APERTURE_SIZE;
	 idx++, addr += VMPGSIZEMAX * 2) {
	unsigned long vpn = vaToVpn (addr);
	unsigned long ppn0 = paToPn (addr);
	unsigned long ppn1 = paToPn (addr + VMPGSIZEMAX);
	mips_tlbwi2 (mktlbhi (vpn, 0),
		     mktlblo (ppn0, TLB_V | TLB_D | TLB_G | TLB_UNCACHED),
		     mktlblo (ppn1, TLB_V | TLB_D | TLB_G | TLB_UNCACHED),
		     (VMPGSIZEMAX * 2 - 1) & TLBHI_VPN2MASK, 
		     idx);
    }

    /* preserve from random replacement */
    if (mips_getwired () < idx)
	mips_setwired (idx);

    if (pcimaster) {
	/* set "target not ready", so other PCI masters can't access us */
	RC33X_PCI_ARB |= RC33X_PCI_ARB_NOT_RDY; 
	mips_wbflush ();

	/* enable own PCI master and target functions (but not target i/o) */
	CONFIG_BIS (PCI_COMMAND_STATUS_REG, (PCI_COMMAND_SERR_ENABLE | 
					     PCI_COMMAND_PARITY_ENABLE |
					     PCI_COMMAND_INVALIDATE_ENABLE |
					     PCI_COMMAND_MASTER_ENABLE | 
					     PCI_COMMAND_MEM_ENABLE));
    }

    pci_cfg_unlock ();
    return pcimaster;
}


int
_pci_cacheline_log2 (void)
{
    return 2;    		/* 4 words == 2^2 */
}


int
_pci_maxburst_log2 (void)
{
    return 2;			/* max 4 words == 2^2 */
}


/*
 * Called to reinitialise the bridge after we've scanned each PCI device
 * and know what is possible.
 */
void
_pci_hwreinit (void)
{
    pci_cfg_lock ();

    if (_pci_bus[0].fast_b2b)
	/* fast back-to-back is supported by all devices */
	CONFIG_BIS (PCI_COMMAND_STATUS_REG, PCI_COMMAND_BACKTOBACK_ENABLE);
    
    if (_pci_bus[0].freq66) {
	/* if every device supported 66MHz we could enable it here */
	/* XXX except there appears to be no way to do this in software */
    }

    /* set latency timer */
    CONFIG_SEL (PCI_MISC_REG);
    PCI_MISC_LTIM_SET (*config_data, _pci_bus[0].def_ltim);

    /* clear "target not ready", we can now respond to other masters */
    RC33X_PCI_ARB &= ~RC33X_PCI_ARB_NOT_RDY; 
    mips_wbflush ();

    pci_cfg_unlock ();
}


/* Flush any read-ahead buffers */
void
_pci_flush (void)
{
    /* can't do this */
}


/* Flush any write-behind buffers */
void
_pci_wbflush (void)
{
    /* can't do this */
}


/* Map the CPU virtual address of an area of local memory to a PCI
   address that can be used by a PCI bus master to access it. */
vm_offset_t
_pci_dmamap (vm_offset_t va, unsigned int len)
{
    return pci_local_mem_pci_base + KVA_TO_PA (va);
}


/* Map the PCI address of an area of local memory to a CPU physical address. */
vm_offset_t
_pci_cpumap (vm_offset_t pcia, unsigned int len)
{
    return pcia - pci_local_mem_pci_base;
}


pcitag_t
_pci_make_tag(int bus, int device, int function)
{
    return ((bus << RC33X_PCI_CONF_BUS_SHFT) 
	    | (device << RC33X_PCI_CONF_DEV_SHFT) 
	    | (function << RC33X_PCI_CONF_DEV_SHFT));
}

void __inline__
_pci_break_tag(pcitag_t tag, int *busp, int *devicep, int *functionp)
{
    if (busp)
	*busp = (tag & RC33X_PCI_CONF_BUS) >> RC33X_PCI_CONF_BUS_SHFT;
    if (devicep)
	*devicep = (tag & RC33X_PCI_CONF_DEV) >> RC33X_PCI_CONF_DEV_SHFT;
    if (functionp) 
	*functionp = (tag & RC33X_PCI_CONF_FUNC) >> RC33X_PCI_CONF_FUNC_SHFT;
}


int
_pci_canscan(pcitag_t tag)
{
    int bus, dev;
    _pci_break_tag (tag, &bus, &dev, NULL);
    return (bus == 0 && (pcireserved & (1 << dev))) ? 0 : 1;
}


pcireg_t
_pci_conf_read(pcitag_t tag, int reg)
{
    unsigned int old_berrc;
    pcireg_t status, data;

    if (_pciverbose >= 3)
	printf ("pci_conf_read: tag=%x, r=%x\n", tag, reg);

    if (reg & 3 || reg < 0 || reg >= 0x100) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: bad reg %x\r\n", reg);
	return ~0;
    }

    pci_cfg_lock ();

    /* disable bus error exception */
    old_berrc = RC33X_BERRC;
    RC33X_BERRC = old_berrc | RC33X_BERRC_EXC_DIS; mips_wbflush ();

    /* clear aborts */
    CONFIG_BIS (SELF_TAG + PCI_COMMAND_STATUS_REG,
		PCI_STATUS_MASTER_ABORT | PCI_STATUS_MASTER_TARGET_ABORT);

    /* read register */
    data = CONFIG_GET (tag | reg);

    /* restore bus error exception enable and clear error bits */
    RC33X_BERRC = old_berrc & ~(RC33X_BERRC_CPU_SRC | RC33X_BERRC_IP_SRC);
    mips_wbflush ();

    /* read status */
    status = CONFIG_GET (SELF_TAG + PCI_COMMAND_STATUS_REG);

    /* reset any errors */
    CONFIG_SET (SELF_TAG + PCI_COMMAND_STATUS_REG, status);

    pci_cfg_unlock ();

    if (_pciverbose >= 1) {
	if (status & PCI_STATUS_MASTER_TARGET_ABORT )
	    _pci_tagprintf (tag, "_pci_conf_read: target abort\r\n");
    }
    return data;
}


void
_pci_conf_write(pcitag_t tag, int reg, pcireg_t data)
{
    unsigned int old_berrc;
    pcireg_t status;

    if (_pciverbose >= 3)
	printf ("pci_conf_write: tag=%x, r=%x d=%x\n", tag, reg, data);

    if (reg & 3 || reg < 0 || reg >= 0x100) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: bad reg %x\r\n", reg);
	return;
    }

    pci_cfg_lock ();

    /* disable bus error exception */
    old_berrc = RC33X_BERRC;
    RC33X_BERRC = old_berrc | RC33X_BERRC_EXC_DIS; mips_wbflush ();    

    /* clear aborts */
    CONFIG_BIS (SELF_TAG + PCI_COMMAND_STATUS_REG,
		PCI_STATUS_MASTER_ABORT | PCI_STATUS_MASTER_TARGET_ABORT);

    /* write register */
    CONFIG_SET (tag | reg, data);

    /* restore bus error exception enable and clear error bits */
    RC33X_BERRC = old_berrc & ~(RC33X_BERRC_CPU_SRC | RC33X_BERRC_IP_SRC);
    mips_wbflush ();

    /* get status */
    status = CONFIG_GET (SELF_TAG + PCI_COMMAND_STATUS_REG);

    /* reset any errors */
    CONFIG_SET (SELF_TAG + PCI_COMMAND_STATUS_REG, status);

    pci_cfg_unlock ();

    if (_pciverbose >= 1) {
	if (status & PCI_STATUS_MASTER_ABORT )
	    _pci_tagprintf (tag, "_pci_conf_write: master abort\r\n");
	if (status & PCI_STATUS_MASTER_TARGET_ABORT )
	    _pci_tagprintf (tag, "_pci_conf_write: target abort\r\n");
    }
}


pcireg_t
_pci_statusread (void)
{
    pcireg_t status;

    pci_cfg_lock ();
    status = CONFIG_GET (SELF_TAG + PCI_COMMAND_STATUS_REG) & 0xffff0000;
    pci_cfg_unlock ();
    return status;
}


void
_pci_statuswrite (pcireg_t val)
{
    pci_cfg_lock ();
    CONFIG_SEL (SELF_TAG + PCI_COMMAND_STATUS_REG);
    *config_data = (*config_data & ~0xffff0000) | (val & 0xffff0000);
    pci_cfg_unlock ();
}


int
_pci_map_io(pcitag_t tag, int reg, vm_offset_t *vap, vm_offset_t *pap)
{
    pcireg_t address;
    vm_offset_t pa;

    if (reg < PCI_MAP_REG_START || reg >= PCI_MAP_REG_END || (reg & 3)) {
	if (_pciverbose >= 1)
	    _pci_tagprintf(tag, "_pci_map_io: bad request\r\n");
	return -1;
    }

    address = _pci_conf_read (tag, reg);

    if ((address & PCI_MAP_IO) == 0) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_io: attempt to memory map a memory region\r\n");
	return -1;
    }

    pa = (address & PCI_MAP_IO_ADDRESS_MASK) - pci_io_space_pci_base;
    *pap = pa;

    *vap = (vm_offset_t) PA_TO_KVA1 (PCI_IO_SPACE + pa);

    if (_pciverbose >= 3)
	_pci_tagprintf (tag, "_pci_map_io: mapping io at virtual 0x%x, physical 0x%x\r\n", *vap, *pap);

    return 0;
}


int
_pci_map_mem(pcitag_t tag, int reg, vm_offset_t *vap, vm_offset_t *pap)
{
    pcireg_t address;
    vm_offset_t pa;

    if (reg == PCI_MAP_REG_ROM) {
	/* expansion ROM */
	address = _pci_conf_read(tag, reg);
	if (!(address & PCI_MAP_ROM)) {
	    _pci_tagprintf (tag, "_pci_map_mem: attempt to map missing rom\r\n");
	    return -1;
	}
	pa = address & PCI_MAP_ROM_ADDRESS_MASK;
    }
    else {
	if (reg < PCI_MAP_REG_START || reg >= PCI_MAP_REG_END || (reg & 3)) {
	    if (_pciverbose >= 1)
		_pci_tagprintf (tag, "_pci_map_mem: bad request\r\n");
	    return -1;
	}
	
	address = _pci_conf_read(tag, reg);

	if ((address & PCI_MAP_IO) != 0) {
	    if (_pciverbose >= 1)
		_pci_tagprintf (tag, "_pci_map_mem: attempt to memory map an I/O region\r\n");
	    return -1;
	}
	
	switch (address & PCI_MAP_MEMORY_TYPE_MASK) {
	case PCI_MAP_MEMORY_TYPE_32BIT:
	case PCI_MAP_MEMORY_TYPE_32BIT_1M:
	    break;
	case PCI_MAP_MEMORY_TYPE_64BIT:
	    if (_pciverbose >= 1)
		_pci_tagprintf (tag, "_pci_map_mem: attempt to map 64-bit region\r\n");
	    return -1;
	default:
	    if (_pciverbose >= 1)
		_pci_tagprintf (tag, "_pci_map_mem: reserved mapping type\r\n");
	    return -1;
	}
	pa = address & PCI_MAP_MEMORY_ADDRESS_MASK;
    }

    pa -= pci_mem_space_pci_base;
    *pap = pa;

    /* only uncached access available... */
    /* note pci_mem_space is a virtual address in kuseg */
    *vap = (vm_offset_t) PCI_MEM_SPACE + pa;

    if (_pciverbose >= 3)
	_pci_tagprintf (tag, "_pci_map_mem: mapping memory at virtual 0x%x, physical 0x%x\r\n", *vap, *pap);

    return 0;
}


/* Return interrupt number */
int
_pci_map_int (pcitag_t tag)
{
#if 0
    static const int sbdintr[4] = {INTR_PCI_A, INTR_PCI_B,
				   INTR_PCI_C, INTR_PCI_C};
    int bus, dev;
#endif
    pcireg_t data;
    int pin;

    /* If we're in satellite mode then we can't receive any PCI interrupts,
       only generate them! */
    if (!pcimaster)
	return -1;

    data = _pci_conf_read (tag, PCI_INTERRUPT_REG);
    pin = PCI_INTERRUPT_PIN (data);
    if (pin == 0)
	return 0;	/* No interrupt used. */

    if (--pin > 3) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_int: bad pin %d\r\n", pin);
	return -1;
    }

#if 1
    /* Although PCI interrupt lines INTA-INTD are fed to CPU interrupt
       inputs INT1,2,4 & 5, there are problems:
       1) All four are ORed together onto CPU INT1, and there
          is no separate mask register to control this, so
	  we can only mask all or none on this bit. 
       2) On the S332 board CPU INT2-5 are not wired, so 
          on that board we have to use INT1. 
       The answer is that we pretend that all PCI cards interrupt on 
       PCI INTA, and rely on drivers to support PCI interrupt sharing. */
    return INTR_PCI_A;
#else
    /* ascend bus hierarchy until we get back to our (host) controller */
    _pci_break_tag (tag, &bus, &dev, NULL);
    while (bus != 0) {
	/* XXX assume interrupt pins just pass straight through */
	tag = _pci_bus[bus].tag;
	_pci_break_tag (tag, &bus, &dev, NULL);
    }

    /* On new S334A/332 interrupt wiring is rotated, but the 5V/3.3V
       connectors are wired identically, so you can't have four cards! */
    switch (dev) {
    case 2:
	/* IDSEL AD12: dev INTA -> board INTA */
	return sbdintr[pin];
    case 4:
	/* IDSEL AD14: dev INTA -> board INTB */
	return sbdintr[(pin + 1) % 4];
    }
    return -1;
#endif
}


pcireg_t
_pci_allocate_mem (pcitag_t tag, vm_size_t size)
{
    pcireg_t address;

    /* allocate downwards, then round to size boundary */
    address = (nextpcimemaddr - size) & ~(size - 1);
    if (address > nextpcimemaddr || address < minpcimemaddr)
	return -1;
    nextpcimemaddr = address;
    return address;
}


pcireg_t
_pci_allocate_io (pcitag_t tag, vm_size_t size)
{
    pcireg_t address;

    /* allocate downwards, then round to size boundary */
    address = (nextpciioaddr - size) & ~(size - 1);
    if (address > nextpciioaddr || address < minpciioaddr)
	return -1;
    nextpciioaddr = address;
    return address;
}


/* 
 * Handle mapping of fixed ISA addresses 
 */

void *
_isa_map_io (unsigned int port)
{
    if (pci_io_space_pci_base != 0
	|| port > 0xffff)
	return 0;
    /* must disable bus error exception if app is accesing random
       locations in ISA space */
    RC33X_BERRC |= RC33X_BERRC_EXC_DIS; mips_wbflush ();
    return PA_TO_KVA1 (PCI_IO_SPACE + port);
}

void *
_isa_map_mem (vm_offset_t addr)
{
    if (pci_mem_space_pci_base != 0
	|| pci_local_mem_pci_base < 0x100000 
	|| addr > 0xfffff)
	return 0;
    /* must disable bus error exception if app is accesing random
       locations in ISA space */
    RC33X_BERRC |= RC33X_BERRC_EXC_DIS; mips_wbflush ();
    /* note: virtual address */
    return (void *)(PCI_MEM_SPACE + addr);
}
