#ident "kit/idt33x/sbdrtpx.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * sbdrtpx.c: low-level interrupt and timer handling for IDT 79S33x board
 */


#include <sys/types.h>
#include <mips/cpu.h>
#include <mips/prid.h>
#include <mips/xcpt.h>
#include "kit_impl.h"

#include <idt/rc3233x.h>
#include "sbd.h"

#include "pthread_impl.h"

/* 
 * Interrupt wiring under MIPS RTPX:
 *  HINT5 - cpu timer / pci intd
 *  HINT4 - pci intc
 *  HINT3 - rc3233x controller
 *  HINT2 - pci intb
 *  HINT1 - pci inta
 *  HINT0 - ext uart
 *  SINT1 - unused/reserved
 *  SINT0 - unused/reserved
 */

/*
 * Note that for this implementation the ipl0 interrupt mask is
 * a structure containing multiple levels of mask, the first
 * word is the cpu status register, the remaining words are
 * equivalent to the multiple masks in the RC3233X interrupt
 * controller.
 */

/* the interrupt mask when at "user" level (ipl0) */
_sbd_intrmask_t		_sbd_ipl0_mask;

/* the CPU status register when at scheduler level (ipl1) */
unsigned int 		_sbd_ipl1_sr;

/* the CPU status register when at interrupt blocking level (iplhi) */
unsigned int		_sbd_iplhi_sr;

/* the interrupt mask with which we were entered (e.g. from IDT/sim),
   see .../kit/pmon/pmonlow.sx */
unsigned int 		_sr_base __attribute__((common));

/* flag to record whether in kernel mode (ipl >= 1) */
int 			_sbd_kernelmode;

/* tell IDT/sim that we handle all interrupts,
   see .../kit/idtsim/simxcpt.sx */
unsigned 		_intrmask = SR_IMASK;

/* table mapping cause register bits to interrupt numbers */
static signed char	crToIntr[256];

/* table of which interrupts are "pulses", and which are "levels" */
static unsigned short	intrpulse[16];

#if !IDT33X_HINT5_TIMER
/* emulate 32-bit count/compare without reset to zero */
static unsigned int	tickbase;
static unsigned int	tickcmp;
#endif

/* 
 * Return number of timer ticks per second
 */
unsigned int
_sbd_tickspersec (void)
{
#if IDT33X_HINT5_TIMER
    /* R4x00 internal counter runs at half pipeline frequency */
    return _sbd_pipefreq () / 2;
#else
    /* RC3233x timer runs at system clock frequency */
    return _sbd_cpufreq ();
#endif
}


/*
 * Return current value of timer (upcount)
 */
unsigned int
_sbd_gettickval (void)
{
#if IDT33X_HINT5_TIMER
    return mips_getcount();
#else
    /* get offset from last interrupt */
    unsigned int ticks;

    /* check for as yet unhandled rollover interrupt (but don't clear it) */
    ticks = RC33X_TMR_COUNT (RC33X_GP0_TMR);
    if (RC33X_INT_PEND (RC33X_INTGRP_TIMER) & (1 << RC33X_GP0_TMR)) {
	/* reread counter in case we just rolled over after first read */
	ticks = RC33X_TMR_COUNT (RC33X_GP0_TMR);
	ticks += tickcmp;
    }

    /* compute 32-bit monotonically increasing count */
    return tickbase + ticks;
#endif
}


/*
 * Set value at which timer should generate interrupt, and
 * clear any pending timer interrupt.
 */
void
_sbd_settickcmp (unsigned int newcmp)
{
#if IDT33X_HINT5_TIMER
    /* set next tick time, and clear interrupt */
    mips_setcompare (newcmp);
#else
    int oticks = RC33X_TMR_COUNT (RC33X_GP0_TMR);
    int delta;
    
    if (RC33X_INT_PEND (RC33X_INTGRP_TIMER) & (1 << RC33X_GP0_TMR)) {
again:
	oticks = RC33X_TMR_COUNT (RC33X_GP0_TMR);
	RC33X_INT_CLEAR (RC33X_INTGRP_TIMER) = 1 << RC33X_GP0_TMR;
	RC33X_INT_CLEAR (RC33X_INTGRP_TOP) = 1 << RC33X_INTGRP_TIMER;
	/* incorporate interrupt into 32-bit accumulator and clear interrupt */
	tickbase += tickcmp;
    }

    /* clock resets to zero every i/u, compare value must be relative 
       to the emulated 32-bit value at time of last reset. */
    delta = newcmp - tickbase;
    if (delta <= 0)
	/* already gone past new compare value, prevent spurious i/u by
	   setting very high compare value (several seconds) */
	delta = INT_MAX;
    else if (delta < 5000)
	/* don't allow too small a delta, or else we can't guarantee
	   to be able to service a timer interrupt before it rolls-over
	   again. */ 
	/* XXX this restricts the accuracy of the timer to approx 100us */
	delta = 5000;

    if (tickcmp != delta) {
	/* set new compare value (leave counter running) */
	RC33X_TMR_COMPARE (RC33X_GP0_TMR) = delta;
	/* check for late interrupt based on old compare value */
	if (RC33X_INT_PEND (RC33X_INTGRP_TIMER) & (1 << RC33X_GP0_TMR))
	    goto again;
	tickcmp = delta;
    }
#endif

    /* Note that the counter might actually be beyond the new compare
       value, and therefore not cause an interrupt until it wraps
       around (many seconds). This is OK though, because the scheduler
       rereads the counter after setting the compare value, and checks
       for this occurrence. */
}


int 
_sbd_intr_howmany ()
{
    return INTR_MAX + 1;
}


/* 
 * Poll specific interrupt, returns 0 if interrupt is not active.
 * Called at ipl1.
 */
int 
_sbd_intrpoll (int intr)
{
    int group = INTR_GROUP (intr);
    unsigned int subintmask = 1 << INTR_SUBINT (intr);
    unsigned int ret;

    if (group == 0) {
	/* check internal cpu cause register */
	if (intr <= INTR_SINT1) {
	    /* read and clear software interrupts (atomically) */
	    mips_setsr (_sbd_iplhi_sr);
	    ret = mips_biccr (1 << intr);
	    mips_setsr (_sbd_ipl1_sr);
	}
	else {
	    /* read hardware interrupts */
	    ret = mips_getcr();
	}
	ret &= subintmask;
    }
    else {
	int pulse = intrpulse[group] & subintmask;

	/* if it's a level interrupt, resample it now */
	if (!pulse) {
	    RC33X_INT_CLEAR (group) = subintmask;
	    RC33X_INT_CLEAR (RC33X_INTGRP_TOP) = 1 << group;
	}

	/* check expansion interrupt pending bits */
	ret = RC33X_INT_PEND (group) & subintmask;

	/* if it's a pulse interrupt clear it when we detect it
	   active, so that we catch any following pulses */
	if (pulse && ret) {
	    RC33X_INT_CLEAR (group) = subintmask;
	    RC33X_INT_CLEAR (RC33X_INTGRP_TOP) = 1 << group;
	}
    }

    return ret != 0;
}


/*
 * Enable ipl0 interrupts which have been blocked, in particular
 * when interrupt handler first installed.
 * Called at ipl1.
 */
void 
_sbd_intrenable (int intr)
{
    int group = INTR_GROUP (intr);
    unsigned int subintmask = 1 << INTR_SUBINT (intr);

    if (group == 0) {
	/* enabled at ipl0; must therefore be disabled at ipl1 */
	mips_atomic_bis (&_sbd_ipl0_mask.sr, subintmask);
	mips_atomic_bic (&_sbd_ipl1_sr, subintmask);
    }
    else {
	/* clear h/w interrupt(s) latch now if this is a "level" interrupt;
	   we don't want to clear a "pulse" interrupt, since we could lose
	   a pulse that occurred between the last intrpoll and now. */
	if (! (intrpulse[group] & subintmask)) {
	    RC33X_INT_CLEAR (group) = subintmask;
	    RC33X_INT_CLEAR (RC33X_INTGRP_TOP) = 1 << group;
	}

	/* enable interrupt */
	_sbd_ipl0_mask.grp[RC33X_INTGRP_TOP] |= (1 << group);
	_sbd_ipl0_mask.grp[group] |= subintmask;
	mips_atomic_bis (&RC33X_INT_MASK (group), subintmask);
    }
}


/* 
 * Take a guess at which ipl this CP0 status register represents. If
 * in doubt return 0. 
 */
int
_sbd_probable_ipl (unsigned int sr)
{
#if IDT33X_HINT5_TIMER
    return !(sr & SR_IE) ? 2 : !(sr & SR_HINT5) ? 1 : 0;
#else
    return !(sr & SR_IE) ? 2 : _sbd_kernelmode ? 1 : 0;
#endif
}


/* 
 * Return next pending interrupt for this cpu, or zero if none.
 * If idle != 0 then we can spin waiting for an interrupt 
 * (possibly going into low-power mode) 
 * Called at ipl1.
 */
int 
_sbd_intrnext (int idle)
{
    int group = 0;
    int intr;
    
    /* switch to "hi" interrupt priority (all interrupts disabled) */
    mips_setsr (_sbd_iplhi_sr);
    
    while (1) {
	unsigned int pend, grppend;
	
	/* find interrupts which we handle that would be pending at ipl0 */
	pend = mips_getcr() & _sbd_ipl0_mask.sr;
	intr = crToIntr[pend >> 8];
	
	/* pick apart expansion interrupts */
	if (intr == INTR_RC3233X_MUX) {
	    /* reload top-level pending register */
	    RC33X_INT_CLEAR (RC33X_INTGRP_TOP) = ~0; mips_wbflush ();
	    grppend = RC33X_INT_PEND (RC33X_INTGRP_TOP);

#if !IDT33X_HINT5_TIMER
	    /* quick check for GP0 timer */
	    if ((grppend & (1 << RC33X_INTGRP_TIMER)) &&
		(RC33X_INT_PEND (RC33X_INTGRP_TIMER) & (1 << RC33X_GP0_TMR))) {
		intr = INTR_CLOCK;
		break;
	    }
#endif

	    intr = 0;
	    grppend &= _sbd_ipl0_mask.grp[RC33X_INTGRP_TOP];
	    while (grppend != 0) {
		group = _mips_clz (grppend) ^ 31;
		grppend &= ~(1 << group);
		pend = RC33X_INT_PEND (group) & _sbd_ipl0_mask.grp[group];
		if (pend != 0) {
		    intr = (group << 4) | (_mips_clz (pend) ^ 31);
		    break;
		}
	    }
	}

	if (intr != 0 || !idle)
	    break;
      
	mips_setsr (_sbd_ipl1_sr);
	/* briefly return to ipl1 to allow any non-RTPX interrupts 
	   to be serviced while we are idle and polling. */
	mips_setsr (_sbd_iplhi_sr);
    }

    if (intr > 0) {
	/* disable the detected interrupt at ipl0 (and greater); 
	   note that clock i/u (intr == -1) is never disabled. */
	unsigned int subintmask = 1 << INTR_SUBINT (intr);
	if (intr < 16) {
	    /* disable cpu interrupt at ipl0 */
	    _sbd_ipl0_mask.sr &= ~subintmask;

	    /* clear software interrupts now */
	    /* (non-atomic but safe at iplhi) */
	    if (intr <= INTR_SINT1)
		(void) mips_biccr (subintmask);
	}
	else {
	    /* disable expansion interrupt s/w mask at ipl0 */
	    if ((_sbd_ipl0_mask.grp[group] &= ~subintmask) == 0)
		_sbd_ipl0_mask.grp[RC33X_INTGRP_TOP] &= ~(1 << group);

	    /* disable in h/w mask too */
	    /* (non-atomic but safe at iplhi: all i/us are disabled) */
	    RC33X_INT_MASK (group) &= ~subintmask;

	    /* clear h/w interrupt latches in case this is a "pulse";
	       we want to latch any following pulses, and it won't
	       hurt for a level interrupt which will be immediately
	       relatched. */
	    RC33X_INT_CLEAR (group) = subintmask;
	    RC33X_INT_CLEAR (RC33X_INTGRP_TOP) = 1 << group;

	    mips_wbflush ();
	}
    }

    /* return to ipl1 */
    mips_setsr (_sbd_ipl1_sr);
    return intr;
}


/* 
 * Board-specific interrupt handler, return 1 if interrupt handled.
 * Called at iplhi.
 */
int 
_sbd_intrhandler (unsigned int pend, struct xcptcontext *xcp)
{
    /* everything is passed to the scheduler */
    return 0;
}


int
_sbd_intrtype (int intr, const intrtype_t *new, intrtype_t *old)
{
    int group = INTR_GROUP (intr);
    unsigned int subintmask = 1 << INTR_SUBINT (intr);

    if (old) {
	if (intrpulse[group] & subintmask)
	    old->mode = (group == RC33X_INTGRP_PIOLOW)
		? INTRTYPE_PULSE_LO : INTRTYPE_PULSE_HI;
	else
	    old->mode = (group == RC33X_INTGRP_PIOLOW)
		? INTRTYPE_LEVEL_LO : INTRTYPE_LEVEL_HI;
    }
	
    /* note that we can't change the active polarity, user must
       just get this right. */
    if (new) {
	switch (new->mode) {
	case INTRTYPE_LEVEL_LO:
	case INTRTYPE_LEVEL_HI:
	    intrpulse[group] &= ~subintmask;
	    break;
	case INTRTYPE_PULSE_LO:
	case INTRTYPE_PULSE_HI:
	    intrpulse[group] |= subintmask;
	    break;
	case INTRTYPE_EDGE_FALLING:
	case INTRTYPE_EDGE_RISING:
	default:
	    break;
	}
    }

    return 0;
}


/* 
 * Initialise interrupt handling for this board 
 * Called at undefined ipl, should return at iplhi
 */
void 
_sbd_intrinit(void)
{
    unsigned int topmask;
    int i;

    /* disable all interrupts at interrupt level and set cpu to iplhi */
    _sbd_iplhi_sr = _sr_base & ~(SR_CU1 | SR_IMASK | SR_IE);
    mips_setsr (_sbd_iplhi_sr);

    /* disable clock, pci & expansion interrupt groups in scheduler */
    _sbd_ipl1_sr = (_sr_base & ~SR_IMASK) | SR_IE;

    /* enable clock & expansion interrupt groups at user level */
#if IDT33X_HINT5_TIMER
    _sbd_ipl0_mask.sr = SR_HINT5 | SR_HINT3;
#else
    _sbd_ipl0_mask.sr = SR_HINT3;
#endif

    /* disable and clear all low level expansion interrupts */
    topmask = 0;
    for (i = RC33X_INTGRP_BUSERR; i <= RC33X_INTGRP_SPI; i++) {
	topmask |= 1 << i;
	RC33X_INT_CLEAR (i) = ~0;
	RC33X_INT_MASK (i) = 0;
	_sbd_ipl0_mask.grp[i] = 0;
    }

    /* indicate interrupts which are pulses that must be latched,
       rather than levels. */
    intrpulse[RC33X_INTGRP_BUSERR]	= ~0;
    intrpulse[RC33X_INTGRP_TIMER] 	= ~0;
    intrpulse[RC33X_INTGRP_PCIERR] 	= ~0;

    /* clear then enable top-level interrupt group */
    RC33X_INT_CLEAR (RC33X_INTGRP_TOP) = topmask;
    RC33X_INT_MASK (RC33X_INTGRP_TOP) = topmask;
    _sbd_ipl0_mask.grp[RC33X_INTGRP_TOP] = 0;	/* but s/w disabled */

    /* build Cause bits to interrupt number mapping table */
    for (i = 0; i < 256; i++) {
	unsigned int cr = i << 8;
	int intr;
	/* test in "priority" order */
#if IDT33X_HINT5_TIMER
	if (cr & CR_HINT5)
	    intr = INTR_CLOCK;
	else
#endif
	if (cr & CR_HINT3)
	    intr = INTR_RC3233X_MUX;
#if !IDT33X_HINT5_TIMER
	else if (cr & CR_HINT5)
	    intr = INTR_PCI_D;
#endif
	else if (cr & CR_HINT4)
	    intr = INTR_PCI_C;
	else if (cr & CR_HINT2)
	    intr = INTR_PCI_B;
	else if (cr & CR_HINT1)
	    intr = INTR_PCI_A;
	else if (cr & CR_HINT0)
	    intr = INTR_DUART;
	else if (cr & CR_SINT1)
	    intr = INTR_SINT1;
	else if (cr & CR_SINT0)
	    intr = INTR_SINT0;
	else
	    intr = 0;
	crToIntr[i] = intr;
    }

#if IDT33X_HINT5_TIMER
    /* set a "safe" compare value and clear timer interrupt */
    mips_setcompare (mips_getcount() + INT_MAX);
#else
    /* start up GP0 timer; clear and enable interrupt */
    RC33X_TMR_CTRL (RC33X_GP0_TMR) = 0;
    RC33X_TMR_COUNT (RC33X_GP0_TMR) = tickbase = 0;
    RC33X_TMR_COMPARE (RC33X_GP0_TMR) = tickcmp = INT_MAX;
    RC33X_TMR_CTRL (RC33X_GP0_TMR) = RC33X_TMR_CTRL_ENB;
    RC33X_INT_CLEAR (RC33X_INTGRP_TIMER) = 1 << RC33X_GP0_TMR;
    RC33X_INT_MASK (RC33X_INTGRP_TIMER) = 1 << RC33X_GP0_TMR;
    mips_wbflush ();
#endif
}
