#ident "kit/idtsim/simrfile.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * idtsim/simrfile.c: Remote file i/o under IDT/sim
 */


#ifdef RFILE

#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sdethread.h>
#include <assert.h>
#include <alloca.h>
#include "file.h"

extern __sdethread_mutex_t _sim_lock;

static int	_rfilerw (struct file *, struct uio *);
static int	_rfileioctl (struct file *, int, void *);
static int	_rfileselect (struct file *, int);
static int	_rfileopen (struct file *, dev_t, int);
static int	_rfileclose (struct file *);
static int*	_rfilelookup (dir_t **, dir_t *, const char **, int, mode_t);
static void	_rfiledfree (dir_t *);

const struct fileops _rfile_ops = 
{
    _rfilerw, 
    _rfilerw, 
    _rfileioctl, 
    _rfileselect, 
    _rfileopen,
    _rfileclose,
    _rfilelookup,
    _rfiledfree
};

struct simf {
    int		sf_fd;
    off_t	sf_offs;
};


static int
_rfilerw (struct file *fp, struct uio *uio)
{
    struct simf *sf = fp->f_data;
    int err = 0;

    if (!sf)
	/* root directory - can't read */
	return EACCES;

    __sdethread_mutex_lock (&_sim_lock);
    if (uio->uio_offset != sf->sf_offs) {
	if (sim_rlseek (sd->sd_fd, uio->uio_offset, SEEK_SET) < 0) {
	    err = ESPIPE;
	    goto done;
	}
    }
    
    while (uio->uio_resid != 0 && err == 0) {
	struct iovec *iov;
	err = uiogetiov (uio, &iov);
	if (err == 0) {
	    ssize_t res;
	    if (uio->uio_rw == UIO_WRITE)
		res = sim_rwrite (sf->sf_fd, iov->iov_base, iov->iov_len);
	    else
		res = sim_rread (sf->sf_fd, iov->iov_base, iov->iov_len);
	    if (res > 0)
		err = uioadvance (res, uio);
	    else if (res < 0 || uio->uio_rw == UIO_WRITE)
		err = EIO;
	    else
		break;
	}
    }

    sf->sf_offs = uio->uio_offset;

 done:
    __sdethread_mutex_unlock (&_sim_lock);
    return err;
}

static int
_rfileioctl (struct file *fp, int cmd, void *arg)
{
    return EINVAL;
}


static int
_rfileselect (struct file *fp, int rw)
{
    /* not used yet, but return always ready */
    return rw;
}



static const dir_t template =
DIRDESC(0, 'r'<<8, S_IFREG|S_IRWXU|S_IRWXG|S_IRWXO, 1024, _rfile_ops, -1);


static int
_rfileopen (struct file *fp, dev_t dev, int fflags)
{
    /* nothing to do */
    return 0;
}


static int
_rfileclose (struct file *fp)
{
    /* nothing to do */
    return 0;
}


static int
_rfilelookup (dir_t *ndpp, dir_t *dp, const char **name, int fflags, mode_t mode)
{
    static int inited = 0;
    const char *nm = *name;
    strict simf *sf;
    dir_t *dp;
    int fd;
    
    /* Handle "." and ".." magic names */
    if (nm[0] == '.') {
	if (nm[1] == '.' && (nm[2] == '\0' || nm[2] == '/')) {
	    /* This isn't going to work if dp->d_parent isn't valid */
	    *name += 2;
	    *ndpp = dp->d_parent;
	    return 0;
	}
	if (nm[1] == '\0' || nm[1] == '/') {
	    *name += 1;
	    *ndpp = dp;
	    return 0;
	}
    }

    __sdethread_mutex_lock (&_sim_lock);

    if (!inited) {
	char ttyname[5];
	strcpy (ttyname, "ttyx");
	ttyname[3] = (dp->d_dev & 0xff) + '0';
	sim_rfileinit (ttyname);
	inited = 1;
    }

    fd = sim_ropen (nm, OFLAGS (fflags));

    __sdethread_mutex_unlock (&_sim_lock);

    if (fd < 0)
	return ENOENT;

    dp = malloc (sizeof (dir_t) + sizeof (struct simf));
    if (!dp) {
	sim_close (fd);
	return ENOMEM;
    }

    *dp = template;
    dp->d_flags = DF_DYNAMIC;
    dp->d_count = 1;
    sf = dp->d_data = dp + 1;
    sf->sf_fd = fd;
    sf->sf_offs = 0;

    /* advance name pointer to end of path */
    *name += strlen (nm);
    *ndpp = dp;
    return 0;
}


static void
_rfiledfree (dir_t *dp)
{
    struct simf *sf = dp->d_data;
    if (sf) {
	__sdethread_mutex_lock (&_sim_lock);
	sim_close (sf->sf_fd);
	__sdethread_mutex_unlock (&_sim_lock);
    }
    if (dp->d_flags & DF_DYNAMIC)
	free (dp);
}

#endif /* RFILE */
