#ident "kit/idtsim/simtty.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * idtsim/simtty.c: Posix style terminal i/o under IDT/sim
 */


#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/init.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <assert.h>
#include "file.h"
#include "tty.h"

#include "kit_impl.h"
#include "sim.h"

static int	ttyread (struct file *, struct uio *);
static int	ttywrite (struct file *, struct uio *);
static int	ttyioctl (struct file *, int, void *);
static int	ttyselect (struct file *, int);
static int	ttyopen (struct file *, dev_t, int);
static int	ttyclose (struct file *);

static const struct fileops _tty_ops = 
{
    ttyread, 
    ttywrite, 
    ttyioctl, 
    ttyselect, 
    ttyopen,
    ttyclose,
};


#define NTTY 4
static struct tty *	ttys[NTTY];
static speed_t		ttyspeeds[NTTY];

/* XXX This table isn't guaranteed correct for all boards.  IDT/sim
   needs to define board/uart-independent speed codes! */

static const struct baudtab {
    int		cbaud;
    speed_t	speed;
} baudtab[] = {
#ifdef IDT134
    {0,		38400},
    {1,		19200},
    {2,		9600},
    {3, 	4800},
    {4,		2400},
    {5,		1200},
    {6,		600},
    {7,		300},
    {8,		150},
    {9,		110},
#else
    {0,		19200},
    {1,		9600},
    {2,		4800},
    {3,		2400},
    {4,		2000},
    {5,		1800},
    {6,		1200},
    {7,		600},
    {8,		300},
    {9,		150},
    {10,	110},
#endif
    {-1,	-1}
};



/*
 * The lowest level device-dependent routines, called by the generic
 * tty code.
 */

static int
ttyparams (struct tty *tp, int set, struct termios *tio)
{
    const struct baudtab *bt;

    if (set) {
	/* CHANGE SETTINGS */

	/* input and output speeds must match */
	if (tio->c_ispeed != tio->c_ospeed)
	    return EINVAL;

	/* XXX since we can't be really sure that the above table
	   is right for every IDT/sim driver we'll quietly
	   ignore this request if the baud rate is unchanged. */
	if (tio->c_ospeed == 0 || tio->c_ospeed == ttyspeeds[tp->t_dev])
	    return 0;

	/* lookup idt/sim code for this speed in table */
	for (bt = baudtab; bt->speed != tio->c_ospeed; bt++)
	    if (bt->speed == -1)
		return EINVAL;

	/* make the change */
	if (sim_ioctl ((int)tp->t_addr, SIM_TIOCBAUD, bt->cbaud) < 0)
	    return EINVAL;

	/* cache the current speed, which can't be read from IDT/sim */
	ttyspeeds[tp->t_dev] = tio->c_ospeed;
    }
    else {
	/* GET SETTINGS */

	/* can't fetch from idt/sim, so synthesise */
	tio->c_cflag = TTYDEF_CFLAG;
	tio->c_ispeed = tio->c_ospeed = ttyspeeds[tp->t_dev];
    }

    return 0;
}


static int
ttyrxpoll (struct tty *tp, int canblock)
{
    char c;
    int n;

    /* never blocks */
    n = sim_read ((int)tp->t_addr, &c, 1);
    if (n == 1)
	return c | 0x100;
    return 0;
}


static int
ttytxrdy (struct tty *tp)
{
    return 1;			/* we can't tell, assume always ready! */
}


static void
ttytx (struct tty *tp, int c)
{
    char cb = c;
    sim_write ((int)tp->t_addr, &cb, 1);
}

static int
ttytxblk (struct tty *tp, const char *bp, int nc)
{
    return sim_write ((int)tp->t_addr, bp, (nc > 16) ? 16 : nc);
}



static int
ttywrite (struct file *fp, struct uio *uio)
{
    /* generic write */
    return _ttywrite ((struct tty *)fp->f_data, uio, fp->f_flag);
}

static int
ttyread (struct file *fp, struct uio *uio)
{
    /* generic read */
    return _ttyread ((struct tty *)fp->f_data, uio, fp->f_flag);
}

static int
ttyioctl (struct file *fp, int cmd, void *arg)
{
    /* generic ioctl */
    return _ttyioctl ((struct tty *)fp->f_data, cmd, arg);
}

static int
ttyselect (struct file *fp, int rw)
{
    /* generic select */
    return _ttyselect ((struct tty *)fp->f_data, rw);
}

static int
ttyopen (struct file *fp, dev_t dev, int fflags)
{
    struct tty *tp;
    int err;

    if (dev >= NTTY)
	return ENXIO;

    /* initialise our structure */
    if (!(tp = ttys[dev])) {
	err = _ttymalloc ("tty(sim)", &tp);
	if (err)
	    return err;
	ttys[dev] = tp;
    }

    if (!(tp->t_state & TS_OPEN)) {
	char ttyname[5];
	int fd;

	/* XXX assumes "conventional" naming, e.g. tty0..ttyN */
	strcpy (ttyname, "ttyx");
	ttyname[3] = dev + '0';
	fd = sim_open (ttyname, OFLAGS(fflags));
	if (fd < 0)
	    return ENXIO;

	/* put SIM into raw, non-blocking mode */
	(void) sim_ioctl (fd, SIM_TIOCRAW, 1);
	(void) sim_ioctl (fd, SIM_FIOCNBLOCK, 1);

	tp->t_dev	= dev;
	tp->t_params	= ttyparams;
	tp->t_rxpoll	= ttyrxpoll;
	tp->t_txrdy	= ttytxrdy;
	tp->t_txblk	= ttytxblk;
	tp->t_tx	= ttytx;
	tp->t_addr	= (void *) fd;

	/* set the current speed, it can't be read from IDT/sim, so
	   this is just a guess! */
	ttyspeeds[dev] = TTYDEF_SPEED;
    }

    /* point file at our private data */
    fp->f_data = tp;

    /* generic open */
    return _ttyopen (tp);
}


static int
ttyclose (struct file *fp)
{
    struct tty *tp = (struct tty *) fp->f_data;
    int error = 0;

    if (tp) {
	error = _ttyclose (tp);
	if (!(tp->t_state & TS_OPEN))
	    sim_close ((int)tp->t_addr);
    }
    return error;
}



/* /dev/tty*:	crw-rw-r-- */
#define TTYMODE	(S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IWOTH)

static void
ttydev_mount (void)
{
    dir_t ttynode;
    int err;

    memset (&ttynode, 0, sizeof (ttynode));
    ttynode.d_mode = TTYMODE;
    ttynode.d_ops = &_tty_ops;

#define ADDDESC(name, dev)				\
    ttynode.d_name = name;				\
    ttynode.d_dev = dev;				\
    err = _mknod ("/dev", &ttynode);			\
    if (err && err != EEXIST) {				\
      _mon_puts ("mknod(/dev/" name ") failed\n");	\
      abort ();						\
    }

    /* add console port (must exist) */
    ADDDESC("console",	0);
    ADDDESC("tty",	0);
    ADDDESC("tty0",	0);

    /* add auxiliary ports */
#if 0
    /* XXX sorry, this won't work: accessing an unknown idt/sim device 
       causes an error message to be output to the console! */
    int d;
    for (d = 1; ; d++) {
	static char name[] = "ttyx";
	int fd;
	name[3] = d + '0';
	if ((fd = sim_open (name, O_RDWR)) < 0)
	    break;
	sim_close (fd);
	ttynode.d_name = name;
	ttynode.d_dev = d;
	if (_mknod ("/dev", &ttynode))
	    break;
    }
#else
    /* assume max of 4 ports */
    ADDDESC("tty1",	1);
    ADDDESC("tty2",	2);
    ADDDESC("tty3",	3);
#endif
}

DECL_INIT_FUNC(IF_MOUNT, ttydev_mount)

/* call this dummy function to load this module */
void _ttydev_load (void);
void _ttydev_load (void) {} 
