#ident "kit/malta/sbdpci-msc.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2003-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * malta/sbdpci-msc.c: Machine-specific functions for PCI autoconfiguration (SOC-IT)
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <mips/xcpt.h>
#include <mips/cpu.h>
#include <pci/pcivar.h>
#include <pci/pcireg.h>
#include "kit_impl.h"

#define printf _mon_printf

#include "sbd.h"
#include "sbdicu.h"
#include "bonito.h"

#include "sbdpci-table.h"
#include <sys/init.h>

#include "msc01_memmap.h"
#include "ssc_memmap.h"
#include "msc01_pci.h"

/* SOC-it SC on all malta core cards puts PCI in the OCP 0 */
/* SOC-it on all malta core cards puts PCI in the IP/IF 2 */
#define SSC_PCI_REG_BASE (SSC_OCP0_REG_BASE + 0x10000)
#define MSC01_PCI_REG_BASE (_sbd_sysctrl_type == SYSCTRL_IS_SOCITSC ?	\
			    SSC_PCI_REG_BASE : MSC01_IP2_REG_BASE)

#define REG32(addr)     (*(volatile unsigned long *)(addr))
#define REG(base,offs)  REG32( (unsigned long)(base) + offs##_##OFS )

extern int _pci_maxbus; /* maximum number of busses we support */
extern int _pciverbose; /* Are we being verbose? */

/* reserve bottom 64KB of i/o space for ISA decodes */
extern pcireg_t _pci_minioaddr;
extern pcireg_t _pci_nextioaddr;

/* reserve bottom 1MB of mem space for ISA decodes */
extern pcireg_t _pci_minmemaddr;
extern pcireg_t _pci_nextmemaddr;


static unsigned long pcireserved;

static const struct pci_bus msc01_socit_pci_bus = {
	0,		/* minimum grant */
	255,		/* maximum latency */
	0,		/* devsel time = fast */
	1,		/* we support fast back-to-back */
	1,		/* we support prefetch */
	0,		/* we don't support 66 MHz */
	0,		/* we don't support 64 bits */
	4000000,	/* bandwidth: in 0.25us cycles / sec */
	1		/* initially one device on bus (i.e. us) */
};

static unsigned long pci_mem_space_pci_base;
static unsigned long pci_local_mem_pci_base;
static unsigned long pci_io_space_pci_base;

/*
 * Called to initialise the bridge at the beginning of time
 */
static int
__msc_pci_hwinit (int initialise)
{
  if (initialise) return -1; /* XXX: We do not handle board bringup yet */
  pcireserved = 0;
  _pci_bus[0] = msc01_socit_pci_bus;
  
  pci_mem_space_pci_base = REG(MSC01_PCI_REG_BASE,MSC01_PCI_SC2PMMAPL);
  pci_local_mem_pci_base = REG(MSC01_PCI_REG_BASE,MSC01_PCI_HEAD4);
  pci_io_space_pci_base = REG(MSC01_PCI_REG_BASE,MSC01_PCI_SC2PIOMAPL);
  
  _pci_minioaddr  = PCI_IO_SPACE + 0x10000;
  _pci_nextioaddr = PCI_IO_SPACE + PCI_IO_SIZE;
  _pci_minmemaddr =  PCI_MEM_SPACE + 0x100000;
  _pci_nextmemaddr = PCI_MEM_SPACE + PCI_MEM_SIZE;  
  return 0;
}

/*
 * Called to reinitialise the bridge after we've scanned each PCI device
 * and know what is possible.
 */
static void
__msc_pci_hwreinit (void)
{
    /* XXX: No idea if there's anything we can do here :-( */
}

/* Map the CPU virtual address of an area of local memory to a PCI
   address that can be used by a PCI bus master to access it. */
static vm_offset_t
__msc_pci_dmamap (vm_offset_t va, unsigned int len)
{
    return KVA_TO_PA (va) + pci_local_mem_pci_base;
}

/* Map the PCI address of an area of local memory to a CPU physical
   address. */
static vm_offset_t
__msc_pci_cpumap (vm_offset_t pcia, unsigned int len)
{
    return pcia - pci_local_mem_pci_base;
}

static pcitag_t
__msc_pci_make_tag(int bus, int device, int function)
{
    pcitag_t tag;
    tag = (bus << 16) | (device << 11) | (function << 8);
    return tag;
}

static void
__msc_pci_break_tag(pcitag_t tag, int *busp, int *devicep, int *functionp)
{
    if (busp) *busp = (tag >> 16) & 255;
    if (devicep) *devicep = (tag >> 11) & 31;
    if (functionp) *functionp = (tag >> 8) & 7;
}

static int
__msc_pci_canscan (pcitag_t tag)
{
    int bus, dev;
    _pci_break_tag (tag, &bus, &dev, NULL);
    return !(bus == 0 && (pcireserved & (1 << dev)));
}

#if !defined(IN_VXWORKS) && !defined(IN_PMON)
#include <sdethread.h>

static __sdethread_mutex_t	pci_cfg_mx = __SDETHREAD_MUTEX_INITIALIZER("pci_cfg");

#define pci_cfg_lock() \
  (void) __sdethread_mutex_lock (&pci_cfg_mx)

#define pci_cfg_unlock() \
  (void) __sdethread_mutex_unlock (&pci_cfg_mx)

#else
#define pci_cfg_lock() 
#define pci_cfg_unlock() 
#endif

/*
 * flush SOC-it(tm) register writes
 */
static void __inline__
bflush (void)
{
    _mips_sync();
}

/*
 * Read the given register from the given device
 */
static pcireg_t
__msc_pci_conf_read(pcitag_t tag, int reg)
{
    int busnum,devnum,func;
    pcireg_t res;
    unsigned long intr;
    
    __msc_pci_break_tag(tag, &busnum, &devnum, &func);
        
    REG(MSC01_PCI_REG_BASE, MSC01_PCI_INTSTAT) =
        MSC01_PCI_INTSTAT_MA_BIT | MSC01_PCI_INTSTAT_TA_BIT;
    
    /* Setup address */
    REG(MSC01_PCI_REG_BASE, MSC01_PCI_CFGADDR) =
        (busnum     << MSC01_PCI_CFGADDR_BNUM_SHF) |
        (devnum     << MSC01_PCI_CFGADDR_DNUM_SHF) |
        (func       << MSC01_PCI_CFGADDR_FNUM_SHF) |
        ((reg >> 2) << MSC01_PCI_CFGADDR_RNUM_SHF);
    bflush();
    res = REG(MSC01_PCI_REG_BASE, MSC01_PCI_CFGDATA);
    
    intr = REG(MSC01_PCI_REG_BASE, MSC01_PCI_INTSTAT);
    if (intr & (MSC01_PCI_INTSTAT_MA_BIT | MSC01_PCI_INTSTAT_TA_BIT) ) {
        REG(MSC01_PCI_REG_BASE, MSC01_PCI_INTSTAT) =
            (MSC01_PCI_INTSTAT_MA_BIT | MSC01_PCI_INTSTAT_TA_BIT);
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: target abort\n");
        return ~0;
    }
    return res;
}

/*
 * Write the given value into the given register of the given device
 */
static void
__msc_pci_conf_write(pcitag_t tag, int reg, pcireg_t data)
{
    int busnum,devnum,func;
    unsigned long intr;
    
    __msc_pci_break_tag(tag, &busnum, &devnum, &func);
        
    pci_cfg_lock();
    REG(MSC01_PCI_REG_BASE, MSC01_PCI_INTSTAT) =
        MSC01_PCI_INTSTAT_MA_BIT | MSC01_PCI_INTSTAT_TA_BIT;
    /* Setup address */
    REG(MSC01_PCI_REG_BASE, MSC01_PCI_CFGADDR) =
        (busnum     << MSC01_PCI_CFGADDR_BNUM_SHF) |
        (devnum     << MSC01_PCI_CFGADDR_DNUM_SHF) |
        (func       << MSC01_PCI_CFGADDR_FNUM_SHF) |
        ((reg >> 2) << MSC01_PCI_CFGADDR_RNUM_SHF);
    bflush();
    
    REG(MSC01_PCI_REG_BASE, MSC01_PCI_CFGDATA) = data;
    bflush();
    intr = REG(MSC01_PCI_REG_BASE, MSC01_PCI_INTSTAT);
    if (intr & (MSC01_PCI_INTSTAT_MA_BIT | MSC01_PCI_INTSTAT_TA_BIT) ) {
        REG(MSC01_PCI_REG_BASE, MSC01_PCI_INTSTAT) =
            (MSC01_PCI_INTSTAT_MA_BIT | MSC01_PCI_INTSTAT_TA_BIT);
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: target abort\n");
    }
    pci_cfg_unlock();
}

/*
 * Perform some IO <-> Memspace mappings
 */
static int
__msc_pci_map_io(pcitag_t tag, int reg, vm_offset_t *vap, vm_offset_t *pap)
{
    pcireg_t address;
    vm_offset_t pa;
    
    if (reg < PCI_MAP_REG_START || reg >= PCI_MAP_REG_END || (reg & 3)) {
	if (_pciverbose >= 1)
	    _pci_tagprintf(tag, "_pci_map_io: bad request\n");
	return -1;
    }
    
    address = _pci_conf_read(tag, reg);
    
    if ((address & PCI_MAP_IO) == 0) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_io: attempt to i/o map a memory region\n");
	return -1;
    }

    pa = (address & PCI_MAP_IO_ADDRESS_MASK) - pci_io_space_pci_base;
    *pap = pa;
    *vap = (vm_offset_t) PA_TO_KVA1 (PCI_IO_SPACE_SOCIT + pa);
    
    if (_pciverbose >= 3)
	_pci_tagprintf(tag, "_pci_map_io: mapping i/o at virtual %08x, physical %08x\n", 
		       *vap, *pap);

    return 0;
}

/*
 * Perform some Mem <-> Memspace mappings
 */
static int
__msc_pci_map_mem(pcitag_t tag, int reg, vm_offset_t *vap, vm_offset_t *pap)
{
    pcireg_t address;
    vm_offset_t pa;

    if (reg == PCI_MAP_REG_ROM) {
	/* expansion ROM */
	address = _pci_conf_read(tag, reg);
	if (!(address & PCI_MAP_ROM)) {
	    _pci_tagprintf (tag, "_pci_map_mem: attempt to map missing rom\n");
	    return -1;
	}
	pa = address & PCI_MAP_ROM_ADDRESS_MASK;
    }
    else {
	if (reg < PCI_MAP_REG_START || reg >= PCI_MAP_REG_END || (reg & 3)) {
	    if (_pciverbose >= 1)
		_pci_tagprintf(tag, "_pci_map_mem: bad request\n");
	    return -1;
	}
	
	address = _pci_conf_read(tag, reg);
	
	if ((address & PCI_MAP_IO) != 0) {
	    if (_pciverbose >= 1)
		_pci_tagprintf(tag, "pci_map_mem: attempt to memory map an I/O region\n");
	    return -1;
	}
	
	switch (address & PCI_MAP_MEMORY_TYPE_MASK) {
	case PCI_MAP_MEMORY_TYPE_32BIT:
	case PCI_MAP_MEMORY_TYPE_32BIT_1M:
	    break;
	case PCI_MAP_MEMORY_TYPE_64BIT:
	    if (_pciverbose >= 1)
		_pci_tagprintf (tag, "_pci_map_mem: attempt to map 64-bit region\n");
	    return -1;
	default:
	    if (_pciverbose >= 1)
		_pci_tagprintf (tag, "_pci_map_mem: reserved mapping type\n");
	    return -1;
	}
	pa = (address & PCI_MAP_MEMORY_ADDRESS_MASK);
    }

    
    pa -= pci_mem_space_pci_base;
    *pap = pa;
    if (address & PCI_MAP_MEMORY_CACHABLE) 
	*vap = (vm_offset_t) PA_TO_KVA0 (PCI_MEM_SPACE_SOCIT + pa);
    else
	*vap = (vm_offset_t) PA_TO_KVA1 (PCI_MEM_SPACE_SOCIT + pa);

    if (_pciverbose >= 3)
	_pci_tagprintf (tag, "_pci_map_mem: mapping memory at virtual 0x%x, physical 0x%x\n", 
			*vap, *pap);
    return 0;
}

/*
 * Final part of pci registration...
 */

static void
__msc_register_pci(void)
{
    if (_sbd_sysctrl_type != SYSCTRL_IS_SOCIT &&
	_sbd_sysctrl_type != SYSCTRL_IS_SOCITSC)
	return;
    REGISTER_STD_PCI_FUNCS(msc);
}

DECL_INIT_FUNC( IF_SYSCTRLDEP, __msc_register_pci );
