#ident "kit/malta/sbdpci.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2001-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * malta/sbdpci.c: Machine-specific functions for PCI autoconfiguration.
 */


#if #pci(yes)

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <mips/xcpt.h>
#include <mips/cpu.h>
#include <pci/pcivar.h>
#include <pci/pcireg.h>

#include "kit_impl.h"
#include "sbd.h"
#include "gt64120.h"

#include "sbdicu.h"

#include "product.h"

#include "sbdpci-table.h"
#include <sys/init.h>

/* one such structure for each PCI bus (allowing for PCI/PCI bridges) */
#define MAXBUS	3
struct pci_bus _pci_bus[MAXBUS];

__attribute__ ((section (".sdata")))
int _pci_maxbus = MAXBUS;

extern int _pciverbose;

int pcimaster;

/* PCI mem space allocation */
pcireg_t _pci_minmemaddr = 0;
pcireg_t _pci_nextmemaddr = 0;

/* PCI i/o space allocation */
pcireg_t _pci_minioaddr = 0;
pcireg_t _pci_nextioaddr = 0;

#define SYSCTRL_SPECIFIC(X) (*(_sbd_pci_table.X))

_pci_funcs_t _sbd_pci_table;

/*
 * Called to initialise at the beginning of time
 */
int
_pci_hwinit (int initialise)
{
  if( _sbd_pci_table._pci_hwinit == 0 ) return -1;
  return SYSCTRL_SPECIFIC(_pci_hwinit)(initialise);
}


/*
 * Called to reinitialise the bridge after we've scanned each PCI device
 * and know what is possible.
 */
void
_pci_hwreinit (void)
{
  SYSCTRL_SPECIFIC(_pci_hwreinit)();
}

/* Map the CPU virtual address of an area of local memory to a PCI
   address that can be used by a PCI bus master to access it. */
vm_offset_t
_pci_dmamap (vm_offset_t va, unsigned int len)
{
  return SYSCTRL_SPECIFIC(_pci_dmamap)(va,len);
}

/* Map the PCI address of an area of local memory to a CPU physical
   address. */
vm_offset_t
_pci_cpumap (vm_offset_t pcia, unsigned int len)
{
  return SYSCTRL_SPECIFIC(_pci_cpumap)(pcia,len);
}

pcitag_t
_pci_make_tag(int bus, int device, int function)
{
  return SYSCTRL_SPECIFIC(_pci_make_tag)(bus,device,function);
}

void 
_pci_break_tag(pcitag_t tag, int *busp, int *devicep, int *functionp)
{
  SYSCTRL_SPECIFIC(_pci_break_tag)(tag,busp,devicep,functionp);
}

pcireg_t
_pci_conf_read(pcitag_t tag, int reg)
{
  return SYSCTRL_SPECIFIC(_pci_conf_read)(tag,reg);
}

void
_pci_conf_write(pcitag_t tag, int reg, pcireg_t data)
{
  SYSCTRL_SPECIFIC(_pci_conf_write)(tag,reg,data);
}

int
_pci_canscan (pcitag_t tag)
{
  return SYSCTRL_SPECIFIC(_pci_canscan)(tag);
}

/*
 * allocate PCI address spaces
 * only applicable if we are the PCI master
 */

static pcireg_t
__COMMON_pci_allocate_mem (pcitag_t tag, size_t size)
{
    pcireg_t address;

    if (!pcimaster) {
	_pci_tagprintf (tag,"PCI: _pci_allocate_mem: not pcimaster\n");
	return -1;
    }

    /* allocate downwards, then round to size boundary */
    address = (_pci_nextmemaddr - size) & ~(size - 1);
    if (address > _pci_nextmemaddr || address < _pci_minmemaddr)
	return -1;
    _pci_nextmemaddr = address;
    return address;
}


static pcireg_t
__COMMON_pci_allocate_io (pcitag_t tag, size_t size)
{
    pcireg_t address;

    if (!pcimaster) {
	_pci_tagprintf (tag,"PCI: _pci_allocate_io: not pcimaster\n");
	return -1;
    }

    /* allocate downwards, then round to size boundary */
    address = (_pci_nextioaddr - size) & ~(size - 1);
    if (address > _pci_nextioaddr || address < _pci_minioaddr)
	return -1;
    _pci_nextioaddr = address;
    return address;
}

int
_pci_map_io(pcitag_t tag, int reg, vm_offset_t *vap, vm_offset_t *pap)
{
  return SYSCTRL_SPECIFIC(_pci_map_io)(tag,reg,vap,pap);
}


int
_pci_map_mem(pcitag_t tag, int reg, vm_offset_t *vap, vm_offset_t *pap)
{
  return SYSCTRL_SPECIFIC(_pci_map_mem)(tag,reg,vap,pap);
}


/* Shared between all PCI implementations for Malta System controllers */
static int
__COMMON_pci_map_int (pcitag_t tag)
{
    static const int sbdintrmap[4] = {INTR_PCI_A, INTR_PCI_B,
				      INTR_PCI_C, INTR_PCI_D};
    pcireg_t data;
    int pin, bus, device;

    data = _pci_conf_read(tag, PCI_INTERRUPT_REG);
    pin = PCI_INTERRUPT_PIN(data);
    if (pin == 0)
	/* No IRQ used. */
	return 0;

    if (--pin > 3) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_int: bad interrupt pin %d\n", pin);
	return -1;
    }

    /* ascend bus hierarchy until we get back to our (host) controller */
    _pci_break_tag (tag, &bus, &device, NULL);
    while (bus != 0) {
	/* XXX assume interrupt pins just pass straight through */
	tag = _pci_bus[bus].tag;
	_pci_break_tag (tag, &bus, &device, NULL);
    }

    switch (device) {
    case 10:	/* south bridge -> INTD */
	return INTR_PCI_D;
    case 11:	/* ethernet -> INTB */
	return INTR_PCI_B;
    case 12:	/* audio -> INTC */
	return INTR_PCI_C;
    case 18:	/* connector 1 -> A,B,C,D */
    case 19:	/* connector 2 -> D,A,B,C */
    case 20:	/* connector 3 -> C,D,A,B */
    case 21:	/* connector 4 -> B,C,D,A */
	return sbdintrmap[(pin + (22 - device)) % 4];
    }

    return -1;
}

int
_pci_map_int (pcitag_t tag)
{
  return SYSCTRL_SPECIFIC(_pci_map_int)(tag);
}

pcireg_t
_pci_allocate_mem (pcitag_t tag, size_t size)
{
  return SYSCTRL_SPECIFIC(_pci_allocate_mem)(tag,size);
}


pcireg_t
_pci_allocate_io (pcitag_t tag, size_t size)
{
  return SYSCTRL_SPECIFIC(_pci_allocate_io)(tag,size);
}

/* 
 * Handle mapping of fixed ISA addresses 
 */

static void *
__COMMON_isa_map_io (unsigned int port)
{
    if (port > 0xffff)
	return 0;
   return PA_TO_KVA1 (PCI_IO_SPACE + port);
}

static void *
__COMMON_isa_map_mem (vm_offset_t addr)
{
    if (addr > 0xfffff)
	return 0;
    return PA_TO_KVA1 (PCI_MEM_SPACE + addr);
}

void *
_isa_map_io (unsigned int port)
{
  return SYSCTRL_SPECIFIC(_isa_map_io)(port);
}

void *
_isa_map_mem (vm_offset_t addr)
{
  return SYSCTRL_SPECIFIC(_isa_map_mem)(addr);
}

static void empty_pci_table(void)
{
  memset(&_sbd_pci_table, 0, sizeof(_pci_funcs_t));
  REGISTER_PCI_FUNC(COMMON,_pci_map_int);
  REGISTER_PCI_FUNC(COMMON,_pci_allocate_io);
  REGISTER_PCI_FUNC(COMMON,_pci_allocate_mem);
  REGISTER_PCI_FUNC(COMMON,_isa_map_io);
  REGISTER_PCI_FUNC(COMMON,_isa_map_mem);
}

DECL_INIT_FUNC( IF_SYSCTRL, empty_pci_table );

#endif
