#ident "kit/mdimon/mdifile.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2002-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mdifile.c: file i/o for MIPS MDI
 */


#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <alloca.h>

#include "file.h"
#include "mdi.h"

static int	_fileread (struct file *, struct uio *);
static int	_filewrite (struct file *, struct uio *);
static int	_fileioctl (struct file *, int, void *);
static int	_fileselect (struct file *, int);
static int	_fileopen (struct file *, dev_t, int);
static int	_fileclose (struct file *);
static int	_filelookup (dir_t **, dir_t *, const char **, int, mode_t);
static void	_filedfree (dir_t *);

const struct fileops _file_ops = 
{
    _fileread, 
    _filewrite, 
    _fileioctl, 
    _fileselect, 
    _fileopen,
    _fileclose,
    _filelookup,
    _filedfree
};


struct mdifile {
    int		mf_fd;
    off_t	mf_offs;
};


static int
_filewrite (struct file *fp, struct uio *uio)
{
    struct mdifile *mf = fp->f_data;
    int err = 0;

    if (uio->uio_offset != mf->mf_offs) {
	if (_mdi_lseek32 (mf->mf_fd, (long)uio->uio_offset, SEEK_SET) < 0)
	    return ESPIPE;
	mf->mf_offs = uio->uio_offset;
    }
    
    while (uio->uio_resid != 0 && err == 0) {
	struct iovec *iov;
	err = uiogetiov (uio, &iov);
	if (err == 0) {
	    size_t res = _mdi_write (mf->mf_fd, iov->iov_base, iov->iov_len);
	    if (res > 0)
		err = uioadvance (res, uio);
	    else
		err = EIO;
	}
    }
    
    mf->mf_offs = uio->uio_offset;
    if (mf->mf_offs > fp->f_dp->d_size)
	fp->f_dp->d_size = mf->mf_offs;

    return err;
}


static int
_fileread (struct file *fp, struct uio *uio)
{
    struct mdifile *mf = fp->f_data;
    ssize_t res;
    int err = 0;

    if (!mf)
	/* root directory - can't read */
	return EACCES;

    if (uio->uio_offset != mf->mf_offs) {
	if (_mdi_lseek32 (mf->mf_fd, (long)uio->uio_offset, SEEK_SET) < 0)
	    return ESPIPE;
	mf->mf_offs = uio->uio_offset;
    }
    
    while (uio->uio_resid != 0 && err == 0) {
	struct iovec *iov;
	err = uiogetiov (uio, &iov);
	if (err == 0) {
	    res = _mdi_read (mf->mf_fd, iov->iov_base, iov->iov_len);
	    if (res > 0)
		err = uioadvance (res, uio);
	    else if (res < 0)
		err = EIO;
	    else 
		break;
	}
    }

    mf->mf_offs = uio->uio_offset;
    return err;
}


static int
_fileioctl (struct file *fp, int cmd, void *arg)
{
    return EINVAL;
}


static int
_fileselect (struct file *fp, int rw)
{
    /* not used yet, but return always ready */
    return rw;
}



static const dir_t template =
DIRDESC(0, 'n'<<8, S_IFREG|S_IRWXU|S_IRWXG|S_IRWXO, 0, _file_ops, -1);

static int
_fileopen (struct file *fp, dev_t dev, int fflags)
{
    struct mdifile *mf = fp->f_data;
    if (mf)
	mf->mf_offs = fp->f_offset;	/* take account of O_APPEND mode */
    return 0;
}


static int
_fileclose (struct file *fp)
{
    /* nothing to do */
    return 0;
}


static int
_filelookup (dir_t **ndpp, dir_t *dp, const char **name, int fflags, 
	     mode_t mode)
{
    const char *nm = *name;
    size_t nmlen = strlen (nm);
    char *buf = 0;
    struct mdifile *mf;
    int fd;
    
    /* Handle "." and ".." magic names */
    if (nm[0] == '.') {
#if 0
	/* no - interpret .. on the host file system */
	if (nm[1] == '.' && (nm[2] == '\0' || nm[2] == '/')) {
	    /* This isn't going to work if dp->d_parent isn't valid */
	    *name += 2;
	    *ndpp = dp->d_parent;
	    return 0;
	}
#endif
	if (nm[1] == '\0' || nm[1] == '/') {
	    *name += 1;
	    *ndpp = dp;
	    return 0;
	}
    }

    if (dp->d_name[0] == 'h' && nm[0] != '/') {
	/* in /host file system, so prepend an '/' */
	buf = malloc (nmlen + 2);
	if (!buf)
	    return ENOMEM;
	buf[0] = '/';
	memcpy (&buf[1], nm, nmlen + 1);
	nm = buf;
    }
    else if (dp->d_name[0] == 't' && nm[0] != '/') {
	/* in /tmp file system, so prepend "/tmp/" */
#define tmpdir "/tmp/"
	buf = malloc (nmlen + sizeof tmpdir);
	if (!buf)
	    return ENOMEM;
	memcpy (&buf[0], tmpdir, sizeof tmpdir - 1);
	memcpy (&buf[sizeof tmpdir - 1], nm, nmlen + 1);
	nm = buf;
    }

    fd = _mdi_open (nm, OFLAGS (fflags), mode);

    if (buf)
	free (buf);

    if (fd < 0)
	return ENOENT;

    dp = malloc (sizeof (dir_t) + sizeof (struct mdifile));
    if (!dp) {
	_mdi_close (fd);
	return ENOMEM;
    }

    *dp = template;
    dp->d_flags = DF_DYNAMIC;
    dp->d_count = 1;

    /* record current file size */
    dp->d_size = _mdi_lseek32 (fd, 0L, SEEK_END);
    (void) _mdi_lseek32 (fd, 0L, SEEK_SET);

    mf = dp->d_data = dp + 1;
    mf->mf_fd = fd;
    mf->mf_offs = 0;

    /* advance name pointer to end of path */
    *name += nmlen;
    *ndpp = dp;
    return 0;
}


static void
_filedfree (dir_t *dp)
{
    struct mdifile *mf = dp->d_data;
    if (mf)
	_mdi_close (mf->mf_fd);
    if (dp->d_flags & DF_DYNAMIC)
	free (dp);
}
