#ident "kit/mdimon/mditty.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2002-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mdimon/mditty.c: Posix style terminal i/o under MIPS MDI "monitor"
 */


#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/init.h>
#include <termios.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "file.h"
#include "tty.h"

#include "kit_impl.h"
#include "mdi.h"

static int	ttyrw (struct file *, struct uio *);
static int	ttyioctl (struct file *, int, void *);
static int	ttyselect (struct file *, int);
static int	ttyopen (struct file *, dev_t, int);
static int	ttyclose (struct file *);

const struct fileops _tty_ops = 
{
    ttyrw, 
    ttyrw, 
    ttyioctl, 
    ttyselect, 
    ttyopen,
    ttyclose,
};




static int
ttyrw (struct file *fp, struct uio *uio)
{
    int err = 0;
    while (uio->uio_resid != 0 && err == 0) {
	struct iovec *iov = uio->uio_iov;
	err = uiogetiov (uio, &iov);
	if (err == 0) {
	    ssize_t res;
	    if (uio->uio_rw == UIO_WRITE)
		res = _mdi_write (1, iov->iov_base, iov->iov_len);
	    else
		res = _mdi_read (0, iov->iov_base, iov->iov_len);
	    if (res >= 0)
		err = uioadvance (res, uio);
	    else
		err = EIO;
	    if (uio->uio_rw == UIO_READ)
		/* return with any data read */
		break;
	}
    }
    return err;
}


static const char ttydefchars[NCCS] = TTYDEF_CHARS;

static int
ttyioctl (struct file *fp, int cmd, void *arg)
{
  switch (cmd) {
  case TIOCGETA: 
    {
      /* return software defaults */
      struct termios *t = (struct termios *)arg;
      t->c_iflag = TTYDEF_IFLAG;
      t->c_oflag = TTYDEF_OFLAG;
      t->c_cflag = TTYDEF_CFLAG;
      t->c_lflag = TTYDEF_LFLAG;
      t->c_ispeed = t->c_ospeed = TTYDEF_SPEED;
      memcpy (&t->c_cc, ttydefchars, NCCS);
    }
    break;

  case TIOCSETA:
  case TIOCSETAW:
  case TIOCSETAF:
    {
      /* check for unsupported modes */
    }
    break;

  case TIOCFLUSH:
  case TIOCDRAIN:
  case TIOCSTART:
    /* quietly ignore */
    break;
    
  case TIOCOUTQ:		/* output queue size */
    *(int *)arg = 0;
    break;
    
  default:
    return EINVAL;
  }
    
  if (! _mdi_isatty (0))
    return ENOTTY;

  return 0;
}

static int
ttyselect (struct file *fp, int rw)
{
  /* always ready */
  return rw;
}

static int
ttyopen (struct file *fp, dev_t dev, int fflags)
{
    if (dev != 0)
	return ENXIO;
    return 0;
}


static int
ttyclose (struct file *fp)
{
    return 0;
}



/* /dev/tty*:	crw-rw-r-- */
#define TTYMODE	(S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IWOTH)

static void
ttydev_mount (void)
{
    dir_t ttynode;
    int err;

    memset (&ttynode, 0, sizeof (ttynode));
    ttynode.d_mode = TTYMODE;
    ttynode.d_ops = &_tty_ops;

#define ADDDESC(name, dev)				\
    ttynode.d_name = name;				\
    ttynode.d_dev = dev;				\
    err = _mknod ("/dev", &ttynode);			\
    if (err && err != EEXIST) {				\
      _mon_puts ("mknod(/dev/" name ") failed\n");	\
      abort ();						\
    }

    ADDDESC("console",	0);
    ADDDESC("tty",	0);
    ADDDESC("tty0",	0);
}

DECL_INIT_FUNC(IF_MOUNT, ttydev_mount)

/* call this dummy function to load this module */
void _ttydev_load (void);
void _ttydev_load (void) {} 
