#ident "kit/mtspmon/mtsprtlx.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2005-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mtspmon/mtsprtlx.c: AP/SP low-level transport 
 */

#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/init.h>
#include <termios.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "file.h"

#include <mips/cpu.h>
#include <mips/mt.h>
#include <mips/xcpt.h>
#include <mips/rtlx.h>
#include "sbd.h"

#include "kit_impl.h"

#define min(x,y) \
    ({ \
    __typeof__(x) __x = (x); \
    __typeof__(y) __y = (y); \
    __x < __y ? __x : __y; \
    })

#define LINUX_CPU 0

#define STDIO_CHANNEL	0
#define DBG_CHANNEL	1
#define SYSIO_CHANNEL	2

/* Magic symbol that vpe.c knows about, uses it to get hold of
   the channels structure */
static void *volatile vpe_shared = NULL;

static struct rtlx_info _rtlx, *rtlx;
static char rtlx_buffer[RTLX_CHANNELS * RTLX_BUFFER_SIZE *
                        RTLX_BUFFERS_PER_CHANNEL];

struct rtlx_fh {
    __sdethread_mutex_t 	fh_mx;
    struct rtlx_channel *fh_chan;
} rtlx_fh[RTLX_CHANNELS];

/* private */
static int lx_open (struct file *fp, dev_t dev, int fflags);
static int lx_close (struct file *fp);
static int lx_readwrite (struct file *fp, struct uio *uio);
static int lx_ioctl (struct file *fp, int cmd, void *arg);
static int lx_select (struct file *fp, int rw);
static int lx_read (struct rtlx_channel *rt, void *buf, int len);
static int lx_write (struct rtlx_channel *lx, const void *buf, int len);

const struct fileops _rtlx_ops = {
    .fo_read = lx_readwrite,
    .fo_write = lx_readwrite,
    .fo_open = lx_open,
    .fo_close = lx_close,
    .fo_ioctl = lx_ioctl,
    .fo_select = lx_select
};


void
_rtlx_set_shared (void *ptr)
{
    /* possibly excessive, but it wont hurt */
    _mips_sync();
    vpe_shared = ptr;
}


void
_rtlx_interrupt_vpe (int tc)
{
    /* make sure all writes have completed before we wake anyone */
    _mips_sync();

    /* send interrupt to TC n */
    mips_mt_dvpe ();
    mips32_mt_settarget (tc);
    mips32_mt_setc0cause (mips32_mt_getc0cause () | CR_SINT0);
    mips_mt_evpe ();
}


/* /dev/lx*:	crw-rw-r-- */
#define LXMODE	(S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IWOTH)

static const char *const lx_name[] = {
    "lx0", "*lxdbg", "*lxsys", "lx3", 
    "lx4", "lx5", "lx6", "lx7"
};


static int rtlx_init_early_done = 0;
    
/* Split the initialisation into two parts, do just the minimum
   required to allow use of the gdb-stub */
void
_rtlx_init_early (void)
{
    int i;
    char *ptr;
    struct rtlx_channel *chan;
    struct rtlx_info *rt;
    
    rt = &_rtlx;
    ptr = rtlx_buffer;

    /* initialise the channels, default buffer size... */
    for (i = 0; i < RTLX_CHANNELS; i++) {
        chan = &rt->channel[i];
        chan->rt_buffer = ptr;
        ptr += RTLX_BUFFER_SIZE;
        chan->lx_buffer = ptr;
        ptr += RTLX_BUFFER_SIZE;
        chan->buffer_size = RTLX_BUFFER_SIZE;
    }

    rt->id = RTLX_ID;
    rtlx = rt;
    _rtlx_set_shared (rtlx);
    _rtlx_interrupt_vpe (LINUX_CPU);

    rtlx_init_early_done = 1;
}

DECL_INIT_FUNC (IF_SYSCTRL, _rtlx_init_early);

/* Second line of initialisation, filesystem can now be used */
void
_rtlx_init (void)
{
    int i;
    dir_t lxnode;

    /* ensure the early initialisation has been done */
    if (!rtlx_init_early_done)
        _rtlx_init_early ();

    memset (&lxnode, 0, sizeof (lxnode));
    lxnode.d_mode = LXMODE;
    lxnode.d_ops = &_rtlx_ops;

    for (i = 0; i < RTLX_CHANNELS; i++) {
	const char *name = lx_name[i];

	/* initialise channel pointers and mutexes */
	__sdethread_mutex_init (&rtlx_fh[i].fh_mx, name);
	rtlx_fh[i].fh_chan = &rtlx->channel[i];

	/* initialise the /dev/lx0..lx7 character device nodes */
	if (name && *name != '*') {
	    lxnode.d_name = name;
	    lxnode.d_dev = i;
	    lxnode.d_data = &rtlx_fh[i];
	    if (_mknod ("/dev", &lxnode)) {
		_mon_puts ("mknod(/dev/lx? failed\n");
		abort ();
	    }
	}
    }
}

DECL_INIT_FUNC (IF_MOUNT, _rtlx_init);

static int 
lx_open (struct file *fp, dev_t dev, int fflags)
{
    struct rtlx_fh *rfh = (struct rtlx_fh *) fp->f_data;
    struct rtlx_channel *rt = rfh->fh_chan;
    int err;
    
    err = __sdethread_mutex_lock (&rfh->fh_mx);
    if (err == 0) {
	/* only one open allowed at a time */
	if (rt->rt_state != RTLX_STATE_OPENED)
	    rt->rt_state = RTLX_STATE_OPENED;
	else
	    err = EBUSY;
	__sdethread_mutex_unlock (&rfh->fh_mx);
    }
    return (err);
}

static int
lx_close (struct file *fp)
{
    struct rtlx_fh *rfh = (struct rtlx_fh *) fp->f_data;
    struct rtlx_channel *rt = rfh->fh_chan;
    rt->rt_state = RTLX_STATE_UNUSED;
    return 0;
}

/* keep the kit'isms in here (taken from pmontftp) */
static int 
lx_readwrite (struct file *fp, struct uio *uio)
{
    struct rtlx_fh *rfh = (struct rtlx_fh *) fp->f_data;
    struct rtlx_channel *rt = rfh->fh_chan;
    int err = 0;
    
    while (uio->uio_resid != 0 && err == 0) {
	struct iovec *iov;
	err = uiogetiov (uio, &iov);
	if (err == 0) {
	    ssize_t res;
	    (void) __sdethread_mutex_lock (&rfh->fh_mx);
	    if (uio->uio_rw == UIO_READ)
		res = lx_read (rt, iov->iov_base, iov->iov_len);
	    else
		res = lx_write (rt, iov->iov_base, iov->iov_len);
	    __sdethread_mutex_unlock (&rfh->fh_mx);
	    if (res > 0)
		err = uioadvance (res, uio);
	    else if (res < 0 || uio->uio_rw == UIO_WRITE)
		err = EIO;
	    else
		break;
	}
    }
    
    return (0);
}

/* read rt_buffer */
static int
lx_read (struct rtlx_channel *rt, void *buf, int len)
{
    unsigned int readp = rt->rt_read;
    int avail;
    unsigned int fl;

    /* find out how many bytes available */
    avail = (rt->rt_write + RTLX_BUFFER_SIZE - readp)
	% RTLX_BUFFER_SIZE;
    
    if (avail < len)
	len = avail;
    if (len <= 0)
	return (0);

    /* first bit: read pointer to end of the buffer */
    fl = min (len, RTLX_BUFFER_SIZE - readp);
    memcpy (buf, (void *)&rt->rt_buffer[readp], fl);
    
    /* second bit: from beginning of the buffer */
    if (fl != len)
	memcpy (buf + fl, (void *)rt->rt_buffer, len - fl);

    /* update the read pointer */
    rt->rt_read = (readp + len) % RTLX_BUFFER_SIZE;

    return (len);
}


static int 
lx_write (struct rtlx_channel *lx, const void *buf, int len)
{
    unsigned int readp = lx->lx_read;
    unsigned int writep = lx->lx_write;
    int space;
    unsigned int fl;
    
    if (readp == writep)
	space = RTLX_BUFFER_SIZE;
    else
	space = (readp + RTLX_BUFFER_SIZE - writep) % RTLX_BUFFER_SIZE;

    /* never fill the buffer completely, so indexes are always equal
       if empty and only empty, or !equal if data available. */
    --space;

    /* total number of bytes to copy */
    if (space < len)
	len = space;
    if (len <= 0)
	return (0);

    if (!buf)
	/* if no buffer, return amount of space free */
	return (len);

    /* first bit: write pointer to end of the buffer */
    fl = min (len, RTLX_BUFFER_SIZE - writep);
    memcpy ((void *)&lx->lx_buffer[writep], buf, fl);
	
    /* second bit: from beginning of the buffer */
    if (fl != len)
	memcpy ((void *)lx->lx_buffer, buf + fl, len - fl);
    
    /* update write pointer */
    lx->lx_write = (writep + len) % RTLX_BUFFER_SIZE;

    /* now interrupt the other side to wake it up */
    _rtlx_interrupt_vpe (LINUX_CPU);
    
    return (len);
}

static int 
lx_ioctl (struct file *fp, int cmd, void *arg)
{
    return EINVAL;
}

static int 
lx_select (struct file *fp, int rw)
{
    /* not used yet, but return always ready */
    return rw;
}


int
_rtlx_read_stdin (void *buff, int len)
{
    struct rtlx_fh *rfh = &rtlx_fh[STDIO_CHANNEL];
    int stat;

    (void) __sdethread_mutex_lock (&rfh->fh_mx);
    stat = lx_read (rfh->fh_chan, buff, len);
    __sdethread_mutex_unlock (&rfh->fh_mx);
    return stat;
}

int 
_rtlx_write_stdout (const void *buff, int len)
{
    struct rtlx_fh *rfh = &rtlx_fh[STDIO_CHANNEL];
    int stat;

    (void) __sdethread_mutex_lock (&rfh->fh_mx);
    stat = lx_write (rfh->fh_chan, buff, len);
    __sdethread_mutex_unlock (&rfh->fh_mx);
    return stat;
}


int 
_rtlx_open_dbg (void)
{
    if (rtlx)
        rtlx->channel[DBG_CHANNEL].rt_state = RTLX_STATE_OPENED;
    return (0);
}

int
_rtlx_close_dbg (void)
{
    if (rtlx)
        rtlx->channel[DBG_CHANNEL].rt_state = RTLX_STATE_UNUSED;
    return (0);
}

int
_rtlx_read_dbg (void *buff, int len)
{
    if (rtlx)
	/* No mutex locking for debug channel, must be only user */
        return (lx_read (&rtlx->channel[DBG_CHANNEL], buff, len));
    return (0);
}

int
_rtlx_write_dbg (const void *buff, int len)
{
    if (rtlx)
	/* No mutex locking for debug channel, must be only user */
        return (lx_write (&rtlx->channel[DBG_CHANNEL], buff, len));
    return (0);
}


/* Warning: no mutex protection for individual sysio operations,
   locking must be provided by caller around each pair of read/write
   calls. */

int
_rtlx_read_sys (void *buff, int len)
{
    struct rtlx_fh *rfh = &rtlx_fh[SYSIO_CHANNEL];

    if (rtlx) {
	int tmp, count = 0;
	char *ptr = (char *)buff;

	/* poll transport for reply */
	while (count  < len) {
	    tmp = lx_read (rfh->fh_chan, ptr, len - count);
	    count += tmp;
	    ptr += tmp;
	}
	
	return (0);
    }

    return (-1);
}

/* returns < 0 on error. Otherwise number of bytes copied. */
int
_rtlx_write_sys (const void *buff, int len)
{
    struct rtlx_fh *rfh = &rtlx_fh[SYSIO_CHANNEL];

    if (rtlx) {
	const char *ptr = (char *)buff;
	int tmp;

	do {
	    tmp = lx_write (rfh->fh_chan, ptr, len);
	    len -= tmp;
	    ptr += tmp;
	} while( len );
	
        return (len);
    }
    return (-1);
}

/* sbdser compatible functions.
   Return result or'd with 0x100 to indicate a valid zero data value.
   Caller is expected to and it with 0xff. */
int
_rtlx_read_1char (void *pdata, int canblock, long reloc)
{
    char c;
    
    if (rtlx)
	/* Warning - no mutex protection for lowest-level console i/o */
        if (lx_read (&rtlx->channel[STDIO_CHANNEL], &c, 1) > 0)
	    return (c | 0x100);
    
    return (0);
}

int
_rtlx_write_1char (void *pdata, char c, long reloc)
{
    if (rtlx)
	/* Warning - no mutex protection for lowest-level console i/o */
        lx_write (&rtlx->channel[STDIO_CHANNEL], &c, 1);
    return (0);
}

#ifdef RTLX_DEBUG
void _rtlx_state(void)
{
	int i;

	_mon_printf("id 0x%lx state %d\n", rtlx->id, rtlx->state);

	for (i = 0; i < RTLX_CHANNELS; i++) {
	    struct rtlx_channel *chan = &rtlx->channel[i];

	    _mon_printf(" rt_state %d lx_state %d buffer_size %d\n",
		   chan->rt_state, chan->lx_state, chan->buffer_size);
	    
	    _mon_printf(" rt_read %d rt_write %d\n",
		   chan->rt_read, chan->rt_write);
	    
	    _mon_printf(" lx_read %d lx_write %d\n",
		   chan->lx_read, chan->lx_write);
	    
	    _mon_printf(" rt_buffer <%s>\n", chan->rt_buffer);
	    _mon_printf(" lx_buffer <%s>\n", chan->lx_buffer);
	}
}
#endif
