/*
 * Copyright (c) 2005-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mtspmon/mtspsyscall.c host I/O via RTLX transport
 */

#define _KERNEL
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/init.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sdethread.h>
#include <mips/rtlx.h>
#include "file.h"
#include "mtsp.h"
#include "kit_impl.h"

static __sdethread_mutex_t sys_mx = __SDETHREAD_MUTEX_INITIALIZER("mtspsys");


/* 
 * Passes syscall args across rtlx transport
 * Returns 0 on success <0 on error
 */
int _mtsp_syscall (const struct mtsp_syscall *sc, const void *args, 
		   struct mtsp_syscall_ret *ret)
{
    int retval;

    (void) __sdethread_mutex_lock (&sys_mx);

    /* write out syscall buffer and args */
    if (_rtlx_write_sys (sc, sizeof(struct mtsp_syscall)) != 0) 
	retval = -1;
    else if (sc->size && _rtlx_write_sys (args, sc->size) != 0)
	retval = -1;
    else
	retval = 0;

    if (retval == 0 && ret) {
	/* poll transport for reply */
	if (_rtlx_read_sys (ret, sizeof(struct mtsp_syscall_ret)) != 0)
	    retval = -1;
    }

    (void) __sdethread_mutex_unlock (&sys_mx);
    return retval;
}

/* return file descriptor, or error, as expected */
int _mtsp_open (const char *fname, int flags, int mode)
{
    struct mtsp_syscall sc;
    struct mtsp_syscall_open op;
    struct mtsp_syscall_ret ret;

    sc.cmd = MTSP_SYSCALL_OPEN;
    sc.abi = MTSP_ABI;
    sc.size = sizeof(struct mtsp_syscall_open);
    
    op.name = fname;
    op.flags = flags;
    op.mode = mode;

    if (_mtsp_syscall (&sc, &op, &ret))
	return -1;

    return ret.retval;
}

int _mtsp_read (int fd, void *buff, int len)
{
    struct mtsp_syscall sc;
    struct mtsp_syscall_read rd;
    struct mtsp_syscall_ret ret;

    sc.cmd = MTSP_SYSCALL_READ;
    sc.abi = MTSP_ABI;
    sc.size = sizeof(struct mtsp_syscall_read);

    rd.fd = fd;
    rd.len = len;
    rd.buff = buff;

    if (_mtsp_syscall(&sc, &rd, &ret))
	return -1;

    return ret.retval;
}

int _mtsp_write (int fd, const void *buff, int len)
{
    struct mtsp_syscall sc;
    struct mtsp_syscall_write wr;
    struct mtsp_syscall_ret ret;

    sc.cmd = MTSP_SYSCALL_WRITE;
    sc.abi = MTSP_ABI;
    sc.size = sizeof(struct mtsp_syscall_write);

    wr.fd = fd;
    wr.len = len;
    wr.buff = (void *)buff;

    if (_mtsp_syscall (&sc, &wr, &ret))
	return -1;

    return ret.retval;
}

int _mtsp_close (int fd)
{
    struct mtsp_syscall sc;
    struct mtsp_syscall_close cl;
    struct mtsp_syscall_ret ret;
    

    sc.cmd = MTSP_SYSCALL_CLOSE;
    sc.abi = MTSP_ABI;
    sc.size = sizeof(struct mtsp_syscall_close);

    cl.fd = fd;

    if (_mtsp_syscall (&sc, &cl, &ret))
	return -1;

    return ret.retval;
}

long _mtsp_lseek32 (int fd, long offset, int where)
{
    struct mtsp_syscall sc;
    struct mtsp_syscall_lseek32 ls;
    struct mtsp_syscall_ret ret;

    sc.cmd = MTSP_SYSCALL_LSEEK32;
    sc.abi = MTSP_ABI;
    sc.size = sizeof(struct mtsp_syscall_lseek32);

    ls.fd = fd;
    ls.offset = offset;
    ls.where = where;

    if (_mtsp_syscall (&sc, &ls, &ret))
	return -1;

    return ret.retval;
}

int _mtsp_ioctl (int fd, int cmd, void *arg)
{
    struct mtsp_syscall sc;
    struct mtsp_syscall_ioctl ioc;
    struct mtsp_syscall_ret ret;

    sc.cmd = MTSP_SYSCALL_IOCTL;
    sc.abi = MTSP_ABI;
    sc.size = sizeof(struct mtsp_syscall_ioctl);

    ioc.fd = fd;
    ioc.cmd = cmd;
    ioc.arg = arg;

    if (_mtsp_syscall (&sc, &ioc, &ret))
	return -1;

    /* ret.retval will contain your return value */
    return ret.retval;
}

int _mtsp_isatty (int a)
{
    /* XXX */
    return (1);
}

int _mtsp_gettime (struct timeval *t)
{
    return (0);
}

void _mtsp_exit (int code)
{
    struct mtsp_syscall sc;
    struct mtsp_syscall_exit ex;

    sc.cmd = MTSP_SYSCALL_EXIT;
    sc.abi = MTSP_ABI;
    sc.size = sizeof(struct mtsp_syscall_exit);

    ex.exitcode = code;

    /* hand-coded syscall to avoid blocking on sys_mx */
    if (__sdethread_mutex_trylock (&sys_mx) == 0
	&& _rtlx_write_sys (&sc, sizeof(struct mtsp_syscall)) == 0
	&& _rtlx_write_sys (&ex, sizeof(struct mtsp_syscall_exit)) == 0) {
	/* success */;
    }
    else {
	/* This might be a really early exit call in which case try to
	   indicate the exit code via the shared rtlx pointer. Set at
	   least one bit, so an exit code of 0 is will be noticed. */
	_rtlx_set_shared ((void *)(0x100 | (code & 0xff)));
	_rtlx_interrupt_vpe (0);
    }
}

time_t _sbd_gettod (void)
{
    struct mtsp_syscall sc;
    struct mtsp_syscall_ret ret;

    sc.cmd = MTSP_SYSCALL_GETTOD;
    sc.abi = MTSP_ABI;
    sc.size = 0;

    if (_mtsp_syscall (&sc, NULL, &ret))
	return -1;
    
    return ret.retval;
}

static long pipefreq = 0;
long _sbd_pipefreq (void)
{
    struct mtsp_syscall sc;
    struct mtsp_syscall_ret ret;

    if (pipefreq == 0) {
	sc.cmd = MTSP_SYSCALL_PIPEFREQ;
	sc.abi = MTSP_ABI;
	sc.size = 0;
	
	if (_mtsp_syscall (&sc, NULL, &ret))
	    return -1;
	pipefreq = ret.retval;
    }
	
    return pipefreq;
}

/* This is a kludge, cpufreq is not actually pipefreq, but it's 
   close enough for now */
long  _sbd_cpufreq (void)
{
    return _sbd_pipefreq ();
}
