#ident "kit/p4032/sbddbg.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * p4032/sbddbg.c: Additional debug support for P4032 board
 */


#include <stddef.h>
#include <signal.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>

#include "kit_impl.h"
#include "sbd.h"
#include "icu.h"
#include "sbdser.h"

#ifdef DBGDBG
#define DBG(fmt, args...) \
	_mon_printf (fmt , ## args);
#else
#define DBG(fmt, args...)
#endif

static int
panichandler (int arg, struct xcptcontext *xcp)
{
    volatile p4032icu * const icu = PA_TO_KVA1 (ICU_BASE);
    if (icu->irr.err & ICU_ERR_DEBUG) {
	/* software debounce debug button */
	do {
	    icu->ctrl.clear = ICU_ERR_DEBUG;
	} while (icu->irr.err & ICU_ERR_DEBUG);
	return SIGTRAP;
    }
    return SIGINT;
}

static int
serhandler (int arg, struct xcptcontext *xcp)
{
    if (_icu_devintr () & (_sbddbgport ? ICU_DEV_UART1 : ICU_DEV_UART0)) {
	if (_dbg_inpoll () == 0x03) 
	    return (SIGTRAP);
    }
    
    return (0);
}

void
_dbg_signals (void)
{
    unsigned int sr;

    _icu_map (_sbddbgport ? INTR_UART1 : INTR_UART0, ICU_HWINT0);
    _icu_mask (_sbddbgport ? INTR_UART1 : INTR_UART0, 1);

    /*
     * intrupt does not enable interrupts so we do them here
     */
    sr = 0;

    /* enable serial interrupt for debug channel */
    (void) intrupt (2, serhandler, 0);
    sr |= SR_IM2;

    /* FIXME program ICU */
    /* enable panic handler for debug button */
    (void) intrupt (6, panichandler, 0);
    sr |= SR_IM6;

    mips_bissr(sr);
}



#ifdef DBGFLASH

#include "flashdev.h"
#include "fcache.h"

#define IS_FLASH(pa) ((pa) >= BOOTPROM_BASE  \
			|| ((pa) >= FSIMM_BASE && (pa) < PCI_MEM_SPACE))

static flashcookie_t	dbg_fcookie;
static paddr_t		dbg_fbase;
static paddr_t		dbg_flimit;

static int
dbg_openflash (paddr_t pa)
{
    struct flashinfo info;

    if (dbg_fcookie && pa >= dbg_fbase && pa < dbg_flimit)
	return 1;

    dbg_fcookie = flash_open (pa);
    if (dbg_fcookie == FLASH_UNKNOWN || dbg_fcookie == FLASH_MISSING)
	return 0;
    
    flash_info (dbg_fcookie, 0, &info);
    dbg_fbase = info.mapbase;
    dbg_flimit = info.mapbase + info.size;
    return 1;
}


void
_dbg_enterhook (void)
{
    (void) _fcache_init ();
}

int
_dbg_exithook (void)
{
    return (_fcache_flush ());
}

int
_dbg_put_byte (vaddr_t va, unsigned char v)
{
    paddr_t	pa;
    int		err;

    err = 0;
    pa = _dbg_vtop (va);
    if (pa == (paddr_t)-1) {
	err = 1;
    }
    else {
	if (IS_FLASH (pa)) {
	    if (!dbg_openflash (pa))
		err = 1;
	    else
		err = _fcache_put_bytes (dbg_fcookie, pa - dbg_fbase, &v, 1);
	}
	else {
	    err = mips_put_byte (va, v);
	}
    }

    DBG ("_dbg_put_byte: mem=%p v=0x%02x", mem, v);
    DBG (" err=%d\n", err);

    return err;
}

unsigned int
_dbg_get_byte (vaddr_t va, int *err)
{
    paddr_t		pa;
    unsigned int	v = 0;
    
    pa = _dbg_vtop (va);
    if (pa == (paddr_t)-1) {
	*err = 1;
    }
    else {
	if (IS_FLASH (pa)) {
	    if (!dbg_openflash (pa))
		*err = 1;
	    else if (_fcache_get_bytes (dbg_fcookie, pa - dbg_fbase, &v, 1))
		*err = 1;
	}
	else
	    v = mips_get_byte ((void *)va, err);
    }

    DBG("_dbg_get_byte: mem=%p", va);
    DBG(" v=0x%02x err=%d\n", v, *err);
    return v;
}

/*
 * this routine is only used when installing single step breakpoints
 * we need to take care of burst writes to ROM that could cause a bus error
 */
int
_dbg_put_word (vaddr_t va, unsigned int v)
{
    paddr_t	pa;
    int		err;

    err = 0;
    pa = _dbg_vtop (va);
    if (pa == (paddr_t)-1) {
	err = 1;
    }
    else {
	if (IS_FLASH (pa))
	    err = 1;		/* fail -> simulate single step */
	else
	    err = mips_put_word ((void *)va, v);
    }

    DBG ("_dbg_put_word: mem=%p v=%08x", va, v);
    DBG (" err=%d\n", err);

    return err;
}

/*
 * this routine is only used when installing single step breakpoints
 * because the flash cache has not been written back yet,
 * we need to coordinate with it...
 */
unsigned int
_dbg_get_word (vaddr_t va, int *err)
{
    paddr_t		pa;
    unsigned int	v;

    pa = _dbg_vtop (va);
    if (pa == (paddr_t)-1) {
	*err = 1;
	v = 0;
    }
    else {
	if (IS_FLASH (pa)) {
	    if (!dbg_openflash (pa))
		*err = 1;
	    else if (_fcache_get_bytes (dbg_fcookie, pa - dbg_fbase, 
					&v, sizeof(v)))
		*err = 1;
	}
	else
	    v = mips_get_word ((void *)va, err);
    }
    DBG("_dbg_get_word: mem=%p", va);
    DBG(" v=0x%08x err=%d\n", v, *err);
    return v;
}

#endif
