#ident "kit/p4032/sbdeeprom.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * P4032/sbdeeprom.c: low-level interface to P4032 E2PROM
 */


#include <stddef.h>
#include <mips/cpu.h>

#include "kit_impl.h"
#include "sbd.h"
#include "eeprom.h"
#include "nvenv.h"

/* eeprom is accessed thru Z80 PIO */
#include "z80pio.h"

/* 93Cxx command layout */
#define CBITS		(EEABITS + 3)
#define AMASK		((1 << EEABITS) - 1)
#define SB		(1 << (EEABITS + 2))
#define COMD(x)		((x) << (EEABITS - 2))

/* 93Cxx command codes */
#define	WRTDIS		(SB | COMD(0x0))
#define	WRITEALL	(SB | COMD(0x1))
#define	ERASEALL	(SB | COMD(0x2))
#define	WRTENB		(SB | COMD(0x3))
#define WRITE(a)	(SB | COMD(0x4) | ((a) & AMASK))
#define READ(a)		(SB | COMD(0x8) | ((a) & AMASK))
#define ERASE(a)	(SB | COMD(0xc) | ((a) & AMASK))

/*
 * Chip requires at least 1 usec delay between signal changes
 */
#define WAIT()	usdelay(1)

#define TOGGLE(bit,on) \
do { \
	WAIT (); if (on) _zpiob_bis (bit); else _zpiob_bic (bit); \
} while (0);


#if __GNUC__ >= 3
/* let auto-inliner make decisions */
#define inline
#endif

/* chip select */
static inline void CS (int on)	{TOGGLE (ZPIOB_E2_CS, on)}

/* shift clock */
static inline void SK (int on)	{TOGGLE (ZPIOB_E2_SK, on)}

/* my data out, chip's data in */
static inline void DOUT (int on) {TOGGLE (ZPIOB_E2_DI, on)}

/* my data in, chip's data out */
static inline unsigned int DIN (void)	
{ 
    WAIT (); 
    return (_zpiob_get () & ZPIOB_E2_DO) ? 1 : 0;
}

/* clock pulse */
static inline void CLK (void)	
{
    SK (1); SK (0);
}


/*
 * send command
 */
static void CMD (unsigned int c)
{
    int n = CBITS;

    SK(0); DOUT(0); CS (1); CLK ();

    while (n-- > 0) {
	DOUT ((c >> n) & 1); CLK ();
    }
    DOUT (0);
}


static void ECMD (void)
{
    int i;

    DOUT (0); CS (0);

    /* do we really need 5 clocks, probably zero or one would do! */
    for (i = 1; i <= 5; i++)
	CLK ();
}


static unsigned short
eeget (unsigned int offs)
{
    unsigned short v = 0;
    int n;

    CMD (READ (offs));
    n = 16;
    while (n-- > 0) {
	CLK ();
	v = (v << 1) | DIN ();
    }
    ECMD ();
    return v;
}


static void
eeput (unsigned int offs, unsigned short val)
{
    int n;

    /* enable writes */
    CMD (WRTENB); ECMD ();

    /* write command + data */
    CMD (WRITE (offs));
    n = 16;
    while (n-- > 0) {
	DOUT ((val >> n) & 1);
	CLK ();
    }
    ECMD ();

    /* raise CS and poll until DIN is active */
    CS (1); CLK ();
    n = 100;	/* 100ms timeout */
    do {
	usdelay (1000);	/* 1ms */
	CLK ();
    } while (!DIN () && n-- > 0);
    ECMD ();

    /* disable writes */
    CMD (WRTDIS); ECMD ();
}



/*
 * Now the glue that interfaces the hardware independent code to 
 * the low-level access functions.
 *
 * Since reading the EEPROM is very slow, we keep a cached copy of the 
 * whole prom. Since the prom is always accessed in 16-bits words, we
 * buffer stores and only flush out a 16-bit word when a new word is 
 * written to.
 */

static union {
    unsigned char	b[EESIZE];
    unsigned short	w[EESIZE/2];
} nvcopy;

#if #endian(little)
#define htol(v) (v)
#define ltoh(v) (v)
#endif
#if #endian(big)
#define htol(v) ({unsigned short _v = v; \
	((_v & 0xff) << 8) | ((_v >> 8) & 0xff);})
#define ltoh(v) htol(v)
#endif

static int nv_inited = 0;
static int nv_laststore;

static void
nv_init (void)
{
    int i;

    if (!nv_inited) {
	for (i = 0; i < EESIZE/2; i++)
	    nvcopy.w[i] = htol(eeget (i));
	nv_laststore = -1;
	nv_inited = 1;
    }
}


unsigned int 
_sbd_nvram_size (unsigned int *nvoffs, unsigned int *nvsize)
{
    if (nvoffs) *nvoffs = NVOFFSET;
    if (nvsize) *nvsize = NVSIZE;
    return EESIZE;
}

int
 _sbd_env_force_reset ()
{
    volatile unsigned int *opt = PA_TO_KVA1(OPTION_BASE);
    return !(*opt & OPTION_ENVNRST);
}

#define _sbd_nvram_force_reset() _sbd_env_force_reset ()


void
_sbd_nvram_flush (int offs)
{
    nv_init ();
    if (offs > 0)
	offs /= 2;
    if (offs != nv_laststore && nv_laststore >= 0)
	eeput (nv_laststore, ltoh(nvcopy.w[nv_laststore]));
    nv_laststore = offs;
}


unsigned char
_sbd_nvram_getbyte(unsigned int offs)
{
    nv_init ();
    return nvcopy.b[offs];
}


void
_sbd_nvram_setbyte(unsigned int offs, unsigned char val)
{
    nv_init ();
    if (nvcopy.b[offs] != val) {
	_sbd_nvram_flush (offs);
	nvcopy.b[offs] = val;
    }
}


unsigned short
_sbd_nvram_getshort (unsigned int offs)
{
    nv_init ();
    return ltoh(nvcopy.w[offs/2]);
}


void
_sbd_nvram_setshort (unsigned int offs, unsigned short val)
{
    nv_init ();
    if (ltoh(nvcopy.w[offs/2]) != val) {
	_sbd_nvram_flush (offs);
	nvcopy.w[offs/2] = htol(val);
    }
}


int
_sbd_nvram_hwtest (void)
{
    unsigned int offs = (NVOFFSET + ENVOFF_TEST) / 2;
    unsigned short x;

    eeput (offs, 0xaaaa);
    if ((x = eeget (offs)) != 0xaaaa)
	return 0;

    eeput (offs, 0x5555);
    if ((x = eeget (offs)) != 0x5555)
	return 0;

    return 1;
}

