#ident "kit/p4032/sbdpanel.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * P4032/sbdpanel.c: interface to the LCD and LED panel displays
 */


#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <mips/cpu.h>

#include "kit_impl.h"

#include "sbd.h"
#include "lcd.h"

static int	lcd_present;
static char	lcd_slider[17];


static inline int
lcd_busy ()
{
    volatile lcddev * dp = PA_TO_KVA1(LCD_BASE);
    usdelay (1);
    return (dp->lcd_stat & LCD_STAT_BUSY) != 0;
}


static int
lcd_wait ()
{
    int timeout = 2000;		/* ~2ms */
    int busy;

    do {
	busy = lcd_busy ();
    } while (busy && --timeout != 0);
    usdelay (1);
    return !busy;
}


static void
lcd_command (unsigned char cmd) 
{
    volatile lcddev *dp = PA_TO_KVA1(LCD_BASE);
    if (!lcd_wait ())
	return;
    dp->lcd_cmd = cmd;
    usdelay (40);
}


static void
lcd_wdata (unsigned char data) 
{
    volatile lcddev *dp = PA_TO_KVA1(LCD_BASE);
    if (!lcd_wait ())
	return;
    dp->lcd_data = data;
    usdelay (40);
}

static unsigned int
lcd_rdata (void)
{
    volatile lcddev *dp = PA_TO_KVA1(LCD_BASE);
    unsigned int data;

    lcd_wait ();
    data = dp->lcd_data & 0xff;
    usdelay (40);
    return data;
}


static void
lcd_message (int row, int col, const char *s, size_t nb, int pad)
{
    int addr;

    addr = ((row == 0) ? 0x00 : 0x40) + col;
    lcd_command (LCD_SET_DDADDR | addr);
    
    while (col++ < 16) {
	if (*s && nb > 0) {
	    lcd_wdata (*s++);
	    nb--;
	}
	else if (pad)
	    lcd_wdata (' ');
	else
	    break;
    }
}


static void
lcd_flash (int on)
{
    lcd_command (LCD_SET_DDADDR | 15);
    lcd_wdata (on ? '*' : ' ');
}


static void
led_message (int row, int col, const char *s, size_t nb, int pad)
{
    volatile unsigned int *led = PA_TO_KVA1 (LED_BASE);
    int i;

    /* ignore row */
    if (col < 4) {
	for (i = 3 - col; i >= 0; i--) {
	    if (*s && nb > 0) {
		led[i] = *s++;
		nb--;
	    }
	    else if (pad)
		led[i] = ' ';
	    else 
		break;
	}
    }
}


static void
led_flash (int on)
{
    if (on)
	_bcr_bis (BCR_LED_ON);
    else
	_bcr_bic (BCR_LED_ON);
}



/* 
 * Public interface here
 */

void
_sbdpanelmessage (int line, int col, const char *str, size_t nb, int pad)
{
    if (lcd_present) {
	const char *ss = str; 
	size_t snb = nb;
	while (*ss == ' ' && snb > 0) {ss++, snb--;} /* skip initial spaces */
	led_message (0, 0, ss, snb, pad);
	lcd_message (line, col, str, nb, pad);
    }
    else {
	led_message (line, col, str, nb, pad);
    }
}


void
_sbdpanelprogress (int percent)
{
    if (lcd_present) {
	static int lastnb = -1;
	int nb = (percent * 16 + 50) / 100;
	
	if (nb < 0)
	    nb = 0;
	else if (nb > 16) 
	    nb = 16;
	if (nb != lastnb) {
	    lcd_message (0, 0, lcd_slider, nb, 1);
	    lastnb = nb;
	}
    }
}


void
_sbdpanelflash (int on)
{
    led_flash (on);
    if (lcd_present)
	lcd_flash (on);
}


void
_sbdpanelclear (void)
{
    led_message (0, 0, "", 0, 1);
    if (lcd_present)
	lcd_command (LCD_CLEAR);
}


void 
_sbdpanelsize (int *prows, int *pcols)
{
    if (lcd_present) {
	/* 2 x 16 LCD display */
	if (prows) *prows = 2;
	if (pcols) *pcols = 16;
    }
    else {
	/* 1 x 4 LED display */
	if (prows) *prows = 1;
	if (pcols) *pcols = 4;
    }
}


int 
_sbdpanelinit (void)
{
    static const unsigned char tpat[4] = {0xaa, 0x55, 0xff, 0x00};
    int i;

    /* should be non-busy (possibly after a delay) */
    lcd_present = lcd_wait ();
    if (!lcd_present)
	return 1;

    lcd_command (LCD_CLEAR);

    /* should be busy */
    lcd_present = lcd_busy();
    if (!lcd_present)
	return 1;

    /* initialise display as per our requirements */
    lcd_command (LCD_FUNC | LCD_FUNC_8BIT | LCD_FUNC_2LINE | LCD_FUNC_5x7);
    lcd_command (LCD_CTRL | LCD_CTRL_DISPLAY);
    lcd_command (LCD_MODE | LCD_MODE_INCR);
    lcd_command (LCD_HOME); 

    /* store pattern into display memory */
    lcd_command (LCD_SET_DDADDR | 0); 
    for (i = 0; i < sizeof(tpat); i++)
	lcd_wdata (tpat[i]); 

    /* read back pattern */
    lcd_command (LCD_SET_DDADDR | 0); 
    for (i = 0; i < sizeof(tpat); i++) {
	/* check dd ram contents */
	if (lcd_rdata () != tpat[i]) {
	    lcd_present = 0;
	    return 1;
	}
    }

    /* it really is there! */
    lcd_command (LCD_CLEAR);

    memset (lcd_slider, 0xff, sizeof(lcd_slider) - 1);
    lcd_slider[sizeof(lcd_slider)-1] = '\0';

    return 1;
}
