#ident "kit/p4032/sbdrtpx.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * sbdrtpx.c: low-level interrupt and timer handling for P4032 board
 */


#include <sys/types.h>
#include <mips/cpu.h>
#include <mips/prid.h>
#include <mips/xcpt.h>
#include "kit_impl.h"
#include "sbd.h"
#include "icu.h"

#include "pthread_impl.h"

/* 
 * Interrupt wiring under MIPS RTPX:
 *  HINT5 - hard counter clock
 *  HINT4 - hard debug/panic interupt
 *  HINT3 - floppy dma
 *  HINT2 - centronics
 *  HINT1 - v962 + fdc
 *  HINT0 - uart0 + uart1 + kbd + gpio + rtc
 *  SINT1 - unused/reserved
 *  SINT0 - unused/reserved
 */

/*
 * Note that the ipl0 interrupt mask includes all three levels of
 * interrupt controller, the 4-bit external pci mask, the 8-bit CPU
 * status register interrupt mask, and the external local device mask.
 *
 * |24    20|19    16|15           8|7            0|
 * |pci mask|  0000  |  $sr imask   |  dev mask    |
 *
 * Note that the sr-mask is aligned to match its position in the status
 * register, and this fact is known by other parts of the kernel.
 */

/* the interrupt mask when at "user" level (ipl0) */
_sbd_intrmask_t		_sbd_ipl0_mask;

/* the CPU status register when at scheduler level (ipl1) */
unsigned int 		_sbd_ipl1_sr;

/* the CPU status register when at interrupt blocking level (iplhi) */
unsigned int		_sbd_iplhi_sr;

/* the interrupt mask with which we were entered (e.g. from PMON),
   see .../kit/pmon/pmonlow.sx */
unsigned int 		_sr_base __attribute__((common));

/* tell PMON that we handle all interrupts (except panic i/u),
   see .../kit/pmon/pmonxlow.sx */
unsigned 		_intrmask = SR_IMASK & ~SR_HINT4;

/* high priority debug button handler on CR_HINT4 */
static int		panicHandler (void);

/* tables mapping interrupt register bits to interrupt numbers */
static signed char	crToIntr[256];
static signed char	devToIntr[256];
static signed char	pciToIntr[16];

/* and vice-versa */
static _sbd_intrmask_t	intrToIrr[INTR_MAX+1];

/* flag set when in _sbd_intrnext() to prevent recursion */
static int		in_intrnext;

/* 
 * Return number of timer ticks per second
 */
unsigned int
_sbd_tickspersec (void)
{
    unsigned int freq = _sbd_pipefreq ();
    if (((mips_getprid () >> 8) & 0xff) == PRID_R4100)
	/* R4100 counter runs at pipeline frequency / 4 */
	freq /= 4;
    else
	/* R4x00/R5x00 counter runs at pipeline frequency / 2 */
	freq /= 2;
    return freq;
}


/*
 * Return current value of timer (upcount)
 */
unsigned int
_sbd_gettickval (void)
{
    return mips_getcount();
}


/*
 * Set value at which timer should generate interrupt, and
 * clear any pending timer interrupt.
 */
void
_sbd_settickcmp (unsigned int v)
{
    /* set next tick time, and clear interrupt */
    mips_setcompare(v);
}


int 
_sbd_intr_howmany ()
{
    return INTR_MAX + 1;
}


/* 
 * Poll specific interrupt, returns 0 if interrupt is not active.
 * Called at ipl1.
 */
int 
_sbd_intrpoll (int intr)
{
    _sbd_intrmask_t irr = intrToIrr[intr];

    if (irr & CR_IMASK) {
	/* check internal cpu cause register */
	irr &= mips_getcr() & CR_IMASK;
    }
    else {
	/* check external icu request registers */
	volatile p4032icu *icu = PA_TO_KVA1 (ICU_BASE);
	if (irr & 0xff) {
	    irr &= (icu->irr.dev & 0xff);
	}
	else if (irr & 0xff0000) {
	    irr &= (icu->irr.pci & 0xf0) << 16;
	}
    }
    return irr != 0;
}


/*
 * Enable ipl0 interupts which have been blocked, in particular
 * when interrupt handler first installed.
 * Called at ipl1.
 */
void 
_sbd_intrenable (int intr)
{
    _sbd_intrmask_t irr = intrToIrr[intr];

    _sbd_ipl0_mask |= irr;
    if (irr & SR_IMASK) {
	/* now enabled at ipl0, so make sure it's disabled at ipl1 */
	_sbd_ipl1_sr &= ~irr;
    }
    else {
	/* set external icu masks to match new value */
	volatile p4032icu *icu = PA_TO_KVA1 (ICU_BASE);
	if (irr & 0xff) {
	    icu->ctrl.mask = _sbd_ipl0_mask & 0xff;
	    mips_wbflush ();
	}
	else if (irr & 0xf00000) {
	    icu->ctrl.pcimask = ((_sbd_ipl0_mask >> 16) & 0xf0)
		| ICU_PCI_FDCDMA;
	    mips_wbflush ();
	}
    }
}


/* 
 * Take a guess at which ipl this CP0 status register represents. If
 * in doubt return 0. 
 */
int
_sbd_probable_ipl (unsigned int sr)
{
    return !(sr & SR_IE) ? 2 : (sr & SR_HINT5) ? 0 : 1;
}


/* 
 * Return next pending interrupt for this cpu, or zero if none.
 * If idle != 0 then we can spin waiting for an interrupt 
 * (possibly going into low-power mode) 
 * Called at ipl1.
 */
int 
_sbd_intrnext (int idle)
{
  volatile p4032icu *icu = PA_TO_KVA1 (ICU_BASE);
  int intr;

  /* switch to "hi" interrupt priority (all interrupts disabled) */
  mips_setsr (_sbd_iplhi_sr);

  in_intrnext = 1;

  while (1) {
      unsigned int pend;

      /* find interrupts which we handle that would be pending at ipl0 */
      pend = mips_getcr() & _sbd_ipl0_mask & _intrmask;
      intr = crToIntr[pend >> 8];
      
      /* special case panic interrupts */
      if (intr == INTR_PANIC) {
	  (void) panicHandler ();
	  continue;
      }
      
      /* pick apart external interrupts multiplexed onto one cpu interrupt */
      if (intr == INTR_DEVMUX)
	  intr = devToIntr[icu->irr.dev & _sbd_ipl0_mask & 0xff];
      else if (intr == INTR_PCIMUX)
	  intr = pciToIntr[(icu->irr.pci >> 4) & (_sbd_ipl0_mask >> 20) & 0x0f];
      
      if (intr != 0 || !idle)
	  break;
      
#if notyet
      /* return to ipl0 to allow non-RTPX interrupts to be serviced, 
	 and so that a wait instruction will get woken. Ideally we
         would short circuit the interrupt which terminates the 
	 wait, to avoid a complete context save & restore. */
      mips_setsr (_sbd_ipl0_sr);
      if (canwait)
	  _mips_wait ();
#else
      /* return to ipl1 to allow non-RTPX interrupts to be serviced */
      mips_setsr (_sbd_ipl1_sr);
#endif
      mips_setsr (_sbd_iplhi_sr);
  }

  in_intrnext = 0;

  if (intr > 0) {
      /* disable the detected interrupt at ipl0; note that clock i/u
	 (intr == -1) is never disabled. */
      _sbd_intrmask_t irr = intrToIrr[intr];

      _sbd_ipl0_mask &= ~irr;

      /* update external hardware masks to reflect new setting */
      if (irr & 0xff) {
	  icu->ctrl.mask = _sbd_ipl0_mask & 0xff;
	  mips_wbflush ();
      }
      if (irr & 0xf00000) {
	  icu->ctrl.pcimask = ((_sbd_ipl0_mask >> 16) & 0xf0) | ICU_PCI_FDCDMA;
	  mips_wbflush ();
      }
  }

  /* return to ipl1 */
  mips_setsr (_sbd_ipl1_sr);
  return intr;
}


/* 
 * Board-specific interrupt handler, return 1 if interrupt handled.
 * Called at iplhi.
 */
int 
_sbd_intrhandler (unsigned int pend, struct xcptcontext *xcp)
{
#ifdef notyet
    /* do nothing if in interrupt poll loop */
    if (in_intrnext)
	return 0;
#endif

    /* we only handle panic interrupts locally */
    if (pend & _intrmask & CR_HINT4)
	return panicHandler ();

    /* everything else is passed to the scheduler */
    return 0;
}


/* 
 * Initialise interrupt handling for this board 
 * Called at undefined ipl, should return at iplhi
 */
void 
_sbd_intrinit(void)
{
    volatile p4032icu *icu = PA_TO_KVA1 (ICU_BASE);
    int i;

    /* disable all interrupts at interrupt level and set cpu to iplhi */
    _sbd_iplhi_sr = _sr_base & ~(SR_CU1 | SR_IMASK | SR_IE);
    mips_setsr (_sbd_iplhi_sr);

    /* disable clock, pci & local dev interrupt groups in scheduler */
    _sbd_ipl1_sr = (_sr_base & ~(SR_HINT5 | SR_HINT1 | SR_HINT0)) | SR_IE;

    /* enable clock, pci & local dev interrupt groups at user level */
    _sbd_ipl0_mask = (_sr_base & SR_IMASK) | SR_HINT5 | SR_HINT1 | SR_HINT0;

    /* disable individual PCI device interrupts, except for
       FDC DMA which gets it's own bit in CPU Cause register */
    icu->ctrl.pcimask = ICU_PCI_FDCDMA;

    /* disable individual local device interrupts */
    icu->ctrl.mask = 0;

    /* wire all local device interrupts to HINT0 */
    icu->ctrl.devxbar0 = ((ICU_HWINT0 << ICU_DEVX0_V96)
			  | (ICU_HWINT0 << ICU_DEVX0_FDC)
			  | (ICU_HWINT0 << ICU_DEVX0_KBD)
			  | (ICU_HWINT0 << ICU_DEVX0_UART0));
		   
    /* ... except centronics which goes to HINT2 */
    icu->ctrl.devxbar1 = ((ICU_HWINT0 << ICU_DEVX1_UART1)
			  | (ICU_HWINT2 << ICU_DEVX1_CENT)
			  | (ICU_HWINT0 << ICU_DEVX1_GPIO)
			  | (ICU_HWINT0 << ICU_DEVX1_RTC));
		   
    /* wire all PCI device interrupts to HINT1 */
    icu->ctrl.pcixbar = ((ICU_HWINT1 << ICU_PCIX_0)
			 | (ICU_HWINT1 << ICU_PCIX_1)
			 | (ICU_HWINT1 << ICU_PCIX_2)
			 | (ICU_HWINT1 << ICU_PCIX_3));

    /* build Cause bits to interrupt number mapping table */
    for (i = 0; i < 256; i++) {
	unsigned int cr = i << 8;
	int intr;
	/* test in "priority" order */
	if (cr & CR_HINT4)
	    intr = INTR_PANIC;
	else if (cr & CR_HINT5)
	    intr = INTR_CLOCK;
	else if (cr & CR_HINT3)
	    intr = INTR_FDCDMA;
	else if (cr & CR_HINT2)
	    intr = INTR_CENT;
	else if (cr & CR_HINT1)
	    intr = INTR_PCIMUX;
	else if (cr & CR_HINT0)
	    intr = INTR_DEVMUX;
	else if (cr & CR_SINT1)
	    intr = INTR_SINT1;
	else if (cr & CR_SINT0)
	    intr = INTR_SINT0;
	else
	    intr = 0;
	crToIntr[i] = intr;
    }

    /* build "local device" i/u bit to interrupt number mapping table */
    for (i = 0; i < 256; i++) {
	int intr;
	if (i & ICU_DEV_CENT)
	    intr = INTR_CENT;
	else if (i & ICU_DEV_V96)
	    intr = INTR_V96;
	else if (i & ICU_DEV_FDC)
	    intr = INTR_FDC;
	else if (i & ICU_DEV_RTC)
	    intr = INTR_RTC;
	else if (i & ICU_DEV_GPIO)
	    intr = INTR_GPIO;
	else if (i & ICU_DEV_UART1)
	    intr = INTR_UART1;
	else if (i & ICU_DEV_UART0)
	    intr = INTR_UART0;
	else if (i & ICU_DEV_KBD)
	    intr = INTR_KBD;
	else
	    intr = 0;
	devToIntr[i] = intr;
    }

    /* build PCI i/u bit to interrupt number mapping table */
    for (i = 0; i < 16; i++) {
	unsigned int pci = i << 4;
	int intr;
	if (pci & ICU_PCI_INT0)
	    intr = INTR_PCI0;
	else if (pci & ICU_PCI_INT1)
	    intr = INTR_PCI1;
	else if (pci & ICU_PCI_INT2)
	    intr = INTR_PCI2;
	else if (pci & ICU_PCI_INT3)
	    intr = INTR_PCI3;
	else
	    intr = 0;
	pciToIntr[i] = intr;
    }

    /*
     * Now build the reverse mapping, from interrupt number
     * to interrupt bit.
     */

    /* HINT4 = Debug/Panic */
    intrToIrr[INTR_PANIC]	= SR_HINT4;

    /* HINT3 = Floppy DMA */
    intrToIrr[INTR_FDCDMA]	= SR_HINT3;

    /* HINT2 = Centronics */
    intrToIrr[INTR_CENT]	= SR_HINT2;

    /* HINT1 = PCI devices */
    intrToIrr[INTR_PCI0]	= ICU_PCI_INT0 << 16;
    intrToIrr[INTR_PCI1]	= ICU_PCI_INT1 << 16;
    intrToIrr[INTR_PCI2]	= ICU_PCI_INT2 << 16;
    intrToIrr[INTR_PCI3]	= ICU_PCI_INT3 << 16;

    /* HINT0 = local devices */
    intrToIrr[INTR_V96]		= ICU_DEV_V96;
    intrToIrr[INTR_FDC]		= ICU_DEV_FDC;
    intrToIrr[INTR_UART0]	= ICU_DEV_UART0;
    intrToIrr[INTR_UART1]	= ICU_DEV_UART1;
    intrToIrr[INTR_KBD]		= ICU_DEV_KBD;
    intrToIrr[INTR_GPIO]	= ICU_DEV_GPIO;
    intrToIrr[INTR_RTC]		= ICU_DEV_RTC;

    /* SINT1/SINT0 = software interrupts */
    intrToIrr[INTR_SINT1]	= SR_SINT1;
    intrToIrr[INTR_SINT0]	= SR_SINT0;

    /* set a "safe" compare value and clear timer interrupt */
    mips_setcompare (mips_getcount() + 0x7fffffff);
}


/*
 * Handle board-specific debug or fatal interrupts
 */
static int 
panicHandler ()
{
    volatile p4032icu *icu = PA_TO_KVA1 (ICU_BASE);
    unsigned int irr = icu->irr.err;
    int sig = 0;
    int i;

    if (irr & ICU_ERR_DEBUG) {
	/* software debounce debug button */
	do {
	    icu->ctrl.clear = ICU_ERR_DEBUG; mips_wbflush ();
	    for (i = 100000; i > 0; i--) continue;
	} while (icu->irr.err & ICU_ERR_DEBUG);
	irr = icu->irr.err;
	sig = SIGINT;
    }

    /* acknowledge (clear) all noticed error interrupts */
    icu->ctrl.clear = irr;
    mips_wbflush ();

    /* handle fatal error conditions here */
    /* XXX could choose to let kernel handle them by returning zero */
    if (irr & ICU_ERR_BUSERR) {
	_mon_printf ("\nBus Error Interrupt\n");
	sig = SIGBUS;
    }
    else if (irr & ICU_ERR_ACFAIL) {
	/* XXX try to send SIGHUP to all processes? */
	_mon_printf ("\nPower Fail\n");
	sig = SIGHUP;
    }

#if 0
    if (_debug_enabled ())
	/* waiting debugger can receive signal number */
	sig = _debug_sighook (sig);
#endif

    /* if still pending, atop now */
    if (sig > 0)
	_debug_breakpoint ();

    return 1;
}


/* 
 * Called by remote debug stub when first initialised to enable "debug"
 * interrupts (h/w intr 4).
 */
void
_dbg_signals (void)
{
    int s = _mips_intdisable ();
    _sbd_ipl0_mask |= SR_HINT4;	/* enable HINT4 in user mode */
    _sbd_ipl1_sr |= SR_HINT4;	/* enable HINT4 in kernel mode */
    _sbd_iplhi_sr &= ~SR_HINT4;	/* disable HINT4 in interrupt mode */
    _intrmask |= SR_HINT4;	/* prevent PMON from handling HINT4 */
    _mips_intrestore (s);
}
