#ident "kit/p5064/pci_machdep.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1997-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * p5064/pci_machdep.c: Machine-specific functions for PCI autoconfiguration.
 */



#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <mips/xcpt.h>
#include <mips/cpu.h>
#include <pci/pcivar.h>
#include <pci/pcireg.h>
#include "kit_impl.h"
#include "sbd.h"
#include "v96xpbc.h"

#if #endian(little)
#define V96X_SWAP_MEM	V96X_SWAP_NONE
#define V96X_SWAP_IO	V96X_SWAP_NONE
#elif #endian(big)
#define V96X_SWAP_MEM	V96X_SWAP_8BIT
#define V96X_SWAP_IO	V96X_SWAP_AUTO
#endif

/* PCI i/o regions in PCI space */
#define PCI_IO_SPACE_PCI_BASE		0x00000000

/* PCI mem regions in PCI space */
#define PCI_MEM_SPACE_PCI_BASE_NEW	0x00000000
#define PCI_LOCAL_MEM_PCI_BASE_NEW	0x80000000
#define PCI_LOCAL_MEM_ISA_BASE_NEW	0x00800000

/* PCI mem regions in PCI space (old configuration) */
#define PCI_MEM_SPACE_PCI_BASE_OLD	0x10000000
#define PCI_LOCAL_MEM_PCI_BASE_OLD	0x00000000
#define PCI_LOCAL_MEM_ISA_BASE_OLD	0x00000000

/* soft versions of above */
static pcireg_t pci_mem_space	       = PCI_MEM_SPACE;
static pcireg_t pci_mem_space_pci_base = PCI_MEM_SPACE_PCI_BASE_NEW;
static pcireg_t pci_local_mem_pci_base = PCI_LOCAL_MEM_PCI_BASE_NEW;
static pcireg_t pci_local_mem_isa_base = PCI_LOCAL_MEM_ISA_BASE_NEW;

static pcireg_t pci_io_space	       = PCI_IO_SPACE;
static pcireg_t pci_io_space_pci_base  = PCI_IO_SPACE_PCI_BASE;

/* PCI mem space allocation (note - skip the 16MB ISA mem space) */
static pcireg_t minpcimemaddr;
static pcireg_t nextpcimemaddr;

/* PCI i/o space allocation (note - bottom 512KB reserved for ISA i/o space) */
static pcireg_t minpciioaddr;
static pcireg_t nextpciioaddr;

static const struct pci_bus v96x_pci_bus = {
	0,		/* minimum grant */
	255,		/* maximum latency */
	0,		/* devsel time = fast */
	1,		/* we support fast back-to-back */
	1,		/* we support prefetch */
	0,		/* we don't support 66 MHz */
	0,		/* we don't support 64 bits */
	4000000,	/* bandwidth: in 0.25us cycles / sec */
	1		/* initially one device on bus (i.e. us) */
};

/* one such structure for each PCI bus (allowing for PCI/PCI bridges) */
#define MAXBUS	3
struct pci_bus _pci_bus[MAXBUS];

__attribute__ ((section (".sdata")))
int _pci_maxbus = MAXBUS;

extern int _pciverbose;

static unsigned char	v96x_vrev;
static int		v96x_autoflush;


/*
 * Called to initialise the bridge at the beginning of time
 */
int
_pci_hwinit (int initialise)
{
    unsigned char * const _v96xp = PA_TO_KVA1(V96XPBC_BASE);
    unsigned int pci_rd0, pci_rd1;
    int pcioldmap = 0;
    char *s;

    /* initialise global data */
    v96x_vrev = V96X_PCI_CC_REV & V96X_PCI_CC_REV_VREV;
    if (v96x_vrev < V96X_VREV_B2) {
	_mon_printf ("V962 revisions < B.2 not supported\n");
	return -1;
    }

    _pci_bus[0] = v96x_pci_bus;

    /* if not initialising PCI for ourselves, then probe the existing setup */
    if (!initialise) {
	/* determine PCI address of our local memory */
	if ((V96X_PCI_MAP1 & V96X_PCI_MAPx_ENABLE)
	    && !(V96X_PCI_BASE1 & V96X_PCI_BASEx_IO))
	    pci_local_mem_pci_base = V96X_PCI_BASE1 & V96X_PCI_BASE1_ADR_BASE;
	else
	    return -1;

	/* is there a separate window from ISA mem space too? */
	if ((V96X_PCI_MAP0 & V96X_PCI_MAPx_ENABLE)
	    && !(V96X_PCI_BASE0 & V96X_PCI_BASEx_IO)
	    && (V96X_PCI_BASE1 & V96X_PCI_BASE1_ADR_BASE) < 0x1000000)
	    pci_local_mem_isa_base = V96X_PCI_BASE0 & V96X_PCI_BASE0_ADR_BASE;
	else
	    pci_local_mem_isa_base = pci_local_mem_pci_base;

	/* determine local address of PCI memory */
	if ((V96X_LB_MAP1 & V96X_LB_MAPx_TYPE) == V96X_LB_TYPE_MEM) {
	    pci_mem_space = V96X_LB_BASE1 & V96X_LB_BASEx_ADR_BASE;
	    pci_mem_space_pci_base =
		(V96X_LB_MAP1 & V96X_LB_MAPx_MAP_ADR) << 16;
	}
	else if ((V96X_LB_MAP0 & V96X_LB_MAPx_TYPE) == V96X_LB_TYPE_MEM) {
	    pci_mem_space = V96X_LB_BASE0 & V96X_LB_BASEx_ADR_BASE;
	    pci_mem_space_pci_base =
		(V96X_LB_MAP0 & V96X_LB_MAPx_MAP_ADR) << 16;
	}
	else
	    return -1;

	/* get separate LB->PCI i/o map */
	if (V96X_LB_BASE2 & V96X_LB_BASEx_ENABLE) {
	    pci_io_space = (V96X_LB_BASE2 & 0xff00) << 16;
	    pci_io_space_pci_base = V96X_LB_MAP2 << 16;
	}
	else
	    return -1;

	/* PCI readahead fifo auto flush at end of burst? */
	v96x_autoflush =
	  (v96x_vrev >= V96X_VREV_C0
	   && ((V96X_FIFO_PRIORITY & V96X_FIFO_PRIORITY_PCI_RD1) >> 
	       V96X_FIFO_PRIORITY_PCI_RD1_SHIFT) == V96X_FIFO_PRI_FLUSHBURST);

	return 0;
    }

    /* We are in control - we define the mappings */
    if (((s = _sbd_getenv ("pcimap")) || (s = _sbd_getenv ("PCIMAP")))
	&& strcmp (s, "old") == 0)
	pcioldmap = 1;
    else
	pcioldmap = 0;

    if (pcioldmap) {
	pci_local_mem_pci_base = PCI_LOCAL_MEM_PCI_BASE_OLD;
	pci_local_mem_isa_base = PCI_LOCAL_MEM_ISA_BASE_OLD;
	pci_mem_space_pci_base = PCI_MEM_SPACE_PCI_BASE_OLD;
	/* point to start and end of region */
	minpcimemaddr  = PCI_MEM_SPACE_PCI_BASE_OLD;
	nextpcimemaddr = PCI_MEM_SPACE_PCI_BASE_OLD + PCI_MEM_SPACE_SIZE;
    }
    else {
	pci_local_mem_pci_base = PCI_LOCAL_MEM_PCI_BASE_NEW;
	pci_local_mem_isa_base = PCI_LOCAL_MEM_ISA_BASE_NEW;
	pci_mem_space_pci_base = PCI_MEM_SPACE_PCI_BASE_NEW;
	/* point to start and end of region (leaving bottom 16M for ISA) */
	minpcimemaddr  = PCI_MEM_SPACE_PCI_BASE_NEW + 0x1000000;
	nextpcimemaddr = PCI_MEM_SPACE_PCI_BASE_NEW + PCI_MEM_SPACE_SIZE;
    }

    /* leave 512Kb at beginning of PCI i/o space for ISA bridge (it
       actually uses only 64K, but this is needed for a ISA DMA 
       h/w fix which needs a higher address bit to spot ISA cycles). */
    pci_io_space = PCI_IO_SPACE;
    pci_io_space_pci_base = PCI_IO_SPACE_PCI_BASE;
    minpciioaddr  = PCI_IO_SPACE_PCI_BASE + 0x80000;
    nextpciioaddr = PCI_IO_SPACE_PCI_BASE + PCI_IO_SPACE_SIZE;

    /* stop the V3 chip from servicing any further PCI requests */
    V96X_PCI_CMD = 0;
    mips_wbflush ();

    /* reset the PCI bus */
    V96X_SYSTEM &= ~V96X_SYSTEM_RST_OUT;
    
    /* enable bridge to PCI and PCI memory accesses, plus error handling */
    V96X_PCI_CMD = V96X_PCI_CMD_MASTER_EN
	| V96X_PCI_CMD_MEM_EN
	| V96X_PCI_CMD_SERR_EN
	| V96X_PCI_CMD_PAR_EN;

    /* clear errors and say we do fast back-to-back transfers */
    V96X_PCI_STAT = V96X_PCI_STAT_PAR_ERR
	| V96X_PCI_STAT_SYS_ERR	
	| V96X_PCI_STAT_M_ABORT	
	| V96X_PCI_STAT_T_ABORT	
	| V96X_PCI_STAT_PAR_REP	
	| V96X_PCI_STAT_FAST_BACK;

    /* Local to PCI aptr 0 - LOCAL:PCI_CONF_SPACE -> PCI:config (1MB) */
    V96X_LB_BASE0 = PCI_CONF_SPACE | V96X_SWAP_IO | V96X_ADR_SIZE_1MB 
	| V96X_LB_BASEx_ENABLE;

    /* Local to PCI aptr 1 - LOCAL:PCI_MEM_SPACE -> PCI:00000000-xxxx */
    V96X_LB_BASE1 = PCI_MEM_SPACE | V96X_SWAP_MEM | V96X_ADR_SIZE_128MB 
	| V96X_LB_BASEx_ENABLE;
    V96X_LB_MAP1 = (pci_mem_space_pci_base >> 16) | V96X_LB_TYPE_MEM;

    /* Local to PCI aptr 2 - LOCAL:PCI_IO_SPACE -> PCI:00000000-xxxx */
    V96X_LB_BASE2 = (PCI_IO_SPACE >> 16) | (V96X_SWAP_IO >> 2)
	| V96X_LB_BASEx_ENABLE;
    V96X_LB_MAP2 = (PCI_IO_SPACE_PCI_BASE >> 16);

    /* PCI to local aptr 1 - PCI:80000000-90000000 -> LOCAL:00000000 */
    /* 256MB window for PCI bus masters to get at our local memory */
    V96X_PCI_BASE1 = pci_local_mem_pci_base | V96X_PCI_BASEx_MEM
	| V96X_PCI_BASEx_PREFETCH;
    V96X_PCI_MAP1 =  0x00000000 | V96X_ADR_SIZE_256MB
	| V96X_SWAP_MEM /*| V96X_PCI_MAPx_RD_POST_INH*/
	| V96X_PCI_MAPx_REG_EN | V96X_PCI_MAPx_ENABLE;

    v96x_autoflush = (v96x_vrev >= V96X_VREV_C0);
    pci_rd1 = v96x_autoflush
	? V96X_FIFO_PRI_FLUSHBURST : V96X_FIFO_PRI_FLUSHALL;

    if (pcioldmap) {
	/* PCI to local aptr 0 - PCI:C0000000-D0000000 -> LOCAL:00000000 */
	/* Alternative non-prefetchable window for PCI bus masters */
	V96X_PCI_BASE0 = 0xC0000000 | V96X_PCI_BASEx_MEM;
	V96X_PCI_MAP0 =  0x00000000 | V96X_ADR_SIZE_256MB
	    | V96X_SWAP_MEM /*| V96X_PCI_MAPx_RD_POST_INH*/
	    | V96X_PCI_MAPx_REG_EN | V96X_PCI_MAPx_ENABLE;
	pci_rd0 = V96X_FIFO_PRI_NOFLUSH;
    }
    else {
	/* PCI to local aptr 0 - PCI:00800000-01000000 -> LOCAL:00000000 */
	/* 8MB window for ISA bus masters to get at our local memory */
	V96X_PCI_BASE0 = pci_local_mem_isa_base | V96X_PCI_BASEx_MEM
	    | V96X_PCI_BASEx_PREFETCH;
	V96X_PCI_MAP0 =  0x00000000 | V96X_ADR_SIZE_8MB
	    | V96X_SWAP_MEM /*| V96X_PCI_MAPx_RD_POST_INH*/
	    | V96X_PCI_MAPx_REG_EN | V96X_PCI_MAPx_ENABLE;
	pci_rd0 = v96x_autoflush
	    ? V96X_FIFO_PRI_FLUSHBURST : V96X_FIFO_PRI_FLUSHALL;
    }

    /* PCI to internal registers - disabled (but avoid address overlap) */
    V96X_PCI_IO_BASE = 0xffffff00 | V96X_PCI_IO_BASE_IO;

    /* Disable PCI_IO_BASE and set optional AD(1:0) to 01b for type1 config */
    V96X_PCI_CFG = V96X_PCI_CFG_IO_DIS | (0x1 << V96X_PCI_CFG_AD_LOW_SHIFT);

    /* setup fifo config:
       empty write fifos at end of burst;
       refill read fifos when not full (any space);
       local bus max burst 8 words (XXX could be >, & be controlled by lb);
       pci bus max burst 256 words */
    V96X_FIFO_CFG = 
	(V96X_FIFO_CFG_BRST_256 << V96X_FIFO_CFG_PBRST_MAX_SHIFT)
	| (V96X_FIFO_CFG_WR_ENDBRST << V96X_FIFO_CFG_WR_LB_SHIFT)
	| (V96X_FIFO_CFG_RD_NOTFULL << V96X_FIFO_CFG_RD_LB1_SHIFT)
	| (V96X_FIFO_CFG_RD_NOTFULL << V96X_FIFO_CFG_RD_LB0_SHIFT)
	| (V96X_FIFO_CFG_BRST_16 << V96X_FIFO_CFG_LBRST_MAX_SHIFT)
	| (V96X_FIFO_CFG_WR_ENDBRST << V96X_FIFO_CFG_WR_PCI_SHIFT)
	| (V96X_FIFO_CFG_RD_NOTFULL << V96X_FIFO_CFG_RD_PCI1_SHIFT)
	| (V96X_FIFO_CFG_RD_NOTFULL << V96X_FIFO_CFG_RD_PCI0_SHIFT);
    
    /* Set fifo priorities: note that on Rev C.0 and above we set the
       read prefetch fifos to flush at the end of a burst, and not to
       retain data like a cache (which causes coherency problems).  For
       Rev B.2 and below we can't do this, so we set them to be
       flushed by any write cycle (inefficient but safer), and we also
       require explicit software flushing of the fifos to maintain
       full coherency (i.e. call pci_flush() from the cache flush
       routines or after modifying uncached descriptors). */

    /* initial setting, may be updated in pci_hwreinit(), below  */
    V96X_FIFO_PRIORITY = 
	V96X_FIFO_PRIORITY_LOCAL_WR /* local->pci write priority (safe) */
	| V96X_FIFO_PRIORITY_PCI_WR /* pci->local write priority (safe) */ 
	| (V96X_FIFO_PRI_NOFLUSH << V96X_FIFO_PRIORITY_LB_RD0_SHIFT)
	| (V96X_FIFO_PRI_NOFLUSH << V96X_FIFO_PRIORITY_LB_RD1_SHIFT)
	| (pci_rd0 << V96X_FIFO_PRIORITY_PCI_RD0_SHIFT)
	| (pci_rd1 << V96X_FIFO_PRIORITY_PCI_RD1_SHIFT);
    
#if 0
    /* enable long local bus timeout, followed by ERR/BTERM */
    V96X_LB_CFG = V96X_LB_CFG_TO_256 | V96X_LB_CFG_ERR_EN;
#endif

    /* clear latched PCI interrupts */
    V96X_LB_ISTAT = 0;		

    /* enable V3 general interrupts */
    V96X_LB_IMASK = V96X_LB_INTR_PCI_RD|V96X_LB_INTR_PCI_WR;

    /* finally unreset the PCI bus */
    V96X_SYSTEM |= V96X_SYSTEM_RST_OUT;

    /* ... and the onboard PCI devices */
    {
	volatile p5064bcr1 * const bcr1 = PA_TO_KVA1 (BCR1_BASE);
	bcr1->eth = BCR1_ENABLE;
	bcr1->scsi = BCR1_ENABLE;
	bcr1->isa = BCR1_ENABLE;
	bcr1->pcmcia = BCR1_ENABLE;
	usdelay (1);
    }

    return 1;
}


/*
 * Called to reinitialise the bridge after we've scanned each PCI device
 * and know what is possible.
 */
void
_pci_hwreinit (void)
{
    char * const _v96xp = PA_TO_KVA1(V96XPBC_BASE);

    if (_pci_bus[0].fast_b2b)
	/* fast back-to-back is supported by all devices */
	V96X_PCI_CMD |= V96X_PCI_CMD_FBB_EN;

    /* Rev B.1+: can now use variable latency timer */
    V96X_PCI_HDR_CFG = _pci_bus[0].def_ltim << V96X_PCI_HDR_CFG_LT_SHIFT;

    if (_pci_bus[0].prefetch && v96x_vrev >= V96X_VREV_C0) {
	/* Rev C.0+: we can safely prefetch from all pci mem devices */
	V96X_LB_BASE1 |= V96X_LB_BASEx_PREFETCH;
	V96X_FIFO_PRIORITY = (V96X_FIFO_PRIORITY & ~V96X_FIFO_PRIORITY_LB_RD1)
	    | (V96X_FIFO_PRI_FLUSHBURST << V96X_FIFO_PRIORITY_LB_RD1_SHIFT);
    }

    /* clear latched PCI interrupts */
    V96X_LB_ISTAT = 0;		

    /* enable PCI read/write error interrupts */
    V96X_LB_IMASK = V96X_LB_INTR_PCI_RD | V96X_LB_INTR_PCI_WR;

}

void
_pci_flush (void)
{
    /* flush read-ahead fifos (not necessary on Rev C.0 and above) */
    if (!v96x_autoflush) {
	char * const _v96xp = PA_TO_KVA1(V96XPBC_BASE);
	V96X_SYSTEM |= 
	    V96X_SYSTEM_LB_RD_PCI1 | V96X_SYSTEM_LB_RD_PCI0 |
	    V96X_SYSTEM_PCI_RD_LB1 | V96X_SYSTEM_PCI_RD_LB0;
    }
}

void
_pci_wbflush (void)
{
    char * const _v96xp = PA_TO_KVA1(V96XPBC_BASE);

    mips_wbflush ();
    /* wait for write FIFO to empty */
    while (V96X_FIFO_STAT & V96X_FIFO_STAT_L2P_WR)
	continue;
}

pcireg_t
_pci_statusread (void)
{
    unsigned char * const _v96xp = PA_TO_KVA1(V96XPBC_BASE);
    return V96X_PCI_STAT;
}

void
_pci_statuswrite (pcireg_t r)
{
    unsigned char * const _v96xp = PA_TO_KVA1(V96XPBC_BASE);

    V96X_PCI_STAT = r;
    (void) V96X_PCI_STAT;
}



/* Map the CPU virtual address of an area of local memory to a PCI
   address that can be used by a PCI bus master to access it. */
vm_offset_t
_pci_dmamap (vm_offset_t va, unsigned int len)
{
    return pci_local_mem_pci_base + KVA_TO_PA (va);
}

/* Map the PCI address of an area of local memory to a CPU physical
   address. */
vm_offset_t
_pci_cpumap (vm_offset_t pcia, unsigned int len)
{
    return pcia - pci_local_mem_pci_base;
}


/* Map the CPU virtual address of an area of local memory to an ISA
   address that can be used by a ISA bus master to access it. */
vm_offset_t
_isa_dmamap (vm_offset_t va, unsigned int len)
{
    unsigned long pa = KVA_TO_PA (va);

    /* restrict ISA DMA access to bottom 8/16MB of local memory */
    if (pa + len > 0x1000000 - pci_local_mem_isa_base)
	return (vm_offset_t)-1;
    return pci_local_mem_isa_base + pa;
}

/* Map the ISA address of an area of local memory to a CPU physical
   address. */
vm_offset_t
_isa_cpumap (vm_offset_t pcia, unsigned int len)
{
    return pcia - pci_local_mem_isa_base;
}


pcitag_t
_pci_make_tag(int bus, int device, int function)
{
    pcitag_t tag;
    tag = (bus << 16) | (device << 11) | (function << 8);
    return tag;
}

void __inline__
_pci_break_tag(pcitag_t tag, int *busp, int *devicep, int *functionp)
{
    if (busp) *busp = (tag >> 16) & 255;
    if (devicep) *devicep = (tag >> 11) & 31;
    if (functionp) *functionp = (tag >> 8) & 7;
}

int
_pci_canscan (pcitag_t tag)
{
    return (1);
}

static pcireg_t
_pci_conf_readn(pcitag_t tag, int reg, int width)
{
    char * const _v96xp = PA_TO_KVA1(V96XPBC_BASE);
    u_int32_t addr, ad_low;
    void *addrp;
    pcireg_t data;
    int bus, device, function;

    if (reg & (width-1) || reg < 0 || reg >= 0x100) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_readn: bad reg 0x%x\r\n", reg);
	return ~0;
    }

    _pci_break_tag (tag, &bus, &device, &function); 
    if (bus == 0) {
	/* Type 0 configuration on onboard PCI bus */
	if (device > 5 || function > 7)
	    return ~0;		/* device out of range */
	addr = (1 << (device+24)) | (function << 8) | reg;
	ad_low = 0;
    }
    else if (v96x_vrev >= V96X_VREV_C0) {
	/* Type 1 configuration on offboard PCI bus */
	if (bus > 255 || device > 31 || function > 7)
	    return ~0;	/* device out of range */
	addr = (bus << 16) | (device << 11) | (function << 8) | reg;
	ad_low = V96X_LB_MAPx_AD_LOW_EN;
    }
    else {
	return ~0;		/* bus out of range */
    }

    /* high 12 bits of address go in map register; set conf space */
    V96X_LB_MAP0 = ((addr >> 16) & V96X_LB_MAPx_MAP_ADR)
	| ad_low | V96X_LB_TYPE_CONF;

    /* clear aborts */
    V96X_PCI_STAT |= V96X_PCI_STAT_M_ABORT | V96X_PCI_STAT_T_ABORT; 

    mips_wbflush ();

    /* low 20 bits of address are in the actual address */
    addrp = PA_TO_KVA1(PCI_CONF_SPACE + (addr&0xfffff));
    switch (width) {
    case 1:
	data = (pcireg_t)*(volatile unsigned char *)addrp;
	break;
    case 2:
	data = (pcireg_t)*(volatile unsigned short *)addrp;
	break;
    default:
    case 4:
	data = (pcireg_t)*(volatile unsigned int *)addrp;
	break;
    }

    if (V96X_PCI_STAT & V96X_PCI_STAT_M_ABORT) {
	V96X_PCI_STAT |= V96X_PCI_STAT_M_ABORT;
#if 0
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: master abort\r\n");
#endif
	return ~0;
    }

    if (V96X_PCI_STAT & V96X_PCI_STAT_T_ABORT) {
	V96X_PCI_STAT |= V96X_PCI_STAT_T_ABORT;
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: target abort\r\n");
	return ~0;
    }

    return data;
}


pcireg_t
_pci_conf_read8(pcitag_t tag, int reg)
{
    return _pci_conf_readn (tag, reg, 1);
}

pcireg_t
_pci_conf_read16(pcitag_t tag, int reg)
{
    return _pci_conf_readn (tag, reg, 2);
}

pcireg_t
_pci_conf_read32(pcitag_t tag, int reg)
{
    return _pci_conf_readn (tag, reg, 4);
}

static void
_pci_conf_writen(pcitag_t tag, int reg, pcireg_t data, int width)
{
    char * const _v96xp = PA_TO_KVA1(V96XPBC_BASE);
    u_int32_t addr, ad_low;
    void *addrp;
    int bus, device, function;

    if (reg & (width-1) || reg < 0 || reg >= 0x80) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_read: bad reg %x\r\n", reg);
	return;
    }

    _pci_break_tag (tag, &bus, &device, &function);

    if (bus == 0) {
	/* Type 0 configuration on onboard PCI bus */
	if (device > 5 || function > 7)
	    return;		/* device out of range */
	addr = (1 << (device+24)) | (function << 8) | reg;
	ad_low = 0;
    }
    else if (v96x_vrev >= V96X_VREV_C0) {
	/* Type 1 configuration on offboard PCI bus */
	if (bus > 255 || device > 31 || function > 7)
	    return;	/* device out of range */
	addr = (bus << 16) | (device << 11) | (function << 8) | reg;
	ad_low = V96X_LB_MAPx_AD_LOW_EN;
    }
    else
	return;			/* bus out of range */

    /* high 12 bits of address go in map register; set conf space */
    V96X_LB_MAP0 = ((addr >> 16) & V96X_LB_MAPx_MAP_ADR)
	| ad_low | V96X_LB_TYPE_CONF;

    /* clear aborts */
    V96X_PCI_STAT |= V96X_PCI_STAT_M_ABORT | V96X_PCI_STAT_T_ABORT; 

    mips_wbflush ();

    /* low 20 bits of address are in the actual address */
    addrp = PA_TO_KVA1(PCI_CONF_SPACE + (addr&0xfffff));
    switch (width) {
    case 1:
	*(volatile unsigned char *)addrp = data;
	break;
    case 2:
	*(volatile unsigned short *)addrp = data;
	break;
    default:
    case 4:
	*(volatile unsigned int *)addrp = data;
	break;
    }

    mips_wbflush ();

    /* wait for write FIFO to empty */
    while (V96X_FIFO_STAT & V96X_FIFO_STAT_L2P_WR)
	continue;

    if (V96X_PCI_STAT & V96X_PCI_STAT_M_ABORT) {
	V96X_PCI_STAT |= V96X_PCI_STAT_M_ABORT;
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_write: master abort\r\n");
    }

    if (V96X_PCI_STAT & V96X_PCI_STAT_T_ABORT) {
	V96X_PCI_STAT |= V96X_PCI_STAT_T_ABORT;
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_conf_write: target abort\r\n");
    }
}

void
_pci_conf_write8(pcitag_t tag, int reg, pcireg_t data)
{
    _pci_conf_writen (tag, reg, data, 1);
}

void
_pci_conf_write16(pcitag_t tag, int reg, pcireg_t data)
{
    _pci_conf_writen (tag, reg, data, 2);
}

void
_pci_conf_write32(pcitag_t tag, int reg, pcireg_t data)
{
    _pci_conf_writen (tag, reg, data, 4);
}



int
_pci_map_port(pcitag_t tag, int reg, unsigned int *port)
{
    pcireg_t address;
    
    if (reg < PCI_MAP_REG_START || reg >= PCI_MAP_REG_END || (reg & 3)) {
	if (_pciverbose >= 1)
	    _pci_tagprintf(tag, "_pci_map_port: bad request\r\n");
	return -1;
    }
    
    address = _pci_conf_read(tag, reg);
    
    if ((address & PCI_MAP_IO) == 0) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_port: attempt to i/o map a memory region\r\n");
	return -1;
    }

    *port = (address & PCI_MAP_IO_ADDRESS_MASK) - pci_io_space_pci_base;
    return 0;
}


int
_pci_map_io(pcitag_t tag, int reg, vm_offset_t *vap, vm_offset_t *pap)
{
    pcireg_t address;
    vm_offset_t pa;
    
    if (reg < PCI_MAP_REG_START || reg >= PCI_MAP_REG_END || (reg & 3)) {
	if (_pciverbose >= 1)
	    _pci_tagprintf(tag, "_pci_map_io: bad request\r\n");
	return -1;
    }
    
    address = _pci_conf_read(tag, reg);
    
    if ((address & PCI_MAP_IO) == 0) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_io: attempt to i/o map a memory region\r\n");
	return -1;
    }

    pa = (address & PCI_MAP_IO_ADDRESS_MASK) - pci_io_space_pci_base;
    *pap = pa;
    *vap = (vm_offset_t) PA_TO_KVA1 (pci_io_space + pa);
    
    if (_pciverbose >= 3)
	_pci_tagprintf(tag, "_pci_map_io: mapping i/o at virtual %08x, physical %08x\n", 
		       *vap, *pap);

    return 0;
}


int
_pci_map_mem(pcitag_t tag, int reg, vm_offset_t *vap, vm_offset_t *pap)
{
    pcireg_t address;
    vm_offset_t pa;

    if (reg == PCI_MAP_REG_ROM) {
	/* expansion ROM */
	address = _pci_conf_read(tag, reg);
	if (!(address & PCI_MAP_ROM)) {
	    _pci_tagprintf (tag, "_pci_map_mem: attempt to map missing rom\r\n");
	    return -1;
	}
	pa = address & PCI_MAP_ROM_ADDRESS_MASK;
    }
    else {
	if (reg < PCI_MAP_REG_START || reg >= PCI_MAP_REG_END || (reg & 3)) {
	    if (_pciverbose >= 1)
		_pci_tagprintf(tag, "_pci_map_mem: bad request\r\n");
	    return -1;
	}
	
	address = _pci_conf_read(tag, reg);
	
	if ((address & PCI_MAP_IO) != 0) {
	    if (_pciverbose >= 1)
		_pci_tagprintf(tag, "pci_map_mem: attempt to memory map an I/O region\r\n");
	    return -1;
	}
	
	switch (address & PCI_MAP_MEMORY_TYPE_MASK) {
	case PCI_MAP_MEMORY_TYPE_32BIT:
	case PCI_MAP_MEMORY_TYPE_32BIT_1M:
	    break;
	case PCI_MAP_MEMORY_TYPE_64BIT:
	    if (_pciverbose >= 1)
		_pci_tagprintf (tag, "_pci_map_mem: attempt to map 64-bit region\r\n");
	    return -1;
	default:
	    if (_pciverbose >= 1)
		_pci_tagprintf (tag, "_pci_map_mem: reserved mapping type\r\n");
	    return -1;
	}
	pa = address & PCI_MAP_MEMORY_ADDRESS_MASK;
    }

    pa -= pci_mem_space_pci_base;
    *pap = pa;
    if (address & PCI_MAP_MEMORY_CACHABLE) 
	*vap = (vm_offset_t) PA_TO_KVA0 (PCI_MEM_SPACE + pa);
    else
	*vap = (vm_offset_t) PA_TO_KVA1 (PCI_MEM_SPACE + pa);

    if (_pciverbose >= 3)
	_pci_tagprintf (tag, "_pci_map_mem: mapping memory at virtual 0x%x, physical 0x%x\r\n", 
			*vap, *pap);
    return 0;
}


/* Return interrupt number */
int
_pci_map_int (pcitag_t tag)
{
#ifdef INTR_PCI0
    static const int sbdintr[4] = {INTR_PCI0, INTR_PCI1,
				   INTR_PCI2, INTR_PCI3};
#endif
    pcireg_t data;
    int pin, bus, device;

    data = _pci_conf_read(tag, PCI_INTERRUPT_REG);
    pin = PCI_INTERRUPT_PIN(data);
    if (pin == 0)
	/* No IRQ used. */
	return 0;

    if (--pin > 3) {
	if (_pciverbose >= 1)
	    _pci_tagprintf (tag, "_pci_map_int: bad pin %d\r\n", pin);
	return -1;
    }

    /* ascend bus hierarchy until we get back to our (host) controller */
    _pci_break_tag (tag, &bus, &device, NULL);
    while (bus != 0) {
	/* XXX assume interrupt pins just pass straight through */
	tag = _pci_bus[bus].tag;
	_pci_break_tag (tag, &bus, &device, NULL);
    }

#ifdef INTR_PCI0
    switch (device) {
    case 13:		/* ethernet */
	return INTR_ETH;
    case 14:		/* scsi */
	return INTR_SCSI;
    case 15:		/* isa bridge */
	pin += 2;
	break;
    case 16:		/* slot 3 */
	pin += 0;
	break;
    case 17:		/* slot 2 */
	pin += 3;
	break;
    case 18:		/* slot 1 */
	pin += 2;
	break;
    default:
	return -1;
    }

    return sbdintr[pin % 4];
#else 
    return -1;
#endif    
}

pcireg_t
_pci_allocate_mem (pcitag_t tag, size_t size)
{
    pcireg_t address;

    /* allocate downwards, then round to size boundary */
    address = (nextpcimemaddr - size) & ~(size - 1);
    if (address > nextpcimemaddr || address < minpcimemaddr)
	return -1;
    nextpcimemaddr = address;
    return address;
}


pcireg_t
_pci_allocate_io (pcitag_t tag, size_t size)
{
    pcireg_t address;

    /* allocate downwards, then round to size boundary */
    address = (nextpciioaddr - size) & ~(size - 1);
    if (address > nextpciioaddr || address < minpciioaddr)
	return -1;
    nextpciioaddr = address;
    return address;
}



/* 
 * Handle mapping of fixed ISA addresses 
 */

void *
_isa_map_io (unsigned int port)
{
    if (pci_io_space_pci_base != 0
	|| port > 0xffff)
	return 0;
   return PA_TO_KVA1 (pci_io_space + port);
}

void *
_isa_map_mem (vm_offset_t addr)
{
    if (pci_mem_space_pci_base != 0
	|| pci_local_mem_pci_base < 0x100000 
	|| addr > 0xfffff)
	return 0;
    return PA_TO_KVA1 (pci_mem_space + addr);
}
