#ident "kit/p5064/sbddimm.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1997-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * P5064/sbddimm.c: DIMM EEPROM support for P5064
 */


#include <stddef.h>
#include <mips/cpu.h>
#include "sbd.h"
#include "pio.h"
#include "dimm.h"

/*
 * This is the code to read the I2C EEROM on the DIMMS
 */
	
#define EEABITS		8	/* 256x8 = 8 address bits */
#define PGSHIFT		7
#define PGMASK		(((0x1 << EEABITS-8) - 1) << 1)
#define PGNUM(o)	(((o) >> PGSHIFT) & PGMASK)
#define OMASK		0xff
#define OFFS(o)		((o) & OMASK)
#define READ		0x01
#define WRITE		0x00
	

/*
 * Max chip frequency = 100 kHz, so we require at least 5 usec
 * delay between signal changes.
 */
#define WAIT() usdelay(6)
	
/* set SCL high */
static inline void SCL_HI(void) {WAIT(); _piob_bis(PIO_I2C_SCL);}
	
/* set SCL low */
static inline void SCL_LO(void) {WAIT(); _piob_bic(PIO_I2C_SCL);}
	
/* set SDA high */
static inline void SDA_HI(void) {WAIT(); _piob_bis(PIO_I2C_SDA);}
	
/* set SDA low */
static inline void SDA_LO(void) {WAIT(); _piob_bic(PIO_I2C_SDA);}

/* disable SDA output driver (tristate) */
static inline void SDI(void) {
    WAIT(); \
    _piob_dir(PIO_OMASK);
    _piob_bic(PIO_I2C_DIR);
}
	
/* enable SDA output driver (tristate) */
static inline void SDO(void) {
    WAIT();
    _piob_dir(PIO_I2C_SDA | PIO_OMASK);
    _piob_bis(PIO_I2C_DIR);
}

/* read SDA */
static inline int SDRD (void) {
    WAIT();
    return (_piob_get() >> 7) & 1;
}
	
/* send a START: SDA high->low when SCL high */
static inline void START(void) {
    SDO();
    SDA_HI();
    SCL_HI();
    SDA_LO();
    SCL_LO();
}

/* send a STOP: SDA low->high when SCL high */
static inline void STOP(void)
{
    SDO();
    SDA_LO();
    SCL_HI();
    SDA_HI();
    SCL_LO();
    SDA_LO();
    SDI();
}

/* receive an ACK: a single 0 */
static inline int GETACK(void)
{
    int v;
    SCL_HI();
    v = SDRD();
    SCL_LO();
    return v ^ 1;
}

/* send an ACK: a single 0 bit */
static inline void SENDACK(void)
{
    SDO();
    SDA_LO();
    SCL_HI();
    SCL_LO();
    SDI();
}
	
/* send a NACK: a single 1 bit */
static inline void SENDNACK(void)
{
    SDO();
    SDA_HI();
    SCL_HI();
    SCL_LO();
    SDA_LO();
    SDI();
}

/* send 8 bit word (note: check for ACK externally) */
static void
i2csend8 (int v)
{
    int i;
    SDO();
    for (i = 0; i < 8; i++) {
	if (v & 0x80)
	    SDA_HI();
	else
	    SDA_LO();
	SCL_HI();
	SCL_LO();
	v <<= 1;
    }
    SDA_LO();
    SDI();
}

/* receive 8 bit word into v0 (note: send ACK/NACK externally) */
static int
i2cget8 (void)
{
    int v;
    int i;
    v = 0;
    for (i = 0; i < 8; i++) {
	SCL_HI();
	v = (v << 1) | SDRD();
	SCL_LO();
    }
    return (v);
}

/* send first command byte: device address & page number */
static int
DADDR(int rw, int dev, int addr)
{
    START();
    i2csend8 (0xa0 | (dev << 1) | ((addr >> PGSHIFT) & PGMASK) | rw);
    return GETACK();
}

/* send two command bytes */
static int
ADDR(int rw, int dev, int offset)
{
    int i, acked;
    /* send first byte: device address */
    for (i = 0; i < 1000; i++) {
	acked = DADDR(rw, dev, offset);
	if (acked)
	    break;
    }

    if (!acked)
	return 0;

    /* send second byte: word offset */
    i2csend8 (offset & OMASK);
    return GETACK();
}

void
_sbd_i2creset (void)
{
    int i;
    for (i = 0; i < 9; i++)
	STOP();
}

int 
_sbd_i2cread (int dev, int offs)
{	

    int v;

    if (!ADDR(WRITE, dev, offs))
	return -1;
    
    if (!DADDR(READ, dev, offs))
	return -1;
	
    v = i2cget8 ();
	
    SENDNACK();
    STOP();
    return v;
}

int 
_sbd_i2cwrite (int dev, int offs, int v)
{	
    if (!ADDR(WRITE, dev, offs))
	return -1;
    
    i2csend8 (v);
	
    if (!GETACK())
	return -1;
    STOP();
    return 0;
}
