#ident "kit/p5064/sbdfreq.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1997-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * P5064/sbdfreq.c: CPU frequency determination for Algorithmics P5064 board
 */


#include <stddef.h>
#include <time.h>
#include <errno.h>
#include <mips/cpu.h>
#include <mips/prid.h>

#include "kit_impl.h"
#include "sbd.h"
#include "rtc.h"

#define RTC_HZ		16
#define RTC_RATE	RTC_RATE_16Hz

/*
 * Dynamically determine CPU pipeline clock frequency by running the
 * on-chip timer against the Real Time Clock periodic timer.
 */

long
_sbd_pipefreq (void)
{
    static long pipefreq = 0;
    volatile p5064icu * const icu = PA_TO_KVA1 (ICU_BASE);
    unsigned long start, cnt, timeout;
    unsigned int osta, ostb;
    int i;

    if (pipefreq > 0)
	return pipefreq;

    /* set RTC periodic interrupt frequency */
    osta = _rtc_bic (RTC_STATUSA, RTC_RATE_MASK);
    _rtc_bis (RTC_STATUSA, RTC_RATE);

    /* enable periodic interrupts */
    ostb = _rtc_bis (RTC_STATUSB, RTCSB_PIE);

    /* run more than once to be sure that we're in the cache */
    for (i = 2, cnt = 0; i != 0; i--) {
	start = cnt;
	_rtc_get (RTC_INTR);		/* clear any pending rtc i/u */
	/* wait until we see an RTC i/u */
	timeout = 10000000 / RTC_HZ;
	while ((icu->irr.dev & ICU_DEV_RTC) == 0)
	    if (--timeout == 0) return 0;
	cnt = mips_getcount ();
    }
    cnt -= start;

    /* restore old rate and interrupt enables */
    _rtc_set (RTC_STATUSA, osta);
    _rtc_set (RTC_STATUSB, ostb);

    /* work out number of counter ticks per sec */
    pipefreq = cnt * RTC_HZ;

    /* pipeline runs at 2x counter rate */
    pipefreq *= 2;

    return pipefreq;
}


struct ratio {char mult; char div;};
static const struct ratio ratios_norml[8] = {{2,1}, {3,1}, {4,1}, {5,1},
					     {6,1}, {7,1}, {8,1}, {1,1}};
static const struct ratio ratios_r4300[8] = {{2,1}, {3,1}, {4,1}, {5,1},
					     {6,1}, {7,1}, {1,1}, {3,2}};
static const struct ratio ratios_r54xx[8] = {{2,1}, {5,2}, {3,1}, {4,1},
					     {5,1}, {6,1}, {7,1}, {1,1}};
static const struct ratio ratios_rm7k2[8] = {{2,1}, {3,1}, {4,1}, {5,2},
					     {6,1}, {7,2}, {8,1}, {9,1}};

/*
 * Return the CPU bus clock frequency (usually same input clock).
 */

long
_sbd_cpufreq (void)
{
    static long cpufreq;
    if (cpufreq == 0) {
	long freq = _sbd_pipefreq ();
#ifdef CFG_ECMASK
	unsigned int prid = mips_getprid();
	unsigned int cfg = mips_getconfig();
	if (((prid >> 16) & 0xff) == 0) {
	    /* pipeline runs at multiple of input clock, so divide down */
	    const struct ratio *ratios;
	    switch ((prid >> 8) & 0xff) {
	    case PRID_R4300:
		ratios = ratios_r4300;
		break;
	    case PRID_R5400:
		ratios = ratios_r54xx;
		break;
	    case PRID_RM7000:
		/* XXX ugly, but may not be including rm7kc0.h */
		if (cfg & 0x00010000) {
		    ratios = ratios_rm7k2;
		    break;
		}
	    default:
		ratios = ratios_norml;
		break;
	    }
	    ratios += (cfg & CFG_ECMASK) >> CFG_ECSHIFT;
	    freq = freq * ratios->div / ratios->mult;
	}
	else
#endif
	{
	    /* MIPS32 doesn't have EC (clock divider bits) in Config */
	    /* XXX so how do we work out the divider */
	}
	cpufreq = freq;
    }
    return cpufreq;
}
