#ident "kit/p5064/sbdfrom.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1997-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * P5064/sbdfrom.c: Flash ROM support for P5064
 */


#ifdef IN_PMON
#include "pmon.h"
#include "sde-compat.h"
#include <assert.h>
#else
#include <sys/types.h>
#include <mips/cpu.h>
#include <stdlib.h>
#include <string.h>
#include "kit_impl.h"
#endif

#include "sbd.h"
#include "sbdfrom.h"
#include "flashdev.h"
#include "flashrom.h"

/* writable cookies */
static struct fromcookie flash_cookie;
static struct fromcookie eprom_cookie;
static struct fromcookie boot_cookie;

/* board-specific code */
static int fromgetparts (flashcookie_t, struct flashparts *);
static int fromsetparts (flashcookie_t, const struct flashparts *);

static const struct flashboard fromboard = {
    fromgetparts,
    fromsetparts
};


flashcookie_t
_sbd_flashopen (paddr_t devaddr)
{
    struct fromcookie *fcp;
    paddr_t base;

    if (devaddr == 0 ||
	(devaddr >= FLASH_BASE && devaddr < FLASH_BASE + 0x100000)) {
	fcp = &flash_cookie;
	base = FLASH_BASE;
    }
    else if (devaddr == 1 || 
	     (devaddr >= EPROM_BASE && devaddr < EPROM_BASE + 0x100000)) {
	fcp = &eprom_cookie;
	base = EPROM_BASE;
    }
    else if (devaddr >= BOOTPROM_BASE && devaddr < BOOTPROM_BASE + 0x100000) {
	/* XXX could be either, depends on link */
	fcp = &boot_cookie;
	base = BOOTPROM_BASE;
    }
    else
	return FLASH_UNKNOWN;

    if (fcp->size == 0) {
	/* programming base address is different from read address */
	fcp->pbase = PA_TO_KVA1 (base) - PROGRAM_OFFS;
	if (IS_KVA0 (fcp))
	    fcp->rbase = PA_TO_KVA0 (base);
	else
	    fcp->rbase = PA_TO_KVA1 (base);
	
	fcp->nbanks = 1;
	
	/* probe for device */
	if (_flashrom_probe (fcp, 0) < 0)
	    return FLASH_MISSING;
	
	/* compute total size */
	fcp->size = fcp->devsize << FROM_NCOLS_LOG2;
    }

    /* check we still fit */
    if ((devaddr & 0xfffff) >= fcp->size)
	return FLASH_MISSING;

    /* plug in board-specific functions */
    fcp->common.board = &fromboard;

    return &fcp->common;
}


/*
 * Return flash partition table 
 */
static int
fromgetparts (flashcookie_t cookie, struct flashparts *parts)
{
    struct fromcookie *fcp = (struct fromcookie *)cookie;

    memset (parts, 0, sizeof (*parts));
    _flashdev_setpart (parts, FLASHPART_RAW, FLASHPART_RAW, 0, fcp->size);
    _flashdev_setpart (parts, FLASHPART_BOOT, FLASHPART_BOOT, 0, 0x68000);
    _flashdev_setpart (parts, FLASHPART_POST, FLASHPART_POST, 
		       0x68000, 0x18000);

    if (cookie == &flash_cookie.common) {
	/* find at least 512 bytes at top of flash for environment */
	struct flashinfo info;
	unsigned int eoffs, esize;

	flash_info (cookie, 0, &info);
	eoffs = info.size; esize = 0; 
	while (esize < 512) {
	    flash_info (cookie, eoffs - 1, &info);
	    eoffs = info.soffs;
	    if (info.sprot) /* must be writable */
		esize = 0;
	    else
		esize += info.ssize;
	}
	_flashdev_setpart (parts, FLASHPART_ENV, FLASHPART_ENV, 
			   eoffs, esize);
	if (eoffs < 0x80000)
	    /* reduce POST size for environment */
	    parts->part[FLASHPART_POST].size = eoffs
		- parts->part[FLASHPART_POST].offs;
	else
	    /* spare partition */
	    _flashdev_setpart (parts, FLASHPART_FFS, FLASHPART_UNDEF, 
			       0x80000, eoffs - 0x80000);
    }
    else {
	if (fcp->size > 0x80000)
	    /* spare partition */
	    _flashdev_setpart (parts, FLASHPART_FFS, FLASHPART_UNDEF, 
			       0x80000, fcp->size - 0x80000);
    }

    return FLASHDEV_OK;
}


static int
fromsetparts (flashcookie_t cookie, const struct flashparts *parts)
{
    /* not supported yet, sorry */
    return FLASHDEV_FAIL;
}
