#ident "kit/p6032/sbdbflash.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2005 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * p6032/sbdbflash.c: Boot flash ROM support for P-6032
 */


#ifdef IN_PMON
#include <sys/types.h>
#include <mips/cpu.h>
#include "pmon.h"
#include <assert.h>
#else
#include <sys/types.h>
#include <mips/cpu.h>
#include <stdlib.h>
#include <string.h>
#include "kit_impl.h"
#endif

#include "flashdev.h"
#include "flashrom.h"
#include "sbd.h"

extern int _flash8_probe (struct fromcookie *, unsigned int);
extern int _flash16_probe (struct fromcookie *, unsigned int);

/* writable cookie */
static struct fromcookie flash_cookie;
static struct fromcookie boot_cookie;

/* board-specific code */
static int fromgetparts (flashcookie_t, struct flashparts *);
static int fromsetparts (flashcookie_t, const struct flashparts *);

static const struct flashboard fromboard = {
    fromgetparts,
    fromsetparts
};


static char * const _bonito = PA_TO_KVA1(BONITO_BASE);

flashcookie_t
_sbd_bflashopen (paddr_t devaddr)
{
    struct fromcookie *fcp;
    paddr_t base;

    if (devaddr == 0
	 || (devaddr >= SOCKET_BASE && devaddr < SOCKET_BASE + SOCKET_SIZE)) {
	fcp = &flash_cookie;
	base = SOCKET_BASE;
    }
    else if (devaddr >= BOOT_BASE && devaddr < BOOT_BASE + BOOT_SIZE
	     && (BONITO_BONPONCFG & BONITO_BONPONCFG_ROMBOOT) == BONITO_BONPONCFG_ROMBOOT_SOCKET) {
	/* XXX boot window maps to start of socket */
	fcp = &boot_cookie;
	base = BOOTPROM_BASE;
    }
    else
	return FLASH_UNKNOWN;

    if (fcp->size == 0) {
	/* always program via flash address */
	fcp->pbase = PA_TO_KVA1 (SOCKET_BASE);

	/* read from there too, but cached */
	if (IS_KVA0 (fcp))
	    fcp->rbase = PA_TO_KVA0 (SOCKET_BASE);
	else
	    fcp->rbase = PA_TO_KVA1 (SOCKET_BASE);
	
	fcp->nbanks = 1;
	
	/* probe for device */
	if (BONITO_BONPONCFG & BONITO_BONPONCFG_ROMCS0WIDTH) {
	    if (_flash16_probe (fcp, 0) < 0)
		return FLASH_MISSING;
	}
	else {
	    if (_flash8_probe (fcp, 0) < 0)
		return FLASH_MISSING;
	}
	
	/* compute total size */
	fcp->size = fcp->devsize;	/* one device */

	/* device may also be mappable */
	fcp->mapbase = base;
    }

    /* check we still fit */
    if ((devaddr & 0xfffff) >= fcp->size)
	return FLASH_MISSING;

    /* plug in board-specific functions */
    fcp->common.board = &fromboard;

    return &fcp->common;
}


/*
 * Return flash partition table 
 */
static int
fromgetparts (flashcookie_t cookie, struct flashparts *parts)
{
    struct fromcookie *fcp = (struct fromcookie *)cookie;

    memset (parts, 0, sizeof (*parts));
    _flashdev_setpart (parts, FLASHPART_RAW, FLASHPART_RAW, 0, fcp->size);
    _flashdev_setpart (parts, FLASHPART_BOOT, FLASHPART_BOOT, 0, 0x68000);
    _flashdev_setpart (parts, FLASHPART_POST, FLASHPART_POST, 
		       0x68000, 0x18000);
#if _SBD_FLASHENV==0
    {
	/* find at least 512 bytes at top of flash for environment */
	struct flashinfo info;
	unsigned int eoffs, esize;

	flash_info (cookie, 0, &info);
	eoffs = info.size; esize = 0; 
	while (esize < 512) {
	    flash_info (cookie, eoffs - 1, &info);
	    eoffs = info.soffs;
	    if (info.sprot) /* must be writable */
		esize = 0;
	    else
		esize += info.ssize;
	}
	_flashdev_setpart (parts, FLASHPART_ENV, FLASHPART_ENV, 
			   eoffs, esize);
	if (eoffs < 0x80000)
	    /* reduce POST size for environment */
	    parts->part[FLASHPART_POST].size = eoffs
		- parts->part[FLASHPART_POST].offs;
	else
	    /* spare partition */
	    _flashdev_setpart (parts, FLASHPART_FFS, FLASHPART_UNDEF, 
			       0x80000, eoffs - 0x80000);
    }
#else
    if (fcp->size > 0x80000)
	/* spare partition */
	_flashdev_setpart (parts, FLASHPART_FFS, FLASHPART_UNDEF, 
			   0x80000, fcp->size - 0x80000);
#endif

    return FLASHDEV_OK;
}


static int
fromsetparts (flashcookie_t cookie, const struct flashparts *parts)
{
    /* not supported yet, sorry */
    return FLASHDEV_FAIL;
}
