#ident "kit/p6032/sbdcache.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * p6032/sbdcache.c: board-specific cache functions for P-6032 board
 */


#include <stddef.h>
#include <mips/cpu.h>
#include "sbd.h"

static char * const _bonito = PA_TO_KVA1(BONITO_BASE);

/*
 * Bonito i/o buffer cache handling
 *
 * The i/o buffer cache snoops on uncached cycles by the MIPS cpu, so
 * we don't need to do anything for uncached CPU acceses (e.g. command
 * and status descriptors). But we do have to explicitly write-back
 * and invalidate buffer cache lines if we have used cached CPU accesses.
 */

/* FIXME these definitions should be in bonito.h! */
#define CACHECMD_INVAL	0
#define CACHECMD_WBINV	1
#define CACHECMD_RDTAG	2
#define CACHECMD_WQFLUSH 3

#define TAG_LOCK	0x80000000
#define TAG_WBACK	0x40000000
#define TAG_PFPEND	0x20000000
#define TAG_PEND	0x10000000
#define TAG_MOD		0x08000000
#define TAG_PFDVAL	0x04000000
#define TAG_DVAL	0x02000000
#define TAG_AVAL	0x01000000
#define TAG_ADDR	0x00ffffff


/*
 * Execute an i/o buffer cache command 
 */
 
static void
bonito_iobc_cmd (unsigned int cmd, unsigned int line)
{
    unsigned int ctrl;
	
    ctrl = (cmd << BONITO_PCICACHECTRL_CACHECMD_SHIFT)
	| (line << BONITO_PCICACHECTRL_CACHECMDLINE_SHIFT);
    
    BONITO_PCICACHECTRL = ctrl;
    BONITO_PCICACHECTRL = ctrl | BONITO_PCICACHECTRL_CMDEXEC;
    while (BONITO_PCICACHECTRL & BONITO_PCICACHECTRL_CMDEXEC)
	continue;
    BONITO_PCICACHECTRL = ctrl;
}


/*
 * Writeback and invalidate any i/o buffer cache lines which overlap
 * the specified address range. 
 */
 
void
_bonito_iobc_wbinv (unsigned int pa, size_t nb)
{
    unsigned int line;
    unsigned int tag;
    
    for (line = 0; line < 4; line++) {
	bonito_iobc_cmd (CACHECMD_RDTAG, line);
	tag = BONITO_PCICACHETAG;
	if (tag & TAG_AVAL) {
	    unsigned int tagaddr = (tag & TAG_ADDR) << 5;
	    if (nb == ~(size_t)0 ||
		(tagaddr < pa + nb && tagaddr + 32 > pa))
		bonito_iobc_cmd (CACHECMD_WBINV, line);
	}
    }
    bonito_iobc_cmd (CACHECMD_WQFLUSH, 0);
}


/*
 * Invalidate (no writeback!) any i/o buffer cache lines which overlap
 * the specified address range. 
 *
 * XXX This function should never be used in a running system, because
 * in a race the line that you pick could be reused by another PCI
 * device, and you could then invalidate a dirty line.
 */
 
void
_bonito_iobc_inval (unsigned int pa, size_t nb)
{
    unsigned int line;
    unsigned int tag;
    
    for (line = 0; line < 4; line++) {
	bonito_iobc_cmd (CACHECMD_RDTAG, line);
	tag = BONITO_PCICACHETAG;
	if (tag & TAG_AVAL) {
	    unsigned int tagaddr = (tag & TAG_ADDR) << 5;
	    if (nb == ~(size_t)0 ||
		(tagaddr < pa + nb && tagaddr + 32 > pa))
		bonito_iobc_cmd (CACHECMD_INVAL, line);
	}
    }
}


void
_bonito_clean_dcache (void *addr, size_t nb)
{
    if (IS_KVA0 (addr)) {
	_bonito_iobc_wbinv (KVA_TO_PA (addr), nb);
	mips_clean_dcache ((vaddr_t)addr, nb);
    }
}


void
_bonito_inval_dcache (void *addr, size_t nb)
{
    if (IS_KVA0 (addr)) {
	/* when invalidating cpu data cache, we still need to make
	   sure that the i/o buffer cache has been written back. */
	_bonito_iobc_wbinv (KVA_TO_PA (addr), nb);
	mips_clean_dcache_nowrite ((vaddr_t)addr, nb);
    }
}
