#ident "kit/p6032/sbdrtpx.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * sbdrtpx.c: low-level interrupt and timer handling for P6032 board
 */


#include <sys/types.h>
#include <mips/cpu.h>
#include <mips/prid.h>
#include <mips/xcpt.h>
#include "kit_impl.h"
#include "sbd.h"
#include "sbdicu.h"
#include "i8259.h"

#include "pthread_impl.h"

/* 
 * Interrupt wiring under MIPS RTPX:
 *  HINT5 - cpu clock
 *  HINT4 - n/c
 *  HINT3 - debug button
 *  HINT2 - n/c
 *  HINT1 - bonito1
 *  HINT0 - bonito0
 *  SINT1 - unused/reserved
 *  SINT0 - unused/reserved
 */

/* the interrupt mask when at "user" level (ipl0) */
_sbd_intrmask_t		_sbd_ipl0_mask;

/* the CPU status register when at scheduler level (ipl1) */
unsigned int 		_sbd_ipl1_sr;

/* the CPU status register when at interrupt blocking level (iplhi) */
unsigned int		_sbd_iplhi_sr;

/* the interrupt mask with which we were entered (e.g. from PMON),
   see .../kit/pmon/pmonlow.sx */
unsigned int 		_sr_base __attribute__((common));

/* tell PMON that we handle all interrupts (except panic i/u),
   see .../kit/pmon/pmonxlow.sx */
unsigned 		_intrmask = SR_IMASK & ~SR_HINT3;

/* high priority debug button handler on CR_HINT3 */
static int		debugHandler (void);

/* table mapping CPU interrupt cause register bits to interrupt numbers */
static signed char	crToIntr[256];

/* pointer to Bonito */
static char * const _bonito = PA_TO_KVA1(BONITO_BASE);

/* pointers to ISA interrupt controllers */
static volatile struct i8259reg * const isamaster = 
			PA_TO_KVA1 (ISAPORT_BASE (ICU1_PORT));

static volatile struct i8259reg * const isaslave =
			PA_TO_KVA1 (ISAPORT_BASE (ICU2_PORT));

static volatile unsigned char * const elcr1 = 
			PA_TO_KVA1 (ISAPORT_BASE (ELCR1_PORT));

static volatile unsigned char * const elcr2 = 
			PA_TO_KVA1 (ISAPORT_BASE (ELCR2_PORT));

/* 
 * Return number of timer ticks per second
 */
unsigned int
_sbd_tickspersec (void)
{
    unsigned int freq = _sbd_pipefreq ();
    if (((mips_getprid () >> 8) & 0xff) == PRID_R4100)
	/* R4100 counter runs at pipeline frequency / 4 */
	freq /= 4;
    else
	/* R4x00/R5x00 counter runs at pipeline frequency / 2 */
	freq /= 2;
    return freq;
}


/*
 * Return current value of timer (upcount)
 */
unsigned int
_sbd_gettickval (void)
{
    return mips_getcount();
}


/*
 * Set value at which timer should generate interrupt, and
 * clear any pending timer interrupt.
 */
void
_sbd_settickcmp (unsigned int v)
{
    /* set next tick time, and clear interrupt */
    mips_setcompare(v);
}


int 
_sbd_intr_howmany ()
{
    return INTR_MAX + 1;
}


/* disable interrupt and clear edge detect latches */
static __inline__ void
intrmaskandclear (int intr)
{
    /* note this is called at iplhi, so we don't need to use atomic
       updates on the mask values */

    if (INTR_IS_CPU (intr)) {
	_sbd_ipl0_mask.sr &= ~(1 << intr);
	/* clear software interrupts */
	if (intr <= INTR_SINT1)
	    (void) mips_biccr (1 << intr);
    }
    else if (INTR_IS_BONITO (intr)) {
	_sbd_ipl0_mask.bonito &= ~(1 << (intr & 31));
	/* clear edge-detect latch and disable interrupt in one go */
	BONITO_INTENCLR = 1 << (intr & 31);
	mips_wbflush ();
    }
    else if (INTR_IS_ISA_SLAVE (intr)) {
	_sbd_ipl0_mask.isaslave &= ~(1 << (intr & 7));
	isaslave->i8259_ocw1 = ~_sbd_ipl0_mask.isaslave;
	isaslave->i8259_cmd = I8259_OCW2 | I8259_SPECIFIC | I8259_EOI
	    | (intr & 7);
	isamaster->i8259_cmd = I8259_OCW2 | I8259_SPECIFIC | I8259_EOI 
	    | 2;
	(void) isamaster->i8259_rr;	/* ensure last PCI cycle complete */
    }
    else {
	_sbd_ipl0_mask.isamaster &= ~(1 << (intr & 7));
	isamaster->i8259_ocw1 = ~_sbd_ipl0_mask.isamaster;
	isamaster->i8259_cmd = I8259_OCW2 | I8259_SPECIFIC | I8259_EOI
	    | (intr & 7);
	(void) isamaster->i8259_rr;	/* ensure last PCI cycle complete */
    }
}


/* 
 * Poll specific interrupt, returns 0 if interrupt is not active.
 * Called at ipl1.
 */
int 
_sbd_intrpoll (int intr)
{
    unsigned int ret;

    if (INTR_IS_CPU (intr)) {
	if (intr <= INTR_SINT1) {
	    /* read and clear software interrupts (atomically) */
	    mips_setsr (_sbd_iplhi_sr);
	    ret = mips_biccr (1 << intr);
	    mips_setsr (_sbd_ipl1_sr);
	}
	else {
	    /* read hardware interrupts */
	    ret = mips_getcr();
	}
	ret &= (1 << intr);
    }
    else if (INTR_IS_BONITO (intr)) {
	ret = BONITO_INTISR & (1 << (intr & 31));
	if (ret) {
	    /* clear edge-detect latch */
	    BONITO_INTENCLR = 1 << (intr & 31);
	    mips_wbflush ();
	}
    }
    else if (INTR_IS_ISA_SLAVE (intr)) {
	ret = isaslave->i8259_rr & (1 << (intr & 7));
	if (ret) {
	    /* clear interrupt */
	    isaslave->i8259_cmd = I8259_OCW2 | I8259_SPECIFIC | I8259_EOI
		| (intr & 7);
	    isamaster->i8259_cmd = I8259_OCW2 | I8259_SPECIFIC | I8259_EOI 
		| 2;
	    (void) isamaster->i8259_rr;	/* ensure last PCI cycle complete */
	}
    }
    else {
	ret = isamaster->i8259_rr & (1 << (intr & 7));
	if (ret) {
	    /* clear interrupt */
	    isamaster->i8259_cmd = I8259_OCW2 | I8259_SPECIFIC | I8259_EOI
		| (intr & 7);
	    (void) isamaster->i8259_rr;	/* ensure last PCI cycle complete */
	}
    }

    return ret != 0;
}


/*
 * Enable ipl0 interupts which have been blocked, in particular
 * when interrupt handler first installed.
 * Called at ipl1.
 */
void 
_sbd_intrenable (int intr)
{
    /* We must use atomic operations on mask "registers" because they
       may be shared with interrupt code running at higher "priority"
       than ipl1. */

    if (INTR_IS_CPU (intr)) {
	/* when enabled at ipl0, make sure it's disabled at ipl1 */
	mips_atomic_bis (&_sbd_ipl0_mask.sr, 1 << intr);
	mips_atomic_bic (&_sbd_ipl1_sr, 1 << intr);
    }
    else if (INTR_IS_BONITO (intr)) {
	/* bonito implements atomic set of mask */
	_sbd_ipl0_mask.bonito |= (1 << (intr & 31));
	BONITO_INTENSET = 1 << (intr & 31);
	mips_wbflush ();
    }
    else if (INTR_IS_ISA_MASTER (intr)) {
	mips_setsr (_sbd_iplhi_sr);
	_sbd_ipl0_mask.isamaster |= 1 << (intr & 7);
	isamaster->i8259_ocw1 = ~_sbd_ipl0_mask.isamaster;
	(void) isamaster->i8259_rr;
	mips_setsr (_sbd_ipl1_sr);
    }
    else {
	mips_setsr (_sbd_iplhi_sr);
	_sbd_ipl0_mask.isaslave |= 1 << (intr & 7);
	isaslave->i8259_ocw1 = ~_sbd_ipl0_mask.isaslave;
	(void) isamaster->i8259_rr;
	mips_setsr (_sbd_ipl1_sr);
    }
}


/* 
 * Take a guess at which ipl this CP0 status register represents. If
 * in doubt return 0. 
 */
int
_sbd_probable_ipl (unsigned int sr)
{
    return !(sr & SR_IE) ? 2 : (sr & SR_HINT5) ? 0 : 1;
}


/* 
 * Return next pending interrupt for this cpu, or zero if none.
 * If idle != 0 then we can spin waiting for an interrupt 
 * (possibly going into low-power mode) 
 * Called at ipl1.
 */
int 
_sbd_intrnext (int idle)
{
  int intr;

  /* switch to "hi" interrupt priority (all interrupts disabled) */
  mips_setsr (_sbd_iplhi_sr);

  while (1) {
      unsigned int pend;

      /* find interrupts which we handle that would be pending at ipl0 */
      pend = mips_getcr() & _sbd_ipl0_mask.sr & _intrmask;
      intr = crToIntr[pend >> 8];
      
      /* special case panic interrupts */
      if (intr == INTR_DEBUG) {
	  (void) debugHandler ();
	  continue;
      }
      
      /* pick apart external interrupts multiplexed onto one cpu interrupt */
      if (intr == INTR_BONITO0) {
	  unsigned int birr;
	  birr = BONITO_INTISR & _sbd_ipl0_mask.bonito;
	  intr = 0;
	  while (birr != 0 && intr == 0) {
	      int bintr = 31 ^ _mips_clz (birr);
	      birr &= ~(1 << bintr);
	      intr = INTR_MBOX0 + bintr;
	      if (intr  == INTR_ISAMASTER) {
		  unsigned int mirr;
		  mirr = isamaster->i8259_rr & _sbd_ipl0_mask.isamaster;
		  intr = 0;
		  while (mirr != 0 && intr == 0) {
		      int mintr = 31 ^ _mips_clz (mirr);
		      mirr &= ~(1 << mintr);
		      intr = INTR_ISAIRQ0 + mintr;
		      if (intr == INTR_ISASLAVE) {
			  unsigned int sirr;
			  sirr = isaslave->i8259_rr & _sbd_ipl0_mask.isaslave;
			  if (sirr != 0)
			      intr = INTR_ISAIRQ8 + 31 - _mips_clz (sirr);
			  else
			      intr = 0;
		      }
		  }
	      }
	  }
      }

      if (intr != 0 || !idle)
	  break;

      /* return to ipl1 to allow non-RTPX interrupts to be serviced */
      mips_setsr (_sbd_ipl1_sr);
      mips_setsr (_sbd_iplhi_sr);
  }

  /* disable and clear the detected interrupt at ipl0; note that clock i/u
     (intr == -1) is never disabled. */
  if (intr > 0)
      intrmaskandclear (intr);

  /* return to ipl1 */
  mips_setsr (_sbd_ipl1_sr);
  return intr;
}


/* 
 * Board-specific interrupt handler, return 1 if interrupt handled.
 * Called at iplhi.
 */
int 
_sbd_intrhandler (unsigned int pend, struct xcptcontext *xcp)
{
    /* we only handle debug interrupts locally */
    if (pend & _intrmask & CR_HINT3)
	return debugHandler ();

    /* everything else is passed to the scheduler */
    return 0;
}


int
_sbd_intrtype (int intr, const intrtype_t *new, intrtype_t *old)
{
    if (old) {
	if (INTR_IS_CPU (intr))
	    old->mode = INTRTYPE_LEVEL_HI;
	else if (INTR_IS_BONITO (intr)) {
	    unsigned int bit = 1 << (intr & 31);
	    old->mode = (BONITO_INTEDGE & bit)
		? ((BONITO_INTPOL & bit)
		   ? INTRTYPE_EDGE_RISING : INTRTYPE_EDGE_FALLING)
		: ((BONITO_INTPOL & bit)
		   ? INTRTYPE_LEVEL_HI : INTRTYPE_LEVEL_LO);
	}
	else if (INTR_IS_ISA_MASTER (intr)) {
	    old->mode = (*elcr1 & (1 << (intr & 7)))
		? INTRTYPE_LEVEL_HI : INTRTYPE_EDGE_RISING;
	}
	else {
	    old->mode = (*elcr2 & (1 << (intr & 7)))
		? INTRTYPE_LEVEL_HI : INTRTYPE_EDGE_RISING;
	}
    }
	
    /* note that we can't change the active polarity, user must
       just get this right. */
    if (new) {
	if (INTR_IS_BONITO (intr)) {
	    unsigned int bit = 1 << (intr & 31);
	    switch (new->mode) {
	    case INTRTYPE_LEVEL_LO:
		BONITO_INTPOL &= ~bit;
		BONITO_INTEDGE &= ~bit;
		break;
	    case INTRTYPE_LEVEL_HI:
		BONITO_INTPOL |= bit;
		BONITO_INTEDGE &= ~bit;
		break;
	    case INTRTYPE_EDGE_FALLING:
		BONITO_INTPOL &= ~bit;
		BONITO_INTEDGE |= bit;
		break;
	    case INTRTYPE_EDGE_RISING:
		BONITO_INTPOL |= bit;
		BONITO_INTEDGE |= bit;
		break;
	    default:
		return 1;
	    }
	}
	else if (INTR_IS_ISA_MASTER (intr)) {
	    unsigned int bit = 1 << (intr & 7);
	    if (new->mode == INTRTYPE_LEVEL_HI)
		*elcr1 |= bit;
	    else if (new->mode == INTRTYPE_EDGE_RISING)
		*elcr1 &= ~bit;
	    else
		return 1;
	}
	else if (INTR_IS_ISA_SLAVE (intr)) {
	    unsigned int bit = 1 << (intr & 7);
	    if (new->mode == INTRTYPE_LEVEL_HI)
		*elcr2 |= bit;
	    else if (new->mode == INTRTYPE_EDGE_RISING)
		*elcr2 &= ~bit;
	    else
		return 1;
	}
	else if (new->mode != INTRTYPE_LEVEL_HI)
	    return 1;
    }

    return 0;
}


/* 
 * Initialise interrupt handling for this board 
 * Called at undefined ipl, should return at iplhi
 */
void 
_sbd_intrinit(void)
{
    int i;

    /* disable all interrupts at interrupt level and set cpu to iplhi */
    _sbd_iplhi_sr = _sr_base & ~(SR_CU1 | SR_IMASK | SR_IE);
    mips_setsr (_sbd_iplhi_sr);

    /* disable clock, bonito level 0 interrupt group in scheduler */
    _sbd_ipl1_sr = (_sr_base & ~(SR_HINT5 | SR_HINT0)) | SR_IE;

    /* enable clock and bonito level 0 interrupt group at user level */
    _sbd_ipl0_mask.sr = (_sr_base & SR_IMASK) | SR_HINT5 | SR_HINT0;

    /* enable only ISA master icu in bonito mask */
    _sbd_ipl0_mask.bonito = ICU_ISA_IRQ;

    /* enable only ISA slave icu in master icu mask */
    _sbd_ipl0_mask.isamaster = 1 << 2;

    /* all interrupts disabled in ISA slave icu */
    _sbd_ipl0_mask.isaslave = 0;

    /* initialise Bonito icu */
    BONITO_INTEDGE 	= (ICU_SYSTEMERR | ICU_MASTERERR
			   | ICU_RETRYERR | ICU_MBOXES);
    BONITO_INTPOL 	= (ICU_UART0_IRQ | ICU_UART1_IRQ | ICU_BIDE_IRQ
			   | ICU_ISA_IRQ | ICU_NMI_IRQ);
    BONITO_INTSTEER	= 0;		/* all routed to CPU Intr0 */
    BONITO_INTENCLR 	= ~0;
    BONITO_INTENSET 	= _sbd_ipl0_mask.bonito;

    /* initialise ISA slave icu */
    isaslave->i8259_dat	= I8259_ICW1 | I8259_NEEDICW4;	/* icw1 */
    isaslave->i8259_dat	= 0x08;				/* icw2 - vector */
    isaslave->i8259_dat	= 0x02;				/* icw3 - slave id */
    isaslave->i8259_dat	= I8259_8086 | I8259_NOBUF;	/* icw4 */
    isaslave->i8259_cmd	= (I8259_OCW3 			/* ocw3 */
			   | I8259_ESMM | I8259_SMM	/* special mask mode */
			   | I8259_RR | I8259_RIR);	/* read IR reg */
    isaslave->i8259_ocw1= ~_sbd_ipl0_mask.isaslave;	/* ocw1 - set mask */
    
    /* initialise ISA master icu */
    isamaster->i8259_cmd = I8259_ICW1 | I8259_NEEDICW4;	/* icw1 */
    isamaster->i8259_dat = 0x00;			/* icw2 - vector */
    isamaster->i8259_dat = 0x04;			/* icw3 - slave id */
    isamaster->i8259_dat = I8259_8086 | I8259_NOBUF;	/* icw4 */
    isamaster->i8259_cmd = (I8259_OCW3 			/* ocw3 */
			   | I8259_ESMM | I8259_SMM	/* special mask mode */
			   | I8259_RR | I8259_RIR);	/* read IR reg */
    isamaster->i8259_ocw1= ~_sbd_ipl0_mask.isamaster;	/* ocw1 - set mask */

    /* build Cause bits to interrupt number mapping table */
    for (i = 0; i < 256; i++) {
	unsigned int cr = i << 8;
	int intr;
	/* test in "priority" order */
	if (cr & CR_HINT3)
	    intr = INTR_DEBUG;
	else if (cr & CR_HINT5)
	    intr = INTR_CLOCK;
	else if (cr & CR_HINT4)
	    intr = INTR_HINT4;
	else if (cr & CR_HINT2)
	    intr = INTR_HINT2;
	else if (cr & CR_HINT1)
	    intr = INTR_BONITO1;
	else if (cr & CR_HINT0)
	    intr = INTR_BONITO0;
	else if (cr & CR_SINT1)
	    intr = INTR_SINT1;
	else if (cr & CR_SINT0)
	    intr = INTR_SINT0;
	else
	    intr = 0;
	crToIntr[i] = intr;
    }

    /* set a "safe" compare value and clear timer interrupt */
    mips_setcompare (mips_getcount() + INT_MAX);

    /* XXX */
    _dbg_signals ();
}


/*
 * Handle board-specific debug or fatal interrupts
 */
static int 
debugHandler ()
{
    /* software debounce debug button (10ms stability required) */
    unsigned int ocr = mips_getcr ();
    unsigned int stable = _sbd_pipefreq () / 2 / 100;
    unsigned int compare;

    compare = mips_getcount () + stable;
    do {
	unsigned int cr = mips_getcr ();
	if ((cr ^ ocr) & CR_HINT3) {
	    /* cause bit changed - start again */
	    compare = mips_getcount () + stable;
	    ocr = cr;
	}
    } while ((int)(compare - mips_getcount ()) > 0);

    _debug_breakpoint ();
    return 1;
}


/* 
 * Called by remote debug stub when first initialised to enable "debug"
 * interrupts (h/w intr 3).
 */
void
_dbg_signals (void)
{
    int s = _mips_intdisable ();
    _sbd_ipl0_mask.sr |= SR_HINT3;	/* enable HINT3 in user mode */
    _sbd_ipl1_sr |= SR_HINT3;		/* enable HINT3 in scheduler mode */
    _sbd_iplhi_sr &= ~SR_HINT3;		/* disable HINT3 in interrupt mode */
    _intrmask |= SR_HINT3;		/* prevent PMON from handling HINT3 */
    _mips_intrestore (s);
}
