#ident "kit/p6032/sbdtune.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2002-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * p6032/sbdtune.c: tune memory system timings
 */


#include <stddef.h>
#include <mips/cpu.h>
#include "sbd.h"
#include "bonito.h"
#include "dimm.h"
#include "kit_impl.h"

static char * const _bonito = PA_TO_KVA1(BONITO_BASE);

#define NSETS 3
/*#define DIMMVERBOSE */

/*
 * bus_clock_period is in ps
 * clock skew is board dependent and in ps
 */

struct dimmsettings {
  int valid;
  int latency;
  int extrascas;
  int extprech;
  int extraswidth;
};


static void
dimm_timings (int dimm, unsigned int bus_clock_period, unsigned int clock_skew,
	      struct dimmsettings *dp, int verbose)
{
  struct {
    int latency;		/* CAS latency setting */
    int tcr;			/* minimum cycle time in ps */
    int tar;			/* access from clock in ps */
  } timings[NSETS], *tp;
  unsigned int cas_latencies;	/* CAS latencies supported by DIMM */
  const int tcr[NSETS] = {9, 23, 25};	/* tcr locations for each supported set */
  int set;
  int latency;
  int tRP, tRAS, tRCD, tRRD;
  int w1, w2;
  int type;

  /* Set the defaults */
  dp->valid = 1;
  dp->latency = 3;
  dp->extrascas = 1;
  dp->extprech = 1;
  dp->extraswidth = 3;

  type = _sbd_dimmread (dimm, 2);

  if (type != 4 && type != 7) {		/* SDRAM/DDR */
    dp->valid = 0;
    return;
  }

  /* FIXME: hardwire DDR for now */
  if (type == 7) {
    dp->latency = 2;
    return;
  }
    

  cas_latencies = _sbd_dimmread (dimm, 18);

#ifdef DIMMVERBOSE
  if (verbose)
    _mon_printf ("DIMM%d cas_latencies=%02x\n", dimm, cas_latencies);
#endif
  
  tp = timings;
  set = 0;

  /* Find the supported timings */
  for (latency = 7; latency > 0; latency--) {
    if (cas_latencies & (1 << (latency-1))) {
      unsigned int r;

      /* DIMM supports this latency setting */
      tp->latency = latency;

      /* convert minimum cycle time to ps without error checking... */
      r = _sbd_dimmread (dimm, tcr[set]);
      tp->tcr = (((r >> 4) & 0x0f)*10+(r & 0x0f)) * 100;
      
      /* convert access from clock to ps without error checking... */
      r = _sbd_dimmread (dimm, tcr[set]+1);

      /* DDR timings are in scaled by a factor of 10 */
      if (type == 7)
	tp->tar = (((r >> 4) & 0x0f)*10 + (r & 0x0f)) * 10;
      else
	tp->tar = (((r >> 4) & 0x0f)*10 + (r & 0x0f)) * 100;
	
#ifdef DIMMVERBOSE
      if (verbose)
	_mon_printf ("DIMM%d: CAS latency %d tCK=%dps tAR=%dps\n",
		     dimm, tp->latency, tp->tcr, tp->tar);
#endif

      tp++;
      if (set++ >= NSETS)
	break;
    }
  }

  /*
   * Check each supported timing from fastest to slowest until we
   * find one we can use
   */
  for (; tp >= timings; tp--) {
    /* Bonito only handles CAS latencies 2 & 3 */
    if (tp->latency != 2 && tp->latency != 3)
      continue;
    if (tp->tcr > bus_clock_period)
      continue;
    if (tp->tar + 1500 + clock_skew > bus_clock_period)
      continue;

    /* Found a usable set of timings */

    /* Set CAS latency */
    dp->latency = tp->latency;
    break;
  }

  tRP = _sbd_dimmread(dimm, 27) * 1000;
  tRRD = _sbd_dimmread(dimm, 28) * 1000;
  tRCD = _sbd_dimmread(dimm, 29) * 1000;
  tRAS = _sbd_dimmread(dimm, 30) * 1000;

#ifdef DIMMVERBOSE
  if (verbose)
    _mon_printf ("DIMM%d tRP=%dps tRAS=%dps tRCD=%dps tRRd=%dps\n",
		 dimm, tRP, tRAS,tRCD, tRRD);
#endif

  /* EXTPRECH */
  if (2 * bus_clock_period >= tRP)
    dp->extprech = 0;

  /* EXTRASCAS */
  if (2 * bus_clock_period >= tRCD)
    dp->extrascas = 0;

  /* EXTRASWIDTH */
  w1 = ((tRAS+999)/bus_clock_period) -
    (5 + (dp->extrascas ? 1 : 0));
  w2 = ((tRRD+999)/bus_clock_period) - 
    (7 + (dp->extrascas ? 1 : 0) + (dp->extprech ? 1 : 0));
  dp->extraswidth = (w1 > w2) ? w1 : w2;

#ifdef DIMMVERBOSE
  if (verbose)
      _mon_printf ("DIMM%d: valid=%d latency=%d extrascas=%d extprech=%d extraswidth=%d\n", 
		 dimm, dp->valid, dp->latency, dp->extrascas,
		 dp->extprech, dp->extraswidth);
#endif
}


void _sbd_tune (void)
{
    if (_sbd_boardtype () == 64) {
	int busps = (1000000000 / (_sbd_bonfreq () / 1000));
	int ps = ((busps+49)/100)*100; /* round to nearest .1ns */
	struct dimmsettings dimms[2], *dp;
	int skew;
	int i;
	int dimmtimings = _sbd_getenv ("dimmtimings") ? 1 : 0;
#ifdef DIMMVERBOSE
	int dimmverbose = _sbd_getenv ("dimmverbose") ? 1 : 0;
#else
#define dimmverbose 0
#endif
	unsigned int sdcfg;
	int extraswidth;
	unsigned int t;

	/* round up to nearest ns */
	BONITO_IODEVCFG = (BONITO_IODEVCFG & ~0xfc000000)
	    | (((busps + 999)/1000) << 26);

	if ((_sbd_pcbrev () & 3) == 0)
	    skew = 1200;
	else
	    skew = 300;

#ifdef DIMMVERBOSE
	if (dimmverbose)
	    _mon_printf ("Checking DIMMS, busclock=%dps(%dps), skew=%dps\n",
			 busps, ps, skew);
#endif

	for (i = 0; i < 2; i++) {
	    dp = &dimms[i];
	    dimm_timings (i, ps, skew, dp, dimmverbose);
	}

	sdcfg = BONITO_SDCFG;
	sdcfg &= ~(BONITO_SDCFG_EXTRDDATA | BONITO_SDCFG_EXTRASCAS | BONITO_SDCFG_EXTPRECH
		   | BONITO_SDCFG_EXTRASWIDTH);
      
	extraswidth = 0;
	for (i = 0; i < 2; i++) {
	    dp = &dimms[i];
	    if (dp->valid == 0)
		continue;
	    if (dp->latency == 3)
		sdcfg |= BONITO_SDCFG_EXTRDDATA;
	    if (dp->extrascas)
		sdcfg |= BONITO_SDCFG_EXTRASCAS;
	    if (dp->extprech)
		sdcfg |= BONITO_SDCFG_EXTPRECH;
	    if (dp->extraswidth > extraswidth)
		extraswidth = dp->extraswidth;
	}

	sdcfg |= extraswidth << BONITO_SDCFG_EXTRASWIDTH_SHIFT;
#ifdef DIMMVERBOSE
	if (dimmverbose)
	    _mon_printf ("New SDCFG = 0x%08x\n", sdcfg);
#endif

	if (dimmtimings) {
#ifndef BONITO_SDCFG_DRAMMODESET
#define BONITO_SDCFG_DRAMMODESET	(1<<21)
#define BONITO_SDCFG_DRAMMODESET_DONE	(1<<31)
#endif	
	    /* Make sure we don't touch SDRAM while doing
	       the SDRAM modeset. */
	    /* FIXME Make sure that we're running this code uncached
	       from ROM or out of cache. */
	    asm (
		"or	%0,%1,%3\n\t"
		"sw	%0,0(%2)\n"
	"1:	 lw	%0,0(%2)\n\t"
		"and	%0,%4\n\t"
		"beqz	%0,1b\n\t"
		"sw	%1,0(%2)\n"
		: 
		: "d" (t),
		"d" (sdcfg),
		"d" (_bonito + BONITO_REGBASE + 0x0c), /* BONITO_SDCFG */
		"i" (BONITO_SDCFG_DRAMMODESET),
		"i" (BONITO_SDCFG_DRAMMODESET_DONE)
		);
	}
    }
}
