#ident "kit/pci/pciio.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1999-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * pciio.c: generic PCI i/o support functions
 */


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <mips/cpu.h>
#include <pci/pcivar.h>
#include <pci/pcireg.h>

#include "kit_impl.h"
#include "sbd.h"

/* Scan the PCI bus for the COUNT'th device which matches *MATCH.  The
 * pci_match structure allows a masked comparison of the device's ID
 * and CLASS registers.  It returns the device's TAG, which can be
 * used to identify it to the other PCI functions.  
 */
 
pcitag_t
_pci_find (const struct pci_match *match, unsigned int count)
{
    pcitag_t tag;
    pcireg_t id, class, misc;
    unsigned int bus, device, func;

    for (bus = 0; bus < _pci_nbus; bus++) {

	for (device = 0; device < 32; device++) {

	    tag = _pci_make_tag (bus, device, 0);
	    if (!_pci_canscan (tag))
		continue;

	    id = _pci_conf_read32 (tag, PCI_ID_REG);
	    if (id == 0 || id == ~(pcireg_t)0)
		continue;

	    misc = _pci_conf_read32 (tag, PCI_MISC_REG);

	    if (PCI_MISC_HDRTYPE (misc) & PCI_MISC_HDRTYPE_MULTI) {
		for (func = 0; func < 8; func++) {
		    tag = _pci_make_tag (bus, device, func);
		    id = _pci_conf_read32 (tag, PCI_ID_REG);
		    if (id == 0 || id == ~(pcireg_t)0)
			continue;
		    class = _pci_conf_read32 (tag, PCI_CLASS_REG);
		    if ((id & match->idmask) == match->id
			&& (class & match->classmask) == match->class
			&& count-- == 0)
			return tag;
		}
	    }
	    else {
		class = _pci_conf_read32 (tag, PCI_CLASS_REG);
		if ((id & match->idmask) == match->id
		    && (class & match->classmask) == match->class
		    && count-- == 0)
		    return tag;
	    }
	}
    }

    return ~(pcitag_t)0;
}


#ifndef _SBD_PCI_CFSPACE_PARTIAL

/* Many PCI controllers don't allow partial word accesses to config
   space.  To support them we provide software byte and halfword 
   access routines. This code assumes little-endian register
   numbering. */

#pragma weak _pci_conf_read8
pcireg_t
_pci_conf_read8 (pcitag_t tag, int reg)
{
    pcireg_t val = _pci_conf_read32 (tag, reg & ~3);
    val >>= (reg & 3) * 8;
    return val & 0xff;
}


#pragma weak _pci_conf_read16
pcireg_t
_pci_conf_read16 (pcitag_t tag, int reg)
{
    pcireg_t val;

    if (reg & 1)
	return ~(pcireg_t)0;
    val = _pci_conf_read32 (tag, reg & ~3);
    val >>= (reg & 3) * 8;
    return val & 0xffff;
}


#pragma weak _pci_conf_write8
void
_pci_conf_write8 (pcitag_t tag, int reg, pcireg_t val)
{
    unsigned int shift = (reg & 3) * 8;
    pcireg_t mask = 0xff;
    pcireg_t xval;

    reg &= ~3;
    mask <<= shift;
    val <<= shift;
    xval = _pci_conf_read32 (tag, reg);
    xval ^= ((xval ^ val) & mask);
    _pci_conf_write32 (tag, reg, xval);
}


#pragma weak _pci_conf_write16
void
_pci_conf_write16 (pcitag_t tag, int reg, pcireg_t val)
{
    unsigned int shift = (reg & 3) * 8;
    pcireg_t mask = 0xffff;
    pcireg_t xval;

    if (reg & 1)
	return;
    reg &= ~3;
    mask <<= shift;
    val <<= shift;
    xval = _pci_conf_read32 (tag, reg);
    xval ^= ((xval ^ val) & mask);
    _pci_conf_write32 (tag, reg, xval);
}
#endif
