#ident "kit/pmon/pmontty.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * pmontty.c: Posix style terminal i/o under LSI PMON
 */


#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/init.h>
#include <termios.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "file.h"
#include "tty.h"
#include "kit_impl.h"
#include "pmon.h"

static int	ttyread (struct file *, struct uio *);
static int	ttywrite (struct file *, struct uio *);
static int	ttyioctl (struct file *, int, void *);
static int	ttyselect (struct file *, int);
static int	ttyopen (struct file *, dev_t, int);
static int	ttyclose (struct file *);

static const struct fileops _tty_ops = 
{
    ttyread, 
    ttywrite, 
    ttyioctl, 
    ttyselect, 
    ttyopen,
    ttyclose,
};


#define NTTY 4
static struct tty *ttys[NTTY];


/*
 * PMON does have a rudimentary type of termio implementation, but
 * it really isn't complete enough: it doesn't even expand tabs on
 * output, and we can't catch ^C in any easy way.  So we use our 
 * own generic routines.
 */


struct convtab {
    unsigned int	pmon;
    long		posix;
};


static const struct convtab baudtab[] = {
    {PMON_B0,		0},
    {PMON_B50,		50},
    {PMON_B75,		75},
    {PMON_B110,		110},
    {PMON_B134,		134},
    {PMON_B150,		150},
    {PMON_B200,		200},
    {PMON_B300,		300},
    {PMON_B600,		600},
    {PMON_B1200,	1200},
    {PMON_B1800,	1800},
    {PMON_B2400,	2400},
    {PMON_B4800,	4800},
    {PMON_B9600,	9600},
    {PMON_B19200,	19200},
    {PMON_B38400,	38400},
    {-1,		-1}};



/*
 * The lowest level device-dependent routines, called by the generic
 * tty code.
 */

static int
ttyparams (struct tty *tp, int set, struct termios *tio)
{
    const struct convtab *bt;
    struct pmontermio pmtio;

    (void) pmon_ioctl ((int)tp->t_addr, PMON_TCGETA, &pmtio);
    if (set) {
	/* CHANGE SETTINGS */
	if (tio->c_ispeed != tio->c_ospeed)
	    return EINVAL;
	
	if (tio->c_ospeed == 0)
	    return 0;

	for (bt = baudtab; bt->posix != tio->c_ospeed; bt++)
	    if (bt->posix == -1)
		break;

	if (bt->pmon == (unsigned)-1) {
	    /* new PMON supports higher and more speeds */ 
	    if (tio->c_ospeed > 3276700 || (tio->c_ospeed % 100) != 0)
		return EINVAL;
	    pmtio.c_cflag = 0x8000 | (tio->c_ospeed / 100);
	}
	else {
	    /* compatible with old PMON */ 
	    pmtio.c_cflag = (pmtio.c_cflag & ~PMON_CBAUD) | bt->pmon;
	}

	if (pmon_ioctl ((int)tp->t_addr, PMON_TCSETAW, &pmtio) < 0)
	    return EINVAL;
    } else {
	/* GET SETTINGS */
	tio->c_cflag = TTYDEF_CFLAG;

	if (pmtio.c_cflag & 0x8000) {
	    tio->c_ispeed = tio->c_ospeed = (pmtio.c_cflag & 0x7fff) * 100;
	}
	else {
	    pmtio.c_cflag &= PMON_CBAUD;
	    for (bt = baudtab; bt->pmon != pmtio.c_cflag; bt++)
		if (bt->pmon == (unsigned)-1)
		    return 0;
	    tio->c_ispeed = tio->c_ospeed = bt->posix;
	}
    }

    return 0;
}


static int
ttyrxpoll (struct tty *tp, int canblock)
{
#if defined(TINYSIM)
    char c;

    if (canblock) {
	pmon_read ((int)tp->t_addr, &c, 1);
	return c | 0x100;
    }
    /* hopefully we'll get a blocking read sometime... */
    return 0;
#else
    int nread = 0;
    char c;

    /* never blocks */
    pmon_ioctl ((int)tp->t_addr, PMON_FIONREAD, &nread);
    if (nread > 0) {
	pmon_read ((int)tp->t_addr, &c, 1);
	return c | 0x100;
    }
    return 0;
#endif
}


static int
ttytxrdy (struct tty *tp)
{
#if defined(TINYSIM)
    static int loops = 0;
    return ((loops++ & 15) == 0) ? 1 : 0;
#else
    return 1;			/* we can't tell, assume always ready! */
#endif
}


static void
ttytx (struct tty *tp, int c)
{
    char cb = c;
    pmon_write ((int)tp->t_addr, &cb, 1);
}


static int
ttytxblk (struct tty *tp, const char *bp, int nc)
{
    return pmon_write ((int)tp->t_addr, bp, (nc > 16) ? 16 : nc);
}



static int
ttywrite (struct file *fp, struct uio *uio)
{
    /* generic write */
    return _ttywrite ((struct tty *)fp->f_data, uio, fp->f_flag);
}

static int
ttyread (struct file *fp, struct uio *uio)
{
    /* generic read */
    return _ttyread ((struct tty *)fp->f_data, uio, fp->f_flag);
}

static int
ttyioctl (struct file *fp, int cmd, void *arg)
{
    /* generic ioctl */
    return _ttyioctl ((struct tty *)fp->f_data, cmd, arg);
}

static int
ttyselect (struct file *fp, int rw)
{
    /* generic select */
    return _ttyselect ((struct tty *)fp->f_data, rw);
}

static int
ttyopen (struct file *fp, dev_t dev, int fflags)
{
    struct tty *tp;
    int err;

    if (dev >= NTTY)
	return ENXIO;

    /* initialise our structure */
    if (!(tp = ttys[dev])) {
	err = _ttymalloc ("tty(pmon)", &tp);
	if (err)
	    return err;
	ttys[dev] = tp;
    }

    if (!(tp->t_state & TS_OPEN)) {
	struct pmontermio pmtio;
	char ttyname[5];
	int fd;

	strcpy (ttyname, "ttyx");
	ttyname[3] = dev + '0';
	fd = pmon_open (ttyname, OFLAGS(fflags));
	if (fd < 0)
	    return ENXIO;

	/* put PMON into raw mode */
	(void) pmon_ioctl (fd, PMON_TCGETA, &pmtio);
	pmtio.c_iflag = pmtio.c_oflag = 0;
	pmtio.c_lflag &= ~(PMON_ISIG | PMON_ICANON | PMON_ECHO | PMON_ECHOE);
	pmtio.c_cc[PMON_VINTR] = _POSIX_VDISABLE;
	(void) pmon_ioctl (fd, PMON_TCSETAW, &pmtio);

	tp->t_dev	= dev;
	tp->t_params	= ttyparams;
	tp->t_rxpoll	= ttyrxpoll;
	tp->t_txrdy	= ttytxrdy;
	tp->t_txblk	= ttytxblk;
	tp->t_tx	= ttytx;
	tp->t_addr	= (void *) fd;
    }

    /* point file at our private data */
    fp->f_data = tp;

    /* generic open */
    return _ttyopen (tp);
}


static int
ttyclose (struct file *fp)
{
    struct tty *tp = (struct tty *) fp->f_data;
    int error = 0;
    
    if (tp) {
	error = _ttyclose (tp);
	if (!(tp->t_state & TS_OPEN))
	    pmon_close ((int)tp->t_addr);
    }
    return error;
}



/* /dev/tty*:	crw-rw-r-- */
#define TTYMODE	(S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IWOTH)

static void
ttydev_mount (void)
{
    dir_t ttynode;
    int d;
    int err;

    memset (&ttynode, 0, sizeof (ttynode));
    ttynode.d_mode = TTYMODE;
    ttynode.d_ops = &_tty_ops;

#define ADDDESC(name, dev)				\
    ttynode.d_name = name;				\
    ttynode.d_dev = dev;				\
    err = _mknod ("/dev", &ttynode);			\
    if (err && err != EEXIST) {				\
      _mon_puts ("mknod(/dev/" name ") failed\n");	\
      abort ();						\
    }

    ADDDESC("console",	0);
    ADDDESC("tty",	0);
    ADDDESC("tty0",	0);

    for (d = 1; ; d++) {
	static char name[] = "ttyx";
	int fd;
	name[3] = d + '0';
	if ((fd = pmon_open (name, O_RDWR)) < 0)
	    break;

	/*
	 * Opening the tty has the side-effect of resetting the tty modes.
	 * PMON may be using this port for debug, so put the channel back
	 * into raw mode before closing
	 */
	{
	  struct pmontermio pmtio;
	  (void) pmon_ioctl (fd, PMON_TCGETA, &pmtio);
	  pmtio.c_iflag = pmtio.c_oflag = 0;
	  pmtio.c_lflag &= ~(PMON_ISIG | PMON_ICANON | PMON_ECHO | PMON_ECHOE);
	  pmtio.c_cc[PMON_VINTR] = _POSIX_VDISABLE;
	  (void) pmon_ioctl (fd, PMON_TCSETAW, &pmtio);
	}

	pmon_close (fd);
	ttynode.d_name = name;
	ttynode.d_dev = d;
	if (_mknod ("/dev", &ttynode))
	    break;
    }
}

DECL_INIT_FUNC(IF_MOUNT, ttydev_mount)

/* call this dummy function to load this module */
void _ttydev_load (void);
void _ttydev_load (void) {}
