#ident "kit/r3kemon/r3kefile.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * r3ketftp.c: TFTP network i/o under R3KEMU
 */


#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <alloca.h>

#include "file.h"
#include "r3kemu.h"

static int	_filerw (struct file *, struct uio *);
static int	_fileioctl (struct file *, int, void *);
static int	_fileselect (struct file *, int);
static int	_fileopen (struct file *, dev_t, int);
static int	_fileclose (struct file *);
static int	_filelookup (dir_t **, dir_t *, const char **, int, mode_t);
static void	_filedfree (dir_t *);

const struct fileops _file_ops = 
{
    _filerw, 
    _filerw, 
    _fileioctl, 
    _fileselect, 
    _fileopen,
    _fileclose,
    _filelookup,
    _filedfree
};


struct r3kef {
    int		rf_fd;
    off_t	rf_offs;
};


static int
_filerw (struct file *fp, struct uio *uio)
{
    struct r3kef *rf = fp->f_data;
    int err = 0;
    
    if (!rf)
	/* root directory - can't read */
	return EACCES;

    if (uio->uio_offset != rf->rf_offs) {
	if (r3ke_lseek (rf->rf_fd, uio->uio_offset, SEEK_SET) < 0)
	    return ESPIPE;
    }
    
    while (uio->uio_resid != 0 && err == 0) {
	struct iovec *iov;
	err = uiogetiov (uio, &iov);
	if (err == 0) {
	    ssize_t res;
	    if (uio->uio_rw == UIO_READ)
		res = r3ke_read (rf->rf_fd, iov->iov_base, iov->iov_len);
	    else
		res = r3ke_write (rf->rf_fd, iov->iov_base, iov->iov_len);
	    if (res > 0)
		err = uioadvance (res, uio);
	    else if (res < 0 || uio->uio_rw == UIO_WRITE)
		err = EIO;
	    else
		break;
	}
    }

    rf->rf_offs = uio->uio_offset;
    return err;
}


static int
_fileioctl (struct file *fp, int cmd, void *arg)
{
    return EINVAL;
}


static int
_fileselect (struct file *fp, int rw)
{
    /* not used yet, but return always ready */
    return rw;
}



static const dir_t template =
DIRDESC(0, 'n'<<8, S_IFREG|S_IRWXU|S_IRWXG|S_IRWXO, 1024, _file_ops, -1);

static int
_fileopen (struct file *fp, dev_t dev, int fflags)
{
    /* nothing to do */
    return 0;
}


static int
_fileclose (struct file *fp)
{
    /* nothing to do */
    return 0;
}


static int
_filelookup (dir_t **ndpp, dir_t *dp, const char **name, int fflags, mode_t mode)
{
    const char *nm = *name;
    struct r3kef *rf;
    int fd;
    
    /* Handle "." and ".." magic names */
    if (nm[0] == '.') {
	if (nm[1] == '.' && (nm[2] == '\0' || nm[2] == '/')) {
	    /* This isn't going to work if dp->d_parent isn't valid */
	    *name += 2;
	    *ndpp = dp->d_parent;
	    return 0;
	}
	if (nm[1] == '\0' || nm[1] == '/') {
	    *name += 1;
	    *ndpp = dp;
	    return 0;
	}
    }

    fd = r3ke_open (nm, OFLAGS (fflags));
    if (fd < 0)
	return ENOENT;

    dp = malloc (sizeof (dir_t) + sizeof (struct r3kef));
    if (!dp) {
	r3ke_close (fd);
	return ENOMEM;
    }

    *dp = template;
    dp->d_flags = DF_DYNAMIC;
    dp->d_count = 1;
    rf = dp->d_data = dp + 1;
    rf->rf_fd = fd;
    rf->rf_offs = 0;

    /* advance name pointer to end of path */
    *name += strlen (nm);
    *ndpp = dp;
    return 0;
}


static void
_filedfree (dir_t *dp)
{
    struct r3kef *rf = dp->d_data;
    if (rf)
	r3ke_close (rf->rf_fd);
    if (dp->d_flags & DF_DYNAMIC)
	free (dp);
}
