#ident "kit/sead2/sbdrtpx.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * sbdrtpx.c: low-level interrupt and timer handling for MIPS SEAD board
 */


#include <sys/types.h>
#include <mips/cpu.h>
#include <mips/prid.h>
#include <mips/xcpt.h>
#include "kit_impl.h"
#include "sbd.h"
#include "sbdicu.h"

#include "pthread_impl.h"

#include <sys/init.h>

/* 
 * Interrupt wiring under MIPS RTPX:
 *  HINT5 - unused
 *  HINT4 - unused
 *  HINT3 - unused
 *  HINT2 - unused
 *  HINT1 - UART1
 *  HINT0 - UART0
 *  SINT1 - unused/reserved
 *  SINT0 - unused/reserved
 */

/* the interrupt mask when at "user" level (ipl0) */
_sbd_intrmask_t		_sbd_ipl0_sr;

/* the CPU status register when at scheduler level (ipl1) */
unsigned int 		_sbd_ipl1_sr;

/* the CPU status register when at interrupt blocking level (iplhi) */
unsigned int		_sbd_iplhi_sr;

/* the interrupt mask with which we were entered (e.g. from TAMON),
   see .../kit/yamon/yamonlow.sx */
unsigned int 		_sr_base __attribute__((common));

/* tell YAMON glue code that we handle all interrupts,
   see .../kit/yamon/yamonxlow.sx */
unsigned 		_intrmask = CR_IMASK;

/* table mapping CPU interrupt cause register bits to interrupt numbers */
static signed char	crToIntr[256];


#ifndef _SBD_CLK_DIV
extern long _sbd_cyclespercount (void);
#define _SBD_CLK_DIV	_sbd_cyclespercount ()
#endif

#ifndef _SBD_CLK_MULT
#define _SBD_CLK_MULT	1
#endif

/* 
 * Return number of timer ticks per second
 */
unsigned int
_sbd_tickspersec (void)
{
    unsigned int freq = _sbd_pipefreq ();

    /* Convert from pipeline frequency to count register frequency. */
    freq = freq * _SBD_CLK_MULT / _SBD_CLK_DIV;

    /* Initialise on-chip counter interrupt */
#ifdef mips_bicconfig
    switch ((mips_getprid () >> 8) & 0xff) {
    case PRID_20KC:
    case PRID_25KF:
#define M20KC_CFG0_TI		0x01000000
	/* Wire timer interrupt to IP5 */
	mips_bicconfig (M20KC_CFG0_TI);
	break;
    }
#endif

    return freq;
}


/*
 * Return current value of timer (upcount)
 */
unsigned int
_sbd_gettickval (void)
{
    return mips_getcount();
}


/*
 * Set value at which timer should generate interrupt, and
 * clear any pending timer interrupt.
 */
void
_sbd_settickcmp (unsigned int v)
{
    /* set next tick time, and clear interrupt */
    mips_setcompare(v);
}


int 
_sbd_intr_howmany ()
{
    return INTR_MAX + 1;
}


/* 
 * Poll specific interrupt, returns 0 if interrupt is not active.
 * Called at ipl1.
 */
int 
_sbd_intrpoll (int intr)
{
    unsigned int ret;

    if (intr <= INTR_SINT1) {
	/* read and clear software interrupts (atomically) */
	mips_setsr (_sbd_iplhi_sr);
	ret = mips_biccr (1 << intr);
	mips_setsr (_sbd_ipl1_sr);
    }
    else {
	/* read hardware interrupts */
	ret = mips_getcr();
    }
    ret &= (1 << intr);
    return ret != 0;
}


/*
 * Enable ipl0 interupts which have been blocked, in particular
 * when interrupt handler first installed.
 * Called at ipl1.
 */
void 
_sbd_intrenable (int intr)
{
    /* We must use atomic operations on mask "registers" because they
       may be shared with interrupt code running at higher "priority"
       than ipl1. */
  
    /* when enabled at ipl0, make sure it's disabled at ipl1 */
    mips_atomic_bis (&_sbd_ipl0_sr, 1 << intr);
    mips_atomic_bic (&_sbd_ipl1_sr, 1 << intr);
}


/* 
 * Take a guess at which ipl this CP0 status register represents. If
 * in doubt return 0. 
 */
int
_sbd_probable_ipl (unsigned int sr)
{
    return !(sr & SR_IE) ? 2 : (sr & SR_HINT5) ? 0 : 1;
}


/* 
 * Return next pending interrupt for this cpu, or zero if none.
 * If idle != 0 then we can spin waiting for an interrupt 
 * (possibly going into low-power mode) 
 * Called at ipl1.
 */
int 
_sbd_intrnext (int idle)
{
  int intr;

  /* switch to "hi" interrupt priority (all interrupts disabled) */
  mips_setsr (_sbd_iplhi_sr);

  while (1) {
      unsigned int pend;

      /* find interrupts which we handle that would be pending at ipl0 */
      pend = mips_getcr() & _sbd_ipl0_sr & _intrmask;

      /* quick map from cause register bits to interrupt number */
      intr = crToIntr[pend >> 8];
      
      if (intr != 0 || !idle)
	  break;

      /* return to ipl1 to allow non-RTPX interrupts to be serviced */
      mips_setsr (_sbd_ipl1_sr);
      mips_setsr (_sbd_iplhi_sr);
  }

  /* disable and clear the detected interrupt at ipl0; 
     note that clock i/u (intr == -1) is never disabled. */
  if (intr > 0) {
      /* at iplhi, so we don't need to use atomic updates on the mask values */
      _sbd_ipl0_sr &= ~(1 << intr);
      /* clear software interrupt "latch" */
      if (intr <= INTR_SINT1)
	  (void) mips_biccr (1 << intr);
  }

  /* return to ipl1 */
  mips_setsr (_sbd_ipl1_sr);
  return intr;
}


/* 
 * Board-specific interrupt handler, return 1 if interrupt handled.
 * Called at iplhi.
 */
int 
_sbd_intrhandler (unsigned int pend, struct xcptcontext *xcp)
{
    /* everything else is passed to the scheduler */
    return 0;
}


int
_sbd_intrtype (int intr, const intrtype_t *new, intrtype_t *old)
{
    if (old) {
	old->mode = INTRTYPE_LEVEL_HI;
    }
	
    /* note that we can't change the active polarity, user must
       just get this right. */
    if (new) {
	if (new->mode != INTRTYPE_LEVEL_HI)
	    return 1;
    }

    return 0;
}


/* 
 * Initialise interrupt handling for this board 
 * Called at undefined ipl, should return at iplhi
 */
void 
_sbd_intrinit(void)
{
    int i;

    /* disable all interrupts at interrupt level and set cpu to iplhi */
    _sbd_iplhi_sr = _sr_base & ~(SR_CU1 | SR_IMASK | SR_IE);
    mips_setsr (_sbd_iplhi_sr);

    /* disable individual interrupts in scheduler, but external
     * interrupts could be enabled */
    _sbd_ipl1_sr = (_sr_base & ~SR_IMASK) | SR_IE;

    /* enable clock interrupt at user level */
    _sbd_ipl0_sr = (_sr_base & SR_IMASK) | SR_HINT5;

    
    /* build Cause bits to interrupt number mapping table */
    for (i = 0; i < 256; i++) {
	unsigned int cr = i << 8;
	int intr;
	/* test in "priority" order */
	if (cr & CR_HINT5)
	    intr = INTR_CLOCK;
	else if (cr & CR_HINT4)
	    intr = INTR_HINT4;
	else if (cr & CR_HINT3)
	    intr = INTR_HINT3;
	else if (cr & CR_HINT2)
	    intr = INTR_HINT2;
	else if (cr & CR_HINT1)
	    intr = INTR_UART1;
	else if (cr & CR_HINT0)
	    intr = INTR_UART0;
	else if (cr & CR_SINT1)
	    intr = INTR_SINT1;
	else if (cr & CR_SINT0)
	    intr = INTR_SINT0;
	else
	    intr = 0;
	crToIntr[i] = intr;
    }

    /* set a "safe" compare value and clear timer interrupt */
    mips_setcompare (mips_getcount() + INT_MAX);
}


/* 
 * Called by remote debug stub when first initialised to enable "debug"
 * interrupts (NMI?).
 */
void
_dbg_signals (void)
{
}
