#ident "kit/share/bremu.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1999-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * bremu.c: emulate a branch instruction
 */


#include <stddef.h>
#include <mips/cpu.h>
#include <mips/fpa.h>
#include <mips/xcpt.h>
#include <mips/instrdef.h>
#include <mips/instr16def.h>
#include <sys/signal.h>
#include "kit_impl.h"

#if __mips >= 4
/* convert condition code register number to csr bit */
static const unsigned int fpacondbit[8] = {
    FPA_CSR_COND0,
    FPA_CSR_COND1,
    FPA_CSR_COND2,
    FPA_CSR_COND3,
    FPA_CSR_COND4,
    FPA_CSR_COND5,
    FPA_CSR_COND6,
    FPA_CSR_COND7
};
#endif

/*
 * Emulate a branch instruction ir (at xcp->epc), which is
 * required when we have to emulate instructions in branch
 * delay slots.
 *
 * Does not change epc due to the instruction - sets *cpc instead.
 */

int
_mips_bremul (struct xcptcontext *xcp, struct fpactx *ctx,
	      unsigned int insn, vaddr_t *cpc, int *skipbd)
{
    union mips_instr ir;
    int bcond, link, likely;

    *skipbd = 0;
    link = likely = 0;
    ir.value = insn;
    switch (ir.itype.op) {
    case ITYPE_special:
	switch (ir.rtype.funct) {
	case RRTYPE_jalr:
	    xcp->regs[ir.rtype.rd] = xcp->epc + 8;
	case RRTYPE_jr:
	    *cpc = xcp->regs[ir.rtype.rs];
	    return 0;
	default:
	    break;
	}
	break;

    case ITYPE_regimm:
	switch (ir.itype.rt) {
	case RITYPE_bltzall:
	    likely = 1;
	case RITYPE_bltzal:
	    link = 1;
	    goto bltz;
	case RITYPE_bltzl:
	    likely = 1;
	case RITYPE_bltz:
	bltz:
	    bcond = (sreg_t)xcp->regs[ir.itype.rs] < 0;
	    goto bret;

	case RITYPE_bgezall:
	    likely = 1;
	case RITYPE_bgezal:
	    link = 1;
	    goto bgez;
	case RITYPE_bgezl:
	    likely = 1;
	case RITYPE_bgez:
	bgez:
	    bcond = (sreg_t)xcp->regs[ir.itype.rs] >= 0;
	    goto bret;

	default:
	    break;
	}
	break;

    case ITYPE_jal:
	xcp->regs[31] = xcp->epc + 8;
    case ITYPE_j:
	*cpc = (xcp->epc & ~0x0fffffff) | (ir.jtype.target << 2);
	return 0;

    case ITYPE_jalx:
	xcp->regs[31] = xcp->epc + 8;
	*cpc = (xcp->epc & ~0x0fffffff) | (ir.jtype.target << 2) | 1;
	return 0;

    case ITYPE_beql:
	likely = 1;
    case ITYPE_beq:
	bcond = (xcp->regs[ir.itype.rs] == xcp->regs[ir.itype.rt]);
	goto bret;

    case ITYPE_bnel:
	likely = 1;
    case ITYPE_bne:
	bcond = (xcp->regs[ir.itype.rs] != xcp->regs[ir.itype.rt]);
	goto bret;

    case ITYPE_blezl:
	likely = 1;
    case ITYPE_blez:
	bcond = ((sreg_t)xcp->regs[ir.itype.rs] <= 0);
	goto bret;

    case ITYPE_bgtzl:
	likely = 1;
    case ITYPE_bgtz:
	bcond = ((sreg_t)xcp->regs[ir.itype.rs] > 0);
	goto bret;

#if !#fpu(no)
    case ITYPE_c1:
	if (ir.itype.rs == COPZRS_bc) {
	    unsigned int fsr = ctx ? ctx->sr : fpa_getsr();
#if __mips >= 4
	    unsigned int cond = fpacondbit[ir.itype.rt >> 2];
#else
	    unsigned int cond = FPA_CSR_COND;
#endif
	    switch (ir.itype.rt & 3) {
	    case COPZRT_BCFL:
		likely = 1;
	    case COPZRT_BCF:
		bcond = !(fsr & cond);
		goto bret;
	    case COPZRT_BCTL:
		likely = 1;
	    case COPZRT_BCT:
		bcond = !!(fsr & cond);
		goto bret;
	    }
	}
	break;
#endif

    case ITYPE_c0:
    case ITYPE_c2:
    case ITYPE_c3:
	/* how can we work these out? */
    default:
	break;
    }
    return SIGILL;

  bret:
    if (bcond) {
	if (link)
	    xcp->regs[31] = xcp->epc + 8;
	*cpc = xcp->epc + (ir.itype.imm << 2) + 4;
    } else {
	*cpc = xcp->epc + 8;
	*skipbd = likely;
    }
    return 0;
}


#if __mips16 || __maybe_mips16

/* convert MIPS16 register to MIPS32 register */
#define mips16_to_mips32(rn)	(((rn) < 2) ? (rn) + 16 : (rn))

/*
 * Similarly for MIPS16, but note that only jumps have delay
 * slots, so we don't have to emulate branches.  Finding the
 * location of the branch delay slot is not so trivial though,
 * so we return its address in *bdpc.
 */

int
_mips16_bremul (struct xcptcontext *xcp, unsigned short insn, 
		vaddr_t *cpc, vaddr_t *bdpc)
{
    union mips16_instr i16;
    vaddr_t pc = xcp->epc;
    int err = 0;
    int offset;

    i16.value = insn;
    switch (i16.itype.op) {
    case ITYPE16_jal:
	offset = mips_get_half ((pc + 2) & ~1, &err);
	if (err)
	    return SIGBUS;
	offset |= (i16.itype.imm & 0x1f) << 21;
	offset |= (i16.itype.imm & 0x3e0) << (16-5);
	*cpc = (pc & ~0x0fffffff) | ((offset & 0x3ffffff) << 2);
	if ((i16.itype.imm & 0x400) == 0)
	    *cpc |= 1;	/* not a jalx, so stay mips16 */
	xcp->regs[31] = VA_TO_REG (pc + 6);
	*bdpc = pc + 4;
	return 0;
	
    case ITYPE16_rr:
	if (i16.rrtype.funct == 0) {
	    *bdpc = pc + 2;
	    switch (i16.rrtype.ry) {
	    case 1:		/* j ra */
		*cpc = xcp->regs[31];
		return 0;
	    case 2:		/* jalr */
		xcp->regs[31] = xcp->epc + 4;
	    case 0:		/* jr */
		*cpc = xcp->regs[mips16_to_mips32(i16.rrtype.rx)];
		return 0;
	    }
	}
	break;
    }
    return SIGILL;
}
#endif
