#ident "kit/share/cp1emu.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * cp1emu.c: a MIPS coprocessor 1 (fpa) instruction emulator
 */


/*
 * A complete emulator for MIPS coprocessor 1 instructions.  This is
 * required for #float(switch) or #float(trap), where it catches all
 * COP1 instructions via the "CoProcessor Unusable" exception.  
 *
 * More surprisingly it is also required for #float(ieee), to help out
 * the hardware fpu at the boundaries of the IEEE-754 representation
 * (denormalised values, infinities, underflow, etc).  It is made
 * quite nasty because emulation of some non-COP1 instructions is
 * required, e.g. in branch delay slots.
 * 
 * Notes: 
 *  1) the IEEE754 library (-le) performs the actual arithmetic;
 *  2) if you know that you won't have an fpu, then you'll get much 
 *     better performance by compiling with -msoft-float!  
 */

#include <stddef.h>
#include <stdio.h>
#include <sys/signal.h>
#include <stdlib.h>

#include <mips/xcpt.h>
#include <mips/cpu.h>
#include <mips/fpa.h>
#include <mips/instrdef.h>

#include <ieee754.h>

#include "kit_impl.h"

#if __mips >= 4 && (__mips != 32 || __mipsfp64)
#define MIPS4_FPU
#else
#undef MIPS4_FPU
#endif

/* Function which emulates the instruction in a branch delay slot. */

static int      mips_dsemul (struct xcptcontext *, 
			     union mips_instr,
			     vaddr_t);

/* Function which emulates a floating point instruction. */

static int      fpa_emu (struct xcptcontext *,
			 struct fpactx *, 
			 union mips_instr);

#ifdef MIPS4_FPU
static int      fpax_emu (struct xcptcontext *,
			  struct fpactx *, 
			  union mips_instr);
#endif

/* The emulated f.p. register set, may be shared with debug stub */

int             _fpaemuactive;
struct fpactx   _fpaemustate;

/* Control registers */

#define FPCREG_RID	0	/* $0  = revision id */
#define FPCREG_CSR	31	/* $31 = csr */

/* Convert Mips rounding mode (0..3) to IEEE library modes. */
static const unsigned char ieee_rm[4] =
{
    IEEE754_RN, IEEE754_RZ, IEEE754_RU, IEEE754_RD
};

#if __mips >= 4
/* convert condition code register number to csr bit */
static const unsigned int fpacondbit[8] = {
    FPA_CSR_COND0,
    FPA_CSR_COND1,
    FPA_CSR_COND2,
    FPA_CSR_COND3,
    FPA_CSR_COND4,
    FPA_CSR_COND5,
    FPA_CSR_COND6,
    FPA_CSR_COND7
};
#endif

#if defined(SR_FR) && (__mips != 32 || __mipsfp64)
#if #endian(big)
#define HIPART(x)	*((unsigned int *)&ctx->regs[x] + 0)
#define LOPART(x)	*((unsigned int *)&ctx->regs[x] + 1)
#else
#define LOPART(x)	*((unsigned int *)&ctx->regs[x] + 0)
#define HIPART(x)	*((unsigned int *)&ctx->regs[x] + 1)
#endif
#endif



static int 
isBranchInstr (union mipsn_instr *i)
{
    switch (i->itype.op) {
    case ITYPE_special:
	switch (i->rtype.funct) {
	case RRTYPE_jalr:
	case RRTYPE_jr:
	    return 1;
	}
	break;

    case ITYPE_regimm:
	switch (i->itype.rt) {
	case RITYPE_bltz:
	case RITYPE_bgez:
	case RITYPE_bltzl:
	case RITYPE_bgezl:
	case RITYPE_bltzal:
	case RITYPE_bgezal:
	case RITYPE_bltzall:
	case RITYPE_bgezall:
	    return 1;
	}
	break;

    case ITYPE_j:
    case ITYPE_jal:
    case ITYPE_jalx:
    case ITYPE_beq:
    case ITYPE_bne:
    case ITYPE_blez:
    case ITYPE_bgtz:
    case ITYPE_beql:
    case ITYPE_bnel:
    case ITYPE_blezl:
    case ITYPE_bgtzl:
	return 1;

    case ITYPE_c0:
    case ITYPE_c1:
    case ITYPE_c2:
    case ITYPE_c3:
	if (i->itype.rs == COPZRS_bc)
	    return 1;
	break;
    }

    return 0;
}



/*
 * Emulate the single floating point instruction pointed at by EPC.
 * Two instructions if the instruction is in a branch delay slot.
 */

static int
cop1Emulate (int xcptno, struct xcptcontext *xcp, struct fpactx *ctx)
{
    union mips_instr ir;
    vaddr_t contpc;
    unsigned int cond;
    int skipbd;
    int err = 0;

    ir.value = mips_get_word (REG_TO_VA (xcp->epc), &err);
    if (err)
	return SIGBUS;

    /* XXX NEC Vr54xx bug workaround */
    if ((XCP_CAUSE (xcp) & CR_BD) && !isBranchInstr (&ir))
	(xcp)->cr = XCP_CAUSE (xcp) & ~CR_BD;

    if (XCP_CAUSE (xcp) & CR_BD) {
	int             sig;

	/* The instruction to be emulated is in a branch delay slot
	 * which means that we have to  emulate the branch instruction
	 * BEFORE we do the cop1 instruction. 
	 *
	 * This branch could be a COP1 branch, but in that
	 * case we would have had a trap for that instruction, 
	 * and would not come through this route.
	 */

	if (sig = _mips_bremul (xcp, ctx, ir.value, &contpc, &skipbd)) {
#ifdef CP1DBG
	    _mon_printf ("failed to emulate branch at %p\n",
			 REG_TO_VA (xcp->epc));
#endif
	    return sig;
	}
	ir.value = mips_get_word (REG_TO_VA (xcp->epc) + 4, &err);
	if (err)
	    return SIGBUS;
    } else {
	contpc = xcp->epc + 4;
    }

  emul:
    switch (ir.itype.op) {
#if defined(SR_FR) && (__mips != 32 || __mipsfp64)
   /* R4000+ 64-bit fpu registers */
#if !#fpu(single)
    case ITYPE_ldc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.itype.rs]) + ir.itype.imm;
	    ctx->regs[ir.itype.rt] = mips_get_dword (va, &err);
	    if (err)
		return SIGBUS;
	}
	break;

    case ITYPE_sdc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.itype.rs]) + ir.itype.imm;
	    if (mips_put_dword (va, ctx->regs[ir.itype.rt]))
		return SIGBUS;
	}
	break;
#endif

    case ITYPE_lwc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.itype.rs]) + ir.itype.imm;
	    _fpreg_t	val;
	    val = mips_get_word (va, &err);
	    if (err)
		return SIGBUS;
	    if (XCP_STATUS (xcp) & SR_FR) {
	      /* load whole register */
	      ctx->regs[ir.itype.rt] =  val;
	    } else if (ir.itype.rt & 1) {
	      /* load to m.s. 32 bits */
#if #fpu(single)
	      /* illegal register in single-float mode */
	      return SIGILL;
#else
	      HIPART(ir.itype.rt & ~1) = val;
#endif
	    } else {
	      /* load to l.s. 32 bits */
	      LOPART(ir.itype.rt) = val;
	    }
	}
	break;

    case ITYPE_swc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.itype.rs]) + ir.itype.imm;
	    unsigned int    val;
	    if ((ir.itype.rt & 1) && !(XCP_STATUS(xcp) & SR_FR)) {
#if #fpu(single)
	      /* illegal register in single-float mode */
	      return SIGILL;
#else
	      /* store from m.s. 32 bits */
	      val = HIPART(ir.itype.rt & ~1);
#endif
	    } else {
	      /* store from l.s. 32 bits */
	      val = LOPART(ir.itype.rt);
	    }
	    if (mips_put_word (va, val))
		return SIGBUS;
	}
	break;
#else /* old 32-bit fpu registers */
    case ITYPE_lwc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.itype.rs]) + ir.itype.imm;
	    ctx->regs[ir.itype.rt] = mips_get_word (va, &err);
	    if (err)
		return SIGBUS;
	}
	break;

    case ITYPE_swc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.itype.rs]) + ir.itype.imm;
	    if (mips_put_word (va, ctx->regs[ir.itype.rt]))
		return SIGBUS;
	}
	break;

    case ITYPE_ldc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.itype.rs]) + ir.itype.imm;
	    unsigned int rt = ir.itype.rt & ~1;
#if #endian(big)
	    ctx->regs[rt + 1] = mips_get_word (va + 0, &err);
	    ctx->regs[rt + 0] = mips_get_word (va + 4, &err);
#else
	    ctx->regs[rt + 0] = mips_get_word (va + 0, &err);
	    ctx->regs[rt + 1] = mips_get_word (va + 4, &err);
#endif
	    if (err)
		return SIGBUS;
	}
	break;

    case ITYPE_sdc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.itype.rs]) + ir.itype.imm;
	    unsigned int rt = ir.itype.rt & ~1;
#if #endian(big)
	    if (mips_put_word (va + 0, ctx->regs[rt + 1]))
		return SIGBUS;
	    if (mips_put_word (va + 4, ctx->regs[rt + 0]))
		return SIGBUS;
#else
	    if (mips_put_word (va + 0, ctx->regs[rt + 0]))
		return SIGBUS;
	    if (mips_put_word (va + 4, ctx->regs[rt + 1]))
		return SIGBUS;
#endif
	}
	break;
#endif

    case ITYPE_c1:
	switch (ir.rrtype.rs) {

#if defined(SR_FR) && (__mips != 32 || __mipsfp64)
#if !#fpu(single)
#if __mips >= 3 || __mips64
	case COPZRS_dmf:
	    /* copregister rd -> gpr[rt] */
	    if (ir.rrtype.rt != 0)
	    {
		xcp->regs[ir.rrtype.rt] = ctx->regs[ir.rrtype.rd];
	    }
	    break;

	case COPZRS_dmt:
	    /* copregister rd <- rt */
	    {
		ctx->regs[ir.rrtype.rd] = xcp->regs[ir.rrtype.rt];
	    }
	    break;
#else /* gp reg size < 8 */
	case COPZRS_dmt:
	    /* allow copregister rd <- $0 only */
	    {
		if (ir.rrtype.rt != 0)
		    return SIGILL;
		ctx->regs[ir.rrtype.rd] = 0;
	    }
	    break;
#endif /* gp reg size < 8 */
#endif /* ! #fpu(single) */

	case COPZRS_mf:
	    /* 32-bit copregister rd -> gpr[rt] */
	    if (ir.rrtype.rt != 0)
	    {
		int value;
		if ((ir.rrtype.rd & 1) && !(XCP_STATUS (xcp) & SR_FR))
		    /* get m.s. bits */
		    value = HIPART(ir.rrtype.rd & ~1);
		else 
		    /* get l.s. bits */
		    value = LOPART(ir.rrtype.rd);
		xcp->regs[ir.rrtype.rt] = value;
	    }
	    break;

	case COPZRS_mt:
	    /* 32-bit copregister rd <- rt */
	    {
		unsigned int value;
		value = xcp->regs[ir.rrtype.rt];
		if ((ir.rrtype.rd & 1) && !(XCP_STATUS (xcp) & SR_FR))
		    /* set m.s. bits only */
		    HIPART(ir.rrtype.rd & ~1) = value;
		else 
		    /* set l.s. bits only */
		    LOPART(ir.rrtype.rd) = value;
	    }
	    break;

#if __mips == 32 || __mips == 64
	case COPZRS_mfh:
	    /* 32-bit copregister rd -> gpr[rt] */
	    if (ir.rrtype.rt != 0)
	    {
		int value;
		value = HIPART(ir.rrtype.rd);
		xcp->regs[ir.rrtype.rt] = value;
	    }
	    break;

	case COPZRS_mth:
	    /* 32-bit copregister rd <- rt */
	    {
		unsigned int value;
		value = xcp->regs[ir.rrtype.rt];
		HIPART(ir.rrtype.rd) = value;
	    }
	    break;
#endif

#else

	case COPZRS_mf:
	    /* copregister rd -> gpr[rt] */
	    if (ir.rrtype.rt != 0)
	    {
		int value = ctx->regs[ir.rrtype.rd];
		xcp->regs[ir.rrtype.rt] = value;
	    }
	    break;

	case COPZRS_mt:
	    /* copregister rd <- rt */
	    {
		unsigned        value;
		value = xcp->regs[ir.rrtype.rt];
		ctx->regs[ir.rrtype.rd] = value;
	    }
	    break;

#if __mips == 32 || __mips == 64
	case COPZRS_mfh:
	    /* 32-bit copregister rd -> gpr[rt] */
	    if (ir.rrtype.rt != 0)
	    {
		int value;
		value = ctx->regs[ir.rrtype.rd | 1];
		xcp->regs[ir.rrtype.rt] = value;
	    }
	    break;

	case COPZRS_mth:
	    /* 32-bit copregister rd <- rt */
	    {
		unsigned int value;
		value = xcp->regs[ir.rrtype.rt];
		ctx->regs[ir.rrtype.rd | 1] = value;
	    }
	    break;
#endif
#endif

	case COPZRS_cf:
	    /* cop control register rd -> gpr[rt] */
	    {
		unsigned        value;

		if (ir.rrtype.rd == FPCREG_CSR) {
		    value = ctx->sr;
#ifdef CSRTRACE
		    _mon_printf ("%p gpr[%d]<-csr=%08x\n",
			     REG_TO_VA (xcp->epc), ir.rrtype.rt, value);
#endif
		} else if (ir.rrtype.rd == FPCREG_RID)
		    value = 0;
		else
		    value = 0;
		if (ir.rrtype.rt)
		    xcp->regs[ir.rrtype.rt] = value;
	    }
	    break;

	case COPZRS_ct:
	    /* copregister rd <- rt */
	    {
		unsigned        value;

		value = xcp->regs[ir.rrtype.rt];

		/* we only have one writable control reg
		 */
		if (ir.rrtype.rd == FPCREG_CSR) {
#ifdef CSRTRACE
		    _mon_printf ("%p gpr[%d]->csr=%08x\n",
			     REG_TO_VA(xcp->epc), ir.rrtype.rt, value);
#endif
		    ctx->sr = value;
		    /* copy new rounding mode to ieee library state! */
		    ieee754_csr.rm = ieee_rm[value & 0x3];
		}
	    }
	    break;

	case COPZRS_bc:
	    if (XCP_CAUSE (xcp) & CR_BD)
		return SIGILL;
	    {
		int		likely = 0;

#if __mips >= 4
		cond = ctx->sr & fpacondbit[ir.rrtype.rt >> 2];
#else
		cond = ctx->sr & FPA_CSR_COND;
#endif
		switch (ir.rrtype.rt & 3) {
		case COPZRT_BCFL:
		    likely = 1;
		case COPZRT_BCF:
		    cond = !cond;
		    break;
		case COPZRT_BCTL:
		    likely = 1;
		case COPZRT_BCT:
		    break;
		default:
		    /* thats an illegal instruction */
		    return SIGILL;
		}

		xcp->cr |= CR_BD;
		if (cond) {
		    /* branch taken: emulate dslot instruction */
		    xcp->epc += 4;
		    contpc = xcp->epc + (ir.itype.imm << 2);

		    ir.value = mips_get_word (REG_TO_VA (xcp->epc), &err);
		    if (err)
			return SIGBUS;

		    switch (ir.itype.op) {
		    case ITYPE_lwc1:
		    case ITYPE_swc1:
#if (__mips >= 2 || __mips64) && !#fpu(single)
		    case ITYPE_ldc1:
		    case ITYPE_sdc1:
#endif
		    case ITYPE_c1:
#ifdef MIPS4_FPU
		    case ITYPE_c1x:
#endif
			/* its one of ours */
			goto emul;
#if __mips >= 4
		    case ITYPE_special:
			if (ir.rrtype.funct == RRTYPE_movci)
			    goto emul;
#endif
		    }

		    /* single step the non-cp1 instruction in the dslot */
		    return mips_dsemul (xcp, ir, contpc);
		} else {
		    /* branch not taken */
		    if (likely)	
			/* branch likely nullifies dslot if not taken */
			contpc += 4;
		    /* else continue & execute dslot as normal insn */
		}
	    }
	    break;

	default:
	    if (!(ir.cp1type.fmt & 0x10))
		return SIGILL;
	    /* a real fpu computation instruction */
	    {
		int             sig;
		if (sig = fpa_emu (xcp, ctx, ir))
		    return sig;
	    }
	}
	break;

#ifdef MIPS4_FPU
    case ITYPE_c1x:
    {
	int             sig;
	if (sig = fpax_emu (xcp, ctx, ir))
	    return sig;
    }
    break;
#endif

#if __mips >= 4
    case ITYPE_special:
	if (ir.rrtype.funct != RRTYPE_movci)
	    return SIGILL;
	cond = fpacondbit[ir.rrtype.rt >> 2];
	if (((ctx->sr & cond) != 0) == ((ir.rrtype.rt & 1) != 0)
	    && ir.rrtype.rd != 0)
	    xcp->regs[ir.rrtype.rd] = xcp->regs[ir.rrtype.rs];
	break;
#endif

    default:
	return SIGILL;
    }

    /* we did it !! */
    xcp->epc = VA_TO_REG (contpc);
    xcp->cr &= ~CR_BD;
    return 0;
}

/*
 * Emulate the arbritrary instruction ir at xcp->epc.  Required when
 * we have to emulate the instruction in a COP1 branch delay slot.  Do
 * not change epc due to the instruction
 *
 * According to the spec:
 * 1) it shouldnt be a branch :-)
 * 2) it can be a COP instruction :-(
 * 3) if we are trying to run a protected memory space we must take
 *    special care on memory access instructions :-(
 */

static union mips_instr dsemul_insns[2];
static vaddr_t		dsemul_cpc;
static unsigned int	dsemul_sr;
static struct xcptaction dsemul_osys;


static int
mips_dsemulret (int xno, struct xcptcontext *xcp)
{
#ifdef DSEMUL_TRACE
    _mon_printf ("desemulret\n");
#endif
    xcptaction (XCPTSYS, &dsemul_osys, 0);
    xcp->epc = VA_TO_REG (dsemul_cpc);
    xcp->sr = dsemul_sr;
    return 0;
}


static int
mips_dsemul (struct xcptcontext *xcp, union mips_instr ir, vaddr_t cpc)
{
    if (ir.value == 0) {	/* a nop is easy */
	xcp->epc = VA_TO_REG (cpc);
	/* BD is no longer valid */
	(xcp)->cr = XCP_CAUSE (xcp) & ~CR_BD;
	return 0;
    }

#ifdef DSEMUL_TRACE
    _mon_printf ("desemul %p %p\n", REG_TO_VA (xcp->epc), cpc);
#endif

    /* the strategy is to push the instruction into memory
     * and put a trap after it which we can catch and jump to 
     * the required address any alternative apart from full 
     * instruction emulation!!.
     */
    {
	struct xcptaction nxa;

	dsemul_insns[0] = ir;
	dsemul_insns[1].ttype.op = ITYPE_special;
	dsemul_insns[1].ttype.funct = RRTYPE_syscall;
	dsemul_cpc = cpc;
	dsemul_sr = xcp->sr;

	nxa.xa_handler = mips_dsemulret;
	nxa.xa_flags = 0;
	xcptaction (XCPTSYS, &nxa, &dsemul_osys);
    }
    xcp->epc = VA_TO_REG (dsemul_insns);
    xcp->sr &= ~SR_IMASK;	/* interrupt disabled inside dsemul */
#if ! #cache(no)
    if (IS_KVA0 (dsemul_insns))
	mips_sync_icache ((vaddr_t) dsemul_insns, sizeof (dsemul_insns));
#endif
    return SIGILL;	/* force out of emulation loop */
}

/* 
 * Conversion table from MIPS compare ops 48-63
 * cond = ieee754dp_cmp(x,y,IEEE754_UN);
 */
static const unsigned char cmptab[8] =
{
    0,						/* cmp_0 (sig) cmp_sf */
    IEEE754_CUN,				/* cmp_un (sig) cmp_ngle */
    IEEE754_CEQ,				/* cmp_eq (sig) cmp_seq */
    IEEE754_CEQ | IEEE754_CUN,			/* cmp_ueq (sig) cmp_ngl  */
    IEEE754_CLT,				/* cmp_olt (sig) cmp_lt */
    IEEE754_CLT | IEEE754_CUN,			/* cmp_ult (sig) cmp_nge */
    IEEE754_CLT | IEEE754_CEQ,			/* cmp_ole (sig) cmp_le */
    IEEE754_CLT | IEEE754_CEQ | IEEE754_CUN,	/* cmp_ule (sig) cmp_ngt */
};

#define SIFROMREG(si,x)	((si) = ctx->regs[x])
#define SITOREG(si,x)	(ctx->regs[x] = (int)(si))

#if __mips64 && !#fpu(single)
#define DIFROMREG(di,x)	((di) = ctx->regs[x])
#define DITOREG(di,x)	(ctx->regs[x] = (di))
#endif

#define SPFROMREG(sp,x)	(sp = (ieee754sp)ctx->regs[x])
#define SPTOREG(sp,x)	(ctx->regs[x] = (_fpreg_t)sp)

#if defined(SR_FR) && (__mips != 32 || __mipsfp64)
#define DPFROMREG(dp,x)	(dp = (ieee754dp)ctx->regs[x])
#define DPTOREG(dp,x)	(ctx->regs[x] = (_fpreg_t)dp)
#else
/* Beware: MIPS COP1 doubles are always little_word endian in registers */
#define DPFROMREG(dp,x)	\
  (dp = (ieee754dp)(((unsigned long long)ctx->regs[(x)+1] << 32) | ctx->regs[x]))
#define DPTOREG(dp,x) \
  (ctx->regs[x] = (_fpreg_t)dp, ctx->regs[(x)+1] = (_fpreg_t)(dp >> 32))
#endif

#ifdef MIPS4_FPU

/*
 * Additional MIPS4 instructions
 */

static ieee754dp fpemu_dp_recip (ieee754dp d)
{
    return ieee754dp_div (ieee754dp_one(0), d);
}

static ieee754dp fpemu_dp_rsqrt (ieee754dp d)
{
    return ieee754dp_div (ieee754dp_one(0), ieee754dp_nsqrt (d));
}

static ieee754sp fpemu_sp_recip (ieee754sp s)
{
    return ieee754sp_div (ieee754sp_one(0), s);
}

static ieee754sp fpemu_sp_rsqrt (ieee754sp s)
{
    return ieee754sp_div (ieee754sp_one(0), ieee754sp_nsqrt (s));
}

static ieee754dp fpemu_dp_madd (ieee754dp r, ieee754dp s, ieee754dp t)
{
    return ieee754dp_add (ieee754dp_mul (s, t), r);
}

static ieee754dp fpemu_dp_msub (ieee754dp r, ieee754dp s, ieee754dp t)
{
    return ieee754dp_sub (ieee754dp_mul (s, t), r);
}

static ieee754dp fpemu_dp_nmadd (ieee754dp r, ieee754dp s, ieee754dp t)
{
    return ieee754dp_neg (ieee754dp_add (ieee754dp_mul (s, t), r));
}

static ieee754dp fpemu_dp_nmsub (ieee754dp r, ieee754dp s, ieee754dp t)
{
    return ieee754dp_neg (ieee754dp_sub (ieee754dp_mul (s, t), r));
}


static ieee754sp fpemu_sp_madd (ieee754sp r, ieee754sp s, ieee754sp t)
{
    return ieee754sp_add (ieee754sp_mul (s, t), r);
}

static ieee754sp fpemu_sp_msub (ieee754sp r, ieee754sp s, ieee754sp t)
{
    return ieee754sp_sub (ieee754sp_mul (s, t), r);
}

static ieee754sp fpemu_sp_nmadd (ieee754sp r, ieee754sp s, ieee754sp t)
{
    return ieee754sp_neg (ieee754sp_add (ieee754sp_mul (s, t), r));
}

static ieee754sp fpemu_sp_nmsub (ieee754sp r, ieee754sp s, ieee754sp t)
{
    return ieee754sp_neg (ieee754sp_sub (ieee754sp_mul (s, t), r));
}

static int
fpax_emu (struct xcptcontext *xcp, struct fpactx *ctx, union mips_instr ir)
{
    unsigned rcsr = 0;	/* resulting csr */

    switch (ir.cp1xtype.fmt) {
    case CP1SFMT:		/* 0 */
    {
	ieee754sp (*handler) (ieee754sp, ieee754sp, ieee754sp);
	ieee754sp fd, fr, fs, ft;

	switch (ir.cp1xtype.funct) {
	case COP1X_lxc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.cp1xtype.fr] + 
				  xcp->regs[ir.cp1xtype.ft]);
 	    _fpreg_t	val;
	    int err = 0;
	    val = mips_get_word (va, &err);
	    if (err)
		return SIGBUS;
	    if (XCP_STATUS(xcp) & SR_FR) {
	      /* load whole register */
	      ctx->regs[ir.cp1xtype.fd] =  val;
	    } else if (ir.cp1xtype.fd & 1) {
#if #fpu(single)
	      /* illegal register in single-float mode */
	      return SIGILL;
#else
	      /* load to m.s. 32 bits */
	      HIPART(ir.cp1xtype.fd & ~1) = val;
#endif
	    } else {
	      /* load to l.s. 32 bits */
	      LOPART(ir.cp1xtype.fd) = val;
	    }
	}
	break;

	case COP1X_sxc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.cp1xtype.fr] + 
				  xcp->regs[ir.cp1xtype.ft]);
	    unsigned int    val;
	    if (XCP_STATUS (xcp) & SR_FR) {
	      /* store whole register */
	      val = ctx->regs[ir.cp1xtype.fs];
	    } else if (ir.cp1xtype.fs & 1) {
#if #fpu(single)
	      /* illegal register in single-float mode */
	      return SIGILL;
#else
	      /* store from m.s. 32 bits */
	      val = HIPART(ir.cp1xtype.fs & ~1);
#endif
	    } else {
	      /* store from l.s. 32 bits */
	      val = LOPART(ir.cp1xtype.fs);
	    }
	    if (mips_put_word (va, val))
		return SIGBUS;
	}
	break;

	case COP1X_madd:	
	    handler = fpemu_sp_madd;
	    goto scoptop;
	case COP1X_msub:
	    handler = fpemu_sp_msub;
	    goto scoptop;
	case COP1X_nmadd:
	    handler = fpemu_sp_nmadd;
	    goto scoptop;
	case COP1X_nmsub:
	    handler = fpemu_sp_nmsub;
	    goto scoptop;

	scoptop:
	    SPFROMREG (fr, ir.cp1xtype.fr);
	    SPFROMREG (fs, ir.cp1xtype.fs);
	    SPFROMREG (ft, ir.cp1xtype.ft);
	    fd = (*handler) (fr, fs, ft);
	    SPTOREG (fd, ir.cp1type.fd);

	copcsr:
	    if (ieee754_cxtest (IEEE754_INEXACT))
		rcsr |= FPA_CSR_INE_X | FPA_CSR_INE_S;
	    if (ieee754_cxtest (IEEE754_UNDERFLOW))
		rcsr |= FPA_CSR_UDF_X | FPA_CSR_UDF_S;
	    if (ieee754_cxtest (IEEE754_OVERFLOW))
		rcsr |= FPA_CSR_OVF_X | FPA_CSR_OVF_S;
	    if (ieee754_cxtest (IEEE754_INVALID_OPERATION))
		rcsr |= FPA_CSR_INV_X | FPA_CSR_INV_S;

	    ctx->sr = (ctx->sr & ~FPA_CSR_ALL_X) | rcsr;
	    if ((ctx->sr >> 5) & ctx->sr & FPA_CSR_ALL_E) {
		/*_mon_printf ("SIGFPE: fpa csr = %08x\n",ctx->sr); */
		return SIGFPE;
	    }

	    break;

	default:
	    return SIGILL;
	}
    }
    break;

#if !#fpu(single)
    case CP1DFMT:		/* 1 */
    {
	ieee754dp (*handler) (ieee754dp, ieee754dp, ieee754dp);
	ieee754dp fd, fr, fs, ft;

	switch (ir.cp1xtype.funct) {
	case COP1X_lxc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.cp1xtype.fr] + 
				  xcp->regs[ir.cp1xtype.ft]);
	    int err = 0;
	    ctx->regs[ir.cp1xtype.fd] = mips_get_dword (va, &err);
	    if (err)
		return SIGBUS;
	}
	break;

	case COP1X_sxc1:
	{
	    void *va = REG_TO_VA (xcp->regs[ir.cp1xtype.fr] + 
				  xcp->regs[ir.cp1xtype.ft]);
	    if (mips_put_dword (va, ctx->regs[ir.cp1xtype.fs]))
		return SIGBUS;
	}
	break;

	case COP1X_madd:	
	    handler = fpemu_dp_madd;
	    goto dcoptop;
	case COP1X_msub:
	    handler = fpemu_dp_msub;
	    goto dcoptop;
	case COP1X_nmadd:
	    handler = fpemu_dp_nmadd;
	    goto dcoptop;
	case COP1X_nmsub:
	    handler = fpemu_dp_nmsub;
	    goto dcoptop;

	dcoptop:
	    DPFROMREG (fr, ir.cp1xtype.fr);
	    DPFROMREG (fs, ir.cp1xtype.fs);
	    DPFROMREG (ft, ir.cp1xtype.ft);
	    fd = (*handler) (fr, fs, ft);
	    DPTOREG (fd, ir.cp1type.fd);
	    goto copcsr;

	default:
	    return SIGILL;
	}
    }
    break;
#endif

    case 0x7:			/* 7 */
    {
	if (ir.cp1xtype.funct != COP1X_sxc1)
	    return SIGILL;
	/* ignore prefx operation */
    }
    break;

    default:
	return SIGILL;
    }

    return 0;
}
#endif



/*
 * Emulate a single COP1 arithmetic instruction.
 */
static int
fpa_emu (struct xcptcontext *xcp, struct fpactx *ctx, union mips_instr ir)
{
    int             rfmt;	/* resulting format */
    unsigned        rcsr = 0;	/* resulting csr */
    unsigned	    cond;
    union {
	ieee754dp       d;
	ieee754sp       s;
	int             w;
#if __mips64
	long long       l;
#endif
    } rv;			/* resulting value */

    switch (rfmt = (ir.cp1type.fmt & 0xf)) {

    case CP1SFMT:{		/* 0 */
	    ieee754sp (*handler) ();

	    switch (ir.cp1type.funct) {
		/* binary ops */
	    case COP1_add:
		handler = ieee754sp_add;
		goto scopbop;
	    case COP1_sub:
		handler = ieee754sp_sub;
		goto scopbop;
	    case COP1_mul:
		handler = ieee754sp_mul;
		goto scopbop;
	    case COP1_div:
		handler = ieee754sp_div;
		goto scopbop;

		/* unary  ops */
#if __mips >= 2 || __mips64
	    case COP1_sqrt:
		handler = ieee754sp_nsqrt;
		goto scopuop;
#endif
#ifdef MIPS4_FPU
	    case COP1_rsqrt:
		handler = fpemu_sp_rsqrt;
		goto scopuop;
	    case COP1_recip:
		handler = fpemu_sp_recip;
		goto scopuop;
#endif
#if __mips >= 4
	    case COP1_movcf:
		cond = fpacondbit[ir.cp1type.ft >> 2];
		if (((ctx->sr & cond) != 0) != ((ir.cp1type.ft & 1) != 0))
		    return 0;
		SPFROMREG (rv.s, ir.cp1type.fs);
		break;
	    case COP1_movz:
		if (xcp->regs[ir.cp1type.ft] != 0)
		    return 0;
		SPFROMREG (rv.s, ir.cp1type.fs);
		break;
	    case COP1_movn:
		if (xcp->regs[ir.cp1type.ft] == 0)
		    return 0;
		SPFROMREG (rv.s, ir.cp1type.fs);
		break;
#endif
	    case COP1_abs:
		handler = ieee754sp_abs;
		goto scopuop;
	    case COP1_neg:
		handler = ieee754sp_neg;
		goto scopuop;
	    case COP1_mov:
		/* an easy one */
		SPFROMREG (rv.s, ir.cp1type.fs);
		break;
		/* binary op on handler */
	      scopbop:
		{
		    ieee754sp       fs, ft;

		    SPFROMREG (fs, ir.cp1type.fs);
		    SPFROMREG (ft, ir.cp1type.ft);

		    rv.s = (*handler) (fs, ft);
		    goto copcsr;
		}
	      scopuop:
		{
		    ieee754sp       fs;

		    SPFROMREG (fs, ir.cp1type.fs);
		    rv.s = (*handler) (fs);
		    goto copcsr;
		}
	      copcsr:
		if (ieee754_cxtest (IEEE754_INEXACT))
		    rcsr |= FPA_CSR_INE_X | FPA_CSR_INE_S;
		if (ieee754_cxtest (IEEE754_UNDERFLOW))
		    rcsr |= FPA_CSR_UDF_X | FPA_CSR_UDF_S;
		if (ieee754_cxtest (IEEE754_OVERFLOW))
		    rcsr |= FPA_CSR_OVF_X | FPA_CSR_OVF_S;
		if (ieee754_cxtest (IEEE754_ZERO_DIVIDE))
		    rcsr |= FPA_CSR_DIV_X | FPA_CSR_DIV_S;
		if (ieee754_cxtest (IEEE754_INVALID_OPERATION))
		    rcsr |= FPA_CSR_INV_X | FPA_CSR_INV_S;
		break;

		/* unary conv ops */
	    case COP1_cvts:
		return SIGILL;	/* not defined */
	    case COP1_cvtd:
#if #fpu(single)
		return SIGILL;	/* not defined */
#else
		{
		    ieee754sp       fs;

		    SPFROMREG (fs, ir.cp1type.fs);
		    rv.d = ieee754dp_fsp (fs);
		    rfmt = CP1DFMT;
		    goto copcsr;
		}
#endif
	    case COP1_cvtw:
		{
		    ieee754sp       fs;

		    SPFROMREG (fs, ir.cp1type.fs);
		    rv.w = ieee754sp_tint (fs);
		    rfmt = CP1WFMT;
		    goto copcsr;
		}

#if __mips >= 2 || __mips64
	    case COP1_round:
	    case COP1_trunc:
	    case COP1_ceil:
	    case COP1_floor:
		{
		    unsigned int    oldrm = ieee754_csr.rm;
		    ieee754sp       fs;

		    SPFROMREG (fs, ir.cp1type.fs);
		    ieee754_csr.rm = ieee_rm[ir.cp1type.funct & 0x3];
		    rv.w = ieee754sp_tint (fs);
		    ieee754_csr.rm = oldrm;
		    rfmt = CP1WFMT;
		    goto copcsr;
		}
#endif /* __mips >= 2 */

#if __mips64 && !#fpu(single)
	    case COP1_cvtl:
		{
		    ieee754sp       fs;

		    SPFROMREG (fs, ir.cp1type.fs);
		    rv.l = ieee754sp_tlong (fs);
		    rfmt = CP1LFMT;
		    goto copcsr;
		}

	    case COP1_roundl:
	    case COP1_truncl:
	    case COP1_ceill:
	    case COP1_floorl:
		{
		    unsigned int    oldrm = ieee754_csr.rm;
		    ieee754sp       fs;

		    SPFROMREG (fs, ir.cp1type.fs);
		    ieee754_csr.rm = ieee_rm[ir.cp1type.funct & 0x3];
		    rv.l = ieee754sp_tlong (fs);
		    ieee754_csr.rm = oldrm;
		    rfmt = CP1LFMT;
		    goto copcsr;
		}
#endif /* __mips64 && !fpu(single) */

	    default:
		if (ir.cp1type.funct >= COP1_cmp_f) {
		    unsigned        cmpop = ir.cp1type.funct - COP1_cmp_f;
		    ieee754sp       fs, ft;

		    SPFROMREG (fs, ir.cp1type.fs);
		    SPFROMREG (ft, ir.cp1type.ft);
		    rv.w = ieee754sp_cmp (fs, ft, cmptab[cmpop & 0x7]);
		    rfmt = -1;
		    if ((cmpop & 0x8) 
			&& ieee754_cxtest (IEEE754_INVALID_OPERATION))
		      rcsr = FPA_CSR_INV_X | FPA_CSR_INV_S;
		} else
		    return SIGILL;
		break;
	    }
	    break;
	}

#if !#fpu(single)
    case CP1DFMT:{
	    ieee754dp (*handler) ();

	    switch (ir.cp1type.funct) {
		/* binary ops */
	    case COP1_add:
		handler = ieee754dp_add;
		goto dcopbop;
	    case COP1_sub:
		handler = ieee754dp_sub;
		goto dcopbop;
	    case COP1_mul:
		handler = ieee754dp_mul;
		goto dcopbop;
	    case COP1_div:
		handler = ieee754dp_div;
		goto dcopbop;

		/* unary  ops */
#if __mips >= 2 || __mips64
	    case COP1_sqrt:
		handler = ieee754dp_nsqrt;
		goto dcopuop;
#endif
#ifdef MIPS4_FPU
	    case COP1_rsqrt:
		handler = fpemu_dp_rsqrt;
		goto dcopuop;
	    case COP1_recip:
		handler = fpemu_dp_recip;
		goto dcopuop;
#endif
#if __mips >= 4
	    case COP1_movcf:
		cond = fpacondbit[ir.cp1type.ft >> 2];
		if (((ctx->sr & cond) != 0) != ((ir.cp1type.ft & 1) != 0))
		    return 0;
		DPFROMREG (rv.d, ir.cp1type.fs);
		break;
	    case COP1_movz:
		if (xcp->regs[ir.cp1type.ft] != 0)
		    return 0;
		DPFROMREG (rv.d, ir.cp1type.fs);
		break;
	    case COP1_movn:
		if (xcp->regs[ir.cp1type.ft] == 0)
		    return 0;
		DPFROMREG (rv.d, ir.cp1type.fs);
		break;
#endif
	    case COP1_abs:
		handler = ieee754dp_abs;
		goto dcopuop;
	    case COP1_neg:
		handler = ieee754dp_neg;
		goto dcopuop;
	    case COP1_mov:
		/* an easy one */
		DPFROMREG (rv.d, ir.cp1type.fs);
		break;

		/* binary op on handler */
	      dcopbop:
		{
		    ieee754dp       fs, ft;

		    DPFROMREG (fs, ir.cp1type.fs);
		    DPFROMREG (ft, ir.cp1type.ft);

		    rv.d = (*handler) (fs, ft);
		    goto copcsr;
		}
	      dcopuop:
		{
		    ieee754dp       fs;

		    DPFROMREG (fs, ir.cp1type.fs);
		    rv.d = (*handler) (fs);
		    goto copcsr;
		}

		/* unary conv ops */
	    case COP1_cvts:
		{
		    ieee754dp       fs;

		    DPFROMREG (fs, ir.cp1type.fs);
		    rv.s = ieee754sp_fdp (fs);
		    rfmt = CP1SFMT;
		    goto copcsr;
		}
	    case COP1_cvtd:
		return SIGILL;	/* not defined */
	    case COP1_cvtw:
		{
		    ieee754dp       fs;

		    DPFROMREG (fs, ir.cp1type.fs);
		    rv.w = ieee754dp_tint (fs);		/* wrong */
		    rfmt = CP1WFMT;
		    goto copcsr;
		}

#if __mips >= 2 || __mips64
	    case COP1_round:
	    case COP1_trunc:
	    case COP1_ceil:
	    case COP1_floor:
		{
		    unsigned int    oldrm = ieee754_csr.rm;
		    ieee754dp       fs;

		    DPFROMREG (fs, ir.cp1type.fs);
		    ieee754_csr.rm = ieee_rm[ir.cp1type.funct & 0x3];
		    rv.w = ieee754dp_tint (fs);
		    ieee754_csr.rm = oldrm;
		    rfmt = CP1WFMT;
		    goto copcsr;
		}
#endif

#if __mips64 && !#fpu(single)
	    case COP1_cvtl:
		{
		    ieee754dp       fs;

		    DPFROMREG (fs, ir.cp1type.fs);
		    rv.l = ieee754dp_tlong (fs);
		    rfmt = CP1LFMT;
		    goto copcsr;
		}

	    case COP1_roundl:
	    case COP1_truncl:
	    case COP1_ceill:
	    case COP1_floorl:
		{
		    unsigned int    oldrm = ieee754_csr.rm;
		    ieee754dp       fs;

		    DPFROMREG (fs, ir.cp1type.fs);
		    ieee754_csr.rm = ieee_rm[ir.cp1type.funct & 0x3];
		    rv.l = ieee754dp_tlong (fs);
		    ieee754_csr.rm = oldrm;
		    rfmt = CP1LFMT;
		    goto copcsr;
		}
#endif /* __mips >= 3 && !fpu(single) */

	    default:
		if (ir.cp1type.funct >= COP1_cmp_f) {
		    unsigned        cmpop = ir.cp1type.funct - COP1_cmp_f;
		    ieee754dp       fs, ft;

		    DPFROMREG (fs, ir.cp1type.fs);
		    DPFROMREG (ft, ir.cp1type.ft);
		    rv.w = ieee754dp_cmp (fs, ft, cmptab[cmpop & 0x7]);
		    rfmt = -1;
		    if ((cmpop & 0x8) 
			&& ieee754_cxtest (IEEE754_INVALID_OPERATION))
		      rcsr = FPA_CSR_INV_X | FPA_CSR_INV_S;
		} else
		    return SIGILL;
		break;
	    }
	    break;
	}
#endif /* !#fpu(single) */

    case CP1WFMT:
	{
	    switch (ir.cp1type.funct) {
	    case COP1_cvts:
		/* convert word to single precision real 
		 */
		rv.s = ieee754sp_fint (ctx->regs[ir.cp1type.fs]);
		rfmt = CP1SFMT;
		goto copcsr;
#if !#fpu(single)
	    case COP1_cvtd:
		/* convert word to double precision real 
		 */
		rv.d = ieee754dp_fint (ctx->regs[ir.cp1type.fs]);
		rfmt = CP1DFMT;
		goto copcsr;
#endif
	    default:
		return SIGILL;
	    }
	    break;
	}

#if __mips64 && !#fpu(single)
    case CP1LFMT:
	{
	    switch (ir.cp1type.funct) {
	    case COP1_cvts:
		/* convert long to single precision real */
		rv.s = ieee754sp_flong (ctx->regs[ir.cp1type.fs]);
		rfmt = CP1SFMT;
		goto copcsr;
	    case COP1_cvtd:
		/* convert long to double precision real */
		rv.d = ieee754dp_flong (ctx->regs[ir.cp1type.fs]);
		rfmt = CP1DFMT;
		goto copcsr;
	    default:
		return SIGILL;
	    }
	    break;
	}
#endif

    default:
	return SIGILL;
    }

    /*
     * Update the fpu CSR register for this operation.
     * If an exception is required, generate a tidy SIGFPE exception,
     * without updating the result register.
     * Note: cause exception bits do not accumulate, they are rewritten
     * for each op; only the flag/sticky bits accumulate.
     */
    ctx->sr = (ctx->sr & ~FPA_CSR_ALL_X) | rcsr;
    if ((ctx->sr >> 5) & ctx->sr & FPA_CSR_ALL_E) {
	/*_mon_printf ("SIGFPE: fpa csr = %08x\n",ctx->sr); */
	return SIGFPE;
    }

    /* 
     * Now we can safely write the result back to the register file.
     */
    switch (rfmt) {
    case -1:
    {
#if __mips >= 4
	cond = fpacondbit[ir.cp1type.fd >> 2];
#else
	cond = FPA_CSR_COND;
#endif
	if (rv.w)
	    ctx->sr |= cond;
	else
	    ctx->sr &= ~cond;
	break;
    }
#if !#fpu(single)
    case CP1DFMT:
	DPTOREG (rv.d, ir.cp1type.fd);
	break;
#endif
    case CP1SFMT:
	SPTOREG (rv.s, ir.cp1type.fd);
	break;
    case CP1WFMT:
	SITOREG (rv.w, ir.cp1type.fd);
	break;
#if __mips64 && !#fpu(single)
    case CP1LFMT:
	DITOREG (rv.l, ir.cp1type.fd);
	break;
#endif
    default:
	abort ();
    }

    return 0;
}

#if ! #fpu(no)
/*
 * Patch up the hardware fpu state when an f.p. exception occurs.  
 */

static int
cop1Patcher (int xcptno, struct xcptcontext *xcp)
{
    struct fpactx  *ctx = &_fpaemustate;
    unsigned        sr;
    int             sig;

    /* reenable Cp1, else fpa_save() will get nested exception */
    sr = mips_bissr (SR_CU1);

    /* get fpu registers and status, then clear pending exceptions */
    fpa_save (ctx);
    fpa_setsr (ctx->sr &= ~FPA_CSR_ALL_X);

    /* get current rounding mode for IEEE library, and emulate insn */
    ieee754_csr.rm = ieee_rm[ctx->sr & 0x3];
    sig = cop1Emulate (xcptno, xcp, ctx);

    /* don't return with f.p. exceptions pending */
    ctx->sr &= ~FPA_CSR_ALL_X;
    fpa_restore (ctx);

    mips_setsr (sr);
    return sig;
}
#endif /* ! #fpu(no) */

/*
 * Emulate the floating point instruction at EPC, and continue
 * to run until we hit a non-fp instruction, or a backward
 * branch.  This cuts down dramatically on the per instruction 
 * exception overhead.
 */

static int
cop1Handler (int xcptno, struct xcptcontext *xcp)
{
    struct fpactx  *ctx = &_fpaemustate;
    reg_t	oldepc, prevepc;
    unsigned int insn;
    int         sig = 0;
    int		err = 0;

    oldepc = xcp->epc;
    do {
	prevepc = xcp->epc;
	insn = mips_get_word (REG_TO_VA (xcp->epc), &err);
	if (err)
	    return SIGBUS;
	if (insn != 0)
	    sig = cop1Emulate (xcptno, xcp, ctx);
	else
	    xcp->epc += 4;	/* skip nops */
    } while (xcp->epc > prevepc && sig == 0);

    /* SIGILL indicates a non-fpu instruction */
    if (sig == SIGILL && xcp->epc != oldepc)
	/* but if epc has advanced, then ignore it */
	sig = 0;

    return sig;
}


void
_cop1_init (int emulate)
{
    extern int      _nofpu;

    if (emulate) {
	/* 
	 * Install cop1 emulator to handle "coprocessor unusable" exception
	 */
	mips_bicsr (SR_CU1);
	xcption (XCPTCPU, cop1Handler);
	_fpaemuactive = 1;	/* tell dbg.c that we are in charge */
	_nofpu = 0;		/* tell setjmp() it "has" an fpu */
    } else {
#if ! #fpu (no)
	/* 
	 * Install cop1 emulator for floating point exceptions only,
	 * i.e. denormalised results, underflow, overflow etc, which
	 * must be emulated in s/w.
	 */
#if ! defined(XCPTFPE_ABSENT)
	/* r4000 or above use dedicate exception */
	xcption (XCPTFPE, cop1Patcher);
#else
	/* r3000 et al use interrupt */
	extern int _sbd_getfpaintr (void);
	int intno = _sbd_getfpaintr ();
	intrupt (intno, cop1Patcher, 0);
	mips_bissr (SR_IM0 << intno);
#endif

#if (#cpu(r4640) || #cpu(r4650)) && !#fpu(single)
	/* For R4640/R4650 compiled *without* the -msingle-float flag,
	   then we share responsibility: the h/w handles the single
	   precision operations, and the trap emulator handles the
	   double precision. We set _fpaemuactive so that dbg.c first
	   fetches the s/w state before saving the h/w state. */
	_fpaemuactive = 1;
	{
	    int i;
	    /* initialise the unused d.p high order words to be NaN */
	    for (i = 0; i < 32; i++)
		_fpaemustate.regs[i] = 0x7ff80bad00000000LL;
	}
#endif /* (r4640 || r4650) && !fpu(single) */
#endif /* ! #fpu (no) */

    }
}
