#ident "kit/share/cw4watch.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * cw4watch.c: LSI CoreWare family watchpoint support for MIPS SDE kit
 */


#if #cache(cw10) || #cache(cw01)
#include <sys/types.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <mips/prid.h>
#if __mips16
#include <mips/instr16def.h>
#endif
#include <mips/watchpoint.h>
#include <string.h>

#include "kit_impl.h"

static int	cw4_wp_init (unsigned int);
#ifdef C0_DCS
static void	cw4_wp_setup (struct mips_watchpoint *);
static int 	cw4_wp_hit (const struct xcptcontext *, vaddr_t *, size_t *);
static int	cw4_wp_set (struct mips_watchpoint *);
static int	cw4_wp_clear (struct mips_watchpoint *);
static void	cw4_wp_remove (void);
static void	cw4_wp_insert (void);
static void	cw4_wp_reset (void);
#endif

#if #cpu(cw401x)
#define	cw4_setbpc(x)	cw401x_setbpc(x)
#define	cw4_setbpcm(x)	cw401x_setbpcm(x)
#define	cw4_setbda(x)	cw401x_setbda(x)
#define	cw4_setbdam(x)	cw401x_setbdam(x)
#define	cw4_setdcs(x)	cw401x_setdcs(x)
#define	cw4_getdcs()	cw401x_getdcs()
#else
#define	cw4_setbpc(x)	cw400x_setbpc(x)
#define	cw4_setbpcm(x)	cw400x_setbpcm(x)
#define	cw4_setbda(x)	cw400x_setbda(x)
#define	cw4_setbdam(x)	cw400x_setbdam(x)
#define	cw4_setdcs(x)	cw400x_setdcs(x)
#define	cw4_getdcs()	cw400x_getdcs()
#endif

const struct mips_watchpoint_funcs _cw4_watchpoint_funcs = {
    cw4_wp_init
#ifdef C0_DCS
    ,
    cw4_wp_setup,
    cw4_wp_hit,
    cw4_wp_set,
    cw4_wp_clear,
    cw4_wp_remove,
    cw4_wp_insert,
    cw4_wp_reset
#endif
};


struct cw4_wp {
    unsigned long	reg;
    unsigned long	mask;
    unsigned long	dcs;
};


#define IWATCH	0
#define DWATCH	1
static struct cw4_wp	cw4_wp[2];
static unsigned int	cw4_dcs;


static int	
cw4_wp_init (unsigned int prid)
{
#ifdef C0_DCS
    return 2;
#else
    return 0;
#endif
}


#ifdef C0_DCS
static void
cw4_wp_setup (struct mips_watchpoint *wp)
{
    if (wp->num == IWATCH)
	wp->capabilities = MIPS_WATCHPOINT_VADDR | MIPS_WATCHPOINT_MASK
	    | MIPS_WATCHPOINT_X;
    else
	wp->capabilities = MIPS_WATCHPOINT_VADDR | MIPS_WATCHPOINT_MASK 
	    | MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W;
    wp->private = &cw4_wp[wp->num];
    wp->maxmask = ~0;
}


static int
cw4_wp_hit (const struct xcptcontext *xcp, vaddr_t *pva, size_t *plen)
{
    int hit = 0;

    if (cw4_dcs & (DCS_DA | DCS_PC)) {
	hit = (cw4_dcs & DCS_PC) ? MIPS_WATCHPOINT_X 
	    : (MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W);
	hit = _mips_watchpoint_address (xcp, hit, pva, plen);
    }

    return hit;
}


static int
cw4_wp_set (struct mips_watchpoint *wp)
{
    struct cw4_wp *pwp = wp->private;
    vaddr_t va = wp->va;

#if __mips16
    if (wp->type == MIPS_WATCHPOINT_X && (va & 1) && wp->len <= 2) {
      /* must point at second half of two word insn */
	union mips16_instr i;
	int err = 0;
	i.value = mips_get_half (va & ~1, &err);
	if (!err) {
	    if (i.itype.op == ITYPE16_extend
		|| i.itype.op == ITYPE16_jal)
		va += 2;
	}
    }
#endif

    /* mask is a "do care" not a "don't care" */
    pwp->mask = ~_mips_watchpoint_calc_mask (va & ~1, wp->len);
    pwp->reg = va & pwp->mask;

    pwp->dcs = 0;
    if (wp->type == MIPS_WATCHPOINT_X) {
	/* always check MIPS16 bit */
	pwp->mask |= (va & 1);
	pwp->dcs |= DCS_PCE;
    }
    else {
      /* catch any access to enclosing word */
	pwp->mask &= ~3;
	wp->mask |= 3;
	if (wp->type & MIPS_WATCHPOINT_R)
	    pwp->dcs |= DCS_DR;
	if (wp->type & MIPS_WATCHPOINT_W)
	    pwp->dcs |= DCS_DW;
    }
    return 1;
}


static int
cw4_wp_clear (struct mips_watchpoint *wp)
{
    struct cw4_wp *pwp = wp->private;
    if (pwp->reg)
	pwp->reg = pwp->dcs = 0;
    return 1;
}


static void	
cw4_wp_remove (void)
{
    if (cw4_wp[IWATCH].reg || cw4_wp[IWATCH].reg) {
	cw4_dcs = cw4_getdcs ();
	cw4_setdcs (0);
    }
}


static void
cw4_wp_insert (void)
{
    unsigned int dcs = 0;

    if (cw4_wp[IWATCH].reg) {
	cw4_setbpc (cw4_wp[IWATCH].reg);
	cw4_setbpcm (cw4_wp[IWATCH].mask);
	dcs |= cw4_wp[IWATCH].dcs;
    }	

    if (cw4_wp[DWATCH].reg) {
	cw4_setbda (cw4_wp[DWATCH].reg);
	cw4_setbdam (cw4_wp[DWATCH].mask);
	dcs |= cw4_wp[DWATCH].dcs;
    }

    if (dcs != 0)
	cw4_setdcs (dcs);
}


static void
cw4_wp_reset (void)
{
    cw4_setdcs (cw4_dcs = 0);
    memset (cw4_wp, 0, sizeof (cw4_wp));
}

#endif /* C0_DCS */
#endif /* #cache(cw10) || #cache(cw01) */
