#ident "kit/share/dbg.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * dbg.c: gdb remote debugger support for MIPS SDE
 */


#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <sys/sdeversion.h>
#include <mips/xcpt.h>
#include <mips/cpu.h>
#include <mips/fpa.h>
#include <mips/instrdef.h>
#include <mips/instr16def.h>
#include <mips/watchpoint.h>

#include "kit_impl.h"
#include "sbd.h"

#ifndef TLB_V
#undef DBGVA		/* experimental "virtual address" support */
#endif

/* FIXME these should probably be global */
#if __mips16 || __maybe_mips16
#define IS_MIPS16(va) ((unsigned int)(va) & 1)
#define MIPS16ADDR(va)  (vaddr_t)((unsigned int)(va) & ~1)
#else
#define IS_MIPS16(va) (0)
#define MIPS16ADDR(va)  (va)
#endif

/* convert MIPS16 register to MIPS32 register */
#define mips16_to_mips32(rn)	(((rn) < 2) ? (rn) + 16 : (rn))

#if (__mips >= 3 && __mips != 32) || __mips64
/* always use 64-bit debug regs on a 64-bit cpu, irrespective of __mips64 */
typedef unsigned long long	dbgreg_t;
#else
typedef unsigned long 		dbgreg_t;
#endif

#define REG_TO_VADDR(r)		(vaddr_t)REG_TO_VA(r)
#define VADDR_TO_REG(v)		VA_TO_REG((void *)v)

#if #cache(no)
# define SYNC_ICACHE(va, sz)
#else
# define SYNC_ICACHE(va, sz)	(mips_sync_icache(va, sz))
#endif

/* Data shared with optional modules must be common or weak extern. */
int _fpaemuactive __attribute__ ((common));		/* Set by cp1emu.c */
struct fpactx _fpaemustate __attribute__ ((common)) ;	/* Set by cp1emu.c */

static int      	dbg_enable;	/* Remote debugging enabled. */
static int             	dbg_waiting;	/* Remote debugger is waiting. */
static int             	dbg_extended;	/* Using extended protocol. */
static unsigned int 	dbg_regsize;	/* Debug register size. */
static char		*dbg_errstr;	/* last debugger error */

/* when doing a single-step disable interrupts if this is set */
int			_dbg_threadsafe_sstep __attribute__ ((common));
static long 		dbg_srsave;	/* saved status register */

/* magic break instructions which we skip over */
extern char 		debug_breakpoint_insn[] asm(".debug_breakpoint_insn");
extern char 		debug_sighook_insn[] asm(".debug_sighook_insn");

/* last watchpoint information */
static int		watch_hit;	/* at watchpoint */
static int		watch_step;	/* stepping over watchpoint */
static vaddr_t		watch_va;	/* watchpoint data address */
static size_t		watch_len;	/* watchpoint data len */
static int		watch_num;	/* # watchpoints */

/* base and offset symbols */
extern char _ftext[], _fdata[], _fbss[];
extern char __attribute__((weak)) _foffs_ftext[0];
extern char __attribute__((weak)) _foffs_fdata[0];
extern char __attribute__((weak)) _foffs_fbss[0];

/* output string to debug port */
static void
_dbg_outs (const char *s)
{
    while (*s)
	_dbg_outch (*s++);
}


/* dummy thread support routines */
#pragma weak _dbg_getthread
int
_dbg_getthread (struct xcptcontext *xcp)
{
    return -1;
}

#pragma weak _dbg_newthread
int
_dbg_newthread (void)
{
    return -1;
}

#pragma weak _dbg_setthread
struct xcptcontext *
_dbg_setthread (struct xcptcontext *xcp, int general, int thread)
{
    return xcp;
}

#pragma weak _dbg_contthread
struct xcptcontext *
_dbg_contthread (struct xcptcontext *xcp)
{
    return xcp;
}

#pragma weak _dbg_querythread
int
_dbg_querythread (int thread)
{
    return 0;
}

#pragma weak _dbg_enterhook
void
_dbg_enterhook (void)
{
}

#pragma weak _dbg_exithook
int
_dbg_exithook (void)
{
    return 0;
}

#pragma weak _dbg_attachhook
void
_dbg_attachhook (void)
{
}

#pragma weak _dbg_detachhook
void
_dbg_detachhook (void)
{
}

#pragma weak _dbg_restart
int
_dbg_restart (struct xcptcontext *xcp, int signo)
{
    return signo;
}


#pragma weak _dbg_getfpactx
int
_dbg_getfpactx (struct xcptcontext *xcp, struct fpactx *fcp)
{
    if (_fpaemuactive)
	/* r4640/r4650 support - preload emulated top half */
	memcpy (fcp, &_fpaemustate, sizeof(struct fpactx));
    if (XCP_STATUS (xcp) & SR_CU1) {
	/* reenable Cp1, else fpa_save() will get CpU exception */	
	mips_bissr (SR_CU1);
	fpa_save (fcp);
    }
    else if (!_fpaemuactive) {
	return 0;
    }
    return 1;
}

#pragma weak _dbg_putfpactx
void
_dbg_putfpactx (struct xcptcontext *xcp, struct fpactx *fcp)
{
    extern int _fpaemuactive;
    extern struct fpactx _fpaemustate;
    if (XCP_STATUS (xcp) & SR_CU1)
	fpa_restore (fcp);
    if (_fpaemuactive)
	memcpy (&_fpaemustate, fcp, sizeof(struct fpactx));
}


#ifdef DBGVA
static int
dbg_get_tlblo (vaddr_t va, tlblo_t *lo, unsigned int *offs)
{
    int idx;

    if (mips_tlb_size () == 0)
	return -1;

#ifdef DBGVA_DYNAMIC
    /* touch this address to get it into the TLB */
    (void)*(volatile unsigned char *)va;
#endif /* DBGVA_DYNAMIC */

#if #mmu(type2)
    {
	tlblo_t even, odd;
	unsigned int pagemask;
	idx = mips_tlbprobe2 (va, &even, &odd, &pagemask);
	if (idx >= 0) {
	    /* convert page mask to page size */
	    pagemask = (pagemask + (1 << TLBHI_VPN2SHIFT)) >> 1;
	    /* select one of the two TlbLo entries */
	    *lo = (va & pagemask) ? odd : even;
	    if (offs)
		/* return offset in page */
		*offs = va & (pagemask - 1);
	}
	return idx;
    }
#elif #mmu(cw401x)
    {
	unsigned int pagemask;
	idx = cw401x_tlbprobe (va, lo, &pagemask);
	if (idx >= 0 && offs)
	    /* return offset in page */
	    *offs = va & (pagemask | ~TLBHI_VPNMASK);
	return idx;
    }
#elif #mmu(type1)
    idx = mips_tlbprobe (va, lo);
    if (idx > 0 & offs)
	*offs = va & VMPGMASK;
#else
    idx = -1;
#endif
    return idx;
}
#endif /* DBGVA */


#pragma weak _dbg_accessible
int
_dbg_accessible (vaddr_t va, size_t len, int writable)
{
    if (IS_KVA01 (va))
      return 1;

#if #cpu(r4640) || #cpu(r4650)
    if (IS_KVA2 (va))
	return 1;
    /* XXX assumes dbound >= ibound */
    if (va + len <= r4650_getdbound())
	return 1;
#elif #cpu(tr410x) || #cpu(cw400x)
    if (!(*(volatile unsigned int *)BSYS_SCR & SCR_TLBEN))
	return 1;			/* all accessible */
#elif defined(DBGVA)
    {
	tlblo_t lo;
	if (dbg_get_tlblo (va, &lo, 0) < 0)
	    return 0;
	if ((lo & TLB_V) == 0)
	    return 0;
	if (writable && (lo & TLB_D) == 0)
	    return 0;
	return 1;
    }
#endif
    return 0;
}


#pragma weak _dbg_cacheable
int
_dbg_cacheable (vaddr_t va)
{
#if #cpu(r4640) || #cpu(r4650)
    {
	/* 512MB/seg => 8 segs @ 4bits/seg => 32 bits */
	unsigned int seg = (unsigned long)va >> 29;
	return ((r4650_getcalg() >> (seg*4)) & 0xf) != CALG_UNCACHED;
    }
#else
    if (IS_KVA0(va))
      return 1;
    if (IS_KVA1(va))
      return 0;
#if #cpu(tr410x) || #cpu(cw400x)
    if (!(*(volatile unsigned int *)BSYS_SCR & SCR_TLBEN))
	return 1;			/* all else cacheable */
#endif
#if defined(DBGVA)
    {
	tlblo_t lo;
	if (dbg_get_tlblo (va, &lo, 0) < 0)
	    return 0;
	if ((lo & TLB_V) == 0)
	    return 0;
#if defined(TLB_CMASK)
	if ((lo & TLB_CMASK) == TLB_UNCACHED)
	    return 0;
#elif defined(TLB_N)
	if (lo & TLB_N)
	    return 0;
#else
#error "Don't know how to check cacheability"
#endif
	return 1;
    }
#else /* !DBGVA */
    return 0;
#endif /* DBGVA */
#endif /* !r4640/50 */
}


#pragma weak _dbg_vtop
paddr_t
_dbg_vtop (vaddr_t va)
{
    if (IS_KVA01(va))
      return KVA_TO_PA(va);

#if #cpu(tr410x) || #cpu(cw400x)
    if (!(*(volatile unsigned int *)BSYS_SCR & SCR_TLBEN))
	return (paddr_t)va;			/* no mapping */
#endif
#if #cpu(r4640) || #cpu(r4650)
    if (IS_KVA2 (va))
	return KVA2_TO_PA(va);
    /* XXX assumes dbound >= ibound and ibase == dbase */
    if (va < r4650_getdbound())
	return r4650_getdbase() + va;
    return (paddr_t)-1;
#elif defined(DBGVA)
    {
	tlblo_t lo;
	unsigned int offs;
	if (dbg_get_tlblo (va, &lo, &offs) < 0)
	    return (paddr_t)-1;
	if ((lo & TLB_V) == 0)
	    return (paddr_t)-1;
	return pnToPa(tlbloPn(lo)) + offs;
    }
#else
    return (paddr_t)-1;
#endif /* DBGVA */
}


#pragma weak _dbg_put_byte
int
_dbg_put_byte (vaddr_t va, unsigned char v)
{
  if (!_dbg_accessible (va, 1, 1))
      return 1;
  return (mips_put_byte ((void *)va, v));
}

#pragma weak _dbg_get_byte
unsigned int
_dbg_get_byte (vaddr_t va, int *err)
{
  if (!_dbg_accessible (va, 1, 0))
      return (*err = 1);
  return (mips_get_byte ((void *)va, err));
}

#pragma weak _dbg_put_half
int
_dbg_put_half (vaddr_t va, unsigned short v)
{
  if (!_dbg_accessible (va, 2, 1))
    return 1;
  return (mips_put_half ((void *)va, v));
}

#pragma weak _dbg_get_half
unsigned int
_dbg_get_half (vaddr_t va, int *err)
{
  if (!_dbg_accessible (va, 2, 0))
    return (*err = 1);
  return (mips_get_half ((void *)va, err));
}

#pragma weak _dbg_put_word
int
_dbg_put_word (vaddr_t va, unsigned int v)
{
  if (!_dbg_accessible (va, 4, 1))
    return 1;
  return (mips_put_word ((void *)va, v));
}

#pragma weak _dbg_get_word
unsigned int
_dbg_get_word (vaddr_t va, int *err)
{
  if (!_dbg_accessible (va, 4, 0))
    return (*err = 1);
  return (mips_get_word ((void *)va, err));
}


/* 
 * Do not even think about setting DBGDBG unless you really know
 * what is going on 
 */
#ifdef DBGDBG
int _dbgdbg = DBGDBG;
#define DBG(level, fmt, args...) \
    do { \
	if ((level) <= _dbgdbg) _dbg_printf (fmt , ## args); \
    } while (0)
#undef NDEBUG
#else
#define DBG(level, fmt, args...)
#endif

#ifndef NDEBUG

/* our own version of assert, to avoid recursive faults */
static void dbg_assfail(char *expr, char *line)
{
    _dbg_printf ("rdbg assertion failure `%s' at line %s.\n", expr, line);
}

#define assfail(expr, line)	dbg_assfail (expr, #line)

#define assert(expression)  \
  ((expression) ? 0 : (assfail (#expression, __LINE__), 0))

#else

#define assert(expression)

#endif


#define MAX_REG_SIZE		sizeof(dbgreg_t)

#if _SDE_MAJOR_ >= 3
#define NUM_REGS 	92	/* Gdb's idea of how many registers there are. */
#define R0_REGNUM	0	/* integer register base */
#define PS_REGNUM	32	/* processor status */
#define LO_REGNUM	33	/* mult/div temp */
#define HI_REGNUM	34	/* mult/div temp */
#define BADVADDR_REGNUM	35	/* bad vaddr */
#define CAUSE_REGNUM	36	/* cause */
#define PC_REGNUM	37	/* program counter */
#define FP0_REGNUM	38	/* FP register base */
#define FCRCS_REGNUM	70	/* FP control/status */
#define FCRIR_REGNUM	71	/* FP implementation/revision */
#define FP_REGNUM	72	/* pseudo frame pointer */
#define ACX_REGNUM	73	/* ACX register */
#define CP0_R0_REGNUM	74	/* CP0 reg $0 */
#define PRID_REGNUM	89	/* PRID register */
#else
#define NUM_REGS 	72	/* Gdb's idea of how many registers there are. */
#define R0_REGNUM	0	/* integer register base */
#define FP0_REGNUM	32	/* FP register base */
#define HI_REGNUM	64	/* mult/div temp */
#define LO_REGNUM	65	/* mult/div temp */
#define PC_REGNUM	66	/* program counter */
#define CAUSE_REGNUM	67	/* cause */
#define PS_REGNUM	68	/* processor status */
#define BADVADDR_REGNUM	69	/* bad vaddr */
#define FCRCS_REGNUM	70	/* FP control/status */
#endif
#define SP_REGNUM	(R0_REGNUM+29)	/* address of top of stack */

/* The maximum i/o buffer size. */
#define BUFMAX		(NUM_REGS * 8 * 2 + 32)
static char     	iobuffer[BUFMAX];

/* 
 * We need these local functions, in case a breakpoint has been
 * set in the libc routines!! 
 */

static int 
dbg_isxdigit (int c)
{
  return (c >= '0' && c <= '9') ||
    (c >= 'a' && c <= 'f') ||
      (c >= 'A' && c <= 'F');
}


static size_t 
dbg_strlen (const char *s)
{
  const char *e = s;
  while (*e)
    e++;
  return e - s;
}

/* not quite standard, return ptr to final <null> in dst string */
static char    *
dbg_strcpy (char *dst, const char *src)
{
  while (*dst++ = *src++)
    continue;
  return dst - 1;
}

static char
dbg_strncmp (const char *str1, const char *str2, int len)
{
    int i;
    
    for (i = 0; i < len; i++) {
        if (*str1 != *str2)
            return (*str1 - *str2);
        str1++;
        str2++;
    }

    return(0);
}

/*
 * convert hex char to number
 */
static int 
hexchToInt (char ch)
{
  if (ch >= '0' && ch <= '9')
    return (ch - '0');
  else if (ch >= 'a' && ch <= 'f')
    return (ch - 'a' + 10);
  else if (ch >= 'A' && ch <= 'F')
    return (ch - 'A' + 10);
  else
    return (0);
}

/* convert number to hex char */
static __inline__ char 
intToHexch (int n)
{
  return "0123456789abcdef"[n & 0xf];
}



/* Wait for a valid remote debug packet of the form
 *	$<packet data>#<checksum>    
 * and reply with a '+' for success, '-' on error. 
 */

static void 
getpacket (char *buffer)
{
  unsigned char   checksum;
  unsigned char   xmitcsum;
  int             count;
  char            ch, ch2;

  while (1) {

    /* wait for the start character, ignore all other characters */
    while ((ch = _dbg_inch ()) != '$')
      continue;

  restart:
    checksum = 0;
    count = 0;

    /* read until a # or buffer is full */
    while (count < BUFMAX - 1) {
      if ((ch = _dbg_inch ()) == '$')
	goto restart;
      if (ch == '#')
	break;
      checksum = checksum + ch;
      buffer[count++] = ch;
    }
    buffer[count] = 0;

    if (ch != '#') {
      DBG (1, "dbg: packet too long: %s\n", buffer);
      goto nak;
    }
	  
    /* get checksum */
    if ((ch = _dbg_inch ()) == '$')
      goto restart;
    if ((ch2 = _dbg_inch ()) == '$')
      goto restart;
    xmitcsum = (hexchToInt (ch) << 4) | hexchToInt (ch2);

    if (checksum == xmitcsum) {
      _dbg_outch ('+');		/* successful transfer */
      return;
    }

    DBG (1, "dbg: bad checksum, sent=%x, csum=%x, buf=%s#%c%c\n", 
	 xmitcsum, checksum, buffer, ch, ch2);

  nak:
    _dbg_outch ('-');		/* error response */
  }
}

/* 
 * Send a remote debug packet and wait for an acknowledge ('+').
 */
static void 
putpacket (char *buffer)
{
  unsigned char   checksum;
  char           *bp;
  char            ch;

  /*  $<packet data>#<checksum>. */
  do {
    _dbg_outch ('$');
    checksum = 0;
    for (bp = buffer; ch = *bp; bp++) {
      _dbg_outch (ch);
      checksum += ch;
    }

    _dbg_outch ('#');
    _dbg_outch (intToHexch (checksum >> 4));
    _dbg_outch (intToHexch (checksum));

  } while (_dbg_inch () != '+');
  
}

/* Scan a hex register (presented in native byte order) */

static char *
scanHexReg (char *s, dbgreg_t *ret)
{
  union {
    char	bytes[MAX_REG_SIZE];
    int		ival;
    dbgreg_t	rval;
  } reg;
  int 		i;
  unsigned rsize = dbg_regsize;

  assert (rsize == sizeof(int) || rsize == sizeof(dbgreg_t));

  for (i = 0; i < rsize; i++) {
    if (!dbg_isxdigit (s[0]) || !dbg_isxdigit (s[1]))
      return 0;
    reg.bytes[i] = (hexchToInt (s[0]) << 4) | hexchToInt (s[1]);
    s += 2;
  }

  if (rsize == sizeof(int))
    *ret = reg.ival;
  else
    *ret = reg.rval;
  return s;
}


/* Scan a hex integer (presented in big endian byte order) */
static char *
scanHexInt (char *s, int *ret)
{
  int	          v = 0;
  int		  negate = 0;

  if (*s == '-') {
      s++;
      negate = 1;
  }

  if (!dbg_isxdigit (*s))
    return 0;

  do {
    v = (v << 4) | hexchToInt (*s++);
  } while (dbg_isxdigit (*s));

  if (negate)
      v = -v;
  *ret = v;
  return s;
}


/* Output a register value in host byte order */
static char *
toHexReg (char *d, dbgreg_t v)
{
  union {
    char	bytes[MAX_REG_SIZE];
    int		ival;
    dbgreg_t	rval;
  } reg;
  int 		i;
  unsigned rsize = dbg_regsize;

  assert (rsize == sizeof(int) || rsize == sizeof(dbgreg_t));
  if (rsize == sizeof(int))
    reg.ival = v;
  else
    reg.rval = v;

  for (i = 0; i < rsize; i++) {
    *d++ = intToHexch (reg.bytes[i] >> 4);
    *d++ = intToHexch (reg.bytes[i]);
  }
  return d;
}


/* Output an integer value in m.s. byte order */
static char *
toHexInt (char *d, unsigned int v)
{
  int leading = 1;
  int i;

  for (i = sizeof(v) * 2; i != 0; i--, v <<= 4) {
      unsigned int nib = v >> 28;
      if (nib == 0 && leading && i != 1)
	  continue;
      *d++ = intToHexch (nib);
      leading = 0;
  }
  return d;
}

#if defined(SR_FR) && __mips != 32
#define BADFLOAT	0x7ff00bad7f000badLL
#else
#define BADFLOAT	0x7ff00bad
#endif

/* f.p. state cache */
static struct xcptcontext *fpawho;
static int		fpachanged;
static struct fpactx	fpactx;


static void
putfpactx (void)
{
  if (fpachanged) {
    _dbg_putfpactx (fpawho, &fpactx);
    fpachanged = 0;
  }
  fpawho = NULL;
}


static void
getfpactx (struct xcptcontext *xcp)
{
  if (fpawho != xcp) {
    if (fpawho)
	putfpactx ();
    if (!_dbg_getfpactx (xcp, &fpactx)) {
	int i;
	for (i = 0; i < 32; i++)
	    fpactx.regs[i] = BADFLOAT;    
    }
    fpawho = xcp;
  }
}



static dbgreg_t
getreg (struct xcptcontext *xcp, unsigned regno)
{
  switch (regno) {
  case R0_REGNUM ... R0_REGNUM+31:
    return xcp->regs[regno];
  case FP0_REGNUM ... FP0_REGNUM+31:
    getfpactx (xcp);
#if defined(SR_FR) && __mips != 32
    if (!(XCP_STATUS (xcp) & SR_FR)) {
	/* Each f.p. reg is 64-bit but it is simulating the old r3000
	   model where each reg is 32-bit.  This is done by selecting
	   one or other half of each even register; the odd numbered
	   64-bit registers are simply unused. */
	unsigned int *regp = (unsigned int *)&fpactx.regs[(regno - FP0_REGNUM) & ~1];
	/* note: f.p. register numbering is little-endian */
	dbgreg_t val = regp[(regno & 1) ^ (BYTE_ORDER != LITTLE_ENDIAN)];
#if (__mips >= 3 || __mips64) && BYTE_ORDER == BIG_ENDIAN
	/* gdb uses only first 4 bytes of each register in this mode 
	 * so make sure that the first 4 bytes contain the significant
	 * bits. */
	val <<= 32;
#else
	val &= 0xffffffff;
#endif
	return val;
    }
#endif /* cpu(r4000) */
    return fpactx.regs[regno - FP0_REGNUM];
  case HI_REGNUM:
    return xcp->mdhi;
  case LO_REGNUM:
    return xcp->mdlo;
  case PC_REGNUM:
#if 0
    if ((XCP_CAUSE (xcp) & CR_BD) && !IS_MIPS16 (xcp->epc))
	return xcp->epc + 4;
#endif
    return xcp->epc;
  case CAUSE_REGNUM:
    return xcp->cr;
  case PS_REGNUM: 
   return xcp->sr;
  case BADVADDR_REGNUM:
   return xcp->vaddr;
  case FCRCS_REGNUM:
    getfpactx (xcp);
    return fpactx.sr;
#ifdef ACX_REGNUM
  case ACX_REGNUM:
    return xcp->mdex;
#endif
#ifdef CP0_R0_REGNUM
  case CP0_R0_REGNUM ... CP0_R0_REGNUM+255:
    /* XXX return current coprocessor 0 register state */
    return (int) _mips_xxc0 (regno - CP0_R0_REGNUM, 0, 0);
#endif
  default:
    return 0;
  }
}


static void
putreg (struct xcptcontext *xcp, unsigned regno, dbgreg_t val)
{
  switch (regno) {
  case R0_REGNUM ... R0_REGNUM+31:
    xcp->regs[regno] = val;
    break;
  case FP0_REGNUM ... FP0_REGNUM+31:
    getfpactx (xcp);
    fpachanged = 1;
#if defined(SR_FR) && __mips != 32
    if (!(XCP_STATUS (xcp) & SR_FR)) {
      /* A mips3 cpu with simulated 32-bit f.p. registers, 
         there are no odd registers, so we have to write
	 the correct half of an even register. */
      unsigned int *regp;
      regp = (unsigned int *)&fpactx.regs[(regno - FP0_REGNUM) & ~1];
#if (__mips >= 3 || __mips64) && BYTE_ORDER == BIG_ENDIAN
      /* gdb uses only first 4 bytes of each register in this mode 
       * so extract the significant bits. */
      val >>= 32;	
#endif
      /* note: register numbering is little-endian */
      regp[(regno & 1) ^ (BYTE_ORDER != LITTLE_ENDIAN)] = val;
      break;
    }
#endif /* cpu(r4000) */
    fpactx.regs[regno - FP0_REGNUM] = val;
    break;
  case HI_REGNUM:
    xcp->mdhi = val;
    break;
  case LO_REGNUM:
    xcp->mdlo = val;
    break;
  case PC_REGNUM:
    xcp->epc = val;
    break;
  case PS_REGNUM: 
    xcp->sr = val;	/* dangerous! */
    break;
  case CAUSE_REGNUM:
    xcp->cr = val;	/* dangerous! */
    break;
  case BADVADDR_REGNUM:
    assert (val == xcp->vaddr);
    break;
#ifdef ACX_REGNUM
  case ACX_REGNUM:
    xcp->mdex = val;
    break;
#endif
#ifdef CP0_R0_REGNUM
  case CP0_R0_REGNUM ... CP0_R0_REGNUM+255:
    /* XXX Set random coprocessor 0 register state - VERY DANGEROUS */
    (void) _mips_xxc0 (regno - CP0_R0_REGNUM, ~0, (reg32_t)val);
    break;
#endif
  case FCRCS_REGNUM:
  default:
    assert (val == 0);
    break;
  }
}

static unsigned long signextend(unsigned long v, int n)
{
    long sv;
    int shift = (sizeof (unsigned long) * 8) - n;
    sv = v;
    return (unsigned long) ((sv << shift) >> shift);
}


/*
 * If the instruction about to be executed is a branch 
 * stash target of branch and return 1 else return 0
 * return -1 on error
 */

static int 
isBranchInstr (struct xcptcontext *x, vaddr_t *pTarget, vaddr_t *pNext)
{
    int err = 0;
    vaddr_t pc;
    union mipsn_instr i;
    union mips16_instr i16;

    pc = REG_TO_VADDR (x->epc);
    if (IS_MIPS16(pc)) {
	i16.value = _dbg_get_half (MIPS16ADDR(pc), &err);
	*pNext = pc + 2;
    }
    else if (pc & 3) {
	err = 1;
    }
    else {
	i.value = _dbg_get_word (pc, &err);
	*pNext = pc + 4;
    }

    if (err) {
    epcinaccessible:
        dbg_errstr = "isBranchInstr: epc inaccessible";
	return -1;
    }

    if (IS_MIPS16(pc)) {
	int extended = 0;
	int offset;
	unsigned long extension = 0;
    reswitch:
	switch (i16.itype.op) {
	case ITYPE16_b:
	    if (extended) {
		offset = i16.itype.imm & 0x1f;
		offset |= extension & 0x7e0;
		offset |= (extension & 0x1f) << 11;
		offset = signextend (offset, 16);
	    }
	    else
		offset = i16.itype.imm;
	    *pTarget = pc + 2 + (offset << 1);
	    return 1;

	case ITYPE16_jal:
	    pc += 2;
	    offset = _dbg_get_half (MIPS16ADDR(pc), &err);
	    if (err)
		goto epcinaccessible;
	    offset |= (i16.itype.imm & 0x1f) << 21;
	    offset |= (i16.itype.imm & 0x3e0) << (16-5);
	    *pTarget = (pc & ~0x0fffffff) | ((offset & 0x3ffffff) << 2);
	    if ((i16.itype.imm & 0x400) == 0)
		*pTarget |= 1;	/* not a jalx, so stay mips16 */
	    *pNext = pc + 4;	/* next insn follows offset+bdslot */
	    return 1;
	    
	case ITYPE16_beqz:
	case ITYPE16_bnez:
	    if (extended) {
		offset = i16.ritype.imm & 0x1f;
		offset |= extension & 0x7e0;
		offset |= (extension & 0x1f) << 11;
		offset = signextend (offset, 16);
	    }
	    else
		offset = i16.ritype.imm;
	    *pTarget = pc + 2 + (offset << 1);
	    return 1;

	case ITYPE16_i8:
	    switch (i16.i8type.funct) {
	    case I8TYPE16_bteqz:
	    case I8TYPE16_btnez:
		if (extended) {
		    offset = i16.i8type.imm & 0x1f;
		    offset |= extension & 0x7e0;
		    offset |= (extension & 0x1f) << 11;
		    offset = signextend (offset, 16);
		}
		else
		    offset = i16.i8type.imm;
		*pTarget = pc + 2 + (offset << 1);
		return 1;
	    }
	    break;

	case ITYPE16_rr:
	    if (i16.rrtype.funct == 0) {
		switch (i16.rrtype.ry) {
		case 1:		/* j ra */
		    *pTarget = REG_TO_VADDR(x->regs[31]);
		    *pNext = pc + 4;	/* next insn after jr + bdslot */
		    return 1;
		case 2:		/* jalr */
		case 0:		/* jr */
		    *pTarget = REG_TO_VADDR(x->regs[mips16_to_mips32(i16.rrtype.rx)]);
		    *pNext = pc + 4;	/* next insn after jr + bdslot */
		    return 1;
		}
	    }
	    break;

	case ITYPE16_extend:
	    extended = 1;
	    extension = i16.exttype.imm; /* we'll manipulate it later */
	    pc += 2;		/* This makes target calculations easier */
	    i16.value = _dbg_get_half (MIPS16ADDR(pc), &err);
	    if (err)
		goto epcinaccessible;
	    *pNext = pc + 2;	/* next insn after extended insn */
	    goto reswitch;
	}

	return 0;
    }

    switch (i.itype.op) {
    case ITYPE_special:
	switch (i.rtype.funct) {
	case RRTYPE_jalr:
	case RRTYPE_jr:
	    *pTarget = REG_TO_VADDR(x->regs[i.rtype.rs]);
	    *pNext = pc + 8;
	    return 1;
	}
	break;

    case ITYPE_j:
    case ITYPE_jal:
	*pTarget = (pc & ~0x0fffffff) | (i.jtype.target << 2);
	*pNext = pc + 8;
	return 1;

    case ITYPE_jalx:
	*pTarget = (pc & ~0x0fffffff) | (i.jtype.target << 2) | 1;
	*pNext = pc + 8;
	return 1;

    case ITYPE_regimm:
	switch (i.itype.rt) {
	case RITYPE_bltz:
	case RITYPE_bgez:
	case RITYPE_bltzl:
	case RITYPE_bgezl:
	case RITYPE_bltzal:
	case RITYPE_bgezal:
	case RITYPE_bltzall:
	case RITYPE_bgezall:
	    *pTarget = pc + (i.itype.imm << 2) + 4;
	    *pNext = pc + 8;
	    return 1;
	}
	break;

    case ITYPE_beq:
    case ITYPE_bne:
    case ITYPE_blez:
    case ITYPE_bgtz:
    case ITYPE_beql:
    case ITYPE_bnel:
    case ITYPE_blezl:
    case ITYPE_bgtzl:
	*pTarget = pc + (i.itype.imm << 2) + 4;
	*pNext = pc + 8;
	return 1;

    case ITYPE_c0:
    case ITYPE_c1:
    case ITYPE_c2:
    case ITYPE_c3:
	if (i.itype.rs == COPZRS_bc) {
	    switch (i.itype.rt & 3) {
	    case COPZRT_BCF:
	    case COPZRT_BCT:
	    case COPZRT_BCFL:
	    case COPZRT_BCTL:
		/* bczf, bczt, bczfl, bcztl */
		*pTarget = pc + (i.itype.imm << 2) + 4;
		*pNext = pc + 8;
		return 1;
	    }
	}
	break;
    }
    return 0;
}

#ifdef SIMULATESSTEP
#if __mips16
/* FIXME */
#error simulated single step not yet supported for mips16
#endif
#define BTAKENOFFSET	2
static int
simulateInstr (struct xcptcontext *x, vaddr_t *pPc,
	       union mips_instr *i, vaddr_t *pTarget, int *pLink)
{
    int err = 0;
    vaddr_t pc;

    pc = REG_TO_VADDR(x->epc);

    *pPc = pc;
    *pLink = -1;			/* default is no link register used */

    i->value = _dbg_get_word (pc, &err);
    if (err) {
	dbg_errstr = "simulateInstr: epc inaccessible";
	return -1;
    }

    switch (i->itype.op) {
    case ITYPE_special:
	switch (i->rtype.funct) {
	case RRTYPE_jalr:
	    *pLink = i->rtype.rd;
	    /* fall through */

	case RRTYPE_jr:
	    *pTarget = REG_TO_VADDR(x->regs[i->rtype.rs]);

	    if (i->rtype.funct == RRTYPE_jalr)
		i->itype.rt = RITYPE_bgezal;
	    else
		i->itype.rt = RITYPE_bgez;
	    i->itype.op = ITYPE_regimm;
	    i->itype.rs = 0;
	    i->itype.imm = BTAKENOFFSET;

	    return 1;
	}
	break;

    case ITYPE_jal:
    case ITYPE_jalx:
        *pLink = 31;			/* FIXME: ra */
	/* fall through */

    case ITYPE_j:
	*pTarget = (pc & ~0x0fffffff) | (i->jtype.target << 2);

	if (i->itype.op == ITYPE_jalx) {
	    *pTarget |= 1;
	    i->itype.rt = RITYPE_bgezal;
	}
	else if (i->itype.op == ITYPE_jal)
	    i->itype.rt = RITYPE_bgezal;
	else
	    i->itype.rt = RITYPE_bgez;
	    
	i->itype.op = ITYPE_regimm;
	i->itype.rs = 0;
	i->itype.imm = BTAKENOFFSET;
	    
	return 1;

    case ITYPE_regimm:
	switch (i->itype.rt) {
	case RITYPE_bltzal:
	case RITYPE_bgezal:
	case RITYPE_bltzall:
	case RITYPE_bgezall:
	    *pLink = 31;			/* FIXME: ra */
	    /* fall through */

	case RITYPE_bltz:
	case RITYPE_bgez:
	case RITYPE_bltzl:
	case RITYPE_bgezl:
	    *pTarget = pc + (i->itype.imm << 2) + 4;

	    i->itype.imm = BTAKENOFFSET;
	    return 1;
	}
	break;

    case ITYPE_beq:
    case ITYPE_bne:
    case ITYPE_blez:
    case ITYPE_bgtz:
    case ITYPE_beql:
    case ITYPE_bnel:
    case ITYPE_blezl:
    case ITYPE_bgtzl:
	*pTarget = pc + (i->itype.imm << 2) + 4;
	i->itype.imm = BTAKENOFFSET;
	return 1;

    case ITYPE_c0:
    case ITYPE_c1:
    case ITYPE_c2:
    case ITYPE_c3:
	if (i->itype.rs == COPZRS_bc) {
	    switch (i->itype.rt & 3) {
	    case COPZRT_BCF:
	    case COPZRT_BCT:
	    case COPZRT_BCFL:
	    case COPZRT_BCTL:
		/* bczf, bczt, bczfl, bcztl */
		*pTarget = pc + (i->itype.imm << 2) + 4;

		i->itype.imm = BTAKENOFFSET;
		return 1;
	    }
	}
	break;
    }
    return 0;
}
#endif /* SIMULATESSTEP */

#define NSSBP	2
static int      nssbp;
static vaddr_t	ssaddr[NSSBP];
static unsigned ssival[NSSBP];
static const unsigned kbp = 0x0000000d;	/* a breakpoint instr code 0 */
static const unsigned kbp16 = 0xe805;	/* a mips16 breakpoint instr code 0 */

#ifdef SIMULATESSTEP
static vaddr_t	simpc, simtarget;
int simlinkreg;
static unsigned	simcode[4];
#endif

static int
ss_addbp (vaddr_t va)
{
    int err = 0;

    if (nssbp >= NSSBP) {
	dbg_errstr = "ss_addbp: too many breakpoints";
	return 1;
    }

    DBG (2, "+ssbp[%d] at %#x ", nssbp, va);

    if (IS_MIPS16(va)) {
	ssival[nssbp] = _dbg_get_half (MIPS16ADDR(va), &err);
    }
    else if (va & 0x3) {
	dbg_errstr = "ss_addbp: bad address";
	return 1;
    }
    else
	ssival[nssbp] = _dbg_get_word (va, &err);

    if (err) {
	dbg_errstr = "ss_addbp: address inaccessible";
	return 1;
    }

    if (IS_MIPS16(va))
	err = _dbg_put_half (MIPS16ADDR(va), kbp16);
    else
	err = _dbg_put_word (va, kbp);
    if (err) {
	dbg_errstr = "ss_addbp: address unwritable";
	return 1;
    }

    if (_dbg_cacheable (va))
	SYNC_ICACHE (va, IS_MIPS16(va) ? 2 : 4);
    
    if (IS_MIPS16(va)) {
	if (_dbg_get_half (MIPS16ADDR(va), &err) != kbp16)
	    err = 1;
	DBG (2, "mips16 %04x->%04x\n", ssival[nssbp], kbp16);
    }
    else {
	if (_dbg_get_word (va, &err) != kbp)
	    err = 1;
	DBG (2, "mips32 %08x->%08x\n", ssival[nssbp], kbp);
    }
    
    if (err) {
	dbg_errstr = "ss_addbp: address not modifiable\n";
	return 1;
    }

    ssaddr[nssbp++] = va;
    return 0;
}


static int
ss_rembp (vaddr_t pc)
{
    int is_ssbp = 0;

    while (nssbp > 0) {
	vaddr_t va = ssaddr[--nssbp];
	DBG (2, "-ssbp at %#x>\n", va);
	if (va == pc)
	    is_ssbp = 1;
	/* ignore errors (this address has been written successfully before) */
	if (IS_MIPS16(va))
	    (void) _dbg_put_half (MIPS16ADDR(va), ssival[nssbp]);
	else
	    (void) _dbg_put_word (va, ssival[nssbp]);
	if (_dbg_cacheable (va))
	    SYNC_ICACHE (va, IS_MIPS16(va) ? 2 : 4);
    }
    return is_ssbp;
}

static int
ss_inplace (struct xcptcontext *x)
{
    vaddr_t target, next;
    int isbranch;
    int err = 0;
    
    /*
     * We have to execute just one instruction:
     * plant a breakpoint after the instr and
     * at the branch target if it is a branch.
     */
    if ((isbranch = isBranchInstr (x, &target, &next)) < 0)
	return 1;

    /* plant at following instruction */
    err = ss_addbp (next);
    /* plant at the target unless it is a branch to self or branch to next */
    if (!err && isbranch && next != target && REG_TO_VADDR(x->epc) != target)
	err = ss_addbp (target);
    return err;
}

#ifdef SIMULATESSTEP
static int
ss_simulate (struct xcptcontext *x)
{
    unsigned int *cp;
    union mips_instr i;
    int isbranch;
    int err = 0;

    /*
     * We have to execute just one instruction:
     * copy the instruction to a temporary location and insert
     * breakpoints after the instr and at the simulated branch
     * target if it is a branch.
     */

    if ((isbranch = simulateInstr (x, &simpc, &i, &simtarget, &simlinkreg)) < 0)
	return 1;

    cp = simcode;
    if (isbranch) {
	*cp++ = (vaddr_t) i.value;
	/* insert the branch delay instruction */
	*cp++ = _dbg_get_word (simpc + 4, &err);
	if (err)
	    return 1;
	/* plant breakpoint after branch delay slot */
	*cp++ = kbp;
	/* plant breakpoint at branch target */
	*cp++ = kbp;
	DBG (2, "simulating branch at %p: %x %x %x %x\n", simcode,
	     simcode[0], simcode[1], simcode[2], simcode[3]);
    } else {
	/* plant at following instruction */
	*cp++ = (vaddr_t) i.value;
	*cp++ = kbp;
	DBG (2, "simulating at %p: %x %x\n", simcode, simcode[0], simcode[1]);
    }
    if (_dbg_cacheable ((vaddr_t)simcode))
	SYNC_ICACHE ((vaddr_t)simcode, (cp - simcode) * sizeof (simcode[0]));
    x->epc = VA_TO_REG(simcode);
    return 0;
}
#endif

static int
ss_setup (struct xcptcontext *x)
{
    if (ss_inplace (x) == 0)
	return 0;
#ifdef SIMULATESSTEP
    if (ss_simulate (x) == 0)
	return 0;
#endif
    return 1;
}

static int
ss_cleanup (struct xcptcontext *x)
{
    vaddr_t pc = REG_TO_VADDR (x->epc);
    int is_ssbp;

    /* remove any single-step break points that we inserted */
    is_ssbp = ss_rembp (pc);
    
#ifdef SIMULATESSTEP
    if (pc == (vaddr_t)&simcode[0]) {
	x->epc = VA_TO_REG (simpc);		/* execution not started */
	is_ssbp = 1;
    }
    else if (pc == (vaddr_t)&simcode[1]) {
	x->epc = VA_TO_REG(simpc + 4);		/* simple single step */
	is_ssbp = 1;
    }
    else if (pc == (vaddr_t)&simcode[2]) {
	x->epc = VA_TO_REG(simpc + 8);		/* branch not taken */
	is_ssbp = 1;
    }
    else if (pc == (vaddr_t)&simcode[3]) {
	x->epc = VA_TO_REG(simtarget);		/* branch taken */
	is_ssbp = 1;
	/* update link register if required */
	if (simlinkreg >= 0)
	    x->regs[simlinkreg] = VA_TO_REG(simpc+8);
    }
#endif

    return is_ssbp;
}



/* This function can be called to force a breakpoint */
void __attribute__((noinline)) 
_debug_breakpoint ()
{
    if (dbg_enable) {
	__asm volatile  (".debug_breakpoint_insn: break 0");
	return;			/* --> type "c" to continue */
    } 				/* --> type "c" to continue */
}


/* This function is called by RTPX when an signal is about to
   be delivered to the current thread. 
   You can choose whether or not to let it be delivered. */
int __attribute__((noinline)) 
_debug_sighook (int signo)
{
    if (dbg_enable) {
	__asm volatile  (".debug_sighook_insn: break 0");
    }
    return signo;		/* --> type "c" to continue */
}


void 
_debug_init (int enable)
{
    int i;

    dbg_enable = enable;

    dbg_regsize = MAX_REG_SIZE;

    /* load stack frame interpreter */
    _xcptstackinfo_load ();

    /* catch all exceptions */
    for (i = 0; i < NXCPT; i++)
	_xcpt_catch (i, 1);		

    /* initialise watchpoint support */
    _mips_watchpoint_init ();
    watch_num = _mips_watchpoint_howmany ();

    if (_dbg_init ())
	_dbg_printf ("\r\nCannot init debug port!\r\n");

    /* arrange to catch additional interrupts/signals */
    _dbg_signals ();

    if (enable > 1)		/* connect now! */
	_debug_breakpoint ();
}


int 
_debug_enabled (void)
{
  return !!dbg_enable;
}


int 
_debug_connected (void)
{
  return dbg_waiting;
}


void
_debug_exit (int status)
{
#ifndef OLDGDB
  if (dbg_waiting) {
    if (status >> 8) {
	iobuffer[0] = 'X';
	status >>= 8;
    }
    else {
	iobuffer[0] = 'W';
    }
    iobuffer[1] = intToHexch (status >> 4);
    iobuffer[2] = intToHexch (status);
    iobuffer[3] = 0;
    putpacket (iobuffer);
  }
#endif
}


#if ! defined(OLDGDB) && 0
static char *
debug_statreg (char *p, unsigned gdbregno, reg_t rval)
{
  *p++ = intToHexch (gdbregno >> 4);
  *p++ = intToHexch (gdbregno);
  *p++ = ':';
  p = toHexReg (p, (dbgreg_t)rval);
  *p++ = ';';
  return p;
}
#endif


static void
debug_status (unsigned int signo, struct xcptcontext *x, int t)
{
  char *p = iobuffer;
#if defined(OLDGDB) || 1
  *p++ = 'S';
  *p++ = intToHexch (signo >> 4);
  *p++ = intToHexch (signo);
  if (watch_hit & (MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W)) {
      *p++ = 'w';
      p = toHexInt (p, watch_va);
      if (watch_len != 4) {
	  *p++ = ',';
	  p = toHexInt (p, watch_len);
      }
      *p++ = ';';
  }
#else
  *p++ = 'T';
  *p++ = intToHexch (signo >> 4);
  *p++ = intToHexch (signo);
  if (t >= 0) {
      p = dbg_strcpy (p, "thread:");
      p = toHexInt (p, t);
      *p++ = ';';
  }
  if (watch_hit & (MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W)) {
      *p++ = ',';
      p = dbg_strcpy (p, "watch:");
      p = toHexInt (p, watch_va);
      if (watch_len != 4) {
	  *p++ = ',';
	  p = toHexInt (p, watch_len);
      }
      *p++ = ';';
  }
  p = debug_statreg (p, PC_REGNUM, x->epc);
  p = debug_statreg (p, SP_REGNUM, x->regs[29]);
  /*p = debug_statreg (p, RA_REGNUM, x->regs[31]);*/
  p = debug_statreg (p, PS_REGNUM, x->sr);
  p = debug_statreg (p, CAUSE_REGNUM, x->cr);
#ifdef FP_REGNUM
  /*p = debug_statreg (p, FP_REGNUM, 0);*/
#endif
#endif
  *p = 0;
}



/* original thread exception context */
static struct xcptcontext *ox;

int
_mips_watchpoint_set_callback (int asid, vaddr_t va, size_t len)
{
    /* get watchpoint physical address range */
    paddr_t ws = _dbg_vtop (va);
    paddr_t we = ws + len;

    /* get exception stack frame limits */
    paddr_t ss = KVA_TO_PA (&asid);
    paddr_t se;
    if (XCP_USRSTACK (ox))
	se = KVA_TO_PA ((vaddr_t)ox->regs[XCP_SP]);
    else
	se = KVA_TO_PA (ox + 1);

    /* check for overlap */
    if (we > ss && se > ws)
	return MIPS_WP_OVERLAP;

    /* XXX can't easily check for overlap with other debugger data,
       but watchpoint should be disabled while in this code - the
       only dangerous window is when saving or restoring the 
       exception stack frame while watchpoints are enabled. */

    return MIPS_WP_OK;
}


/* Slow CRC32 routine, avoiding 1024 byte lookup table. */
static int
crc32 (vaddr_t mem, int len, unsigned int *crcp)
{
  unsigned int crc = 0xffffffff;
  int err = 0;

  while (len--) {
      unsigned int c;
      int i;
      c = _dbg_get_byte (mem++, &err);
      if (err)
	  return 0;
      c = ((crc >> 24) ^ c) << 24;
      for (i = 8; i != 0; --i)
	  c = (c << 1) ^ ((c & 0x80000000) ? 0x04c11db7 : 0);
      crc = (crc << 8) ^ c;
  }
  *crcp = crc;
  return 1;
}



int 
_debug_handler (int signo, struct xcptcontext *x)
{
  struct xcptcontext *gx, *cx, *nx;
  static int	inhere = 0;
  int		sstep, was_sstep;
  int		nsig;
  int		restart;
  int		addr;
  vaddr_t	membase;
  vaddr_t	mem;
  int		length;
  int		i;
  int		err = 0;
  int		th, oth, nth;
  int		threadcheck, endthreadcheck;

  if (!dbg_enable)
    return signo;
  
  if (inhere) {
#ifdef DBGVA
      /* XXX use some flag to say this is OK */
      if (signo == SIGSEGV) {
	  int isbranch;
	  vaddr_t target, next;

	  /*
	   * This was a reference to an unmappable address from the debugger
	   * simply skip the instruction and continue (we assume the routine
	   * knows what it was doing and can cope with the consequences).
	   * We also ignore the possibility that it was the pc that caused
	   * the trap.
	   */
	  if ((isbranch = isBranchInstr (x, &target, &next)) < 0)
	      return SIGKILL;

	  if (isbranch != 0) {
	      /*
	       * FIXME the miss happened in a branch delay slot
	       * without iterpreting the branch, we can't decide
	       * where to go, so just kill the debugger
	       */
	      DBG (1, "SIGSEGV in branch delay slot\n");
	      return SIGKILL;
	  }
	  x->epc = next;
	  return 0;
      }
#endif

    /* recursive debug */
    DBG (1, "Recursive debug trap: signo %d pc=%p\n", signo, 
	 REG_TO_VA (x->epc));
    return SIGKILL;
  }

  ++inhere;
  signo &= 0xff;
  DBG (1, "dbg: signo %d at %p\n", signo, REG_TO_VA (x->epc));

  /* save original signal which caused trap */
  x->signo = signo << 8;

  if (signo == SIGTRAP)
      signo = 0;

  /* save current application signal */
  x->signo |= signo;

  if (!dbg_waiting) {
#ifndef RDBG_DONT_WRITE
    /* Announce our first unexpected trap */
    if (signo != 0)
	xcpt_show (x);
    _dbg_printf ("Awaiting remote debugger...");
#endif

    /*
     * Announce via the debug port that we are going 
     * into remote debug mode.  If there a debugger
     * is already attached it won't hurt, but if there 
     * isn't then someone had better attach one.
     */
    _dbg_outs ("Connect remote debugger to this line...");

    watch_va = watch_len = 0;
    watch_hit = watch_step = 0;

    _dbg_attachhook ();
  }

  was_sstep = ss_cleanup (x);

  if (was_sstep && _dbg_threadsafe_sstep)
      /* restore previous interrupt enable */
      x->sr = dbg_srsave;

  if (!was_sstep) {
      vaddr_t va;
      size_t len;

      /* remove watchpoints */
      _mips_watchpoint_remove ();

      /* did we hit a watchpoint? */
      if (watch_hit = _mips_watchpoint_hit (x, &va, &len)) {

	  if (watch_hit & MIPS_WATCHPOINT_INEXACT) {
	      /* yes - but not one that we requested, try to step over it */
	      DBG (1, "inexact watchpoint: single-stepping\n",
		   REG_TO_VA (x->epc));
	      if (ss_setup (x) == 0) {
		  if (_dbg_threadsafe_sstep) {
		      dbg_srsave = x->sr;
		      x->sr &= ~SR_IE;
		  }
		  watch_step = 1;
		  inhere--;
		  return 0;
	      }
	  }	      

	  /* yes - a direct hit */
	  DBG (1, "watchpoint hit: va=0x%lx len=%lu\n", va, len);
	  watch_va = va;
	  watch_len = len;
	  watch_step = 0;
      }
  }
  else {
      if (watch_step && signo == 0) {
	  /* oontinue after stepping over watchpoint */
	  DBG (1, "end of watchpoint step: continuing\n");
	  _mips_watchpoint_insert ();
	  watch_step = 0;
	  inhere--;
	  return 0;
      }

      DBG (1, "normal debug trap\n");
      watch_hit = watch_step = 0;
  }

  _dbg_enterhook ();
  
  oth = th = _dbg_getthread (ox = x); 
  threadcheck = dbg_waiting;
  endthreadcheck = 0;

status:
  /* if gdb is waiting then tell gdb about any new threads 
     since we last spoke to it. */
  if (threadcheck) {
      if (!was_sstep && (nth = _dbg_newthread ()) >= 0) {
	  if (!(nx = _dbg_setthread (x, 1, nth)))
	      goto status;
	  x = nx;
	  th = nth;
	  putfpactx ();
      }
      else {
	  /* no more new threads, or this is the end of a single step */
	  if (endthreadcheck)
	      threadcheck = 0;
	  else
	      endthreadcheck = 1;
	  x = ox;
	  th = oth;
      }
  }

  if ((x->signo & 0xff) == 0 && x->epc != 0) {
      void * pc = REG_TO_VA (x->epc);
      if (pc == debug_breakpoint_insn) {
	  /* we're at our own break point, quietly STEP over it */
	  x->epc += IS_MIPS16 (pc) ? 2 : 4;
      }
      else if (pc == debug_sighook_insn) {
	  /* we're at signal hook break point, quietly STEP over it */
	  /* update old signum too, and note to pass back into a0 */
	  int sig = x->regs[XCP_A0] & 0xff;
	  x->signo = 0x10000 | (sig << 8) | sig;
	  x->epc += IS_MIPS16 (pc) ? 2 : 4;
      }
  }

  if (threadcheck)
      (void) ss_cleanup (x);	/* XXX do we need this */

  gx = cx = x;
  sstep = 0;

  if (dbg_waiting) {
    /* debugger was already waiting, report our new status. */
    debug_status ((cx->signo >> 8) & 0xff, cx, th);
    putpacket (iobuffer);
  }

  dbg_waiting = 0;	/* remote debugger no longer waiting */

  /* mark fpa registers as invalid */
  fpawho = NULL;
  fpachanged = 0;

  restart = 0;
  while (!restart) {
    unsigned       regno;
    dbgreg_t	   regval;
    char           *p = iobuffer;
    char	   req;
    int		   is_binary = 0;

    getpacket (p);
    DBG (3, "<req: %.40s> ", p);

    req = *p;

    if (req != 'c')
	threadcheck = 0;

    switch (req) {
    case '?':
      /* ?		Report status */
      gx = cx = x;
      if (p[1])
	  scanHexInt (p + 1, (int *)&dbg_regsize);
      debug_status ((x->signo >> 8) & 0xff, cx, th);
      break;

    case '!':
      dbg_extended = 1;
      break;

    case 'r':
      if (p[1] == '\0') {
	  /* r = reset the target */
	  /* XXX not implemented */
	  goto status;
      }
      /* FALL THROUGH */

    case 'p':
      /* pNN[:SS]	Return the value of CPU register NN (size SS) */
      if (!(p = scanHexInt (++p, (int *)&regno)))
	goto error;
      if (*p == ':') {
	if (!(p = scanHexInt (++p, (int *)&dbg_regsize)))
	  goto error;
      }
      p = toHexReg (iobuffer, getreg (gx, regno));
      *p = 0;
      break;

    case 'R':
      /* RNN[:SS],	Set the value of CPU register NN (size SS) */
      if (dbg_extended && p[3] == '\0') {
	  /* RXX = "restart" the application */
	  /* XXX not implemented, how do we reinitialise .data? */
	  goto status;
      }
      /* FALL THROUGH */

    case 'P':
      /* PNN[:SS]=	Set the value of CPU register NN (size SS) */
      if (!(p = scanHexInt (++p, (int *)&regno)))
	goto error;
      if (*p == ':') {
	if (!(p = scanHexInt (++p, (int *)&dbg_regsize)))
	  goto error;
      }
      if (*p != ((req == 'P') ? '=' : ','))
	goto error;
      if (!(p = scanHexReg (++p, &regval)))
	goto error;
      putreg (gx, regno, regval);
      goto success;

    case 'g':
      /* g		Return the value of all CPU registers */
      if (*++p) {
	if (!(p = scanHexInt (p, (int *)&dbg_regsize)))
	  goto error;
      }
      p = iobuffer;
      for (regno = 0; regno < NUM_REGS; regno++)
	p = toHexReg (p, getreg (gx, regno)); 
      *p = 0;
      break;

    case 'H':
      /* H[gc]nn		Set general/cont thread nn */
      if ((p[1] == 'g' || p[1] == 'c') && dbg_strlen(&p[2]) <= 17) {
	  int general = *++p == 'g';
	  p = scanHexInt (++p, &nth);
	  if (!(nx = _dbg_setthread (x, general, nth))) {
	      dbg_strcpy (iobuffer, "E99 cannot set thread");
	      break;
	  }
	  if (general) {
	      if (nx != gx)
		  /* flush fpa context */
		  putfpactx ();
	      gx = nx;
	  }
	  else {
	      if (nth >= 0)
		  th = nth;
	      cx = nx;
	  }
	  goto success;
      }

      /* H/J		Set the value of all CPU registers 64/32 wide */
    case 'J':
      dbg_regsize = (req == 'H') ? 8 : 4;
      /* drop through */

    case 'G':
      /* G		Set the value of all the CPU registers */
      if (dbg_strlen (++p) != NUM_REGS * dbg_regsize * 2) {
	dbg_strcpy (iobuffer, "E22 bad set-registers data size");
	break;
      }
      for (regno = 0; regno < NUM_REGS; regno++) {
	dbgreg_t regval;
	if (!(p = scanHexReg (p, &regval)))
	  goto error;
	putreg (gx, regno, regval);
      }
      goto success;

    case 'm':
      /* mAA..AA,LLLL	Read LLLL bytes at address AA..AA */
      if (!(p = scanHexInt (++p, &addr)))
	goto error;
      if (*p++ != ',')
	goto error;
      if (!scanHexInt (p, &length))
	goto error;
      if (length >= BUFMAX / 2)
	goto error;
      if (!_dbg_accessible (addr, length, 0))
	goto error;
      p = iobuffer;
      for (mem = addr, i = 0; i < length; mem++, i++) {
	  unsigned char m;
	  m = _dbg_get_byte (mem, &err);
	  if (err)
	      goto fault;
	  *p++ = intToHexch (m >> 4);
	  *p++ = intToHexch (m);
      }
      *p = 0;
      break;

    case 'X':
      /* XAA..AA,LLLL: Write LLLL escaped binary bytes at address AA.AA */
      is_binary = 1;
      /* fall through */

    case 'M':
      /* MAA..AA,LLLL:	Write LLLL bytes at address AA.AA return OK */
      if (!(p = scanHexInt (++p, &addr)))
	goto error;
      if (*p++ != ',')
	goto error;
      if (!(p = scanHexInt (p, &length)))
	goto error;
      if (*p++ != ':')
	goto error;
      if (! is_binary && dbg_strlen (p) != length * 2) {
	dbg_strcpy (iobuffer, "E22 bad set-memory data size");
	break;
      }
      if (length == 0)
	goto success;
      membase = (vaddr_t)addr;
      if (!_dbg_accessible (membase, length, 1))
	goto error;
      for (mem = membase, i = 0; i < length; mem++, i++) {
	  unsigned char v;
	  if (is_binary) {
	    v = *p++;
	    if (v == 0x7d)
	      v = 0x20 ^ *p++;
	  }
	  else {
	    if (!dbg_isxdigit (p[0]) || !dbg_isxdigit (p[1]))
	      goto error;
	    v = (hexchToInt (p[0]) << 4) | hexchToInt (p[1]);
	    p += 2;
	  }
	  if (_dbg_put_byte (mem, v))
	      goto fault;
	  if (_dbg_get_byte (mem, &err) != v || err)
	      goto fault;
      }
      if (length > 0 && _dbg_cacheable (membase))
	  SYNC_ICACHE (membase, length);
      goto success;

    case 's':
      /* sAA..AA	Step one instruction (from AA..AA, optional) */
      sstep = 1;
      goto contaddr;

    case 'S':
      /* Ssig;AA..AA	Step one instruction (from AA..AA, optional) */
      sstep = 1;
      /* FALL THROUGH */

    case 'C':
      /* Csig;AA..AA	Continue execution (from AA..AA, optional) */
      if (! (p = scanHexInt (++p, &nsig)))
	  goto error;
      if (nsig >= NSIG)
	  goto error;
      if (*p == '\0')
	  goto cont;
      if (*p++ != ';')
	  goto error;
      x->signo = (x->signo & ~0xff) | nsig;
      /* FALL THOUGH */

    case 'c':
      /* cAA..AA	Continue execution (from AA..AA, optional) */
    contaddr:
      if (scanHexInt (++p, &addr)) {
	  membase = (vaddr_t)addr;
	  /* FIXME This test is probably too restrictive */
	  if (IS_MIPS16 (membase)) {
	      if (!_dbg_accessible (MIPS16ADDR (membase), 2, 0))
		  goto error;
	  }
	  else {
	      if ((membase & 3) || !_dbg_accessible (membase, 4, 0))
		  goto error;
	  }
	  cx->epc = VADDR_TO_REG (membase);
      }
    cont:
      dbg_waiting = 1;
      restart = 1;
      break;

    case 'k':
      /* kSS		Send signal SS (default: SIGKILL & disconnect) */
      if (*++p == '\0')
	{
	  /* kill program (immediately) & disconnect */
	  DBG (1, "dbg: got kill\n");
	  nsig = SIGKILL;
	  restart = 1;
	}
      else 
	{
	  /* just record the signal, don't continue yet */
	  if (!scanHexInt (p, &nsig))
	    goto error;
	  DBG (1, "dbg: got signo = %d\n", nsig);
	  if (nsig >= NSIG)
	    goto error;
	  goto success;
	}
      x->signo = (x->signo & ~0xff) | nsig;
      break;

    case 'D':
	putpacket ("OK");
	/* drop through */

    case 'd':
      /* d		Remote debugger disconnecting: continue */
      dbg_regsize = MAX_REG_SIZE; /* reset on disconnect */
      restart = 1;
      break;

    case 'T':
	{
	    int thread;
	    if (!scanHexInt (++p, &thread))
		goto error;
	    if (_dbg_querythread (thread))
		goto success;
	    dbg_strcpy (iobuffer, "E99 Thread is dead\n");
	}
      break;

    case 'q':
        if (dbg_strncmp (p, "qCRC:", 5) == 0) {
	    /* qCRC:lma,size */
	    if ((p = scanHexInt (p+5, &addr))
		&& *p++ != ',' && (p = scanHexInt (p, &length))
		&& *p == '\0') 
	    {
		unsigned int crc;
		membase = (vaddr_t)addr;
		if (! _dbg_accessible (membase, length, 0))
		  goto fault;
		if (! crc32 (membase, length, &crc))
		  goto fault;
		p = iobuffer;
		*p++ = 'C';
		p = toHexInt (p, crc);
		*p++ = '\0';
	      }
	} 
	else if (dbg_strncmp (p, "qOffsets", 8) == 0
		 && _foffs_fdata != 0
		 && _foffs_fdata != _fdata) {
	    /* qOffsets  return string in form "Text=hexnum;Data=hexnum;Bss=hexnum"
	       sign extended to 64bits */
	    ptrdiff_t offs;

	    p = dbg_strcpy (iobuffer, "Text=");
	    offs = (ptrdiff_t) (_ftext - _foffs_ftext);
	    p = toHexInt (p, (int)(offs >> 31));
	    p = toHexInt (p, (int)offs);
            
	    p = dbg_strcpy (p, ";Data=");
	    offs = (ptrdiff_t) (_fdata - _foffs_fdata);
	    p = toHexInt (p, (int)(offs >> 31));
	    p = toHexInt (p, (int)offs);

	    p = dbg_strcpy (p, ";Bss=");
	    offs = (ptrdiff_t) (_fbss - _foffs_fbss);
	    p = toHexInt (p, (int)(offs >> 31));
	    p = toHexInt (p, (int)offs);
	    
	    *p++ = '\0';
	} 
	else {
	    /* Unsupported 'q' request */
	    iobuffer[0] = '\0';
	}
	break;

    case 'W':
    case 'w':
      /* [Ww]AA,LL,TT[,N]		insert/delete hardware watchoint */
      if (!p[1]) {
	  /* test for hit data watchpoint */
	  p = iobuffer;
	  if (watch_num > 0) {
	      if (cx == ox && 
		  (watch_hit & (MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W))) {
		  *p++ = '1';
		  if (watch_va != 0) {
		      *p++ = ',';
		      p = toHexInt (p, watch_va);
		      if (watch_len != 4) {
			  *p++ = ',';
			  p = toHexInt (p, watch_len);
		      }
		  }
	      }
	      else 
		  *p++ = '0';
	  }
	  *p++ = '\0';
      }
      else if (p[1] == '?') {
	  /* watchpoint capability request */
	  unsigned int caps = 0;
	  int i;
	  p = iobuffer;
	  p = toHexInt (p, watch_num);
	  for (i = 0; i < watch_num; i++) {
	      unsigned int cap = _mips_watchpoint_capabilities (i);
 	      *p++ = ',';
	      p = toHexInt (p, cap);
	      caps |= cap;
	  }
	  if (caps & MIPS_WATCHPOINT_MASK) {
	      for (i = 0; i < watch_num; i++) {
		  vaddr_t mask = _mips_watchpoint_mask (i);
		  *p++ = ',';
		  if (mask != ~(vaddr_t)0)
		      p = toHexInt (p, _mips_watchpoint_mask (i));
	      }
	  }
	  *p++ = '\0';
      }
      else {
	  static const unsigned types[4] = {
	      MIPS_WATCHPOINT_W, 
	      MIPS_WATCHPOINT_R, 
	      MIPS_WATCHPOINT_R|MIPS_WATCHPOINT_W, 
	      MIPS_WATCHPOINT_X
	  }; 
	      
	  /* watchpoint set/clear request */
	  int length, type;
	  int wpnum = 0;
	  int ok;

	  if ((p = scanHexInt (++p, &addr))
	      && *p++ == ','
	      && (p = scanHexInt (p, &length))
	      && *p++ == ','
	      && (p = scanHexInt (p, &type))
	      && (*p == '\0'
		  || (*p++ != ','
		      && (p = scanHexInt (p, &wpnum))))
	      && *p == '\0')
	      membase = (vaddr_t)addr;
	  else
	      goto error;

	  if (!_dbg_accessible (membase, 4, 0))
	      goto error;
	  if (type < 0 || type > 3 || length <= 0 || wpnum < 0)
	      goto error;
	  if (type == 3 && length == 1)
	      length = IS_MIPS16 (addr) ? 2 : 4;
	  if (req == 'W')
	      ok = _mips_watchpoint_set (types[type], -1, 
					 addr, _dbg_vtop (addr),
					 length);
	  else
	      ok = _mips_watchpoint_clear (types[type], -1, addr, length);
	  switch (ok) {
	  case MIPS_WP_OK:
	  case MIPS_WP_NOMATCH:	/* XXXX */
	      goto success;
	  case MIPS_WP_NONE:
	      dbg_strcpy (iobuffer, "E99 target does not support watchpoints");
	      break;
	  case MIPS_WP_NOTSUP:
	      dbg_strcpy (iobuffer, "E99 cannot handle this watchpoint");
	      break;
	  case MIPS_WP_INUSE:
	      dbg_strcpy (iobuffer, "E99 target out of watchpoint resources");
	      break;
	  case MIPS_WP_OVERLAP:
	      dbg_strcpy (iobuffer, "E99 watchpoint overlaps debugger data");
	      break;
	  case MIPS_WP_BADADDR:
	      dbg_strcpy (iobuffer, "E99 bad watchpoint address");
	      break;
	  default:
	      dbg_strcpy (iobuffer, "E99 internal watchpoint problem");
	      break;
	  }
      }
      break;

    fault:
      dbg_strcpy (iobuffer, "E14 bad memory address");
      break;

    error:
      dbg_strcpy (iobuffer, "E22 invalid arguments");
      break;
	  
    success:
      dbg_strcpy (iobuffer, "OK");
      break;
	  
    default:
      /* ignore unknown requests (gdb must cope with the consequences) */
      iobuffer[0] = '\0';
      break;
    }				/* switch */

    /* reply to the request */
    if (!restart) {
	DBG (3, "<rep: %.40s> ", iobuffer);
	putpacket (iobuffer);
    }
  }

  /*
   * Return from exception restoring new state;
   * apart from single-stepping it would be easy.
   */

  x = _dbg_contthread (cx);

  DBG (1, "restarting at %p (sig=%d)\n", REG_TO_VA (x->epc), x->signo & 0xff);

  /* Restore floating point registers, if they were changed */
  putfpactx ();

  if (threadcheck && dbg_waiting && !sstep)
      goto status;

  if (sstep) {
      if (ss_setup (x)) {
	  char *s;
	  s = dbg_strcpy (iobuffer, "E14 ");
	  dbg_strcpy (s, dbg_errstr);
	  DBG (3, "<rep: %.40s> ", iobuffer);
	  putpacket (iobuffer);
#ifndef CONT_ON_SSERR
	  goto status;
#endif
      }
  }

  was_sstep = sstep;
  
  if (_dbg_exithook ()) {
      dbg_strcpy (iobuffer, "E14 _dbg_exithook failed");
      DBG (3, "<rep: %.40s> ", iobuffer);
      putpacket (iobuffer);
      goto status;
  }
      
  if (!dbg_waiting)
      _dbg_detachhook ();

  /* prevent interrupts during a single-step */
  if (sstep && _dbg_threadsafe_sstep) {
      dbg_srsave = x->sr;
      x->sr &= ~SR_IE;
  }

  if (x->signo & 0x10000) {
      /* Returning to _dbg_sighook: cause new signal number to be
	 returned. */
      char * pc = REG_TO_VA (x->epc);
      if (pc >= debug_sighook_insn
	  && pc <= debug_sighook_insn + 4) {
	  x->regs[XCP_A0] = x->regs[XCP_V0] = x->signo & 0xff;
	  x->signo = 0;
      }
  }

  /* make sure all output messages are complete before restarting */
  _dbg_drain ();

  /* insert watchpoints if not single-stepping */
  if (!sstep)
      _mips_watchpoint_insert ();

  /* continue program, at specified signal handler */
  inhere--;
  return _dbg_restart (x, x->signo & 0xff);
}


/* write a text string to the host console while in debug mode */
void
_dbg_write (const char *s, size_t n)
{
    int o;
    iobuffer[0] = 'O';
    while (n != 0) {
	o = 1;
	while (n != 0 && o < BUFMAX - 32) {
	    char c = *s++;
	    if (c != '\r') {
#if _SDE_MAJOR_ >= 3
	      /* gdb 4.16+ encodes each character in hex */
	      iobuffer[o++] = intToHexch(c >> 4);
	      iobuffer[o++] = intToHexch(c);
#else
	      /* gdb 4.15 sends plain characters, but #, $ and * are used
		 by protocol, so have to be omitted! */
	      if (c != '$' && c != '#' && c != '*')
		iobuffer[o++] = c;
#endif
	    }
	    n--;
	}
	iobuffer[o++] = '\0';
	putpacket (iobuffer);
    }
}

void
_rdebug_load ()
{
}
