#ident "kit/share/env.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1997-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * env.c: getenv, setenv, etc
 */


#include <string.h>
#include <stdlib.h>
#include <sdethread.h>
#include "kit_impl.h"

extern char **environ;

static __sdethread_mutex_t env_mx = __SDETHREAD_MUTEX_INITIALIZER("env");
static int env_once;
static int env_alloced;

/*
 * Simply add a new name/value pair at the end of the environ
 * array.
 */

static int
_putenv (const char *name, const char *value)
{
    char **envp;
    char *entry;
    const char *s;
    int	cnt;
    int l_value;
    int l_name;

    /* count entries in environ */
    for (envp = environ, cnt = 0; *envp; ++envp, ++cnt)
	continue;
    
    if (env_alloced) {
	/* simply increase malloced space */
	envp = (char **) realloc (environ, (sizeof(char *) * (cnt + 2)));
	if (!envp)
	    return (-1);
    } 
    else {
	/* new malloced space, copy old non-malloced version */
	envp = (char **) malloc ((sizeof(char *) * (cnt + 2)));
	if (!envp)
	    return(-1);
	memcpy (envp, environ, cnt * sizeof(char *));
	env_alloced = 1;
    }

    /* find name length (no '=' allowed) */
    for (s = name; *s && *s != '='; s++)
	continue;
    l_name = s - name;
	
    /* find value length */
    l_value = strlen (value);
    
    /* name + '=' + value + '\0' */
    if (!(entry = malloc (l_name + l_value + 2)))
	return (-1);
    
    /* add new entry to environ */
    envp[cnt] = entry;
    envp[cnt + 1] = NULL;
    environ = envp;

    /* copy NAME=VAL into entry */
    memcpy (entry, name, l_name); 
    entry += l_name;
    *entry++ = '=';
    memcpy (entry, value, l_value + 1);
    
    return (0);
}


/*
 * Find a named entry and return a pointer to the value, and it's
 * index in the environ[] array.  
 */

static char *
_findenv (const char *name, int *offset)
{
    const char *s;
    char **envp;
    int l_name;
    
    if (env_once == 0) {
	if (! environ || !environ[0]) {
	    /* initialise from non-volatile memory */
	    const char *err = _sbd_envinit ();
	    if (err)
		_mon_printf ("\nNon-volatile environment: %s\n", err);
	    _sbd_mapenv (_putenv);
	}

	/* add board's monitor name to initial environment */
	for (envp = environ; *envp; envp++)
	    if (strncmp (*envp, "MONITOR=", sizeof "MONITOR=" - 1) == 0)
		break;
	if (! *envp) {
	    extern const char _mon_name[];
	    _putenv ("MONITOR", _mon_name);
	}

	env_once = 1;
    }

    if (!environ)
	return NULL;

    /* find name length */
    for (s = name; *s && *s != '='; s++)
	continue;
    l_name = s - name;

    for (envp = environ; *envp; envp++) {
	char *entry = *envp;
	if (strncmp (entry, name, l_name) == 0 && entry[l_name] == '=') {
	    if (offset)
		*offset = envp - environ;
	    return (&entry[l_name+1]);
	}
    }

    return (NULL);
}



/*
 * Here are the standard user-callable C library functions.
 */

int
setenv (const char *name, const char *value, int rewrite)
{
    char **envp;
    char *entry;
    int offset;
    int ret;

    __sdethread_mutex_lock (&env_mx);

    /* find if already exists */
    if ((entry = _findenv (name, &offset))) {
#if 0
	if (strcmp (entry, value) == 0) {
	    /* no change */
	    __sdethread_mutex_unlock (&env_mx);
	    return (0);
	}
#endif

	if (!(rewrite & 1)) {
	    /* rewrite not permitted */
	    __sdethread_mutex_unlock (&env_mx);
	    return (0);
	}

	if (strlen (entry) >= strlen (value)) { 
	    /* old value is longer; copy over */
	    strcpy (entry, value);
	    goto store;
	}

	/* we're going to replace it, so remove the old one */
	for (envp = &environ[offset]; envp[0] = envp[1]; ++envp)
	    continue;
    } 

    /* add name=value to environ[] */
    if ((ret = _putenv (name, value)) != 0) {
	__sdethread_mutex_unlock (&env_mx);
	return ret;
    }

store:
    /* optionally set name=value in NVRAM */
    if (rewrite & 2)
	ret = _sbd_setenv (name, value, rewrite & 1);
    __sdethread_mutex_unlock (&env_mx);
    return ret;
}


static void _unsetenv (const char *name, int rewrite)
{
    char **envp;
    int	offset;

    __sdethread_mutex_lock (&env_mx);

    /* delete all local copies */
    while (_findenv(name, &offset)) {
	for (envp = &environ[offset]; envp[0] = envp[1]; ++envp)
	    continue;
    }

    /* delete non-volatile copy */
    if (rewrite)
	_sbd_unsetenv (name);

    __sdethread_mutex_unlock (&env_mx);
}

/*
 * unsetenv(name) --
 *	Delete environmental variable "name".
 */
void unsetenv (const char *name)
{
    _unsetenv (name, 0);
}

/*
 * deleteenv(name) --
 *	Delete environmental variable "name", including
 *	removing from permanent storage
 */
#pragma weak deleteenv
void deleteenv (const char *name)
{
    _unsetenv (name, 1);
}

/*
 * getenv --
 *	Returns ptr to value associated with name, if any, else NULL.
 */
char *getenv (const char *name)
{
    char *val;

    /* just return local copy */
    __sdethread_mutex_lock (&env_mx);
    val = _findenv (name, NULL);
    __sdethread_mutex_unlock (&env_mx);
    return val;
}

