#ident "kit/share/fcache.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * share/fcache.c: Generic FLASH cache handling for debug support
 */


#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include <mips/cpu.h>
#include "kit_impl.h"
#include "flashdev.h"
#include "fcache.h"

#ifdef DBGDBG
#define DBG(fmt, args...) \
	_dbg_printf (fmt , ## args);
#else
#define DBG(fmt, args...)
#endif

/* LINESIZE must be:
 *	a power of 2 becuase the tag is generated using LINEMASK
 *	<= 32 because that is how many dirty bits there are
 */
#define LINESIZE	16
#define LINEMASK	(LINESIZE-1)

#define NCACHE		32

static struct fcache {
    unsigned int	tag;		/* tag for entry (offs & LINEMASK) */
    unsigned int	valid;		/* cache entry valid */
    unsigned int	dirty;		/* dirty bytes in this line */
    flashcookie_t	cookie;		/* flash device cookie */
    unsigned char	data[LINESIZE];
} fcache[NCACHE];


static unsigned char *	fcache_buf;
static unsigned int	fcache_bsize;

void
_fcache_init (void)
{
    memset (fcache, 0, sizeof (fcache));
}


int
_fcache_put_bytes (flashcookie_t cookie, unsigned int offs, 
		   const void *buf, unsigned int nb)
{
    const unsigned char *bp = buf;
    struct fcache *cp, *ep, *fp;
    unsigned int tag;
    unsigned int loffs;
    unsigned int n;
    int i;
    
    while (nb != 0) {
	tag = offs & ~LINEMASK;
	loffs = offs & LINEMASK;
	
	n = LINESIZE - loffs;
	if (n > nb)
	    n = nb;
      
	DBG ("cacheput tag=0x%x loffs=0x%x foffs=0x%x: ", tag, loffs, offs);
      
	/* is this line in the cache ? */
	ep = fp = 0;
	for (cp = fcache; cp < &fcache[NCACHE]; cp++) {
	    if (cp->valid && cp->cookie == cookie && cp->tag == tag)
		break;
	    if (cp->valid == 0)
		fp = cp;
	    else if (cp->dirty == 0)
		ep = cp;
	}

	if (cp >= &fcache[NCACHE]) {
	    /* didn't find the line so create a new one if we can */

	    if (!(cp = fp) && !(cp = ep)) {
		/* XXX should we flush a line and continue? */
		DBG ("cache full\n");
		return 1;
	    }

	    /* initialise the line */
	    DBG ("(init) ");	

	    if (flash_readbytes (cookie, tag, cp->data, LINESIZE, 
				 FLASHDEV_READ_CODE) != FLASHDEV_OK) {
		DBG ("read error\n");
		return 1;
	    }

	    cp->cookie =  cookie;
	    cp->tag = tag;
	    cp->dirty = 0;
	    cp->valid = ~0;
	}

	/* update bytes in line */
	for (i = n; i != 0; i--, loffs++, bp++) {
	    if (cp->data[loffs] != *bp) {
		cp->dirty |= (1 << loffs);
		cp->data[loffs] = *bp;
	    }
	}
	DBG (" dirty=0x%x\n", cp->dirty);

	offs += n;
	nb -= n;
    }

    return 0;
}


int
_fcache_get_bytes (flashcookie_t cookie, unsigned int offs, 
		   void *buf, unsigned int nb)
{
    unsigned char *bp = buf;
    struct fcache *cp;
    unsigned int tag, loffs, n;

    while (nb != 0) {
	tag = offs & ~LINEMASK;
	loffs = offs & LINEMASK;
	
	n = LINESIZE - loffs;
	if (n > nb)
	    n = nb;
      
	DBG ("cacheget tag=0x%x loffs=0x%x foffs=0x%x: ", tag, loffs, offs);

	/* is this line in the cache ? */
	for (cp = fcache; cp < &fcache[NCACHE]; cp++) {
	    if (cp->valid && cp->cookie == cookie && cp->tag == tag)
		break;
	}

	if (cp < &fcache[NCACHE]) {
	    /* line in the cache so return data from there */
	    memcpy (bp, &cp->data[loffs], n);
	    DBG ("hit\n");
	}
	else {
	    /* else read flash directly */
	    if (flash_readbytes (cookie, offs, bp, n, FLASHDEV_READ_CODE) 
		!= FLASHDEV_OK) {
		DBG ("read err\n");
		return 1;
	    }
	    DBG ("miss\n");
	}

	offs += n;
	bp += n;
	nb -= n;
  }

  return 0;
}


int
_fcache_flush (void)
{
    struct fcache *cp, *mcp;
    struct flashinfo info;
    int err = 0;

    /* search for line that needs to be written out */
    for (cp = fcache; cp < &fcache[NCACHE]; cp++) {
	if (cp->valid == 0 || cp->dirty == 0)
	    continue;

	/* get info about sector containing this line */
	flash_info (cp->cookie, cp->tag, &info);
	
	if (info.ssize > fcache_bsize) {
	    fcache_buf = realloc (fcache_buf, info.ssize);
	    fcache_bsize = info.ssize;
	}

	/* read sector into buffer */
	if (flash_readbytes (cp->cookie, info.soffs, fcache_buf, 
			     info.ssize, FLASHDEV_READ_CODE) != FLASHDEV_OK) {
	    err = 1;
	    continue;
	}
    
	/* merge all remaining cache lines in this sector into the buffer */
	for (mcp = cp; mcp < &fcache[NCACHE]; mcp++) {
	    if (mcp->valid && mcp->dirty && mcp->cookie == cp->cookie 
		&& info.soffs <= mcp->tag 
		&& mcp->tag < info.soffs + info.ssize) {
		unsigned int offs = mcp->tag - info.soffs;
		DBG ("Adding line for %p to sector buffer, offset=0x%x\n", 
		     cp->tag, offs);
		memcpy (&fcache_buf[offs], mcp->data, LINESIZE);
		mcp->dirty = 0;
	    }
	}

	DBG ("Programming sector at 0x%x from %p, size %u\n",
	     info.soffs, fcache_buf, info.ssize);
	if (flash_programbytes (cp->cookie, info.soffs, fcache_buf, 
				info.ssize, FLASHDEV_PROG_CODE) != FLASHDEV_OK)
	    err = 1;
    }

#ifdef DBGDBG
    /* Paranoia :-) */
    {
	int i;
	for (cp = fcache; cp < &fcache[NCACHE]; cp++) {
	    if (cp->valid && cp->dirty) {
		/* dirty line still in cache! */
		DBG ("fcacheflush: dirty line 0x%x: ", cp->tag);
		for (i = 0; i < LINESIZE; i++)
		    DBG ("%c%02x", (cp->dirty & (1 << i)) ? '>' : ' ', 
			 cp->data[i]);
		DBG ("\n");
		err = 1;
	    }
	}
    }
#endif

    return err;
}
