/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * share/file.h: POSIX file i/o data structures
 */


#ifndef _SYS_FILE_H
#define _SYS_FILE_H

#include <sys/cdefs.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <sys/unistd.h>

#include <sdethread.h>

/*
 * File descriptor table.
 * One entry for each open file or socket.
 */
struct file {
    	struct file 	*f_next;	/* free list pointer */
	short		f_flag;		/* see fcntl.h */
	unsigned char	f_type;		/* descriptor type */
#define DTYPE_FREE	0
#define DTYPE_DNODE	1
#define DTYPE_SOCKET	2
#define DTYPE_PIPE	3
	short		f_count;	/* open reference count */
	off_t		f_offset;	/* file offset */
	struct dirnode *f_dp;		/* directory entry/inode */
	const struct fileops *f_ops;	/* file i/o ops */
	void *		f_data;		/* private data */
};

typedef struct file file_t;

/* 
 * Directory node.
 * A cross between a *nix inode and directory entry.
 * It holds the file name and file status combined.
 */
struct dirnode {
    	const char *		d_name;		/* file name */
	__sdethread_mutex_t 	d_mx;		/* thread protection */
	short			d_count;	/* reference count */
    	dev_t			d_dev;		/* device number */
	mode_t			d_mode;		/* file mode (see stat.h) */
	off_t			d_size;		/* file size */
	const struct fileops *	d_ops;		/* file i/o ops */
    	void *			d_data;		/* private data */
	nlink_t			d_nlink;	/* link count */
        ino_t			d_ino;		/* file inode */
    	unsigned short		d_flags;	/* misc flags */
    	struct dirnode *	d_parent;	/* parent directory */
    	struct dirnode *	d_sibling;	/* sibling entry */
    	struct dirnode *	d_link;		/* hard link */
};

typedef struct dirnode dir_t;

/* d_flags bits */
#define DF_DYNAMIC		0x0001

/* for some backwards compatibility */
#define dirdesc dirnode

#define DIRDESC(name,dev,mode,size,ops,data) \
    {name, __SDETHREAD_MUTEX_INITIALIZER("dnode"),		\
     0, dev, mode, size, &ops, (void *)data, 1}


/*
 * The low-level file i/o operations.
 */
struct fileops {
    /* file i/o operations */
    int		(*fo_read)	(file_t *, struct uio *);
    int		(*fo_write)	(file_t *, struct uio *);
    int		(*fo_ioctl)	(file_t *, int, void *);
    int		(*fo_select)	(file_t *, int);
    int		(*fo_open)	(file_t *, dev_t, int);
    int		(*fo_close)	(file_t *);
    /* dirnode operations */
    int		(*fo_lookup) 	(dir_t **, dir_t *, const char **, int, mode_t);
    void	(*fo_dfree) 	(dir_t *);
    int		(*fo_init) 	(dir_t *);
    int		(*fo_mknod) 	(dir_t *, const dir_t *);
    int		(*fo_stat) 	(dir_t *, file_t *, struct stat *);
    int		(*fo_remove) 	(dir_t *, dir_t *);
    long	(*fo_pathconf) 	(dir_t *, int, long *);
};

/* allocate a new file i/o handle, and optional file descriptor */
int		_fileio_falloc (file_t **, int *);

/* allocate a file descriptor for this file handle */
int		_fileio_fdalloc (file_t *, int, int *);

/* delete a file descriptor */
void		_fileio_fdremove (int);

/* return the (locked) file handle for this file descriptor */
int 		_fileio_fget (int, file_t **);

/* release the lock on this file handle (and free if ref count == 0) */
void		_fileio_frelease (file_t *);

/* allocate a new file i/o handle, and optional file descriptor */
int		_fileio_lookup (dir_t **, const char *, int, mode_t);

/* extra flag bits for lookup */
#define FNODREF         0x80000000	/* don't dreference hard link */

/* special function to add a new "filename" to a directory */
extern int	_mknod (const char *, const dir_t *);

#define OPEN_MAX	64
#define NFILE 		(OPEN_MAX + 2)

/* enter blocked signal system call state */
#define _syscall_enter() 		\
   __sdethread_sigstate_t _save; 	\
   __sdethread_sigdisable (&_save);

/* leave blocked signal system call state */
#define _syscall_leave(err,retval) 		\
   (__sdethread_sigrestore (&_save), (err)	\
    ? ((errno = (err)), -1)		      	\
    : (retval))
		       

#endif /*_SYS_FILE_H*/
