#ident "kit/share/flashdev.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

#if #feature(flashdev)

/*
 * Copyright (c) 1999-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * flashdev.c: Posix style access to raw Flash ROM
 */


#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/init.h>
#include <stdlib.h>
#include <sdethread.h>
#include <assert.h>
#include <string.h>
#include "file.h"
#include "kit_impl.h"
#include "flashdev.h"

static int	flashrw (struct file *, struct uio *);
static int	flashioctl (struct file *, int, void *);
static int	flashselect (struct file *, int);
static int	flashopen (struct file *, dev_t, int);
static int	flashclose (struct file *);

static const struct fileops flash_ops = 
{
    flashrw, 
    flashrw, 
    flashioctl, 
    flashselect, 
    flashopen,
    flashclose,
};


struct flashd {
    flashcookie_t	cookie;
    struct flashpart	part;
    int			refcnt;
    unsigned int	flags;
};


#ifndef FLASHNDEV
#define FLASHNDEV	4
#endif

#define NFLASH		(FLASHNPART * FLASHNDEV)

#define fdev(dev)	((dev) / FLASHNPART)
#define fdevpart(dev)	((dev) % FLASHNPART)

static struct flashd *flashd[NFLASH];
static __sdethread_mutex_t flash_mx = __SDETHREAD_MUTEX_INITIALIZER("flash");

static int
flash_seterrno (int fret)
{
    switch (fret) {
    case FLASHDEV_OK:
	return 0;
    case FLASHDEV_FAIL:
	/* parameter error */
	return EINVAL;
    case FLASHDEV_PROTECTED:
	/* can't program protected sector */
	return EACCES;
    case FLASHDEV_PARTIAL:
	/* can't erase partial sector */
	return EACCES;
    case FLASHDEV_NOMEM:
	/* no memory for buffer */
	return ENOMEM;
    case FLASHDEV_FATAL:
    default:
	/* fatal programming error */
	return EIO;
    }
}


static int
flashrw (struct file *fp, struct uio *uio)
{
    struct flashd *flash = (struct flashd *)fp->f_data;
    int ret = FLASHDEV_OK;
    size_t nb;
    int err = 0;
    
    if (uio->uio_offset >= flash->part.size)
	return ENXIO;		/* ??? */

    /* bounds checking */
    if (uio->uio_offset + uio->uio_resid > flash->part.size) {
	if (uio->uio_rw == UIO_WRITE)
	    return ENXIO;	/* ??? */
	nb = flash->part.size - uio->uio_offset;
    }
    else
	nb = uio->uio_resid;

    __sdethread_mutex_lock (&flash_mx);
    while (nb != 0 && err == 0) {
	struct iovec *iov;
	size_t len;

	err = uiogetiov (uio, &iov);
	if (err != 0)
	    break;

	len = iov->iov_len;
	if (len > nb)
	    len = nb;

	if (uio->uio_rw == UIO_READ)
	    ret = flash_readbytes (flash->cookie, 
				   flash->part.offs + uio->uio_offset, 
				   iov->iov_base, len, flash->flags);
	else
	    ret = flash_programbytes (flash->cookie, 
				      flash->part.offs + uio->uio_offset, 
				      iov->iov_base, len, flash->flags);

	if (ret == FLASHDEV_OK)
	    err = uioadvance (len, uio);
	else
	    err = flash_seterrno (ret);

	nb -= len;
    }
    __sdethread_mutex_unlock (&flash_mx);
    return 0;
}

static int
flashioctl (struct file *fp, int cmd, void *arg)
{
    struct flashd *flash = (struct flashd *)fp->f_data;
    int ret = FLASHDEV_OK;

    __sdethread_mutex_lock (&flash_mx);
    switch (cmd) {
    case FLASHIOINFO: {
	struct flashinfo *info = arg;
	ret = flash_info (flash->cookie, info->soffs, info);
	break;
    }

    case FLASHIOGPART:
	memcpy (arg, &flash->part, sizeof (flash->part));
	break;

    case FLASHIOGPARTS:
	ret = flash_parts (flash->cookie, (struct flashparts *)arg);
	break;

    case FLASHIOSPARTS:
	ret = flash_setparts (flash->cookie, (const struct flashparts *)arg);
	break;

    case FLASHIOGFLGS:
	*(unsigned int *)arg = flash->flags;
	break;

    case FLASHIOSFLGS:
	flash->flags = *(unsigned int *)arg;
	break;

    case FLASHIOERASEDEV: {
	struct flashinfo info;
	unsigned int offs;

	if (flash->part.type == FLASHPART_RAW) {
	    /* try quick erase of whole device */
	    ret = flash_erasedevice (flash->cookie);
	    if (ret == FLASHDEV_OK || ret != FLASHDEV_FAIL)
		break;
	}

	/* erase partition sector by sector */
	for (offs = flash->part.offs; offs < flash->part.size; 
	     offs += info.ssize) {
	    flash_info (flash->cookie, offs, &info);
	    ret = flash_erasesector (flash->cookie, offs);
	    if (ret != FLASHDEV_OK)
		break;
	}
	break;
    }

    case FLASHIOERASESECT:
	ret = flash_erasesector (flash->cookie, *(unsigned int *)arg);
	break;

    case FLASHIOFLUSH:
	ret = flash_flush (flash->cookie);
	break;

    default:
	__sdethread_mutex_unlock (&flash_mx);
	return EINVAL;
    }
    __sdethread_mutex_unlock (&flash_mx);
    return flash_seterrno (ret);
}


static int
flashselect (struct file *fp, int rw)
{
    /* always ready! */
    return rw;
}


static int
flashopen (struct file *fp, dev_t dev, int fflags)
{
    struct flashd *flash;

    if (dev >= NFLASH)
	return ENXIO;

    /* allocate our structure */
    __sdethread_mutex_lock (&flash_mx);
    if (!(flash = flashd[dev])) {
	flash = flashd[dev] = malloc (sizeof (*flash));
	if (!flash)
	    return ENOMEM;
	memset (flash, 0, sizeof (*flash));
    }

    /* initialise on first open */
    if (flash->refcnt == 0) {
	struct flashparts parts;

	flash->cookie = flash_open (fdev (dev));
	if (flash->cookie == FLASH_UNKNOWN) {
	    __sdethread_mutex_unlock (&flash_mx);
	    return ENXIO;
	}
	if (flash->cookie == FLASH_MISSING) {
	    __sdethread_mutex_unlock (&flash_mx);
	    return ENODEV;
    	}

	/* get our partition information */
	flash_parts (flash->cookie, &parts);
	flash->part = parts.part[fdevpart(dev)];

	/* set default programming mode (byte stream) */
	flash->flags = FLASHDEV_PROG_MERGE | FLASHDEV_PROG_STREAM;
    }

    /* point file at our private data */
    flash->refcnt++;
    fp->f_data = flash;
    __sdethread_mutex_unlock (&flash_mx);
    return 0;
}


static int
flashclose (struct file *fp)
{
    struct flashd *flash = (struct flashd *)fp->f_data;

    __sdethread_mutex_lock (&flash_mx);
    flash_flush (flash->cookie);
    if (--flash->refcnt == 0) {
	/* final close */
    }
    __sdethread_mutex_unlock (&flash_mx);
    return 0;
}



/*
 * Dynamically create /dev entries for all flash devices
 */

/* /dev/flash*:	crw-rw-r-- */
#define FLASHMODE	(S_IFBLK|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)

/* call this function to load this module */
static void
flashdev_mount (void)
{
    static int inited = 0;
    dir_t flashnode;
    char fname[8] = "flashnp\0";
    const char partc[] = "\0btefghi";
    flashcookie_t cookie;
    struct flashparts parts;
    int d, p;

    if (inited)
	return;
    inited = 1;

    memset (&flashnode, 0, sizeof (flashnode));
    flashnode.d_name = fname;
    flashnode.d_mode = FLASHMODE;
    flashnode.d_ops = &flash_ops;

    for (d = 0; d < FLASHNDEV; d++) {

	/* check that flash is known and present */
	cookie = flash_open (d);
	if (cookie == FLASH_UNKNOWN)
	    break;
	if (cookie == FLASH_MISSING)
	    continue;
	fname[5] = d + '0';

	/* scan each part and add non-empty ones as a device */
	flash_parts (cookie, &parts);
	for (p = 0; p < FLASHNPART; p++) {
	    if (parts.part[p].size != 0) {
		flashnode.d_dev = d * FLASHNPART + p;
		flashnode.d_size = parts.part[p].size;
		fname[6] = partc[p];
		if (_mknod ("/dev", &flashnode)) {
		    _mon_printf ("mknod(/dev/%s) failed\n", fname);
		    abort ();
		}
	    }
	}
    }
}

DECL_INIT_FUNC(IF_MOUNT, flashdev_mount)


/* call this dummy function to load this module */
void
_flashdev_load (void)
{
}

#endif /* #feature(flashdev) */
