/*
 * Copyright (c) 1996-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * share/flashdev.h: generic FLASH support definitions
 */


#ifndef _flashdev_H_
#define _flashdev_H_

#ifdef __cplusplus
extern "C" {
#endif

/* include ioctl based interface for common structures */
#include <sys/flashio.h>

/*
 * A cookie passed to the device routines that may be used
 * to control system/device specific information (eg write protection)
 * Note: private to flash routines, caller should treat as opaque.
 */
struct flashcookie {
    const struct flashdev *	dev;		/* dev access functions */
    const struct flashboard *	board;		/* board access functions */
};

/* this is what the caller should use */
typedef const struct flashcookie *flashcookie_t;

/* this function returns the cookie for the flash at the specified 
   physical address, or index (0..n) */
extern flashcookie_t	_sbd_flashopen (paddr_t);

/* call this for POSIX /dev/flash device interface */
extern void		_flashdev_load (void);

/* missing: valid address or index, but no device found */
#define FLASH_MISSING	(flashcookie_t)0

/* unknown: unknown address or device index */
#define FLASH_UNKNOWN	(flashcookie_t)-1

/* magic number preceding partition table if stored in flash */
#define FLASHPART_MAGIC		0x77049081

/* define a partition structure dynamically */
#define _flashdev_setpart(parts, idx, t, o, s) \
  do { \
    struct flashpart *_p = &(parts)->part[idx]; \
    _p->type = t; _p->offs = o; _p->size = s; \
  } while (0)
    
/*
 * Vectors to device specific routines
 */
struct flashdev {
    int	(*devinfo) (flashcookie_t, unsigned int, struct flashinfo *);
    int	(*deverasedevice) (flashcookie_t);
    int	(*deverasesector) (flashcookie_t, unsigned int);
    int	(*devprogrambytes) (flashcookie_t, unsigned int, 
			    const void *, unsigned int, int);
    int	(*devreadbytes) (flashcookie_t, unsigned int, 
			 void *, unsigned int, unsigned int);
};


/*
 * Vectors to board specific routines
 */
struct flashboard {
    int (*getparts) (flashcookie_t, struct flashparts *);
    int (*setparts) (flashcookie_t, const struct flashparts *);
};

/* cover macros */
#define flash_open(paddr) \
    _sbd_flashopen (paddr)

#define flash_info(cookie, offs, infop) \
    (*(cookie)->dev->devinfo) (cookie, offs, infop)

#define flash_erasedevice(cookie) \
    (*(cookie)->dev->deverasedevice) (cookie)

#define flash_erasesector(cookie, soffs) \
    (*(cookie)->dev->deverasesector) (cookie, soffs)

#define flash_programbytes(cookie, offs, buf, len, flags) \
    (*(cookie)->dev->devprogrambytes) (cookie, offs, buf, len, flags)

#define flash_flush(cookie) \
    (*(cookie)->dev->devprogrambytes) (cookie, 0, 0, 0, 0)

#define flash_readbytes(cookie, offs, buf, len, flags) \
    (*(cookie)->dev->devreadbytes) (cookie, offs, buf, len, flags)

#define flash_parts(cookie, parts) \
    (*(cookie)->board->getparts) (cookie, parts)

#define flash_setparts(cookie, parts) \
    (*(cookie)->board->setparts) (cookie, parts)

/* status returned by above functions */
#define FLASHDEV_OK		 0	/* success */
#define FLASHDEV_FAIL		-1	/* parameter error */
#define FLASHDEV_PROTECTED	-2	/* can't program protected sector */
#define FLASHDEV_PARTIAL	-3	/* can't erase partial sector */
#define FLASHDEV_NOMEM		-4	/* no memory for buffer */
#define FLASHDEV_FATAL		-5	/* fatal programming error */

/* flag bits for flash_programbytes() */
#define FLASHDEV_PROG_REBOOT	FLASHFLGS_REBOOT
#define FLASHDEV_PROG_NOCOPY	FLASHFLGS_NOCOPY
#define FLASHDEV_PROG_MERGE	FLASHFLGS_MERGE
#define FLASHDEV_PROG_CODE_EB	FLASHFLGS_CODE_EB
#define FLASHDEV_PROG_CODE_EL	FLASHFLGS_CODE_EL
#define FLASHDEV_PROG_CODE	FLASHFLGS_CODE
#define FLASHDEV_PROG_STREAM	FLASHFLGS_STREAM

/* flag bits for flash_readbytes() */
#define FLASHDEV_READ_CODE_EB	FLASHFLGS_CODE_EB
#define FLASHDEV_READ_CODE_EL	FLASHFLGS_CODE_EL
#define FLASHDEV_READ_CODE	FLASHFLGS_CODE
#define FLASHDEV_READ_STREAM	FLASHFLGS_STREAM

/* internal flag bits */
#define _FLASHDEV_SECTOR	0x4000	/* sector access (internal flag) */
#define _FLASHDEV_CMD		0x8000	/* command access (internal flag) */

#define FLASHDEV_MODE \
	(_FLASHDEV_SECTOR | _FLASHDEV_CMD | \
	 FLASHFLGS_STREAM | FLASHFLGS_CODE_EB | FLASHFLGS_CODE_EL)

#ifdef __cplusplus
}
#endif

#endif /* _flashdev_H_ */
