#ident "kit/share/intrupt.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * intrupt.c: generic MIPS low-level interrupt handler 
 */


/*
 * This file implements generic MIPS interrupt handling according to
 * a simple (but common) model, where each interrupt input is assigned
 * an interrupt priority level (ipl).  The intraction() function allows
 * the ipl for each input to be defined, whilst the simpler intrupt()
 * function sets up a default mapping as follows:
 *
 *	    INPUT	     CAUSE REG		IPL
 *	h/w interrupt 5		IP7		8 <-HIGHEST
 *	h/w interrupt 4		IP6		7 
 *	h/w interrupt 3		IP5		6 
 *	h/w interrupt 2		IP4		5 
 *	h/w interrupt 1		IP3		4 
 *	h/w interrupt 0		IP2		3 
 *	s/w interrupt 1		IP1		2 
 *	s/w interrupt 0		IP0		1 
 *	      -      		 -		0 <-LOWEST
 *
 * The CPU can be set to any ipl between 0 and 8 (inclusive): being
 * at a level implies that all interrupts of that level and below
 * are masked out, and all above are enabled.  Therefore ipl 0 means
 * that all interrupts are enabled, and ipl 8 means all are disabled.
 *
 * When an interrupt handler is called, the CPU is automatically set
 * to its corresponding ipl for the duration of that call (i.e. allowing 
 * nested interrupts of a higher priority).
 * 
 * Some hardware configurations may require second level interrupt 
 * handlers that examine external interrupt request registers, e.g. if 
 * several external interrupts are multiplexed onto one cpu interrupt 
 * line.  In this case interrupts should be installed and caught via 
 * a set of cover functions which arrange to catch the interrupts first 
 * and do any further vectoring that is required. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <sys/errno.h>
#include <sys/signal.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include "kit_impl.h"

#define intrblock()	_mips_intdisable ()
#define intrsetmask(sr)	_mips_intrestore (sr)

static int intrhandler (int, struct xcptcontext *);
static int intr_default (int, struct xcptcontext *);

/*
 * One entry per CPU interrupt input (0-7)
 */
static struct intraction intrtab[8] = {
    {intr_default, 0}, {intr_default, 1}, {intr_default, 2}, {intr_default, 3},
    {intr_default, 4}, {intr_default, 5}, {intr_default, 6}, {intr_default, 7}
};

/*
 * One entry per priority level (0-8)
 */
static struct ipl {
    unsigned short	mask;	/* i/u mask */
    unsigned short	bits;	/* i/u bits */
} ipltab[9];

/* Bit mask of all interrupts that we know about. */
unsigned int	_intrmask;

/* Base value of status register on startup */
unsigned int	_sr_base __attribute__((common));

/* Table indexed by cause register to give highest prio intrupt. */
static signed char	priotab[256];

static struct xcptaction oldxact;



/*
 * intraction(): installs interrupt handler and (optionally) returns
 * old handler.  The handler is called with new->ia_arg as its 
 * first argument, and the xception frame as its second.
 */
int intraction (unsigned int intrno, const struct intraction *new,
		struct intraction *old)
{
    static int xcpt_set = 0;
    struct intraction *ia;
    struct ipl *ip, *oip;
    unsigned int ibit, sr;
    int i;

    if (intrno > 7 || new->ia_ipl > 8)
      return EINVAL;

    ia = &intrtab[intrno];
    if (ia->ia_ipl && new->ia_ipl && new->ia_ipl != ia->ia_ipl)
      return EINVAL;

    /* block all interrupts */
    sr = intrblock ();

    if (!xcpt_set) {
	/* first time through: set initial ipl masks
	   to the what we inherited, except ipl 8, which
	   we leave to mean block all interrupts. */
	for (oip = ipltab; oip < &ipltab[8]; oip++)
	    oip->mask = _sr_base & SR_IMASK;
    }

    ibit = CR_IP0 << intrno;
    if (new->ia_ipl) {
	/* put interrupt at new ipl */
	ip = &ipltab[new->ia_ipl];
	ip->bits |= ibit;

	/* enable interrupt at all lower ipls; disable at self + higher */
	for (oip = ipltab; oip < &ipltab[8]; oip++) {
	    if (oip < ip)
		oip->mask |= ibit;
	    else
		oip->mask &= ~ibit;
	}

	/* don't actually enable yet */
	/*sr |= ibit;*/		
    } else {
	/* remove interrupt from old ipl */
	unsigned int obit = _sr_base & ibit;

	ip = &ipltab[ia->ia_ipl];
	ip->bits &= ~ibit;

	/* reset interrupt masks to initial state */
	for (oip = ipltab; oip < &ipltab[8]; oip++)
	    oip->mask = (oip->mask & ~ibit) | obit;
	
	/* also reset C0_STATUS to its initial state */
	sr = (sr & ~ibit) | obit;
    }

    /* store new interrupt handler in table */
    if (old) {
      *old = *ia;
      if (ia->ia_handler == intr_default)
	old->ia_handler = XCPT_DFL;
    }
    *ia = *new;

    /* recompute intrmask & priotab */
    _intrmask = 0;
    memset (priotab, -1, sizeof(priotab));

    /* work down from prio 8 to 1 */
    for (ip = &ipltab[8]; ip >= &ipltab[1]; ip--) {

	ibit = ip->bits;
	if (ibit == 0)
	    continue;

	_intrmask |= ibit;	/* record all known intrupts */
	ibit >>= 8;

	/* For each possible set of interrupt inputs, see if this 
	   ipl has a matching intrupt and is the highest priority. */
	for (i = 0; i < 256; i++) {
	    unsigned short bits = ibit & i;
	    int intr;

	    if (bits == 0 || priotab[i] != -1)
		/* no matching bits, or already grabbed by higher prio */
		continue;

	    /* find the highest prio intrupt at this ipl */
	    for (intr = 7; intr >= 0; intr--)
		if (bits & (1 << intr))
		    break;
	    assert (intr >= 0);

	    /* and record it in the table */
	    priotab[i] = intr;
	}
    }

    if (_intrmask == 0) {
	/* no interrupts left: remove handler */
	(void) xcptaction (XCPTINTR, &oldxact, NULL);
	xcpt_set = 0;
    } else {
	/* install low-level xcption handler */
	if (!xcpt_set) {
	    struct xcptaction xact;
	    xact.xa_handler = intrhandler;
	    xact.xa_flags = 0;
	    (void) xcptaction (XCPTINTR, &xact, &oldxact);
	    xcpt_set = 1;
	}
    }

    /* restore the i/u mask and forcibly enable global i/us */
    intrsetmask (sr | SR_IE);

    return 0;
}


/*
 * intrupt(): simpler way to install interrupt handler; returns old handler.
 */
xcpt_t intrupt (unsigned int intrno, xcpt_t handler, int arg)
{
    struct intraction new, old;
    int err;

    if (handler == XCPT_DFL) {
	/* remove handler */
	new.ia_handler = intr_default;
	new.ia_arg = intrno;
	new.ia_ipl = 0;
    } else {
	/* add handler */
	new.ia_handler = handler;
	new.ia_arg = arg;
	new.ia_ipl = intrno + 1;
    }

    if (err = intraction (intrno, &new, &old)) {
	extern int errno;		/* XXX shouldn't set errno */
	errno = err;
	return XCPT_ERR;
    }

    return old.ia_handler;
}



/* 
 * *NIX style spl (set priority level) functions.
 * Sets cpu ipl from 0 (all ints enabled), to 8 (all ints disabled)
 *
 * The modus operandi is usually:
 *	{
 *		int s = spln(5);	// block out level 5 ints and below
 *		// critical region here
 *		(void) splx(s); 	// return to previous level
 *	}
 */


/*
 * splx(): put a new interrupt mask in status register, return old value
 */
inline unsigned int splx (unsigned int imask)
{
    register unsigned int sr, osr;
    osr = sr = mips_getsr ();
    sr ^= (sr ^ imask) & SR_IMASK;    /* quick bit insert */ 
    mips_setsr (sr);
    return osr;
}


/* 
 * spln(): set cpu interrupt priority to numeric ipl (0 to 8) 
 */
unsigned int spln (unsigned int ipl)
{
    return splx (ipl >= 8 ? 0 : ipltab[ipl].mask);
}


/*
 * siron(): switch on software interrupt (0 or 1)
 */
void siron (unsigned int intrno)
{
    unsigned int sr;

    sr = intrblock ();
    mips_biscr (CR_SINT0 << (intrno & 1));
    intrsetmask (sr);
}


/*
 * siroff(): switch off software interrupt (0 or 1)
 */
void siroff (unsigned int intrno)
{
    unsigned int sr;

    sr = intrblock ();
    mips_biccr (CR_SINT0 << (intrno & 1));
    intrsetmask (sr);
}


int
intrpending (unsigned int intrno)
{
    if (intrno > 7)
      return 0;
    return (mips_getcr() & (CR_IP0 << intrno)) != 0;
}



/*
 * Low level interrupt exception handler.
 */
static int intrhandler (int xcptno, struct xcptcontext *xcp)
{
    unsigned int cr = XCP_CAUSE (xcp) & XCP_STATUS (xcp) & CR_IMASK;
    struct intraction *ia;
    int sig = 0;
    int intr;

    while (cr != 0) {

	if ((cr & _intrmask) == 0) {
	    _mon_printf ("\nUnregistered interrupt: epc=%p, sr=%x, cr=%x\n",
			 REG_TO_VA (xcp->epc), XCP_STATUS (xcp),
			 XCP_CAUSE (xcp));
	    return SIGKILL;
	}

	/* find highest priority interrupt bit number */
	intr = priotab[(cr & _intrmask) >> 8];
	ia = &intrtab[intr];

	/* call its handler at its ipl */
	splx (ipltab[ia->ia_ipl].mask);
	sig = (ia->ia_handler) (ia->ia_arg, xcp);
	splx (0);

	/* check for a signal request */
	if (sig)
	    return sig;

	/* fetch new cause register */
	cr = mips_getcr () & XCP_STATUS (xcp) & CR_IMASK;
    }

    return 0;
}


/*
 * Default interrupt handler (shouldn't get called)
 */
static int
intr_default (int arg, struct xcptcontext *xcp)
{
    if (oldxact.xa_handler != XCPT_DFL)
	return (*oldxact.xa_handler) (XCPTINTR, xcp);

    _mon_printf ("\nUnexpected interrupt %d: epc=%p, sr=%x, cr=%x\n", arg,
		 REG_TO_VA (xcp->epc), XCP_STATUS (xcp), XCP_CAUSE (xcp));
    return SIGKILL;
}


