#ident "kit/share/lr33watch.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * lr33watch.c: LR33000 family watchpoint support for MIPS SDE kit
 */


#if #cpu(lr33k)
#include <sys/types.h>
#include <lsi/lr33000.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <mips/prid.h>
#include <mips/watchpoint.h>
#include <string.h>

#include "kit_impl.h"

static int	lr33k_wp_init (unsigned int);
static void	lr33k_wp_setup (struct mips_watchpoint *);
static int 	lr33k_wp_hit (const struct xcptcontext *, vaddr_t *, size_t *);
static int	lr33k_wp_set (struct mips_watchpoint *);
static int	lr33k_wp_clear (struct mips_watchpoint *);
static void	lr33k_wp_remove (void);
static void	lr33k_wp_insert (void);
static void	lr33k_wp_reset (void);

const struct mips_watchpoint_funcs _lr33k_watchpoint_funcs = {
    lr33k_wp_init,
    lr33k_wp_setup,
    lr33k_wp_hit,
    lr33k_wp_set,
    lr33k_wp_clear,
    lr33k_wp_remove,
    lr33k_wp_insert,
    lr33k_wp_reset
};


struct lr33k_wp {
    unsigned long	reg;
    unsigned long	dcic;
};


#define IWATCH	0
#define DWATCH	1
static struct lr33k_wp	lr33k_wp[2];
static unsigned int	lr33k_dcic;


static int	
lr33k_wp_init (unsigned int prid)
{
    switch ((prid >> 8) & 0xff) {
    case PRID_LR33K:
	/* lr33k iwatch/dwatch supported */
	return 2;
    default:
	/* unrecognised */
	return -1;
    }
}


static void
lr33k_wp_setup (struct mips_watchpoint *wp)
{
    if (wp->num == IWATCH)
	/* Iwatch */
	wp->capabilities = MIPS_WATCHPOINT_VADDR | MIPS_WATCHPOINT_WORD
	    | MIPS_WATCHPOINT_X;
    else
	/* Dwatch */
	wp->capabilities = MIPS_WATCHPOINT_VADDR | MIPS_WATCHPOINT_WORD 
	    | MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W;
    wp->private = &lr33k_wp[wp->num];
}


static int
lr33k_wp_hit (const struct xcptcontext *xcp, vaddr_t *pva, size_t *plen)
{
    int hit = 0;

    if (XCP_CODE (xcp) == XCPTBP && (lr33k_dcic & (DCIC_DA | DCIC_PC))) {
	hit = (lr33k_dcic & DCIC_PC) ? MIPS_WATCHPOINT_X 
	    : (MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W);
	hit = _mips_watchpoint_address (xcp, hit, pva, plen);
    }					 

    return hit;
}


static int
lr33k_wp_set (struct mips_watchpoint *wp)
{
    struct lr33k_wp *pwp = wp->private;

    if (((wp->va & 3) + wp->len - 1) > 3)
	return 0;
    wp->mask = 3;

    pwp->reg = wp->va & ~3;
    if (wp->type & MIPS_WATCHPOINT_X)
	pwp->dcic |= DCIC_PCE;
    if (wp->type & MIPS_WATCHPOINT_R)
	pwp->dcic |= DCIC_DR;
    if (wp->type & MIPS_WATCHPOINT_W)
	pwp->dcic |= DCIC_DW;

    return 1;
}


static int
lr33k_wp_clear (struct mips_watchpoint *wp)
{
    struct lr33k_wp *pwp = wp->private;
    pwp->reg = pwp->dcic = 0;
    return 1;
}


static void	
lr33k_wp_remove (void)
{
    lr33k_dcic = lr33k_xchdcic (0);
}


static void
lr33k_wp_insert (void)
{
    unsigned int dcic = 0;

    if (lr33k_wp[IWATCH].reg) {
	lr33k_setbpc (lr33k_wp[IWATCH].reg);
	dcic |= lr33k_wp[IWATCH].dcic;
    }	

    if (lr33k_wp[DWATCH].reg) {
	lr33k_setbda (lr33k_wp[DWATCH].reg);
	dcic |= lr33k_wp[DWATCH].dcic;
    }

    if (dcic != 0)
	lr33k_setdcic (dcic);
}


static void
lr33k_wp_reset (void)
{
    memset (lr33k_wp, 0, sizeof (lr33k_wp));
    lr33k_setdcic (lr33k_dcic = 0);
}
#endif
