#ident "kit/share/m16entry.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1998-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * m16entry.c: a VERY SLOW implementation of MIPS16 entry/exit insns
 */


/*
 * This needs to be rewritten as a fast, low-level assembler exception
 * handler, but at least serves as an example.
 *
 * Note: this module had better not be compiled with -mentry or
 * we'll be into infinite recursion here!  
 */

#include <stddef.h>
#include <mips/cpu.h>
#include <mips/fpa.h>
#include <mips/m16c1.h>
#include <mips/xcpt.h>
#include <sys/signal.h>
#include "kit_impl.h"

#if (__mips >= 3 && __mips != 32) || __mips64
#define mips_put_reg mips_put_dword
#define mips_get_reg mips_get_dword
#else
#define mips_put_reg mips_put_word
#define mips_get_reg mips_get_word
#endif

/* Link to next handler */

static struct xcptaction chainxa;

/* Note this code relies on the low level xcption handler leaving a
   32-byte hole between the bottom of the user stack and the top of
   the xcptcontext frame! */

static int
mips16_entry_handler (int xcptno, struct xcptcontext *xcp)
{
    int err = 0;

    if (xcp->epc & 1) {
	unsigned short insn;

	insn = mips_get_half (REG_TO_VA (xcp->epc & ~1), &err);
	if (err)
	    return SIGBUS;

	if ((insn & 0xf81f) == 0xe809 && (insn & 0x0c0) != 0x0c0) {
	    unsigned int aregs, sregs, rreg;

	    /* The entry/exit instruction should never be in a 
	       branch delay slot. */
	    if (XCP_CAUSE (xcp) & CR_BD)
		return SIGILL;

	    aregs = (insn & 0x700) >> 8;
	    sregs = (insn & 0x0c0) >> 6;
	    rreg =  (insn & 0x020);

	    /* stack must be dword aligned */
	    if (xcp->regs[XCP_SP] & 7)
		return SIGBUS;

	    if (aregs < 5) {
		/* entry pseudo-instruction */
		void *sp = REG_TO_VA (xcp->regs[XCP_SP]);
		int i;

		/* store up to four arg registers */
		for (i = 0; i < aregs; i++) {
#if __mips64 && ! (_MIPS_SIM==_ABIO32 || defined(__mips_algabi))
		    err |= mips_put_dword (sp + i * 8, xcp->regs[XCP_A0 + i]);
#else
		    err |= mips_put_word (sp + i * 4, xcp->regs[XCP_A0 + i]);
#endif
		}

		/* save return address */
		if (rreg) {
		    sp -= sizeof(reg_t);
		    err |= mips_put_reg (sp, xcp->regs[XCP_RA]);
		}

		/* save s0 and s1 */
		if (sregs != 0) {
		    sp -= sizeof(reg_t);
		    err |= mips_put_reg (sp, xcp->regs[XCP_S0]);
		    if (sregs == 2) {
			sp -= sizeof(reg_t);
			err |= mips_put_reg (sp, xcp->regs[XCP_S1]);
		    }
		}

		/* decrement user stack pointer and skip entry insn */
		if (err == 0) {
		    xcp->regs[XCP_SP] -= 32;
		    xcp->epc += 2;
		}
	    }
	    else {
		/* exit pseudo-instruction */
		void *sp = REG_TO_VA (xcp->regs[XCP_SP] + 32);

		/* restore return address */
		if (rreg) {
		    sp -= sizeof (reg_t);
		    xcp->regs[XCP_RA] = mips_get_reg (sp, &err);
		}

		/* restore s0 and s1 */
		if (sregs != 0) {
		    sp -= sizeof (reg_t);
		    xcp->regs[XCP_S0] = mips_get_reg (sp, &err);
		    if (sregs == 2) {
			sp -= sizeof (reg_t);
			xcp->regs[XCP_S1] = mips_get_reg (sp, &err);
		    }
		}

		/* floating-point return value (see mips16.sx) */
#if #fpu(no)
		if (aregs != 7)
		    return SIGILL;
#else
		if (aregs == 5) {
		    /* copy single-precision value in $a0 into $f0 */
		    unsigned int val = xcp->regs[XCP_A0];
		    _mips16_lwc1 (&val);
		}
		else if (aregs == 6) {
		    /* copy double-precision value in $a0/a1 into $f0/f1 */
		    _mips16_ldc1 (&xcp->regs[XCP_A0]);
		}
#endif

		/* pop stack frame and return to caller */
		if (err == 0) {
		    xcp->epc = xcp->regs[XCP_RA];
		    xcp->regs[XCP_SP] += 32;
		}
	    }

	    /* all done! */
	    return err ? SIGBUS : 0;
	}
    }

    /* not handled, let next handler have a go */
    if (chainxa.xa_handler != XCPT_DFL)
	return (*chainxa.xa_handler) (xcptno, xcp);
    return SIGILL;
}


void
_mips16_entry_init (void)
{
    static int inited = 0;
    if (!inited) {
	struct xcptaction xa;
	xa.xa_handler = mips16_entry_handler;
	xa.xa_flags = 0;
	xcptaction (XCPTRI, &xa, &chainxa);
	inited = 1;
    }
}
