#ident "kit/share/m32watch.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * m32watch.c: MIPS32 family watchpoint support for MIPS SDE kit
 */


#if #cache(m32)
#include <sys/types.h>
#include <string.h>
#include <mips/m32c0.h>
#include <mips/m32xcpt.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <mips/prid.h>
#include <mips/watchpoint.h>

#include "kit_impl.h"

static int	m32wp_init (unsigned int);
static void	m32wp_setup (struct mips_watchpoint *);
static int 	m32wp_hit (const struct xcptcontext *, vaddr_t *, size_t *);
static int	m32wp_set (struct mips_watchpoint *);
static int	m32wp_clear (struct mips_watchpoint *);
static void	m32wp_remove (void);
static void	m32wp_insert (void);
static void	m32wp_reset (void);

const struct mips_watchpoint_funcs _m32_watchpoint_funcs = {
    m32wp_init,
    m32wp_setup,
    m32wp_hit,
    m32wp_set,
    m32wp_clear,
    m32wp_remove,
    m32wp_insert,
    m32wp_reset
};


struct m32wp {
    unsigned long 	lo;
    unsigned long 	hi;
};

unsigned int m32_dbgreg;

#define NWP	8
static struct m32wp m32wp[8];
static int m32nwp;

static int	
m32wp_init (unsigned int prid)
{
    if ((prid >> 16) & 0xff == 0) 
	/* unrecognised */
	return -1;

    if (! (mips32_getconfig1 () & CFG1_WR))
	/* mips32 but watchpoints not supported */
	return 0;

    /* count how many have we got */
    for (m32nwp = 0; m32nwp < NWP; ) {
      m32nwp++;
      if (!(mips32_getwatchhi (m32nwp - 1) & WATCHHI_M))
	break;	/* no more watchpoints */
    }

    return m32nwp;
}

static void
m32wp_setup (struct mips_watchpoint *wp)
{
    unsigned int mask;

    wp->private = &m32wp[wp->num];

    /* MIPS32 watchpoints are minimum of dword aligned, virtual addr match */
    wp->capabilities = MIPS_WATCHPOINT_DWORD | MIPS_WATCHPOINT_VADDR;
    
    /* determine watchpoint types supported by this register */
    mips32_setwatchlo (wp->num, WATCHLO_I | WATCHLO_R | WATCHLO_W);
    wp->capabilities |= (mips32_getwatchlo (wp->num)
			 & (WATCHLO_I | WATCHLO_R | WATCHLO_W));

    /* clear watchpoint */
    mips32_setwatchlo (wp->num, 0);

    /* determine mask availability and size */
    mips32_setwatchhi (wp->num, WATCHHI_MASK);
    mask = mips32_getwatchhi (wp->num) & WATCHHI_MASK;
    if (mask != 0)
	wp->capabilities |= MIPS_WATCHPOINT_MASK;
    wp->maxmask = mask | 0x7;

    /* clear "hit" bits */
    mips32_setwatchhi (wp->num, WATCHHI_I | WATCHHI_R | WATCHHI_W);
}


static int
m32wp_hit (const struct xcptcontext *xcp, vaddr_t *pva, size_t *plen)
{
    int hit = 0;
    int i;

    if (XCP_CODE (xcp) == XCPTWATCH) {
	if (! (XCP_CAUSE (xcp) & CR_WP) ) {
#ifdef DBG_DIB
	    /* Use Jade debug register to distinguish */
	    if (m32_dbgreg & DBG_DIB)
		hit |= MIPS_WATCHPOINT_X;
	    if (m32_dbgreg & DBG_DDBS)
		hit |= MIPS_WATCHPOINT_W;
	    if (m32_dbgreg & DBG_DDBL)
		hit |= MIPS_WATCHPOINT_R;
#else
	    /* Use MIPS32r2 watchhi bits to distinguish */
	    for (i = 0; i < m32nwp; i++)
		hit |= (m32wp[i].hi 
			& (WATCHHI_I | WATCHHI_R | WATCHHI_W));
#endif
	    if (hit == 0)
		hit = MIPS_WATCHPOINT_X | MIPS_WATCHPOINT_R 
		    | MIPS_WATCHPOINT_W;
	    /* Use generic code to fully decode */
	    hit = _mips_watchpoint_address (xcp, hit, pva, plen);
	}
    }

    return hit;
}


static int
m32wp_set (struct mips_watchpoint *wp)
{
    struct m32wp *pwp = wp->private;

    /* must fit in one watchpoint reg */
    if (wp->mask > wp->maxmask)
	return 0;

    pwp->lo = wp->va & ~wp->mask & WATCHLO_VA;
    pwp->lo |= wp->type & (WATCHLO_I | WATCHLO_R | WATCHLO_W);

    pwp->hi = wp->mask & WATCHHI_MASK;
    if (wp->asid < 0)
	pwp->hi |= WATCHHI_G;
    else 
	pwp->hi |= (wp->asid << WATCHHI_ASIDSHIFT) & WATCHHI_ASIDMASK;

    return 1;
}


static int
m32wp_clear (struct mips_watchpoint *wp)
{
    struct m32wp *pwp = wp->private;
    if (pwp->lo)
	pwp->lo = pwp->hi = 0;
    return 1;
}


static void	
m32wp_remove (void)
{
    int count = 0;
    int i;

#ifdef DBG_DIB
    m32_dbgreg = mips32_getdebug ();
#endif

    for (i = 0; i < m32nwp; i++) {
	if (m32wp[i].lo) {
	    m32wp[i].hi = mips32_getwatchhi (i);
	    mips32_setwatchlo (i, 0);
	    mips32_setwatchhi (i, WATCHHI_I | WATCHHI_R | WATCHHI_W);
	    count++;
	}
    }

    if (count > 0)
	_mips_watchlo = _mips_watchhi = 0;
}


static void
m32wp_insert (void)
{
    struct m32wp *pwp = m32wp;
    int i;

    /* wp0 handled by exception handler */
    if (pwp->lo != 0) {
	_mips_watchhi = pwp->hi;
	_mips_watchlo = pwp->lo;
    }

    /* all "other" watchpoints */
    for (pwp = &m32wp[1], i = 1; i < m32nwp; pwp++, i++) {
	if (pwp->lo != 0) {
	    mips32_setwatchhi (i, pwp->hi);
	    mips32_setwatchlo (i, pwp->lo);
	}
    }
}

static void
m32wp_reset (void)
{
    int i;

    _mips_watchlo = _mips_watchhi = 0;
    for (i = 0; i < m32nwp; i++) {
	mips32_setwatchhi (i, WATCHHI_I | WATCHHI_R | WATCHHI_W);
	mips32_setwatchlo (i, 0);
    }
#ifdef DBG_DIB
    mips32_setdebug (0);
#endif
    memset (m32wp, 0, sizeof (m32wp));
}
#endif
