#ident "kit/share/mfs.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mfs.c: a simple memory-based, read-only Posix file system.
 * See ../$MONITOR/root.c for the structures which it traverses.
 */


#define _KERNEL
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/dirent.h>
#include <sys/errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "file.h"

static int	_mfsread (struct file *, struct uio *);
static int	_mfswrite (struct file *, struct uio *);
static int	_mfsioctl (struct file *, int, void *);
static int	_mfsselect (struct file *, int);
static int	_mfsopen (struct file *, dev_t, int);
static int	_mfsclose (struct file *);
static int	_mfslookup (dir_t **, dir_t *, const char **, int, mode_t);
static int	_mfsinit (dir_t *);
static int	_mfsmknod (dir_t *, const dir_t *);
static int	_mfsstat (dir_t *, file_t *, struct stat *);

const struct fileops _mfs_ops = 
{
    _mfsread, 
    _mfswrite, 
    _mfsioctl, 
    _mfsselect, 
    _mfsopen,
    _mfsclose,
    _mfslookup,
    0,				/* dfree */
    _mfsinit,
    _mfsmknod,
    _mfsstat,
};

static ino_t ino = 1;

static int
_mfsread (struct file *fp, struct uio *uio)
{
    dir_t *nd, *dd;
    struct dirent *dp = 0;
    off_t doff = 0;
    size_t prevdlen;
    int err = 0;

    for (nd = (dir_t *) fp->f_dp->d_data; nd && uio->uio_resid != 0; nd = nd->d_sibling) {
	size_t nlen = strlen (nd->d_name);
	size_t dlen, allocdlen;

	/* calculate record length, must a 4 byte multiple */
	dlen = (sizeof (struct dirent) - sizeof(dp->d_name) + nlen + 1 + 3)
	    & ~3;

	if (uio->uio_offset <= doff) {
	    /* current entry lies at or beyond current file offset */

	    if (dlen > uio->uio_resid)
		/* not enough space in buffer for whole record */
		break;
	    
	    /* allocate buffer to build dir entry */
	    /* round allocation up to 64 bytes */
	    allocdlen = (dlen + 63) & ~63;
	    if (!dp || allocdlen > prevdlen) {
		prevdlen = allocdlen;
		if (dp)
		    dp = realloc (dp, allocdlen);
		else
		    dp = malloc (allocdlen);
		if (!dp) {
		    err = ENOMEM;
		    break;
		}
	    }

	    /* build dir entry */
	    dp->d_reclen = dlen;
	    dd = nd->d_link;
	    dp->d_fileno = dd->d_ino;
	    dp->d_type = IFTODT (dd->d_mode);
	    dp->d_namlen = nlen;
	    strcpy (dp->d_name, nd->d_name);

	    /* copy to user */
	    err = uiomove (dp, dlen, uio);
	    if (err)
		break;
	}

	doff += dlen;
    }

    if (dp)
	free (dp);
    return err;
}

static int
_mfswrite (struct file *fp, struct uio *uio)
{
    return EISDIR;
}

static int
_mfsioctl (struct file *fp, int cmd, void *args)
{
    return EISDIR;
}

    
static int
_mfsselect (struct file *fp, int rw)
{
    return FREAD | FWRITE;
}

static int
_mfsopen (struct file *fp, dev_t dev, int fflags)
{
    return 0;
}

static int
_mfsclose (struct file *fp)
{
    return 0;
}


static int
_mfslookup (dir_t **ndpp, dir_t *dp, const char **name, 
	    int fflags, mode_t mode)
{
    const char *f = *name, *e;
    dir_t *nd;
    int len;
    
    /* skip over extra path separators */
    if (*f == '/') {
	while (*++f == '/')
	    continue;
	/* if reached end of name, then return current dirnode */
	if (!*f) {
	    *name = f;
	    *ndpp = dp;
	    return 0;
	}
    }

    /* pick out the current name component */
    for (e = f; *e && *e != '/'; e++)
	continue;
    len = e - f;
    assert (len > 0);

    /* search directory */
    for (nd = (dir_t *) dp->d_data; nd; nd = nd->d_sibling) {
	if (strncmp (nd->d_name, f, len) == 0 && nd->d_name[len] == '\0')
	    break;
    }

    if (!nd) {
	/* not found */
	if (*e == '\0' && (fflags & O_CREAT))
	    /* attempt to create a file */
	    return EROFS;
	/* non-existent file */
	return ENOENT;
    }
    
    /* advance name pointer past match */
    *name = e;

    /* get the real dirnode that this dir entry refers to */
    if (!(fflags & FNODREF))
	nd = nd->d_link;

    /* bump node reference count */
    __sdethread_mutex_lock (&nd->d_mx);
    nd->d_count++;
    __sdethread_mutex_unlock (&nd->d_mx);

    /* return new directory node */
    *ndpp = nd;
    return 0;
}


static int
_mfsmknod (dir_t *pdp, const dir_t *node)
{
    int nlen = strlen (node->d_name);
    dir_t *dp, *ldp = NULL;
    char *namep;

    /* XXX more extensive thread protection required here */
    __sdethread_mutex_lock (&pdp->d_mx);

    for (dp = (dir_t *) pdp->d_data; dp; dp = dp->d_sibling) {
	if (strcmp (dp->d_name, node->d_name) == 0) {
	    __sdethread_mutex_unlock (&pdp->d_mx);
	    return EEXIST;
	}
	ldp = dp;
    }

    /* allocate new node and name */
    dp = malloc (sizeof(dir_t) + nlen + 1);
    if (!dp) {
	__sdethread_mutex_unlock (&pdp->d_mx);
	return ENOMEM;
    }

    *dp = *node;
    dp->d_name = namep = (char *)(dp + 1);
    strcpy (namep, node->d_name);
    dp->d_sibling = NULL;
    dp->d_link = dp;
    dp->d_flags = DF_DYNAMIC;
    dp->d_ino = ino++;
    dp->d_nlink = 1;
    __sdethread_mutex_init (&dp->d_mx, "dnode");

    /* append to end of directory */
    if (ldp)
	ldp->d_sibling = dp;
    else
	pdp->d_data = dp;

    /* bump parent directory link count and size */
    pdp->d_size += 24;
    if (S_ISDIR (dp->d_mode))
	pdp->d_nlink++;

    __sdethread_mutex_unlock (&pdp->d_mx);
    return 0;
}


static int
_mfsstat (dir_t *dp, file_t *fp, struct stat *stb)
{
    stb->st_ino = dp->d_ino;
    stb->st_size = dp->d_size;
    stb->st_nlink = dp->d_nlink;
    return 0;
}


/* Traverse the compiled-in directory tree, and link siblings together */
static int
_mfsinit (dir_t *dp)
{
    dir_t *nd = dp->d_data;
    dir_t *ld = NULL;
    int ret = 0;

    if (nd->d_parent)
	/* directory contants already scanned */
	return 0;
    
    dp->d_ino = ino++;
    if (!dp->d_parent)
	/* top-level dir, must be root */
	dp->d_parent = dp;
    
    for (; nd->d_name; nd++) {

	/* bump parent directory size */
	dp->d_size += 24;
	nd->d_parent = dp;
	nd->d_link = nd;

	if (S_ISDIR (nd->d_mode) && nd->d_ops->fo_init) { 
	    /* handle "." and ".." magic links */
	    if (nd->d_name && nd->d_name[0] == '.') {
		if (nd->d_name[1] == '\0') {
		    /* "." hard link to parent dir */
		    nd->d_link = dp;
		    nd->d_link->d_nlink++;
		}
		else if (nd->d_name[1] == '.' && nd->d_name[2] == '\0') {
		    /* ".." hard link to parent's parent */
		    nd->d_link = dp->d_parent;
		    nd->d_link->d_nlink++;
		}
	    }

	    /* initialise the sub-directory */
	    if ((*nd->d_ops->fo_init) (nd) < 0)
		ret = -1;
	}
	else {
	    /* not a directory */
	    nd->d_ino = ino++;
	    nd->d_nlink = 1;
	}
	/* link onto sibling chain */
	if (ld)
	    ld->d_sibling = nd;
	ld = nd;
    }
    return ret;
}
