#ident "kit/share/mon.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2005 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * share/mon.c: simple integer-only, unbuffered printf direct to hardware
 */

 
#include <sys/types.h>
#include <stdarg.h>
#include <ctype.h>

#include "kit_impl.h"

#define SIGNED	0x01
#define ALLSIGN	0x02
#define LEFT	0x04
#define ALT	0x08
#define UPPER	0x10
#define ZEROPAD	0x20
#define SPACE	0x40
#define PREC	0x80


static size_t
mystrlen (const char *s)
{
    const char *e = s;
    while (*e)
	++e;
    return e - s;
}


static char *
convert (int val, unsigned int base, unsigned int prec, 
	 unsigned int width, unsigned int flags)
{
    const char *digits = (flags & UPPER) 
	? "0123456789ABCDEF" 
	: "0123456789abcdef";
    static char buf[20];
    unsigned int n;
    int i;

    /* quick limit check */
    if (prec > sizeof(buf) - 4)
	prec = sizeof(buf) - 4;
    if (width > sizeof(buf) - 1)
	width = sizeof(buf) - 1;

    /* get absolute value */
    if (flags & SIGNED && val < 0)
	n = -val;
    else
	n = val;

    i = sizeof(buf);
    buf[--i] = '\0';
#define fldwidth	(sizeof(buf) - i - 1)

    /* generate digit string backwards */
    while (n != 0) {
	buf[--i] = digits[n % base];
	n /= base;
    }

    /* expand to minimum precision with leading zeros */
    while (fldwidth < prec)
	buf[--i] = '0';

    if (flags & ZEROPAD) {
	if (flags & ALT && val != 0)
	    /* leave room for base prefix */
	    width -= (base == 8) ? 1 : (base == 16) ? 2 : 0;
	/* pad to minimum width with leading zeros */
	while (fldwidth < width)
	    buf[--i] = '0';
    }

    if (flags & ALT && val != 0) {
	/* add base prefix */
	if (base == 8)
	    buf[--i] = '0';
	else if (base == 16)
	    buf[--i] = (flags & UPPER) ? 'X' : 'x', buf[--i] = '0';
    }

    /* add sign prefix */
    if (flags & SIGNED && val < 0)
	buf[--i] = '-';
    else if (flags & ALLSIGN)
	buf[--i] = '+';
    else if (flags & SPACE)
	buf[--i] = ' ';

#undef fldwidth

    return (&buf[i]);
}

#define BUFSIZE 128

struct strbuf {
    char *op;
    char *ep;
    char buf[BUFSIZE];
};

static void
sbflush (struct strbuf *sb)
{
    if (sb->op != sb->buf) {
	*sb->op = '\0';
	_mon_puts (sb->op = sb->buf);
    }
}

static void
sbputc (struct strbuf *sb, char c)
{
    if (sb->op >= sb->ep)
	sbflush (sb);
    *sb->op++ = c;
}

void
_mon_vprintf (const char *format, va_list ap)
{
    struct strbuf sb;
    int c;
    unsigned int base;
    int arg;
    unsigned int notflag, flags, width, prec, len;
    char *s;

    sb.op = sb.buf;
    sb.ep = sb.buf + BUFSIZE - 2;
#define _vpputc(c) sbputc (&sb, c)

    while (1) {
	if ((c = *format++) == '\0')
	    break;

	if (c == '\n') {
	    _vpputc ('\r');
	    _vpputc ('\n');
	    continue;
	}

	if (c != '%') {
	    _vpputc (c);
	    continue;
	}

	flags = width = 0;
	prec = 1;

	/* handle initial flags */
	notflag = 0;
	while (!notflag) {
	    c = *format++;
	    switch (c) {
	    case '0':
		flags |= ZEROPAD;
		break;
	    case '-':
		flags |= LEFT;
		break;
	    case ' ':
		flags |= SPACE;
		break;
	    case '+':
		flags |= ALLSIGN;
		break;
	    case '#':
		flags |= ALT;
		break;
	    default:
		notflag = 1;
		break;
	    }
	}

	if (flags & LEFT)
	    flags &= ~ZEROPAD;

	/* handle width */
	if (c == '*') {
	    width = va_arg (ap, int);
	    c = *format++;
	}
	else {
	    while (isdigit (c)) {
		width = width * 10 + c - '0';
		c = *format++;
	    }
	}

	/* handle precision */
	if (c == '.') {
	    flags |= PREC;
	    c = *format++;
	    if (c == '*') {
		prec = va_arg (ap, unsigned int);
		c = *format++;
	    }
	    else {
		prec = 0;
		while (isdigit (c)) {
		    prec = prec * 10 + c - '0';
		    c = *format++;
		}
	    }
	}

	/* handle size modifiers (redundant) */
	if (c == 'l' || c == 'h')
	    c = *format++;

	if (c == '\0')
	    break;

	switch (c) {
	case 'p':
	    flags |= ALT;
	    if (!(flags & PREC))
		prec = sizeof(void *)/4;
	    /* drop through */
	case 'X':
	    flags |= UPPER;
	    /* drop through */
	case 'x':
	    base = 16;
	    goto doconvert;
	case 'o':
	    base = 8;
	    goto doconvert;
	case 'd':
	case 'i':
	    flags |= SIGNED;
	    /* drop through */
	case 'u':
	    base = 10;
	    goto doconvert;
	doconvert:
	    /* XXX note assume short=int=long; don't handle "long long" */
	    arg = va_arg(ap, int);
	    s = convert (arg, base, prec, width, flags);
	    prec = len = mystrlen (s);
	    goto dostring;
	case 's':
	    s = va_arg(ap, char *);
	    if (!s)
		s = "(null)";
	    len = mystrlen (s);
	    if (!(flags & PREC) || len < prec)
		prec = len;
	dostring:
	    if (len < width && !(flags & LEFT)) {
		/* pad on left with spaces */
		do {
		    _vpputc (' ');
		} while (++len < width);
	    }
	    /* write the result */
	    while (prec-- != 0)
		_vpputc (*s++);
	    /* pad on right with spaces */
	    while (len++ < width)
		_vpputc (' ');
	    break;
	case 'c':
	    arg = va_arg(ap, int);
	    _vpputc (arg);
	    break;
	default:
	    _vpputc (c);
	    break;
	}
    }

    sbflush (&sb);
}


void
_mon_printf (const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    _mon_vprintf (fmt, ap);
    va_end(ap);
}
