#ident "kit/share/muldivem.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1998-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * muldivem.c: trap-based emulator for missing h/w multiplier (e.g. cw4010)
 */


/*
 * This needs to be rewritten as a fast, low-level assembler exception
 * handler, but at least serves as an example.
 *
 * Note that code compiled -mno-mul will not need this, because
 * it will call emulation routines directly.  This is only for
 * library code which can't be recompiled.
 *
 * XXX Note: we don't emulate 64-bit dmult or ddiv because all 
 * (known) 64-bit cpus do have a h/w multiplier.
 */

#include <stddef.h>
#include <mips/cpu.h>
#include <mips/fpa.h>
#include <mips/xcpt.h>
#include <sys/signal.h>
#include "kit_impl.h"

/* Link to next handler */

static struct xcptaction chainxa;

/* Library functions from libc */

extern unsigned long long __mulsidi3 (unsigned int, unsigned int);
extern unsigned long long __umulsidi3 (unsigned int, unsigned int);
extern unsigned int __udivmodsi4 (unsigned int, unsigned int, unsigned int *);

/* convert MIPS16 register to MIPS32 register */
#define mips16_to_mips32(rn)	(((rn) < 2) ? (rn) + 16 : (rn))

static int
muldiv_handler (int xcptno, struct xcptcontext *xcp)
{
    unsigned int insn;
    int rs, rt, rd;
    int err = 0;

#if __mips16 || __maybe_mips16
    if (xcp->epc & 1) {
	insn = mips_get_half (REG_TO_VA (xcp->epc & ~1), &err);
	if (err) return SIGBUS;

	if ((insn & 0xf81c) == 0xe818) {
	    rs = mips16_to_mips32 ((insn >> 8) & 0x7);
	    rt = mips16_to_mips32 ((insn >> 5) & 0x7);
	    goto muldiv;
	}
	else if ((insn & 0xf8fe) == 0xe810) {
	    rs = rd = mips16_to_mips32 ((insn >> 8) & 0x7);
	    goto hilo;
	}
	else {
	    goto unrecognised;
	}
    }
    else 
#endif
    {
	insn = mips_get_word (REG_TO_VA (xcp->epc), &err);
	if (err) return SIGBUS;

	if ((insn & 0xfc00fffc) == 0x00000018) {
	    unsigned long long tmp;
	    int quo;
	    unsigned int mod;

	    rs = xcp->regs[(insn >> 11) & 0x1f];
	    rt = xcp->regs[(insn >> 16) & 0x1f];

#if __mips16 || __maybe_mips16
	muldiv:
#endif
	    /* A mul/div instruction should never be in a 
	       branch delay slot. */
	    if (XCP_CAUSE (xcp) & CR_BD)
		return SIGILL;
	    rs = xcp->regs[rs];
	    rt = xcp->regs[rt];
	    switch (insn & 0x3) {
	    case 0x0:
		/* mult */
		tmp = __mulsidi3 (rs, rt);
		xcp->mdhi = (int)(tmp >> 32);
		xcp->mdlo = (int)tmp;
		break;
	    case 1:
		/* multu */
		tmp = __umulsidi3 (rs, rt);
		xcp->mdhi = (int)(tmp >> 32);
		xcp->mdlo = (int)tmp;
		break;
	    case 2:
		/* div */
		if (rt != 0) {
		    int sign = 0;
		    if (rs < 0) {
			sign = 3;
			rs = -rs;
		    }
		    if (rt < 0) {
			sign ^= 1;
			rt = -rt;
		    }
		    quo = __udivmodsi4 (rs, rt, &mod);
		    xcp->mdlo = (sign & 1) ? -quo : quo;
		    xcp->mdhi = (sign & 2) ? -mod : mod;
		}
		break;
	    case 3:
		/* divu */
		if (rt != 0) {
		    quo = __udivmodsi4 (rs, rt, &mod);
		    xcp->mdlo = quo;
		    xcp->mdhi = mod;
		}
		break;
	    }
	}
	else if ((insn & 0xfc1f07fc) == 0x00000010) {
	    /* A mul/div instruction should never be in a 
	       branch delay slot. */
	    if (XCP_CAUSE (xcp) & CR_BD)
		return SIGILL;
	    rs = (insn >> 21) & 0x1f;
	    rd = (insn >> 11) & 0x1f;
#if __mips16 || __maybe_mips16
	hilo:
#endif
	    switch (insn & 0x3) {
	    case 0:
		/* mfhi */
		xcp->regs[rd] = xcp->mdhi;
		break;
	    case 1:
		/* mthi */
		xcp->mdhi = xcp->regs[rs];
		break;
	    case 2:
		/* mflo */
		xcp->regs[rd] = xcp->mdlo;
		break;
	    case 3:
		/* mtlo */
		xcp->mdlo = xcp->regs[rs];
		break;
	    }
	}
	else {
	    goto unrecognised;
	}
    }

    /* advance over emulated insn */
    xcp->epc += (xcp->epc & 1) ? 2 : 4;
    return 0;

unrecognised:
    /* not handled, let next handler have a go */
    if (chainxa.xa_handler != XCPT_DFL)
	return (*chainxa.xa_handler) (xcptno, xcp);
    return SIGILL;
}


void
_muldiv_emu_init (void)
{
    static int inited = 0;
    if (!inited) {
	struct xcptaction xa;
	xa.xa_handler = muldiv_handler;
	xa.xa_flags = 0;
	xcptaction (XCPTRI, &xa, &chainxa);
	inited = 1;
    }
}
