#ident "kit/share/nintrupt.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1997-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * intrupt.c: MIPS low-level interrupt handler 
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <sys/errno.h>
#include <sys/signal.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include "kit_impl.h"
#include "sbd.h"

#ifndef INTR_MAX
#define INTR_MAX	64
#endif

/* the interrupt mask with which we were entered */
unsigned int		_sr_base __attribute__((common));

/* The PROM monitor exception trampolines use this to decide 
   which i/us to pass back to the monitor (see ../pmon/pmonxlow.sx) */
unsigned int		_intrmask;

/* the interrupt mask when at non-interrupt level (ipl0) */
extern unsigned int	_sbd_ipl0_sr;

/* the CPU status register when at normal interrupt level (ipl1) */
extern unsigned int	_sbd_ipl1_sr;

/* the CPU status register when at interrupt blocking level (ipl2) */
extern unsigned int	_sbd_ipl2_sr;

/* One entry per interrupt */
static struct intraction intrtab[INTR_MAX];

/* Previous interrupt exception handler */
static struct xcptaction oldxact;



static void
_intrupt_init (void)
{
    static int intrhandler (int, struct xcptcontext *);
    struct xcptaction xact;

    _sbd_intrinit ();

    xact.xa_handler = intrhandler;
    xact.xa_flags = 0;
    (void) xcptaction (XCPTINTR, &xact, &oldxact);
}


/*
 * intraction(): installs interrupt handler and (optionally) returns
 * old handler.  The handler is called with new->ia_arg as its 
 * first argument, and the xception frame as its second.
 */
int intraction (unsigned int intr, const struct intraction *new,
		struct intraction *old)
{
    static int xcpt_set = 0;
    struct intraction *ia;
    struct ipl *ip, *oip;
    unsigned int ibit, sr;
    int i;

    if (intr == 0 || intr > INTR_MAX)
      return EINVAL;
    ia = &intrtab[intr-1];

    /* block *all* interrupts */
    mips_setsr (_sbd_ipl2_sr);

    if (!xcpt_set) {
	_intrupt_init ();
	xcpt_set = 1;
    }

    /* store new interrupt handler in table */
    if (old) 
      *old = *ia;
    *ia = *new;

    /* if this is a valid handler, enable the interrupt */
    if (new->ia_handler) {
	/* tell low-level code that we're interested in this interrupt */
	_sbd_intrenable (intr);
	/* tell prom trampoline too */
	_intrmask |= irr;
    }

    /* reenable interrupts */
    mips_setsr (_sbd_ipl1_sr | (_sbd_ipl0_sr & SR_IMASK));

    return 0;
}


/*
 * intrupt(): simpler way to install interrupt handler; returns old handler.
 */
xcpt_t intrupt (unsigned int intrno, xcpt_t handler, int arg)
{
    struct intraction new, old;
    int err;

    memset (&new, 0, sizeof (new));
    if (handler != XCPT_DFL) {
	/* add handler */
	new.ia_handler = handler;
	new.ia_arg = arg;
    }

    if (err = intraction (intrno, &new, &old)) {
	extern int errno;		/* XXX shouldn't set errno */
	errno = err;
	return XCPT_ERR;
    }

    if (old.ia_handler == 0)
	return XCPT_DFL;

    return old.ia_handler;
}


int
intrpending (intrset_t *pend)
{
    mips_setsr (_sbd_ipl1_sr);
    _sbd_intrpending (pend);
    mips_setsr (_sbd_ipl1_sr | (_sbd_ipl0_sr & SR_IMASK));
    return 0;
}


int
intrprocmask (int how, const intrset_t *new, intrset_t *old)
{
    int err;
    mips_setsr (_sbd_ipl1_sr);
    err = _sbd_intrprocmask (how, new, old);
    mips_setsr (_sbd_ipl1_sr | (_sbd_ipl0_sr & SR_IMASK));
    return err ? EINVAL : 0;
}



/*
 * Low level interrupt exception handler, called at ipl2
 */
static int intrhandler (int xcptno, struct xcptcontext *xcp)
{
    int sig = 0;
    int intr;

    if (_sbd_intrhandler (0, xcp))
	/* board-specific code handled it, but we may have been 
	   at ipl1, so return after handling it. */
	return 0;

    /* switch to ipl1 (enabling non-user interrupts) */
    mips_setsr (_sbd_ipl1_sr);

    /* continue while user interrupts are pending */
    while ((intr = _sbd_intrnext (0)) && sig == 0) {
	if (intr < 0) {
	    /* the timer clock tick is magic */
	    sig = _timertick (xcp);
	}
	else {
	    /* a normal interrupt */
	    struct intraction *ia = &intrtab[intr - 1];
	    if (ia->ia_handler) {
		intrset_t omask;

		/* disable other interrupts, as requested */
		_sbd_intrprocmask (INTR_DISABLE, &ia->ia_mask, &omask);

		/* call handler with other interrupts reenabled */
		mips_setsr (_sbd_ipl1_sr | (_sbd_ipl0_sr & SR_IMASK));
		sig = (ia->ia_handler) (ia->ia_arg, xcp);
		mips_setsr (_sbd_ipl1_sr);

		/* restore old interrupt mask and reenable this interrupt */
		intraddset (&omask, intr);
		_sbd_intrprocmask (INTR_SETMASK, &omask, 0);
	    }
	}
    }

    /* insert new interrupt mask into saved status register */
    xcp->sr ^= (xcp->sr ^ _sbd_ipl0_sr) & SR_IMASK;
    return sig;
}
