#ident "kit/share/paneldev.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

#if #feature(paneldev)

/*
 * Copyright (c) 2000-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * paneldev.c: Posix style access to raw alphanumeric display
 */


#define _KERNEL
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/panelio.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/init.h>
#include <stdlib.h>
#include <string.h>
#include <sdethread.h>
#include <assert.h>
#include "file.h"
#include "kit_impl.h"

static int	panelwrite (struct file *, struct uio *);
static int	panelioctl (struct file *, int, void *);
static int	panelselect (struct file *, int);
static int	panelopen (struct file *, dev_t, int);
static int	panelclose (struct file *);

static const struct fileops panel_ops = 
{
    NULL, 
    panelwrite, 
    panelioctl, 
    panelselect, 
    panelopen,
    panelclose,
};


static __sdethread_mutex_t panel_mx = __SDETHREAD_MUTEX_INITIALIZER("panel");
static struct panelinfo panel_info;
static struct panelcoord panel_coord;
static int panel_opt = PANELOPT_PAD;
static size_t panelsize;
static char * panelbuf;

static int
panelwrite (struct file *fp, struct uio *uio)
{
    size_t nb = panelsize;
    int err;

    if (nb > uio->uio_resid)
	nb = uio->uio_resid;

    __sdethread_mutex_lock (&panel_mx);
    err = uiomove (panelbuf, nb, uio);
    if (err == 0) {
	_sbdpanelmessage (panel_coord.row, panel_coord.col, panelbuf, nb, panel_opt);
	uio->uio_resid = 0;
    }
    __sdethread_mutex_unlock (&panel_mx);

    return err;
}

static int
panelioctl (struct file *fp, int cmd, void *arg)
{
    __sdethread_mutex_lock (&panel_mx);
    switch (cmd) {
    case PANELIOINFO: {
	struct panelinfo *info = arg;
	*info = panel_info;
	break;
    }

    case PANELIOSMODE: {
	const struct panelmode *mode = arg;
	_sbdpanelmode (mode, NULL);
	break;
    }

    case PANELIOGMODE: {
	struct panelmode *mode = arg;
	_sbdpanelmode (NULL, mode);
	break;
    }

    case PANELIOCLEAR: {
	_sbdpanelclear ();
	break;
    }

    case PANELIOPROGRESS: {
	int *percent = arg;
	_sbdpanelprogress (*percent);
	break;
    }

    case PANELIOSCOORD: {
	struct panelcoord *coord = arg;
	panel_coord = *coord;
	break;
    }

    default:
	__sdethread_mutex_unlock (&panel_mx);
	return EINVAL;
    }

    __sdethread_mutex_unlock (&panel_mx);
    return 0;
}


static int
panelselect (struct file *fp, int rw)
{
    /* always ready! */
    return rw;
}


static int
panelopen (struct file *fp, dev_t dev, int fflags)
{

    if (dev > 0)
	return ENXIO;
    return 0;
}


static int
panelclose (struct file *fp)
{
    return 0;
}



/*
 * Create /dev/panel device
 */

/* /dev/panel*:	c-w--w--w- */
#define PANELMODE	(S_IFCHR|S_IWUSR|S_IWGRP|S_IWOTH)

/* call this function to load this module */
static void
paneldev_mount (void)
{
    static int inited = 0;
    dir_t node;

    if (inited)
	return;
    inited = 1;

    if (_sbdpanelinit ()) {

	_sbdpanelinfo (&panel_info);

	memset (&node, 0, sizeof (node));
	node.d_name = "panel";
	node.d_mode = PANELMODE;
	node.d_ops = &panel_ops;
	node.d_size = panel_info.rows * panel_info.cols;

	panelsize = node.d_size;
	panelbuf = malloc (panelsize);
	if (!panelbuf)
	    return;

	if (_mknod ("/dev", &node)) {
	    _mon_puts ("mknod(/dev/panel) failed\n");
	    abort ();
	}
    }
}

/* initialise panel device at MOUNT stage of startup */
DECL_INIT_FUNC(IF_MOUNT, paneldev_mount)

/* dummy _sbdpanelmode() for old board kits */
#pragma weak _sbdpanelmode=_dummy_sbdpanelmode
void _dummy_sbdpanelmode (const struct panelmode *, struct panelmode *);
void _dummy_sbdpanelmode (const struct panelmode *new, struct panelmode *old)
{
    if (old) {
	/* default mode for old board kits */
	memset (old, 0, sizeof (struct panelmode));
	old->options = panel_opt;
	old->brighton = old->contrast = 100;
    }
    if (new)
	panel_opt = new->options;
}


/* dummy _sbdpanelinfo() for old board kits */
#pragma weak _sbdpanelinfo=_dummy_sbdpanelinfo
void _dummy_sbdpanelinfo (struct panelinfo *info);
void _dummy_sbdpanelinfo (struct panelinfo *info)
{
    int rows, cols;
    _sbdpanelsize (&rows, &cols);
    info->type = PANELTYPE_ALPHA;
    info->flags = 0;
    info->rows = rows;
    info->cols = cols;
}


/* call this dummy function to load this module */
void _paneldev_load (void);
void _paneldev_load (void) {}

#endif /* #feature(paneldev) */
