#ident "kit/share/posix.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * posix.c: a Posix-like system call emulation for MIPS SDE
 */


#define _KERNEL
#include <sys/cdefs.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/filedesc.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#undef _KERNEL
#define _XOPEN_SOURCE 500
#include <sys/stat.h>
#include <sys/init.h>
#include <unistd.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <paths.h>

#define _KERNEL
#include "file.h"
#include "kit_impl.h"

/* from <sys/fcntl.h>, etc */
int	open (const char *, int, mode_t);
int	creat (const char *, mode_t);
int	fcntl (int, int, int);
int	ioctl (int, int, void *);

/* limit on total number of open files */
#define MAXFILES	(OPEN_MAX * 4)

/* the global file descriptor table and free list */
static __sdethread_mutex_t files_mx = __SDETHREAD_MUTEX_INITIALIZER("files");
static file_t	files[NDFILE];	/* initial static table */
static file_t *	freefiles;
static int nfiles;

/* initial file descriptor table */
struct filedesc0 _fileio_fd0;

/* single-process get file descriptors function */
#pragma weak _fileio_getfdp
struct filedesc *
_fileio_getfdp ()
{
    return &_fileio_fd0.fd_fd;
}


/* Add a new reference to a dirnode handle. */
static dir_t *
duse (dir_t *dp)
{
    if (dp) {
        __sdethread_mutex_lock (&dp->d_mx);
	dp->d_count++;
        __sdethread_mutex_unlock (&dp->d_mx);
    }
    return dp;
}


/* Close a dirnode handle, optionally freeing it. */
static void
drelease (dir_t *dp)
{
    if (dp) {
	int unrefed;
        __sdethread_mutex_lock (&dp->d_mx);
	unrefed = --dp->d_count == 0;
	__sdethread_mutex_unlock (&dp->d_mx);
	/* XXX There is a window here where another thread could do a
	   lookup and get a new reference to this directory entry.  If
	   that is possible in a file system, then its "dfree" code
	   should relock the mutex and check that the count is still
	   zero before freeing the data structure. */
	if (unrefed && dp->d_ops->fo_dfree)
	    (*dp->d_ops->fo_dfree) (dp);
    }
}


/* Free a file handle. */
static void
_fileio_ffree (file_t *fp)
{
    /* must be unreferenced */
    assert (fp->f_count == 0);

    /* unlock associated dirnode, if any */
    if (fp->f_dp) {
	drelease (fp->f_dp);
	fp->f_dp = NULL;
    }
    fp->f_type = DTYPE_FREE;

    /* don't free memory, put back on free file list */
    __sdethread_mutex_lock (&files_mx);
    fp->f_next = freefiles;
    freefiles = fp;
    nfiles--;
    __sdethread_mutex_unlock (&files_mx);
}


/* Allocate a file handle and optional file descriptor. */
int
_fileio_falloc (file_t **fpp, int *fdp)
{
    file_t *fp;

    /* get first file handle off free list */
    __sdethread_mutex_lock (&files_mx);
    if (fp = freefiles)
	freefiles = fp->f_next;
    else {
	/* none free - allocate a new one */
	if (nfiles >= MAXFILES
	     || !(fp = malloc (sizeof (file_t)))) {
	    __sdethread_mutex_unlock (&files_mx);
	    if  (fdp) 
		*fdp = -1;
	    *fpp = NULL;
	    return ENFILE;
	}
	memset (fp, 0, sizeof (file_t));
    }
    ++nfiles;
    __sdethread_mutex_unlock (&files_mx);

    /* initialise fields */
    fp->f_dp = NULL;
    fp->f_ops = NULL;
    fp->f_data = NULL;
    fp->f_offset = 0;

    /* safe because no-one else points to fp yet */
    fp->f_count = 1;

    /* allocate and attach to a file descriptor */
    if (fdp) {
	int err = _fileio_fdalloc (fp, 0, fdp);
	if (err) {
	    _fileio_frelease (fp);
	    return err;
	}
    }

    /* return with reference count == 2 */
    *fpp = fp;
    return 0;
}


/* Allocate user file descriptor number, and store file handle
   in descriptor table. */
int
_fileio_fdalloc (file_t *fp, int want, int *result)
{
    struct filedesc *fdp = _fileio_getfdp ();
    struct file **newofile;
    char *newofileflags;
    int lim;
    int last, fd, nfiles, i;

    lim = OPEN_MAX;	/* XXX could be per process limit */

    /*
     * Search for a free descriptor starting at the higher
     * of want or fd_freefile.  If that fails, consider
     * expanding the ofile array.
     */

    __sdethread_mutex_lock (&files_mx);

    for (;;) {
	last = lim < fdp->fd_nfiles ? lim : fdp->fd_nfiles;
	if ((fd = want) < fdp->fd_freefile)
	    fd = fdp->fd_freefile;
	for (; fd < last; fd++) {
	    if (fdp->fd_ofiles[fd] == NULL) {
		fdp->fd_ofiles[fd] = fp;
		fdp->fd_ofileflags[fd] = 0;
		fp->f_count++;		/* bump file ref count */
		if (fd > fdp->fd_lastfile)
		    fdp->fd_lastfile = fd;
		if (want <= fdp->fd_freefile)
		    fdp->fd_freefile = fd;
		__sdethread_mutex_unlock (&files_mx);
		*result = fd;
		return (0);
	    }
	}

	/*
	 * No space in current array.  Expand?
	 */
	if (fdp->fd_nfiles >= lim) {
	    __sdethread_mutex_unlock (&files_mx);
	    *result = -1;
	    return (EMFILE);
	}

	if (fdp->fd_nfiles < NDEXTENT)
	    nfiles = NDEXTENT;
	else
	    nfiles = 2 * fdp->fd_nfiles;
	newofile = (struct file **) malloc (nfiles * OFILESIZE);
	if (!newofile) {
	    __sdethread_mutex_unlock (&files_mx);
	    *result = -1;
	    return (EMFILE);
	}
	newofileflags = (char *) &newofile[nfiles];

	/*
	 * Copy the existing ofile and ofileflags arrays
	 * and zero the new portion of each array.
	 */
	memcpy (newofile, fdp->fd_ofiles, 
		(i = sizeof(struct file *) * fdp->fd_nfiles));
	memset((char *)newofile + i, 0, nfiles * sizeof(struct file *) - i);
	memcpy(newofileflags, fdp->fd_ofileflags, 
	       (i = sizeof(char) * fdp->fd_nfiles));
	memset(newofileflags + i, 0, nfiles * sizeof(char) - i);
	if (fdp->fd_nfiles > NDFILE)
	    free (fdp->fd_ofiles);
	fdp->fd_ofiles = newofile;
	fdp->fd_ofileflags = newofileflags;
	fdp->fd_nfiles = nfiles;
    }
}


/* Remove a file descriptor from the file table */
void
_fileio_fdremove (int fd)
{
    struct filedesc *fdp = _fileio_getfdp ();

    __sdethread_mutex_lock (&files_mx);
    if (fd >= 0 && fd < fdp->fd_nfiles) {
	struct file *fp = fdp->fd_ofiles[fd];
	if (fp)
	    fp->f_count--; /* decrement file ref count */
	fdp->fd_ofiles[fd] = NULL;
	while (fdp->fd_lastfile > 0
	       && fdp->fd_ofiles[fdp->fd_lastfile] == NULL)
	    fdp->fd_lastfile--;
	if (fd < fdp->fd_freefile)
	    fdp->fd_freefile = fd;
    }
    __sdethread_mutex_unlock (&files_mx);
}


/*
 * Copy a filedesc structure.
 */
struct filedesc *
_fileio_fdcopy (struct filedesc *fdp)
{
    struct filedesc *newfdp;
    struct file **fpp;
    int i;

    newfdp = malloc (sizeof(struct filedesc0));
    if (!newfdp)
	return NULL;
    
    /* don't let any changes happen while copying */
    __sdethread_mutex_lock (&files_mx);

    memcpy (newfdp, fdp, sizeof(struct filedesc));

    newfdp->fd_cdir = duse(fdp->fd_cdir);
    newfdp->fd_rdir = duse(fdp->fd_rdir);
    newfdp->fd_refcnt = 1;
    
    /*
     * If the number of open files fits in the internal arrays
     * of the open file structure, use them, otherwise allocate
     * additional memory for the number of descriptors currently
     * in use.
     */
    if (newfdp->fd_lastfile < NDFILE) {
	newfdp->fd_ofiles = ((struct filedesc0 *) newfdp)->fd_dfiles;
	newfdp->fd_ofileflags =
	    ((struct filedesc0 *) newfdp)->fd_dfileflags;
	i = NDFILE;
    } else {
	/*
	 * Compute the smallest multiple of NDEXTENT needed
	 * for the file descriptors currently in use,
	 * allowing the table to shrink.
	 */
	i = newfdp->fd_nfiles;
	while (i > 2 * NDEXTENT && i >= newfdp->fd_lastfile * 2)
	    i /= 2;
	newfdp->fd_ofiles = malloc (i * OFILESIZE);
	if (!newfdp->fd_ofiles) {
	    __sdethread_mutex_unlock (&files_mx);
	    free (newfdp);
	    return NULL;
	}
	newfdp->fd_ofileflags = (char *) &newfdp->fd_ofiles[i];
    }

    newfdp->fd_nfiles = i;
    memcpy (newfdp->fd_ofiles, fdp->fd_ofiles, i * sizeof(struct file **));
    memcpy (newfdp->fd_ofileflags, fdp->fd_ofileflags, i * sizeof(char));

    /* bump file ref counts */
    fpp = newfdp->fd_ofiles;
    for (i = newfdp->fd_lastfile; i-- >= 0; fpp++)
	if (*fpp != NULL)
	    (*fpp)->f_count++;

    __sdethread_mutex_unlock (&files_mx);
    return (newfdp);
}


/*
 * Release a filedesc structure.
 */
void
_fileio_fdfree (struct filedesc *fdp)
{
    struct file **fpp;
    int i;
    
    __sdethread_mutex_lock (&files_mx);
    --fdp->fd_refcnt;
    __sdethread_mutex_unlock (&files_mx);

    if (fdp->fd_refcnt > 0)
	return;

    /* no-one else can be referencing this filedesc table now */
    fpp = fdp->fd_ofiles;
    for (i = fdp->fd_lastfile; i-- >= 0; fpp++)
	if (*fpp) 
	    _fileio_frelease (*fpp);
    if (fdp->fd_nfiles > NDFILE)
	free (fdp->fd_ofiles);
    drelease (fdp->fd_cdir);
    if (fdp->fd_rdir)
	drelease (fdp->fd_rdir);
    free (fdp);
}



/* Return the (locked) file handle attached to file descriptor. */
int
_fileio_fget (int fd, file_t **fpp)
{
    struct filedesc *fdp = _fileio_getfdp ();
    file_t *fp;

    __sdethread_mutex_lock (&files_mx);
    if (fd < 0 || fd >= fdp->fd_nfiles || !(fp = fdp->fd_ofiles[fd])) {
	__sdethread_mutex_unlock (&files_mx);
	*fpp = NULL;
	return EBADF;
    }
    fp->f_count++;	/* bump file ref count for caller */
    __sdethread_mutex_unlock (&files_mx);
    *fpp = fp;
    return 0;
}


void
_fileio_frelease (file_t *fp)
{
    if (fp) {
	/* decrement ref count */
	__sdethread_mutex_lock (&files_mx);
	--fp->f_count;	/* remove caller's reference */
	__sdethread_mutex_unlock (&files_mx);
	if (fp->f_count == 0) {
	    /* no-one else can have a pointer to this now */
	    if (fp->f_ops)
		(*fp->f_ops->fo_close) (fp);
	    _fileio_ffree (fp);
	}
    }
}



/*
 * A simple minded "recursive" directory lookup.
 */

int
_fileio_lookup (dir_t **dpp, const char *name, int fflags, mode_t mode)
{
    struct filedesc *fdp;
    dir_t *dp, *odp, *ndp;
    int newdev;
    int err = 0;

    /* must be a name to look up */
    if (!*name)
	return ENOENT;

    fdp = _fileio_getfdp ();
    if (*name == '/' || !fdp->fd_cdir) {
	/* strip all leading /s */
	while (*++name == '/') continue;
	dp = duse (fdp->fd_rdir);
    } else {
	dp = duse (fdp->fd_cdir);
    }
    odp = dp;

    while (1) {
	/* if no remaining name return the current directory */
	if (!*name)
	    break;

	/* lookup next section of the name */
	/* (note fo_lookup must have bumped ref count on returned dirnode) */
	err = (*dp->d_ops->fo_lookup) (&ndp, dp, &name, fflags, mode);
	if (err)
	    goto done;

	/* have we switched to a new file system or device */
	newdev = (ndp->d_ops != dp->d_ops) || (ndp->d_dev != dp->d_dev);

	/* release old dirnode */
	drelease (odp); 
	odp = dp = ndp;
	
	/* no more path left */
	if (!*name) {
	    if (S_ISDIR (dp->d_mode) && newdev) {
		/* looking up root dir of new device */
		name = ".";
		continue;
	    }
	    /* we've found the file */
	    break;
	}

	/* must be a directory */
	if (!S_ISDIR (dp->d_mode)) {
	    err = ENOTDIR;
	    goto done;
	}

	/* must now be at a file separator */
	if (*name++ != '/') {
	    err = ENOENT;
	    goto done;
	}
    }

    *dpp = dp;
    return 0;

 done:
    if (odp)
	drelease (odp);
    return err;
}



int
open (const char *name, int oflags, mode_t mode)
{
    int fflags = FFLAGS (oflags);
    file_t *fp = 0;
    dir_t *dp = 0;
    int fd = -1;
    int err;

    _syscall_enter();

    err = _fileio_falloc (&fp, &fd);
    if (err)
	goto error;

    err = _fileio_lookup (&dp, name, fflags, mode);
    if (err)
	goto error;

    if (((fflags & FWRITE) && !(dp->d_mode & S_IWUSR))
	|| ((fflags & FREAD) && !(dp->d_mode & S_IRUSR))) {
	err = EACCES;
	goto error;
    }
    
    /* initialise file descriptor */
    fp->f_type = DTYPE_DNODE;
    fp->f_flag = fflags & FMASK;
    fp->f_offset = 0;
    fp->f_ops = dp->d_ops;
    fp->f_data = dp->d_data;
    fp->f_dp = dp;
    dp = 0;			/* dp now in *fp */

    err = (*fp->f_ops->fo_open) (fp, fp->f_dp->d_dev, fflags);

 error:
    if (err != 0)
	_fileio_fdremove (fd);
    if (dp)
	drelease (dp);
    _fileio_frelease (fp);
    return _syscall_leave (err, fd);
}


/* not the POSIX mknod! */
int
_mknod (const char *path, const dir_t *node)
{
    dir_t *dp;
    int err;

    /* lookup the leading pathname */
    err = _fileio_lookup (&dp, path, FWRITE, 0);
    if (err == 0) {
	if (!(dp->d_mode & S_IWUSR) || !dp->d_ops->fo_mknod)
	    /* directory must be writable and support "mknod" */
	    err = EACCES;
	else
	    /* add the new node */
	    err = (*dp->d_ops->fo_mknod) (dp, node);
	drelease (dp);
    }

    return err;
}


#if defined(FREE_SDE) && 0
/* Users may easily disable the nag by setting this variable to one */
int _sde_stdout_once;

static __sdethread_once_t announce_once = __SDETHREAD_ONCE_INIT;

static void
announce ()
{
    const char freemsg[] = "\r\n"
	"  Built using MIPS Technologies' Free GNU Toolkit\r\n"
	"For a supported and more configurable product visit\r\n"
	"       http://www.mips.com/products/sde.html\r\n\n";

    if (!_sde_stdout_once)
	write (STDERR_FILENO, freemsg, sizeof(freemsg) - 1);
}
#endif


static ssize_t
_fileio_doio (int fd, const struct iovec *iovp, 
	      int iovcnt, int rw, off_t offs)
{
    file_t *fp;
    struct iovec aiov[UIO_SMALLIOV], *iov = aiov;
    struct iovec *needfree = NULL;
    struct uio uio;
    unsigned int iovlen;
    ssize_t ret = 0;
    size_t len;
    int err, i;

    _syscall_enter ();

#if defined(FREE_SDE) && 0
    if (rw == UIO_WRITE && (fd == STDOUT_FILENO || fd == STDERR_FILENO))
	__sdethread_once (&announce_once, announce);
#endif

    if (iovcnt <= 0 || iovcnt > IOV_MAX) {
	err = EINVAL;
	goto done;
    }

    /* allocate local iovec */
    iovlen = iovcnt * sizeof(struct iovec);
    if (iovcnt > UIO_SMALLIOV) {
	iov = needfree = malloc(iovlen);
	if (!iov) {
	    err = ENOMEM;
	    goto done;
	}
    }

    /* copy iovec to local storage, so we can update in place */
    memcpy (iov, iovp, iovlen);

    err = _fileio_fget (fd, &fp);
    if (err)
	goto done;

    if (!(fp->f_flag & (rw == UIO_READ ? FREAD : FWRITE))) {
	err = EACCES;
	goto done;
    }

    if (fp->f_dp && S_ISDIR (fp->f_dp->d_mode)) {
	err = EISDIR;
	goto done;
    }

    uio.uio_iov = iov;
    uio.uio_iovcnt = iovcnt;
    uio.uio_segflg = UIO_USERSPACE;
    uio.uio_rw = rw;

    /* negative offset means use current file offset */
    if (offs < 0) {
	if (rw == UIO_WRITE && fp->f_flag & FAPPEND && fp->f_dp)
	    uio.uio_offset = fp->f_dp->d_size;
	else
	    uio.uio_offset = fp->f_offset;
    }
    else
	uio.uio_offset = offs;

    /* calculate total length */
    len = 0;
    for (i = 0; i < iovcnt; i++, iov++) {
	if (iov->iov_len > SSIZE_MAX ||
	    (len += iov->iov_len) > SSIZE_MAX) {
	    err = EINVAL;
	    goto done;
	}
    }
    uio.uio_resid = len;
    
    if (rw == UIO_READ)
	err = (*fp->f_ops->fo_read) (fp, &uio);
    else
	err = (*fp->f_ops->fo_write) (fp, &uio);

    /* XXX handle ERESTART here? */

    /* update file offset */
    if (offs < 0)
	fp->f_offset = uio.uio_offset;
    ret = len - uio.uio_resid;

 done:
    if (needfree)
	free (needfree);
    _fileio_frelease (fp);
    return _syscall_leave (err, ret);
}


ssize_t
pwritev (int fd, const struct iovec *iov, int iovcnt, off_t offs)
{
    if (offs < 0) {
	errno = EINVAL;
	return -1;
    }
    return _fileio_doio (fd, iov, iovcnt, UIO_WRITE, offs);
}


ssize_t
writev (int fd, const struct iovec *iov, int iovcnt)
{
    return _fileio_doio (fd, iov, iovcnt, UIO_WRITE, (off_t)-1);
}


ssize_t
pwrite (int fd, const void *buf, size_t len, off_t offs)
{
    struct iovec iov[1];

    if (offs < 0) {
	errno = EINVAL;
	return -1;
    }
    iov[0].iov_base = (void *)buf;
    iov[0].iov_len = len;
    return _fileio_doio (fd, iov, 1, UIO_WRITE, offs);
}


ssize_t
write (int fd, const void *buf, size_t len)
{
    struct iovec iov[1];
    iov[0].iov_base = (void *)buf;
    iov[0].iov_len = len;
    return _fileio_doio (fd, iov, 1, UIO_WRITE, (off_t)-1);
}


ssize_t
preadv (int fd, const struct iovec *iov, int iovcnt, off_t offs)
{
    return _fileio_doio (fd, iov, iovcnt, UIO_READ, offs);
}


ssize_t
readv (int fd, const struct iovec *iov, int iovcnt)
{
    return _fileio_doio (fd, iov, iovcnt, UIO_READ, (off_t)-1);
}


ssize_t
pread (int fd, void *buf, size_t len, off_t offs)
{
    struct iovec iov[1];
    iov[0].iov_base = buf;
    iov[0].iov_len = len;
    return _fileio_doio (fd, iov, 1, UIO_READ, offs);
}


ssize_t
read (int fd, void *buf, size_t len)
{
    struct iovec iov[1];
    iov[0].iov_base = buf;
    iov[0].iov_len = len;
    return _fileio_doio (fd, iov, 1, UIO_READ, (off_t)-1);
}


int 
ioctl (int fd, int cmd, void *args)
{
    file_t *fp;
    int err;

    _syscall_enter ();
    err = _fileio_fget (fd, &fp);
    if (err == 0)
	err = ((*fp->f_ops->fo_ioctl) (fp, cmd, args));
    _fileio_frelease (fp);
    return _syscall_leave (err, 0);
}


int 
fcntl (int fd, int cmd, int arg)
{
    file_t *fp;
    int tmp, ret;
    int err;

    _syscall_enter ();
    
    err = _fileio_fget (fd, &fp);
    if (err)
	goto done;

    switch (cmd) {
    case F_GETFL:
	ret = OFLAGS (fp->f_flag);
	break;

    case F_SETFL:
	ret = 0;
	fp->f_flag &= ~FCNTLFLAGS;
	fp->f_flag |= FFLAGS (arg) & FCNTLFLAGS;
	tmp = fp->f_flag & FNONBLOCK;
	err = (*fp->f_ops->fo_ioctl)(fp, FIONBIO, &tmp);
	if (err)
	    break;	/* failed */
	tmp = fp->f_flag & FASYNC;
	err = (*fp->f_ops->fo_ioctl)(fp, FIOASYNC, &tmp);
	if (err == 0)
	    break;	/* succeeded */
	fp->f_flag &= ~FNONBLOCK;
	tmp = 0;
	(void) (*fp->f_ops->fo_ioctl)(fp, FIONBIO, &tmp);
	break;

    case F_GETFD:
    case F_SETFD:
	/* quietly ignore getfd/setfd for now */
	ret = 0;
	break;

    case F_DUPFD:
	if (arg < 0 || arg >= OPEN_MAX) {
	    err = EBADF;
	    break;
	} 
	err = _fileio_fdalloc(fp, arg, &ret);
	break;

    default:
	err = EINVAL;
	break;
    }

 done:
    _fileio_frelease (fp);
    return _syscall_leave (err, ret);
}


static int
mkstat (file_t *fp, dir_t *dp, struct stat *stb)
{
    int err;

    /* construct default entry */
    memset (stb, 0, sizeof (struct stat));
    if (dp) {
	stb->st_mode = dp->d_mode;
	stb->st_rdev = dp->d_dev;
	stb->st_ino = dp->d_ino;
	stb->st_size = dp->d_size;
	stb->st_nlink = dp->d_nlink;
    }

    /* allow possible file sys override */
    if (fp && fp->f_ops->fo_stat)
	err = ((*fp->f_ops->fo_stat) (dp, fp, stb));
    else if (dp && dp->d_ops->fo_stat)
	err = ((*dp->d_ops->fo_stat) (dp, fp, stb));
    else
	err = 0;
    return err;
}


int
fstat (int fd, struct stat *stb)
{
    file_t *fp;
    int err;

    _syscall_enter ();

    err = _fileio_fget (fd, &fp);
    if (err == 0)
	err = mkstat (fp, fp->f_dp, stb);
    _fileio_frelease (fp);

    return _syscall_leave (err, 0);
}


int 
stat (const char *name, struct stat *stb)
{
    dir_t *dp;
    int err;

    _syscall_enter ();

    err = _fileio_lookup (&dp, name, FREAD, 0);
    if (err == 0) {
	err = mkstat (NULL, dp, stb);
	drelease (dp);
    }

    return _syscall_leave (err, 0);
}


int 
lstat (const char *name, struct stat *stb)
{
    /* at the moment lstat == stat */
    return stat (name, stb);
}


static int
dopathconf (dir_t *dp, int pc, long *retval)
{
    if (!dp)
	return EINVAL;

    if (dp->d_ops->fo_pathconf) {
	int err = ((*dp->d_ops->fo_pathconf) (dp, pc, retval));
	if (err)
	    return err;
    }

    switch (pc) {
    case _PC_CHOWN_RESTRICTED:
	*retval = 1;
	break;
    case _PC_NO_TRUNC:
	*retval = 0;
    case _PC_LINK_MAX:
	*retval = _POSIX_LINK_MAX;
    case _PC_VDISABLE:
	*retval = 1;
    case _PC_MAX_CANON:
	*retval = _POSIX_MAX_CANON;
    case _PC_MAX_INPUT:
	*retval = _POSIX_MAX_INPUT;
    case _PC_NAME_MAX:
	*retval = _POSIX_NAME_MAX;
    case _PC_PATH_MAX:
	*retval = _POSIX_PATH_MAX;
    case _PC_PIPE_BUF:
    case _PC_ASYNC_IO:
    case _PC_PRIO_IO:
    case _PC_SYNC_IO:
	/* not supported */
	*retval = -1;
    default:
	/* unknown request */
	return EINVAL;
   }
   return 0;
}


long
fpathconf (int fd, int pc)
{
    file_t *fp;
    long ret;
    int err;

    _syscall_enter ();

    err = _fileio_fget (fd, &fp);
    if (err == 0)
	err = dopathconf (fp->f_dp, pc, &ret);
    _fileio_frelease (fp);

    return _syscall_leave (err, ret);
}


long
pathconf (const char *name, int pc)
{
    dir_t *dp;
    long ret;
    int err;

    _syscall_enter ();

    err = _fileio_lookup (&dp, name, FREAD, 0);
    if (err == 0) {
	err = dopathconf (dp, pc, &ret);
	drelease (dp);
    }

    return _syscall_leave (err, ret);
}


int
chdir (const char *name)
{
    struct filedesc *fdp;
    dir_t *dp = 0;
    dir_t *odp;
    int err;

    _syscall_enter ();

    err = _fileio_lookup (&dp, name, FREAD, 0);
    if (err)
	goto done;
    
    if (!S_ISDIR (dp->d_mode)) {
	err = ENOTDIR;
	goto done;
    }

    fdp = _fileio_getfdp ();

    /* swap cdir and new dirnode pointers */
    __sdethread_mutex_lock (&files_mx);
    odp = fdp->fd_cdir;
    fdp->fd_cdir = dp;
    __sdethread_mutex_unlock (&files_mx);
    
    if (odp)
	drelease (odp);

 done:
    if (err && dp)
	drelease (dp);
    return _syscall_leave (err, 0);
}


off_t
lseek (int fd, off_t offset, int whence)
{
    file_t *fp;
    off_t ret;
    int err;

    _syscall_enter ();

    err = _fileio_fget (fd, &fp);

    if (err == 0) {
	switch (whence) {
	case SEEK_SET:
	    ret = 0;
	    break;
	case SEEK_END:
	    if (fp->f_dp) {
		ret = fp->f_dp->d_size;
		break;
	    }
	    /* fall through */
	case SEEK_CUR:
	    ret = fp->f_offset;
	    break;
	default:
	    err = EINVAL;
	    goto err;
	}
	ret += offset;
	if (ret < 0)
	    err = EINVAL;
	else
	    fp->f_offset = ret;
    }

err:
    _fileio_frelease (fp);
    return _syscall_leave (err, ret);
}


int
close (int fd)
{
    file_t *fp;
    int err;

    _syscall_enter ();
    
    err = _fileio_fget (fd, &fp);
    if (err == 0)
	/* remove from open file table, but note that
	   _fileio_frelease() actually does the job of telling the
	   file device and freeing the file descriptor */
	_fileio_fdremove (fd);
    _fileio_frelease (fp);

    return _syscall_leave (err, 0);
}


int
dup2 (int from, int to)
{
    file_t *fp;
    int err;

    _syscall_enter ();

    if (to < 0 || to >= OPEN_MAX) {
	err = EBADF;
	goto error;
    }

    err = _fileio_fget (from, &fp);
    if (err == 0 && from != to) {
	struct filedesc *fdp;
	file_t *tfp;

	fdp = _fileio_getfdp ();
	if (to >= fdp->fd_nfiles) {
	    /* need to grow file descriptor table */
	    int i;
	    err = _fileio_fdalloc(fp, to, &i);
	    if (err || i == to)
		/* total failure or success */
		goto done;
	    /* got wrong descriptor, free it drop through */
	    _fileio_fdremove (to);
	}

	/* got enough space in table, just exchange file descriptors */
	__sdethread_mutex_lock (&files_mx);
	tfp = fdp->fd_ofiles[to];
	if (fp != tfp) {
	    fdp->fd_ofiles[to] = fp;
	    fp->f_count++; /* bump file ref count */
	}
	if (to > fdp->fd_lastfile)
	    fdp->fd_lastfile = to;
	__sdethread_mutex_unlock (&files_mx);

	/* release old file (if any) note that we didn't
	   increment its old ref count, so this will free 
	   it if no-one else references it. */
	if (fp != tfp)
	    _fileio_frelease (tfp);
    }
	
 done:
    _fileio_frelease (fp);

 error:
    return _syscall_leave (err, to);
}


int 
dup (int fd)
{
    return fcntl (fd, F_DUPFD, 0);
}


int
getdtablesize ()
{
    return OPEN_MAX;
}


/*
 * Return next non-zero length iovec
 */
int
_uiogetiov(uio, iovp)
    struct uio *uio;
    struct iovec **iovp;
{
    struct iovec *iov;
    
    if (uio->uio_resid == 0)
#ifdef DIAGNOSTIC
	panic("_uiogetiov: zero resid");
#else
        return EINVAL;
#endif
    
     do {
	 iov = uio->uio_iov;
	 if (iov->iov_len > 0) {
	     *iovp = iov;
	     return 0;
	 }
	 uio->uio_iov++;
	 uio->uio_iovcnt--;
     } while (uio->uio_iovcnt != 0);

#ifdef DIAGNOSTIC
    panic("_uiogetiov: bad iovcnt");
#endif
    return EINVAL;
}


/*
 * Advance uio pointers by cnt bytes
 */
int
_uioadvance(len, uio)
    size_t len;
    struct uio *uio;
{
    struct iovec *iov = uio->uio_iov;
    
    if (iov->iov_len < len)
#ifdef DIAGNOSTIC
	panic("uioadvance: small iov");
#else
        return (EINVAL);
#endif
    iov->iov_base += len;
    iov->iov_len -= len;
    uio->uio_resid -= len;
    uio->uio_offset += len;
    return (0);
}


int
_uiomove(cp, n, uio)
    void *cp;
    size_t n;
    struct uio *uio;
{
    struct iovec *iov;
    size_t cnt;
    int error = 0;
    
#ifdef DIAGNOSTIC
    if (uio->uio_rw != UIO_READ && uio->uio_rw != UIO_WRITE)
	panic("uiomove: mode");
#endif
    
    while (n > 0 && uio->uio_resid && error == 0) {
	error = _uiogetiov (uio, &iov);
	if (error)
	    break;
	cnt = iov->iov_len;
	if (cnt > n)
	    cnt = n;
	if (uio->uio_rw == UIO_READ)
	    memcpy (iov->iov_base, cp, cnt);
	else
	    memcpy (cp, iov->iov_base, cnt);
	error = _uioadvance (cnt, uio);
	cp += cnt;
	n -= cnt;
    }
    return error;
}


/*
 * Give next character to user as result of read.
 */
int
_ureadc(c, uio)
    int c;
    struct uio *uio;
{
    struct iovec *iov;
    int error;
    
    error = _uiogetiov (uio, &iov);
    if (error == 0) {
	*(unsigned char *)iov->iov_base = c;
	error = _uioadvance (1, uio);
    }
    return error;
}


/*
 * Get next character from user as result of write.
 */
int
_uwritec(cp, uio)
    int *cp;
    struct uio *uio;
{
    struct iovec *iov;
    int error;
    
    error = _uiogetiov (uio, &iov);
    if (error == 0) {
	*cp = *(unsigned char *)iov->iov_base;
	error = _uioadvance (1, uio);
    }
    return error;
}


void
_scandevs ()
{
    struct filedesc *fdp;
    struct file **fpp;
    file_t *fp;
    int i;

    /* XXX only scans current process */
    fdp = _fileio_getfdp ();

    __sdethread_mutex_lock (&files_mx);
    fpp = fdp->fd_ofiles;
    for (i = fdp->fd_lastfile; i-- >= 0; fpp++)
	if ((fp = *fpp) && fp->f_ops) 
	    (*fp->f_ops->fo_select) (fp, FREAD | FWRITE);
    __sdethread_mutex_unlock (&files_mx);
}

/* pointer to _scandevs function */
__attribute__((common)) void (*__scandevs) (void);

/* dummy versions of select(2) support functions */
WEAK_ALIAS(_selrecord);
WEAK_PROTOTYPE(_selrecord);
void
WEAK_NAME(_selrecord) (struct selinfo *sip)
{
}

WEAK_ALIAS(_selwakeup);
WEAK_PROTOTYPE(_selwakeup);
void
WEAK_NAME(_selwakeup) (struct selinfo *sip)
{
}

/* The posix layer relies on an initialised ttydev layer. We force this
   via an explicit reference to _ttydev_load. */
extern void _ttydev_load (void);
static __attribute((__used__)) void *ttydev_init_dummy = &_ttydev_load;

static void
_posix_init ()
{
    extern dir_t _root;
    struct filedesc *fdp;
    file_t *fp;
    int fd;
    
    /* initialise builtin file system */
    if (_root.d_ops->fo_init)
	(_root.d_ops->fo_init) (&_root);

    /* initialise file descriptor table */
    fdp = &_fileio_fd0.fd_fd;
    fdp->fd_ofiles = _fileio_fd0.fd_dfiles;
    fdp->fd_ofileflags = _fileio_fd0.fd_dfileflags;
    fdp->fd_nfiles = NDFILE;
    fdp->fd_cdir = duse (&_root);
    fdp->fd_rdir = duse (&_root);
    fdp->fd_cmask = 0755;
    fdp->fd_refcnt++;

    /* attach _scandevs function */
    __scandevs = _scandevs;

    /* now mount any extra file systems */
    _call_init_funcs (IF_MOUNT);

    /* build initial free file list */
    for (fp = files; fp < &files[NDFILE]; fp++)
	_fileio_ffree (fp);

    /* open STDIN, STDOUT & STDERR */
    fd = open (_PATH_CONSOLE, O_RDWR, 0);
    if (fd != 0) {
	_mon_puts (_PATH_CONSOLE " open failed\n");
	abort ();
    }

    dup2 (fd, 1);
    dup2 (fd, 2);

    /* change to host's current directory, if there is one */
    (void) chdir ("/cdir");
}

DECL_INIT_FUNC(IF_FILESYS, _posix_init)
