#ident "kit/share/profil.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * profil.c: pc-sampling based profiling 
 */


#include <stddef.h>
#include <assert.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <mips/xcpt.h>
#include <mips/cpu.h>
#include "kit_impl.h"

unsigned long __profil_hertz (void);

#ifdef _SDE_EXT_PROFIL

/* Dummy profiling hooks for cases where the pc-sample histogram is
   collected by some external tool, such as a simulator. */

#pragma weak _sbd_profil
void
_sbd_profil (int start)
{
}


int
profil (char *buf, int size, int offset, int scale)
{
    _sbd_profil (scale != 0);
    return -1;
}


unsigned long
__profil_hertz (void)
{
    return 0;
}

#else

/* A slow, generic, implementation of statistical pc-sample profiling. 
   
   Ideally we would short-circuit the C xcption, intrupt and timer
   code, and implement addupc() in the low-level assembler exception
   handler. But that code is general purpose and doesn't know how to
   handle the various timer devices, so.... */

#ifndef PROFHZ
#define PROFHZ	100		/* default to 100Hz sampling */
#endif

static unsigned short	*pbuf;
static unsigned short	*plim;
static unsigned long 	psize;
static unsigned long    poffset;
static int		pscale;

static void
addupc (struct timeval *dtv, struct xcptcontext *xcp)
{
    register unsigned short *cnt;
    register unsigned long offs;

    /* get offset from base of code and check it's within limits */
    if (pscale != 0 
	&& (offs = ((unsigned long)xcp->epc & ~1) - poffset) < psize) {
	/* scale to sample offset (32x32->64 bit mult) */
	offs = ((unsigned long long)offs * pscale) >> 16;
	/* get pointer to sample bucket for this pc */
	cnt = &pbuf[offs / sizeof(pbuf[0])];
	/* bump counter, unless it would overflow */
	if (*cnt != (unsigned short)0xffff)
	    ++(*cnt);
    }
}


int
profil (char *buf, int size, int offset, int scale)
{
    struct itimerval it;
    int oscale;

    oscale = pscale;
    pscale = 0;
    if (scale != 0 && oscale == 0) {
	/* START TIMER */
	it.it_interval.tv_sec = 0;
	it.it_interval.tv_usec = 1000000/PROFHZ;
	it.it_value = it.it_interval;
	it.it_func = addupc;
	setitimer (ITIMER_PROF, &it, 0);
    }
    else if (scale == 0 && oscale != 0) {
	/* STOP TIMER */
	timerclear (&it.it_value); 
	timerclear (&it.it_interval); 
	setitimer (ITIMER_PROF, &it, 0);
    }
    pbuf = (unsigned short *)buf;
    plim = (unsigned short *)(buf + size);
    psize = size;
    poffset = offset;
    pscale = scale;
    return 0;
}


unsigned long
__profil_hertz (void)
{
    return PROFHZ;
}

#endif
