#ident "kit/share/r4650watch.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * r4650watch.c: R4650 family watchpoint support for MIPS SDE kit
 */


#if #cache(r4k) || #cache(rc32)
#include <sys/types.h>
#include <mips/r4650.h>
#include <mips/r4kxcpt.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <mips/prid.h>
#include <mips/watchpoint.h>

#include "kit_impl.h"

static int	r4650wp_init (unsigned int);
static void	r4650wp_setup (struct mips_watchpoint *);
static int 	r4650wp_hit (const struct xcptcontext *, vaddr_t *, size_t *);
static int	r4650wp_set (struct mips_watchpoint *);
static int	r4650wp_clear (struct mips_watchpoint *);
static void	r4650wp_remove (void);
static void	r4650wp_insert (void);
static void	r4650wp_reset (void);

const struct mips_watchpoint_funcs _r4650_watchpoint_funcs = {
    r4650wp_init,
    r4650wp_setup,
    r4650wp_hit,
    r4650wp_set,
    r4650wp_clear,
    r4650wp_remove,
    r4650wp_insert,
    r4650wp_reset
};


struct r4650wp {
    unsigned long	reg;
    unsigned int	mask;
};


#define IWATCH	0
#define DWATCH	1
static struct r4650wp r4650wp[2];


static int	
r4650wp_init (unsigned int prid)
{
    switch ((prid >> 8) & 0xff) {
    case PRID_R4650:
    case PRID_RC3236X:
	/* r4650 iwatch/dwatch don't care bits */
	r4650wp[IWATCH].mask = 3;
	r4650wp[DWATCH].mask = 7;
	return 2;
    default:
	/* unrecognised */
	return -1;
    }
}

static void
r4650wp_setup (struct mips_watchpoint *wp)
{
    wp->private = &r4650wp[wp->num];
    wp->capabilities =  MIPS_WATCHPOINT_VADDR | MIPS_WATCHPOINT_DWORD;
    if (wp->num == IWATCH)
	wp->capabilities |= MIPS_WATCHPOINT_X;
    else
	wp->capabilities |= MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W;
}


static int
r4650wp_hit (const struct xcptcontext *xcp, vaddr_t *pva, size_t *plen)
{
    unsigned int cr = XCP_CAUSE (xcp);
    int hit = 0;

    if ((cr & CR_XMASK)  == CR_XCPT(EXC_WATCH) 
	&& (r4650wp[IWATCH].reg || r4650wp[DWATCH].reg)) {
	hit = (cr & CR_IW) ? MIPS_WATCHPOINT_X : MIPS_WATCHPOINT_W;
	hit = _mips_watchpoint_address (xcp, hit, pva, plen);
    }

    return hit;
}


static int
r4650wp_set (struct mips_watchpoint *wp)
{
    struct r4650wp *pwp = wp->private;

    if (((wp->va & pwp->mask) + wp->len - 1) > pwp->mask)
	return 0;
    wp->mask = pwp->mask;

    pwp->reg = wp->va & ~pwp->mask;
    if (wp->type & MIPS_WATCHPOINT_X)
	pwp->reg |= IWATCH_EN;
    if (wp->type & MIPS_WATCHPOINT_R)
	pwp->reg |= DWATCH_R;
    if (wp->type & MIPS_WATCHPOINT_W)
	pwp->reg |= DWATCH_W;

    return 1;
}


static int
r4650wp_clear (struct mips_watchpoint *wp)
{
    struct r4650wp *pwp = wp->private;
    pwp->reg = 0;
    return 1;
}


static void	
r4650wp_remove (void)
{
    if (r4650wp[IWATCH].reg || r4650wp[DWATCH].reg) {
	_mips_watchlo = _mips_watchhi = 0;
	r4650_setiwatch (0);
	r4650_setdwatch (0);
    }
}


static void
r4650wp_insert (void)
{
    if (r4650wp[IWATCH].reg || r4650wp[DWATCH].reg) {
	/* exception handler reloads */
	_mips_watchlo = r4650wp[IWATCH].reg;
	_mips_watchhi = r4650wp[DWATCH].reg;
    }
}


static void
r4650wp_reset (void)
{
    r4650wp[IWATCH].reg = r4650wp[DWATCH].reg = 0;
    _mips_watchlo = _mips_watchhi = 0;
    r4650_setiwatch (0);
    r4650_setdwatch (0);
}
#endif
