#ident "kit/share/r4kclock.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * r4kclock.c: generic R4000 timing support functions
 */


/*
 * This file will usually be #include'd into an SBD skeleton file
 * which must define _sbd_pipefreq () which returns the frequency
 * at which this R4000's pipeline runs (i.e. counter frequency * 2).
 */

#include <stddef.h>
#include <sys/time.h>
#include <sys/clock.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <mips/prid.h>
#include <assert.h>

#include "kit_impl.h"
#include "sbd.h"

static clock_t		(*tickfunc)(long, struct xcptcontext *);
static clock_t		tick;
static clock_t		maxtick;
static unsigned long	cycpertick;
static unsigned long	cycpermsec;
static unsigned long	lastcount;

clock_t			_clk_tck;


static int
clkintr (int arg, struct xcptcontext *xcp)
{
    unsigned long nextcount; 
    long cycles, ticks;
    clock_t newtick;

    /* get number of cycles since last tick */
    cycles = mips_getcount () - lastcount;
    if (cycles < cycpertick)
	cycles = cycpertick;	/* !!! paranoia !!! */

    /* round down to number of whole ticks */
    ticks = cycles / cycpertick;
    cycles -= cycles % cycpertick;

    /* note counter value at last tick */
    lastcount += cycles;

    /* call timer tick function */
    newtick = (*tickfunc) (ticks, xcp);
    
    /* update parameters if they have changed */
    if (newtick != tick) {
	tick = newtick;
	cycpertick = tick / 1000 * cycpermsec;
    }

    /* set next tick point */
    mips_setcompare (nextcount = lastcount + cycpertick);
    
    /* There is a race between reading count and writing compare
       whereby we could set compare to a value "below" the new
       clock value.  Check again to avoid an 80 sec. wait */
    cycles = mips_getcount() - nextcount;
    while (cycles > 0) {
	mips_setcompare (nextcount += cycpertick);
	cycles = mips_getcount() - nextcount;
    }

    return 0;
}




void
_sbd_settimer (clock_t newtick)
{
    unsigned long nextcount;
    long cycles;

    assert (newtick <= maxtick);
    if (newtick == tick)
	return;

    tick = newtick;
    cycpertick = tick / 1000 * cycpermsec;

    /* set next tick point */
    mips_setcompare (nextcount = lastcount + cycpertick);
	
    /* make sure we don't wait 80 secs */
    cycles = mips_getcount() - nextcount;
    while (cycles > 0) {
	mips_setcompare (nextcount += cycpertick);
	cycles = mips_getcount() - nextcount;
    }
}


clock_t
_sbd_gettimer ()
{
    /* return how many microseconds have passed since last interrupt */
    if (cycpermsec == 0)
	return 0;
    else {
#if 0
	unsigned long long ticks = mips_getcount () - lastcount;
	ticks = ticks * 1000 / cycpermsec;
	return (clock_t)ticks;
#else
	/* round down to millisecond, avoiding 64-bit arith */
	unsigned long ticks = mips_getcount () - lastcount;
	ticks = ticks / cycpermsec;
	return (clock_t)(ticks * 1000);
#endif
    }
}


/* Default version of _sbd_cyclespercount, can be overridden by
   specific target sbd files. */

long __sbd_cyclespercount (void);

long _sbd_cyclespercount (void) 
  __attribute__ ((weak, alias ("__sbd_cyclespercount")));

long
__sbd_cyclespercount (void)
{
    switch ((mips_getprid () >> 8) & 0xff) {
    case PRID_R4100:
	/* The Vr4100 counter runs at quarter pipeline frequency */
	return 4;
    case PRID_20KC:
    case PRID_25KF:
	/* 20Kc & 20Kf count 1:1 with pipeline clock */
	return 1;
    }

#if (__mips == 32 || __mips == 64) && mips_isa_rev >= 2
    {
	int ratio;
	__asm __volatile ("rdhwr %0,$3"	\
			  : "=d" (ratio));
	if (ratio != 0)
	    return ratio;
    }
#endif

    /* Most other MIPS-based CPUs on this planet count at 1:2
       pipeline clocks. */
    return 2;
}

#ifndef _SBD_CLK_DIV
#define _SBD_CLK_DIV	_sbd_cyclespercount ()
#endif

#ifndef _SBD_CLK_MULT
#define _SBD_CLK_MULT	1
#endif

clock_t
_sbd_inittimer (clock_t (*func)(long, struct xcptcontext *))
{
    if (!_clk_tck) {
	_clk_tck = _sbd_pipefreq ();

	/* convert from pipeline frequency to count register frequency */
	_clk_tck = (_clk_tck * _SBD_CLK_MULT) / _SBD_CLK_DIV;

	/* Initialise on-chip counter interrupt */
#ifdef mips_bicconfig
	switch ((mips_getprid () >> 8) & 0xff) {
	case PRID_20KC:
	case PRID_25KF:
#define M20KC_CFG0_TI		0x01000000
	    /* Wire timer interrupt to IP5 */
	    mips_bicconfig (M20KC_CFG0_TI);
	    break;
	}
#endif
    }

    /* clock cycles per msec (round to nearest) */
    cycpermsec = (_clk_tck + 500) / 1000;
    if (cycpermsec == 0)
	return 0;

    /* max # of usecs per 32-bit count */
    maxtick = 0x7fffffffUL / cycpermsec * 1000;
    
    /* initialise compare register (thereby clearing any interrupt) */
    lastcount = mips_getcount ();
    mips_setcompare (lastcount - 1);
    
    tickfunc = func;
    intrupt (7, clkintr, 0);
    mips_bissr (SR_IM7);	/* enable interrupt now */

    return maxtick;
}
