#ident "kit/share/s2681clk.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * share/s2681clk.c: common timer code using S2681 UART.
 */



#include <stddef.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/clock.h>
#include <sys/signal.h>
#include <sys/termios.h>
#include <sys/fcntl.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <stdlib.h>
#include <assert.h>

#include "sbd.h"
#include "s2681.h"
#include "kit_impl.h"

/* 
 * The timer crystal frequency is S2681XTAL (normally 3.6864Mhz).
 * We prescale this by 16 to give to give a counter input frequency 
 * of 230,400 Hz.  The interrupt happens only every other time the
 * count reaches zero, so we pretend that the input clock is half
 * this again (i.e. 115,200 Hz).
 * 
 * Note that we can't read the timer, so we set the lowest freq
 * to 50Hz, so that clock() etc give at least that resolution.
 * This is also small enough that we can calcuate the counter value
 * accurately without overflow in 32 bits.
 */

#define S2681HZ		(S2681XTAL / 16 / 2)	/* effective input freq */
#define MAXTICK		(1000000 / 50)		/* fake lowest freq = 50 Hz */

static clock_t		(*tickfunc)();
static clock_t		tick;

/* timer resolution in Hz */
clock_t			_clk_tck = 50;

unsigned char		_s2681_acr[2];

#ifndef pre_s2681
#define pre_s2681()	0
#define post_s2681(x)
#endif
#ifndef intenable_s2681
#define intenable_s2681(x)
#endif



void
_sbd_settimer (clock_t newtick)
{
    volatile struct s2681dev *sd = (struct s2681dev *) PA_TO_KVA1(DUART_BASE);
    unsigned long counter, x;

    /* work out how many counter cycles are required for this tick length */
    assert (newtick <= MAXTICK);
    tick = newtick;
    /* MAXTICK * S2681HZ just fits in 32 bits */
    counter = (tick * S2681HZ) / 1000000; 
    assert (counter >= 0x0002 && counter <= 0xffff);

    /* reprogram timer (note this also resets the timer, 
       so don't do it too often). */
    
    x = pre_s2681 ();
    sd->g_ctl = counter;	mips_wbflush ();
    sd->g_ctu = counter >> 8;	mips_wbflush ();
    (void) sd->g_start;
    post_s2681 (x);
}


static int
s2681intr (int arg, struct xcptcontext *xcp)
{
    volatile struct s2681dev *sd = (struct s2681dev *) PA_TO_KVA1(DUART_BASE);
    clock_t newtick;
    unsigned long x;

    x = pre_s2681 ();
    if (!(sd->g_isr & ISR_CT)) {
	post_s2681 (x);
	return 0;
    }

    /* clear timer interrupt latch */
    (void) sd->g_stop;
    post_s2681 (x);
    
    /* call timer tick function and reprogram if necessary */
    newtick = (*tickfunc) (1, xcp);
    if (newtick != tick)
	_sbd_settimer (newtick);

    return 0;
}

clock_t
_sbd_gettimer ()
{
    /* 
     * We can't read the timer because we only get an interuupt
     * every OTHER time that it reaches zero, and we can't tell
     * on which half of the cycle it is!
     */
    return 0;
}


struct s2681acr {
    unsigned short	speed;
    unsigned char	set;
};

static const struct s2681acr s2681acr[] = {
    {50,	0},
    {75,	1},
    {110,	2},
    {134,	2},
    {150,	1},
    {220,	0},
    {300,	2},
    {600,	2},
    {1050,	1},
    {1200,	0},
    {2400, 	2},
    {4800,	2},
    {1800,	1},
    {7200,	0},
    {9600,	2},
    {19200,	1},
    {38400,	0},
    {0}
};


static unsigned int
findset (const char *devname)
{
    const struct s2681acr *acr;
    int speed = 9600;
    struct termios tio;
    int fd; 

    fd = open (devname, O_RDONLY);
    if (fd >= 0) {
	tcgetattr (fd, &tio);
	close (fd);
	speed = cfgetospeed (&tio);
    }
    
    for (acr = s2681acr; acr->speed != speed; acr++) {
	if (acr->speed == 0)
	    abort ();
    }

    return acr->set;
}

    
clock_t
_sbd_inittimer (clock_t (*func)())
{
    volatile struct s2681dev *sd = (struct s2681dev *) PA_TO_KVA1(DUART_BASE);
    unsigned long x;
    unsigned int set0, set1;

    /* try to intuit the setting of the write-only ACR register,
       which depends on the baud rate of both UART channels. */

    set0 = findset ("/dev/tty0");
    set1 = findset ("/dev/tty1");
    if (set0 + set1 == 1)		/* incompatible baud rates */
	abort ();

    x = pre_s2681 ();

    /* XXX this assumes that we're using s2681.s, which maintains
       a soft copy of the write-only ACR register in _s2681_acr.
       If we're running under a monitor and the user changes the baud
       rate under our feet then our ctmode setting will be clobbered 
       (hopefully unlikely) */

    /* set timer mode, with /16 prescaler */
    _s2681_acr[0] &= ~(ACR_CTMODE | ACR_SET2);
    _s2681_acr[0] |= ACR_T_XTAL_16;
    if ((set0 | set1) & 1)
	_s2681_acr[0] |= ACR_SET2;
	
    sd->g_acr = _s2681_acr[0];	mips_wbflush ();

    /* load max counter */
    sd->g_ctl = 0xff;		mips_wbflush ();
    sd->g_ctu = 0xff;		mips_wbflush ();
    (void) sd->g_start;

    /* clear any pending timer interrupt */
    (void) sd->g_stop;

    /* enable counter/timer interrupts */
    sd->g_imr = ISR_CT;	 	mips_wbflush ();

    post_s2681 (x);

    /* any other board-specific enables */
    intenable_s2681 (sd);

    /* install our interrupt handler */
    tickfunc = func;
    intrupt (INTR_DUART, s2681intr, 0);

    return MAXTICK;
}
