#ident "kit/share/sbrk.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * sbrk.c: a generic sbrk() emulation for MIPS SDE
 */


#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sdethread.h>
#include <mips/cpu.h>
#include "kit_impl.h"

/* _minbrk and _maxbrk can be set by startup code, or by a linker
   script, so we don't want them in bss where they'll get cleared, so
   they can't be common, but they must be capable of being
   overridden. */
void *		_minbrk __attribute__((weak)) = 0;
void *		_maxbrk __attribute__((weak)) = 0;

extern int	errno;
extern char     _end[];

static __sdethread_mutex_t sbmx = __SDETHREAD_MUTEX_INITIALIZER("sbrk");
static void *	curbrk = 0;
static void *	curbrk_hi;
static int pagesize = 4096;

int
getpagesize (void)
{
    return pagesize;
}

#ifndef MAXSTACK
#define MAXSTACK	(8 * pagesize)
#endif

static void
adjust_pagesize (ptrdiff_t size) {
    while (((size / 64) < pagesize) && (pagesize > 16))
	pagesize /= 2;
}

/*
 * The _sbd_memlayout() function returns a pointer to a phys memory
 * region table, but note that at present sbrk() only uses the first
 * entry.
 *
 * This function can be overridden by the board-specific code
 * if it has some other way to determine the real size of 
 * physical memory (e.g. reading the memory controller).
 */

#pragma weak _sbd_memlayout=_stub_sbd_memlayout
const struct sbd_region *_stub_sbd_memlayout (void);

const struct sbd_region *
_stub_sbd_memlayout (void)
{
    static struct sbd_region mem[2];
    extern paddr_t _mem_top;

    mem[0].type = SBD_MEM_RAM;
    mem[0].base = 0x00000000;
    if (_mem_top)
	/* _mem_top is usually set by monitor-specific startup code */
	mem[0].size = _mem_top;
    else if (_maxbrk)
	/* if we don't know the memory size, use the end of the heap */
	mem[0].size = KVA_TO_PA (_maxbrk);
    else{
	/* final resort, assume stack points to top of memory */
	void * sp;
	__asm__ ("move %0,$sp" : "=d" (sp));
	mem[0].size = KVA_TO_PA (sp);
    }

    return mem;
}


/* 
 * Initialise the sbrk heap. 
 *
 * This function is hard-wired to the idea that the code is linked to
 * KSEG0 or KSEG1 addresses. It could just about cope with being
 * linked to run in KUSEG, as long as there's a one-to-one mapping
 * from virtual to physical address. If you are playing real virtual
 * memory games then the functions in the module will have to be
 * replaced.
 */

void
_sbrk_init (void)
{
    const struct sbd_region * layout;
    void * minva,  * maxva;
    paddr_t rbase, rtop, min, max;

    if (curbrk)
	return;

    if (_minbrk)
	/* user specified heap start */
	minva = _minbrk;
    else
	/* usually heap starts after data & bss segment */
	minva = _end;

    if (_maxbrk) {
	/* user specified heap top */
	maxva = _maxbrk;
	adjust_pagesize (maxva - minva);
    } else {
	/* usually stack is at top of memory, and
	   heap grows up towards base of stack */
	char * sp;
	__asm__ ("move %0,$sp" : "=d" (sp));
	adjust_pagesize (sp - (char *)minva);
	maxva = sp - MAXSTACK;
    }

    /* convert min/max to physical addresses */
    if (IS_KVA01 (minva))
	min = KVA_TO_PA (minva);
    else
	/* use virtual address */
	min = (paddr_t) minva;

    if (IS_KVA01 (maxva))
	max = KVA_TO_PA (maxva);
    else
	max = (paddr_t) maxva;

    /* determine physical memory layout */
    layout = _sbd_memlayout ();

    /* base of heap must be inside memory region #0 */
    rbase = layout[0].base;
    rtop = rbase + layout[0].size;
    if (min < rbase || min >= rtop) {
	if (rbase >= KVA_TO_PA (_end))
	    /* no overlap of region with data - use region base */
	    min = rbase;
	else
	    /* can't determine a good heap base */
	    /* XXX could try _end in case of bad _minbrk setting */
	    return;
    }

    /* end of heap must be inside memory region #0 (and above base) */
    if (max < min || max >= rtop) {
	if (rtop > min)
	    /* use top of region as top of heap */
	    /* XXX what about poss overlap with stack? */
	    max = rtop;
	else
	    /* can't determine a good heap top */
	    return;
    }

    /* put minbrk/maxbrk in same kernel virtual segment as data */
    if (IS_KVA1 (_end)) {
	/* kseg1: uncached data segment */
	_minbrk = PA_TO_KVA1 (min);
	_maxbrk = PA_TO_KVA1 (max);
    }
    else if (IS_KVA0 (_end)) {
	/* kseg0: cached data segmnt */
	_minbrk = PA_TO_KVA0 (min);
	_maxbrk = PA_TO_KVA0 (max);
    }
    else {
	/* kuseg: use virtual addresses */
	_minbrk = (void *) min;
	_maxbrk = (void *) max;
    }
    
    curbrk = _minbrk;
    curbrk_hi = _maxbrk;
}


long
_sbrk_totalpages ()
{
    const struct sbd_region *layout;
    long total;

    __sdethread_mutex_lock (&sbmx);
    layout = _sbd_memlayout ();
    for (total = 0; layout->type != SBD_MEM_END; ++layout)
	if (layout->type == SBD_MEM_RAM)
	    total += layout->size / pagesize;
    __sdethread_mutex_unlock (&sbmx);

    return total;
}


long
_sbrk_availpages ()
{
    long npg;

    __sdethread_mutex_lock (&sbmx);
    if (!curbrk)
	_sbrk_init ();
    if (!curbrk)
	npg = 0;
    else
	npg = (curbrk_hi - curbrk) / pagesize;
    __sdethread_mutex_unlock (&sbmx);
    return npg;
}

void *
_sbrkhi (int n)
{
    void *newbrk, *p;
    int err = 0;

    __sdethread_mutex_lock (&sbmx);

    if (!curbrk) {
	_sbrk_init ();
	if (!curbrk) {
	    err = ENOMEM;
	    goto done;
	}
    }

    p = curbrk_hi;
    newbrk = curbrk_hi - n;
    if (n > 0) {
	if (newbrk > curbrk_hi || newbrk < curbrk) {
	    err = ENOMEM;
	    goto done;
	}
    } else {
	if (newbrk < curbrk_hi || newbrk > _maxbrk) {
	    err = EINVAL;
	    goto done;
	}
    }
    curbrk_hi = newbrk;

done:
    __sdethread_mutex_unlock (&sbmx);
    if (err) {
	errno = err;
	return (void *)-1;
    }
    return p;
}

void *
_sbrk (int n)
{
    void *newbrk, *p;
    int err = 0;

    __sdethread_mutex_lock (&sbmx);

    if (!curbrk) {
	_sbrk_init ();
	if (!curbrk) {
	    err = ENOMEM;
	    goto done;
	}
    }

    p = curbrk;
    newbrk = curbrk + n;
    if (n > 0) {
	if (newbrk < curbrk || newbrk > curbrk_hi) {
	    err = ENOMEM;
	    goto done;
	}
    } else {
	if (newbrk > curbrk || newbrk < _minbrk) {
	    err = EINVAL;
	    goto done;
	}
    }
    curbrk = newbrk;

done:
    __sdethread_mutex_unlock (&sbmx);
    if (err) {
	errno = err;
	return (void *)-1;
    }
    return p;
}

void *
sbrk (int n)
{
    void *p;

    p = _sbrk(n);

    /* sbrk is defined to return zeroed pages */
    if (n > 0 && p != (void *)-1)
	memset (p, 0, n);

    return p;
}
