#ident "kit/share/timer.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * timer.c: high-resolution timing functions for MIPS SDE
 */


#include <stddef.h>
#include <assert.h>
#include <sys/time.h>
#include <sys/clock.h>
#include <sys/errno.h>
#include <sys/signal.h>
#include <mips/xcpt.h>
#include <mips/cpu.h>
#include "kit_impl.h"
#include <sdethread.h>

/*
 * Bump a timeval by a number of usec's.
 */
#define BUMPTIME(t, usec) { \
	register struct timeval *tp = (t); \
	tp->tv_sec += (usec) / 1000000; \
	tp->tv_usec += (usec) % 1000000; \
	if (tp->tv_usec >= 1000000) { \
		tp->tv_usec -= 1000000; \
		tp->tv_sec++; \
	} \
}

/* current time of day */
static struct timeval curtime;

/* current interval timers */
static struct itimerval itimers[NITIMER];
static int nitimers;

/* board-specific info */
static clock_t tick;	/* current microsecs per tick */
static clock_t maxtick;	/* coarsest resolution (in usecs) */
#define MINTICK	1000	/* finest resolution == .0001 secs == 1000 Hz */

/* the set of resolutions which we support */
static const clock_t resolutions[] = {
    				/* frequency	repeat rate */
    20000000,			/*   0.05Hz	20.0   secs */
    10000000,			/*   0.1 Hz	10.0   secs */
     5000000,			/*   0.2 Hz	 5.0   secs */
     2000000,			/*   0.5 Hz	 2.0   secs */
     1000000,			/*   1   Hz	 1.0   secs */
      500000,			/*   2   Hz	  .5   secs */
      200000,			/*   5   Hz	  .2   secs */
      100000,			/*  10   Hz	  .1   secs */
       50000,			/*  20   Hz	  .05  secs */
       20000,			/*  50   Hz	  .02  secs */
       10000,			/* 100   Hz	  .01  secs */
        5000,			/* 200   Hz	  .005 secs */
        2000,			/* 500   Hz	  .002 secs */
        1000,			/*1000   Hz	  .001 secs */
	   0};		

static const int itimersig[NITIMER] = {
    SIGALRM, SIGVTALRM, SIGPROF, SIGUSR1
};

extern int errno;



/*
 * Check that a proposed value to load into the .it_value or
 * .it_interval part of an interval timer is acceptable, and
 * fix it to have at least minimal value (i.e. if it is less
 * than the resolution of the clock, round it up.)
 */
static int
itimerfix(struct timeval *tv)
{

	if (tv->tv_sec < 0 || tv->tv_sec > 100000000 ||
	    tv->tv_usec < 0 || tv->tv_usec >= 1000000)
		return (EINVAL);
	if (tv->tv_sec == 0 && tv->tv_usec != 0 
	    && tv->tv_usec < MINTICK)
		tv->tv_usec = MINTICK;
	return (0);
}


/*
 * Add and subtract routines for timevals.
 * N.B.: subtract routine doesn't deal with
 * results which are before the beginning,
 * it just gets very confused in this case.
 * Caveat emptor.
 */
static void
timevalfix(struct timeval *t1)
{

	if (t1->tv_usec < 0) {
		t1->tv_sec--;
		t1->tv_usec += 1000000;
	}
	if (t1->tv_usec >= 1000000) {
		t1->tv_sec++;
		t1->tv_usec -= 1000000;
	}
}

static void
timevaladd(struct timeval *t1, struct timeval *t2)
{

	t1->tv_sec += t2->tv_sec;
	t1->tv_usec += t2->tv_usec;
	timevalfix(t1);
}

static void
timevalsub(struct timeval *t1, struct timeval *t2)
{

	t1->tv_sec -= t2->tv_sec;
	t1->tv_usec -= t2->tv_usec;
	timevalfix(t1);
}



static clock_t
clocktick (long ticks, struct xcptcontext *xcp)
{
    struct itimerval *it;
    clock_t usec;

    usec = ticks * tick;
    BUMPTIME (&curtime, usec);

    for (it = itimers; it < &itimers[NITIMER]; it++) {
	if (!timerisset (&it->it_value)
	    || timercmp (&curtime, &it->it_value, <))
	    continue;

	if (it->it_func) {
	    struct timeval dtv = curtime;
	    timevalsub (&dtv, &it->it_value);
	    (*it->it_func) (&dtv, xcp);
	} else {
	    _sig_raise (0, itimersig[it - itimers], xcp);
	}

	if (!timerisset (&it->it_interval)) {
	    /* timer is finished */
	    timerclear (&it->it_value);
	    nitimers--;
	    continue;
	}
	
	/* advance timer to next tick */
	while (1) {
	    timevaladd (&it->it_value, &it->it_interval);
	    if (timercmp (&it->it_value, &curtime, >))
		break;
	}
    }

    if (nitimers == 0)
	tick = maxtick;
    return tick;
}



/* 
 * A function to find the coarsest resolution from the above list 
 * which will satisfy all of the outstanding interval timers. 
 */

static clock_t
bestresolution ()
{
    struct itimerval *it;
    const clock_t *res;
    clock_t request;
    int match;

    if (nitimers == 0)
	return maxtick;

    for (res = resolutions; *res > MINTICK; res++) {
	if (*res > maxtick)
	    continue;

	match = 0;
	for (it = itimers; it < &itimers[NITIMER]; it++) {
	    if (!timerisset (&it->it_value))
		continue;
	    if (timerisset (&it->it_interval)) {
		if (it->it_interval.tv_sec <= 3600)
		    /* up to an hour: be precise */
		    request = it->it_interval.tv_sec * 1000000
			+ it->it_interval.tv_usec;
		else
		    /* sod the accuracy, avoid overflow */
		    request = maxtick;
	    }
	    else
		request = it->it_value.tv_usec;
	    if (*res == request || (*res < request && request % *res == 0))
		match++;
	}

	if (match == nitimers)
	    break;
    }
    
    return *res;
}


static int
timerstart ()
{
    const clock_t *res;

    curtime.tv_usec = 0;
    curtime.tv_sec = _sbd_gettod ();
    if (curtime.tv_sec < 0)	
	/* handle error from _sbd_gettod */
	curtime.tv_sec = 0;

    maxtick = _sbd_inittimer (clocktick);
    if (maxtick == 0)
	return 0;
    for (res = resolutions; *res > maxtick; res++)
	continue;
    assert (*res != 0);
    tick = maxtick = *res;
    _sbd_settimer (tick);
    return 1;
}
    


int
setitimer (unsigned which, struct itimerval *value, struct itimerval *ovalue)
{
    struct itimerval aitv;
    struct itimerval *it;
    int s;

    if (which >= NITIMER) {
	errno = EINVAL;
	return -1;
    }

    /* Return old value of timer, if requested */
    if (ovalue)
	getitimer (which, ovalue);

    if (!value)
	return 0;

    aitv = *value;
    if (itimerfix (&aitv.it_value) || itimerfix (&aitv.it_interval)) {
	errno = EINVAL;
	return -1;
    }
	
    if (which != ITIMER_PROF && which != ITIMER_USER)
	aitv.it_func = 0;	/* compatibility */

    s = __sdethread_intr_disable_all ();

    if (tick == 0) {
	if (!timerstart ()) {
	    errno = EINVAL;
	    return -1;
	}
    }

    it = &itimers[which];
    if (timerisset (&aitv.it_value)) {
	/* Set the timer. */
        if (!timerisset (&it->it_value))
	    nitimers++;
	timevaladd (&aitv.it_value, &curtime);
    }
    else {
	/* Clear the timer. */
        if (timerisset (&it->it_value))
	    nitimers--;
    }
    *it = aitv;

    /* Work out the new tick and program it. */
    tick = bestresolution ();
    _sbd_settimer (tick);

    __sdethread_intr_enable_all ();

    return 0;
}


int
getitimer (unsigned which, struct itimerval *value)
{
    struct itimerval aitv;
    int s;

    if (which >= NITIMER) {
	errno = EINVAL;
	return -1;
    }

    /* get timer, and time since last tick */
    s = __sdethread_intr_disable_all ();

    aitv = itimers[which];
    if (timerisset (&aitv.it_value)) {
	if (timercmp(&aitv.it_value, &curtime, <))
	    timerclear(&aitv.it_value);
	else
	    timevalsub(&aitv.it_value, &curtime);
    }
    __sdethread_intr_setmask (s);

    *value = aitv;
    return 0;
}


/*
 * Return the current "time", as accurate as the hardware will
 * give us.
 */
int
gettimeofday (struct timeval *tvp, struct timezone *tzp)
{
    clock_t usec;

    (void) __sdethread_intr_disable_all ();

    if (tick == 0) {
	if (!timerstart ()) {
	    __sdethread_intr_enable_all ();
	    errno = ENODEV;
	    return -1;
	}
    }

    *tvp = curtime;
    usec = _sbd_gettimer ();
    BUMPTIME (tvp, usec);
    __sdethread_intr_enable_all ();

    if (tzp)
	tzp->tz_minuteswest = tzp->tz_dsttime = 0;

    return 0;
}


/* Dummy _sbd_settod, in case not supported on board */
#pragma weak _sbd_settod=_stub_sbd_settod
void _stub_sbd_settod (time_t);
void _stub_sbd_settod (time_t t)
{
}


int
settimeofday (const struct timeval *tvp, const struct timezone *tzp)
{
    clock_t usec;
    int s;

    /* adjust current wall-clock time */
    s = __sdethread_intr_disable_all ();
    usec = _sbd_gettimer ();
    curtime = *tvp;
    curtime.tv_sec -= usec / 1000000;
    curtime.tv_usec -= usec % 1000000;
    if (curtime.tv_usec < 0) {
	curtime.tv_sec--;
	curtime.tv_usec += 1000000;
    }
    (void) __sdethread_intr_setmask (s);

    /* program seconds part into real-time clock */
    _sbd_settod (tvp->tv_sec);

    return 0;
}

void __sdethread_intr_enable_all (void)
{
    spln (0);
}

int __sdethread_intr_disable_all (void)
{
    return (spln (8));
}

int __sdethread_intr_setmask (int s)
{
    return (splx (s));
}
