/*
 * tty.c: generic tty i/o routines
 *
 * This code is adapted liberally from the NetBSD/BSD4.4 lite kernel,
 * but modified substantially to work in a polled i/o environment.
 * Then further enhanced to work in a Pthreads environment.
 * 
 * Copyright (c) 1996-2002,2005 MIPS Technologies, Inc. All rights reserved.
 *
 * Copyright (c) 1982, 1986, 1990, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 * (c) UNIX System Laboratories, Inc.
 * All or some portions of this file are derived from material licensed
 * to the University of California by American Telephone and Telegraph
 * Co. or Unix System Laboratories, Inc. and are reproduced herein with
 * the permission of UNIX System Laboratories, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)tty.c	8.8 (Berkeley) 1/21/94
 */

#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <termios.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "file.h"
#include "tty.h"
#include "kit_impl.h"

/* Data shared with optional modules must be common or weak extern. */
volatile int	_sig_interrupt_syscall __attribute__ ((common));
int 		_rtpx_sync_tty __attribute__ ((common));

/*
 * Table with character classes and parity. The 8th bit indicates parity,
 * the 7th bit indicates the character is an alphameric or underscore (for
 * ALTWERASE), and the low 6 bits indicate delay type.  If the low 6 bits
 * are 0 then the character needs no special processing on output; classes
 * other than 0 might be translated or (not currently) require delays.
 */
#define	E	0x00	/* Even parity. */
#define	O	0x80	/* Odd parity. */
#define	PARITY(c)	(char_type[c] & O)

#define	ALPHA	0x40	/* Alpha or underscore. */
#define	ISALPHA(c)	(char_type[(c) & TTY_CHARMASK] & ALPHA)

#define	CCLASSMASK	0x3f
#define	CCLASS(c)	(char_type[c] & CCLASSMASK)

#define	BS	BACKSPACE
#define	CC	CONTROL
#define	CR	RETURN
#define	NA	ORDINARY | ALPHA
#define	NL	NEWLINE
#define	NO	ORDINARY
#define	TB	TAB
#define	VT	VTAB

static const unsigned char char_type[] = {
	E|CC, O|CC, O|CC, E|CC, O|CC, E|CC, E|CC, O|CC,	/* nul - bel */
	O|BS, E|TB, E|NL, O|CC, E|VT, O|CR, O|CC, E|CC, /* bs - si */
	O|CC, E|CC, E|CC, O|CC, E|CC, O|CC, O|CC, E|CC, /* dle - etb */
	E|CC, O|CC, O|CC, E|CC, O|CC, E|CC, E|CC, O|CC, /* can - us */
	O|NO, E|NO, E|NO, O|NO, E|NO, O|NO, O|NO, E|NO, /* sp - ' */
	E|NO, O|NO, O|NO, E|NO, O|NO, E|NO, E|NO, O|NO, /* ( - / */
	E|NA, O|NA, O|NA, E|NA, O|NA, E|NA, E|NA, O|NA, /* 0 - 7 */
	O|NA, E|NA, E|NO, O|NO, E|NO, O|NO, O|NO, E|NO, /* 8 - ? */
	O|NO, E|NA, E|NA, O|NA, E|NA, O|NA, O|NA, E|NA, /* @ - G */
	E|NA, O|NA, O|NA, E|NA, O|NA, E|NA, E|NA, O|NA, /* H - O */
	E|NA, O|NA, O|NA, E|NA, O|NA, E|NA, E|NA, O|NA, /* P - W */
	O|NA, E|NA, E|NA, O|NO, E|NO, O|NO, O|NO, O|NA, /* X - _ */
	E|NO, O|NA, O|NA, E|NA, O|NA, E|NA, E|NA, O|NA, /* ` - g */
	O|NA, E|NA, E|NA, O|NA, E|NA, O|NA, O|NA, E|NA, /* h - o */
	O|NA, E|NA, E|NA, O|NA, E|NA, O|NA, O|NA, E|NA, /* p - w */
	E|NA, O|NA, O|NA, E|NO, O|NO, E|NO, E|NO, O|CC, /* x - del */
	/*
	 * Meta chars; should be settable per character set;
	 * for now, treat them all as normal characters.
	 */
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
	NA,   NA,   NA,   NA,   NA,   NA,   NA,   NA,
};
#undef	BS
#undef	CC
#undef	CR
#undef	NA
#undef	NL
#undef	NO
#undef	TB
#undef	VT

/* Macros to clear/set/test flags. */
#define	SET(t, f)	(t) |= (f)
#define	CLR(t, f)	(t) &= ~(f)
#define	ISSET(t, f)	((t) & (f))



/*
 * Character list handling.  For SDE we just use fixed size
 * circular buffers of bytes and bits.  The circular byte buffer
 * holds the character, the circular bit buffer indicates that the
 * corresponding character is "quoted".
 */

static int
clinit (struct clist *clp, unsigned char *buf, unsigned int size)
{
    /* the quote bitmap lies immediately after the byte buffer */
    clp->c_buf = buf;
    clp->c_quote = &buf[size];
    clp->c_first = clp->c_cc = 0;
    clp->c_limit = size - 1;
    return 0;
}


static int
clalloc (struct clist *clp, unsigned int size)
{
    unsigned char *buf;

    /* must be power of two */
    if (size & (size - 1)) 
	abort ();

    buf = malloc (CLBUFSIZE (size));
    if (!buf)
	return -1;
    return clinit (clp, buf, size);
}


static void
clfree (struct clist *clp)
{
    if (clp->c_buf) {
	free (clp->c_buf);
	clp->c_buf = 0;
    }
}


#define isset(b, i)	((b)[(i) >> 3] & (1 << ((i) & 7)))
#define setbit(b, i)	(b)[(i) >> 3] |= (1 << ((i) & 7))
#define clrbit(b, i)	(b)[(i) >> 3] &= ~(1 << ((i) & 7))


#define getc _ttygetc    
int
getc (struct clist *cl)
{
    int c;

    if (cl->c_cc == 0)
	return -1;

    c = cl->c_buf[cl->c_first];
    if (isset (cl->c_quote, cl->c_first))
	c |= TTY_QUOTE;
    cl->c_first = (cl->c_first + 1) & cl->c_limit;
    cl->c_cc--;
    return (c);
}

#define getcblk _ttygetcblk
int
getcblk (struct clist *cl, char **bp)
{
    int cc;

    if (cl->c_cc == 0)
	return -1;
    *bp = (char *)&cl->c_buf[cl->c_first];
    cc = cl->c_limit + 1 - cl->c_first;
    if (cc > cl->c_cc)
	cc = cl->c_cc;
    return (cc);
}

#define delcblk _ttydelcblk
void
delcblk (struct clist *cl, int nc)
{
    cl->c_first = (cl->c_first + nc) & cl->c_limit;
    cl->c_cc -= nc;
}


#define putc _ttyputc
int 
putc (int c, struct clist *cl)
{
    unsigned int index;

    if (cl->c_cc >= cl->c_limit)
	return -1;
    
    index = (cl->c_first + cl->c_cc) & cl->c_limit;
    cl->c_buf[index] = c;
    if (c & TTY_QUOTE)
	setbit (cl->c_quote, index);
    else
	clrbit (cl->c_quote, index);
    cl->c_cc++;
    return 0;
}

#define unputc _ttyunputc
int 
unputc (struct clist *cl)
{
    unsigned int index;
    int c;

    if (cl->c_cc == 0)
	return -1;

    cl->c_cc--;
    index = (cl->c_first + cl->c_cc) & cl->c_limit;
    c = cl->c_buf[index];
    if (isset (cl->c_quote, index))
	c |= TTY_QUOTE;
    return c;
}


#define flushq _ttyflushq
void
flushq (struct clist *cl)
{
    cl->c_cc = 0;
}


/*
 * Put the chars in the from queue on the end of the to queue.
 */
static void
catq (struct clist *from, struct clist *to)
{
    int c;
    
    while ((c = getc(from)) >= 0)
	putc(c, to);
}


/*
 * Return the index of the first character in the list or -1 if none.
 *
 * Callers must not allow getc's to happen between firstc's and getc's
 * so that the pointer becomes invalid.  Note that interrupts are NOT
 * masked.
 *
 * *c is set to the FIRST character
 */

static int
firstc(struct clist *clp, int *c)
{
    if (clp->c_cc == 0)
	return -1;
    *c = clp->c_buf[clp->c_first];
    if (isset(clp->c_quote, clp->c_first))
	*c |= TTY_QUOTE;
    return clp->c_first;
}


/*
 * Given an index into the clist return the index of the next character.
 * to the next character in the list or return NULL if no more chars.
 * 
 * *c is set to the next character.
 */
static int
nextc(struct clist *clp, int index, int *c)
{
    index = (index + 1) & clp->c_limit;
    if (index == ((clp->c_first + clp->c_cc) & clp->c_limit))
	return -1;
    
    *c = clp->c_buf[index];
    if (isset(clp->c_quote, index))
	*c |= TTY_QUOTE;
    return index;
}
    


#define ttylock _ttylock
void
ttylock (struct tty *tp)
{
    if (__sdethread_mutex_lock (&tp->t_mx))
	abort();
}


#define ttyunlock _ttyunlock
int
ttyunlock (struct tty *tp, int err)
{
    if (__sdethread_mutex_unlock (&tp->t_mx))
	abort();
    return err;
}


/*
 * Start moving characters from the output queue to the device.
 */

static int
ttstart (struct tty *tp)
{
    char *bp;
    int c, nc;

    /* interrupt driven output - just kick off */
    if (tp->t_start) {
	(tp->t_start) (tp);
	return 0;
    }

    if (!ISSET (tp->t_state, TS_BUSY)) {
	/* polled output - loop while tx ready */
	SET (tp->t_state, TS_BUSY);
	while ((*tp->t_txrdy) (tp)) {
	    if (tp->t_state & TS_XON_PEND) {
		tp->t_state &= ~TS_XON_PEND;
		(*tp->t_tx) (tp, tp->t_cc[VSTART]);
	    }
	    else if (tp->t_state & TS_XOFF_PEND) {
		tp->t_state &= ~TS_XOFF_PEND;
		(*tp->t_tx) (tp, tp->t_cc[VSTOP]);
	    }
	    else {
		if (ISSET (tp->t_state, TS_TTSTOP))
		    break;
		if (tp->t_txblk) {
		    if ((nc = getcblk (&tp->t_outq, &bp)) < 0)
			break;
		    nc = (*tp->t_txblk) (tp, bp, nc);
		    if (nc > 0)
			delcblk (&tp->t_outq, nc);
		}
		else {
		    if ((c = getc (&tp->t_outq)) < 0)
			break;
		    (*tp->t_tx) (tp, c);
		}
	    }
	}
	CLR (tp->t_state, TS_BUSY);
    }

    return 0;
}



#define ttsetwater _ttsetwater
void
ttsetwater (struct tty *tp)
{
    /* output queue low and high water marks */
    /* leave enough room for tab expansion */
    tp->t_hiwat = tp->t_outq.c_limit - 9;
    tp->t_lowat = tp->t_hiwat / 5;
    if (tp->t_lowat < 32)
	tp->t_lowat = 32;

    /* input queue limit (at which we junk input!) */
    tp->t_hog = tp->t_rawq.c_limit - 1;
}


/*
 * Send stop character on input overflow.
 */
static void
ttyblock(struct tty *tp)
{
    register int total;

    total = tp->t_rawq.c_cc + tp->t_canq.c_cc;
#if 0
    if (tp->t_rawq.c_cc > tp->t_hog) {
	/* XXX BSD semantics, but is this sensible!! */
	/* other end is ignoring XOFF, junk all input in raw buffer */
	ttyflush(tp, FREAD | FWRITE);
	CLR(tp->t_state, TS_TBLOCK);
    }
#endif
    /*
     * Block further input iff: current input > threshold
     * AND input is available to user program.
     */
    if (total >= tp->t_hog * 3 / 4 &&
	!ISSET(tp->t_state, TS_TBLOCK) &&
	(!ISSET(tp->t_lflag, ICANON) || tp->t_canq.c_cc > 0)) {
	if (ISSET(tp->t_iflag, IXOFF) &&
	    tp->t_cc[VSTOP] != _POSIX_VDISABLE) {
	    SET(tp->t_state, TS_TBLOCK|TS_XOFF_PEND);
	    CLR(tp->t_state, TS_XON_PEND);
	    ttstart(tp);
	}
	if (ISSET(tp->t_cflag, CRTS_IFLOW) && tp->t_rxflow) {
	    SET(tp->t_state, TS_TBLOCK);
	    (*tp->t_rxflow) (tp, 0);
	}
    }
}


static void
ttyunblock(struct tty *tp)
{
    if (ISSET(tp->t_iflag, IXOFF) && tp->t_cc[VSTART] != _POSIX_VDISABLE) {
	SET(tp->t_state, TS_XON_PEND);
	CLR(tp->t_state, TS_TBLOCK|TS_XOFF_PEND);
	ttstart(tp);
    }
    if (ISSET(tp->t_cflag, CRTS_IFLOW) && tp->t_rxflow) {
	CLR(tp->t_state, TS_TBLOCK);
	(*tp->t_rxflow) (tp, 1);
    }
}



/*
 * Wait for something to occur from the tty
 */
#define ttysleep _ttysleep
int
ttysleep (struct tty *tp, int waitflags, int vtime)
{
    int ret = 0;

    if (tp->t_wait) {
	/* interrupt based wait */
	if (vtime != 0)
	    ret = (tp->t_wait) (tp, waitflags, vtime);
    }
    else {
	/* polled wait */
	if (vtime != 0)
	    SET (tp->t_state, waitflags);
	(void) ttyunlock(tp, 0);
	_scandevs();
	ttylock(tp);
	if (_sig_interrupt_syscall)
	    ret = EINTR;
    }
    return ret;
}


/*
 * Wait for output to drain
 */
#define ttywait _ttywait
int
ttywait (struct tty *tp)
{
    int ret = 0;

    while (tp->t_outq.c_cc
	   && (ISSET(tp->t_state, TS_CARR_ON) || ISSET(tp->t_cflag, CLOCAL))
	   && ret == 0)
	ret = ttysleep (tp, TS_TXDRAIN, -1);
    return ret;
}
	

#define ttwakeup _ttwakeup
void
ttwakeup (struct tty *tp, int rw)
{
    if (tp->t_wakeup)
	(*tp->t_wakeup) (tp, rw);
    else {
	if (rw & FREAD)
	    CLR (tp->t_state, TS_RXWAIT);
	if (rw & FWRITE)
	    CLR (tp->t_state, TS_TXWAIT | TS_TXDRAIN);
    }
    if (rw & FREAD)
	_selwakeup (&tp->t_rsel);
    if (rw & FWRITE)
	_selwakeup (&tp->t_wsel);
    if (ISSET(tp->t_state, TS_ASYNC))
	_sig_raise (tp->t_pgrp, SIGIO, NULL);
}


/*
 * Flush tty read and/or write queues
 */
#define ttyflush _ttyflush
void
ttyflush(struct tty *tp, int rw)
{
    if (rw & FREAD) {
	flushq(&tp->t_canq);
	flushq(&tp->t_rawq);
	tp->t_rocount = 0;
	tp->t_rocol = 0;
	CLR(tp->t_state, TS_LOCAL);
    }
    if (rw & FWRITE) {
	CLR(tp->t_state, TS_TTSTOP);
	if (tp->t_txstop)
	    (*tp->t_txstop) (tp, 1);
	flushq(&tp->t_outq);
    }
    ttwakeup (tp, rw);

    /*
     * If we flushed the input queue after blocking reception,
     * unblock now.
     */
    if (ISSET(tp->t_state, TS_TBLOCK) && rw & FREAD)
	ttyunblock(tp);
}



/*
 * Output a single character on a tty, doing output processing
 * as needed (expanding tabs, newline processing, etc.).
 * Returns < 0 if succeeds, otherwise returns char to resend.
 * Must be recursive.
 */
static int
ttyoutput (int c, struct tty *tp)
{
    int col, notout;
    long oflag = tp->t_oflag; 
    
    if (!ISSET(oflag, OPOST)) {
	if (ISSET(tp->t_lflag, FLUSHO))
	    return (-1);
	if (putc(c, &tp->t_outq))
	    return (c);
	return (-1);
    }

    /*
     * Do tab expansion if OXTABS is set.  Special case if we external
     * processing, we don't do the tab expansion because we'll probably
     * get it wrong.  If tab expansion needs to be done, let it happen
     * externally.
     */
    CLR(c, ~TTY_CHARMASK);
    if (c == '\t' &&
	    ISSET(oflag, OXTABS) && !ISSET(tp->t_lflag, EXTPROC)) {
	c = 8 - (tp->t_column & 7);
	if (ISSET(tp->t_lflag, FLUSHO)) {
	    notout = 0;
	} else {
	    for (notout = c; notout != 0; notout--)
		if (putc (' ', &tp->t_outq) < 0)
		    break;
	}
	tp->t_column += c - notout;
	return (notout ? '\t' : -1);
    }

    if (c == CEOT && ISSET(oflag, ONOEOT))
	return (-1);

    /*
     * Newline translation: if ONLCR is set,
     * translate newline into "\r\n".
     */
    if (c == '\n' && ISSET(tp->t_oflag, ONLCR)) {
	if (putc('\r', &tp->t_outq))
	    return (c);
    }

    if (!ISSET(tp->t_lflag, FLUSHO) && putc(c, &tp->t_outq))
	return (c);

    col = tp->t_column;
    switch (CCLASS(c)) {
    case BACKSPACE:
	if (col > 0)
	    --col;
	break;
    case CONTROL:
	break;
    case NEWLINE:
    case RETURN:
	col = 0;
	break;
    case ORDINARY:
	++col;
	break;
    case TAB:
	col = (col + 8) & ~7;
	break;
    }
    tp->t_column = col;
    return (-1);
}


/*
 * Echo a typed character to the terminal.
 */
static void
ttyecho(int c, struct tty *tp)
{

    int echoc;

    if (!ISSET(tp->t_state, TS_CNTTB))
	CLR(tp->t_lflag, FLUSHO);

    if ((!ISSET(tp->t_lflag, ECHO) &&
	 (!ISSET(tp->t_lflag, ECHONL) || c == '\n')) ||
	ISSET(tp->t_lflag, EXTPROC))
	return;

    echoc = c & TTY_CHARMASK;
    if (ISSET(tp->t_lflag, ECHOCTL) &&
	(echoc <= 037 && c != '\t' && c != '\n' || echoc == 0177)) {
	(void)ttyoutput('^', tp);
	if (echoc == 0177)
	    echoc = '?';
	else
	    echoc += 'A' - 1;
    }
    (void)ttyoutput(echoc, tp);
}


/*
 * ttyretype --
 *	Reprint the rawq line.  Note, it is assumed that c_cc has already
 *	been checked.
 */
static void
ttyretype(struct tty *tp)
{
	register int i;
	int c;

	/* Echo the reprint character. */
	if (tp->t_cc[VREPRINT] != _POSIX_VDISABLE)
		ttyecho(tp->t_cc[VREPRINT], tp);

	(void)ttyoutput('\n', tp);

	for (i = firstc(&tp->t_canq, &c); i >= 0; i = nextc(&tp->t_canq, i, &c))
		ttyecho(c, tp);
	for (i = firstc(&tp->t_rawq, &c); i >= 0; i = nextc(&tp->t_rawq, i, &c))
		ttyecho(c, tp);

	tp->t_rocount = tp->t_rawq.c_cc;
	tp->t_rocol = 0;
}


/*
 * Back over cnt characters, erasing them.
 */
static void
ttyrubo(struct tty *tp, int cnt)
{

	while (cnt-- > 0) {
		(void)ttyoutput('\b', tp);
		(void)ttyoutput(' ', tp);
		(void)ttyoutput('\b', tp);
	}
}


/*
 * Rubout one character from the rawq of tp
 * as cleanly as possible.
 */
static void
ttyrub(int c, struct tty *tp)
{
    int savecol;
    int tabc;
    int i;

    if (!ISSET(tp->t_lflag, ECHO) || ISSET(tp->t_lflag, EXTPROC))
	return;
    CLR(tp->t_lflag, FLUSHO);
    if (ISSET(tp->t_lflag, ECHOE)) {
	if (tp->t_rocount == 0) {
	    /*
	     * Screwed by ttwrite; retype
	     */
	    ttyretype(tp);
	    return;
	}
	if (c == ('\t' | TTY_QUOTE) || c == ('\n' | TTY_QUOTE))
	    ttyrubo(tp, 2);
	else {
	    CLR(c, ~TTY_CHARMASK);
	    switch (CCLASS(c)) {
	    case ORDINARY:
		ttyrubo(tp, 1);
		break;
	    case BACKSPACE:
	    case CONTROL:
	    case NEWLINE:
	    case RETURN:
	    case VTAB:
		if (ISSET(tp->t_lflag, ECHOCTL))
		    ttyrubo(tp, 2);
		break;
	    case TAB:
		if (tp->t_rocount < tp->t_rawq.c_cc) {
		    ttyretype(tp);
		    return;
		}
		savecol = tp->t_column;
		SET(tp->t_state, TS_CNTTB);
		SET(tp->t_lflag, FLUSHO);
		tp->t_column = tp->t_rocol;
		for (i = firstc(&tp->t_rawq, &c); i >= 0;
		     i = nextc(&tp->t_rawq, i, &tabc))
		    ttyecho(tabc, tp);
		CLR(tp->t_lflag, FLUSHO);
		CLR(tp->t_state, TS_CNTTB);

		/* savecol will now be length of the tab. */
		savecol -= tp->t_column;
		tp->t_column += savecol;
		if (savecol > 8)
		    savecol = 8; /* overflow screw */
		while (--savecol >= 0)
		    (void)ttyoutput('\b', tp);
		break;
	    default:		/* XXX */
		abort ();
	    }
	}
    } else
	ttyecho(tp->t_cc[VERASE], tp);
    --tp->t_rocount;
}



#define ttyinput _ttyinput
int _ttyinput(int, struct tty *);

/* Is 'c' a line delimiter ("break" character)? */
#define	TTBREAKC(c)							\
	((c) == '\n' || ((c) == cc[VEOF] ||				\
	(c) == cc[VEOL] || (c) == cc[VEOL2]) && (c) != _POSIX_VDISABLE)


/*
 * Reinput pending characters after state switch
 */
static void
ttypend(struct tty *tp)
{
    struct clist tq;
    register int c;
    
    CLR(tp->t_lflag, PENDIN);
    SET(tp->t_state, TS_TYPEN);
    tq = tp->t_rawq;
    tp->t_rawq.c_first = (tp->t_rawq.c_first + tp->t_rawq.c_cc)
	& tp->t_rawq.c_limit;
    tp->t_rawq.c_cc = 0;
    while ((c = getc(&tq)) >= 0)
	ttyinput(c, tp);
    CLR(tp->t_state, TS_TYPEN);
}


/*
 * Process input of a single character received on a tty.
 */
int
_ttyinput(int c, struct tty *tp)
{
    int iflag, lflag;
    unsigned char *cc;
    unsigned int err;
    int i;

    /*
     * If input is pending take it first.
     */
    lflag = tp->t_lflag;
    if (ISSET(lflag, PENDIN))
	ttypend(tp);

    CLR(c, TTY_QUOTE);

    cc = tp->t_cc;
    iflag = tp->t_iflag;

    /* Handle exceptional conditions (break, parity, framing). */
    if (err = (ISSET(c, TTY_ERRORMASK))) {
	CLR(c, TTY_ERRORMASK);
	if (ISSET(err, TTY_FE) && !c) {	/* Break. */
	    if (ISSET(iflag, IGNBRK))
		goto endcase;
	    else if (ISSET(iflag, BRKINT) &&
		     ISSET(lflag, ISIG) &&
		     (cc[VINTR] != _POSIX_VDISABLE))
		c = cc[VINTR];
	    else if (ISSET(iflag, PARMRK))
		goto parmrk;
	} else if (ISSET(err, TTY_PE) &&
		   ISSET(iflag, INPCK) || ISSET(err, TTY_FE)) {
	    if (ISSET(iflag, IGNPAR))
		goto endcase;
	    else if (ISSET(iflag, PARMRK)) {
	    parmrk:
		(void)putc(0377 | TTY_QUOTE, &tp->t_rawq);
		(void)putc(0 | TTY_QUOTE, &tp->t_rawq);
		(void)putc(c | TTY_QUOTE, &tp->t_rawq);
		goto endcase;
	    } else
		c = 0;
	}
    } else if (c == 0377 &&
	       (ISSET(iflag, INPCK|PARMRK|IGNPAR|ISTRIP)) == (INPCK|PARMRK))
	if (putc(0377|TTY_QUOTE, &tp->t_rawq) < 0)
	    return 0;
    
    /*
     * In tandem mode, check high water mark.
     */
    if (ISSET(iflag, IXOFF))
	ttyblock(tp);

    if (!ISSET(tp->t_state, TS_TYPEN) && ISSET(iflag, ISTRIP))
	CLR(c, 0x80);
    
    if (!ISSET(lflag, EXTPROC)) {
	/*
	 * Check for literal nexting very first
	 */
	if (ISSET(tp->t_state, TS_LNCH)) {
	    SET(c, TTY_QUOTE);
	    CLR(tp->t_state, TS_LNCH);
	}

	/*
	 * Scan for special characters.  This code
	 * is really just a big case statement with
	 * non-constant cases.  The bottom of the
	 * case statement is labeled ``endcase'', so goto
	 * it after a case match, or similar.
	 */
	
	/*
	 * Control chars which aren't controlled
	 * by ICANON, ISIG, or IXON.
	 */
	if (ISSET(lflag, IEXTEN)) {
	    if (CCEQ(cc[VLNEXT], c)) {
		if (ISSET(lflag, ECHO)) {
		    if (ISSET(lflag, ECHOE)) {
			(void)ttyoutput('^', tp);
			(void)ttyoutput('\b', tp);
		    } else
			ttyecho(c, tp);
		}
		SET(tp->t_state, TS_LNCH);
		goto endcase;
	    }
	    
	    if (CCEQ(cc[VDISCARD], c)) {
		if (ISSET(lflag, FLUSHO))
		    CLR(tp->t_lflag, FLUSHO);
		else {
		    ttyflush(tp, FWRITE);
		    ttyecho(c, tp);
		    if (tp->t_rawq.c_cc + tp->t_canq.c_cc)
			ttyretype(tp);
		    SET(tp->t_lflag, FLUSHO);
		}
		goto startoutput;
	    }
	}
	
	/*
	 * Signals.
	 */
	if (ISSET(lflag, ISIG)) {
	    if (CCEQ(cc[VINTR], c) || CCEQ(cc[VQUIT], c)
		|| CCEQ(cc[VSUSP], c)) {
		if (!ISSET(lflag, NOFLSH))
		    ttyflush(tp, FREAD | FWRITE);
		else
		    ttwakeup(tp, FREAD | FWRITE);
		ttyecho(c, tp);
		_sig_raise (tp->t_pgrp,
			    CCEQ(cc[VINTR], c) ? SIGINT  
			    : CCEQ(cc[VQUIT], c) ? SIGQUIT
			    : SIGTSTP,
			    NULL);
		goto endcase;
	    }
	}
	
	/*
	 * Handle start/stop characters.
	 */
	if (ISSET(iflag, IXON)) {
	    if (CCEQ(cc[VSTOP], c)) {
		if (!ISSET(tp->t_state, TS_TTSTOP)) {
		    SET(tp->t_state, TS_TTSTOP);
		    if (tp->t_txstop) 
			(*tp->t_txstop) (tp, 0);
		    return (0);
		}
		if (!CCEQ(cc[VSTART], c))
		    return (0);
		/*
		 * if VSTART == VSTOP then toggle
		 */
		goto endcase;
	    }
	    if (CCEQ(cc[VSTART], c))
		goto restartoutput;
	}

	/*
	 * IGNCR, ICRNL, & INLCR
	 */
	if (c == '\r') {
	    if (ISSET(iflag, IGNCR))
		goto endcase;
	    else if (ISSET(iflag, ICRNL))
		c = '\n';
	} else if (c == '\n' && ISSET(iflag, INLCR))
	    c = '\r';
    }
    if (!ISSET(tp->t_lflag, EXTPROC) && ISSET(lflag, ICANON)) {
	/*
	 * From here on down canonical mode character
	 * processing takes place.
	 */

	/*
	 * erase (^H / ^?)
	 */
	if (CCEQ(cc[VERASE], c)) {
	    if (tp->t_rawq.c_cc)
		ttyrub(unputc(&tp->t_rawq), tp);
	    goto endcase;
	}

	/*
	 * kill (^U)
	 */
	if (CCEQ(cc[VKILL], c)) {
	    if (ISSET(lflag, ECHOKE) && tp->t_rawq.c_cc == tp->t_rocount)
		while (tp->t_rawq.c_cc)
		    ttyrub(unputc(&tp->t_rawq), tp);
	    else {
		ttyecho(c, tp);
		if (ISSET(lflag, ECHOK) ||
		    ISSET(lflag, ECHOKE))
		    ttyecho('\n', tp);
		flushq(&tp->t_rawq);
		tp->t_rocount = 0;
	    }
	    CLR(tp->t_state, TS_LOCAL);
	    goto endcase;
	}

	/*
	 * word erase (^W)
	 */
	if (CCEQ(cc[VWERASE], c)) {
	    int alt = ISSET(lflag, ALTWERASE);
	    int ctype;

	    /*
	     * erase whitespace
	     */
	    while ((c = unputc(&tp->t_rawq)) == ' ' || c == '\t')
		ttyrub(c, tp);
	    if (c == -1)
		goto endcase;
	    /*
	     * erase last char of word and remember the
	     * next chars type (for ALTWERASE)
	     */
	    ttyrub(c, tp);
	    c = unputc(&tp->t_rawq);
	    if (c == -1)
		goto endcase;
	    if (c == ' ' || c == '\t') {
		(void)putc(c, &tp->t_rawq);
		goto endcase;
	    }
	    ctype = ISALPHA(c);
	    /*
	     * erase rest of word
	     */
	    do {
		ttyrub(c, tp);
		c = unputc(&tp->t_rawq);
		if (c == -1)
		    goto endcase;
	    } while (c != ' ' && c != '\t' &&
		     (alt == 0 || ISALPHA(c) == ctype));
	    (void)putc(c, &tp->t_rawq);
	    goto endcase;
	}

	/*
	 * reprint line (^R)
	 */
	if (CCEQ(cc[VREPRINT], c)) {
	    ttyretype(tp);
	    goto endcase;
	}
    }

    /*
     * Check for input buffer overflow
     */
    if (tp->t_rawq.c_cc + tp->t_canq.c_cc >= tp->t_hog) {
	if (ISSET(iflag, IMAXBEL)) {
	    if (tp->t_outq.c_cc < tp->t_hiwat)
		(void)ttyoutput(CTRL('g'), tp);
	} else
	    ttyflush(tp, FREAD | FWRITE);
	goto endcase;
    }

    /*
     * Put data char in q for user and
     * wakeup on seeing a line delimiter.
     */
    if (putc(c, &tp->t_rawq) >= 0) {
	if (!ISSET(lflag, ICANON)) {
	    ttwakeup(tp, FREAD);
	    ttyecho(c, tp);
	    goto endcase;
	}
	if (TTBREAKC(c)) {
	    tp->t_rocount = 0;
	    catq(&tp->t_rawq, &tp->t_canq);
	    ttwakeup(tp, FREAD);
	} else if (tp->t_rocount++ == 0)
	    tp->t_rocol = tp->t_column;

	ttyecho(c, tp);
	if (CCEQ(cc[VEOF], c) && ISSET(lflag, ECHO)) {
	    /*
	     * Place the cursor over the '^' of the ^D.
	     */
	    i = ISSET(lflag, ECHOCTL) ? 2 : 1;
	    while (i-- > 0)
		(void)ttyoutput('\b', tp);
	}
    }

 endcase:
    /*
     * IXANY means allow any character to restart output.
     */
    if (ISSET(tp->t_state, TS_TTSTOP) &&
	!ISSET(iflag, IXANY) && cc[VSTART] != cc[VSTOP])
	return (0);

 restartoutput:
    CLR(tp->t_lflag, FLUSHO);
    CLR(tp->t_state, TS_TTSTOP);

 startoutput:
    return (ttstart(tp));
}



static void
ttpoll (struct tty *tp)
{
    int block, mustpoll, c;

    /* True if we can block waiting for input */
    block = (ISSET (tp->t_state, TS_RXWAIT)
	     /*&& !ISSET (tp->t_state, TS_NBIO)*/
	     && tp->t_outq.c_cc == 0);
    CLR (tp->t_state, TS_RXWAIT);

    /* 
     * There is no point in polling for input if we're just going
     * to throw it away because our input buffer is full! 
     * But we must do so if we are doing input processing
     * of any sort.
     */
    mustpoll = 	(ISSET(tp->t_lflag, ICANON|IEXTEN|ISIG)
		 || ISSET(tp->t_iflag, IXON));

    while (mustpoll || tp->t_rawq.c_cc + tp->t_canq.c_cc < tp->t_hog) {
	(void) ttyunlock (tp, 0);
	c = (*tp->t_rxpoll) (tp, block);
	ttylock (tp);
	if (c == 0)
	    break;
	ttyinput (c, tp);
	block = 0;
    }
}


static int
ttnread(struct tty *tp)
{
    int nread;
    
    if (ISSET(tp->t_lflag, PENDIN))
	ttypend(tp);

    nread = tp->t_canq.c_cc;
    if (!ISSET(tp->t_lflag, ICANON)) {
	nread += tp->t_rawq.c_cc;
	if (nread < tp->t_cc[VMIN] && !tp->t_cc[VTIME])
	    nread = 0;
    }
    return (nread);
}



int
_ttyioctl (struct tty *tp, unsigned int cmd, void *data)
{
    int ret = 0;

    ttylock(tp);
    switch (cmd) {		/* Process the ioctl. */
    case FIOASYNC:		/* set/clear async i/o */
	if (*(int *)data)
	    SET(tp->t_state, TS_ASYNC);
	else
	    CLR(tp->t_state, TS_ASYNC);
	break;
    case FIONBIO:		/* set/clear non-blocking i/o */
	break;
    case FIONREAD:		/* get # bytes to read */
	ttysleep (tp, TS_RXWAIT, 0);	/* poll */
	*(int *)data = ttnread(tp);
	break;
    case TIOCFLUSH: {		/* flush buffers */
	register int flags = *(int *)data;
	if (flags == 0)
	    flags = FREAD | FWRITE;
	else
	    flags &= FREAD | FWRITE;
	ttyflush(tp, flags);
	break;
    }
    case TIOCDRAIN:		/* wait till output drained */
	ret = ttywait(tp);
	break;
    case TIOCGETA: {		/* get termios struct */
	struct termios *t = (struct termios *)data;
	memcpy (t, &tp->t_termios, sizeof(struct termios));
	break;
    }
    case TIOCOUTQ:		/* output queue size */
	*(int *)data = tp->t_outq.c_cc;
	break;
    case TIOCSETA:		/* set termios struct */
    case TIOCSETAW:		/* drain output, set */
    case TIOCSETAF: {		/* drn out, fls in, set */
	register struct termios *t = (struct termios *)data;

	if (cmd == TIOCSETAW || cmd == TIOCSETAF) {
	    if (ret = ttywait(tp))
		break;
	    if (cmd == TIOCSETAF)
		ttyflush(tp, FREAD);
	}
	if (!ISSET(t->c_cflag, CIGNORE)) {
	    /*
	     * Set device hardware.
	     */
	    if (tp->t_params && (ret = (*tp->t_params)(tp, 1, t)))
		break;
	    if (!ISSET(tp->t_state, TS_CARR_ON) &&
		ISSET(tp->t_cflag, CLOCAL) &&
		!ISSET(t->c_cflag, CLOCAL)) {
		CLR(tp->t_state, TS_ISOPEN);
		SET(tp->t_state, TS_WOPEN);
		ttwakeup(tp, FREAD | FWRITE);
	    }
	    tp->t_cflag = t->c_cflag;
	    tp->t_ispeed = t->c_ispeed;
	    tp->t_ospeed = t->c_ospeed;
	    ttsetwater(tp);
	    if (t->c_ospeed == 0)
		_sig_raise (tp->t_pgrp, SIGHUP, 0);
	}
	if (cmd != TIOCSETAF) {
	    if (ISSET(t->c_lflag, ICANON) != ISSET(tp->t_lflag, ICANON))
		if (ISSET(t->c_lflag, ICANON)) {
		    /* ICANON on: reinput raw queue */
		    SET(tp->t_lflag, PENDIN);
		    /*ttwakeup(tp);*/
		} else {
		    /* ICANON off: add canon queue to raw queue */
		    struct clist tq;
		    catq(&tp->t_rawq, &tp->t_canq);
		    tq = tp->t_rawq;
		    tp->t_rawq = tp->t_canq;
		    tp->t_canq = tq;
		    CLR(tp->t_lflag, PENDIN);
		}
	}
	tp->t_iflag = t->c_iflag;
	tp->t_oflag = t->c_oflag;
	tp->t_lflag = t->c_lflag 
	    | ISSET(tp->t_lflag, PENDIN) | ISSET(tp->t_lflag, EXTPROC);
	memcpy (tp->t_cc, t->c_cc, sizeof(t->c_cc));
	break;
    }
    case TIOCSPGRP:		/* set process group */
	tp->t_pgrp = *(int *)data;
	break;
    case TIOCGPGRP:		/* get process group */
	*(int *)data = tp->t_pgrp;
	break;
    case TIOCSTART:		/* start output, like ^Q */
	if (ISSET(tp->t_state, TS_TTSTOP) || ISSET(tp->t_lflag, FLUSHO)) {
	    CLR(tp->t_lflag, FLUSHO);
	    CLR(tp->t_state, TS_TTSTOP);
	    ttstart(tp);
	}
	break;
    case TIOCSTOP:		/* stop output, like ^S */
	if (!ISSET(tp->t_state, TS_TTSTOP)) {
	    SET(tp->t_state, TS_TTSTOP);
	    if (tp->t_txstop)
		(*tp->t_txstop) (tp, 0);
	}
	break;
    case TIOCSENDBRK:		/* send a break signal */
	if (tp->t_sendbrk)
	    (*tp->t_sendbrk) (tp, *(int *)data);
	break;
    default:
	ret = ENOTTY;
	break;
    }

    return ttyunlock (tp, ret);
}



/* fixed buffers for console tty */
static struct tty	tty1;
static unsigned char	raw1[CLBUFSIZE (CLISTSIZE1)];
static unsigned char	can1[CLBUFSIZE (CLISTSIZE1)];
static unsigned char	out1[CLBUFSIZE (CLISTSIZE1)];
static int		first = 1;


int
_ttyclalloc (struct tty *tp)
{
    if (!tp->t_rawq.c_buf) {
	if (tp == &tty1) {
	    clinit (&tp->t_rawq, raw1, CLISTSIZE1);
	    clinit (&tp->t_canq, can1, CLISTSIZE1);
	    clinit (&tp->t_outq, out1, CLISTSIZE1);
	}
	else {
	    if (clalloc (&tp->t_rawq, CLISTSIZE) < 0 
		|| clalloc (&tp->t_canq, CLISTSIZE) < 0
		|| clalloc (&tp->t_outq, CLISTSIZE) < 0) {
		clfree (&tp->t_rawq);
		clfree (&tp->t_canq);
		clfree (&tp->t_outq);
		return ENOMEM;
	    }
	}
    }
    return 0;
}


void
_ttyclfree (struct tty *tp)
{
    if (tp != &tty1) {
	clfree (&tp->t_rawq);
	clfree (&tp->t_canq);
	clfree (&tp->t_outq);
    }
}

int
_ttymalloc (const char *name, struct tty **tpp)
{
    struct tty *tp;
    int ret;

    if (first) {
	tp = &tty1;
	first = 0;
    }
    else {
	tp = (struct tty *) malloc (sizeof (struct tty));
	if (!tp)
	    return ENOMEM;
    }
    memset (tp, 0, sizeof (struct tty));

    __sdethread_mutex_init (&tp->t_mx, name);

    if ((ret = _ttyclalloc (tp)) != 0) {
	_ttyfree (tp);
	return ret;
    }

    *tpp = tp;
    return 0;
}


void
_ttyfree (struct tty *tp)
{
    __sdethread_mutex_destroy (&tp->t_mx);
    _ttyclfree (tp);
    if (tp == &tty1)
	first = 1;
    else
	free (tp);
}



int
_ttyread (struct tty *tp, struct uio *uio, int flag)
{
    size_t oresid = uio->uio_resid;
    unsigned char *cc = tp->t_cc;
    unsigned int lflag = tp->t_lflag;
    int vmin, vtime;
    struct clist *qp;
    int err = 0;
    int c;

    ttylock(tp);
    
    /* take pending input first */
    if (ISSET(lflag, PENDIN))
	ttypend(tp);

    if (ISSET (lflag, ICANON)) {
	qp = &tp->t_canq;
	vmin = 1;
	vtime = -1;
    }
    else {
	qp = &tp->t_rawq;
	vtime = tp->t_cc[VTIME];
	if (vtime == 0) {
	    vmin = tp->t_cc[VMIN];
	    if (vmin > uio->uio_resid)
		vmin = uio->uio_resid;
	}
	else
	    vmin = 1;
    }

    _sig_interrupt_syscall = 0;
    while (qp->c_cc < vmin) {

	int carrier = ISSET(tp->t_state, TS_CARR_ON) ||
	    ISSET(tp->t_cflag, CLOCAL);
	if (!carrier && ISSET(tp->t_state, TS_ISOPEN))
	    return ttyunlock (tp, 0);	/* EOF */

	err = ttysleep (tp, TS_RXWAIT, vtime);
	if (err != 0 && err != ETIME) {
	    if (err == EINTR)
		/* XXX or should we just continue */
		break;
	    return ttyunlock (tp, err);
	}

	if (qp->c_cc == 0 && (flag & FNONBLOCK))
	    return ttyunlock (tp, EWOULDBLOCK);

	if (qp->c_cc < vmin && !ISSET (lflag, ICANON)) {
	    if (tp->t_cc[VMIN] > 0 && tp->t_cc[VTIME] > 0) {
		/* MIN > 0, TIME > 0: inter-byte timer */
		if (qp->c_cc > 0) {
		    /* wait until at least one byte arrives */
		    if (vmin == 1) {
			/* then minimum is relevant */
			vmin = tp->t_cc[VMIN];
			if (vmin > uio->uio_resid)
			    vmin = uio->uio_resid;
			continue;
		    }
		    /* else inter-byte timer fired */
		    break;
		}
	    }
	    else if (vtime > 0) {
		/* MIN = 0, TIME > 0: elapsed timer fired */
		break;
	    }
	    /* MIN > 0, TIME = 0: timer irrelevant */
	    /* MIN = 0, TIME = 0: shouldn't happen */
	}
    }

    while (uio->uio_resid != 0 && (c = getc (qp)) >= 0) {
	if (ISSET(lflag, ISIG) && CCEQ(cc[VDSUSP], c)) {
	    _sig_raise (tp->t_pgrp, SIGTSTP, NULL);
	    break;
	}

	if (ISSET(lflag, ICANON) && CCEQ(cc[VEOF], c))
	    break;

	err = ureadc (c, uio);
	if (err)
	    break;

	if (ISSET(lflag, ICANON) && TTBREAKC(c))
	    break;
    }

    if (ISSET(tp->t_state, TS_TBLOCK) && tp->t_rawq.c_cc < tp->t_hog / 2)
	ttyunblock (tp);

    ttysleep (tp, TS_RXWAIT, 0);    /* final poll */

    /* Only return EINTR if no characters have been 
       transferred to the user buffer, to distinguish
       from end of file. */
    if (err == EINTR && uio->uio_resid != oresid)
	err = 0;

    return ttyunlock (tp, err);
}


int
_ttywrite (struct tty *tp, struct uio *uio, int flag)
{
    size_t oresid = uio->uio_resid;
    int err = 0;
    int c;

    ttylock(tp);

    _sig_interrupt_syscall = 0;
    while (uio->uio_resid) {
	if (ISSET(tp->t_lflag, FLUSHO)) {
	    /*
	     * currently discarding output, but
	     * continue polling input to check for restart
	     */
	    ttysleep (tp, TS_RXWAIT, 0);	/* poll */
	    break;
	}

	/* wait for space in output queue */
	while (tp->t_outq.c_cc >= tp->t_hiwat) {
	    ttstart(tp);
	    if (flag & FNONBLOCK)
		err = EWOULDBLOCK;
	    else
		err = ttysleep (tp, TS_TXWAIT, -1);
	    if (err) {
		/* ignore error if any transferred */
		if (uio->uio_resid != oresid)
		    err = 0;
		return ttyunlock (tp, err);
	    }
	}
	    
	err = uwritec (&c, uio);

	tp->t_rocount = 0;
	if (CCLASS (c))
	    /* fancy character: get ttyoutput to handle it */
	    (void) ttyoutput (c, tp);
	else {
	    /* simple character: just stuff on output queue */
	    /* XXX should try to do a block at a time */
	    (void) putc (c, &tp->t_outq);
	    tp->t_column++;
	}

	ttysleep (tp, TS_RXWAIT, 0);		/* poll */
    }

    /* start the transmitter running */
    (void) ttstart (tp);

    /* In a polled environment, it isn't safe to return with output
       still pending.  If user has set NBIO however, we assume that
       they know what they are doing. */
    if (!(flag & FNONBLOCK)
	&& (!ISSET (tp->t_state, TS_INTRUPT) || _rtpx_sync_tty))
	(void) ttywait (tp);

    return ttyunlock (tp, err);
}


int
_ttyselect (struct tty *tp, int rw)
{
    int res = 0;

    ttylock(tp);

    if (rw & FREAD) {
	if (!ISSET (tp->t_state, TS_INTRUPT))
	    ttpoll (tp);
	if (ttnread(tp) > 0
	    || (!ISSET(tp->t_cflag, CLOCAL) &&
		!ISSET(tp->t_state, TS_CARR_ON)))
	    res |= FREAD;
	else
	    _selrecord (&tp->t_rsel);
    }

    if (rw & FWRITE) {
	if (!ISSET (tp->t_state, TS_INTRUPT))
	    ttstart (tp);
	if (tp->t_outq.c_cc <= tp->t_lowat)
	    res |= FWRITE;
	else
	    _selrecord (&tp->t_wsel);
    }

    return ttyunlock (tp, res);
}



static const char ttydefchars[NCCS] = TTYDEF_CHARS;

void
_ttychars (struct tty *tp)
{
    memcpy (&tp->t_cc, ttydefchars, NCCS);
}

int
_ttyopen (struct tty *tp)
{
    ttylock(tp);
    
    if (!ISSET (tp->t_state, TS_ISOPEN)) {
	assert (tp->t_count == 0);
	/* set software defaults */
	tp->t_iflag = TTYDEF_IFLAG;
	tp->t_oflag = TTYDEF_OFLAG;
	tp->t_lflag = TTYDEF_LFLAG;
	_ttychars (tp);
	if (tp->t_params) {
	    /* get cflags and speed settings from existing line */
	    (void) (*tp->t_params) (tp, 0, &tp->t_termios);
	    (void) (*tp->t_params) (tp, 1, &tp->t_termios);
	}
	ttsetwater (tp);
	SET (tp->t_state, TS_ISOPEN);
    }

    tp->t_count++;

    return ttyunlock (tp, 0);
}


/* line discipline close */
int
_ttylclose (struct tty *tp, int flag)
{
    ttylock(tp);
    if (flag & FNONBLOCK)
	ttyflush (tp, FREAD | FWRITE);
    else {
	if (ttywait (tp))
	    ttyflush (tp, FREAD);
    }
    return ttyunlock (tp, 0);
}


/* final device close */
int
_ttyclose (struct tty *tp)
{
    ttylock(tp);
    if (--tp->t_count == 0) {
	ttyflush (tp, FREAD | FWRITE);
	if (tp->t_close)
	    (*tp->t_close) (tp);
	/* clear all state bits except interrupt flag */
	tp->t_state &= TS_INTRUPT;
	tp->t_pgrp = 0;
    }
    return ttyunlock (tp, 0);
}


/*
 * Handle modem control transition on a tty.
 * Flag indicates new state of carrier.
 * Returns 0 if the line should be turned off, otherwise 1.
 */
int
_ttymodem(struct tty *tp, int flag)
{
    if (!ISSET(tp->t_state, TS_WOPEN) && ISSET(tp->t_cflag, MDMBUF)) {
	/*
	 * MDMBUF: do flow control according to carrier flag
	 */
	if (flag) {
	    CLR(tp->t_state, TS_TTSTOP);
	    ttstart(tp);
	} else if (!ISSET(tp->t_state, TS_TTSTOP)) {
	    SET(tp->t_state, TS_TTSTOP);
	    if (tp->t_txstop)
		(*tp->t_txstop) (tp, 0);
	}
    } else if (flag == 0) {
	/*
	 * Lost carrier.
	 */
	CLR(tp->t_state, TS_CARR_ON);
	if (ISSET(tp->t_state, TS_ISOPEN) &&
	    !ISSET(tp->t_cflag, CLOCAL)) {
	    _sig_raise (tp->t_pgrp, SIGHUP, NULL);
	    ttyflush(tp, FREAD | FWRITE);
	    return (0);
	}
    } else {
	/*
	 * Carrier now on.
	 */
	SET(tp->t_state, TS_CARR_ON);
	ttwakeup(tp, FREAD | FWRITE);
    }
    return (1);
}


#ifdef notyet
/*
 * Default modem control routine (for other line disciplines).
 * Return argument flag, to turn off device on carrier drop.
 */
int
_nullmodem (struct tty *tp, int flag)
{
    
    if (flag)
	SET(tp->t_state, TS_CARR_ON);
    else {
	CLR(tp->t_state, TS_CARR_ON);
	if (ISSET(tp->t_state, TS_ISOPEN) &&
	    !ISSET(tp->t_cflag, CLOCAL)) {
	    _sig_raise (tp->t_pgrp, SIGHUP, NULL);
	    ttyflush(tp, FREAD | FWRITE);
	    return (0);
	}
    }
    return (1);
}
#endif
