/*
 * Copyright (c) 1996-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * share/tty.h: POSIX serial i/o data structures
 */


#ifndef _SYS_TTY_H
#define _SYS_TTY_H

#include <sdethread.h>
#include <sys/select.h>
#include <sys/termios.h>

/*
 * share/tty.h 
 */

struct clist {
    short		c_limit;
    short		c_first;
    short		c_cc;
    unsigned char	*c_quote;
    unsigned char	*c_buf;
};


struct tty {
    unsigned short	t_state;
    dev_t		t_dev;
    pid_t		t_pgrp;
    struct termios	t_termios;
    struct clist	t_rawq;
    struct clist	t_canq;
    struct clist	t_outq;

    /* interface to low-level port driver: */
    /* set port parameters */
    int			(*t_params)(struct tty *, int, struct termios *);
    /* polled i/o interface */
    int			(*t_rxpoll)(struct tty *, int);
    int			(*t_txrdy)(struct tty *);
    void		(*t_tx)(struct tty *, int);
    int			(*t_txblk)(struct tty *, const char *, int);
    /* interrupt i/o interface */
    void		(*t_start)(struct tty *);
    int			(*t_wait)(struct tty *, int, int);
    void		(*t_wakeup)(struct tty *, int);
    void		(*t_rxflow)(struct tty *, int);
    void		(*t_txstop)(struct tty *, int);
    void		(*t_sendbrk)(struct tty *, int);
    void		(*t_close)(struct tty *);

    short		t_column;
    short		t_rocount, t_rocol;
    short		t_lowat, t_hiwat, t_hog;
    short		t_count;
    void *		t_addr;
    __sdethread_mutex_t t_mx;

    struct selinfo	t_rsel;
    struct selinfo	t_wsel;
};

#define	t_cc		t_termios.c_cc
#define	t_cflag		t_termios.c_cflag
#define	t_iflag		t_termios.c_iflag
#define	t_ispeed	t_termios.c_ispeed
#define	t_lflag		t_termios.c_lflag
#define	t_oflag		t_termios.c_oflag
#define	t_ospeed	t_termios.c_ospeed

/* TTY flags in t_state */
#define	TS_ISOPEN	0x0001	/* tty is open */
#define	TS_OPEN		TS_ISOPEN	/* history */
#define TS_ASYNC	0x0002	/* async i/o selected (SIGIO) */
#define TS_CARR_ON	0x0004	/* carrier enabled */
#define TS_TTSTOP	0x0008	/* stopped by XOFF */
#define TS_TBLOCK	0x0010	/* we've sent an XOFF */
#define	TS_RXWAIT	0x0020	/* ttread is waiting for input */
#define TS_BUSY		0x0040	/* ttstart is busy outputting */
#define TS_INTRUPT	0x0080	/* tty driver is interrupt driven */
#define	TS_TXWAIT	0x0100	/* waiting for output queue to lower */
#define	TS_TXDRAIN	0x0200	/* waiting for output queue to drain */
#define TS_XON_PEND	0x0400	/* an XON is pending */
#define TS_XOFF_PEND	0x0800	/* an XOFF is pending */
#define	TS_CNTTB	0x1000	/* counting tab width, ignore FLUSHO. */
#define	TS_LNCH		0x2000	/* next char is literal */
#define TS_TYPEN	0x4000	/* reinputting pending input */
#define TS_WOPEN	0x8000	/* waiting for open */
#define TS_LOCAL	(TS_CNTTB | TS_LNCH | TS_TYPEN)

#define CLISTSIZE1	256	/* size of first (console) clists */
#define CLISTSIZE	1024	/* size of clist for others */
#define CLBUFSIZE(n)	((n) + ((n) / 8))

/* Character type information. */
#define	ORDINARY	0
#define	CONTROL		1
#define	BACKSPACE	2
#define	NEWLINE		3
#define	TAB		4
#define	VTAB		5
#define	RETURN		6

/* Flags on a character passed to _ttyinput. */
#define	TTY_CHARMASK	0x000000ff	/* Character mask */
#define	TTY_QUOTE	0x00000100	/* Character quoted */
#define	TTY_ERRORMASK	0xff000000	/* Error mask */
#define	TTY_FE		0x01000000	/* Framing error or BREAK condition */
#define	TTY_PE		0x02000000	/* Parity error */

struct uio;

int		_ttymalloc (const char *, struct tty **);
void		_ttyfree (struct tty *);
int		_ttyclalloc (struct tty *);
void		_ttyclfree (struct tty *);
void		_ttylock (struct tty *);
int		_ttyunlock (struct tty *, int);
int		_ttyread (struct tty *, struct uio *, int);
int		_ttywrite (struct tty *, struct uio *, int);
int		_ttyioctl (struct tty *, unsigned int, void *);
int		_ttyselect (struct tty *, int);
int		_ttyopen (struct tty *);
int		_ttylclose (struct tty *, int);
int		_ttyclose (struct tty *);
int		_ttywait (struct tty *);
int		_ttysleep (struct tty *, int, int);
void		_ttwakeup (struct tty *, int);
void		_ttyflush (struct tty *, int);
void		_ttsetwater (struct tty *);
void		_ttychars (struct tty *);
int		_ttymodem (struct tty *, int);
int		_nullmodem (struct tty *, int);

/* clist support for interrupt drivers */
int		_ttyinput (int, struct tty *);
int		_ttygetc (struct clist *);
int		_ttygetcblk (struct clist *, char **);
void		_ttydelcblk (struct clist *, int);
int		_ttyputc (int, struct clist *);
int		_ttyunputc (struct clist *);
void		_ttyflushq (struct clist *);

#endif /*_SYS_TTY_H*/
