#ident "kit/share/unaligned.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1999-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * unaligned.c: trap-based emulator for unaligned loads and stores
 */


/*
 * NOTE: this code doesn't handle floating-point load/stores;
 *	 it is very slow, and is no substitute for porting your code
 *       to avoid unaligned accesses.
 *
 * TODO: Keep history table so that user can view a list of instructions that
 *	 need to be rewritten.
 */

#include <stddef.h>
#include <mips/cpu.h>
#include <mips/fpa.h>
#include <mips/xcpt.h>
#include <mips/prid.h>
#include <mips/instrdef.h>
#include <mips/instr16def.h>
#include <sys/signal.h>
#include "kit_impl.h"

/* convert MIPS16 register to MIPS32 register */
#define mips16_to_mips32(rn)	(((rn) < 2) ? (rn) + 16 : (rn))

typedef union {
    short		int16;
    long		int32;
    long long		int64;
    unsigned short	uint16;
    unsigned long	uint32;
    unsigned long long	uint64;
} __attribute__((packed)) unaligned_t;


static int
unaligned_handler (int xcptno, struct xcptcontext *xcp)
{
    vaddr_t pc = xcp->epc;
    unaligned_t *up;
    int err = 0;
    int sig;

#if __mips16 || __maybe_mips16
    if (pc & 1) {
	union mips16_instr i16;
	int rx, ry;
	unsigned short extension = 0;
	short offset;

	i16.value = mips_get_half (pc & ~1, &err);
	if (err) return SIGBUS;
	pc += 2;

	if (XCP_CAUSE (xcp) & CR_BD) {
	    /* instruction in branch-delay slot: emulate branch and
	       fetch the delay slot instruction */
	    vaddr_t bdpc;
	    if (sig = _mips16_bremul (xcp, i16.value, &pc, &bdpc))
		return sig;
	    i16.value = mips_get_half (bdpc & ~1, &err);
	    if (err) return SIGBUS;
	}
	else if (i16.itype.op == ITYPE16_extend) {
	    /* extend instruction: save extension & fetch next insn */
	    extension = i16.exttype.imm; /* we'll manipulate it later */
	    i16.value = mips_get_half (pc & ~1, &err);
	    if (err) return SIGBUS;
	    pc += 2;
	}

	rx = mips16_to_mips32(i16.rritype.rx);
	ry = mips16_to_mips32(i16.rritype.ry);

	/* zero-extend 5-bit offset */
	offset = i16.rritype.imm & 0x1f;
	offset |= extension & 0x7e0;
	offset |= (extension & 0x1f) << 11;

	up = (unaligned_t *) (REG_TO_VA (xcp->regs[rx]) + offset);

	switch (i16.itype.op) {
	case ITYPE16_lh:
	    xcp->regs[ry] = up->int16;
	    break;
	case ITYPE16_lhu:
	    xcp->regs[ry] = up->uint16;
	    break;
	case ITYPE16_lw:
	    xcp->regs[ry] = up->int32;
	    break;
#if (__mips >= 3 && __mips != 32) || __mips64
	case ITYPE16_lwu:
	    xcp->regs[ry] = up->uint32;
	    break;
	case ITYPE16_ld:
	    xcp->regs[ry] = up->uint64;
	    break;
#endif

	case ITYPE16_sh:
	    up->uint16 = xcp->regs[ry];
	    break;
	case ITYPE16_sw:
	    up->uint32 = xcp->regs[ry];
	    break;
#if (__mips >= 3 && __mips != 32) || __mips64
	case ITYPE16_sd:
	    up->uint64 = xcp->regs[ry];
	    break;
#endif

	default:
	    return SIGILL;
	}
    }
    else 
#endif
    {
	union mips_instr ir;
	int rs, rt;

	ir.value = mips_get_word ((void *)pc, &err);
	if (err) return SIGBUS;
	pc += 4;

	if (XCP_CAUSE (xcp) & CR_BD) {
	    /* instruction in branch-delay slot: emulate branch and
	       fetch the delay slot instruction */
	    vaddr_t contpc;
	    int skipbd;

	    if (sig = _mips_bremul (xcp, NULL, ir.value, &contpc, &skipbd)) {
		/* XXX NEC Vr54xx bug workaround */
		if (sig == SIGILL &&
		    ((mips_getprid() >> 8) & 0xff) == PRID_R5400) {
		    (xcp)->cr = XCP_CAUSE (xcp) & ~CR_BD;
		    goto notbd;
		}
		return sig;
	    }

	    if (skipbd) {
		/* failing branch-likely, skip delay slot */
		pc = contpc;	/* continue from branch target */
		goto done;
	    }

	    ir.value = mips_get_word ((void *)pc, &err);
	    if (err) return SIGBUS;
	    pc = contpc;	/* continue from branch target */
	}

    notbd:
	rs = ir.itype.rs;
	rt = ir.itype.rt;
	up = (unaligned_t *) (REG_TO_VA (xcp->regs[rs]) + ir.itype.imm);

	switch (ir.itype.op) {
	case ITYPE_lh:
	    xcp->regs[rt] = up->int16;
	    break;
	case ITYPE_lhu:
	    xcp->regs[rt] = up->uint16;
	    break;
	case ITYPE_lw:
	    xcp->regs[rt] = up->int32;
	    break;
#if (__mips >= 3 && __mips != 32) || __mips64
	case ITYPE_lwu:
	    xcp->regs[rt] = up->uint32;
	    break;
	case ITYPE_ld:
	    xcp->regs[rt] = up->uint64;
	    break;
#endif

	case ITYPE_sh:
	    up->uint16 = xcp->regs[rt];
	    break;
	case ITYPE_sw:
	    up->uint32 = xcp->regs[rt];
	    break;
#if (__mips >= 3 && __mips != 32) || __mips64
	case ITYPE_sd:
	    up->uint64 = xcp->regs[rt];
	    break;
#endif

	case ITYPE_ll:
	case ITYPE_lld:
	case ITYPE_sc:
	case ITYPE_scd:
	case ITYPE_cache:
	    /* load-linked/store-conditional and cache insns not emulated */

	case ITYPE_lwc1:
	case ITYPE_ldc1:
	case ITYPE_swc1:
	case ITYPE_sdc1:
	    /* floating-point load/store not emulated */
	    return SIGBUS;

	default:
	    return SIGILL;
	}
    }

 done:
    xcp->epc = VA_TO_REG (pc);
    return 0;
}


/* 
 * Call this function to install the handler.
 */
void
_mips_unaligned_init (void)
{
    static int inited = 0;
    if (!inited) {
	struct xcptaction xa;
	xa.xa_handler = unaligned_handler;
	xa.xa_flags = 0;
	xcptaction (XCPTADEL, &xa, NULL);
	xcptaction (XCPTADES, &xa, NULL);
	inited = 1;
    }
}


/* initialise module just before calling main() */
#include <sys/init.h>
DECL_INIT_FUNC(IF_PREMAIN, _mips_unaligned_init)

/* include module by referring to this symbol */
void _unaligned_load (void);
void _unaligned_load (void) {}
