#ident "kit/share/watch.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * watch.c: switchable watchpoint support for MIPS SDE kit
 */


#include <sys/types.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <mips/watchpoint.h>

#include "kit_impl.h"

#if #cache(m32)
extern struct mips_watchpoint_funcs	_m32_watchpoint_funcs;
#endif
#if #cache(r54) && 0
/* XXX should support Vr5432 debug mode - for now use r4000 compatibility */
extern struct mips_watchpoint_funcs	_r54_watchpoint_funcs;
#endif
#if #cache(r4k) || #cache(r54)
extern struct mips_watchpoint_funcs	_r4k_watchpoint_funcs;
#endif
#if #cache(r3k)
extern struct mips_watchpoint_funcs	_r3k_watchpoint_funcs;
#endif
#if #cache(r4k) || #cache(rc32)
extern struct mips_watchpoint_funcs	_r4650_watchpoint_funcs;
#endif
#if #cache(r5k)
extern struct mips_watchpoint_funcs	_r5k_watchpoint_funcs;
#endif
#if #cache(rm7k)
extern struct mips_watchpoint_funcs	_rm7k_watchpoint_funcs;
#endif
#if #cpu(lr33k)
extern struct mips_watchpoint_funcs	_lr33k_watchpoint_funcs;
#endif
#if #cache(cw10) || #cache(cw01)
extern struct mips_watchpoint_funcs	_cw4_watchpoint_funcs;
#endif

static const struct mips_watchpoint_funcs *watchpoint_funcs[] = {
#if #cache(m32)
    &_m32_watchpoint_funcs,
#endif
#if #cache(r54) && 0
    &_r54_watchpoint_funcs,
#endif
#if #cache(r4k) || #cache(r54)
    &_r4k_watchpoint_funcs,
#endif
#if #cache(r4k) || #cache(rc32)
    &_r4650_watchpoint_funcs,
#endif
#if #cache(r3k)
    &_r3k_watchpoint_funcs,
#endif
#if #cache(r5k)
    &_r5k_watchpoint_funcs,
#endif
#if #cache(rm7k)
    &_rm7k_watchpoint_funcs,
#endif
#if #cpu(lr33k)
    &_lr33k_watchpoint_funcs,
#endif
#if #cache(cw10) || #cache(cw01)
    &_cw4_watchpoint_funcs,
#endif
    0
};


/* current cpu functions */
static const struct mips_watchpoint_funcs *wfn;

/* list of current watchpoints */
static struct mips_watchpoint watchpoints[WATCHPOINT_MAX];
static int nwatchpoints = -1;

/*
 * Initialise the watchpoint subsystem
 */
int
_mips_watchpoint_init (void)
{
    const struct mips_watchpoint_funcs * const *wfnp;
    unsigned int prid;
    int found = 0;

    if (wfn)
	/* already initialised */
	return nwatchpoints;

    prid = mips_getprid ();
    for (wfnp = watchpoint_funcs; *wfnp; wfnp++)
	if ((found = (*wfnp)->wp_init (prid)) >= 0)
	    break;

    if (found > 0) {
	int i;
	wfn = *wfnp;

	/* initialise each watchpoint resource */
	nwatchpoints = found;
	for (i = 0; i < found; i++) {
	    watchpoints[i].num = i;
	    wfn->wp_setup (&watchpoints[i]);
	}

	/* reset cpu-specific facilities */
	wfn->wp_reset ();
    }

    /* return number of resources */
    return nwatchpoints;
}


/*
 * Return the number of watchpoint resources.
 * XXX redundant function
 */
int
_mips_watchpoint_howmany (void)
{
    return nwatchpoints;
}


/*
 * Return the capabilities of the numbered watchpoint resource (0 .. n-1)
 */
int
_mips_watchpoint_capabilities (int wpnum)
{
    if (wpnum < 0 || wpnum >= nwatchpoints)
	return 0;
    return watchpoints[wpnum].capabilities;
}


/*
 * Return the maximal mask of the numbered watchpoint resource (0 .. n-1)
 */
vaddr_t
_mips_watchpoint_mask (int wpnum)
{
    if (wpnum < 0 || wpnum >= nwatchpoints)
	return 0;
    return watchpoints[wpnum].maxmask;
}


/*
 * Return non-zero set of "hit" flags if we've just hit a watchpoint
 * and optionally return the virtual address and byte length of 
 * the access which triggered it.
 */
int
_mips_watchpoint_hit (const struct xcptcontext *xcp, vaddr_t *pva, 
		      size_t *plen)
{
    int hit;
    vaddr_t va;
    size_t len;

    if (!wfn)
	return 0;

    if (!pva)
	pva = &va;

    if (!plen)
	plen = &len;

    hit = wfn->wp_hit (xcp, pva, plen);

    if (hit) {
	struct mips_watchpoint *wp;
	for (wp = watchpoints; wp < &watchpoints[nwatchpoints]; wp++) {
	    if ((wp->type & hit) && wp->len) {
		vaddr_t wpa, hita;

		wpa = wp->va;
		hita = *pva;
		if (!(wp->capabilities & MIPS_WATCHPOINT_VADDR)) {
		    wpa = KVA_TO_PA (wpa);
		    hita = KVA_TO_PA (hita);
		}

		/* do address ranges overlap */
		if (wpa < hita + *plen && hita < wpa + wp->len)
		    return hit | (wp->type & MIPS_WATCHPOINT_VADDR);
	    }
	}
	hit |= MIPS_WATCHPOINT_INEXACT;
    }

    return hit;
}


/*
 * Remove watchpoints after a debug exception
 */
void
_mips_watchpoint_remove (void)
{
    if (wfn)
	wfn->wp_remove ();
}


/*
 * Insert watchpoints before returning from debug exception
 */
void
_mips_watchpoint_insert (void)
{
    if (wfn)
	wfn->wp_insert ();
}



/*
 * Chance for debug host to complain if the address range
 * covered by the watchpoint would overlap its own data.
 */
#pragma weak _mips_watchpoint_set_callback
int
_mips_watchpoint_set_callback (int asid, vaddr_t va, size_t len)
{
    return MIPS_WP_OK;
}


/* 
 * Compute a don't care mask for the region bounding ADDR 
 * and ADDR + LEN - 1. 
 */
vaddr_t
_mips_watchpoint_calc_mask (addr, len)
     vaddr_t addr;
     size_t len;
{
  vaddr_t mask;
  int i;

  mask = addr ^ (addr + len - 1);

  for (i = 32; i >= 0; i--)
    if (mask == 0)
      break;
    else
      mask >>= 1;

  mask = (~(vaddr_t)0) >> i;

  return mask;
}


/*
 * Add a new watchpoint 
 */
int
_mips_watchpoint_set (int type, int asid, vaddr_t va, paddr_t pa, size_t len)
{
    struct mips_watchpoint *wp;
    int res = MIPS_WP_NOTSUP;

    if (nwatchpoints <= 0)
	return MIPS_WP_NONE;

    for (wp = watchpoints; wp < &watchpoints[nwatchpoints]; wp++) {
	if ((wp->capabilities & type) == type) {
	    if (wp->len != 0)
		res = MIPS_WP_INUSE;
	    else {
		wp->type = type;
		wp->asid = asid;
		wp->va = va;
		if ((wp->capabilities & MIPS_WATCHPOINT_VADDR) || pa)
		    wp->pa = pa;
		else if (IS_KVA01 (va))
		    wp->pa = KVA_TO_PA (va);
		else
		    /* can't translate mapped virtual address */
		    return MIPS_WP_BADADDR;
		wp->len = len;
		wp->mask = _mips_watchpoint_calc_mask (va, len);
		if (!wfn->wp_set (wp))
		    res = MIPS_WP_NOTSUP;
		else
		    res = _mips_watchpoint_set_callback (wp->asid, 
							 wp->va & ~wp->mask, 
							 wp->mask + 1);
		if (res != MIPS_WP_OK)
		    wp->len = 0;
		break;
	    }
	}
    }

    return res;
}


/*
 * Remove an existing watchpoint 
 */
int
_mips_watchpoint_clear (int type, int asid, vaddr_t va, size_t len)
{
    struct mips_watchpoint *wp;

    for (wp = watchpoints; wp < &watchpoints[nwatchpoints]; wp++) {
	if (wp->type == type && wp->asid == asid
	    && wp->va == va && wp->len == len) {
	    wp->len = 0;
	    wfn->wp_clear (wp);
	    return MIPS_WP_OK;
	}
    }

    return MIPS_WP_NOMATCH;
}


/*
 * Remove all watchpoints
 */
void
_mips_watchpoint_reset (void)
{
    if (wfn) {
	struct mips_watchpoint *wp;
	for (wp = watchpoints; wp < &watchpoints[nwatchpoints]; wp++)
	    wp->len = 0;
	wfn->wp_reset ();
    }
}
