#ident "kit/share/watchsup.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * watchsup.c: watchpoint support code for MIPS SDE kit
 */


#include <sys/types.h>
#include <stdlib.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <mips/instrdef.h>
#include <mips/instr16def.h>
#include <mips/watchpoint.h>

#include "kit_impl.h"


/* convert MIPS16 register to MIPS32 register */
#define mips16_to_mips32(rn)	(((rn) < 2) ? (rn) + 16 : (rn))

/* size of data access for all load/store instructions */
static const unsigned char datalen[64] = {
    0, 0, 0, 0,		/* 00: */
    0, 0, 0, 0,		/* 04: */
    0, 0, 0, 0,		/* 08: */
    0, 0, 0, 0,		/* 12: */
    0, 0, 0, 0,		/* 16: */
    0, 0, 0, 0,		/* 20: */
    0, 0, 7, 7,		/* 24: xx  xx    ldl  ldr */
    0, 0, 0, 0,		/* 28: */
    1, 2, 3, 4,		/* 32: lb   lh   lwl  lw   */
    1, 2, 3, 4,		/* 36: lbu  lhu  lwr  lwu  */
    1, 2, 3, 4,		/* 40: sb   sh   swl  sw   */
    7, 7, 3, 4,		/* 44: sdl  sdr  swr  cache*/
    4, 4, 4, 4,		/* 48: ll   lwc1 lwc3 lwc3 */
    8, 8, 8, 8,		/* 52: lld  ldc1 ldc2 ld   */
    4, 4, 4, 4,		/* 56: sc   swc1 swc2 swc3 */
    8, 8, 8, 8		/* 60: scd  sdc1 sdc2 sd   */
};


/* 
 * Return the address referenced by the trapped instruction, and the
 * size of the access.
 */
int
_mips_watchpoint_address (const struct xcptcontext *xcp, int type, 
			  vaddr_t *pva, size_t *plen)
{
    vaddr_t pc = (vaddr_t) xcp->epc;
    vaddr_t va = 0;
    int len = 0;

    if ((pc & 3) == 0) {
	if (XCP_CAUSE (xcp) & CR_BD)
	    pc += 4;
	if (type & (MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W)) {
	    union mipsn_instr i;
	    int err = 0;
	    type &= ~(MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W);
	    i.value = mips_get_word ((void *)pc, &err);
	    if (err == 0 && (len = datalen[i.itype.op])) {
		/* calculate virtual address */
		va = (vaddr_t) xcp->regs[i.itype.rs] + i.itype.imm;

		/* handle left/right magic */
		if (len == 3 || len == 7) {
		    size_t plen;
		    switch (i.itype.op) {
		    case ITYPE_lwl:
		    case ITYPE_swl:
		    case ITYPE_ldl:
		    case ITYPE_sdl:
#if #endian(big)
			plen = len + 1 - (va & len);
#else
			va &= ~len;
			plen = (va & len) + 1;
#endif
			break;
		    case ITYPE_ldr:
		    case ITYPE_sdr:
		    case ITYPE_lwr:
		    case ITYPE_swr:
#if #endian(big)
			va &= ~len;
			plen = (va & len) + 1;
#else
			plen = len + 1 - (va & len);
#endif
			break;
		    default:
			abort();
		    }
		    len = plen;
		}

		if (i.itype.op == ITYPE_ldl || i.itype.op == ITYPE_ldr)
		    /* special cases */
		    type |= MIPS_WATCHPOINT_R;
		else if (i.rrtype.op == ITYPE_special2 && 
			 i.rrtype.funct == SPEC2_lxs &&
			 i.rrtype.shamt == LXS_lwxs)
		{	
		    /* special case: SmartMIPS lwxs */
		    type |= MIPS_WATCHPOINT_R;
		    va = (vaddr_t) xcp->regs[i.rrtype.rs]
			+ xcp->regs[i.rrtype.rt] * 4;
		    len = 4;
		}
		else if (i.cp1xtype.op == ITYPE_c1x &&
			 (i.cp1xtype.funct == COP1X_lxc1 ||
			  i.cp1xtype.funct == COP1X_sxc1))
		{
		    /* special case: lxc1 & sxc1 */
		    type |= (i.cp1xtype.funct == COP1X_lxc1)
			? MIPS_WATCHPOINT_R : MIPS_WATCHPOINT_W;
		    va = (vaddr_t) xcp->regs[i.cp1xtype.fr]
			+ xcp->regs[i.cp1xtype.ft];
		    len = 4 << i.cp1xtype.fmt;
		}
		else
		    /* normal case */
		    type |= ((i.itype.op & 8) ? MIPS_WATCHPOINT_W
			     : MIPS_WATCHPOINT_R);
	    }
	}
	if (type == MIPS_WATCHPOINT_X) {
	    /* not a load/store, must be an instruction fetch breakpoint */
	    va = pc;
	    len = 4;
	}
    }
#if __mips16 || __maybe_mips16
    else {
	/* oh my god, it's MIPS16! */
	union mips16_instr i;

	/* fetch insn at watchpoint and compute size of insn */
	pc &= ~1;
	i.value = mips_get_half (pc, &err);
	if (i.itype.op == ITYPE16_extend
	    || i.itype.op == ITYPE16_jal)
	    len = 4;
	else
	    len = 2;

	if (XCP_CAUSE (xcp) & CR_BD) {
	    /* in branch delay: skip jump, dslot must be unextended */
	    pc += len;
	    len = 2;
	}

	if (type == MIPS_WATCHPOINT_X) {
	    /* execution watchpoint at pc, len is correct */
	    va = pc;
	}
	else if (type & (MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W)) {
	    vaddr_t max;
	    int extended = 0;
	    unsigned int extension = 0;
	    int shift = 0;
	    reg_t rv;
	    short offset;
	    int len = 0;
	    int err = 0;

	    type &= ~(MIPS_WATCHPOINT_R | MIPS_WATCHPOINT_W);

	again:
	    i.value = mips_get_half (pc, &err);
	    if (err == 0)
		switch (i.itype.op) {

		case ITYPE16_extend:
		    /* extended instruction */
		    extension = i.exttype.imm;
		    extended = 1;
		    pc += 2;
		    goto again;

#if 0
		/* bdslot already handled */
		case ITYPE16_jal:
		    /* jal: assume load/store in delay slot */
		    pc += 4;
		    goto again;

		case ITYPE16_rr:
		    if (i.rrtype.funct == RRTYPE16_jr) {
			/* jr/jalr: assume load/store in delay slot */
			pc += 2;
			goto again;
		    }
		    break;
#endif

		case ITYPE16_sb:
		    type |= MIPS_WATCHPOINT_W;
		case ITYPE16_lb:
		case ITYPE16_lbu:
		    shift = 0;
		    goto rri;

		case ITYPE16_sh:
		    type |= MIPS_WATCHPOINT_W;
		case ITYPE16_lh:
		case ITYPE16_lhu:
		    shift = 1;
		    goto rri;

		case ITYPE16_sw:
		    type |= MIPS_WATCHPOINT_W;
		case ITYPE16_lw:
		case ITYPE16_lwu:
		    shift = 2;
		    goto rri;

		case ITYPE16_sd:
		    type |= MIPS_WATCHPOINT_W;
		case ITYPE16_ld:
		    shift = 3;
		    goto rri;

		rri:
		    rv = xcp->regs[mips16_to_mips32(i.rritype.rx)];
		ri:
		    offset = i.rritype.imm & 0x1f;
		    goto calc;

		case ITYPE16_lwpc:
		    rv = xcp->epc;
		    offset = i.ritype.imm & 0xff;
		    shift = 2;
		    goto calc;

		case ITYPE16_swsp:
		swsp:
		    type |= MIPS_WATCHPOINT_W;
		case ITYPE16_lwsp:
		    shift = 2;
		sdsp:
		    rv = xcp->regs[XCP_SP];
		    offset = i.ritype.imm & 0xff;
		    goto calc;

		calc:
		    if (extended) {
			offset &= 0x1f;
			offset |= extension & 0x7e0;
			offset |= (extension & 0x1f) << 11;
		    }
		    else {
			offset <<= shift;
		    }
		    va = (vaddr_t) rv + offset;
		    len = 1 << shift;
		    /* round down to aligned boundary */
		    va &= ~(len - 1);
		    if (!(type & MIPS_WATCHPOINT_W))
			type |= MIPS_WATCHPOINT_R;
		    break;
		    
		    
		case ITYPE16_i8:
		    switch (i.i8type.funct) {
		    case I8TYPE16_swrasp:
			goto swsp;

		    case I8TYPE16_svrs:
			va = xcp->regs[XCP_SP];
			if (i.i8type.imm & 0x80)
			    /* save: use incoming sp. */
			    type |= MIPS_WATCHPOINT_W;
			else {
			    /* restore: use adjusted sp. */
			    type |= MIPS_WATCHPOINT_R;
			    offset = i.i8type.imm & 0x0f;
			    if (extended)
				offset |= extension & 0xf0;
			    else if (offset == 0)
				offset = 128 / 8;
			    va += offset * 8;
			}

			max = va;
			if (extended) {
			    /* args */
			    switch (extension & 0xf) {
			    case 14:				/* a3 */
				max += 4;
			    case 12: case 13:			/* a2 */
				max += 4;
			    case 8: case 9: case 10:		/* a1 */
				max += 4 ;
			    case 4: case 5: case 6: case 7:	/* a0 */
				max += 4;
			    }
			    /* astatic */
			    switch (extension & 0xf) {	
			    case 11:
				va -= 4;
			    case 3: case 7:
				va -= 4;
			    case 2: case 6: case 10:
				va -= 4;
			    case 1: case 5: case 9: case 13:
				va -= 4;
			    }
			    /* xsregs */
			    va -= 4 * ((extension >> 8) & 0x7);
			}
			if (i.i8type.imm & 0x40) 		/* ra */
			    va -= 4;
			if (i.i8type.imm & 0x20) 		/* s0 */
			    va -= 4;
			if (i.i8type.imm & 0x10) 		/* s1 */
			    va -= 4;
			len = max - va;
			break;
		    }
		    break;

		case ITYPE16_i64:
		    switch (i.i64type.funct) {
		    case I64TYPE16_sdsp:
			type |= MIPS_WATCHPOINT_W;
		    case I64TYPE16_ldsp:
			shift = 3;
			rv = xcp->regs[XCP_SP];
			goto ri;

		    case I64TYPE16_sdrasp:
			type |= MIPS_WATCHPOINT_W;
			shift = 3;
			goto sdsp;
			
		    case I64TYPE16_ldpc:
			shift = 3;
			rv = xcp->epc;
			goto ri;
		    }
		}
	}
    }
#endif

    if (plen)
	*plen = len;
    if (pva)
	*pva = va;
    return type;
}
