#ident "kit/share/xcpt.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * xcpt.c: C-level exception handling for MIPS SDE
 */


#include <stdlib.h>
#include <assert.h>
#include <sys/errno.h>
#include <sys/signal.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include "kit_impl.h"

static const unsigned char xcpcsig[] =
{
#define XCPCDEF(NAME, NUMBER, SIG, DESC) SIG ,
#include <mips/xcptcpu.h>
#undef XCPCDEF
    0
};

static const unsigned char xcptsig[] =
{
#define XCPTDEF(NAME, NUMBER, SIG, BRIEF, DESC)	SIG ,
#include <mips/xcptcpu.h>
#undef XCPTDEF
    0
};

int
_xcpt_signo (const struct xcptcontext *xcp)
{
    int sig = xcpcsig[XCP_CLASS(xcp)];
    if (sig == 0)
	sig = xcptsig[XCP_CODE(xcp)];
    return sig;
}

/*
 * The default xcption handler converts an xcption into a POSIX signal
 * to deliver to the current `process'.
 */
static int
xcpt_default (int xcptno, struct xcptcontext *xcp)
{
    int             sig = _xcpt_signo (xcp);

    switch (sig) {
#ifdef DCS_PC
    case SIGBUS:
	/* some LSI parts signal watchpoints with a bus error */
	if ((cw400x_getdcs () & DCS_PC) || (cw400x_getdcs () & DCS_DA))
	    sig = SIGTRAP;
	break;
#endif

    case SIGABRT:
	/* catch well known break/trap instructions */
	{
	    unsigned int bcode;
	    int err = 0;

#if __mips16 || __maybe_mips16
	    if ((unsigned int)xcp->epc & 1) 
	    {
		/* MIPS16 break */
		unsigned short *ip;
		unsigned short i;
		
		ip = (unsigned short *) REG_TO_VA (xcp->epc & ~1);

#ifndef DBGVA
		/* accessing epc in mapped region could fault */
		if (!IS_KVA01 (ip)) {
		    sig = SIGSEGV;
		    break;
		}
#endif

		if (XCP_CAUSE (xcp) & CR_BD) {
		    /* skip the jr or jal */
		    i = mips_get_half (ip, &err);
		    if ((i & 0xf800) == 0x1800)	
			ip += 2;			/* jal */
		    else
			ip += 1;			/* jr */
		}
		
		if (!err)
		    /* get the break insn */
		    i = mips_get_half (ip, &err);

		if (err) {
		    sig = SIGBUS;
		    break;
		}
	    
		bcode = (i >> 5) & 0x3f;
	    }
	    else
#endif
	    {
		/* MIPS32 break */
		unsigned int *ip;
		unsigned int i;

		ip = (unsigned int *) REG_TO_VA (xcp->epc);
		if (XCP_CAUSE (xcp) & CR_BD)
		    ip++;
		
		if (((unsigned int)xcp->epc & 3)
#ifndef DBGVA
		    /* accessing epc in mapped region could fault */
		    || !IS_KVA01 (ip)
#endif
		    ) {
		    sig = SIGSEGV;
		    break;
		}

		i = mips_get_word (ip, &err);
		if (err) {
		    sig = SIGBUS;
		    break;
		}
		
		if (xcptno == XCPTBP) {
		    /* extract break code number */
		    bcode = (i >> 16) & 0x3ff;
		}
		else if ((i & 0xfc000000) == 0x00000000) {
		    /* extract trap code number */
		    bcode = (i >> 6) & 0x3ff;
		    if (bcode == 0) 
			bcode = 7; /* old gas compatibility */
		}
		else
		    bcode = 7;

	    }

	    switch (bcode) {
	    case 0:		/* dbg.c breakpoint */
	    case 5:		/* gdb breakpoint */
		sig = SIGTRAP;
		break;
	    case 1:		/* SDE libc abort() */
		sig = SIGABRT;
		break;
	    case 6:		/* overflow check */
	    case 7:		/* zero divide */
	    case 1023:		/* multiply overflow */
		sig = SIGFPE;
		break;
	    default:
		sig = SIGILL;
		break;
	    }
	}
	break;
    }

    return sig;
}


/*
 * Inverse mapping from signal number to exceptions.  Used to ensure
 * that our exception handler catches those exceptions which might cause 
 * that signal.
 */

void
_xcpt_sigcatch (int signo, int enable)
{
    int x;
    for (x = 0; x < NXCPT; x++)
	if (signo <= 0 || xcptsig[x] == signo)
	    _xcpt_catch (x, enable);
}



static xcpt_t   xcpthandlertab[NXCPT] =
{
    &xcpt_default, &xcpt_default, &xcpt_default, &xcpt_default,
    &xcpt_default, &xcpt_default, &xcpt_default, &xcpt_default,
    &xcpt_default, &xcpt_default, &xcpt_default, &xcpt_default,
    &xcpt_default, &xcpt_default, &xcpt_default, &xcpt_default,
#if NXCPT > 16
    &xcpt_default, &xcpt_default, &xcpt_default, &xcpt_default,
    &xcpt_default, &xcpt_default, &xcpt_default, &xcpt_default,
    &xcpt_default, &xcpt_default, &xcpt_default, &xcpt_default,
    &xcpt_default, &xcpt_default, &xcpt_default, &xcpt_default,
#endif
};


/*
 * OLD simple interface
 */
xcpt_t
xcption (enum xcptcode xcptno, xcpt_t handler)
{
    struct xcptaction new, old;
    int err;

    new.xa_handler = handler;
    new.xa_flags = 0;

    if (err = xcptaction (xcptno, &new, &old)) {
	extern int errno;
	errno = err;		/* XXX shouldn't set errno */
	return XCPT_ERR;
    }

    return old.xa_handler;
}


int
xcptaction (enum xcptcode xcptno, struct xcptaction *new, struct xcptaction *old)
{
    extern unsigned int _xcpt_flags[];

    if (xcptno < 0 || xcptno >= NXCPT)
	return EINVAL;

    if (old) {
	old->xa_handler = xcpthandlertab[xcptno];
	if (old->xa_handler == xcpt_default)
	    old->xa_handler = XCPT_DFL;
	old->xa_flags = _xcpt_flags[xcptno];
    }

    if (new) {
	/* 
	 * You may *want* to put an exception handler in kseg2/kuseg,
	 * but that would make things very tricky!
	 */
	if (new->xa_handler == XCPT_DFL)
	    new->xa_handler = xcpt_default;
	else if (!IS_KVA01 (new->xa_handler))
	    return EINVAL;

	/* commit to table (note this should be an atomic set) */
	xcpthandlertab[xcptno] = new->xa_handler;
	_xcpt_flags[xcptno] = new->xa_flags;

	/* hook or unhook our exception handler */
	_xcpt_catch (xcptno, new->xa_handler != xcpt_default);
    }

    return 0;
}



/*
 * Entrypoint, called from xcptlow.s.
 */
int
_xcpt_deliver (struct xcptcontext *xcp)
{
    int xcode, sig;

#if defined(XCPTCACHEERR) && defined(XCPC_CACHEERR)
    if (XCP_CLASS (xcp) == XCPC_CACHEERR)
	xcode = XCPTCACHEERR;
    else
	xcode = XCP_CODE (xcp);
#else
    xcode = XCP_CODE (xcp);
#endif
    if (sig = (xcpthandlertab[xcode]) (xcode, xcp)) {
	if (sig = _debug_handler (sig, xcp))
	    _sig_raise (0, sig, xcp);
    }
    _sig_deliver (xcp);
    return 0;
}
