#ident "kit/share/xcptstack.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1999-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * xcptstack.c: return the successive entries in the call stack
 */


#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <mips/instrdef.h>
#if __mips16 || __maybe_mips16
#include <mips/instr16def.h>
#endif
#include "kit_impl.h"

#undef DBGXCPT
#ifdef DBGXCPT
#define DBG(fmt, args...) \
	_mon_printf (fmt , ## args);
#else
#define DBG(fmt, args...)
#endif

/* instruction code for 32-bit "jr ra" */
#define JR_RA	((ITYPE_special << 26) | (REG_RA << 21) | RRTYPE_jr)

/* max insns in prologue */
#define MAX_PROLOGUE	32		

/* "fixed" register numbers */ 
#define REG_SP	29		/* frame pointer == s8 */
#define REG_FP	30		/* stack pointer */
#define REG_RA	31		/* return addres */

static struct savedreg {
    void	*addr;
    short	base;
    short	size;
    reg_t	val;
} regs[32];


/* FIXME these should probably be global */
#define IS_MIPS16(va) ((unsigned long)(va) & 1)
#define MIPS16ADDR(va)  ((void *)((unsigned long)(va) & ~1))

/* convert MIPS16 register to MIPS32 register */
#define mips16_to_mips32(rn)	(((rn) < 2) ? (rn) + 16 : (rn))


static unsigned int
mips_get_instruction (void *va, int *err)
{
#if __mips16 || __maybe_mips16
    if (IS_MIPS16(va))
	return mips_get_half (MIPS16ADDR(va), err);
#endif
    if ((unsigned long)va & 3) {
	*err = 1;
	return 0;
    }
    return mips_get_word (va, err);
}

static int
getsavedreg (int rn, reg_t *rv)
{
    int saved = 0;

    if (regs[rn].addr) {
	int err = 0;
#if (__mips >= 3 && __mips != 32) || __mips64
	if (regs[rn].size == 8)
	  regs[rn].val = mips_get_dword (regs[rn].addr, &err);
	else
#endif
	  regs[rn].val = mips_get_word (regs[rn].addr, &err);
	regs[rn].addr = 0;
	saved = !err;
    }
    *rv = regs[rn].val;
    return saved;
}



static void
setregsave (int rn, int base, void *addr, int size)
{
    DBG("reg%d stored at %p (from reg%d) size=%d\n",
	rn, addr, base, size);
    regs[rn].addr = addr;
    regs[rn].base = base;
    regs[rn].size = size;
}


static void
adjregsaves (int obase, int nbase, int adjust)
{
    int r;

    for (r = 0; r < 32; r++) {
	if (regs[r].addr && regs[r].base == obase) {
	    DBG("adjust: reg%d was at %p now at %p (from r%d)\n",
		r, regs[r].addr, regs[r].addr += adjust, adjust);
	    regs[r].addr += adjust;
	    regs[r].base = nbase;
	}
    }
}

static unsigned long
xcpsignature (struct xcptcontext *xcp)
{
    unsigned long sig;
    unsigned long *p;

    /* This is probably very mathematically unsound... */
    for (sig = 0, p = (unsigned long *)xcp; p < (unsigned long *)(xcp+1); p++)
	sig = sig ^ *p;
    return sig;
}

/*
 * Return stack information for a given depth
 */
enum XcptStackTraceStates
xcptstackinfo (struct xcptcontext *xcp, int atdepth, void **vap)
{
    static void *pc, *sp, *fp, *ra;
    static unsigned long xcpsig;
    static int depth;

    union mipsn_instr insn;
#if __mips16 || __maybe_mips16
    union mips16_instr insn16;
    int width;
    unsigned int extended = 0, extension = 0;
    unsigned int seen_adjsp = 0;
    int immediate;
    int reg;
#endif

    extern void _xcptcall(void);
    void *va = 0;
    void *base;
    unsigned long mask;
    int cont, stksize, adjust;
    int instrlen;
    reg_t rv;
    int rn;
    int err = 0;

    if (atdepth == 0 ||
	atdepth != depth+1 ||
	xcpsignature(xcp) != xcpsig) {
	xcpsig = xcpsignature (xcp);
	memset (regs, 0, sizeof (regs));
	sp = REG_TO_VA (xcp->regs[REG_SP]);
	fp = REG_TO_VA (xcp->regs[REG_FP]);
	ra = REG_TO_VA (xcp->regs[REG_RA]);
	pc = REG_TO_VA (xcp->epc);
	DBG ("Restarting stack backtrace pc=%p ra=%p sp=%p fp=%p\n",
	     MIPS16ADDR(pc), MIPS16ADDR(ra), sp, fp);
 	depth = -1;
    }

    DBG ("depth=%d atdepth=%d\n", depth, atdepth);
    for ( ;depth != atdepth; depth++) {
	stksize = adjust = 0;

	/* stop if stack pointer is invalid */
	if (((unsigned long)sp & (sizeof(reg_t) - 1)) || !IS_KVA01 (sp)) {
	    *vap = pc;
	    ra = 0;
	    break;
	}

#if __mips16 || __maybe_mips16
	instrlen = IS_MIPS16(pc) ? 2 : 4;
#else
	instrlen = 4;
#endif

	DBG ("Scanning for start of function, instrlen=%d\n", instrlen);

	/*
	 * Find the beginning of the current subroutine by scanning backwards
	 * from the current PC for the end of the previous subroutine.
	 */
	for (va = pc - instrlen; ; va -= instrlen) {
	    if (!IS_KVA01 (va))
		goto stackerr;
#if __mips16 || __maybe_mips16
	    if (IS_MIPS16 (va)) {
		insn16.value = mips_get_instruction (va, &err);
		if (err)
		    goto stackerr;
		DBG ("va=%p 0x%04x\n", MIPS16ADDR(va), insn16.value);
     		/* On MIPS16, any one of the following is likely to be the
		   start of a function:
		   entry
		   addiu sp,-n
		   daddiu sp,-n
		   extend -n followed by 'addiu sp,+n' or 'daddiu sp,+n'  */
		if (((insn16.value & 0xf81f) == 0xe809 &&
		     (insn16.value & 0x700) != 0x700) 				/* entry */
		    || (insn16.value & 0xff80) == 0x6480			/* save */
		    || (insn16.value & 0xff80) == 0x6380			/* addiu sp,-n */
		    || (insn16.value & 0xff80) == 0xfb80			/* daddiu sp,-n */
		    || ((insn16.value & 0xf810) == 0xf010 && seen_adjsp))	/* extend -n */
		    break;
		/* XXX extend+save test required */
		else if ((insn16.value & 0xff00) == 0x6300	/* addiu sp */
			 || (insn16.value & 0xff00) == 0xfb00)	/* daddiu sp */
		    seen_adjsp = 1;
		else
		    seen_adjsp = 0;
		continue;
	    }
#endif
	    insn.value = mips_get_instruction (va, &err);
	    if (err)
		goto stackerr;
	    DBG ("va=%p 0x%08x\n", MIPS16ADDR(va), insn.value);
	    if (insn.value == JR_RA) {
		va += 2*instrlen; /* skip back over "jr ra" & delay slot */
		break;
	    }
	}

	if (!IS_MIPS16(pc)) {
	    /* skip over nulls which might separate .o files */
	    DBG ("Skipping nulls va=%p\n", va);
	    while (mips_get_instruction(va, &err) == 0 && va < pc)
		va += instrlen;
	}

	if (pc == _xcptcall) {
	    /* in _xcptcall, get saved registers from xcpframe */
	    struct xcptcontext *nxcp;
#if _MIPS_SIM==_ABIO32
	    /* 40: XCPTCONTEXT
	       24: xcpt_gen() argspace			(16 bytes)
	       16: _xcptcall() saved ra, rounded up	( 8 bytes)
	       00: _xcptcall() argspace 		(16 bytes) */
	    nxcp = sp + 40;
#elif _MIPS_SIM==_ABIO64
	    /* 72: XCPTCONTEXT
	       40: xcpt_gen() argspace			(32 bytes)
	       32: _xcptcall() saved ra, rounded up	( 8 bytes)
	       00: _xcptcall() argspace 		(32 bytes) */
	    nxcp = sp + 72;
#else
	    /* N32, N64, etc */
	    /* 16: XCPTCONTEXT
	       16: xcpt_gen() argspace 			( 0 bytes)
	       00: _xcptcall() saved ra, rounded up	(16 bytes) */
	    nxcp = sp + 16;
#endif
	    DBG ("Following exception frame from _xcptcall\n");
	    fp = REG_TO_VA (nxcp->regs[XCP_S8]);
	    sp = REG_TO_VA (nxcp->regs[XCP_SP]);
	    ra = REG_TO_VA (nxcp->regs[XCP_RA]);
	    pc = REG_TO_VA (nxcp->epc);
	    *vap = pc;
	    continue;
	}

	/* scan forwards to find stack size and any saved registers */
	mask = 0;
	DBG ("Examine function prologue va=%p\n", MIPS16ADDR(va));
	for (cont = (pc - va)/instrlen; cont > 0 ; va += instrlen, cont--) {
#if __mips16 || __maybe_mips16
	    if (IS_MIPS16(va)) {
		insn16.value = mips_get_instruction(va, &err);
		if (err) 
		    goto stackerr;

		DBG ("va=%p 0x%04x op=%d cont=%d\n", MIPS16ADDR(va), insn16.value, insn16.itype.op, cont);
		switch (insn16.itype.op) {
		case ITYPE16_b:
		case ITYPE16_beqz:
		case ITYPE16_bnez:
		    cont = 1;	/* stop now */
		    break;

		case ITYPE16_jal:
		    /* jal/jalx is a 32bit instruction with delay slot */

		    va += instrlen; /* skip remainder of instruction */
		    cont--;

		    if (cont > 2)
			cont = 2;	/* stop after next */
		    break;
		case ITYPE16_i8:
		    switch (insn16.i8type.funct) {
		    case I8TYPE16_bteqz:
		    case I8TYPE16_btnez:
			cont = 1;	/* stop now */
			break;

		    case I8TYPE16_swrasp:
			width = 4;
		    sxrasp:
			if (!(mask & (1 << REG_RA))) {
			    mask |= 1 << REG_RA;
			    if (extended)
				immediate = (short) (((extension & 0x001f) << 11) |
						     (extension & 0x07e0) |
						     (insn16.i8type.imm & 0x1f));
			    else
				immediate = ((unsigned)insn16.i8type.imm) << 2;
			    DBG("sxrasp sp=%p immediate=%d extended=%d\n",
				sp, immediate, extended);
			    setregsave (REG_RA, REG_SP, sp + immediate, width);
			}
			break;
			
		    case I8TYPE16_adjsp:
		    adjsp:
		    if (extended)
			immediate = (short) (((extension & 0x001f) << 11) |
					     (extension & 0x07e0) |
					     (insn16.i8type.imm & 0x1f));
		    else
			immediate = insn16.i8type.imm << 3;
		    DBG("adjsp: immediate=%d\n", immediate);
		    /* adjust previous saves, based on old val */
		    adjregsaves (REG_SP, REG_SP, -immediate);
		    /* but we assume that current sp is now correct */
		    stksize += -immediate;
		    break;
		    }
		    break;

		case ITYPE16_swsp:
		    /* look for saved registers and store a pointer to them */
		    width = 4;
		sxsp:
		    reg = mips16_to_mips32(insn16.ritype.rx);
		    if (!(mask & (1 << reg))) {
			mask |= 1 << reg;
			if (extended)
			    immediate = (short) (((extension & 0x001f) << 11) |
						 (extension & 0x07e0) |
						 (insn16.i8type.imm & 0x1f));
			else
			    immediate = ((unsigned)insn16.i8type.imm) << 2;
			setregsave (reg, REG_SP, sp + immediate, width);
		    }
		    break;
	    
		case ITYPE16_rr:
		    switch (insn16.rrtype.funct) {
		    case RRTYPE16_jr:
			if (cont > 1)
			    cont = 2;
			break;
		    case RRTYPE16_break:
			cont = 1;
			break;
		    case RRTYPE16_entry:
			/* FIXME */
			;
		    }
		    break;

		case ITYPE16_extend:
		    extended = 1;
		    extension = insn16.exttype.imm; /* we'll manipulate it later */
		    continue;		/* so extended doesn't get reset */

		case ITYPE16_i64:
		    switch (insn16.i64type.funct) {
		    case I64TYPE16_sdsp:
			width = 8;
			goto sxsp;
			
		    case I64TYPE16_sdrasp:
			width = 8;
			goto sxrasp;

		    case I64TYPE16_dadjsp:
			goto adjsp;
		    }

		}

		extended = 0;
		continue;

	    }
#endif
	    insn.value = mips_get_instruction(va, &err);
	    if (err) 
		goto stackerr;

	    switch (insn.itype.op) {
	    case ITYPE_special:
		switch (insn.rrtype.funct) {
		case RRTYPE_jr:
		case RRTYPE_jalr:
		    if (cont > 2)
			cont = 2; /* stop after next instruction */
		    break;
		
		case RRTYPE_syscall:
		case RRTYPE_break:
		    cont = 1; /* stop now */
		    break;

		case RRTYPE_or:
		case RRTYPE_daddu:
		case RRTYPE_addu:
		    /* one of RS or RT must be $0, for this to be a MOVE */
		    if (insn.rrtype.rt == 0)
			rn = insn.rrtype.rs;
		    else if (insn.rrtype.rs == 0) 
			rn = insn.rrtype.rt;
		    else
			break;
		
		    if (insn.rrtype.rd == REG_FP && rn == REG_SP) {
			/* move $fp,$sp */
			/* check for use of alloca ($sp != $fp) */
			if (fp > sp) {
			    /* base previous register-saves on more reliable $fp */
			    adjregsaves (REG_SP, REG_FP, adjust = fp - sp);
			    /* get the correct value of $sp at this point */
			    sp += adjust;
			}
		    }
		    break;
		}
		break;

	    case ITYPE_regimm:
		switch (insn.itype.rt) {
		case RITYPE_bltz:
		case RITYPE_bgez:
		case RITYPE_bltzl:
		case RITYPE_bgezl:
		case RITYPE_bltzal:
		case RITYPE_bgezal:
		case RITYPE_bltzall:
		case RITYPE_bgezall:
		    if (cont > 2)
			cont = 2;		/* stop after next instruction */
		    break;
		
		case RITYPE_tgei:
		case RITYPE_tgeiu:
		case RITYPE_tlti:
		case RITYPE_tltiu:
		case RITYPE_teqi:
		case RITYPE_tnei:
		    cont = 1;		/* stop now */
		    break;
		}
		break;

	    case ITYPE_c0:
	    case ITYPE_c1:
	    case ITYPE_c2:
	    case ITYPE_c3:
		if (insn.itype.rs == COPZRS_bc)
		    switch (insn.itype.rt & 3) {
		    case COPZRT_BCF:
		    case COPZRT_BCT:
		    case COPZRT_BCFL:
		    case COPZRT_BCTL:
			if (cont > 2)
			    cont = 2;	/* stop after next instruction */
		    }
		break;

	    case ITYPE_sw:
	    case ITYPE_sd:
		/* look for saved registers and store a pointer to them */
		if (insn.itype.rs == REG_SP)
		    base = sp;
		else if (insn.itype.rs == REG_FP)
		    base = fp;
		else
		    break;
		if (!(mask & (1 << insn.itype.rt))) {
		    mask |= 1 << insn.itype.rt;
		    setregsave (insn.itype.rt, insn.itype.rs,
				base + insn.itype.imm,
				(insn.itype.op == ITYPE_sw) ? 4 : 8);
		}
		break;
	    
	    case ITYPE_addiu:
	    case ITYPE_daddiu:
		/* stack or frame pointer adjustment? */
		if (insn.itype.rs == REG_SP) {
		    if (insn.itype.rt == REG_SP) {	/* addu $sp,size */
			/* adjust previous saves, based on old val */
			adjregsaves (REG_SP, REG_SP, -insn.itype.imm);
			/* but we assume that current sp is now correct */
			stksize += -insn.itype.imm;
		    } else if (insn.itype.rt == REG_FP) { /* addu $fp,$sp,size */
			/* check for use of alloca ($sp + size != $fp) */
			if (fp > sp + insn.itype.imm) {
			    /* base previous register-saves on more reliable $fp */
			    adjust = fp - (sp + insn.itype.imm);
			    adjregsaves (REG_SP, REG_FP, adjust);
			    /* we get the correct value of $sp at this point */
			    sp += adjust;
			}
		    }
		}
		break;
	    }
	}

	/* fetch saved fp and ra registers from stack */
	DBG ("Examination stopped at %p\n", MIPS16ADDR(va));

	if (getsavedreg (REG_FP, &rv))
	    fp = REG_TO_VA (rv);
	if (getsavedreg (REG_RA, &rv))
	    ra = REG_TO_VA (rv);

	*vap = pc;
	if (ra) {
	    if (pc == ra && stksize == 0)
		return XcptStackTraceLoop;
	    else {
		pc = ra; ra = 0;
		sp += stksize;
		DBG ("Now pc=%p ra=%p sp=%p fp=%p stksize=%d\n", MIPS16ADDR(pc), MIPS16ADDR(ra), sp, fp, stksize);
	    }
	}
	else
	    break;
    }

    return (depth != atdepth) ? XcptStackTraceLast : XcptStackTracePC;

 stackerr:
    *vap = va;
    return XcptStackTraceBadPC;
}


/* call this function to force this code to be loaded */
void
_xcptstackinfo_load ()
{
}
