#ident "kit/yamon/yamontty.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2001-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * yamontty.c: Posix style terminal i/o under MIPS YAMON
 *
 * Not used or tested yet. Outstanding problems: there's only access
 * to the "tty0" port; there's no API to change baud rate etc; I don't
 * know if getchar() allows us to handle input like control-c, xon/xoff, 
 * etc, or whether YAMON itself intercepts and interprets it.
 */


#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/init.h>
#include <string.h>
#include <termios.h>
#include <assert.h>
#include <stdlib.h>
#include "file.h"
#include "tty.h"
#include "kit_impl.h"
#include "yamon.h"

static int	ttyread (struct file *, struct uio *);
static int	ttywrite (struct file *, struct uio *);
static int	ttyioctl (struct file *, int, void *);
static int	ttyselect (struct file *, int);
static int	ttyopen (struct file *, dev_t, int);
static int	ttyclose (struct file *);

static const struct fileops _tty_ops = 
{
    ttyread, 
    ttywrite, 
    ttyioctl, 
    ttyselect, 
    ttyopen,
    ttyclose,
};


/* Only 1 port supported by API currently! */
#define NTTY 1
static struct tty *ttys[NTTY];




/*
 * The lowest level device-dependent routines, called by the generic
 * tty code.
 */

/* set/get device paramaters */
static int
ttyparams (struct tty *tp, int set, struct termios *tio)
{
    if (set) {
	/* CHANGE SETTINGS - only 38400 baud supported */
	if (tio->c_ispeed != tio->c_ospeed
	    || tio->c_ospeed != 38400)
	    return EINVAL;
    } else {
	/* GET SETTINGS */
	tio->c_cflag = TTYDEF_CFLAG;
	tio->c_ispeed = tio->c_ospeed = 38400;
    }

    return 0;
}


/* poll for input */
static int
ttyrxpoll (struct tty *tp, int canblock)
{
    char c;

    if (yamon_getchar (tp->t_dev, &c))
	return c | 0x100;
    return 0;
}


/* test for transmit ready */
static int
ttytxrdy (struct tty *tp)
{
    /* we can't tell, assume always ready! */
    return 1;
}


/* transmit one character */
static void
ttytx (struct tty *tp, int c)
{
    char cb = c;
    yamon_print_count (tp->t_dev, &cb, 1);
}


/* transmit buffer */
static int
ttytxblk (struct tty *tp, const char *bp, int nc)
{
    /* limit to 16 bytes to allow input poll for XON/XOFF etc */
    if (nc > 16) 
	nc = 16;
    yamon_print_count (tp->t_dev, bp, nc);
    return nc;
}



/* 
 * Top-level file i/o interface - indirect to generic "tty" code
 * in ../share/tty.c
 */

static int
ttywrite (struct file *fp, struct uio *uio)
{
    /* generic write */
    return _ttywrite ((struct tty *)fp->f_data, uio, fp->f_flag);
}

static int
ttyread (struct file *fp, struct uio *uio)
{
    /* generic read */
    return _ttyread ((struct tty *)fp->f_data, uio, fp->f_flag);
}

static int
ttyioctl (struct file *fp, int cmd, void *arg)
{
    /* generic ioctl */
    return _ttyioctl ((struct tty *)fp->f_data, cmd, arg);
}

static int
ttyselect (struct file *fp, int rw)
{
    /* generic select */
    return _ttyselect ((struct tty *)fp->f_data, rw);
}

static int
ttyopen (struct file *fp, dev_t dev, int fflags)
{
    struct tty *tp;
    int err = 0;

    if (dev >= NTTY)
	return ENXIO;

    /* initialise our structure */
    if (!(tp = ttys[dev])) {
	err = _ttymalloc ("tty(yamon)", &tp);
	if (err)
	    return err;
	ttys[dev] = tp;
    }

    if (!(tp->t_state & TS_OPEN)) {
	tp->t_dev	= dev;
	tp->t_params	= ttyparams;
	tp->t_rxpoll	= ttyrxpoll;
	tp->t_txrdy	= ttytxrdy;
	tp->t_txblk	= ttytxblk;
	tp->t_tx	= ttytx;
	tp->t_addr	= (void *) 0;	/* device-specific data */
    }

    /* point file at our private data */
    fp->f_data = tp;

    /* generic open */
    return _ttyopen (tp);
}


static int
ttyclose (struct file *fp)
{
    int error = 0;
    if (fp->f_data)
	error = _ttyclose ((struct tty *) fp->f_data);
    return error;
}



/* /dev/tty*:	crw-rw-r-- */
#define TTYMODE	(S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IWOTH)

static void
ttydev_mount (void)
{
    dir_t ttynode;
    int err;

    memset (&ttynode, 0, sizeof (ttynode));
    ttynode.d_mode = TTYMODE;
    ttynode.d_ops = &_tty_ops;

#define ADDDESC(name, dev)				\
    ttynode.d_name = name;				\
    ttynode.d_dev = dev;				\
    err = _mknod ("/dev", &ttynode);			\
    if (err && err != EEXIST) {				\
      _mon_puts ("mknod(/dev/" name ") failed\n");	\
      abort ();						\
    }

    ADDDESC("console",	0);
    ADDDESC("tty",	0);
    ADDDESC("tty0",	0);
}

DECL_INIT_FUNC(IF_MOUNT, ttydev_mount)

/* call this dummy function to load this module */
void _ttydev_load (void);
void _ttydev_load (void) {}
