#ident "lib/c/std/malloc.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * malloc.c : stdlib function
 */

#define _POSIX_C_SOURCE 199309L

#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/posix.h>
#include <sdethread.h>

/*
 * Yet another malloc/free/realloc
 *
 * for small blocks uses
 *      bucket allocator Chris Kingsley Berkely Malloc
 * for large blocks uses 
 *      cartesian tree "Fast Fits" C.J Stephenson IBM Sys Journal
 */

/*#define MAXBUCKETBLK (pagesize/2) */
#define MAXBUCKETBLK	16

/* this is non POSIX !! */
extern void    *_sbrk (size_t nbytes);
extern void    *_sbrkhi (size_t nbytes);

static void    *getcore (size_t nbytes, int hi_p);

static __sdethread_once_t	malloc_once = __SDETHREAD_ONCE_INIT;
static unsigned long      pagesize;	/* of this machine */
static unsigned long      pagemask;

#define PAGEROUND_UP(x) (((x) + pagemask) & ~pagemask)
#define PAGEROUND_DOWN(x) ((x) & ~pagemask)
#define PAGEALIGNP(x)	(((unsigned long)(x) & pagemask) == 0)

static struct {
    unsigned long   bmem;
    unsigned long   tmem;
    void           *lower;
    void           *upper;
    void           *lower_hi;
    void           *upper_hi;
} minfo;

#define nextblk(p,size) (((void *)(p)) + size)

/*
 * Alignment is done simply by making all malloc
 * request a multiple of this size
 */
#define ALIGN	sizeof(double)

/*
 * We keep several buckets which are blists of small blocks
 * in each bucket is a list of free blocks
 */
struct bblk {
    struct bblk    *next;
};

#define NBUCKET 16

struct bucket {
  struct bblk    *first;	/* freelist  */
  int             bsize;	/* size of entries in this bucket */
  int             nalloc;	/* number allocated via this bucket */
  __sdethread_mutex_t mx;		/* lock on bucket */
};
static struct bucket bucket[NBUCKET];

/* 
 * FBUCKET holds tree allocated data pending free
 * TBUCKET holds free tree nodes
 * MBUCKET0 holds free malloc'able blocks
 */

#define FBUCKET	1
#define TBUCKET 2
#define MBUCKET0 3		/* bucketSize == 1 << 3 */

/*
 * Generate a free list of blocks of a given size
 * works best for blocks which are a divisible by page size
 * the list is highly suitable for putting in buckets.
 */

static struct bblk *
genblist (size_t size)
{
    struct bblk    *first, *bb;

    if (size < pagesize) {
	int             nblk = pagesize / size;
	int             n;

	if ((first = (struct bblk *)getcore (pagesize, 1)) == NULL)
	    return NULL;
	minfo.bmem += pagesize;

	for (bb = first, n = 1; n < nblk; n++, bb = bb->next)
	    bb->next = (struct bblk *)nextblk (bb, size);
	bb->next = NULL;
    } else {			/* just generate one entry */
	size = PAGEROUND_UP (size);
	if ((first = (struct bblk *)getcore (size, 1)) == NULL)
	    return NULL;
	minfo.bmem += size;
	first->next = NULL;
    }
    return first;
}

static void    *
bucketAlloc (int bno)
{
    struct bucket  *b = &bucket[bno];
    struct bblk    *bb;

    assert (bno < NBUCKET && (b->bsize || bno == FBUCKET));

    if(__sdethread_mutex_lock(&b->mx) != 0)
        return NULL;
    
    if (!(bb = b->first)) {
	bb = b->first = genblist (b->bsize);
	if (!bb) {
	    __sdethread_mutex_unlock(&b->mx);
	    return NULL;
	}
    }
    b->first = bb->next;
    b->nalloc++;
    __sdethread_mutex_unlock(&b->mx);
    
    return (void *)bb;
}

static void 
bucketFree (int bno, struct bblk *bb)
{
    struct bucket  *b = &bucket[bno];

    assert (bno < NBUCKET && (b->bsize || bno == FBUCKET));

    if(__sdethread_mutex_lock(&b->mx) == 0) {
      bb->next = b->first;
      b->first = bb;
      b->nalloc--;
      __sdethread_mutex_unlock(&b->mx);
    }
}

/*
 * the tree allocations routines maintain a tree of available
 * space only. treeFree is the main point by which you add free space
 * to the tree. treeAalloc allocates space out of the tree.
 */

/*
 * Tree nodes are created for free blocks only
 */

struct tnode {
    struct tnode   *left;	/* Left tree pointer */
    struct tnode   *right;	/* Right tree pointer */
    void           *block;	/* Ptr to the data block */
    size_t          size;	/* Size of the data block */
};
static struct tnode *troot;

#define weight(x)	((x)?x->size:0)

#if MAXBUCKETBLK >= 8
#define SMALLEST_TBLK	(MAXBUCKETBLK + ALIGN)
#else
#define SMALLEST_TBLK	16
#endif

#define freeTnode(p) bucketFree(TBUCKET,(struct bblk *)(p));

static struct tnode *
allocTnode ()
{
    struct tnode   *p = (struct tnode *)bucketAlloc (TBUCKET);

    if (!p)			/* cant do much if we cant free! */
	abort ();
    return p;
}

static void     demote (struct tnode **p);
static void     insert (void *p, size_t len, struct tnode **tree);
static void     delete (struct tnode **p);
static int      tfillup (size_t nbytes);


/*
 * Allocate nbytes or more out of the tree return ptr to
 * bytes (first word is assigned actual bytes alloced)
 */
static void    *
treeAlloc (size_t nbytes)
{
    struct tnode   *a;		/* ptr to node to be allocated */
    struct tnode   **p;		/* address of ptr to node */
    size_t          left_weight;
    size_t          right_weight;
    struct tnode   *left_son;
    struct tnode   *right_son;
    void           *retblock;	/* Address returned to the user */
    size_t          retsize;	/* Actual size returned to user */

    /*
     * ensure that at least one block is big enough to satisfy
     *  the request.
     */
    if (weight (troot) < nbytes) {
	/*
	 * the largest block is not enough. 
	 */
	if (!tfillup (nbytes))
	    return (NULL);
    }

    /*
     * search down through the tree until a suitable block is
     * found.  At each decision point, select the better
     * fitting node.
     */

    p = (struct tnode * *)&troot;
    a = *p;
    left_son = a->left;
    right_son = a->right;
    left_weight = weight (left_son);
    right_weight = weight (right_son);

    while (left_weight >= nbytes || right_weight >= nbytes) {
	if (left_weight <= right_weight) {
	    if (left_weight >= nbytes) {
		p = &a->left;
		a = left_son;
	    } else {
		p = &a->right;
		a = right_son;
	    }
	} else {
	    if (right_weight >= nbytes) {
		p = &a->right;
		a = right_son;
	    } else {
		p = &a->left;
		a = left_son;
	    }
	}
	left_son = a->left;
	right_son = a->right;
	left_weight = weight (left_son);
	right_weight = weight (right_son);
    }				/*while */

    /*
     * allocate storage from the selected node.
     */

    /*
     * split the node, allocating nbytes from the top.
     */
    if (a->size - nbytes >= SMALLEST_TBLK) {
	struct tnode   *x = allocTnode ();

	x->block = nextblk (a->block, nbytes);
	x->size = a->size - nbytes;
	x->left = a->left;
	x->right = a->right;
	/*
	 * the node pointed to by *p has become smaller;
	 * move it down to its appropriate place in
	 * the tree.
	 */
	*p = x;
	demote (p);
	retblock = a->block;
	retsize = nbytes;
	freeTnode (a);
    } else {
	retblock = a->block;
	retsize = a->size;
	delete (p);
    }
    *(size_t *)retblock = retsize;	/* Argh this is horrible */
    return retblock;
}


/*
 * Return a block to the free space tree.
 * 
 * algorithm:
 *      Starting at the root, search for and coalesce free blocks
 *      adjacent to one given.  When the appropriate place in the
 *      tree is found, insert the given block.
 */
static void 
treeFree (void *ptr, size_t nbytes)
{
    struct tnode   **np;	/* For deletion from free list */
    struct tnode   *neighbor;	/* Node to be coalesced */
    void           *neigh_block;	/* Ptr to potential neighbor */
    size_t          neigh_size;	/* Size of potential neighbor */

    assert (nbytes >= SMALLEST_TBLK && (nbytes & (ALIGN - 1)) == 0);
    /*
     * Search the tree for the correct insertion point for this
     *  node, coalescing adjacent free blocks along the way.
     */
    np = &troot;
    neighbor = *np;
    while (neighbor != NULL) {
	neigh_block = (void *)neighbor->block;
	neigh_size = neighbor->size;
	if (ptr < neigh_block) {
	    if (ptr + nbytes == neigh_block) {
		/*
		 * Absorb and delete right neighbor
		 */
		nbytes += neigh_size;
		delete (np);
	    } else if (ptr + nbytes > neigh_block) {
		/*
		 * The block being freed overlaps
		 * another block in the tree.  This
		 * is bad news.
		 */
		abort ();
	    } else {
		/*
		 * Search to the left
		 */
		np = &neighbor->left;
	    }
	} else if (ptr > neigh_block) {
	    if (neigh_block + neigh_size == ptr) {
		/*
		 * Absorb and delete left neighbor
		 */
		ptr = neigh_block;
		nbytes += neigh_size;
		delete (np);
	    } else if (neigh_block + neigh_size > ptr) {
		/*
		 * This block has already been freed
		 */
		abort ();
	    } else {
		/*
		 * search to the right
		 */
		np = &neighbor->right;
	    }
	} else {
	    /*
	     * This block has already been freed
	     * as "ptr == neigh_block"
	     */
	    abort ();
	}			/*else */
	neighbor = *np;
    }				/*while */

    /*
     * Insert the new node into the free space tree
     */
    insert (ptr, nbytes, &troot);
}

/*
 * fill the tree with at least nbytes by using getcore
 */
static int 
tfillup (size_t nbytes)
{
    void           *p;

    /* round nbytes to pagesz */
    nbytes = PAGEROUND_UP (nbytes);

    if ((p = getcore (nbytes, 0))) {
	minfo.tmem += nbytes;
	treeFree (p, nbytes);
	return 1;
    }

    return 0;
}

/*
 * Insert a new node in a cartesian tree or subtree, placing it
 *      in the correct position with respect to the existing nodes.
 *
 * algorithm:
 *      Starting from the root, a binary search is made for the new
 *      node. If this search were allowed to continue, it would
 *      eventually fail (since there cannot already be a node at the
 *      given address); but in fact it stops when it reaches a node in
 *      the tree which has a length less than that of the new node (or
 *      when it reaches a null tree pointer).  The new node is then
 *      inserted at the root of the subtree for which the shorter node
 *      forms the old root (or in place of the null pointer).
 */

static void 
insert (
	   void *p,		/* block to insert */
	   size_t len,		/* Length of new node */
	   struct tnode **tree)
{				/* Address of ptr to root */
    struct tnode   *x;
    struct tnode   **left_hook;	/* Temp for insertion */
    struct tnode   **right_hook;	/* Temp for insertion */
    struct tnode   *newnode;

    assert ((len & (ALIGN - 1)) == 0);
    assert (((size_t)p & (ALIGN - 1)) == 0);

    x = *tree;

    /*
     * Search for the first node which has a weight less
     *  than that of the new node; this will be the
     *  point at which we insert the new node.
     */

    while (weight (x) >= len) {
	if (p < x->block)
	    tree = &x->left;
	else
	    tree = &x->right;
	x = *tree;
    }

    /*
     * Perform root insertion. The variable x traces a path through
     *  the tree, and with the help of left_hook and right_hook,
     *  rewrites all links that cross the territory occupied
     *  by p.  Note that this improves performance under
     *  paging.
     */

    newnode = allocTnode ();
    *tree = newnode;
    left_hook = &newnode->left;
    right_hook = &newnode->right;

    newnode->left = NULL;
    newnode->right = NULL;
    newnode->block = p;
    newnode->size = len;

    while (x != NULL) {
	/*
	 * Remark:
	 *      The name 'left_hook' is somewhat confusing, since
	 *      it is always set to the address of a .right link
	 *      field.  However, its value is always an address
	 *      below (i.e., to the left of) p. Similarly
	 *      for right_hook. The values of left_hook and
	 *      right_hook converge toward the value of p,
	 *      as in a classical binary search.
	 */
	if (x->block < p) {
	    /*
	     * rewrite link crossing from the left
	     */
	    *left_hook = x;
	    left_hook = &x->right;
	    x = x->right;
	} else {
	    /*
	     * rewrite link crossing from the right
	     */
	    *right_hook = x;
	    right_hook = &x->left;
	    x = x->left;
	}			/*else */
    }				/*while */

    *left_hook = *right_hook = NULL;	/* clear remaining hooks */

}				/*insert */


/*
 * Delete a node from a cartesian tree. p is the address of
 *      a pointer to the node which is to be deleted.
 *
 * algorithm:
 *      The left and right sons of the node to be deleted define two
 *      subtrees which are to be merged and attached in place of the
 *      deleted node.  Each node on the inside edges of these two
 *      subtrees is examined and longer nodes are placed above the
 *      shorter ones.
 *
 * On entry:
 *      *p is assumed to be non-null.
 */

static void 
delete (struct tnode **p)
{
    struct tnode   *x;
    struct tnode   *left_branch;	/* left subtree of deleted node */
    struct tnode   *right_branch;	/* right subtree of deleted node */

    x = *p;
    left_branch = x->left;
    right_branch = x->right;

    while (left_branch != right_branch) {
	/*
	 * iterate until left branch and right branch are
	 * both NULL.
	 */
	if (weight (left_branch) >= weight (right_branch)) {
	    /*
	     * promote the left branch
	     */
	    *p = left_branch;
	    p = &left_branch->right;
	    left_branch = left_branch->right;
	} else {
	    /*
	     * promote the right branch
	     */
	    *p = right_branch;
	    p = &right_branch->left;
	    right_branch = right_branch->left;
	}			/*else */
    }				/*while */
    *p = NULL;
    freeTnode (x);
}				/*delete */


/*
 * Demote a node in a cartesian tree, if necessary, to establish
 *      the required vertical ordering.
 *
 * algorithm:
 *      The left and right subtrees of the node to be demoted are to
 *      be partially merged and attached in place of the demoted node.
 *      The nodes on the inside edges of these two subtrees are
 *      examined and the longer nodes are placed above the shorter
 *      ones, until a node is reached which has a length no greater
 *      than that of the node being demoted (or until a null pointer
 *      is reached).  The node is then attached at this point, and
 *      the remaining subtrees (if any) become its descendants.
 *
 * on entry:
 *   a. All the nodes in the tree, including the one to be demoted,
 *      must be correctly ordered horizontally;
 *   b. All the nodes except the one to be demoted must also be
 *      correctly positioned vertically.  The node to be demoted
 *      may be already correctly positioned vertically, or it may
 *      have a length which is less than that of one or both of
 *      its progeny.
 *   c. *p is non-null
 */

static void 
demote (struct tnode **p)
{
    struct tnode   *x;		/* addr of node to be demoted */
    struct tnode   *left_branch;
    struct tnode   *right_branch;
    size_t          wx;

    x = *p;
    left_branch = x->left;
    right_branch = x->right;
    wx = weight (x);

    while (weight (left_branch) > wx || weight (right_branch) > wx) {
	/*
	 * select a descendant branch for promotion
	 */
	if (weight (left_branch) >= weight (right_branch)) {
	    /*
	     * promote the left branch
	     */
	    *p = left_branch;
	    p = &left_branch->right;
	    left_branch = *p;
	} else {
	    /*
	     * promote the right branch
	     */
	    *p = right_branch;
	    p = &right_branch->left;
	    right_branch = *p;
	}			/*else */
    }				/*while */

    *p = x;			/* attach demoted node here */
    x->left = left_branch;
    x->right = right_branch;
}				/*demote */

/*
 * Finally we have malloc/free/realloc
 */

/*
 * We directly proceed all blocks of allocated malloc data with a mov.
 * Note we also use it as a struct blist with length indic
 */
struct mov {
    unsigned short  magic;
    unsigned char   reservd;
    unsigned char   bucket;	/* 0 to NBUCKET - 1 */
    size_t          length;
};

#define ALLOCMAGIC 0xfef7	/* must make pattern an invalid ptr */

#define bucketSize(b) (1<<b)	/* bucket[b].bsize */

static struct mov zeroAlloc[2];	/* we point zero allocs here */

static void 
minit ()
{
    void           *p;
    int             b;

    pagesize = getpagesize ();
    pagemask = pagesize - 1;

    /* move the brk up to a page aligned place and mark where
     * we start
     */
    p = _sbrkhi (0);
    p = _sbrkhi ((unsigned)p - PAGEROUND_DOWN ((unsigned)p));
    minfo.upper_hi = p;
    p = _sbrk (0);
    p = _sbrk (PAGEROUND_UP ((unsigned)p) - (unsigned)p);
    minfo.lower = p;

    /* now fill in some bucket info */
    bucket[TBUCKET].bsize = sizeof (struct tnode);
    
    __sdethread_mutex_init(&bucket[TBUCKET].mx, "malloc_tb");
    __sdethread_mutex_init(&bucket[FBUCKET].mx, "malloc_fb");
    for (b = MBUCKET0; b < NBUCKET; b++) {
      bucket[b].bsize = 1 << b;
      __sdethread_mutex_init(&bucket[b].mx, "malloc_b");
    }
}

/*
 * Really free the entries on the FBUCKET free list
 */
static void 
tcompact ()
{
    struct bblk    *b;
    struct bblk    *bn;

    for (b = bucket[FBUCKET].first; b; b = bn) {
	bn = b->next;
	treeFree ((void *)b, ((struct mov *)b)->length);
    }
    bucket[FBUCKET].first = NULL;
}


void           *
malloc (size_t nbytes)
{
    struct mov     *mov;
    int             b;

    /* initialise things on first call */
    __sdethread_once (&malloc_once, minit);

    if (nbytes == 0)
	return (void *)&zeroAlloc[1];	/* into disaster */

    /* add malloc overhead */
    nbytes += sizeof (struct mov);

    if (nbytes <= MAXBUCKETBLK) {
	/* grab one out of the bucket list */

	for (b = MBUCKET0; nbytes > bucketSize (b); b++)
	    continue;
	assert (b < NBUCKET);
	
	mov = (struct mov *)bucketAlloc (b);
	if (mov) 
	  goto done;
    } 
    
    /* too big for bucket or no bucket space left */
    if (nbytes <= SMALLEST_TBLK)
      /* minimum number of bytes */
      nbytes = SMALLEST_TBLK;
    else
      /* make the number of bytes a multiple of ALIGN bytes */
      nbytes = (nbytes + (ALIGN - 1)) & ~(ALIGN - 1);

    if(__sdethread_mutex_lock(&bucket[FBUCKET].mx) != 0)
      return NULL;

    /* clear up any outstanding tree frees */
    if (bucket[FBUCKET].first)
      tcompact ();
    
    mov = (struct mov *)treeAlloc (nbytes);
    if (!mov) {
      __sdethread_mutex_unlock(&bucket[FBUCKET].mx);
      return NULL;
    }
    bucket[FBUCKET].nalloc++;	/* look as if we did */

    __sdethread_mutex_unlock(&bucket[FBUCKET].mx);
    
    /* UGH first word was filled in by treeAlloc with length 
     */
    nbytes = *(size_t *)mov;
    b = FBUCKET;
    
done:
    mov->magic = ALLOCMAGIC;
    mov->bucket = b;
    mov->length = nbytes;
    
    
    return (void *)(mov + 1);
}

void 
free (void *p)
{
    if (p != NULL) {
	struct mov     *mov = ((struct mov *)p) - 1;

	if (mov == zeroAlloc)
	    return;

	if (mov->magic != ALLOCMAGIC) 	/* bogus */
	    abort ();

	if ((void *)mov < minfo.lower ||
	    (void *)mov > minfo.upper_hi) /* paranoia */
	    abort ();

	assert (mov->bucket < NBUCKET);

	/* Always do a bucket free: we clear out the
	 * freed tree allocated data in malloc().
	 */
	bucketFree (mov->bucket, (struct bblk *)mov);
    }
}


void           *
realloc (void *p, size_t nbytes)
{
    struct mov     *mov;
    int             b;
    size_t          obytes;
    void           *np;

    if (p == NULL || (mov = ((struct mov *)p) - 1) == zeroAlloc)
	return malloc (nbytes);

    if (mov->magic != ALLOCMAGIC)
      /* oh god! realloc on a free block -> means "compaction" 
       * all mallocs refer to this by saying see the man page
       * for the old malloc. We could go and traipse down
       * the bucket free lists and unfree the freed block?
       */
      abort ();

    if ((b = mov->bucket) != FBUCKET) {
	if (nbytes < bucketSize (b) - sizeof (struct mov)) {
	    /* fits in same bucket */
	    mov->length = nbytes + sizeof (struct mov);
	    return p;
	}
    }

    obytes = mov->length - sizeof (struct mov);
    free (p);
    np = malloc (nbytes);
    if (np && np != p)
	memcpy (np, p, (nbytes < obytes) ? nbytes : obytes);
    return np;
}

/*
 * getcore --- grab a integral number of pages
 */
static void    *
getcore (size_t nbytes, int hi_p)
{
    void           *p;
    unsigned long  off;

    assert (PAGEALIGNP (nbytes));

    if (hi_p) {
        /* check brk alignment */
        p = _sbrkhi (0);

        /* offset to achieve page alignment */
        off = (unsigned long)p - PAGEROUND_DOWN ((unsigned long)p);
        nbytes += off;

        p = _sbrkhi (nbytes);
	if ((int)p == -1)
	    return NULL;

	assert (PAGEALIGNP (p));

	minfo.lower_hi = p;
	return p;
    } else {
        /* check brk alignment */
        p = _sbrk (0);

        /* offset to achieve page alignment */
        off = PAGEROUND_UP ((unsigned long)p) - (unsigned long)p;
        nbytes += off;

        p = _sbrk (nbytes);
	if ((int)p == -1)
	    return NULL;

	assert (PAGEALIGNP (p));

	minfo.upper = p + nbytes;
	return p + off;
    }
}


#ifdef MALLOCSTATS

#include <stdio.h>

/*
 * Diagnostic routines
 */
static          depth = 0;

static void 
prtree (struct tnode *p, void *cp)
{
    int             n;

    if (depth == 0) {
        fprintf (stderr, "prtree. p %p cp %s\n", p, (char *)cp);
	/*
	 * } else {
	 * fprintf(stderr, "prtree. p %x depth %d\n", p, depth);
	 */
    }

    if (p != NULL) {
	depth++;
	prtree (p->left, NULL);
	depth--;

	for (n = 0; n < depth; n++) {
	    fprintf (stderr, "   ");
	}
	fprintf (stderr,
		 "(%p): (left = %p, right = %p, block = %p, size = %lu)\n",
		 p, p->left, p->right, p->block, p->size);

	depth++;
	prtree (p->right, NULL);
	depth--;
    }
}

void 
mallocstats ()
{
    int             bno;
    int             fcount;

    fprintf (stderr, "Malloc low area %p to %p %ld bytes, high area %p to %p %ld bytes (bucket %ld tree %ld)\n",
	     minfo.lower, minfo.upper, minfo.upper - minfo.lower,
	     minfo.lower_hi, minfo.upper_hi, minfo.upper_hi - minfo.lower_hi,
	     minfo.bmem, minfo.tmem);

    for (bno = 0; bno < NBUCKET; bno++) {
	struct bblk    *bb;

	__sdethread_mutex_lock (&bucket[bno].mx);
	for (bb = bucket[bno].first, fcount = 0; bb; bb = bb->next, fcount++)
	    continue;
	__sdethread_mutex_unlock (&bucket[bno].mx);

	if (fcount > 0 || bucket[bno].nalloc)
	  fprintf (stderr, "Bucket %d Size %d Nalloc %d Nfree %d Mem %d\n",
		   bno, bucket[bno].bsize, bucket[bno].nalloc, fcount,
		   (fcount + bucket[bno].nalloc) * bucket[bno].bsize);
    }

    __sdethread_mutex_lock (&bucket[FBUCKET].mx);
    prtree (troot, "root");
    __sdethread_mutex_unlock (&bucket[FBUCKET].mx);
}
#endif
