#ident "lib/c/std/strtod.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ./strtod.c : stdlib function
 */


#include <stdlib.h>
#include <values.h>

#if defined(__NO_FLOAT)

double strtod(const char *s, char **endp)
{
    abort ();
}

#elif defined(__SOFT_FLOAT) || defined(__SINGLE_FLOAT)

double strtod(const char *s, char **endp)
{
    extern double __strtod(const char *, char **);
    return __strtod (s, endp);
}

#else

#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <float.h>

/*
 * strtod for IEEE doubles 
 */

#if _IEEE==1
#define DOIEEEHEX
static double _twoemax = 9007199254740992.;	/*2^53*/
#else
static double _twoemax = 72057594037927936.;	/*2^56*/
#endif

#if defined(DOIEEEHEX)
#include <machine/endian.h>
/* precomputed values for exp5 which are IEEE round to nearest
 * produced by repeated multiplication (with accumulated error !!)
*/
#if BYTE_ORDER == BIG_ENDIAN
/* use a union with a double so the compiler uses the correct alignment. */
union ieeehex {
    struct { unsigned int h,l; };
    double d;
};
#define IEEEHEXK(hi,lo)	{(hi),(lo)}
#else
union ieeehex {
    struct { unsigned int l,h; };
    double d;
};

#define IEEEHEXK(hi,lo)	{(lo),(hi)}
#endif

static const union ieeehex _pow5[] = {
    IEEEHEXK( 0x40140000, 0x00000000 ),	/* 5 */
    IEEEHEXK( 0x40390000, 0x00000000 ),	/* 25 */
    IEEEHEXK( 0x40838800, 0x00000000 ),	/* 625 */
    IEEEHEXK( 0x4117d784, 0x00000000 ),	/* 390625 */
    IEEEHEXK( 0x4241c379, 0x37e08000 ),	/* 152587890625 */
    IEEEHEXK( 0x4493b8b5, 0xb5056e17 ),	/* 2.3283064365387e+022 */
    IEEEHEXK( 0x49384f03, 0xe93ff9f6 ),	/* 5.42101086242753e+044 */
    IEEEHEXK( 0x52827748, 0xf9301d33 ),	/* 2.93873587705572e+089 */
    IEEEHEXK( 0x65154fdd, 0x7f73bf3f )	/* 8.63616855509445e+178 */
};
#define POW5(x) (*(double *)(&_pow5[x]))
#else
static const double	_pow5[]	= {
	5.,
	25.,
	625.,
	390625.,
	152587890625.,
	23283064365386962890625.,
#if _IEEE==1
	5.4210108624275231e+044,
	2.9387358770557196e+089,
	8.6361685550944492e+178,
#endif
};

#define POW5(x) (_pow5[x])
#endif

double strtod(const char *nptr, char **endp)
{
    const char *s = nptr;
    int mneg = 0;
    int expt = 0;
    int	nz = 0;			/* result should be non zero */
    int some = 0;
    double	v = 0;
    
    while(isspace(*s) ) 
      s++;
    
    if(*s == '-') {
	mneg = 1;
	s++;
    }else if(*s == '+') {
           s++;
    }

    /* grab ipart unlikely to overflow 
     */
    while( isdigit(*s) ) {
	int inc = *s++ - '0';
	nz |= inc;
	if(v < _twoemax)
	  v = v * 10 + inc;
	else
	  expt++;
	some = 1;
    }
    
    if( *s == '.') {
	s++;
	
	/* grab fractional part decrementing exponent
	 */
	while( isdigit(*s) ) {
	    int inc = *s++ - '0';
	    nz |= inc;
	    if(v < _twoemax) {
		v = v * 10 + inc;
		expt--;
	    }
	    some = 1;
	}
    }

    if (some && (*s == 'e' || *s == 'E')) {
	int e 		= 0;
	int eneg	= 0;
	
	s++;
	
	/* grab exponent */
	if( *s == '-') {
	    eneg = 1;
	    s++;
	} else if( *s == '+') {
	    s++;
	}
	while(isdigit(*s))
	  e = e * 10 + (*s++ - '0');
	if(eneg)
	  e = -e;
	expt += e;
    }
    
    if(endp)
      *endp = (char *) (some ? s : nptr);

    if(expt) {
	int c;
	double	flexp = 1.0;
	int	bexp = expt;

	if(expt < 0)
	  expt = -expt;
	
	for (c = 0; expt && c < sizeof(_pow5)/sizeof(_pow5[0]); c++) {
	    if (expt & 01) 
	      flexp *= POW5(c);
	    expt >>= 1; 
	}
	
	if (bexp < 0) 
	  v /= flexp;
	else 
	  v *= flexp;
	
	v = ldexp(v, bexp);
    }
    
    if (isinf (v)) {
	errno = ERANGE;
    }
    else if (v == 0.0 && nz) {
	v = 0.0;
	errno = ERANGE;
    }

    if(mneg)
      v = -v;
   
    return v;
}

#endif
