#ident "lib/c/std/strtod.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2004-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ./strtof.c : stdlib function
 */


#include <stdlib.h>
#include <stdint.h>
#include <values.h>

#if defined(__NO_FLOAT)

float strtof (const char *s, char **endp)
{
    abort ();
}

#elif defined(__SOFT_FLOAT)

float strtof (const char *s, char **endp)
{
    extern float __strtof(const char *, char **);
    return __strtof (s, endp);
}

#else

#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <float.h>

/*
 * strtof for IEEE floats
 */

static float _twoemax = 16777216.;	/*2^24*/

/* precomputed values for exp5 which are IEEE round to nearest
 * produced by repeated multiplication (with accumulated error !!)
*/
#if 0
static const float	_pow5[]	= {
	5.f,
	25.f,
	625.f,
	390625.f,
	152587890624.f,
	23283064365386962890624.f,
};
#define POW5(x) (_pow5[x])
#else
static const int32_t _pow5[] = {
    0x40a00000,	/* 5. */
    0x41c80000,	/* 25. */
    0x441c4000,	/* 625. */
    0x48bebc20,	/* 390625. */
    0x520e1bca,	/* 152587894785.p*/
    0x649dc5ae	/* 23283065137950552489985. */
};
#define POW5(x) (*(float *)(&_pow5[x]))
#endif

float strtof (const char *nptr, char **endp)
{
    const char *s = nptr;
    int mneg = 0;
    int expt = 0;
    int	nz = 0;			/* result should be non zero */
    int some = 0;
    float v = 0.0f;
    
    while (isspace (*s)) 
      s++;
    
    if (*s == '-') {
	mneg = 1;
	s++;
    }else if(*s == '+'){
           s++;
    }
    
    /* grab ipart unlikely to overflow */
    while( isdigit(*s) ) {
	int inc = *s++ - '0';
	nz |= inc;
	if (v < _twoemax)
	  v = v * 10 + inc;
	else
	  expt++;
	some = 1;
    }
    
    if( *s == '.') {
	s++;
	
	/* grab fractional part decrementing exponent
	 */
	while( isdigit(*s) ) {
	    int inc = *s++ - '0';
	    nz |= inc;
	    if (v < _twoemax) {
		v = v * 10 + inc;
		expt--;
	    }
	    some = 1;
	}
    }

    if (some && (*s == 'e' || *s == 'E')) {
	int e 		= 0;
	int eneg	= 0;
	
	s++;
	
	/* grab exponent */
	if (*s == '-') {
	    eneg = 1;
	    s++;
	} else if ( *s == '+') {
	    s++;
	}
	while (isdigit (*s))
	  e = e * 10 + (*s++ - '0');
	if (eneg)
	  e = -e;
	expt += e;
    }
    
    if(endp)
      *endp = (char *) (some ? s : nptr);

    if (expt) {
	float	flexp = 1.0f;
	int	bexp = expt;
	int c;

	if (expt < 0)
	  expt = -expt;
	
	for (c = 0; expt && c < sizeof(_pow5)/sizeof(_pow5[0]); c++) {
	    if (expt & 1) 
	      flexp *= POW5(c);
	    expt >>= 1; 
	}
	
	if (bexp < 0) 
	  v /= flexp;
	else 
	  v *= flexp;
	
	v = ldexpf (v, bexp);
    }
    
    if (mneg)
      v = -v;
    
    if (isinf (v)) {
	errno = ERANGE;
    }
    else if (v == 0.0f && nz) {
	v = 0.0f;
	errno = ERANGE;
    }
    
    return v;
}

#endif
