#ident "lib/c/stdio/_doprnt.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * vfprintf: standard I/O library
 */


#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include "stdiolib.h"

/* --- Constants --- */
#define MAX_FCONVERSION	512	/* largest possible real conversion 	*/
#define MAX_ICONVERSION	64	/* largest possible binary conversion inc zero fill */
#define MAX_EXPT	5	/* largest possible exponent field */
#define MAX_FRACT	39	/* largest possible fraction field */

#define LEFT_JUSTIFY	0x0001	/* conversion flags */
#define ALWAYS_SIGN	0x0002
#define PREFIX_SPACE	0x0004
#define ALTERNATE_FORM	0x0008
#define ZERO_FILL	0x0010
#define CHAR		0x0080	/* length modifiers */
#define SHORT		0x0100
#define LONG		0x0200
#define LONGDOUBLE	0x0400
#define LONGLONG	0x0800
#define SIGNED		0x1000	/* signed integer argument */
#define UPPER_CASE	0x4000	/* %X, %E, %G */

#define UNSPECIFIED	(-1)	/* for precision */

#define POINTER_BASE	16	/* constant for %p conversion */

typedef unsigned long long iutype;	/* unsigned max int type */
typedef signed long long istype;	/*   signed max int type */
typedef double		rtype;		/* real type */

/* --- Macros --- */
#define TESTFLAG(x)	(flags & (x))
#define CLEARFLAG(x)	(flags &= ~(x))
#define SETFLAG(x)	(flags |= (x))
#define isdigit(c)	((c) >= '0' && (c) <= '9')
#define _tolower(c)	((c) - 'A' + 'a')


static int __sputn(const char *buf,size_t n,FILE *fp)
{
  while(n--) {
    int c = *buf++;
    if (putc_unlocked(c,fp) == EOF)
	return EOF;
  }
  return 0;
}


/* convert integers */
static int _iconvert(FILE *stream,
		     iutype number,int base,
		     int width, int precision, unsigned flags);


#ifndef __NO_FLOAT
/* convert floats */
static int _fconvert(FILE *stream, rtype arg, int fmtch,
		     int width, int prec, unsigned flags);

#define _isNan(arg) (arg != arg)
#endif

static int _outcnvrt(FILE *stream,char *p, int n, 
		     int width, unsigned flags);

#define _IO_ERRVAL -1

int vfprintf(FILE *stream,const char *format,va_list ap)
{
    int 	c;
    int totalCharsOutput = 0;
    
    _arg_assert(format);

    flockfile (stream);

    _io_assert(stream, _IO_READ);

    while(1) {
      int 	width;
	int 	precision;
	int	base;
	int 	ncw;
	unsigned flags;
	
	if( (c = *format++) == '\0')
	  break;
	
	if(c != '%') {
	  putc_unlocked(c, stream);
	  totalCharsOutput++;
	  continue;
	} 
	c = *format;
	if (c == '%') {
	  putc_unlocked(c, stream);
	  format++;
	  totalCharsOutput++;
	  continue;
	}
	
	precision = UNSPECIFIED;
	flags = 0;
	ncw = 0;
	
	/* --- Extract the flags --- */
	while (1) {
	    if (c == '-')
	      SETFLAG(LEFT_JUSTIFY);
	    else if (c == '+')
	      SETFLAG(ALWAYS_SIGN);
	    else if (c == ' ')
	      SETFLAG(PREFIX_SPACE);
	    else if (c == '#')
	      SETFLAG(ALTERNATE_FORM);
	    else if (c == '0')
	      SETFLAG(ZERO_FILL);
	    else
	      break;
	    c = *++format;
	}
    
	/* The minimum field width */
	if (c == '*') {
	    width = va_arg(ap, int);
	    c = *++format;
	} else {
	    width = 0;
	    while (isdigit(c)) {
		width = 10 * width + c - '0';
		c = *++format;
	    }
	}
	if (width < 0) {
	    SETFLAG(LEFT_JUSTIFY);
	    width = 0 - width;
	}
	
	/* Conversion precision */
	if (c == '.') {
	    if ((c = *++format) == '*') {
		precision = va_arg(ap, int);
		c = *++format;
	    }
	    else {		/* PENDING - this should handle +/- signs */
		precision = 0;
		while (isdigit(c)) {
		    precision = 10 * precision + c - '0';
		    c = *++format;
		}
	    }
	    if (precision < 0)
	      precision = UNSPECIFIED;
	}
	    
	/* length modifiers */
	if (c == 'h') {
	    if(format[1] == 'h') {
		SETFLAG(CHAR);
		++format;
	    } else 
		SETFLAG(SHORT);
	    c = *++format;
	} else if (c == 'l') {
	    if(format[1] == 'l') {
		SETFLAG(LONGLONG);
		++format;
	    } else 
	      SETFLAG(LONG);
	    c = *++format;
	} else if (c == 'q') {
	    /* BSD unix quad format */
	    SETFLAG(LONGLONG);
	    c = *++format;
	} else if (c == 'L') {
	    SETFLAG(LONGDOUBLE);
	    c = *++format;
	}
	
	/* --- Fetch argument type --- */
	format++;
    
	/* --- Resolve mutually exclusive flags --- */
	if (TESTFLAG(ALWAYS_SIGN))
	  CLEARFLAG(PREFIX_SPACE);
	if (TESTFLAG(LEFT_JUSTIFY))
	  CLEARFLAG(ZERO_FILL);

	/* process by type */
	switch (c) {
	case 'D':		/* XXX - not Ansi */
	    SETFLAG(LONG);
	case 'd':
	case 'i':
	    SETFLAG(SIGNED);
	    CLEARFLAG(ALTERNATE_FORM);
	    base = 10;
	    goto bincnv;
	
	case 'O':		/* XXX - not Ansi */
	    SETFLAG(LONG);
	case 'o':
	    base = 8;
	    goto bincnv;
	
	case 'U':		/* XXX - not Ansi */
	    SETFLAG(LONG);
	case 'u':
	    CLEARFLAG(ALTERNATE_FORM);
	    base = 10;
	    goto bincnv;

	case 'X':
	    SETFLAG(UPPER_CASE);
	case 'x':
	    base = 16;
	    goto bincnv;

	case 'P':
	    SETFLAG(UPPER_CASE);
	case 'p':
	    SETFLAG(ALTERNATE_FORM);
	    if (sizeof(char *) == sizeof(long long))
	      SETFLAG(LONGLONG);
	    else
	      SETFLAG(LONG);
	    base = POINTER_BASE;
	    goto bincnv;

	    /* do conversion binary number to ascii digits 
	     * base =8,10,16
	     */
	bincnv:
	    {
		iutype l;
	    
		/* Set default precision */
		if (precision == UNSPECIFIED)
		  precision = 1;
		else
		  CLEARFLAG(ZERO_FILL);
	    
		if (TESTFLAG(SIGNED)) {
		    if (TESTFLAG(LONGLONG))
		      l = va_arg(ap, long long);
		    else if (TESTFLAG(LONG))
		      l = va_arg(ap, long);
		    else if (TESTFLAG(SHORT) || TESTFLAG(CHAR))
		     /* promoted to int */
		      l = va_arg(ap, int);
		    else
		      l = va_arg(ap, int);
		} else {
		    if (TESTFLAG(LONGLONG))
		      l = va_arg(ap, unsigned long long);
		    else if (TESTFLAG(LONG))
		      l = va_arg(ap, unsigned long);
		    else if (TESTFLAG(SHORT) || TESTFLAG(CHAR))
		      l = va_arg(ap, unsigned int);
		    else
		      l = va_arg(ap, unsigned int);
		}
	    
		ncw = _iconvert(stream, l, base, width, precision, flags);
	    }
	    break;
	    
	case 'f':
	case 'E':
	case 'e':
	case 'G':
	case 'g':
	    /* --- Set default precision --- */
	    {
#ifdef __NO_FLOAT
		(void) va_arg(ap, long long);
		ncw = _outcnvrt(stream, "(float)", 7, width, flags);
#else
		volatile rtype d;

		if (TESTFLAG(LONGDOUBLE))
		  d = va_arg(ap, long double);
		else
		  d = va_arg(ap, double);
		ncw = _fconvert(stream, d, c, width, precision, flags);
#endif
	    }
	    break;

	case 'c':
	    {
		char c = va_arg(ap, int);
		ncw = _outcnvrt(stream,&c, 1, width, flags);
	    }
	    break;
	    
	case 's':
	    {
		char *s = va_arg(ap, char *);
		char *p = s;
	    
		if(s == 0) 
		  p = s = "(null)"; /* hmm */
	    
		if (precision == UNSPECIFIED) {
		    p += strlen(s);
		} else {
		    while (--precision >= 0 && *p)
		      p++;
		}
		ncw = _outcnvrt(stream,s, p - s, width, flags);
	    }
	    break;
	case 'n':
	    if (TESTFLAG(LONGLONG))
	      *va_arg(ap, long long *) = totalCharsOutput;
	    else if (TESTFLAG(LONG))
	      *va_arg(ap, long *) = totalCharsOutput;
	    else if (TESTFLAG(SHORT))
	      *va_arg(ap, short *) = totalCharsOutput;
	    else if (TESTFLAG(CHAR))
	      *va_arg(ap, char *) = totalCharsOutput;
	    else
	      *va_arg(ap, int *) = totalCharsOutput;
	    ncw = 0;
	    break;
	default:
	    /* error return? */
	    ncw = 0;
	    break;
	}
	totalCharsOutput += ncw;
    }

    funlockfile(stream);
    return ferror (stream) ? EOF : totalCharsOutput;
}


int vprintf (const char *format, va_list ap)
{
    return vfprintf (stdout, format, ap);
}


static int _iconvert(FILE *stream,iutype number, int base, int width, 
		     int precision, unsigned flags)
{
    char buf[MAX_ICONVERSION];
    char *p = &buf[MAX_ICONVERSION - 1];
    int digit;
    int negative = 0;
    unsigned u;

    /* --- Get the sign right --- */
    if (TESTFLAG(SIGNED))
      if ((istype) number < 0LL) {
	  negative = 1;
	  number = (iutype) (0LL - (istype) number);
      }

    /* --- Build the number back to front (at the back of the buffer) --- 
     */
    if(sizeof(iutype) != sizeof(unsigned) && (u=number) == number) {
	/* calc using u */
	while (u != 0) {
	    digit = u % base;
	    if (digit < 10)
	      digit += '0';
	    else
	      digit += (TESTFLAG(UPPER_CASE) ? 'A' : 'a') - 10;
	    *p-- = digit;
	    u /= base;
	    precision--;
	}
    } else {
	while (number != 0) {
	    digit = number % base;
	    if (digit < 10)
	      digit += '0';
	    else
	      digit += (TESTFLAG(UPPER_CASE) ? 'A' : 'a') - 10;
	    *p-- = digit;
	    number /= base;
	    precision--;
	}
    }
    
    /* --- Ensure enough precision --- */
    while (--precision >= 0)
      *p-- = '0';

    /* --- Zero fill if needed (but leave room for prefix and sign) --- */
    if (TESTFLAG(ZERO_FILL)) {
	precision = width - (&buf[MAX_ICONVERSION - 1] - p);
	if (TESTFLAG(ALTERNATE_FORM))
	  if (base == 8 && *(p+1) != '0')
	    precision--;
	  else {
	    if (base == 16)
	      precision -= 2; 
	  }
	if (TESTFLAG(SIGNED))
	  if (negative || TESTFLAG(ALWAYS_SIGN | PREFIX_SPACE))
	    precision--;
	while (--precision >= 0)
	  *p-- = '0';
    }

    /* --- Generate alternate forms for %o and %x --- */
    if (TESTFLAG(ALTERNATE_FORM)) {
	if (base == 8 && *(p+1) != '0')
	  *p-- = '0';
	else if (base == 16) {
	    *p-- = TESTFLAG(UPPER_CASE) ? 'X' : 'x';
	    *p-- = '0';
	}
    }

    /* --- Generate signs --- */
    if (TESTFLAG(SIGNED)) {
	if (negative)
	  *p-- = '-';
	else if (TESTFLAG(ALWAYS_SIGN))
	  *p-- = '+';
	else if (TESTFLAG(PREFIX_SPACE))
	  *p-- = ' ';
    }
    
    p++;
    return _outcnvrt(stream,p, &buf[MAX_ICONVERSION] - p, width, flags);
}

static const char spaceData[32] = "                                ";
static const char zeroData[32] = "00000000000000000000000000000000";

static int repl(FILE *stream,size_t n,const char *data)
{
  while(n) {
    size_t nc = n;
    
    if(n > 32)
      nc = 32;
    n -= nc;
    
    if (__sputn(data,nc,stream) == EOF)
	return EOF;
  }

  return 0;
}

#define spaces(n,stream) repl(stream,n,spaceData)
#define zeroes(n,stream) repl(stream,n,zeroData)



/* output exactly n characters from p in a field of given width
 * flags provide left or right justify pad with space
 */
static int _outcnvrt(FILE *stream,char *p, int n, int width, unsigned flags)
{
    int ret = 0;

    /* pad to right justify */
    if (!TESTFLAG(LEFT_JUSTIFY) && n < width)
	ret = spaces(width-n,stream);
    
    if (ret != EOF)
	ret = __sputn(p,n,stream);

    /* pad to left justify */
    if (ret != EOF && TESTFLAG(LEFT_JUSTIFY) && n < width)
	spaces(width-n,stream);
    
    return (n > width) ? n : width;
}



#ifndef __NO_FLOAT

union ureal {
    rtype d; 
    struct {
#if defined(__ns32000__) || defined(i386) || defined(__MIPSEL__)
	unsigned manl:32;
	unsigned manh:20;
	unsigned exp:11;
	unsigned sign:1;
#else
	unsigned sign:1;
	unsigned exp:11;
	unsigned manh:20;
	unsigned manl:32;
#endif
    } s;
};

/*
 * _finite arg not Infinity or Nan
 */
static int _finite(rtype d)
{
    union ureal u = {.d = d};
    return (u.s.exp != 0x7ff);
}

static int _isneg(rtype d)
{
    union ureal u = {.d = d};
    return (u.s.sign);
}

/*
 * Floating point output, cvt() onward lifted from BSD sources:
 *
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Chris Torek.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

extern double modf(double, double *);
#define	to_char(n)	((n) + '0')
#define	to_digit(c)	((c) - '0')

static int cvt (rtype arg, int prec, unsigned flags, char *signp, int fmtch,
		char *startp, char *endp);
static char *rounded (double fract, int *exp, char *start, char *end, 
		      char ch, char *signp);
static char *exponent(char *p, int exp, int fmtch);


static int _fconvert(FILE *stream, rtype arg, int fmtch,
		     int width, int prec, unsigned flags)
{
    char	buf[MAX_FCONVERSION+1], *cp;
    char	sign;
    int		size;
    int		fldsz;
    int		fpprec;
    int 	ret;
    int		pad_zeroes;

    if( !_finite(arg) ) {
	if( _isNan(arg) )
	  size = _outcnvrt(stream,"NaN",3, width,flags);
	else if( arg < 0) 
	  size = _outcnvrt(stream,"-Inf",4, width,flags);
	else
	  size = _outcnvrt(stream,"+Inf",4, width,flags);
	return size;
    }

    fpprec = 0;
    if (prec == UNSPECIFIED)
      prec = 6;
    else if (prec > MAX_FRACT) {
	if (fmtch != 'g' && fmtch != 'G' || TESTFLAG(ALTERNATE_FORM))
	  fpprec = prec - MAX_FRACT;
	prec = MAX_FRACT;
    }

    /* leave room for sign at start of buffer */
    cp = buf + 1;

    /*
     * cvt may have to round up before the "start" of
     * its buffer, i.e. ``intf("%.2f", (double)9.999);'';
     * if the first character is still NUL, it did.
     * softsign avoids negative 0 if _double < 0 but
     * no significant digits will be shown.
     */
    *cp = '\0';
    size = cvt (arg, prec, flags, &sign, fmtch,
		cp, buf + sizeof(buf));
    if (*cp == '\0')
      cp++;

    if (TESTFLAG (ZERO_FILL)) {
	pad_zeroes = width - size;
	
	if (!sign && !TESTFLAG (ALWAYS_SIGN))
	    pad_zeroes++;
	
	while (--pad_zeroes > 0) {
	    *--cp = '0';
	    size++;
	}
    }
    
    if (sign)
      *--cp = sign, size++;
    else if (TESTFLAG(ALWAYS_SIGN))
      *--cp = '+', size++;
    else if (TESTFLAG(PREFIX_SPACE))
      *--cp = ' ', size++;

    fldsz = size + fpprec;
    ret = 0;

    /* pad to right justify */
    if (!TESTFLAG (LEFT_JUSTIFY) && !TESTFLAG (ZERO_FILL) && fldsz < width)
	ret = spaces (width - fldsz, stream);

    /* copy conversion result */
    if (ret != EOF)
	ret = __sputn(cp, size, stream);

    /* add extra precision */
    if(ret != EOF && fpprec)
	ret = zeroes(fpprec,stream);

    /* pad to left justify */
    if (ret != EOF && TESTFLAG(LEFT_JUSTIFY) && fldsz < width)
	spaces (width - fldsz, stream);
    
    return fldsz > width ? fldsz : width;
}


static int cvt (rtype number, int prec, unsigned flags, char *signp, int fmtch,
		char *startp, char *endp)
{
	register char *p, *t;
	register double fract;
	double integer, tmp;
	int dotrim, expcnt, gformat;

	dotrim = expcnt = gformat = 0;
	if (_isneg (number)) {
		number = -number;
		*signp = '-';
	} else
		*signp = 0;

	fract = modf(number, &integer);

	/* get an extra slot for rounding. */
	t = ++startp;

	/*
	 * get integer portion of number; put into the end of the buffer; the
	 * .01 is added for modf(356.0 / 10, &integer) returning .59999999...
	 */
	for (p = endp - 1; integer; ++expcnt) {
		tmp = modf(integer / 10, &integer);
		*p-- = to_char((int)((tmp + .01) * 10));
	}
	switch (fmtch) {
	case 'f':
		/* reverse integer into beginning of buffer */
		if (expcnt)
			for (; ++p < endp; *t++ = *p);
		else
			*t++ = '0';
		/*
		 * if precision required or alternate flag set, add in a
		 * decimal point.
		 */
		if (prec || TESTFLAG(ALTERNATE_FORM))
			*t++ = '.';
		/* if requires more precision and some fraction left */
		if (fract) {
			if (prec)
				do {
					fract = modf(fract * 10, &tmp);
					*t++ = to_char((int)tmp);
				} while (--prec && fract);
			if (fract)
				startp = rounded (fract, (int *)NULL, startp,
				    t - 1, (char)0, signp);
		}
		for (; prec--; *t++ = '0');
		break;
	case 'e':
	case 'E':
eformat:	if (expcnt) {
			*t++ = *++p;
			if (prec || TESTFLAG(ALTERNATE_FORM))
				*t++ = '.';
			/* if requires more precision and some integer left */
			for (; prec && ++p < endp; --prec)
				*t++ = *p;
			/*
			 * if done precision and more of the integer component,
			 * round using it; adjust fract so we don't re-round
			 * later.
			 */
			if (!prec && ++p < endp) {
				fract = 0;
				startp = rounded ((double)0, &expcnt, startp,
				    t - 1, *p, signp);
			}
			/* adjust expcnt for digit in front of decimal */
			--expcnt;
		}
		/* until first fractional digit, decrement exponent */
		else if (fract) {
			/* adjust expcnt for digit in front of decimal */
			for (expcnt = -1;; --expcnt) {
				fract = modf(fract * 10, &tmp);
				if (tmp)
					break;
			}
			*t++ = to_char((int)tmp);
			if (prec || TESTFLAG(ALTERNATE_FORM))
				*t++ = '.';
		}
		else {
			*t++ = '0';
			if (prec || TESTFLAG(ALTERNATE_FORM))
				*t++ = '.';
		}
		/* if requires more precision and some fraction left */
		if (fract) {
			if (prec)
				do {
					fract = modf(fract * 10, &tmp);
					*t++ = to_char((int)tmp);
				} while (--prec && fract);
			if (fract)
				startp = rounded (fract, &expcnt, startp,
				    t - 1, (char)0, signp);
		}
		/* if requires more precision */
		for (; prec--; *t++ = '0');

		/* unless alternate flag, trim any g/G format trailing 0's */
		if (gformat && !TESTFLAG(ALTERNATE_FORM)) {
			while (t > startp && *--t == '0');
			if (*t == '.')
				--t;
			++t;
		}
		t = exponent(t, expcnt, fmtch);
		break;
	case 'g':
	case 'G':
		/* a precision of 0 is treated as a precision of 1. */
		if (!prec)
			++prec;
		/*
		 * ``The style used depends on the value converted; style e
		 * will be used only if the exponent resulting from the
		 * conversion is less than -4 or greater than the precision.''
		 *	-- ANSI X3J11
		 */
		if (expcnt > prec || (!expcnt && fract && fract < .0001)) {
			/*
			 * g/G format counts "significant digits, not digits of
			 * precision; for the e/E format, this just causes an
			 * off-by-one problem, i.e. g/G considers the digit
			 * before the decimal point significant and e/E doesn't
			 * count it as precision.
			 */
			--prec;
			fmtch -= 2;		/* G->E, g->e */
			gformat = 1;
			goto eformat;
		}
		/*
		 * reverse integer into beginning of buffer,
		 * note, decrement precision
		 */
		if (expcnt)
			for (; ++p < endp; *t++ = *p, --prec);
		else
			*t++ = '0';
		/*
		 * if precision required or alternate flag set, add in a
		 * decimal point.  If no digits yet, add in leading 0.
		 */
		if (prec || TESTFLAG(ALTERNATE_FORM)) {
			dotrim = 1;
			*t++ = '.';
		}
		else
			dotrim = 0;
		/* if requires more precision and some fraction left */
		if (fract) {
			if (prec) {
			        do {
				    fract = modf(fract * 10, &tmp);
				    *t++ = to_char((int)tmp);
				} while(!tmp && !expcnt);
				while (--prec && fract) {
					fract = modf(fract * 10, &tmp);
					*t++ = to_char((int)tmp);
				}
			}
			if (fract)
				startp = rounded (fract, (int *)NULL, startp,
				    t - 1, (char)0, signp);
		}
		/* alternate format, adds 0's for precision, else trim 0's */
		if (TESTFLAG(ALTERNATE_FORM))
			for (; prec--; *t++ = '0');
		else if (dotrim) {
			while (t > startp && *--t == '0');
			if (*t != '.')
				++t;
		}
	}
	return (t - startp);
}


static char *
rounded (double fract, int *exp, char *start, char *end, char ch, char *signp)
{
	double tmp;

	if (fract)
		(void)modf(fract * 10, &tmp);
	else
		tmp = to_digit(ch);
	if (tmp > 4)
		for (;; --end) {
			if (*end == '.')
				--end;
			if (++*end <= '9')
				break;
			*end = '0';
			if (end == start) {
				if (exp) {	/* e/E; increment exponent */
					*end = '1';
					++*exp;
				}
				else {		/* f; add extra digit */
				*--end = '1';
				--start;
				}
				break;
			}
		}
	/* ISO C90 says that -0 should indeed be output */
#if 0
	/* ``"%.3f", (double)-0.0004'' gives you a negative 0. */
	else if (*signp == '-')
		for (;; --end) {
			if (*end == '.')
				--end;
			if (*end != '0')
				break;
			if (end == start)
				*signp = 0;
		}
#endif
	return (start);
}

static char *
exponent(char *p, int exp, int fmtch)
{
	register char *t;
	char expbuf[MAX_FCONVERSION];

	*p++ = fmtch;
	if (exp < 0) {
		exp = -exp;
		*p++ = '-';
	}
	else
		*p++ = '+';
	t = expbuf + MAX_FCONVERSION;
	if (exp > 9) {
		do {
			*--t = to_char(exp % 10);
		} while ((exp /= 10) > 9);
		*--t = to_char(exp);
		for (; t < expbuf + MAX_FCONVERSION; *p++ = *t++);
	}
	else {
		*p++ = '0';
		*p++ = to_char(exp);
	}
	return (p);
}
#endif /* !__NO_FLOAT */
