#ident "lib/c/stdio/_fwrite.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * _fwrite.c : standard I/O library
 */

/* --- _fwrite.c --- */
/* Writes directly from a buffer into a stream */
/* Notes:  Attempts to buffer requests but writes directly if impossible*/

#include <sys/types.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "stdiolib.h"

#define _IO_ERRVAL 0

/* --- The following value determines whether _fwrite() requests go	--- */
/* --- through the buffer or are done directly.  As set, the limit	--- */
/* --- means that small requests (less than half the stream's buffer	--- */
/* --- size) will use buffering; larger requests are done directly.	--- */
/* --- Empirical testing shows this is reasonable; more should be done.	--- */
#define BUFFER_LIMIT	(stream->_io_bufsiz >> 1)


static void
to_io (unsigned char *p, size_t bytes, FILE *stream)
{
    if (_io_testflag(stream, _IO_LBUF)) {
	unsigned char *nl;
	while (bytes != 0 && (nl = memchr (p, '\n', bytes))) {
	    unsigned int n = ++nl - p;
	    memcpy(stream->_io._io_ptr, p, n);
	    stream->_io._io_ptr += n;
	    stream->_io_count -= n;
	    bytes -= n;
	    _flspbuf (stream);
	    p = nl;
	}
    }
    
    if (bytes != 0) {
	memcpy(stream->_io._io_ptr, p, bytes);
	stream->_io._io_ptr += bytes;
	stream->_io_count -= bytes;
    }
}

static int
to_fs (unsigned char *p, size_t bytes, FILE *stream)
{
    size_t remain = bytes;
    ssize_t nw;

    while (remain > 0) {
	nw = write(stream->_io_fd, p, remain);
	if (nw < 0 && errno != EINTR)
	    break;
	if (nw > 0) {
	    p += nw;
	    remain -= nw;
	}
    }
    return bytes - remain;
}


#define TO_IO(bytes)	to_io(p, bytes, stream)
#define TO_FS(bytes)	to_fs(p, bytes, stream)

size_t _fwrite(const void *ptr,size_t size,size_t nmemb,FILE *stream)
{
    unsigned char *p = (unsigned char *) ptr;
    long total_size;
    int buffer_limit;
    size_t items;
    size_t n;

    if ((total_size = (long) nmemb * (long) size) == 0)
      return 0;

    _arg_assert(p != NULL);

    flockfile(stream);
    
    _io_assert(stream, _IO_READ);

    /* --- Virgin stream: set things up --- */
    if (stream->_io_count == _IO_NEVER)
      _o_init(stream);

    if (total_size == 1) {
	/* quick single character output for C++ */
        putc_unlocked (*(char *)ptr, stream);
	funlockfile(stream);
	return 1;
    }

    /* --- Most common case(?): into the buffer --- */
    buffer_limit = (stream->_io_count < BUFFER_LIMIT)
      ? stream->_io_count : BUFFER_LIMIT;
    if (total_size <= buffer_limit) {
	TO_IO(total_size);
	funlockfile(stream);
	return nmemb;
    }

    /* --- Doesn't fit: flush what we've accumulated --- */
    _flspbuf(stream);

    /* --- Huge objects: write() in manageable sections --- */
    items = 0;
    if (size > SSIZE_MAX) {
	while (nmemb-- > 0) {
	    size_t obj_size = size;

	    while (obj_size > SSIZE_MAX) {
		if ((n = TO_FS(SSIZE_MAX)) != SSIZE_MAX) {
		    _io_setflag(stream, _IO_ERR);
		    funlockfile(stream);
		    return items;
		}
		obj_size -= SSIZE_MAX;
		p += SSIZE_MAX;
	    }
	    if ((n = TO_FS(obj_size)) != obj_size) {
		_io_setflag(stream, _IO_ERR);
		funlockfile(stream);
		return items;
	    }
	    p += obj_size;
	    items++;
	}
	funlockfile(stream);
	return items;
    }

    /* --- Lots of small objects: write() in chunks --- */
    {
	size_t items_per_chunk;
	size_t chunk_size;
	size_t cutoff = stream->_io_count;

	items_per_chunk = SSIZE_MAX / size;
	if ((chunk_size = size * items_per_chunk) > cutoff)
	  cutoff = chunk_size;
	while ((size_t) total_size > cutoff) {
	    if ((n = TO_FS(chunk_size)) != chunk_size) {
		_io_setflag(stream, _IO_ERR);
		funlockfile(stream);
		return items + n / size;
	    } else {
		items += items_per_chunk;
	    }
	    total_size -= chunk_size;
	    p += chunk_size;
	}
	if ((long)(chunk_size = total_size) > BUFFER_LIMIT) {
	    if ((n = TO_FS(chunk_size)) != chunk_size)
	      _io_setflag(stream, _IO_ERR);
	    funlockfile(stream);
	    return items + n / size;
	}
	TO_IO(chunk_size);
	funlockfile(stream);
	return nmemb;
    }
}
