#ident "lib/c/stdio/_iobfuncs.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2006-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * _iobfuncs.c : simple console i/o for standard I/O library
 * 
 * Pointed to by the default _io_funcs structure in _iobdata.c.  That
 * structure * will be overridden, and the full-blown versions of
 * these functions in _iobproc.c called, iff fopen() or fdopen() are
 * used by the application.
 */

#include <stdio.h>
#include <stdlib.h>
#include "stdiolib.h"

#define _IO_ERRVAL EOF

extern void	_mon_write (const char *buf, unsigned int nb);
extern void	_mon_putc (char c);
extern int	_mon_getc (int canblock);

/* stdout line buffer */
#define STDOUT_BUFSIZ 128
static char	stdout_buf [STDOUT_BUFSIZ];

int _filbuf_simple (struct _iob *stream)
{
    int c;

    if (stream->_io_count == _IO_NEVER - 1)
	/* initialise stream */
	stream->_io._io_char = EOF;

#if 0
    if (_io_testflag(stream, _IO_EOF))
	return EOF;
#endif

    /* simple streams are always unbuffered */
    stream->_io_count = 0;

    if (stream->_io._io_char != EOF) { 
	/* return ungetc()ed char */
	c = stream->_io._io_char;
	stream->_io._io_char = EOF;
    } else {
	/* flush stdout buffer (the only possible buffered stream) */
	struct _iob *ostream = stdout;
	if (_io_testflag (ostream, _IO_LBUF))
	    _fflush_simple (ostream);

	/* TODO: line editing, signal handling */
	c = _mon_getc (1);
	switch (c) {
	case '\r':
	    c = '\n';
	    break;
#if 0
	case '\004':	/* ctrl-d */
	case '\032':	/* ctrl-z */
	    _io_setflag(stream, _IO_EOF);
	    c = EOF;
	    break;
#endif
	}
	if (_io_testflag (stream, _IO_LBUF))
	    _mon_putc (c);
    }
    return c;
}

int _flsbuf_simple (int c, struct _iob *stream)
{
    /* unbuffered stream (stderr): write one char */
    if (stream->_io_bufsiz == 0) {
	char cb = c; 
	_mon_write (&cb, 1);
	stream->_io_count = 0;
	return c;
    }

    /* note: _io_count has just been decremented */
    if (++stream->_io_count == _IO_NEVER) {
	/* first output to stream */
	if (! stream->_io_buf) {
	    /* attach local buffer, not a malloced one; this must be
	       stdout, since stderr is unbuffered. */
	    stream->_io._io_ptr = stream->_io_buf = stdout_buf;
	    stream->_io_count = stream->_io_bufsiz = STDOUT_BUFSIZ;
	    _io_setflag (stream, _IO_MYBUF); /* prevent free on buffer */
	}
    }

    /* output buffered data (must be buffer-full) */
    if (stream->_io_count == 0)
	_fflush_simple (stream);

    /* add new character to buffer */
    --stream->_io_count;
    *stream->_io._io_ptr++ = c;

    if (c == '\n' && _io_testflag (stream, _IO_LBUF))
	/* flush line-feed too */
	_fflush_simple (stream);

    return c;
}

int _fflush_simple (struct _iob *stream)
{
    if (_io_testflag(stream, _IO_WRITE)) {
	/* flush line buffer on output stream */
	if (stream->_io_count >= 0) {
	    if (stream->_io_count != stream->_io_bufsiz)
		/* flush stdout line buffer */
		_mon_write (stream->_io_buf, 
			    stream->_io_bufsiz - stream->_io_count);
	    stream->_io._io_ptr = stream->_io_buf;
	    stream->_io_count = stream->_io_bufsiz;
	}
    }
    else {
	/* flush ungetc buffer on input stream */
	stream->_io._io_char = EOF;
    }
    return 0;
}

size_t _fwrite_simple (const void *ptr, size_t size, size_t nmemb,
		       struct _iob *stream)
{
    long total_size;

    if ((total_size = (long) nmemb * (long) size) == 0)
      return 0;

    if (_io_testflag (stream, _IO_READ) != 0) {
	/* XXX is this overkill? */
	_io_setflag (stream, _IO_ERR);
	errno = EBADF;
	return 0;
    }

    _arg_assert(ptr != NULL);

    flockfile (stream);

#if 0
    /* XXX Is this workaround still necessary? */
    if (total_size == 1)
	/* single byte buffered output for GNU C++ library */
        putc_unlocked (*(char *)ptr, stream);
    else
#endif
    {
	/* output pending line buffered data */
	_fflush_simple (stream);

	/* output user data without buffering */
	_mon_write (ptr, total_size);
    }

    funlockfile (stream);

    return nmemb;
}
 
int _fsync_simple (struct _iob *stream, int mbz)
{
    /* XXX is this overkill? */
    if (_io_testflag (stream, mbz) != 0) {
	_io_setflag (stream, _IO_ERR);
	errno = EBADF;
	return 0;
    }
    return 1;
}
