#ident "lib/c/stdio/_iobproc.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * _iobproc.c : standard I/O library
 */

/* --- _iobproc.c --- */
/* Manages allocation of FILEs */
/* Notes:  Includes _ioballoc() and _iobfree()*/

#include <sdethread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "stdiolib.h"

static __sdethread_mutex_t iob_mx = __SDETHREAD_MUTEX_INITIALIZER("iob");

/*
 * The default console-only version of the _io_funcs structure is in
 * _iobdata.c. This version will override it only when one of the
 * functions in this module is (indirectly) called by the application,
 * i.e. via fopen() or fdopen().
 */

const struct _iobfuncs _io_funcs = {
    .filbuf = _filbuf,
    .flsbuf = _flsbuf,
    .fflush = _fflush,
    .fwrite = _fwrite,
    .fsync = _fsync,
};


/* _ioballoc - allocates an i/o block (FILE)	*/
/* Input:  none					*/
/* Output: Pointer to a currently free FILE	*/

FILE *_ioballoc()
{
  struct _iov *iov;
  FILE *fp;

  __sdethread_mutex_lock (&iob_mx);
  for (iov = _libc_data->_iov; iov; iov = iov->next) {
      for (fp = iov->iob; fp < &iov->iob[IOV_CHUNK]; fp++)
	  if (fp->_io_count == _IO_FREE)
	      break;
      if (fp < &iov->iob[IOV_CHUNK])
	  break;
  }
  if (!iov) {
      iov = malloc (sizeof (struct _iov));
      if (!iov) {
	  __sdethread_mutex_unlock (&iob_mx);
	  return NULL;
      }
      memset (iov, 0, sizeof (struct _iov));
      for (fp = iov->iob; fp < &iov->iob[IOV_CHUNK]; fp++)
	  fp->_io_count = _IO_FREE;
      iov->next = _libc_data->_iov;
      _libc_data->_iov = iov;
      fp = &iov->iob[0];
  }
  fp->_io_count = _IO_NEVER;
  __sdethread_mutex_unlock (&iob_mx);
  return fp;
}

/* _iobfree - frees an i/o block (FILE)	*/
/* Input:  A stream			*/
/* Output: None	(but FILE is freed)	*/

void _iobfree(FILE *stream)
{
  stream->_io_count = _IO_FREE;
}


/* _fsync - check a file handle's mode and reverse direction if required */
int _fsync (struct _iob *stream, int mbz)
{
    int valid = _io_valid (stream);

    if (valid && _io_testflag (stream, mbz) != 0) {
	assert (mbz == _IO_READ || mbz == _IO_WRITE);
	if (_io_testflag (stream, _IO_UPDATE))
	    /* flush update stream when changing direction */
	    stream->_io_funcs->fflush(stream);		
	else
	    /* normal stream cannot change direction */
	    valid = 0;
    }
    if (! valid) {
	_io_setflag (stream, _IO_ERR);
	errno = EBADF;
    }
    return valid;
}
