#ident "lib/c/stdio/flock.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * flock.c : interface SDE stdio to underlying RTOS locks
 *
 * Hide the SDE internal stdio data structures from the RTOS
 * implementation layer.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sdethread.h>
#include "stdiolib.h"

/* Make the sdethread support functions weak externs, so that we don't
   pull in the whole thread library just to satisfy these. */
#pragma weak ___sdethread_once
#pragma weak __sdethread_flock_init
#pragma weak __sdethread_flock_create
#pragma weak __sdethread_flock_lock
#pragma weak __sdethread_flock_trylock
#pragma weak __sdethread_flock_unlock
#pragma weak __sdethread_flock_release
#pragma weak __sdethread_flock_free
#pragma weak __sdethread_threading

static __sdethread_once_t flock_once = __SDETHREAD_ONCE_INIT;

static void flock_dtor (void);

static void
flock_init ()
{
    /* Register a cleanup handler for when a thread exits or is
       otherwise terminated. */
    __sdethread_flock_init (flock_dtor);
}


/* Return the file i/o lock handle for given stream.
   May be NULL for streams which don't require locks. */
static __sdethread_flock_t
fpgetlock (FILE * fp)
{
    __sdethread_flock_t fpl;

    if (_io_testflag (fp, _IO_STRING))
	/* no locking for sprintf/sscanf etc */
	return NULL;

    if (! &__sdethread_threading)
	/* no thread library present, disable all further calls */
	return NULL;

    /* Initialise flock interface to thread library */
    __sdethread_once (&flock_once, flock_init);

    if (!(fpl = fp->_io_lock))
	/* Allocate new lock */
	fpl = __sdethread_flock_create ((__sdethread_flock_t *)&fp->_io_lock, fp->_io_fd);

    return fpl;
}

/* Lock a stdio file handle - recursive locking is allowed. */
void
flockfile (FILE * fp)
{
    __sdethread_flock_t fpl = fpgetlock (fp);
    if (fpl)
	__sdethread_flock_lock (fpl);
}


/* Try to lock a stdio file handle - recursive locking is allowed. */
int
ftrylockfile (FILE * fp)
{
    __sdethread_flock_t fpl = fpgetlock (fp);
    if (fpl)
	return __sdethread_flock_trylock (fpl);
    return 0;
}


/* Unlock a stdio file handle. */
void
funlockfile (FILE * fp)
{
    __sdethread_flock_t fpl;
    if (fpl =  fp->_io_lock)
	return __sdethread_flock_unlock (fpl);
}

/* Worker function for _flock_dtor below. Release an flock owned by
   current thread when it exits, but don't free the lock, as other
   threads may be waiting for it. */
static int
flock_cleanup (FILE *fp)
{
    __sdethread_flock_t fpl;
    if (fpl = fp->_io_lock) 
	__sdethread_flock_release (fpl);
    return 0;
}

/* Callback destructor function to clean up all file locks held by a
   thread when it exits or is forcibly terminated. */
static void
flock_dtor (void)
{
    _fwalk (flock_cleanup);
}


/* Release and free an flock structure when a stream is really closed.
   Typically called by fclose(). Will be called with stream already
   flocked. */
void
_flockfree (FILE *fp)
{
    __sdethread_flock_t fpl;
    if (fpl = fp->_io_lock) {
	fp->_io_lock = NULL;
	__sdethread_flock_free (fpl);
    }
}
