#ident "lib/c/stdio/fread.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * fread.c : standard I/O library
 */

/* --- fread.c --- */
/* Reads directly from a stream into a buffer */
/* Notes:  Attempts to buffer requests but reads directly if impossible	*/

#include <sys/types.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "stdiolib.h"

#define _IO_ERRVAL 0

/* --- The following value determines whether fread() requests go	--- */
/* --- through the buffer or are done directly.  As set, the limit	--- */
/* --- means that small requests (less than half the stream's buffer	--- */
/* --- size) will use buffering; larger requests are done directly.	--- */
/* --- Empirical testing shows this is reasonable; more should be done.	--- */
#define BUFFER_LIMIT	(stream->_io_bufsiz >> 1)

#define FROM_IO(bytes) {			\
	memcpy(p, stream->_io._io_ptr, bytes);	\
	p += bytes;				\
	stream->_io._io_ptr += bytes;		\
	stream->_io_count -= bytes;		\
	}
#define FROM_FS(bytes)	({			\
	ssize_t _nr;				\
	do {					\
	    _nr = read(stream->_io_fd, p, bytes); \
	} while (_nr < 0 && errno == EINTR);	\
	_nr;					\
})

size_t (fread)(void *ptr,size_t size,size_t nmemb,FILE *stream)
{
    unsigned long total_size;
    unsigned long bytes_left;
    unsigned char *p = (unsigned char *) ptr;
    
    if ((total_size = (long) nmemb * (long) size) == 0)
      return 0;

    _arg_assert(p != NULL);

    flockfile (stream);

    _io_assert(stream, _IO_WRITE);

    /* --- Virgin stream: set things up --- */
    if (stream->_io_count == _IO_NEVER)
      _i_init(stream);

    /* --- Most common case(?): from the buffer --- */
    if (total_size <= stream->_io_count) {
	FROM_IO(total_size);
	funlockfile (stream);
	return nmemb;
    }

    bytes_left = total_size;
    /* --- Empty the buffer --- */
    if (stream->_io_count > 0) {
	bytes_left -= stream->_io_count;
	FROM_IO(stream->_io_count);
    }

    if (_io_testflag(stream, _IO_EOF)) {
	funlockfile (stream);
	return (total_size - bytes_left) / size;
    }

    /* --- Small objects: try filling the stdio buffer ONCE --- 
     */
    if (bytes_left < BUFFER_LIMIT) {
	int c;
	
	--stream->_io_count;	/* expected by _filbuf() */
	if ((c = stream->_io_funcs->filbuf(stream)) == EOF) {
	    funlockfile (stream);
	    return (total_size - bytes_left) / size;
	}
	*p++ = c; bytes_left--;
	
	if (bytes_left <= stream->_io_count) {
	    FROM_IO(bytes_left);
	    funlockfile (stream);
	    return nmemb;
	}
	/* --- strange how do we get here given BUFFER_LIMIT def
	 */
	if (stream->_io_count > 0) {
	    bytes_left -= stream->_io_count;
	    FROM_IO(stream->_io_count);
	}
    }

    /* This errs on safe size and expects read to take an int not a size_t
     */
    
    while(bytes_left > 0) {
	ssize_t	n;

	/* flush all line-buffered output before reading from terminal */
	if (_io_testflag(stream, _IO_LBUF))
	    _fwalk (_flslbuf);

	n = (bytes_left > INT_MAX) ? INT_MAX : bytes_left;

	if ((n = FROM_FS(n)) <= 0) {
	    if (n < 0) {
		_io_setflag(stream, _IO_ERR);
	    } else {
		_io_setflag(stream, _IO_EOF);
	    }
	    funlockfile (stream);
	    return (total_size - bytes_left) / size;
	}
	p += n;
	bytes_left -= n;
    }
    funlockfile (stream);
    return nmemb;
}
