/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * stdiolib.h : standard I/O library internal
 */

#ifndef __STDIOLIB_H
#define __STDIOLIB_H

#include <stdarg.h>
#include <errno.h>

/* --- Constants --- */
#define _IO_NEVER	(-1)	/* MAGIC count for virgin streams */
#define _IO_FREE	(-3)	/* MAGIC count for unused streams */
#define P_tmpdir	"/tmp"	/* directory for temporary files */

struct _iobfuncs {
    int		(*filbuf) (struct _iob *stream);
    int 	(*flsbuf) (int c, struct _iob *stream);
    int 	(*fflush) (struct _iob *stream);
    size_t 	(*fwrite) (const void *ptr, size_t size, size_t nmemb, 
			   struct _iob *stream);
    int		(*fsync) (struct _iob *stream, int mbz);
};

extern const struct _iobfuncs _io_funcs;
extern const struct _iobfuncs _io_strfuncs;
extern const struct _iobfuncs _io_strallocfuncs;

/* --- Stream initialisers --- */
#define STDIN_INIT	{_IO_NEVER, (union _iou){NULL}, NULL,	\
		   	 0, _IO_READ|_IO_LBUF|_IO_TTY, 0,	\
			 NULL, &_io_funcs}

#define STDOUT_INIT	{_IO_NEVER, (union _iou){NULL}, NULL,	\
		   	 BUFSIZ, _IO_WRITE|_IO_LBUF|_IO_TTY, 1,	\
			 NULL, &_io_funcs}

#define STDERR_INIT	{_IO_NEVER, (union _iou){NULL}, NULL,	\
		   	 0, _IO_WRITE, 2,			\
			 NULL, &_io_funcs}

#define EMPTY_INIT	{_IO_FREE, (union _iou){NULL}, NULL,	\
		  	 0, 0 , 0,				\
			 NULL, NULL}

/* --- Stream table list --- */
#define IOV_CHUNK	4
struct _iov {
    struct _iov		*next;
    FILE		iob[IOV_CHUNK];
};

/* --- Prototypes --- */
void	_bufalloc	(FILE *stream);
void	_buffree	(FILE *stream);
void	_cleanup	(void);
int	_flspbuf	(FILE *stream);
int	_flslbuf	(FILE *stream);
FILE	*_ioballoc	(void);
void	_iobfree	(FILE *stream);
void	_i_init		(FILE *stream);
void	_o_init		(FILE *stream);
int	_valmode	(const char *mode);

int	_filbuf_simple	(FILE *stream);
int	_filbuf		(FILE *stream);
int	_flsbuf_simple	(int, FILE *stream);
int	_flsbuf		(int, FILE *stream);
int	_fflush_simple	(FILE *stream);
int	_fflush		(FILE *stream);
size_t	_fwrite_simple	(const void *ptr, size_t size, size_t nmemb,
			 struct _iob *stream);
size_t	_fwrite		(const void *ptr, size_t size, size_t nmemb,
			 struct _iob *stream);
int	_fsync_simple	(struct _iob *stream, int mbz);
int	_fsync		(struct _iob *stream, int mbz);

/* --- Macros --- */
#define _io_valid(f)	((f)->_io_count != _IO_FREE)

#ifndef NDEBUG
#define _arg_assert(condition)			\
  do {						\
    if (!(condition)) {				\
	errno = EINVAL;				\
    	return (_IO_ERRVAL);			\
    }						\
  } while (0)
#else
#define _arg_assert(condition)
#endif

#define _io_assert(stream, mbz)				\
  do {							\
      if (! (stream)->_io_funcs->fsync (stream, mbz)) {	\
          funlockfile (stream);				\
    	  return (_IO_ERRVAL);				\
      }							\
  } while (0)

#define _io_assert_unlocked(stream, mbz)		\
  do {							\
      if (! (stream)->_io_funcs->fsync (stream, mbz)) {	\
    	  return (_IO_ERRVAL);				\
      }							\
  } while (0)


#endif /* !defined __STDIOLIB_H */
