/*
 * Copyright (c) 1987 Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Arthur Olson.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <limits.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <libc_thread.h>

#ifndef TZNAME_MAX
#define TZNAME_MAX 32
#endif

#define	SECSPERMIN	60
#define	MINSPERHOUR	60
#define	HOURSPERDAY	24
#define	DAYSPERWEEK	7
#define	DAYSPERNYEAR	365
#define	DAYSPERLYEAR	366
#define	SECSPERHOUR	(SECSPERMIN * MINSPERHOUR)
#define	SECSPERDAY	((long) SECSPERHOUR * HOURSPERDAY)
#define	MONSPERYEAR	12

#define	TM_SUNDAY	0
#define	TM_MONDAY	1
#define	TM_TUESDAY	2
#define	TM_WEDNESDAY	3
#define	TM_THURSDAY	4
#define	TM_FRIDAY	5
#define	TM_SATURDAY	6

#define	TM_JANUARY	0
#define	TM_FEBRUARY	1
#define	TM_MARCH	2
#define	TM_APRIL	3
#define	TM_MAY		4
#define	TM_JUNE		5
#define	TM_JULY		6
#define	TM_AUGUST	7
#define	TM_SEPTEMBER	8
#define	TM_OCTOBER	9
#define	TM_NOVEMBER	10
#define	TM_DECEMBER	11
#define	TM_SUNDAY	0

#define	TM_YEAR_BASE	1900

#define	EPOCH_YEAR	1970
#define	EPOCH_WDAY	TM_THURSDAY

/*
** Accurate only for the past couple of centuries;
** that will probably do.
*/

#define	isleap(y) ( (((y) % 4) == 0 && ((y) % 100) != 0) || ((y) % 400) == 0)

struct rule {
	int		r_type;		/* type of rule--see below */
	int		r_day;		/* day number of rule */
	int		r_week;		/* week number of rule */
	int		r_mon;		/* month number of rule */
	long		r_time;		/* transition time of rule */
};

#define RT_NONE			0
#define	RT_JULIAN_DAY		1	/* Jn - Julian day */
#define	RT_DAY_OF_YEAR		2	/* n - day of year */
#define	RT_MONTH_NTH_DAY_OF_WEEK	3	/* Mm.n.d - month, week, day of week */


static const int	mon_lengths[2][MONSPERYEAR] = {
	31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31,
	31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

static const int	year_lengths[2] = {
	DAYSPERNYEAR, DAYSPERLYEAR
};

/*
** Given a pointer into a time zone string, scan until a character that is not
** a valid character in a zone name is found.  Return a pointer to that
** character.
*/

static const char *getzname(register const char *strp)
{
    char	c;

    while ((c = *strp) != '\0' && !isdigit(c) && c != ',' && c != '-' &&
	   c != '+')
      ++strp;
    return strp;
}

/*
** Given a pointer into a time zone string, extract a number from that string.
** Check that the number is within a specified range; if it is not, return
** NULL.
** Otherwise, return a pointer to the first character not part of the number.
*/

static const char *getnum(const char * strp,
			  int * const nump,const int min,const int max)
{
    char	c;
    int	num;

    if (strp == NULL || !isdigit(*strp))
      return NULL;
    num = 0;
    while ((c = *strp) != '\0' && isdigit(c)) {
	num = num * 10 + (c - '0');
	if (num > max)
	  return NULL;		/* illegal value */
	++strp;
    }
    if (num < min)
      return NULL;		/* illegal value */
    *nump = num;
    return strp;
}

/*
** Given a pointer into a time zone string, extract a number of seconds,
** in hh[:mm[:ss]] form, from the string.
** If any error occurs, return NULL.
** Otherwise, return a pointer to the first character not part of the number
** of seconds.
*/

static const char *getsecs(const char *	strp,long * const secsp)
{
    int	num;

    strp = getnum(strp, &num, 0, HOURSPERDAY);
    if (strp == NULL)
      return NULL;
    *secsp = num * SECSPERHOUR;
    if (*strp == ':') {
	++strp;
	strp = getnum(strp, &num, 0, MINSPERHOUR - 1);
	if (strp == NULL)
	  return NULL;
	*secsp += num * SECSPERMIN;
	if (*strp == ':') {
	    ++strp;
	    strp = getnum(strp, &num, 0, SECSPERMIN - 1);
	    if (strp == NULL)
	      return NULL;
	    *secsp += num;
	}
    }
    return strp;
}

/*
** Given a pointer into a time zone string, extract an offset, in
** [+-]hh[:mm[:ss]] form, from the string.
** If any error occurs, return NULL.
** Otherwise, return a pointer to the first character not part of the time.
*/

static const char *getoffset(const char *strp,long * const offsetp)
{
    int	neg;

    if (*strp == '-') {
	neg = 1;
	++strp;
    } else if (isdigit(*strp) || *strp++ == '+')
      neg = 0;
    else	return NULL;	/* illegal offset */
    strp = getsecs(strp, offsetp);
    if (strp == NULL)
      return NULL;		/* illegal time */
    if (neg)
      *offsetp = -*offsetp;
    return strp;
}

/*
** Given a pointer into a time zone string, extract a rule in the form
** date[/time].  See POSIX section 8 for the format of "date" and "time".
** If a valid rule is not found, return NULL.
** Otherwise, return a pointer to the first character not part of the rule.
*/

static const char *getrule(const char *strp,struct rule * const	rulep)
{
    if (*strp == 'J') {
	/*
	 ** Julian day.
	 */
	rulep->r_type = RT_JULIAN_DAY;
	++strp;
	strp = getnum(strp, &rulep->r_day, 1, DAYSPERNYEAR);
    } else if (*strp == 'M') {
	/*
	 ** Month, week, day.
	 */
	rulep->r_type = RT_MONTH_NTH_DAY_OF_WEEK;
	++strp;
	strp = getnum(strp, &rulep->r_mon, 1, MONSPERYEAR);
	if (strp == NULL)
	  return NULL;
	if (*strp++ != '.')
	  return NULL;
	strp = getnum(strp, &rulep->r_week, 1, 5);
	if (strp == NULL)
	  return NULL;
	if (*strp++ != '.')
	  return NULL;
	strp = getnum(strp, &rulep->r_day, 0, DAYSPERWEEK - 1);
    } else if (isdigit(*strp)) {
	/*
	 ** Day of year.
	 */
	rulep->r_type = RT_DAY_OF_YEAR;
	strp = getnum(strp, &rulep->r_day, 0, DAYSPERLYEAR - 1);
    } else
      return NULL;	/* invalid format */
    if (strp == NULL)
      return NULL;
    if (*strp == '/') {
	/*
	 ** Time specified.
	 */
	++strp;
	strp = getsecs(strp, &rulep->r_time);
    } else	
      rulep->r_time = 2 * SECSPERHOUR; /* default = 2:00:00 */
    return strp;
}

/*
** Given the Epoch-relative time of January 1, 00:00:00 GMT, in a year, the
** year, a rule, and the offset from GMT at the time that rule takes effect,
** calculate the Epoch-relative time that rule takes effect.
*/

static time_t transtime(
			const time_t				janfirst,
			const int				year,
			const struct rule * const		rulep,
			const long				offset)
{
    int	leapyear;
    time_t	value;
    int	i;
    int		d, m1, yy0, yy1, yy2, dow;

    leapyear = isleap(year);
    switch (rulep->r_type) {
    default:
	return -1;
    case RT_JULIAN_DAY:
	/*
	 ** Jn - Julian day, 1 == January 1, 60 == March 1 even in leap
	 ** years.
	 ** In non-leap years, or if the day number is 59 or less, just
	 ** add SECSPERDAY times the day number-1 to the time of
	 ** January 1, midnight, to get the day.
	 */
	value = janfirst + (rulep->r_day - 1) * SECSPERDAY;
	if (leapyear && rulep->r_day >= 60)
	  value += SECSPERDAY;
	break;

    case RT_DAY_OF_YEAR:
	/*
	 ** n - day of year.
	 ** Just add SECSPERDAY times the day number to the time of
	 ** January 1, midnight, to get the day.
	 */
	value = janfirst + rulep->r_day * SECSPERDAY;
	break;

    case RT_MONTH_NTH_DAY_OF_WEEK:
	/*
	 ** Mm.n.d - nth "dth day" of month m.
	 */
	value = janfirst;
	for (i = 0; i < rulep->r_mon - 1; ++i)
	  value += mon_lengths[leapyear][i] * SECSPERDAY;

	/*
	 ** Use Zeller's Congruence to get day-of-week of first day of
	 ** month.
	 */
	m1 = (rulep->r_mon + 9) % 12 + 1;
	yy0 = (rulep->r_mon <= 2) ? (year - 1) : year;
	yy1 = yy0 / 100;
	yy2 = yy0 % 100;
	dow = ((26 * m1 - 2) / 10 +
	       1 + yy2 + yy2 / 4 + yy1 / 4 - 2 * yy1) % 7;
	if (dow < 0)
	  dow += DAYSPERWEEK;

	/*
	 ** "dow" is the day-of-week of the first day of the month.  Get
	 ** the day-of-month (zero-origin) of the first "dow" day of the
	 ** month.
	 */
	d = rulep->r_day - dow;
	if (d < 0)
	  d += DAYSPERWEEK;
	for (i = 1; i < rulep->r_week; ++i) {
	    if (d + DAYSPERWEEK >=
		mon_lengths[leapyear][rulep->r_mon - 1])
	      break;
	    d += DAYSPERWEEK;
	}

	/*
	 ** "d" is the day-of-month (zero-origin) of the day we want.
	 */
	value += d * SECSPERDAY;
	break;
    }

    /*
     ** "value" is the Epoch-relative time of 00:00:00 GMT on the day in
     ** question.  To get the Epoch-relative time of the specified local
     ** time on that day, add the transition time and the current offset
     ** from GMT.
     */
    return value + rulep->r_time + offset;
}

static __sdethread_mutex_t	tz_mx = __SDETHREAD_MUTEX_INITIALIZER("tz");
static char		tz_stdnam[TZNAME_MAX] = "UCT";
static char		tz_dstnam[TZNAME_MAX] = "UCT";
long			_tz_stdoffset = 0;
long 			_tz_dstoffset = 0;
static struct rule 	tz_dststart;
static struct rule 	tz_dstend;
static char		tz_hasdst;

#define tz_stdoffset	_tz_stdoffset
#define tz_dstoffset	_tz_dstoffset

#if defined(__mips) && (__GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ >= 7))
 __attribute__ ((section (".sdata")))
#endif
char 			*tzname[2] = { tz_stdnam, tz_dstnam};

/*
** Given a POSIX section 8-style TZ string, fill in the rule tables as
** appropriate.
*/

static int tzparse(const char *name)
{
    const char *		stdname;
    const char *		dstname;
    int				stdlen;
    int				dstlen;
    
    tzname[0] = tz_stdnam;
    tzname[1] = tz_dstnam;
    memset(&tz_dststart,0,sizeof(tz_dststart));
    memset(&tz_dstend,0,sizeof(tz_dstend));
    
    stdname = name;
    name = getzname(name);
    stdlen = name - stdname;
    if (stdlen < 3 || stdlen >= TZNAME_MAX)
      return -1;
    strncpy(tz_stdnam,stdname,stdlen);
    if (*name == '\0')
      return -1;
    
    name = getoffset(name, &tz_stdoffset);
    if (name == NULL)
      return -1;
    if (*name != '\0') {
	dstname = name;
	name = getzname(name);
	dstlen = name - dstname; /* length of DST zone name */
	if (dstlen < 3 || dstlen >= TZNAME_MAX)
	  return -1;
	strncpy(tz_dstnam,dstname,dstlen);
	
	if (*name != '\0' && *name != ',' ) {
	    name = getoffset(name, &tz_dstoffset);
	    if (name == NULL)
	      return -1;
	} else	
	  tz_dstoffset = tz_stdoffset - SECSPERHOUR;
	tz_hasdst = 1;
	
	if (*name == ',') {
	    ++name;
	    if ((name = getrule(name, &tz_dststart)) == NULL)
	      return -1;
	    if (*name++ != ',')
	      return -1;
	    if ((name = getrule(name, &tz_dstend)) == NULL)
	      return -1;
	} 
	if (*name != '\0')
	  return -1;
    } else {
	strcpy(tz_dstnam,"   ");
	tz_dstoffset = 0; 
	tz_hasdst = 0;
    }
    return 0;
}

/*
 * reset tz state to the default initial state
*/
static void tzreset()
{
    tzparse("UTC0");
}

void tzset(void)
{
    const char *tz = getenv("TZ");
    static char	*tz_envval;
    
    __sdethread_mutex_lock (&tz_mx);
    if(tz && *tz) {
	if(tz_envval) {
	    if(strcmp(tz_envval,tz) == 0) {
		__sdethread_mutex_unlock (&tz_mx);
		return;
	    }
	    tz_envval = realloc(tz_envval,strlen(tz)+1);
	    strcpy(tz_envval,tz);
	} else {
	    tz_envval = malloc(strlen(tz)+1);
	    strcpy(tz_envval,tz);
	}
	if(tzparse(tz) == 0) {
	    __sdethread_mutex_lock (&tz_mx);
	    return;
	}
    }
    if(tz_envval) {
	free(tz_envval);
	tz_envval = 0;
	tzreset();
    }
    __sdethread_mutex_unlock (&tz_mx);
}


struct tm *localtime_r (const time_t *tp, struct tm *tm)
{
    extern struct tm *_offtime_r(const time_t *tp,int offset, struct tm *);
    time_t	tgmt = *tp;
    int		dst;
    
    tzset();
    
    if(tz_hasdst) {
	if(tz_dststart.r_type != RT_NONE) {
	    time_t	tjan = 0;
	    time_t	tyr = tgmt;
	    time_t 	yl;
	    time_t	stim;
	    time_t	etim;
	    int		yr;
	    
	    for(yr = EPOCH_YEAR;
		tyr >= (yl=year_lengths[isleap(yr)] * SECSPERDAY); yr++) {
		
		tyr -= yl;
		tjan +=  yl;
	    }
	    stim = transtime(tjan,yr,&tz_dststart,0);
	    etim = transtime(tjan,yr,&tz_dstend,0);
	    if(stim < etim) {
		if(tgmt >= stim && tgmt < etim)
		  dst = 1;
		else
		  dst = 0;
	    } else {
		if(tgmt >= etim && tgmt < stim)
		  dst = 0;
		else
		  dst = 1;
	    }
	} else
	  dst = -1;
    } else
      dst = 0;
    
    tm = _offtime_r (tp, dst==1?-tz_dstoffset:-tz_stdoffset, tm);
    tm->tm_isdst = dst;
    tm->tm_zone = 0;
    return tm;
}

struct tm *localtime(const time_t *tp)
{
    static struct tm tm;
    _THREAD_PRIVATE_KEY(localtime);
    struct tm *tmp = _THREAD_PRIVATE(localtime, tm, NULL);
    if (tmp)
	return localtime_r (tp, tmp);
    return NULL;
}

time_t	mktime(struct tm *tm)
{
    int yr;
    int mn;
    time_t secs;
    struct	tm *ntm;

    if (tm->tm_sec > 61 || 
	tm->tm_min > 59 || 
	tm->tm_hour > 23 || 
	tm->tm_mday > 31 || 
	tm->tm_mon > 12 ||
	tm->tm_year < 70) {
	return (time_t)-1;
    }
    
    /*
     * Sum up seconds from beginning of year
     */
    secs = tm->tm_sec;
    secs += tm->tm_min * SECSPERMIN;
    secs += tm->tm_hour * SECSPERHOUR;
    secs += (tm->tm_mday-1) * SECSPERDAY;
    
    for (mn = 0; mn < tm->tm_mon; mn++)
      secs += mon_lengths[isleap(tm->tm_year+1900)][mn] * SECSPERDAY;
    
    for(yr=1970; yr < tm->tm_year + 1900; yr++)
      secs += year_lengths[isleap(yr)]*SECSPERDAY;
    
    if(tm->tm_isdst == 0) {
	secs -= tz_stdoffset;
	ntm = localtime(&secs);
    } else if(tm->tm_isdst > 0) {
	tm->tm_isdst = 1;
	secs -= tz_dstoffset;
	ntm = localtime(&secs);
    } else {
	/* try to determine if daylight saving was set at this time 
	 */
	secs -= tz_stdoffset;
	ntm = localtime(&secs);
	if(ntm->tm_mday != tm->tm_mday ||
	   ntm->tm_hour != tm->tm_hour ||
	   ntm->tm_min  != tm->tm_min ||
	   ntm->tm_sec  != tm->tm_sec) {
	    secs += tz_stdoffset;
	    secs -= tz_dstoffset;
	}
    }
    /* work secs back into a tm to get all the fields correct 
     */
    if(ntm->tm_year != tm->tm_year ||
       ntm->tm_mon  != tm->tm_mon ||
       ntm->tm_mday != tm->tm_mday ||
       ntm->tm_hour != tm->tm_hour ||
       ntm->tm_min  != tm->tm_min ||
       ntm->tm_sec  != tm->tm_sec ||
       (tm->tm_isdst >= 0 && ntm->tm_isdst != tm->tm_isdst))
       return -1;
    
    *tm = *ntm;
    return secs;
}
