#ident "lib/ieee/sp_fstr.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2004-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <ctype.h>
#include <errno.h>
#include <assert.h>

#include "ieee754int.h"
#include "ieee754sp.h"

/*
 * strtof for IEEE doubles 
*/

static const int32_t f5pow[] = {
    0x40a00000,	/* 5. */
    0x41c80000,	/* 25. */
    0x441c4000,	/* 625. */
    0x48bebc20,	/* 390625. */
    0x520e1bca,	/* 152587894785. */
    0x649dc5ae	/* 23283065137950552489985. */
};

#define F5POW(x)	(*(ieee754sp *)&f5pow[x])	/* 5^x */
#define F5POW_MAXX	(sizeof(f5pow)/sizeof(f5pow[0]))


/* required constats */
#define FZERO		ieee754sp_zero(0)
#define FONE		ieee754sp_one(0)
#define FTEN		ieee754sp_ten(0)
#define FADD(x,y)	ieee754sp_add(x,y)
#define FNEG(x)		ieee754sp_neg(x)
#define FMUL(x,y)	ieee754sp_mul(x,y)
#define FDIV(x,y)	ieee754sp_div(x,y)
#define FCMPLT(x,y)	(ieee754sp_order(x,y) < 0)
#define FLO(i)		ieee754sp_fint(i)
#define FLDEXP(x,n)	ieee754sp_nldexp(x,n)
#define ISINF(x)	ieee754sp_isinf(x)

ieee754sp ieee754sp_nfstr(const char *nptr, char **endp)
{
    const char *s = nptr;
    int mneg = 0;
    int expt = 0;
    int nz   = 0;
    int some = 0;
    ieee754sp v = FZERO;

    while(isspace(*s) ) 
      s++;
    
    if(*s == '-') {
	mneg = 1;
	s++;
    }else if(*s == '+'){
           s++;
    }
    
    while( isdigit(*s) ) {
	int inc = *s++ - '0';
	nz |= inc;
	v = FADD(FMUL(v,FTEN),FLO(inc));
	some = 1;
    }
    
    if( *s == '.') {
	s++;
	
	/* grab fractional part decrementing exponent
	 */
	while( isdigit(*s) ) {
	    int inc = *s++ - '0';
	    nz |= inc;
	    v = FADD(FMUL(v,FTEN),FLO(inc));
	    expt--;
	    some = 1;
	}
    }

    if(some && (*s == 'e' || *s == 'E')) {
	int e 		= 0;
	int eneg	= 0;
	
	s++;
	
	/* grab exponent */
	if( *s == '-') {
	    eneg = 1;
	    s++;
	} else if( *s == '+') {
	    s++;
	}
	while(isdigit(*s))
	  e = e * 10 + (*s++ - '0');
	if(eneg)
	  e = -e;
	expt += e;
    }
    
    if(endp)
      *endp = (char *) (some ? s : nptr);

    if(expt) {
	ieee754sp flexp = FONE;
	int	bexp = expt;
	int c;

	if(expt < 0)
	  expt = -expt;
	
	for (c = F5POW_MAXX; c;  c--) {
	    if (expt & (1 << (c-1)))
	      flexp = FMUL(flexp,F5POW(c-1));
	}
	if (bexp < 0) 
	  v = FDIV(v,flexp);
	else 
	  v = FMUL(v,flexp);
	v = FLDEXP(v, bexp);
    }
    
    if (UNLIKELY (ISINF (v))) {
	errno = ERANGE;
    }
    else if (UNLIKELY (v == FZERO && nz)) {
	v = FZERO;
	errno = ERANGE;
    }

    if (mneg)
      v = FNEG(v);
    
    return v;
}
