#ident "lib/ieee/sp_sqrt.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/* IEEE754 floating point arithmetic
 * single precision square root
 */
/*
 * Copyright (c) 1996-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <limits.h>
#include <assert.h>
#include <stdarg.h>
#include <errno.h>
#include "ieee754sp.h"

#define nan	ieee754sp_snan(0)

ieee754sp
ieee754sp_nsqrt (ieee754sp x)
{
    int s,q,t,i;
    uint32_t r;
    COMPXSP;
	
    /* take care of Inf and NaN */

    EXPLODEXSP;

    /* x == INF or NAN? */
    if (! CLASS_NORM_P (xc)) {
#if IEEE754_CLASS_CASE
	switch (xc) {
	case IEEE754_CLASS_QNAN:
	case IEEE754_CLASS_SNAN:
	    /* sqrt(Nan) = Nan */
	    return ieee754sp_nanxcpt (x, "sqrt", x);
	case IEEE754_CLASS_ZERO:
	    /* sqrt(0) = 0 */
	    return x;		
	case IEEE754_CLASS_INF:
	    if (xs) {
		/* sqrt(-Inf) = Nan */
		errno = EDOM;
		return ieee754sp_nanxcpt (nan, "sqrt", x);
	    }
	    /* sqrt(+Inf) = Inf */
	    return x;
	case IEEE754_CLASS_DNORM:
	    SPDNORMX;
	}
#else
        if (xc & IEEE754_CLASS_ZERO)
	    /* sqrt(0) = 0 */
	    return x;		
	if (xc & (IEEE754_CLASS_QNAN | IEEE754_CLASS_SNAN))
	    /* sqrt(Nan) = Nan */
	    return ieee754sp_nanxcpt (x, "sqrt", x);
	if (xc & IEEE754_CLASS_INF) {
	    if (xs) {
		/* sqrt(-Inf) = Nan */
		errno = EDOM;
		return ieee754sp_nanxcpt (nan, "sqrt", x);
	    }
	    /* sqrt(+Inf) = Inf */
	    return x;
	}
        assert (xc & IEEE754_CLASS_DNORM);
	SPDNORMX;
#endif
    }

    if (xs) {
	/* sqrt(-x) = Nan */
	errno = EDOM;
	return ieee754sp_nanxcpt (nan, "sqrt", x);
    }

    if (xe & 1)	/* odd xe, double x to make it even */
	xm += xm;
    xe >>= 1;	/* xe = [xe/2] */

    /* generate sqrt(x) bit by bit */
    xm += xm;
    q = s = 0;		/* q = sqrt(x) */
    r = 0x01000000;	/* r = moving bit from right to left */

    while (r!=0) {
	t = s+r; 
	if (t<=xm) { 
	    s    = t+r; 
	    xm  -= t; 
	    q   += r; 
	} 
	xm += xm;
	r>>=1;
    }

    if (xm!=0) {
	switch (IEEE754_GET_RM ()) {
	case IEEE754_RP:
	    q += 2;
	    break;
	case IEEE754_RN:
	    q += (q&1);
	    break;
	}
    }

    x = (q>>1)+0x3f000000;
    x += (xe << SP_MBITS);
    return x;
}

#if ! IEEE754_GCCSOFTFLOAT
ieee754sp_u
ieee754sp_sqrt (ieee754sp_u x)
{
    ieee754sp_u r;
    r.bits = ieee754sp_nsqrt (x.bits);
    return r;
}
#endif
