/*
 * Copyright (c) 1983, 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char rcsid[] = "$OpenBSD: syslog.c,v 1.9 2000/01/02 23:35:58 hugh Exp $";
#endif /* LIBC_SCCS and not lint */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <sys/uio.h>
#include <sys/un.h>
#include <netdb.h>
#ifndef USE_UNIXSOCK
#include <netinet/in.h>
#include <arpa/inet.h>
#endif
#include <errno.h>
#include <fcntl.h>
#include <paths.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include "thread_private.h"

struct logstate {
    int		file;			/* fd for log */
    int		facility;		/* default facility code */
    int		mask;			/* mask of priorities to be logged */
    int		connected;		/* have done connect */
    int		opened;			/* have done openlog() */
    int		stat;			/* status bits, set by openlog() */
    const char *tag;			/* string to tag the entry with */
};

extern char	*__progname;		/* Program name, from crt0. */

static void	disconnectlog __P((void)); /* disconnect from syslogd */
static void	connectlog __P((void));	/* (re)connect to syslogd */

_THREAD_PRIVATE_KEY(syslog);
static struct logstate logstate = {-1, LOG_USER, 0xff};

#ifndef USE_STRFTIME
static const char amon_names[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
  };
#endif

/*
 * syslog, vsyslog --
 *	print message on log file; output is intended for syslogd(8).
 */
void
#ifdef __STDC__
syslog(int pri, const char *fmt, ...)
#else
syslog(pri, fmt, va_alist)
	int pri;
	char *fmt;
	va_dcl
#endif
{
	va_list ap;

#ifdef __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	vsyslog(pri, fmt, ap);
	va_end(ap);
}

void
vsyslog(pri, fmt, ap)
	int pri;
	register const char *fmt;
	va_list ap;
{
	struct logstate *ls = _THREAD_PRIVATE(syslog, logstate, NULL);
	register int cnt;
	register char ch, *p, *t;
	time_t now;
	int fd, saved_errno;
#define	TBUF_LEN	2048
#define	FMT_LEN		1024
	char *xpri, *stdp, tbuf[TBUF_LEN], fmt_cpy[FMT_LEN];
	int tbuf_left, fmt_left, prlen;

#define	INTERNALLOG	LOG_ERR|LOG_CONS|LOG_PERROR|LOG_PID
	/* Check for invalid bits. */
	if (pri & ~(LOG_PRIMASK|LOG_FACMASK)) {
		syslog(INTERNALLOG,
		    "syslog: unknown facility/priority: %x", pri);
		pri &= LOG_PRIMASK|LOG_FACMASK;
	}

	/* Check priority against setlogmask values. */
	if (!(LOG_MASK(LOG_PRI(pri)) & ls->mask))
		return;

	saved_errno = errno;

	/* Set default facility if none specified. */
	if ((pri & LOG_FACMASK) == 0)
		pri |= ls->facility;

	/* Build the message. */
	
	/*
 	 * Although it's tempting, we can't ignore the possibility of
	 * overflowing the buffer when assembling the "fixed" portion
	 * of the message.  Strftime's "%h" directive expands to the
	 * locale's abbreviated month name, but if the user has the
	 * ability to construct to his own locale files, it may be
	 * arbitrarily long.
	 */
	(void)time(&now);

	p = tbuf;  
	tbuf_left = TBUF_LEN;
	
#define	DEC()	\
	do {					\
		if (prlen >= tbuf_left)		\
			prlen = tbuf_left - 1;	\
		p += prlen;			\
		tbuf_left -= prlen;		\
	} while (0)

	prlen = snprintf(p, tbuf_left, "<%d>", pri);
	DEC();

	xpri = p;

#ifdef USE_STRFTIME
	prlen = strftime(p, tbuf_left, "%h %e %T ", localtime(&now));
#else
	{
	    /* avoid dragging in all of strftime */
	    struct tm *tm = localtime(&now);
	    prlen = snprintf(p, tbuf_left, "%.3s % 2d %02d:%02d:%02d ", 
			     amon_names[tm->tm_mon], tm->tm_mday,
			     tm->tm_hour, tm->tm_min, tm->tm_sec);
	}
#endif
	DEC();

	stdp = p;

	if (ls->tag == NULL)
#if __AlgNET__
	    (void) pthread_getname_np (pthread_self(), &ls->tag);
#else
	    ls->tag = __progname;
#endif
	if (ls->tag != NULL) {
		prlen = snprintf(p, tbuf_left, "%s", ls->tag);
		DEC();
	}
	if (ls->stat & LOG_PID) {
		prlen = snprintf(p, tbuf_left, "[%d]", getpid());
		DEC();
	}
	if (ls->tag != NULL) {
		if (tbuf_left > 1) {
			*p++ = ':';
			tbuf_left--;
		}
		if (tbuf_left > 1) {
			*p++ = ' ';
			tbuf_left--;
		}
	}

	/* 
	 * We wouldn't need this mess if printf handled %m, or if 
	 * strerror() had been invented before syslog().
	 */
	for (t = fmt_cpy, fmt_left = FMT_LEN; (ch = *fmt); ++fmt) {
		if (ch == '%' && fmt[1] == 'm') {
		        char buf[64];
			++fmt;
			prlen = snprintf(t, fmt_left, "%s",
			    strerror_r(saved_errno, buf, sizeof(buf)));
			if (prlen >= fmt_left)
				prlen = fmt_left - 1;
			t += prlen;
			fmt_left -= prlen;
		} else {
			if (fmt_left > 1) {
				*t++ = ch;
				fmt_left--;
			}
		}
	}
	*t = '\0';

	prlen = vsnprintf(p, tbuf_left, fmt_cpy, ap);
	DEC();
	cnt = p - tbuf;

	/* Output to stderr if requested. */
	if (ls->stat & LOG_PERROR) {
		struct iovec iov[2];

		iov[0].iov_base = stdp;
		iov[0].iov_len = cnt - (stdp - tbuf);
		iov[1].iov_base = "\n";
		iov[1].iov_len = 1;
		(void)writev(STDERR_FILENO, iov, 2);
	}

	/* Get connected, output the message to the local logger. */
	if (!ls->opened)
		openlog(ls->tag, ls->stat, 0);

	connectlog();
	if (send(ls->file, tbuf, cnt, 0) >= 0)
		return;

	/*
	 * If the send() failed, the odds are syslogd was restarted.
	 * Make one (only) attempt to reconnect to /dev/log.
	 */
	disconnectlog();
	connectlog();
	if (send(ls->file, tbuf, cnt, 0) >= 0)
		return;

	/*
	 * Output the message to the console; don't worry about blocking,
	 * if console blocks everything will.  Make sure the error reported
	 * is the one from the syslogd failure.
	 */
	if (ls->stat & LOG_CONS &&
	    (fd = open(_PATH_CONSOLE, O_WRONLY, 0)) >= 0) {
		struct iovec iov[2];
		
		iov[0].iov_base = xpri;
		iov[0].iov_len = cnt - (xpri - tbuf);
		iov[1].iov_base = "\r\n";
		iov[1].iov_len = 2;
		(void)writev(fd, iov, 2);
		(void)close(fd);
	}
}

static void
disconnectlog()
{
	struct logstate *ls = _THREAD_PRIVATE(syslog, logstate, NULL);

	/*
	 * If the user closed the FD and opened another in the same slot,
	 * that's their problem.  They should close it before calling on
	 * system services.
	 */
	if (ls->file != -1) {
		close(ls->file);
		ls->file = -1;
	}
	ls->connected = 0;		/* retry connect */
}

static void
connectlog()
{
	struct logstate *ls = _THREAD_PRIVATE(syslog, logstate, NULL);

#if defined(USE_UNIXSOCK)
	struct sockaddr_un SyslogAddr;	/* AF_UNIX address of local logger */
	if (ls->file == -1) {
		if ((ls->file = socket(AF_UNIX, SOCK_DGRAM, 0)) == -1)
			return;
		(void)fcntl(ls->file, F_SETFD, 1);
	}
	if (ls->file != -1 && !ls->connected) {
		memset(&SyslogAddr, '\0', sizeof(SyslogAddr));
		SyslogAddr.sun_len = sizeof(SyslogAddr);
		SyslogAddr.sun_family = AF_UNIX;
		(void)strncpy(SyslogAddr.sun_path, _PATH_LOG,
		    sizeof(SyslogAddr.sun_path) - 1);
		SyslogAddr.sun_path[sizeof(SyslogAddr.sun_path) - 1] = '\0';
		if (connect(ls->file, (struct sockaddr *)&SyslogAddr,
		    sizeof(SyslogAddr)) == -1) {
			(void)close(ls->file);
			ls->file = -1;
		} else
			ls->connected = 1;
	}
#elif 0
	/* Use UDP loopback - should always work */
	struct sockaddr_in SyslogAddr;
	if (ls->file == -1) {
	    if ((ls->file = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		return;
	    (void)fcntl(ls->file, F_SETFD, 1);
	}
	if (ls->file != -1 && !ls->connected) {
	    SyslogAddr.sin_family = AF_INET;
	    SyslogAddr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
	    SyslogAddr.sin_port = htons (514);
	    SyslogAddr.sin_len = sizeof (SyslogAddr);
	    /* XXX A UDP connect will always succeed, even if there is
	       no server running and bound to that port, so we should
	       probably first try binding to the port and if that
	       succeeds then the server isn't present. */
	    if (connect(ls->file, (struct sockaddr *)&SyslogAddr, 
			sizeof(SyslogAddr)) < 0) {
		(void)close(ls->file);
		ls->file = -1;
	    } else
		ls->connected = 1;
	}
#else
	/* Syslog daemon not implemented yet */
	ls->file = -1;
	ls->connected = 0;
#endif
}

void
openlog(ident, logstat, logfac)
	const char *ident;
	int logstat, logfac;
{
	struct logstate *ls = _THREAD_PRIVATE(syslog, logstate, NULL);
	if (ident != NULL)
		ls->tag = ident;
	ls->stat = logstat;
	if (logfac != 0 && (logfac &~ LOG_FACMASK) == 0)
		ls->facility = logfac;

	if (ls->stat & LOG_NDELAY)	/* open immediately */
		connectlog();

	ls->opened = 1;	/* ident and facility has been set */
}

void
closelog()
{
	struct logstate *ls = _THREAD_PRIVATE(syslog, logstate, NULL);
	(void)close(ls->file);
	ls->file = -1;
	ls->connected = 0;
}

/* setlogmask -- set the log mask level */
int
setlogmask(pmask)
	int pmask;
{
	struct logstate *ls = _THREAD_PRIVATE(syslog, logstate, NULL);
	int omask;

	omask = ls->mask;
	if (pmask != 0)
		ls->mask = pmask;
	return (omask);
}
