/* 
 * "Machine-dependent" glue for BSD kernel running on AlgRTX.
 * Mainly interrupt handling, softnet and startup code.
 *
 * Copyright (c) 2001 Algorithmics Ltd
 *
 * This program is NOT free software, it is supplied under the terms
 * of the AlgNET License Agreement, a copy of which is available at:
 *
 *  http://www.algor.co.uk/algor/info/algnet-license.pdf
 *
 * Any company which has obtained and signed a valid AlgNET license
 * may use and modify this software internally and use (without
 * restrictions) any derived binary.  You may not, however,
 * redistribute this in whole or in part as source code, nor may you
 * modify or remove any part of this copyright message.  
 */

/*	$OpenBSD: machdep.c,v 1.39 2000/10/27 00:16:04 mickey Exp $	*/
/*
 * Copyright (c) 1988 University of Utah.
 * Copyright (c) 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department, The Mach Operating System project at
 * Carnegie-Mellon University and Ralph Campbell.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	from: @(#)machdep.c	8.3 (Berkeley) 1/12/94
 *      $Id: machdep.c,v 1.39 2000/10/27 00:16:04 mickey Exp $
 */


/* from: Utah Hdr: machdep.c 1.63 91/04/24 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <mips/cpu.h>	/* MIPS SDE header */
#include <sys/map.h>
#include <sys/device.h>
#include <sys/proc.h>
#include <sys/buf.h>
#include <sys/reboot.h>
#include <sys/conf.h>
#include <sys/file.h>
#include <sys/clist.h>
#include <sys/timeout.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/msgbuf.h>
#include <sys/ioctl.h>
#include <sys/tty.h>

#include <vm/vm.h>
#include <vm/vm_kern.h>
#include <vm/pmap.h>

#include <net/netisr.h>

#include <machine/pio.h>
#define _ALGRTX_BUS_DMA_PRIVATE
#include <machine/bus.h>
#include <machine/autoconf.h>

#include "kit_impl.h"		/* MIPS SDE header */

/* Use C library malloc */
#undef malloc
#undef free
extern void *malloc __P((size_t));
extern void free __P((void *));

/*
 * safepri is a safe priority for sleep to set for a spin-wait
 * during autoconfiguration or after a panic.
 */
int	safepri = IPL_NONE;

/* if 1, still working on cold-start */
int	cold = 1;

/* dynamic timeouts */
#define NTIMEOUT 20
static struct timeout timeouttab[NTIMEOUT];
int	ntimeout = NTIMEOUT;
struct timeout *timeouts = timeouttab;

/*
 * Return the best possible estimate of the time in the timeval
 * to which tvp points.
 */
void
microtime(tvp)
	register struct timeval *tvp;
{
	int s = splclock();
	static struct timeval lasttime;

	*tvp = time;
	if (tvp->tv_sec == lasttime.tv_sec &&
	    tvp->tv_usec <= lasttime.tv_usec &&
	    (tvp->tv_usec = lasttime.tv_usec + 1) >= 1000000) {
		tvp->tv_sec++;
		tvp->tv_usec -= 1000000;
	}
	lasttime = *tvp;
	splx(s);
}


/* INTERRUPT HANDLING */

/* Each thread has a current simulate "interrupt priority level" or
 * ipl.  When at that ipl it should prevent any other thread at the
 * same or lower ipl from running, including "interrupt" routines.
 *
 * For now we only implement three levels: zero, softnet and 
 * everything else.  */

/* We must be careful to always lock and unlock these mutexes in
   the same order to avoid deadlock. */

/* XXX I'm not happy with this technique, because a high priority
   interrupt routine has to wait for all lower level spl'ed code to
   finish before it can lock every mutec. Perhaps we should actually
   change each thread's priority to match its IPL, but how do we then
   guarantee to block other threads spl'ed to the same priority level
   - purely by making them FIFO scheduled? */

static pthread_mutex_t	splsoftclock_mx;
static pthread_mutex_t	splsoftnet_mx;
static pthread_mutex_t	splnet_mx;

/* The softnet interrupt is implemented via a condition variable */
static pthread_cond_t	softnet_cd = PTHREAD_COND_INITIALIZER;
static int		softnet_pend;

/* Each AlgRTX doorbell has one or more "interrupt" handlers
   chained to it. */
struct intrhand {
    LIST_ENTRY(intrhand) ichain;
    int			ipl;
    int			(*handler)(void *);
    void		*arg;
    const char		*name;
};

struct interrupt {
    LIST_HEAD(,intrhand) head;
    pthread_t		tid;		/* monitor thread */
    pthread_doorbell_t	db;		/* algrtx doorbell */
    int			maxipl;
};

static struct interrupt *intrtable;

/* The generic hardware "interrupt" handler thread, one run
   for each interrupt. */

static void *
intrupt_thread (void *arg)
{
    struct interrupt *in = arg;
    struct intrhand *ih;
    sigset_t sm;

    /* disable all signals in this thread */
    sigfillset (&sm);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);

    splsoftclock ();
    while (1) {
	pthread_doorbell_wait (&in->db, &splsoftclock_mx);
	LIST_FOREACH(ih, &in->head, ichain) {
	    int s = splx (ih->ipl);
	    (*ih->handler) (ih->arg);
	    splx (s);
	}
    }
    return 0;
}


/* The softnet "interrupt" handler thread */

static void *
softnet_thread (void *arg)
{
    sigset_t sm;

    /* disable all signals in this thread */
    sigfillset (&sm);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);

    splsoftnet ();
    while (1) {
	while (softnet_pend == 0) {
	    splx (IPL_SOFTCLOCK);		/* drop to splsoftclock */
	    pthread_cond_wait (&softnet_cd, &splsoftclock_mx);
	    splsoftnet ();			/* raise to splsoftnet */
	}
	softnet_pend = 0;

#ifdef INET
	if (netisr & (1 << NETISR_ARP)) {
	    netisr &= ~(1 << NETISR_ARP);
	    arpintr();
	}
	if (netisr & (1 << NETISR_IP)) {
	    netisr &= ~(1 << NETISR_IP);
	    ipintr();
	}
#endif
#ifdef INET6
	if (netisr & (1 << NETISR_IPV6)) {
	    netisr &= ~(1 << NETISR_IPV6);
	    ip6intr();
	}
#endif
#ifdef NETATALK
	if (netisr & (1 << NETISR_ATALK)) {
	    netisr &= ~(1 << NETISR_ATALK);
	    atintr();
	}
#endif
#ifdef NS
	if (netisr & (1 << NETISR_NS)) {
	    netisr &= ~(1 << NETISR_NS);
	    nsintr();
	}
#endif
#ifdef ISO
	if (netisr & (1 << NETISR_ISO)) {
	    netisr &= ~(1 << NETISR_ISO);
	    clnlintr();
	}
#endif
#if NPPP > 0
	if(netisr & (1 << NETISR_PPP)) {
	    netisr &= ~(1 << NETISR_PPP);
	    pppintr();
	}
#endif
#if NBRIDGE > 0
	if(netisr & (1 << NETISR_BRIDGE)) {
	    netisr &= ~(1 << NETISR_BRIDGE);
	    bridgeintr();
	}
#endif
    }
    return 0;
}


/* This key returns a pointer to the current thread's ipl */
static pthread_key_t ipl_key;


static int *
_kern_getcuripl ()
{
    int *curipl;
    if (!(curipl = (int *) pthread_getspecific (ipl_key))) {
	curipl = malloc (sizeof (int));
	if (!curipl)
	    panic ("malloc curipl");
	*curipl = IPL_NONE;
	pthread_setspecific (ipl_key, curipl);
    }	
    return curipl;
}


static int
_kern_splworker (curipl, ipl)
    int *curipl;
    int ipl;
{
    int oldipl;

    oldipl = *curipl;
    if (ipl != oldipl) {
	if (ipl > oldipl) {
	    if (ipl >= IPL_SOFTCLOCK && oldipl < IPL_SOFTCLOCK)
		pthread_mutex_lock (&splsoftclock_mx);
	    if (ipl >= IPL_SOFTNET && oldipl < IPL_SOFTNET)
		pthread_mutex_lock (&splsoftnet_mx);
	    if (ipl >= IPL_NET && oldipl < IPL_NET)
		pthread_mutex_lock (&splnet_mx);
	}
	else {
	    if (ipl < IPL_NET && oldipl >= IPL_NET)
		pthread_mutex_unlock (&splnet_mx);
	    if (ipl < IPL_SOFTNET && oldipl >= IPL_SOFTNET)
		pthread_mutex_unlock (&splsoftnet_mx);
	    if (ipl < IPL_SOFTCLOCK && oldipl >= IPL_SOFTCLOCK)
		pthread_mutex_unlock (&splsoftclock_mx);
	}
	*curipl = ipl;
    }
    return oldipl;
}


int
_kern_spl (ipl)
{
    int *curipl = _kern_getcuripl ();
    if (ipl > *curipl)	/* only raise */
	return _kern_splworker (curipl, ipl);
    return *curipl;
}


int
_kern_splx (ipl)
{
    int *curipl = _kern_getcuripl ();
    if (ipl != *curipl)
	return _kern_splworker (curipl, ipl);
    return *curipl;
}


/* Install an "interrupt" handler */
void *
_kern_intr_establish (int inum, int ipl, int (*handler)(void *), 
		     void *arg, char *name)
{
    struct intrhand *ih;
    struct interrupt *in;
    pthread_doorbellattr_t dba;
    pthread_attr_t tha;
    struct sched_param sp;
    int s;

    ih = (struct intrhand *) malloc (sizeof (struct intrhand));
    if (!ih)
	panic ("kern_intr_establish");

    ih->ipl = ipl;
    ih->handler = handler;
    ih->arg = arg;
    ih->name = name;

    in = &intrtable[inum];
    s = splhigh ();
    if (in->maxipl == 0) {
	LIST_INIT (&in->head);
	in->maxipl = ipl;

	pthread_doorbellattr_init (&dba);
	pthread_doorbellattr_setevn (&dba, inum);
	pthread_doorbell_init (&in->db, &dba);
	pthread_doorbellattr_destroy (&dba);

	/* Start thread */
	pthread_attr_init(&tha);
    
	/* set FIFO scheduling */
	pthread_attr_setdetachstate (&tha, PTHREAD_CREATE_DAEMON_NP);
	pthread_attr_setinheritsched (&tha, PTHREAD_EXPLICIT_SCHED);
	pthread_attr_setschedpolicy (&tha, SCHED_FIFO);
	
	/* set priority based on IPL */
	pthread_attr_getschedparam (&tha,&sp);
	sp.sched_priority = sched_get_priority_min (SCHED_FIFO) + ipl;
	pthread_attr_setschedparam (&tha, &sp);

	/* XXX this intr could be handling several devices */
	pthread_attr_setname_np (&tha, name);
    
	if (pthread_create (&in->tid, &tha, intrupt_thread, in) != 0)
	    panic ("intrthread create");
	
	pthread_attr_destroy (&tha);
    }
    else if (ipl > in->maxipl) {
	in->maxipl = ipl;
	/* XXX Change priority of running thread? */
    }
    LIST_INSERT_HEAD (&in->head, ih, ichain);
    (void) splx (s);

    return ih;
}


/* Remove an "interrupt" handler */
void
_kern_intr_disestablish (void *cookie)
{
    struct intrhand *ih = cookie;
    int s = splhigh ();
    LIST_REMOVE (ih, ichain);
    (void) splx (s);
}


/* Schedule a softnet "interrupt" */
/* XXX must be called at splnet or above */
void
setsoftnet ()
{
#ifdef DIAGNOSTIC
    if (*_kern_getcuripl () < IPL_SOFTNET)
	panic ("setsoftnet");
#endif
    if (!softnet_pend) {
	softnet_pend = 1;
	pthread_cond_signal (&softnet_cd);
    }
}


/* Delay for usec microseconds. */
void
_kern_delay (usec)
    int usec;
{
    extern int errno;
    struct timespec ts;

    ts.tv_sec = usec / 1000000;
    ts.tv_nsec = usec * 1000;
    while (nanosleep (&ts, &ts) < 0 && errno == EINTR)
	continue;
}

void _bsd_startup __P((void));

/*
 * cpu_startup: allocate memory for variable-sized tables,
 * initialize cpu, and do autoconfiguration.
 */
void
cpu_startup()
{
    pthread_attr_t tha;
    pthread_mutexattr_t mxa;
    struct sched_param sp;
    vm_offset_t maxaddr;
    void *base;
    int nintr;

    /* 
     * Create per-thread ipl key 
     */
    if (pthread_key_create (&ipl_key, free) != 0)
	panic ("ipl_key");

    /* 
     * Initialise spl mutexes
     */
    pthread_mutexattr_init (&mxa);
    pthread_mutexattr_setprotocol (&mxa, PTHREAD_PRIO_INHERIT);
    pthread_mutexattr_setname_np (&mxa, "splsoftclock");
    pthread_mutex_init (&splsoftclock_mx, &mxa);
    pthread_mutexattr_setname_np (&mxa, "splsoftnet");
    pthread_mutex_init (&splsoftnet_mx, &mxa);
    pthread_mutexattr_setname_np (&mxa, "splnet");
    pthread_mutex_init (&splnet_mx, &mxa);
    pthread_mutexattr_destroy (&mxa);

    /* initialise condition variable */
    pthread_cond_setname_np (&softnet_cd, "softnet");

    /*
     * Allocate and initialise low-level error message buffer.
     */
    initmsgbuf (pthread_getpages_np (NBPG), NBPG);
    
#define VM_SIZE	(VM_KMEM_SIZE + VM_MBUF_SIZE + NBPG*4)
    /* Initialise kernel "memory map" */
    base = pthread_getpages_np (VM_SIZE);
    if (!base)
	panic ("no mem for kernel_map");
    kmem_init ((vm_offset_t)base, (vm_offset_t)base + VM_SIZE);

    /* Initialise kernel malloc pool */
    kmeminit ();

    /* 
     * Allocate interrupt table 
     */
    nintr = _sbd_intr_howmany ();
    intrtable = (struct interrupt *) malloc(nintr * sizeof (struct interrupt));
    bzero(intrtable, nintr * sizeof (struct interrupt));

    /*
     * Allocate mbuf cluster refcnt table.
     */
    mclrefcnt = (char *) malloc(NMBCLUSTERS+CLBYTES/MCLBYTES);
    bzero(mclrefcnt, NMBCLUSTERS+CLBYTES/MCLBYTES);

    mb_map = kmem_suballoc(kernel_map, (void *)&mbutl, &maxaddr,
			   VM_MBUF_SIZE, FALSE);

    /*
     * Start softnet "interrupt" thread
     */
    pthread_attr_init(&tha);
    
    /* set FIFO scheduling */
    pthread_attr_setdetachstate (&tha, PTHREAD_CREATE_DAEMON_NP);
    pthread_attr_setinheritsched (&tha, PTHREAD_EXPLICIT_SCHED);
    pthread_attr_setschedpolicy (&tha, SCHED_FIFO);
    
    /* set minumum FIFO priority + ipl */
    pthread_attr_getschedparam (&tha,&sp);
    sp.sched_priority = sched_get_priority_min (SCHED_FIFO) + IPL_SOFTNET;
    pthread_attr_setschedparam (&tha, &sp);

    pthread_attr_setname_np (&tha, "softnet");
    
    if (pthread_create (NULL, &tha, softnet_thread, NULL) != 0)
	panic ("softnet_thread create");

    pthread_attr_destroy (&tha);

    /* Initialise timeouts */
    timeout_init();

    splhigh ();	

    if (config_rootfound("mainbus", "mainbus") == 0)
	panic("no mainbus found");

    cold = 0;
    
    spl0();
}



/* XXX Bus DMA mapping.

   This feature isn't used by any of the P4032 drivers, but may be
   needed in the future. In particular it could be a way to handle the
   Bonito IOBC cleanly, see _bus_dmamap_sync(). */

/*
 * Common function for DMA map creation.  May be called by bus-specific
 * DMA map creation functions.
 */
int
_bus_dmamap_create(t, size, nsegments, maxsegsz, boundary, flags, dmamp)
	bus_dma_tag_t t;
	bus_size_t size;
	int nsegments;
	bus_size_t maxsegsz;
	bus_size_t boundary;
	int flags;
	bus_dmamap_t *dmamp;
{
	struct algrtx_bus_dmamap *map;
	void *mapstore;
	size_t mapsize;

	/*
	 * Allocate and initialize the DMA map.  The end of the map
	 * is a variable-sized array of segments, so we allocate enough
	 * room for them in one shot.
	 *
	 * Note we don't preserve the WAITOK or NOWAIT flags.  Preservation
	 * of ALLOCNOW notifies others that we've reserved these resources,
	 * and they are not to be freed.
	 *
	 * The bus_dmamap_t includes one bus_dma_segment_t, hence
	 * the (nsegments - 1).
	 */
	mapsize = sizeof(struct algrtx_bus_dmamap) +
	    (sizeof(bus_dma_segment_t) * (nsegments - 1));
	if ((mapstore = _kern_malloc(mapsize, M_DEVBUF,
	    (flags & BUS_DMA_NOWAIT) ? M_NOWAIT : M_WAITOK)) == NULL)
		return (ENOMEM);

	bzero(mapstore, mapsize);
	map = (struct algrtx_bus_dmamap *)mapstore;
	map->_dm_size = size;
	map->_dm_segcnt = nsegments;
	map->_dm_maxsegsz = maxsegsz;
	map->_dm_boundary = boundary;
	map->_dm_flags = flags & ~(BUS_DMA_WAITOK|BUS_DMA_NOWAIT);
	map->dm_mapsize = 0;		/* no valid mappings */
	map->dm_nsegs = 0;		/* no valid mappings */

	*dmamp = map;
	return (0);
}

/*
 * Common function for DMA map destruction.  May be called by bus-specific
 * DMA map destruction functions.
 */
void
_bus_dmamap_destroy(t, map)
	bus_dma_tag_t t;
	bus_dmamap_t map;
{

	_kern_free(map, M_DEVBUF);
}

/*
 * Common function for loading a DMA map with a linear buffer.  May
 * be called by bus-specific DMA map load functions.
 */
int
_bus_dmamap_load_buffer(t, map, buf, buflen, p, flags, lastaddrp, segp, first)
	bus_dma_tag_t t;
	bus_dmamap_t map;
	void *buf;
	bus_size_t buflen;
	struct proc *p;
	int flags;
	bus_addr_t *lastaddrp;
	int *segp;
	int first;
{
	bus_size_t sgsize;
	bus_addr_t curaddr, lastaddr, baddr, bmask;
	caddr_t vaddr = buf;
	int seg;

	lastaddr = *lastaddrp;
	bmask = ~(map->_dm_boundary - 1);

	for (seg = *segp; buflen > 0; ) {

		sgsize = buflen;

		/* Check for maximum segment size */
		if (sgsize > map->_dm_maxsegsz)
		    	sgsize = map->_dm_maxsegsz;

		/* Get the physical address for this segment */
		curaddr = bus_dmamap_vtop (t, vaddr, sgsize);

		/* Make sure we don't cross any boundaries. */
		if (map->_dm_boundary > 0) {
			baddr = (curaddr + map->_dm_boundary) & bmask;
			if (sgsize > (baddr - curaddr))
				sgsize = (baddr - curaddr);
		}

		/* Add to table */
		if (first) {
			map->dm_segs[seg].ds_addr = curaddr;
			map->dm_segs[seg].ds_len = sgsize;
			map->dm_segs[seg]._ds_vaddr = vaddr;
			first = 0;
		} else {
		    /* Probably no point in this optimisation for AlgRTX */
			if (curaddr == lastaddr &&
			    (map->dm_segs[seg].ds_len + sgsize) <=
			     map->_dm_maxsegsz &&
			    (map->_dm_boundary == 0 ||
			     (map->dm_segs[seg].ds_addr & bmask) ==
			     (curaddr & bmask)))
				map->dm_segs[seg].ds_len += sgsize;
			else {
				if (++seg >= map->_dm_segcnt)
					break;
				map->dm_segs[seg].ds_addr = curaddr;
				map->dm_segs[seg].ds_len = sgsize;
				map->dm_segs[seg]._ds_vaddr = vaddr;
			}
		}

		lastaddr = curaddr + sgsize;
		vaddr += sgsize;
		buflen -= sgsize;
	}

	*segp = seg;
	*lastaddrp = lastaddr;

	/*
	 * Did we fit?
	 */
	if (buflen != 0)
		return (EFBIG);		/* XXX better return value here? */

	return (0);
}

/*
 * Common function for loading a direct-mapped DMA map with a linear
 * buffer.  Called by bus-specific DMA map load functions with the
 * OR value appropriate for indicating "direct-mapped" for that
 * chipset.
 */
int
_bus_dmamap_load(bus_dma_tag_t t, bus_dmamap_t map, void *buf,
    bus_size_t buflen, struct proc *p, int flags)
{
	bus_addr_t lastaddr;
	int seg, error;

	/*
	 * Make sure that on error condition we return "no valid mappings".
	 */
	map->dm_mapsize = 0;
	map->dm_nsegs = 0;

	if (buflen > map->_dm_size)
		return (EINVAL);

	seg = 0;
	error = _bus_dmamap_load_buffer(t, map, buf, buflen,
	    p, flags, &lastaddr, &seg, 1);
	if (error == 0) {
		map->dm_mapsize = buflen;
		map->dm_nsegs = seg + 1;
		
		/*
		 * For linear buffers, we support marking the mapping
		 * as COHERENT.
		 */
		if (IS_KVA1 (buf))
			map->_dm_flags |= ALGOR_DMAMAP_COHERENT;
	}
	return (error);
}


/*
 * Like _bus_dmamap_load(), but for mbufs.
 */
int
_bus_dmamap_load_mbuf(t, map, m0, flags)
	bus_dma_tag_t t;
	bus_dmamap_t map;
	struct mbuf *m0;
	int flags;
{
	bus_addr_t lastaddr;
	int seg, error, first;
	struct mbuf *m;

	/*
	 * Make sure that on error condition we return "no valid mappings."
	 */
	map->dm_mapsize = 0;
	map->dm_nsegs = 0;

	if ((m0->m_flags & M_PKTHDR) == 0)
		panic("_bus_dmamap_load_mbuf: no packet header");

	if (m0->m_pkthdr.len > map->_dm_size)
		return (EINVAL);

	first = 1;
	seg = 0;
	error = 0;
	for (m = m0; m != NULL && error == 0; m = m->m_next) {
		error = _bus_dmamap_load_buffer(t, map,
		    m->m_data, m->m_len, NULL, flags, &lastaddr, &seg, first);
		first = 0;
	}
	if (error == 0) {
		map->dm_mapsize = m0->m_pkthdr.len;
		map->dm_nsegs = seg + 1;
	}
	return (error);
}

/*
 * Like _bus_dmamap_load(), but for uios.
 */
int
_bus_dmamap_load_uio(t, map, uio, flags)
	bus_dma_tag_t t;
	bus_dmamap_t map;
	struct uio *uio;
	int flags;
{

	panic("_bus_dmamap_load_uio: not implemented");
}

/*
 * Like _bus_dmamap_load(), but for raw memory allocated with
 * bus_dmamem_alloc().
 */
int
_bus_dmamap_load_raw(t, map, segs, nsegs, size, flags)
	bus_dma_tag_t t;
	bus_dmamap_t map;
	bus_dma_segment_t *segs;
	int nsegs;
	bus_size_t size;
	int flags;
{

	panic("_bus_dmamap_load_raw: not implemented");
}

/*
 * Common function for unloading a DMA map.  May be called by
 * bus-specific DMA map unload functions.
 */
void
_bus_dmamap_unload(t, map)
	bus_dma_tag_t t;
	bus_dmamap_t map;
{

	/*
	 * No resources to free; just mark the mappings as
	 * invalid.
	 */
	map->dm_mapsize = 0;
	map->dm_nsegs = 0;
	map->_dm_flags &= ~ALGOR_DMAMAP_COHERENT;
}

/*
 * Common function for DMA map synchronization.  May be called
 * by bus-specific DMA map synchronization functions.
 */
void
_bus_dmamap_sync(t, map, offset, len, ops)
	bus_dma_tag_t t;
	bus_dmamap_t map;
	bus_addr_t offset;
	bus_size_t len;
	int ops;
{
	bus_size_t minlen;
	int i;

	/*
	 * Mixing PRE and POST operations is not allowed.
	 */
	if ((ops & (BUS_DMASYNC_PREREAD|BUS_DMASYNC_PREWRITE)) != 0 &&
	    (ops & (BUS_DMASYNC_POSTREAD|BUS_DMASYNC_POSTWRITE)) != 0)
		panic("_bus_dmamap_sync: mix PRE and POST");

#ifdef DIAGNOSTIC
	if (offset >= map->dm_mapsize)
		panic("_bus_dmamap_sync: bad offset %lu (map size is %lu)",
		    offset, map->dm_mapsize);
	if (len == 0 || (offset + len) > map->dm_mapsize)
		panic("_bus_dmamap_sync: bad length");
#endif

	/*
	 * If the mapping is of COHERENT DMA-safe memory, no cache
	 * flush is necessary.
	 */
	if (map->_dm_flags & ALGOR_DMAMAP_COHERENT)
		return;

	/*
	 * No cache flushes are necessary if we're only doing
	 * POSTREAD or POSTWRITE (i.e. not doing PREREAD or PREWRITE).
	 */
	if ((ops & (BUS_DMASYNC_PREREAD|BUS_DMASYNC_PREWRITE)) == 0)
		return;

	/*
	 * Flush data cache for PREREAD.  This has the side-effect
	 * of invalidating the cache.  Done at PREREAD since it
	 * causes the cache line(s) to be written back to memory.
	 *
	 * Flush data cache for PREWRITE, so that the contents of
	 * the data buffer in memory reflect reality.
	 *
	 * Give the test above, we know we're doing one of these
	 * two operations, so no additional tests are necessary.
	 */

	for (i = 0; i < map->dm_nsegs && len != 0; i++) {
		/* Find the beginning segment. */
		if (offset >= map->dm_segs[i].ds_len) {
			offset -= map->dm_segs[i].ds_len;
			continue;
		}

		/*
		 * Now at the first segment to sync; nail
		 * each segment until we have exhausted the
		 * length.
		 */
		minlen = len < map->dm_segs[i].ds_len - offset ?
		    len : map->dm_segs[i].ds_len - offset;
		mips_clean_cache ((vaddr_t)map->dm_segs[i]._ds_vaddr + offset,
				  len);
		offset = 0;
		len -= minlen;
	}
}

/*
 * Common function for DMA-safe memory allocation.  May be called
 * by bus-specific DMA memory allocation functions.
 */
int
_bus_dmamem_alloc(t, size, alignment, boundary, segs, nsegs, rsegs, flags)
	bus_dma_tag_t t;
	bus_size_t size, alignment, boundary;
	bus_dma_segment_t *segs;
	int nsegs;
	int *rsegs;
	int flags;
{
       void * b;

       if ((b = _kern_malloc(size, M_DEVBUF,
	    (flags & BUS_DMA_NOWAIT) ? M_NOWAIT : M_WAITOK)) == NULL)
		return (ENOMEM);

	segs[0]._ds_vaddr = b;
	segs[0].ds_len = PAGE_SIZE;
	segs[0].ds_addr = bus_dmamap_vtop (t, (caddr_t) b, PAGE_SIZE);
	*rsegs = 1;

	return (0);
}

/*
 * Common function for freeing DMA-safe memory.  May be called by
 * bus-specific DMA memory free functions.
 */
void
_bus_dmamem_free(t, segs, nsegs)
	bus_dma_tag_t t;
	bus_dma_segment_t *segs;
	int nsegs;
{
    int seg;

    for (seg = 0; seg < nsegs; seg++)
	_kern_free(segs[seg]._ds_vaddr, M_DEVBUF);
}

/*
 * Common function for mapping DMA-safe memory.  May be called by
 * bus-specific DMA memory map functions.
 */
int
_bus_dmamem_map(t, segs, nsegs, size, kvap, flags)
	bus_dma_tag_t t;
	bus_dma_segment_t *segs;
	int nsegs;
	size_t size;
	caddr_t *kvap;
	int flags;
{
        if (nsegs != 1)
	    panic("_bus_dmamem_map");;

	/* Pass back base address as an uncached virtual address, 
	   safe for descriptor access, etc. */
    	if (flags & BUS_DMA_COHERENT) {
	    /* clean cache before passing uncached address to caller */
	    mips_clean_cache((vaddr_t)segs[0]._ds_vaddr, segs[0].ds_len);
	    *kvap = KVA0_TO_KVA1 (segs[0]._ds_vaddr);
	}
	else
	    *kvap = segs[0]._ds_vaddr;
	return (0);
}

/*
 * Common function for unmapping DMA-safe memory.  May be called by
 * bus-specific DMA memory unmapping functions.
 */
void
_bus_dmamem_unmap(t, kva, size)
	bus_dma_tag_t t;
	caddr_t kva;
	size_t size;
{

	/* Nothing to do */
}

/*
 * Common functin for mmap(2)'ing DMA-safe memory.  May be called by
 * bus-specific DMA mmap(2)'ing functions.
 */
int
_bus_dmamem_mmap(t, segs, nsegs, off, prot, flags)
	bus_dma_tag_t t;
	bus_dma_segment_t *segs;
	int nsegs, off, prot, flags;
{
	panic ("_bus_dmamem_mmap: not implemented");
}


/* Arrange for kernel to be started automatically */

#include <sys/init.h>
extern void _kern_init (void);
DECL_INIT_FUNC (IF_PREMAIN, _kern_init)
