/* 
 * Copyright (c) 2001 Algorithmics Ltd
 *
 * An implementation of BSD sleep/wakeup on top of AlgRTX thread primitives. 
 *
 * This program is NOT free software, it is supplied under the terms
 * of the AlgNET License Agreement, a copy of which is available at:
 *
 *  http://www.algor.co.uk/algor/info/algnet-license.pdf
 *
 * Any company which has obtained and signed a valid AlgNET license
 * may use and modify this software internally and use (without
 * restrictions) any derived binary.  You may not, however,
 * redistribute this in whole or in part as source code, nor may you
 * modify or remove any part of this copyright message.  
 */

/*	$OpenBSD: kern_synch.c,v 1.28 2000/08/03 14:36:36 mickey Exp $	*/
/*	$NetBSD: kern_synch.c,v 1.37 1996/04/22 01:38:37 christos Exp $	*/

/*-
 * Copyright (c) 1982, 1986, 1990, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 * (c) UNIX System Laboratories, Inc.
 * All or some portions of this file are derived from material licensed
 * to the University of California by American Telephone and Telegraph
 * Co. or Unix System Laboratories, Inc. and are reproduced herein with
 * the permission of UNIX System Laboratories, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)kern_synch.c	8.6 (Berkeley) 1/21/94
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/timeout.h>

int	lbolt;			/* once a second sleep address */

static void
schedcpu(arg)
	void *arg;
{
    	/* lbolt wakeup once per second */
	struct timeout *to = (struct timeout *)arg;
	wakeup((caddr_t)&lbolt);
	timeout_add(to, hz);
}


void
scheduler_start()
{
	static struct timeout schedcpu_to;
	timeout_set(&schedcpu_to, schedcpu, &schedcpu_to);
	schedcpu(&schedcpu_to);
}


/*
 * We're only hashing with 7 bits of the address; everything is
 * aligned to 8, lots of things are aligned to greater powers
 * of 2.  Shift right by 8, i.e. drop the bottom 256 worth.
 */
#define TABLESIZE	128
#define LOOKUP(x)	(((long)(x) >> 8) & (TABLESIZE - 1))
static struct slpque {
	struct slpchan	*sq_head;
	struct slpchan **sq_tailp;
} slpque[TABLESIZE];

static pthread_mutex_t	slpque_mx = PTHREAD_MUTEX_INITIALIZER;

struct slpchan {
    struct slpchan	*sl_next;
    void		*sl_ident;
    pthread_mutex_t	sl_mx;
    pthread_cond_t	sl_cond;
    int			sl_condwait;
    pthread_cond_t	sl_intrcond;
    int			sl_intrcondwait;
};
    

/*
 * During autoconfiguration or after a panic, a sleep will simply
 * lower the priority briefly to allow interrupts, then return.
 * The priority to be used (safepri) is machine-dependent, thus this
 * value is initialized and maintained in the machine-dependent layers.
 * This priority will typically be 0, or the lowest priority
 * that is safe for use on the interrupt stack; it can be made
 * higher to block network software interrupts after panics.
 */
int safepri;


/*
 * General sleep call.  Suspends the current thread until a wakeup is
 * performed on the specified identifier.  The thread will then be made
 * runnable, but the priority is ignored for AlgRTX.  Sleeps at most timo/hz 
 * seconds (0 means no timeout).  
 * If pri includes PCATCH flag, signals are checked.  Returns 0 if
 * awakened, EWOULDBLOCK if the timeout expires.  If PCATCH is set and a
 * signal needs to be delivered, ERESTART is returned if the current system
 * call should be restarted if possible, and EINTR is returned if the system
 * call should be interrupted by the signal (return EINTR).
 */
int
tsleep(ident, priority, wmesg, timo)
	void *ident;
	int priority, timo;
	char *wmesg;
{
	struct slpque *qp;
	struct slpchan *sl;
	struct timespec timo_ts, *ts;
	int catch = priority & PCATCH;
	int ret, s;
	extern int cold;

	if (cold || panicstr) {
		/*
		 * After a panic, or during autoconfiguration,
		 * just give interrupts a chance, then just return;
		 * don't run any other procs or panic below,
		 * in case this is the idle process and already asleep.
		 */
		int s = splx(safepri);
		splx(s);
		return (0);
	}

#ifdef DIAGNOSTIC
	if (ident == NULL)
	    panic("tsleep");
#endif

	pthread_mutex_lock (&slpque_mx);

	/* find the identifier */
	qp = &slpque[LOOKUP(ident)];
	for (sl = qp->sq_head; sl != NULL; sl = sl->sl_next)
	    if (sl->sl_ident == ident)
		break;

	if (sl == NULL) {
	    pthread_mutexattr_t mxa;
	    pthread_condattr_t cda;

	    /* create new identifier */
	    MALLOC (sl, struct slpchan *, sizeof (*sl), M_SLEEP, M_WAITOK);
	    if (qp->sq_head == 0)
		qp->sq_head = sl;
	    else
		*qp->sq_tailp = sl;
	    *(qp->sq_tailp = &sl->sl_next) = 0;

	    sl->sl_ident = ident;
	    sl->sl_condwait = sl->sl_intrcondwait = 0;

	    /* create mutex and assign wmesg as name */
	    pthread_mutexattr_init (&mxa);
	    if (wmesg)
		pthread_mutexattr_setname_np (&mxa, wmesg);
	    pthread_mutex_init (&sl->sl_mx, &mxa);
	    pthread_mutexattr_destroy (&mxa);

	    /* create uninterruptible condition variable */
	    pthread_condattr_init (&cda);
	    if (wmesg)
		pthread_condattr_setname_np (&cda, wmesg);
	    pthread_cond_init (&sl->sl_cond, &cda);
	    pthread_condattr_destroy (&cda);

	    /* create interruptible condition variable */
	    pthread_condattr_init (&cda);
	    if (wmesg)
		pthread_condattr_setname_np (&cda, wmesg);
	    pthread_condattr_setinterruptible_np (&cda, 1);
	    pthread_cond_init (&sl->sl_intrcond, &cda);
	    pthread_condattr_destroy (&cda);
	}

	pthread_mutex_unlock (&slpque_mx);

	/* set up timeout */
	if (timo)
	    hztoabs (ts = &timo_ts, timo);
	else
	    ts = NULL;

	pthread_mutex_lock (&sl->sl_mx);
	s = spl0 ();
	if (catch) {
	    sl->sl_intrcondwait++;
	    ret = pthread_cond_timedwait (&sl->sl_intrcond, &sl->sl_mx, ts);
	    sl->sl_intrcondwait--;
	}
	else {
	    sl->sl_condwait++;
	    ret = pthread_cond_timedwait (&sl->sl_cond, &sl->sl_mx, ts);
	    sl->sl_condwait--;
	}
	(void) splx (s);
	pthread_mutex_unlock (&sl->sl_mx);

	if (ret != 0) {
	    /* convert ETIME to EWOULDBLOCK */
	    if (ret == ETIME)
		ret = EWOULDBLOCK;
	    /* only other error should be EINTR */
	    else if (! catch || ret != EINTR)
		panic ("tsleep condwait");
	}
	

	return (ret);
}


/*
 * Short-term, non-interruptable sleep.
 */
void
sleep(ident, priority)
	void *ident;
	int priority;
{
    tsleep(ident, priority, "sleep", 0);
}


/*
 * Remove a process from its wait queue
 */
void
unsleep(p)
	register struct proc *p;
{
    /* XXX do we need this? */
    panic("unsleep");
}


/*
 * Make all processes sleeping on the specified identifier runnable.
 */
void
wakeup(ident)
	void *ident;
{
    struct slpque *qp;
    struct slpchan *sl;

    /* find the identifier */
    pthread_mutex_lock (&slpque_mx);
    qp = &slpque[LOOKUP(ident)];
    for (sl = qp->sq_head; sl != NULL; sl = sl->sl_next)
	if (sl->sl_ident == ident)
	    break;
    pthread_mutex_unlock (&slpque_mx);
    
    /* nothing to do if no-one has slept on it yet */
    if (!sl)
	return;
    
    /* wakeup blocked threads */
    pthread_mutex_lock (&sl->sl_mx);
    if (sl->sl_condwait != 0)
	pthread_cond_broadcast (&sl->sl_cond);
    if (sl->sl_intrcondwait != 0)
	pthread_cond_broadcast (&sl->sl_intrcond);
    pthread_mutex_unlock (&sl->sl_mx);
}
