/* 
 * An implementation of the BSD timeout facility using AlgRTX 
 * timing facilities.
 *
 * Copyright (c) 2001 Algorithmics Ltd
 *
 * This program is NOT free software, it is supplied under the terms
 * of the AlgNET License Agreement, a copy of which is available at:
 *
 *  http://www.algor.co.uk/algor/info/algnet-license.pdf
 *
 * Any company which has obtained and signed a valid AlgNET license
 * may use and modify this software internally and use (without
 * restrictions) any derived binary.  You may not, however,
 * redistribute this in whole or in part as source code, nor may you
 * modify or remove any part of this copyright message.  
 */

/*	$OpenBSD$	*/

/*
 * Copyright (c) 2000 Artur Grabowski <art@openbsd.org>
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/timeout.h>


/*
 * Timeouts are kept on a queue. The to_ts is the wallclock time
 * when the timeout should be called.
 *
 * In the future we might want to build a timer wheel to improve the speed
 * of timeout_add (right now it's linear). See "Redesigning the BSD Callout 
 * and Timer Facilities" by Adam M. Costello and Geroge Varghese.
 */

static TAILQ_HEAD(,timeout) timeout_todo;	/* Queue of timeouts. */
static TAILQ_HEAD(,timeout) timeout_static;	/* Static pool of timeouts. */

/*
 * All lists are locked with the same lock.
 */
static pthread_mutex_t	timeout_lock;
static pthread_cond_t	timeout_cond = PTHREAD_COND_INITIALIZER;

#define timeout_list_lock(s) \
	do { pthread_mutex_lock (&timeout_lock); } while (0)
#define timeout_list_unlock(s) \
	do { pthread_mutex_unlock(&timeout_lock); } while (0)

static void * timeout_thread (void *);

void
timeout_init()
{
	pthread_attr_t tha;
	pthread_mutexattr_t mxa;
	struct sched_param sp;
	int i;

	TAILQ_INIT(&timeout_todo);
	TAILQ_INIT(&timeout_static);
	
	for (i = 0; i < ntimeout; i++)
		TAILQ_INSERT_HEAD(&timeout_static, &timeouts[i], to_list);

	/* initialise timeout mutex */
	pthread_mutexattr_init (&mxa);
	pthread_mutexattr_setprotocol (&mxa, PTHREAD_PRIO_INHERIT);
	pthread_mutexattr_setname_np (&mxa, "ktimemx");
	pthread_mutex_init (&timeout_lock, &mxa);
	pthread_mutexattr_destroy (&mxa);

	/* initialise condition variable */
	pthread_cond_setname_np (&timeout_cond, "ktick");

	/* create timeout thread */
	pthread_attr_init(&tha);
	
	/* set FIFO scheduling */
	pthread_attr_setdetachstate (&tha, PTHREAD_CREATE_DAEMON_NP);
	pthread_attr_setinheritsched (&tha, PTHREAD_EXPLICIT_SCHED);
	pthread_attr_setschedpolicy (&tha, SCHED_FIFO);
	
	/* set minumum FIFO priority + 1 */
	pthread_attr_getschedparam (&tha,&sp);
	sp.sched_priority = sched_get_priority_min (SCHED_FIFO) + 1;
	pthread_attr_setschedparam (&tha, &sp);

	pthread_attr_setname_np (&tha, "ktimout");
    
    	if (pthread_create (NULL, &tha, timeout_thread, NULL) != 0)
	    panic ("timeout_thread create");

	pthread_attr_destroy (&tha);
}


void
timeout_set(to, fn, arg)
	struct timeout *to;
	void (*fn)(void *);
	void *arg;
{

	to->to_func = fn;
	to->to_arg = arg;
	to->to_flags = TIMEOUT_INITIALIZED;
}

void
timeout_add(new, to_ticks)
	struct timeout *new;
	int to_ticks;
{
	struct timeout *to;

#ifdef DIAGNOSTIC
	if (to_ticks < 0)
		panic("timeout_add: to_ticks < 0");
#endif

	timeout_list_lock(&s);

	_RTPX_ELOG ("timeout_add %p(%p) in %d ticks", new->to_func, new->to_arg,
		    to_ticks);

	/*
	 * First we prepare the new timeout so that we can return right
	 * after the insertion in the queue (makes the code simpler).
	 */

	/* If this timeout was already on a queue we remove it. */
	if (new->to_flags & TIMEOUT_ONQUEUE)
		TAILQ_REMOVE(&timeout_todo, new, to_list);
	else
		new->to_flags |= TIMEOUT_ONQUEUE;

	/* Initialize the time here, it won't change. */
	hztoabs(&new->to_ts, to_ticks);

	/*
	 * Walk the list of pending timeouts and find an entry which
	 * will timeout after we do, insert the new timeout there.
	 */
	TAILQ_FOREACH(to, &timeout_todo, to_list) {
	        if (timespeccmp (&to->to_ts, &new->to_ts, >)) {
			TAILQ_INSERT_BEFORE(to, new, to_list);
			goto out;
		}
	}

	/* We can only get here if we're the last (or only) entry */
	TAILQ_INSERT_TAIL(&timeout_todo, new, to_list);

out:
	/* Wake up the alarm thread if the new request is head of the
	   list, because the wait must be restarted with the new
	   (earlier) timeout. */
	if (TAILQ_FIRST (&timeout_todo) == new)
	    pthread_cond_signal (&timeout_cond);

	timeout_list_unlock(s);
}

void
timeout_del(to)
	struct timeout *to;
{
	timeout_list_lock(&s);
	if (to->to_flags & TIMEOUT_ONQUEUE) {
		TAILQ_REMOVE(&timeout_todo, to, to_list);
		to->to_flags &= ~TIMEOUT_ONQUEUE;
	}
	timeout_list_unlock(s);
}

static void *
timeout_thread (void *ign)
{
    sigset_t sm;
    struct timeout *to;
    void (*fn) __P((void *));
    void *arg;
    int s;

    /* disable all signals in this thread */
    sigfillset (&sm);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);

    pthread_mutex_lock (&timeout_lock);
    while (1) {
	struct timespec now, *ts;

	/* get time of first timeout on list */
	if (TAILQ_EMPTY (&timeout_todo))
	    ts = NULL;
	else
	    ts = &(TAILQ_FIRST (&timeout_todo)->to_ts);

	/* wait until timeout or wakeup */
	pthread_cond_timedwait (&timeout_cond, &timeout_lock, ts);

	clock_gettime (CLOCK_REALTIME, &now);
	
	while ((to = TAILQ_FIRST(&timeout_todo)) != NULL &&
	       timespeccmp (&to->to_ts, &now, <=)) {

		TAILQ_REMOVE(&timeout_todo, to, to_list);
		to->to_flags &= ~TIMEOUT_ONQUEUE;

		fn = to->to_func;
		arg = to->to_arg;

		if (to->to_flags & TIMEOUT_STATIC)
			TAILQ_INSERT_HEAD(&timeout_static, to, to_list);

		_RTPX_ELOG ("timeout calling %p(%p)", fn, arg);

		pthread_mutex_unlock (&timeout_lock);

		s = splsoftclock ();
		fn(arg);
		splx (s);

		pthread_mutex_lock (&timeout_lock);

		clock_gettime (CLOCK_REALTIME, &now);
	}
    }
    return 0;
}

/*
 * Legacy interfaces. timeout() and untimeout()
 *
 * Kill those when everything is converted. They are slow and use the
 * static pool (which causes (potential and real) problems).
 */

void
timeout(fn, arg, to_ticks)
	void (*fn) __P((void *));
	void *arg;
	int to_ticks;
{
	struct timeout *to;

	if (to_ticks <= 0)
		to_ticks = 1;

	/*
	 * Get a timeout struct from the static list.
	 */
	timeout_list_lock(&s);

	to = TAILQ_FIRST(&timeout_static);
	if (to == NULL)
		panic("timeout table full");
	TAILQ_REMOVE(&timeout_static, to, to_list);

	timeout_list_unlock(s);

	timeout_set(to, fn, arg);
	to->to_flags |= TIMEOUT_STATIC;
	timeout_add(to, to_ticks);
}

void
untimeout(fn, arg)
	void (*fn) __P((void *));
	void *arg;
{
	struct timeout *to;

	timeout_list_lock(&s);
	TAILQ_FOREACH(to, &timeout_todo, to_list) {
		if (to->to_func == fn && to->to_arg == arg) {
#ifdef DIAGNOSTIC
			if ((to->to_flags & TIMEOUT_ONQUEUE) == 0)
				panic("untimeout: not TIMEOUT_ONQUEUE");
			if ((to->to_flags & TIMEOUT_STATIC) == 0)
				panic("untimeout: not static");
#endif
			TAILQ_REMOVE(&timeout_todo, to, to_list);
			to->to_flags &= ~TIMEOUT_ONQUEUE;
			/* return it to the static pool */
			TAILQ_INSERT_HEAD(&timeout_static, to, to_list);
			break;
		}
	}
	timeout_list_unlock(s);
}


/*
 * Compute number of hz until specified time.  Used to
 * compute the second argument to timeout_add() from an absolute time.
 */
int
hzto(tvp)
	struct timeval *tvp;
{
	struct timespec ts;
	struct timeval tv;
	unsigned long ticks;
	long sec, usec;

	/*
	 * If the number of usecs in the whole seconds part of the time
	 * difference fits in a long, then the total number of usecs will
	 * fit in an unsigned long.  Compute the total and convert it to
	 * ticks, rounding up and adding 1 to allow for the current tick
	 * to expire.  Rounding also depends on unsigned long arithmetic
	 * to avoid overflow.
	 *
	 * Otherwise, if the number of ticks in the whole seconds part of
	 * the time difference fits in a long, then convert the parts to
	 * ticks separately and add, using similar rounding methods and
	 * overflow avoidance.  This method would work in the previous
	 * case but it is slightly slower and assumes that hz is integral.
	 *
	 * Otherwise, round the time difference down to the maximum
	 * representable value.
	 *
	 * If ints have 32 bits, then the maximum value for any timeout in
	 * 10ms ticks is 248 days.
	 */
	clock_gettime (CLOCK_REALTIME, &ts);
	TIMESPEC_TO_TIMEVAL (&tv, &ts);
	sec = tvp->tv_sec - tv.tv_sec;
	usec = tvp->tv_usec - tv.tv_usec;
	if (usec < 0) {
		sec--;
		usec += 1000000;
	}
	if (sec < 0 || (sec == 0 && usec <= 0)) {
		ticks = 0;
	} else if (sec <= LONG_MAX / 1000000)
		ticks = (sec * 1000000 + (unsigned long)usec + (tick - 1))
		    / tick + 1;
	else if (sec <= LONG_MAX / hz)
		ticks = sec * hz
		    + ((unsigned long)usec + (tick - 1)) / tick + 1;
	else
		ticks = LONG_MAX;
	if (ticks > INT_MAX)
		ticks = INT_MAX;
	return ((int)ticks);
}


void hztoabs(ts, ticks)
    struct timespec *ts;
    int ticks;
{
    clock_gettime (CLOCK_REALTIME, ts);
    ts->tv_sec += ticks / hz;
    ts->tv_nsec += (ticks % hz) * tick * 1000;
    if (ts->tv_nsec >= 1000000000L) {
	ts->tv_sec++;
	ts->tv_nsec -= 1000000000L;
    }
}
