/* 
 * An implementation of the BSD system log device driver which
 * works in the the AlgRTX/MIPS SDE file i/o system.
 *
 * Copyright (c) 2001 Algorithmics Ltd
 *
 * This program is NOT free software, it is supplied under the terms
 * of the AlgNET License Agreement, a copy of which is available at:
 *
 *  http://www.algor.co.uk/algor/info/algnet-license.pdf
 *
 * Any company which has obtained and signed a valid AlgNET license
 * may use and modify this software internally and use (without
 * restrictions) any derived binary.  You may not, however,
 * redistribute this in whole or in part as source code, nor may you
 * modify or remove any part of this copyright message.  
 */

/*	$OpenBSD: subr_log.c,v 1.6 2000/02/22 19:28:03 deraadt Exp $	*/
/*	$NetBSD: subr_log.c,v 1.11 1996/03/30 22:24:44 christos Exp $	*/

/*
 * Copyright (c) 1982, 1986, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)subr_log.c	8.1 (Berkeley) 6/10/93
 */

/*
 * Error log buffer for kernel printf's.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/vnode.h>
#include <sys/ioctl.h>
#include <sys/msgbuf.h>
#include <sys/file.h>
#include <sys/select.h>
#ifndef __AlgRTX__
#include <sys/signalvar.h>
#endif
#include <sys/syslog.h>
#include <sys/conf.h>

#define LOG_RDPRI	(PZERO + 1)

#define LOG_ASYNC	0x04
#define LOG_RDWAIT	0x08

static struct logsoftc {
	int	sc_state;		/* see above for possibilities */
	struct	selinfo sc_selp;	/* process waiting on select call */
	int	sc_pgid;		/* process/group for async I/O */
#if __AlgRTX__
    	pthread_mutex_t sc_mx;		/* protecting mutex */
	pthread_cond_t sc_cd;		/* condition wait */
#else
	uid_t	sc_siguid;		/* uid for process that set sc_pgid */
	uid_t	sc_sigeuid;		/* euid for process that set sc_pgid */
#endif
} logsoftc;

int	log_open;			/* also used in log() */
int	msgbufmapped;			/* is the message buffer mapped */
int	msgbufenabled;			/* is logging to the buffer enabled */
struct	msgbuf *msgbufp;		/* the mapped buffer, itself. */

#ifndef offsetof
#define offsetof(type, member) ((size_t)(&((type *)0)->member))
#endif

#ifdef __AlgRTX__

/* /dev/klog:   cr--r--r-- */
#define LOGMODE	(S_IFCHR|S_IRUSR|S_IRGRP|S_IROTH)

static int  	logread (struct file *, struct uio *);
static int	logwrite (struct file *, struct uio *);
static int      logioctl (struct file *, int, void *);
static int      logselect (struct file *, int);
static int      logopen (struct file *, dev_t, int);
static int      logclose (struct file *);

/* POSIX file i/o ops for /dev/tty[pqr]x */
static const struct fileops _log_ops = {
    logread,
    logwrite,
    logioctl,
    logselect,
    logopen,
    logclose,
};
#endif

void
initmsgbuf(buf, bufsize)
	caddr_t buf;
	size_t bufsize;
{
	register struct msgbuf *mbp;
	long new_bufs;

	/* Sanity-check the given size. */
	if (bufsize < sizeof(struct msgbuf))
		return;

	mbp = msgbufp = (struct msgbuf *)buf;

	new_bufs = bufsize - offsetof(struct msgbuf, msg_bufc);
	if ((mbp->msg_magic != MSG_MAGIC) || (mbp->msg_bufs != new_bufs) ||
	    (mbp->msg_bufr < 0) || (mbp->msg_bufr >= mbp->msg_bufs) ||
	    (mbp->msg_bufx < 0) || (mbp->msg_bufx >= mbp->msg_bufs)) {
		/*
		 * If the buffer magic number is wrong, has changed
		 * size (which shouldn't happen often), or is
		 * internally inconsistent, initialize it.
		 */

		bzero(buf, bufsize);
		mbp->msg_magic = MSG_MAGIC;
		mbp->msg_bufs = new_bufs;
	}

	/* mark it as ready for use. */
	msgbufmapped = msgbufenabled = 1;

#ifdef __AlgRTX__
	{
	    pthread_mutexattr_t mxa;
	    pthread_condattr_t cda;
	    dir_t lognode;

	    pthread_mutexattr_init (&mxa);
	    pthread_mutexattr_setname_np (&mxa, "klogmx");
	    pthread_mutex_init (&logsoftc.sc_mx, &mxa);
	    pthread_mutexattr_destroy (&mxa);
	    
	    pthread_condattr_init (&cda);
	    pthread_condattr_setname_np (&cda, "klog");
	    pthread_condattr_setinterruptible_np (&cda, 1);
	    pthread_cond_init (&logsoftc.sc_cd, &cda);
	    pthread_condattr_destroy (&cda);

	    memset (&lognode, 0, sizeof (lognode));
	    lognode.d_mode = LOGMODE;
	    lognode.d_ops = &_log_ops;
	    lognode.d_name = "klog";
	    if (_mknod ("/dev", &lognode))
		panic ("initmsgbug/mknod");
	}
#endif
}


#ifdef __AlgRTX__
void
logwakeup()
{
    pthread_mutex_lock (&logsoftc.sc_mx);
    if (log_open) {
	selwakeup (&logsoftc.sc_selp);
	if (logsoftc.sc_state & LOG_ASYNC)
	    pgsignal (logsoftc.sc_pgid, SIGIO, 0);
	if (logsoftc.sc_state & LOG_RDWAIT)
	    pthread_cond_broadcast (&logsoftc.sc_cd);
    }
    pthread_mutex_unlock (&logsoftc.sc_mx);
}


void
logputchar (c)
    int c;
{
    struct msgbuf *mbp = msgbufp;
    pthread_mutex_lock (&logsoftc.sc_mx);
    mbp->msg_bufc[mbp->msg_bufx++] = c;
    mbp->msg_bufl = min(mbp->msg_bufl+1, mbp->msg_bufs);
    if (mbp->msg_bufx < 0 || mbp->msg_bufx >= mbp->msg_bufs)
	mbp->msg_bufx = 0;
    /* If the buffer is full, keep the most recent data. */
    if (mbp->msg_bufr == mbp->msg_bufx) {
	if (++mbp->msg_bufr >= mbp->msg_bufs)
	    mbp->msg_bufr = 0;
    }
    pthread_mutex_unlock (&logsoftc.sc_mx);
}


static int
logopen (struct file *fp, dev_t dev, int flag)
{
    int error = 0;
    pthread_mutex_lock (&logsoftc.sc_mx);
    if (log_open)
	error = EBUSY;
    else
	log_open = 1;
    pthread_mutex_unlock (&logsoftc.sc_mx);
    return (error);
}

static int
logclose(struct file *fp)
{
    log_open = 0;
    logsoftc.sc_state = 0;
    return 0;
}

static int
logread (struct file *fp, struct uio *uio)
{
    struct msgbuf *mbp = msgbufp;
    long l;
    int error = 0;
    
    pthread_mutex_lock (&logsoftc.sc_mx);
    while (mbp->msg_bufr == mbp->msg_bufx) {
	if (fp->f_flag & IO_NDELAY)
	    error = EWOULDBLOCK;
	else {
	    logsoftc.sc_state |= LOG_RDWAIT;
	    error = pthread_cond_wait (&logsoftc.sc_cd, &logsoftc.sc_mx);
	}
	if (error) {
	    pthread_mutex_unlock (&logsoftc.sc_mx);
	    return (error);
	}
    }
    logsoftc.sc_state &= ~LOG_RDWAIT;
    pthread_mutex_unlock (&logsoftc.sc_mx);

    while (uio->uio_resid > 0) {
	l = mbp->msg_bufx - mbp->msg_bufr;
	if (l < 0)
	    l = mbp->msg_bufs - mbp->msg_bufr;
	l = min(l, uio->uio_resid);
	if (l == 0)
	    break;
	error = uiomove((caddr_t)&mbp->msg_bufc[mbp->msg_bufr],
			(int)l, uio);
	if (error)
	    break;
	mbp->msg_bufr += l;
	if (mbp->msg_bufr < 0 || mbp->msg_bufr >= mbp->msg_bufs)
	    mbp->msg_bufr = 0;
    }
    return (error);
}


static int
logwrite (struct file *fp, struct uio *uio)
{
    return (EIO);
}


static int
logselect (struct file *fp, int rw)
{
    int ret = 0;

    if (rw == FREAD) {
	pthread_mutex_lock (&logsoftc.sc_mx);
	if (msgbufp->msg_bufr != msgbufp->msg_bufx)
	    ret = 1;
	pthread_mutex_unlock (&logsoftc.sc_mx);
	if (ret == 0)
	    selrecord(p, &logsoftc.sc_selp);
    }
    return (ret);

}


static int
logioctl (struct file *fp, int com, void *data)
{
    int error = 0;
    long l;
    
    pthread_mutex_lock (&logsoftc.sc_mx);

    switch (com) {

    /* return number of characters immediately available */
    case FIONREAD:
	l = msgbufp->msg_bufx - msgbufp->msg_bufr;
	if (l < 0)
	    l += msgbufp->msg_bufs;
	*(int *)data = l;
	break;

    case FIONBIO:
	break;

    case FIOASYNC:
	if (*(int *)data)
	    logsoftc.sc_state |= LOG_ASYNC;
	else
	    logsoftc.sc_state &= ~LOG_ASYNC;
	break;

    case TIOCSPGRP:
	logsoftc.sc_pgid = *(int *)data;
	break;

    case TIOCGPGRP:
	*(int *)data = logsoftc.sc_pgid;
	break;

    default:
	error = EINVAL;
	break;
    }

    pthread_mutex_unlock (&logsoftc.sc_mx);
    return (error);
}

#else
/*ARGSUSED*/
int
logopen(dev, flags, mode, p)
	dev_t dev;
	int flags, mode;
	struct proc *p;
{
	if (log_open)
		return (EBUSY);
	log_open = 1;
	return (0);
}

/*ARGSUSED*/
int
logclose(dev, flag, mode, p)
	dev_t dev;
	int flag, mode;
	struct proc *p;
{

	log_open = 0;
	logsoftc.sc_state = 0;
	return (0);
}

/*ARGSUSED*/
int
logread(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	register struct msgbuf *mbp = msgbufp;
	register long l;
	register int s;
	int error = 0;

	s = splhigh();
	while (mbp->msg_bufr == mbp->msg_bufx) {
		if (flag & IO_NDELAY) {
			splx(s);
			return (EWOULDBLOCK);
		}
		logsoftc.sc_state |= LOG_RDWAIT;
		error = tsleep((caddr_t)mbp, LOG_RDPRI | PCATCH,
			       "klog", 0);
		if (error) {
			splx(s);
			return (error);
		}
	}
	splx(s);
	logsoftc.sc_state &= ~LOG_RDWAIT;

	while (uio->uio_resid > 0) {
		l = mbp->msg_bufx - mbp->msg_bufr;
		if (l < 0)
			l = mbp->msg_bufs - mbp->msg_bufr;
		l = min(l, uio->uio_resid);
		if (l == 0)
			break;
		error = uiomove((caddr_t)&mbp->msg_bufc[mbp->msg_bufr],
			(int)l, uio);
		if (error)
			break;
		mbp->msg_bufr += l;
		if (mbp->msg_bufr < 0 || mbp->msg_bufr >= mbp->msg_bufs)
			mbp->msg_bufr = 0;
	}
	return (error);
}

/*ARGSUSED*/
int
logselect(dev, rw, p)
	dev_t dev;
	int rw;
	struct proc *p;
{
	int s = splhigh();

	switch (rw) {

	case FREAD:
		if (msgbufp->msg_bufr != msgbufp->msg_bufx) {
			splx(s);
			return (1);
		}
		selrecord(p, &logsoftc.sc_selp);
		break;
	}
	splx(s);
	return (0);
}

void
logwakeup()
{
	if (!log_open)
		return;
	selwakeup(&logsoftc.sc_selp);
	if (logsoftc.sc_state & LOG_ASYNC)
		csignal(logsoftc.sc_pgid, SIGIO,
		    logsoftc.sc_siguid, logsoftc.sc_sigeuid);
	if (logsoftc.sc_state & LOG_RDWAIT) {
		wakeup((caddr_t)msgbufp);
		logsoftc.sc_state &= ~LOG_RDWAIT;
	}
}

/*ARGSUSED*/
int
logioctl(dev, com, data, flag, p)
	dev_t dev;
	u_long com;
	caddr_t data;
	int flag;
	struct proc *p;
{
	long l;
	int s;

	switch (com) {

	/* return number of characters immediately available */
	case FIONREAD:
		s = splhigh();
		l = msgbufp->msg_bufx - msgbufp->msg_bufr;
		splx(s);
		if (l < 0)
			l += msgbufp->msg_bufs;
		*(int *)data = l;
		break;

	case FIONBIO:
		break;

	case FIOASYNC:
		if (*(int *)data)
			logsoftc.sc_state |= LOG_ASYNC;
		else
			logsoftc.sc_state &= ~LOG_ASYNC;
		break;

	case TIOCSPGRP:
		logsoftc.sc_pgid = *(int *)data;
		logsoftc.sc_siguid = p->p_cred->p_ruid;
		logsoftc.sc_sigeuid = p->p_ucred->cr_uid;
		break;

	case TIOCGPGRP:
		*(int *)data = logsoftc.sc_pgid;
		break;

	default:
		return (-1);
	}
	return (0);
}
#endif
