/* 
 * Copyright (c) 2001 Algorithmics Ltd
 *
 * A much modified version of the original BSD pseudo-tty module, to
 * work with the AlgRTX/MIPS SDE file i/o system.  
 *
 * This program is NOT free software, it is supplied under the terms
 * of the AlgNET License Agreement, a copy of which is available at:
 *
 *  http://www.algor.co.uk/algor/info/algnet-license.pdf
 *
 * Any company which has obtained and signed a valid AlgNET license
 * may use and modify this software internally and use (without
 * restrictions) any derived binary.  You may not, however,
 * redistribute this in whole or in part as source code, nor may you
 * modify or remove any part of this copyright message.  
 */

/*	$OpenBSD: tty_pty.c,v 1.7 1997/11/30 21:41:03 deraadt Exp $	*/
/*	$NetBSD: tty_pty.c,v 1.33.4.1 1996/06/02 09:08:11 mrg Exp $	*/

/*
 * Copyright (c) 1982, 1986, 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)tty_pty.c	8.4 (Berkeley) 2/20/95
 */

/*
 * Pseudo-teletype Driver
 * (Actually two drivers, requiring two entries in 'cdevsw')
 */
#include "pty.h"		/* XXX */
#if NPTY > 0

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/ioctl.h>
#include <sys/proc.h>
#include <sys/tty.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/kernel.h>
#include <sys/signalvar.h>
#include <sys/uio.h>
#include <sys/conf.h>

#if NPTY == 1
#undef NPTY
#define	NPTY	4		/* crude XXX */
#endif

#if NPTY > 48
#error Max 48 NPTYs supported 
#endif

#define BUFSIZ 100		/* Chunk size iomoved to/from user */

/*
 * pts == /dev/tty[pqr]?
 * ptc == /dev/pty[pqr]?
 */
static struct pt_softc {
    struct tty  pt_tty;
    int		pt_flags;
    struct	selinfo pt_selr, pt_selw;
    u_char	pt_send;
    u_char	pt_ucntl;
    char	pt_cname[6];
    char	pt_sname[6];
    int		pt_cwait;
    pthread_cond_t pt_ccond;
    pthread_cond_t pt_scond;
} pt_softc[NPTY];

#define	PF_PKT		0x08		/* packet mode */
#define	PF_STOPPED	0x10		/* user told stopped */
#define	PF_REMOTE	0x20		/* remote and flow controlled input */
#define	PF_NOSTOP	0x40
#define PF_UCNTL	0x80		/* user control mode */

#define minor(dev)	(dev)

void		ptyattach (int);

static void	ptcwakeup (struct tty *, int);
static void	ptsstart (struct tty *);
static void	ptstxstop (struct tty *, int);
static int	ptswait (struct tty *, int, int);
static void	ptswake (struct tty *, int);

static int  	ptsread (struct file *, struct uio *);
static int	ptswrite (struct file *, struct uio *);
static int      ptsioctl (struct file *, int, void *);
static int      ptsselect (struct file *, int);
static int      ptsopen (struct file *, dev_t, int);
static int      ptsclose (struct file *);

/* POSIX file i/o ops for /dev/tty[pqr]x */
static const struct fileops _pts_ops = {
    ptsread,
    ptswrite,
    ptsioctl,
    ptsselect,
    ptsopen,
    ptsclose,
};

static int  	ptcread (struct file *, struct uio *);
static int	ptcwrite (struct file *, struct uio *);
static int      ptcioctl (struct file *, int, void *);
static int      ptcselect (struct file *, int);
static int      ptcopen (struct file *, dev_t, int);
static int      ptcclose (struct file *);

/* POSIX file i/o ops for /dev/pty[pqr]x */
static const struct fileops _ptc_ops = {
    ptcread,
    ptcwrite,
    ptcioctl,
    ptcselect,
    ptcopen,
    ptcclose,
};

/* /dev/tty*:   crw-rw-r-- */
#define TTYMODE	(S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IWOTH)

/*
 * Establish ptys in the system.
 */
void
ptyattach(int n)
{
    static char ptsname[] = {"ttyXX"};
    static char ptcname[] = {"ptyXX"};
    pthread_condattr_t cda;
    pthread_mutexattr_t mxa;
    dir_t ptsnode;
    dir_t ptcnode;
    struct pt_softc *pti;
    struct tty *tp;
    int i;

    memset (&ptsnode, 0, sizeof (ptsnode));
    ptsnode.d_mode = TTYMODE;
    ptsnode.d_ops = &_pts_ops;
    ptsnode.d_name = ptsname;

    memset (&ptcnode, 0, sizeof (ptcnode));
    ptcnode.d_mode = TTYMODE;
    ptcnode.d_ops = &_ptc_ops;
    ptcnode.d_name = ptcname;

    pthread_mutexattr_init (&mxa);
    pthread_condattr_init (&cda);
    pthread_condattr_setinterruptible_np (&cda, 1);

    for (pti = pt_softc, i = 0; i < NPTY; i++, pti++) {
	ptsname[3] = ptcname[3] = "pqr"[i / 16];
	ptsname[4] = ptcname[4] = "0123456789abcdef"[i % 16];
	ptsnode.d_dev = ptcnode.d_dev = i;
	if (_mknod ("/dev", &ptsnode))
	    panic ("mknod(/dev/%s) failed\n", ptsname);
	if (_mknod ("/dev", &ptcnode))
	    panic ("mknod(/dev/%s) failed\n", ptcname);

	strcpy (pti->pt_cname, ptcname);
	strcpy (pti->pt_sname, ptsname);

	/* setup hooks for interrupt driven tty driver */
	tp = &pti->pt_tty;
	tp->t_state     = TS_INTRUPT;
	tp->t_dev	= i;
	tp->t_start	= ptsstart;
	tp->t_txstop	= ptstxstop;
	tp->t_wait	= ptswait;
	tp->t_wakeup	= ptswake;
	tp->t_addr	= pti;		/* back pointer */

	/* initialise mutex */
	__sdethread_mutex_init (&tp->t_mx, pti->pt_cname);

	/* initialise controller cond variable */
	pthread_condattr_setname_np (&cda, pti->pt_cname);
	pthread_cond_init (&pti->pt_ccond, &cda);

	/* initialise slace cond variable */
	pthread_condattr_setname_np (&cda, pti->pt_sname);
	pthread_cond_init (&pti->pt_scond, &cda);
    }

    pthread_mutexattr_destroy (&mxa);
    pthread_condattr_destroy (&cda);
}



/* 
 * AlgRTX tty intrrupt driver synchronisation glue
 */

static void
ptycleanup (void *arg)
{
    struct tty *tp = arg;
    __sdethread_mutex_unlock (&tp->t_mx);
}


static int
ptswait (struct tty *tp, int waitflags, int tenths)
{
    struct pt_softc *pti = tp->t_addr;
    struct timespec ts, *tsp = NULL;
    int res;

    tp->t_state |= waitflags;
    if (tenths >= 0) {
	clock_gettime (CLOCK_REALTIME, &ts);
	ts.tv_sec += tenths / 10;
	ts.tv_nsec += (tenths % 10) * 100*1000*1000;
	if (ts.tv_nsec >= 1000000000)
	    ts.tv_sec++, ts.tv_nsec -= 1000000000;
	tsp = &ts;
    }

    pthread_cleanup_push (ptycleanup, tp);
    res = pthread_cond_timedwait (&pti->pt_scond, (pthread_mutex_t *)&tp->t_mx, tsp);
    pthread_cleanup_pop (0);

    return res;
}

static void
ptswake (struct tty *tp, int flags)
{
    int wakeup = 0;
    if (flags & FWRITE) {
	if (tp->t_state & TS_TXWAIT && tp->t_outq.c_cc <= tp->t_lowat) {
	    tp->t_state &= ~TS_TXWAIT;
	    wakeup = 1;
	}
	if (tp->t_state & TS_TXDRAIN && tp->t_outq.c_cc == 0) {
	    tp->t_state &= ~TS_TXDRAIN;
	    wakeup = 1;
	}
	if (tp->t_state & TS_WOPEN && tp->t_state & TS_CARR_ON) {
	    tp->t_state &= ~TS_WOPEN;
	    wakeup = 1;
	}
    }
    if (flags & FREAD) {
	if (tp->t_state & TS_RXWAIT) {
	    tp->t_state &= ~TS_RXWAIT;
	    wakeup = 1;
	}
    }
    if (wakeup) {
	/* XXX could use different cond variables for read and write */
	struct pt_softc *pti = tp->t_addr;
	pthread_cond_broadcast (&pti->pt_scond);
    }
}


static int
ptcsleep (struct tty *tp, int flag)
{
    struct pt_softc *pti = tp->t_addr;
    int res;

    pti->pt_cwait |= flag;
    while (pti->pt_cwait & flag) {
	pthread_cleanup_push (ptycleanup, tp);
	res = pthread_cond_wait (&pti->pt_ccond, (pthread_mutex_t *)&tp->t_mx);
	pthread_cleanup_pop (0);
    }
    return res;
}


static void
ptcwakeup (struct tty *tp, int flag)
{
    struct pt_softc *pti = tp->t_addr;

    if (flag & FREAD)
	_selwakeup(&pti->pt_selr);
    if (flag & FWRITE)
	_selwakeup(&pti->pt_selw);
    if (pti->pt_cwait & flag) {
	pti->pt_cwait &= ~flag;
	pthread_cond_broadcast (&pti->pt_ccond);
    }
}



/*
 * Slave (/dev/tty[pqr]x) device 
 */

static int
ptsopen (struct file *fp, dev_t dev, int flag)
{
    struct pt_softc *pti;
    register struct tty *tp;
    int error = 0;
    
    if (minor(dev) >= NPTY)
	return (ENXIO);

    pti = &pt_softc[minor(dev)];
    tp = &pti->pt_tty;
    
    _ttylock (tp);

    /* initialise on first open */
    if ((tp->t_state & (TS_ISOPEN | TS_WOPEN)) == 0) {
	tp->t_iflag = TTYDEF_IFLAG;
	tp->t_oflag = TTYDEF_OFLAG;
	tp->t_lflag = TTYDEF_LFLAG;
	tp->t_cflag = TTYDEF_CFLAG & ~CLOCAL;
	tp->t_ispeed = tp->t_ospeed = TTYDEF_SPEED;
	if (error = _ttyclalloc (tp))
	    return error;
	_ttsetwater(tp);
	_ttychars(tp);
    }

    /* is controller already open */
    if (tp->t_rxpoll)
	tp->t_state |= TS_CARR_ON;
    else
	tp->t_state &= ~TS_CARR_ON;

    /* wait for controller */
    while ((tp->t_state & TS_CARR_ON) == 0) {
	if (flag & FNONBLOCK)
	    break;
	error = _ttysleep(tp, TS_WOPEN, -1);
	if (error)
	    goto done;
    }

    /* success */
    tp->t_state |= TS_ISOPEN;
    tp->t_count++;
    ptcwakeup(tp, FREAD|FWRITE);
    fp->f_data = tp;

 done:
    return _ttyunlock (tp, error);
}


static int
ptsclose(struct file *fp)
{
    struct tty *tp = fp->f_data;
    int error;

    if (!tp)
	return 0;

    /* line close */
    error = _ttylclose (tp, fp->f_flag);

    /* device close */
    error |= _ttyclose (tp);

    /* wakeup controller */
    _ttylock (tp);
    ptcwakeup(tp, FREAD|FWRITE);
    return _ttyunlock (tp, error);
}


static int
ptsread (struct file *fp, struct uio *uio)
{
    struct tty *tp = fp->f_data;
    struct pt_softc *pti = tp->t_addr;
    int error = 0;

    _ttylock (tp);
again:
    if (pti->pt_flags & PF_REMOTE) {
	if (tp->t_canq.c_cc == 0) {
	    if (fp->f_flag & FNONBLOCK) {
		error = EWOULDBLOCK;
		goto done;
	    }
	    error = _ttysleep(tp, TS_RXWAIT, -1);
	    if (error)
		goto done;
	    goto again;
	}
	while (tp->t_canq.c_cc > 1 && uio->uio_resid > 0)
	    if (ureadc(_ttygetc(&tp->t_canq), uio) < 0) {
		error = EFAULT;
		break;
	    }
	if (tp->t_canq.c_cc == 1)
	    (void) _ttygetc(&tp->t_canq);
	if (tp->t_canq.c_cc)
	    goto done;
    } else 
	if (tp->t_rxpoll) {	/* XXX not thread-safe */
	    _ttyunlock (tp, 0);
	    error = _ttyread (tp, uio, fp->f_flag);
	    _ttylock (tp);
	}
    ptcwakeup(tp, FWRITE);
 done:
    return _ttyunlock (tp, error);
}


/*
 * Write to pseudo-tty.
 * Wakeups of controlling tty will happen
 * indirectly, when tty driver calls ptsstart.
 */
static int
ptswrite (struct file *fp, struct uio *uio)
{
    struct tty *tp = fp->f_data;
    
    if (!tp->t_rxpoll)	/* XXX not thread-safe */
	return (EIO);	/* controller gone */
    return _ttywrite (tp, uio, fp->f_flag);
}


static int
ptsselect (struct file *fp, int rw)
{
    return _ttyselect ((struct tty *) fp->f_data, rw);
}




/*
 * Start output on pseudo-tty.
 * Wake up process selecting or sleeping for input from controlling tty.
 */
static void
ptsstart(struct tty *tp)
{
    struct pt_softc *pti = tp->t_addr;

    if (tp->t_state & TS_TTSTOP)
	return;
    if (pti->pt_flags & PF_STOPPED) {
	pti->pt_flags &= ~PF_STOPPED;
	pti->pt_send = TIOCPKT_START;
    }
    ptcwakeup(tp, FREAD);
}


static void
ptstxstop (struct tty *tp, int flush)
{
    struct pt_softc *pti = tp->t_addr;
    int flag;

    /* note: FLUSHREAD and FLUSHWRITE already ok */
    if (flush == 0) {
	flush = TIOCPKT_STOP;
	pti->pt_flags |= PF_STOPPED;
    } else
	pti->pt_flags &= ~PF_STOPPED;
    pti->pt_send |= flush;
    /* change of perspective */
    flag = 0;
    if (flush & FREAD)
	flag |= FWRITE;
    if (flush & FWRITE)
	flag |= FREAD;
    ptcwakeup(tp, flag);
}



/*
 * Controller / master (/dev/pty[pqr]x) device 
 */

static int
ptcopen (struct file *fp, dev_t dev, int flag)
{
    struct pt_softc *pti;
    struct tty *tp;
    
    if (minor(dev) >= NPTY)
	return (ENXIO);
    pti = &pt_softc[minor(dev)];
    tp = &pti->pt_tty;

    _ttylock (tp);

    if (tp->t_rxpoll)
	return _ttyunlock (tp, EIO);		/* already open */

    tp->t_rxpoll = (void *)1;
    tp->t_lflag &= ~EXTPROC;
    fp->f_data = tp;

    _ttymodem (tp, 1);				/* carrier on! */

    pti->pt_flags = 0;
    pti->pt_send = 0;
    pti->pt_ucntl = 0;

    return _ttyunlock (tp, 0);
}


static int
ptcclose (struct file *fp)
{
    struct tty *tp = fp->f_data;

    if (!tp)
	return 0;

    _ttylock (tp);
    _ttymodem (tp, 0);		/* carrier off! */
    tp->t_state &= ~TS_CARR_ON;
    tp->t_rxpoll = 0;		/* mark closed */
    return _ttyunlock (tp, 0);
}


static int
ptcread (struct file *fp, struct uio *uio)
{
    struct tty *tp = fp->f_data;
    struct pt_softc *pti = tp->t_addr;
    char *bp;
    int error = 0, cc;
    
    /*
     * We want to block until the slave
     * is open, and there's something to read;
     * but if we lost the slave or we're NBIO,
     * then return the appropriate error instead.
     */
    
    _ttylock (tp);
    for (;;) {
	if (tp->t_state&TS_ISOPEN) {
	    if (pti->pt_flags&PF_PKT && pti->pt_send) {
		error = ureadc((int)pti->pt_send, uio);
		if (error)
		    goto done;
		if (pti->pt_send & TIOCPKT_IOCTL) {
		    cc = min(uio->uio_resid,
			     sizeof(tp->t_termios));
		    uiomove((caddr_t) &tp->t_termios,
			    cc, uio);
		}
		pti->pt_send = 0;
		goto done;
	    }
	    if (pti->pt_flags&PF_UCNTL && pti->pt_ucntl) {
		error = ureadc((int)pti->pt_ucntl, uio);
		if (error == 0)
		    pti->pt_ucntl = 0;
		goto done;
	    }
	    if (tp->t_outq.c_cc && (tp->t_state&TS_TTSTOP) == 0)
		break;	/* ready to be read */
	}
	if ((tp->t_state&TS_CARR_ON) == 0)
	    goto done;	/* EOF */
	if (fp->f_flag & FNONBLOCK) {
	    error = EWOULDBLOCK;
	    goto done;
	}
	error = ptcsleep (tp, FREAD);
	if (error)
	    goto done;
    }

    if (pti->pt_flags & (PF_PKT|PF_UCNTL))
	error = ureadc(0, uio);

    while (uio->uio_resid > 0 && error == 0) {
	cc = _ttygetcblk (&tp->t_outq, &bp);
	if (cc <= 0)
	    break;
	if (cc > uio->uio_resid)
	    cc = uio->uio_resid;
	error = uiomove(bp, cc, uio);
	_ttydelcblk (&tp->t_outq, cc);
    }

    if (tp->t_outq.c_cc <= tp->t_lowat)
	_ttwakeup (tp, FWRITE);

 done:
    return _ttyunlock (tp, error);
}


static int
ptcwrite (struct file *fp, struct uio *uio)
{
    struct tty *tp = fp->f_data;
    struct pt_softc *pti = tp->t_addr;
    u_char *cp = NULL;
    int cc = 0;
    u_char locbuf[BUFSIZ];
    int cnt = 0;
    int error = 0;

    _ttylock (tp);
 again:
    if ((tp->t_state&TS_ISOPEN) == 0)
	goto block;
    
    if (pti->pt_flags & PF_REMOTE) {
	if (tp->t_canq.c_cc)
	    goto block;
	while (uio->uio_resid > 0
	       && tp->t_canq.c_cc < tp->t_hog - 1) {
	    int c;
	    error = uwritec (&c, uio);
	    if (error)
		goto done;
	    (void) _ttyputc (c, &tp->t_canq);
	}
	(void) _ttyputc (0, &tp->t_canq);
	_ttwakeup(tp, FREAD);
	goto done;
    }
    
    while (uio->uio_resid > 0) {
	if (cc == 0) {
	    cc = min(uio->uio_resid, BUFSIZ);
	    cp = locbuf;
	    error = uiomove((caddr_t)cp, cc, uio);
	    if (error)
		goto done;
	}
	while (cc > 0) {
	    if ((tp->t_rawq.c_cc + tp->t_canq.c_cc) >= tp->t_hog - 2 &&
		(tp->t_canq.c_cc > 0 || !(tp->t_iflag&ICANON))) {
		/* not enough room, wakeup reader and wait */
		_ttwakeup (tp, FREAD);
		goto block;
	    }
	    _ttyinput (*cp++, tp);
	    cnt++;
	    cc--;
	}
	cc = 0;
    }
    goto done;
    
 block:
    /*
     * Come here to wait for slave to open, for space
     * in outq, or space in rawq.
     */
    if ((tp->t_state&TS_CARR_ON) == 0) {
	error = EIO;
	goto done;
    }
    if (fp->f_flag & FNONBLOCK) {
	/* adjust for data copied in but not written */
	uio->uio_resid += cc;
	if (cnt == 0)
	    error = EWOULDBLOCK;
	goto done;
    }
    error = ptcsleep (tp, FWRITE);
    if (error) {
	/* adjust for data copied in but not written */
	uio->uio_resid += cc;
	goto done;
    }
    goto again;

 done:
    return _ttyunlock (tp, error);
}


static int
ptcselect (struct file *fp, int rw)
{
    struct tty *tp = fp->f_data;
    struct pt_softc *pti = tp->t_addr;
    int result = 0;
    
    _ttylock (tp);
    
    if ((tp->t_state&TS_CARR_ON) == 0)
	return _ttyunlock (tp, 1);
    
    switch (rw) {
	
    case FREAD:
	if ((tp->t_state&TS_ISOPEN) &&
	    tp->t_outq.c_cc && (tp->t_state&TS_TTSTOP) == 0) {
	    result = 1;
	    break;
	}
	/* FALLTHROUGH */
	
    case 0:					/* exceptional */
	if ((tp->t_state&TS_ISOPEN) &&
	    (((pti->pt_flags & PF_PKT) && pti->pt_send) ||
	     ((pti->pt_flags & PF_UCNTL) && pti->pt_ucntl))) {
	    result = 1;
	    break;
	}
	_selrecord(&pti->pt_selr);
	break;
	
	
    case FWRITE:
	if (tp->t_state&TS_ISOPEN) {
	    if (pti->pt_flags & PF_REMOTE) {
		if (tp->t_canq.c_cc == 0) {
		    result = 1;
		    break;
		}
	    } else {
		if (tp->t_rawq.c_cc + tp->t_canq.c_cc < tp->t_hog-2
		    || (tp->t_canq.c_cc == 0
			&& (tp->t_iflag & ICANON)))
		    result = 1;
		break;
	    }
	}
	_selrecord(&pti->pt_selw);
	break;
    }

    return _ttyunlock (tp, result);
}


static int
ptyioctl (struct file *fp, int cmd, void *data, int ctl)
{
    struct tty *tp = fp->f_data;
    struct pt_softc *pti = tp->t_addr;
    u_char *cc = tp->t_cc;
    int stop, error = 0;

    /*
     * IF CONTROLLER STTY THEN MUST FLUSH TO PREVENT A HANG.
     * ttywflush(tp) will hang if there are characters in the outq.
     */
    if (cmd == TIOCEXT) {
	/*
	 * When the EXTPROC bit is being toggled, we need
	 * to send an TIOCPKT_IOCTL if the packet driver
	 * is turned on.
	 */
	_ttylock (tp);
	if (*(int *)data) {
	    if (pti->pt_flags & PF_PKT) {
		pti->pt_send |= TIOCPKT_IOCTL;
		ptcwakeup(tp, FREAD);
	    }
	    tp->t_lflag |= EXTPROC;
	} else {
	    if ((tp->t_lflag & EXTPROC) &&
		(pti->pt_flags & PF_PKT)) {
		pti->pt_send |= TIOCPKT_IOCTL;
		ptcwakeup(tp, FREAD);
	    }
	    tp->t_lflag &= ~EXTPROC;
	}
	return _ttyunlock (tp, 0);
    }

    if (ctl) {
	_ttylock (tp);
	switch (cmd) {

	case TIOCGPGRP:
	    /*
	     * We aviod calling ttioctl on the controller since,
	     * in that case, tp must be the controlling terminal.
	     */
	    *(int *)data = tp->t_pgrp;
	    goto done;

	case TIOCPKT:
	    if (*(int *)data) {
		if (pti->pt_flags & PF_UCNTL)
		    return (EINVAL);
		pti->pt_flags |= PF_PKT;
	    } else
		pti->pt_flags &= ~PF_PKT;
	    goto done;

	case TIOCUCNTL:
	    if (*(int *)data) {
		if (pti->pt_flags & PF_PKT)
		    return (EINVAL);
		pti->pt_flags |= PF_UCNTL;
	    } else
		pti->pt_flags &= ~PF_UCNTL;
	    goto done;

	case TIOCREMOTE:
	    if (*(int *)data)
		pti->pt_flags |= PF_REMOTE;
	    else
		pti->pt_flags &= ~PF_REMOTE;
	    _ttyflush(tp, FREAD|FWRITE);
	    goto done;

	case TIOCSETD:
	case TIOCSETA:
	case TIOCSETAW:
	case TIOCSETAF:
	    _ttyflushq (&tp->t_outq);
	    break;

	case TIOCSIG:
	    if (*(unsigned int *)data >= NSIG) {
		error = EINVAL;
	    }
	    else {
		if ((tp->t_lflag&NOFLSH) == 0)
		    _ttyflush(tp, FREAD|FWRITE);
		pgsignal(tp->t_pgrp, *(unsigned int *)data, 1);
	    }
	    goto done;
	}

	(void) _ttyunlock (tp, 0);
    }

#if 0
    /* linesw ioctl XXX do this _ttyioctl? */
    error = nullioctl (tp, cmd, data);
    if (error == ENOTTY)
	error = _ttyioctl(tp, cmd, data);
#else
    error = _ttyioctl(tp, cmd, data);
#endif

    _ttylock (tp);

    if (error == ENOTTY) {
	if (pti->pt_flags & PF_UCNTL &&
	    (cmd & ~0xff) == UIOCCMD(0)) {
	    if (cmd & 0xff) {
		pti->pt_ucntl = (u_char)cmd;
		ptcwakeup(tp, FREAD);
	    }
	    error = 0;
	    goto done;
	}
    }

    /*
     * If external processing and packet mode send ioctl packet.
     */
    if ((tp->t_lflag&EXTPROC) && (pti->pt_flags & PF_PKT)) {
	switch(cmd) {
	case TIOCSETA:
	case TIOCSETAW:
	case TIOCSETAF:
	    pti->pt_send |= TIOCPKT_IOCTL;
	    ptcwakeup(tp, FREAD);
	default:
	    break;
	}
    }
    stop = (tp->t_iflag & IXON) && CCEQ(cc[VSTOP], CTRL('s')) 
	&& CCEQ(cc[VSTART], CTRL('q'));
    if (pti->pt_flags & PF_NOSTOP) {
	if (stop) {
	    pti->pt_send &= ~TIOCPKT_NOSTOP;
	    pti->pt_send |= TIOCPKT_DOSTOP;
	    pti->pt_flags &= ~PF_NOSTOP;
	    ptcwakeup(tp, FREAD);
	}
    } else {
	if (!stop) {
	    pti->pt_send &= ~TIOCPKT_DOSTOP;
	    pti->pt_send |= TIOCPKT_NOSTOP;
	    pti->pt_flags |= PF_NOSTOP;
	    ptcwakeup(tp, FREAD);
	}
    }
 done:
    return _ttyunlock (tp, error);
}


static int
ptcioctl (struct file *fp, int cmd, void *data)
{
    return ptyioctl (fp, cmd, data, 1);
}


static int
ptsioctl (struct file *fp, int cmd, void *data)
{
    return ptyioctl (fp, cmd, data, 0);
}
#endif /* NPTY > 0 */
