/* 
 * kick.c: Network interface configuration (kickstart)
 * 
 * Copyright (c) 2001 Algorithmics Ltd - all rights reserved.
 *
 * Attempts to use DHCP, BOOTP. RARP or finally saved environment
 * (DHCP & RARP not implemented yet)
 *
 * This program is NOT free software, it is supplied under the terms
 * of the AlgRTX License Agreement, a copy of which is available at:
 *
 *  http://www.algor.co.uk/algor/info/algrtx-license.pdf
 *
 * Any company which has obtained and signed a valid AlgRTX license
 * may use and modify this software internally and use (without
 * restrictions) any derived binary.  You may not, however,
 * redistribute this in whole or in part as source code, nor may you
 * modify or remove any part of this copyright message.  
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/queue.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/route.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ifaddrs.h>
#include <netdb.h>
#include <stdlib.h>
#include <pthread.h>
#include <stdio.h>
#include <alloca.h>
#include <setjmp.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <err.h>

/*#define DEBUG*/

#include "bootp.h"
#include "bootparams.h"

/* list of per-interface settings */
static SLIST_HEAD(, bootparams) bootphead = SLIST_HEAD_INITIALIZER(bootphead);

static __inline__ void
setsin (struct sockaddr_in *sin, int family, u_long addr)
{
    bzero (sin, sizeof (*sin));
    sin->sin_len = sizeof (*sin);
    sin->sin_family = family;
    sin->sin_addr.s_addr = addr;
}


static int
ia_getenv (const char *name, struct in_addr *ia)
{
    char *s;
    if (name && (s = getenv (name)))
	return inet_aton (s, ia);
    return 0;
}

static void
ia_setenv (struct ifaddrs *ifa, const char *name, const struct in_addr ia)
{
    char fullname[32];
    snprintf (fullname, sizeof fullname, "%s_%s", name, ifa->ifa_name);
    setenv (fullname, inet_ntoa (ia), 3);
}

static void
ia_delenv (struct ifaddrs *ifa, const char *name)
{
    char fullname[32];
    snprintf (fullname, sizeof fullname, "%s_%s", name, ifa->ifa_name);
    deleteenv (fullname);
    deleteenv (name);
}

/*
 * Add an IP address to an address list structure, ignoring
 * duplicates. 
 */
static int
bootp_addaddr (struct bootpaddrs *ba, void *addr)
{
    struct in_addr a;
    int i;

    memcpy (&a, addr, sizeof (struct in_addr));
    /* ignore duplicates */
    for (i = 0; i < ba->naddr; i++)
	if (ba->addrs[i].s_addr == a.s_addr)
	    return 0;
    if (ba->naddr < BOOT_MAXADDR) {
	ba->addrs[ba->naddr++] = a;
	return 1;
    }
    return 0;
}


static int
bootp_addaddrstr (struct bootpaddrs *ba, const char *src)
{
    struct in_addr a;
    char *str = strdup (src);
    char *sp, *cp;
    int any = 0;

    sp = str + strspn (str, " \t");
    while ((cp = strsep (&sp, " \t")) && *cp && inet_aton (cp, &a))
	any |= bootp_addaddr (ba, &a);
    free (str);
    return any;
}


static char *
bootp_getaddrstr (struct bootpaddrs *ba)
{
    char *str = malloc (ba->naddr * sizeof "xxx.xxx.xxx.xxx ");
    char *s = str;
    int i;

    for (i = 0; i < ba->naddr; i++) {
	if (i)
	    *s++ = ' ';
	strcpy (s, inet_ntoa (ba->addrs[i]));
	s += strlen (s);
    }
    return str;
}



/*
 * Fetch interface parameters from non-volatile enviroment 
 */
static int
check_env (struct ifaddrs *ifa, struct bootparams *bootp)
{
    struct in_addr a;
    char buf[32];
    char *val;

    /* Per-interface parameters */

    snprintf (buf, sizeof(buf), "netaddr_%s", ifa->ifa_name);
    if (ia_getenv (buf, &bootp->addr) 
	|| (bootp->have & BOOT_ONLYIF
	    && ia_getenv ("netaddr", &bootp->addr)))
	bootp->have |= BOOT_ADDR;

    snprintf (buf, sizeof(buf), "netmask_%s", ifa->ifa_name);
    if (ia_getenv (buf, &bootp->mask) 
	|| (bootp->have & BOOT_ONLYIF
	    && ia_getenv ("netmask", &bootp->mask)))
	bootp->have |= BOOT_MASK;

    snprintf (buf, sizeof(buf), "broadcast_%s", ifa->ifa_name);
    if (ia_getenv (buf, &bootp->broadaddr) 
	|| (bootp->have & BOOT_ONLYIF
	    && ia_getenv ("broadcast", &bootp->broadaddr)))
	bootp->have |= BOOT_BROADADDR;

    /* Host-wide parameters */

    /* XXX The lower-case/upper-case environment variable names are
       confusing, Unix "convention" is to use upper-case, but
       PMON uses lower-case. We try to support both. */

    if ((val = getenv ("HOSTNAME")) || (val = getenv ("hostname"))) {
	bootp->have |= BOOT_HOSTNAME;
	if (bootp->hostname)
	    free (bootp->hostname);
	bootp->hostname = strdup (val);
    }

    if ((val = getenv ("LOCALDOMAIN")) || (val = getenv ("localdomain"))) {
	bootp->have |= BOOT_DOMAINNAME;
	if (bootp->domain)
	    free (bootp->domain);
	bootp->domain = strdup (val);
    }

    if ((val = getenv ("GATEWAY")) || (val = getenv ("gateway"))) {
	if (bootp_addaddrstr (&bootp->gateip, val))
	    bootp->have |= BOOT_GATEIP;
    }

    if ((val = getenv ("NAMESERVER")) || (val = getenv ("nameserver"))) {
	if (bootp_addaddrstr (&bootp->dnsip, val))
	    bootp->have |= BOOT_DNSIP;
    }

    if (val = getenv ("LOGSERVER")) {
	if (bootp_addaddrstr (&bootp->logip, val))
	    bootp->have |= BOOT_LOGIP;
    }

    return 1;
}



#define MINTMO	4	/* 4 seconds initial timout */
#define MAXTMO	64	/* 64 seconds max timeout */
#define MAX_MSG (3 * 512)

/* Returns true if n_long's on the same net */
#define	SAMENET(a1, a2, m) ((a1.s_addr & m.s_addr) == (a2.s_addr & m.s_addr))

static const char vm_zero[4];
static const char vm_rfc1048[4] = VM_RFC1048;
static const char vm_cmu[4] = VM_CMU;

static const char * const vend[] = {
	vm_zero,			/* try no explicit vendor type first */
	vm_rfc1048,			/* try variable format */
	vm_cmu				/* very unlikely CMU format */
};

/* Local forwards */
static  void boot_bootp (struct bootparams *);
static	int bootprecv (struct bootparams *, struct bootp *, int);
static	void vend_cmu (struct bootparams *, u_char *, u_int);
static	void vend_rfc1048 (struct bootparams *, u_char *, u_int);

/* XXX not thread safe */
static	sigjmp_buf jmpb;

static void
finish (int signo)
{
    siglongjmp (jmpb, 1);
}


static int
check_bootp (int s, struct ifaddrs *ifa, struct bootparams *bootp)
{
    struct ifaliasreq addreq;
    struct bootp wbp;
    u_char rbuf[MAX_MSG];
    struct sockaddr_in clnt, srvr;
    time_t tmo;
    struct timeval start, tlast, twhen, tleft;
    fd_set rfds;
    void (*osigint)(int);
    volatile int notified;
    int recvcnt = 0;
    int nvend = 0;
    int cc, n;

    bzero(&wbp, sizeof(wbp));
    wbp.bp_op = BOOTREQUEST;

    /* get the ethernet address */
    if (ifa->ifa_addr && ifa->ifa_addr->sa_family == AF_LINK) {
	struct sockaddr_dl *sdl = (struct sockaddr_dl *) ifa->ifa_addr;
	if (sdl->sdl_type == IFT_ETHER 
	    && sdl->sdl_alen <= sizeof (wbp.bp_chaddr)) {
	    wbp.bp_htype = HTYPE_ETHERNET;
	    wbp.bp_hlen = sdl->sdl_alen;
	    memcpy (wbp.bp_chaddr, LLADDR(sdl), sdl->sdl_alen);
	}
	else {
	    warnx ("bad link-level address for `%s'", ifa->ifa_name);
	    return 0;
	}
    }
    else {
	warnx ("no link-level address for `%s'", ifa->ifa_name);
	return 0;
    }

    osigint = signal(SIGINT, finish);
    if (sigsetjmp (jmpb, 0)) {
	if (notified)
	    fprintf (stderr, "\ninterrupted\n");
	goto quit;
    }

    /* set interface address to 0.0.0.0 broadaddr 255.255.255.255 */
    bzero (&addreq, sizeof(addreq));
    strncpy(addreq.ifra_name, ifa->ifa_name, IFNAMSIZ);
    setsin ((struct sockaddr_in *)&addreq.ifra_addr, 
	    AF_INET, htonl (INADDR_ANY));
    setsin ((struct sockaddr_in *)&addreq.ifra_broadaddr, AF_INET, 
	    htonl ((INADDR_ANY & IN_CLASSA_NET)
		   | (INADDR_BROADCAST & IN_CLASSA_HOST)));
    if (ioctl(s, SIOCAIFADDR, &addreq) < 0) {
	warn ("ioctl (%s, SIOCAIFADDR, dummy)", ifa->ifa_name);
	goto quit;
    }

    n = 1;
    if (setsockopt (s, SOL_SOCKET, SO_BROADCAST, &n, sizeof (n)) < 0) {
	warn ("setsockopt(BROADCAST)");
	goto quit;
    }

    if (setsockopt (s, SOL_SOCKET, SO_DONTROUTE, &n, sizeof (n)) < 0) {
	warn ("setsockopt(DONTROUTE)");
	goto quit;
    }

    setsin (&clnt, AF_INET, htonl (INADDR_ANY));
    clnt.sin_port = htons (IPPORT_BOOTPC);
    if (bind (s, (struct sockaddr *)&clnt, sizeof(clnt)) < 0) {
	warn ("bind");
	goto quit;
    }

    setsin (&srvr, AF_INET, 
	    htonl ((INADDR_ANY & IN_CLASSA_NET)
		   | (INADDR_BROADCAST & IN_CLASSA_HOST)));
    srvr.sin_port = htons (IPPORT_BOOTPS);

    notified = 0;
    
    /* get start time and compute "random" request id */
    gettimeofday (&start, NULL);
    for (bootp->xid = start.tv_sec, n = wbp.bp_hlen - 1; n >= 0; n--)
	bootp->xid = (bootp->xid << 5) + wbp.bp_chaddr[n];

    do {
	timerclear (&tleft);
	tmo = MINTMO;
	recvcnt = 0;

	while (tmo <= MAXTMO) {

	    if (tleft.tv_sec < 0
		|| (tleft.tv_sec == 0 && tleft.tv_usec <= 0)) {

		/* Timeout expired */

		if ((nvend != 0 || tmo > MINTMO) && !notified) {
		    fprintf (stderr, "%s: waiting for bootp/dhcp", 
			     ifa->ifa_name);
		    notified = 1;
		}
		if (notified)
		    fprintf (stderr, ".");

		/* Create new request */
		wbp.bp_xid = htonl (++bootp->xid);
		gettimeofday (&tlast, NULL);
		wbp.bp_secs = htons (tlast.tv_sec - start.tv_sec);
		wbp.bp_unused = htons (0x8000); /* broadcast reply */
		bzero(wbp.bp_file, sizeof(wbp.bp_file));

		bcopy(vend[nvend], wbp.bp_vend, sizeof(u_int32_t));
		wbp.bp_vend[sizeof(u_int32_t)] = TAG_END;

		/* If we need vendor-specific info, cycle for next time */
		if ((bootp->need & ~bootp->have) & ~BOOT_ADDR)
		    nvend = (nvend + 1) % (sizeof(vend) / sizeof(vend[0]));

		/* Send request into the void */
		cc = sendto (s, &wbp, sizeof(wbp), 0,
			     (struct sockaddr *)&srvr, sizeof(srvr));
		if (cc != sizeof(wbp)) {
		    warn ("sendto");
		    goto quit;
		}

		/* Set timeout in TMO secs from now, 
		   and add in "random" usec delay. */
		tleft.tv_sec = tmo;
		tleft.tv_usec = tlast.tv_usec;

		/* calc time elapsed since request sent */
		timeradd (&tlast, &tleft, &twhen);
	    }

	    /* Wait for someone to reply */
	    FD_ZERO (&rfds);
	    FD_SET (s, &rfds);
	    n = select (s + 1, &rfds, 0, 0, &tleft);
	    if (n < 0 && errno != EINTR) {
		warn ("select");
		goto quit;
	    }

	    if (n == 0) {
#ifdef DEBUG
 printf ("select: timeout\n");
#endif
		/* timeout: quit if we rcvd any valid pcakets */
		if (recvcnt > 0)
		    break;
		timerclear (&tleft);	/* trigger resend */
		tmo *= 2;		/* bump timeout */
	    } else if (n > 0) {
		cc = recv (s, rbuf, sizeof(rbuf), 0);
#ifdef DEBUG
 printf ("recv: %d\n", cc);
#endif
		if (cc < 0) {
		    warn ("recv");
		    goto quit;
		}

		/* Got a packet, process it */
		cc = bootprecv (bootp, (struct bootp *)rbuf, cc);
		if (cc >= 0)
		    ++recvcnt;

		/* don't wait any longer if we have got all we want */
		if ((bootp->have & bootp->need) == bootp->need)
		    break;

		/* XXX will fail if time is manually changed! */
		gettimeofday (&tlast, NULL);

		/* compute time remaining until next timeout */
		timersub (&twhen, &tlast, &tleft);
	    }
	}
    } while (((bootp->have & bootp->need) != bootp->need) && tmo < MAXTMO);

    if (notified)
	fprintf (stderr, "\n");

quit:
    /* delete dummy bootp address */
    (void) ioctl(s, SIOCDIFADDR, &addreq);
    (void) signal (SIGINT, osigint);
    return recvcnt > 0;
}


/* Returns 0 if this is the packet we're waiting for else -1 */
static int
bootprecv (struct bootparams *bootp, struct bootp *bp, int rsize)
{
    const int hdrsize = sizeof(*bp) - sizeof (bp->bp_vend);
    char *s;
    u_long ul;

    if (rsize < hdrsize + 4 || ntohl(bp->bp_xid) != bootp->xid) {
#ifdef DEBUG
	if (rsize < hdrsize + 4)
	    printf("bootprecv: %s expected %d bytes, got %d\n",
		   bootp->name, hdrsize + 4, rsize);
	else 
	    printf("bootprecv: %s expected xid 0x%x, got 0x%x\n",
		   bootp->name, bootp->xid, ntohl(bp->bp_xid));
#endif
	return (-1);
    }

#ifdef DEBUG
    printf("bootprecv: %s got %d bytes from %s\n", bootp->name, rsize,
	   inet_ntoa (bp->bp_siaddr));
#endif
    bootp->have = 0;

    /* Pick up our ip address (and compute natural netmask) */
    if (bp->bp_yiaddr.s_addr != 0) {
	u_long addr;

	bootp->have |= BOOT_ADDR;
	bootp->addr = bp->bp_yiaddr;
#ifdef DEBUG
	printf("ipaddr: %s\n", inet_ntoa(bootp->addr));
#endif
	
	addr = ntohl (bootp->addr.s_addr);
	if (IN_CLASSA(addr))
	    bootp->nmask.s_addr = htonl (IN_CLASSA_NET);
	else if (IN_CLASSB(addr))
	    bootp->nmask.s_addr = htonl (IN_CLASSB_NET);
	else
	    bootp->nmask.s_addr = htonl (IN_CLASSC_NET);
#ifdef DEBUG
	printf("netmask: %s\n", inet_ntoa(bootp->nmask));
#endif
    }

    if (bp->bp_siaddr.s_addr != 0) {
	bootp->have |= BOOT_BOOTIP;
	bootp->bootip = bp->bp_siaddr;
    }

    if (bp->bp_file[0] != 0) {
	bootp->have |= BOOT_BOOTFILE;
	if (bootp->bootfile)
	    free (bootp->bootfile);
	bp->bp_file[sizeof (bp->bp_file) - 1] = '\0';	/* terminate */
	bootp->bootfile = strdup (bp->bp_file);
    }

    /* compute returned size of vendor specific area */
    rsize -= hdrsize;

    /* Parse vendor specific info */
    if (bcmp(vm_cmu, bp->bp_vend, sizeof(vm_cmu)) == 0)
	vend_cmu(bootp, bp->bp_vend, rsize);
    else if (bcmp(vm_rfc1048, bp->bp_vend, sizeof(vm_rfc1048)) == 0)
	vend_rfc1048(bootp, bp->bp_vend, rsize);
    else if (bcmp(vm_zero, bp->bp_vend, sizeof(vm_zero)) != 0) {
	bcopy(bp->bp_vend, &ul, sizeof(ul));
	warnx ("unknown vendor type 0x%x", ul);
    }

    /* Nothing more to do if we don't know our ip address yet */
    if ((bootp->have & BOOT_ADDR) == 0)
	return (-1);

    /* Check subnet mask against net mask; toss if bogus */
    if ((bootp->have & BOOT_MASK) != 0 &&
	(bootp->nmask.s_addr & bootp->mask.s_addr) != bootp->nmask.s_addr) {
#ifdef DEBUG
	printf("subnet mask (%s) bad\n", inet_ntoa(bootp->mask));
#endif
	bootp->have &= ~BOOT_MASK;
    }

    /* Use natural subnet mask if no other defined */
    if ((bootp->have & BOOT_MASK) == 0)
	bootp->mask = bootp->nmask;

#ifdef DEBUG
    if (bootp->have & BOOT_MASK)
	printf("subnet mask: %s\n", inet_ntoa(bootp->mask));
    if (bootp->have & BOOT_GATEIP) {
	printf("gateways: %s\n", s = bootp_getaddrstr (&bootp->gateip));
	free (s);
    }
    if (bootp->have & BOOT_DNSIP) {
	printf("dns servers: %s\n", s = bootp_getaddrstr (&bootp->dnsip));
	free (s);
    }
    if (bootp->have & BOOT_SYSLOGIP) {
	printf("log servers: %s\n", s = bootp_getaddrstr (&bootp->syslog));
	free (s);
    }
#if 0
    if (bootp->have & BOOT_NTPIP) {
	printf("ntp servers: %s\n", s = bootp_getaddrstr (&bootp->ntp));
	free (s);
    }
#endif
    if (bootp->have & BOOT_BOOTIP)
	printf("boot ip: %s\n", inet_ntoa(bootp->bootip));
    if (bootp->have & BOOT_HOSTNAME)
	printf ("hostname: %s\n", bootp->hostname);
    if (bootp->have & BOOT_DOMAINNAME)
	printf ("domainname: %s\n", bootp->domain);
    if (bootp->have & BOOT_BOOTFILE)
	printf("boot file: %s\n", bootp->bootfile);
#endif

    return (0);
}


static void
vend_cmu (struct bootparams *bootp, u_char *cp, u_int len)
{
    struct cmu_vend *vp;

    if (len < sizeof (struct cmu_vend)) {
#ifdef DEBUG
	printf("vend_cmu: expected %d bytes, got %d\n",
	       sizeof(struct cmu_vend), len);
#endif
	return;
    }

    vp = (struct cmu_vend *)cp;
    
    if (vp->v_smask.s_addr != 0) {
	bootp->have |= BOOT_MASK;
	bootp->mask = vp->v_smask;
    }
    if (vp->v_dgate.s_addr != 0) {
	bootp->have |= BOOT_GATEIP;
	bootp->gateip.naddr = 1;
	bootp->gateip.addrs[0] = vp->v_dgate;
    }
    if (vp->v_dns1.s_addr != 0) {
	bootp->have |= BOOT_DNSIP;
	bootp->dnsip.naddr = 1;
	bootp->dnsip.addrs[0] = vp->v_dns1;
    }
}


static void
vend_rfc1048 (struct bootparams *bootp, u_char *cp, u_int len)
{
    struct in_addr a;
    u_char *ep;
    int size;
    u_char tag;
    int i;

    ep = cp + len;

    /* Step over magic cookie */
    cp += sizeof(u_int32_t);

    while (cp < ep) {
	tag = *cp++;
	size = *cp++;
	if (tag == TAG_END)
	    break;
#ifdef DEBUG	
	printf("vend_rfc1048: got tag %d size %d\n",
	       tag, size);
#endif
	switch (tag) {
	case TAG_SUBNET_MASK:
	    if (size == sizeof(struct in_addr)) {
		bootp->have |= BOOT_MASK;
		bcopy(cp, &bootp->mask, size);
	    }
	    break;
	case TAG_IP_BROADADDR:
	    if (size == sizeof(struct in_addr)) {
		bootp->have |= BOOT_BROADADDR;
		bcopy(cp, &bootp->broadaddr, size);
	    }
	    break;
	case TAG_GATEWAY:
	    while (size >= sizeof(struct in_addr)) {
		if (!(bootp->have & BOOT_GATEIP))
		    bootp->gateip.naddr = 0;
		bootp->have |= BOOT_GATEIP;
		bootp_addaddr (&bootp->gateip, cp);
		size -= sizeof(struct in_addr);
		cp += sizeof(struct in_addr);
	    }
	    break;
	case TAG_DOMAIN_SERVER:
	    while (size >= sizeof(struct in_addr)) {
		if (!(bootp->have & BOOT_DNSIP))
		    bootp->dnsip.naddr = 0;
		bootp->have |= BOOT_DNSIP;
		bootp_addaddr (&bootp->dnsip, cp);
		size -= sizeof(struct in_addr);
		cp += sizeof(struct in_addr);
	    }
	    break;
	case TAG_LOG_SERVER:
	    while (size >= sizeof(struct in_addr)) {
		if (!(bootp->have & BOOT_LOGIP))
		    bootp->logip.naddr = 0;
		bootp->have |= BOOT_LOGIP;
		bootp_addaddr (&bootp->logip, cp);
		size -= sizeof(struct in_addr);
		cp += sizeof(struct in_addr);
	    }
	    break;
#if 0
	case TAG_NTP_SERVER:
	    while (size >= sizeof(struct in_addr)) {
		if (!(bootp->have & BOOT_NTPIP))
		    bootp->numntp = 0;
		bootp->have |= BOOT_NTPIP;
		bootp_addaddr (&bootp->ntpip, cp);
		size -= sizeof(struct in_addr);
		cp += sizeof(struct in_addr);
	    }
	    break;
#endif
	case TAG_HOST_NAME:
	    bootp->have |= BOOT_HOSTNAME;
	    if (bootp->hostname)
		bootp->hostname = realloc (bootp->hostname, size + 1);
	    else
		bootp->hostname = malloc (size + 1);
	    if (bootp->hostname) {
		memcpy (bootp->hostname, cp, size);
		bootp->hostname[size] = '\0';
	    }
	    break;
	case TAG_DOMAIN_NAME:
	    bootp->have |= BOOT_DOMAINNAME;
	    if (bootp->domain)
		bootp->domain = realloc (bootp->domain, size + 1);
	    else
		bootp->domain = malloc (size + 1);
	    if (bootp->domain) {
		memcpy (bootp->domain, cp, size);
		bootp->domain[size] = '\0';
	    }
	    break;
	}
	cp += size;
    }
}



/*
 * Create and initialise bootparam structure.
 */
static struct bootparams *
create_bootparams (struct ifaddrs *ifa, int numif)
{
    struct bootparams *bootp;
    char buf[32];
    char *val;

    bootp = malloc (sizeof (struct bootparams));
    if (!bootp)
	return NULL;

    memset (bootp, 0, sizeof (struct bootparams));
    if (numif == 1)
	bootp->have |= BOOT_ONLYIF;
    bootp->need = BOOT_ADDR;	/* XXX minimum requirement */
    bootp->name = strdup (ifa->ifa_name);

    snprintf (buf, sizeof(buf), "ifconfig_%s", ifa->ifa_name);
    if ((val = getenv (buf))
	|| (bootp->have & BOOT_ONLYIF
	    && (val = getenv ("ifconfig")))) {
	if (strcmp (val, "env") == 0)
	    bootp->have |= BOOT_IFCONFIG_ENV;
	else if (strcmp (val, "bootp") == 0)
	    bootp->have |= BOOT_IFCONFIG_BOOTP;
	else if (strcmp (val, "dhcp") == 0)
	    bootp->have |= BOOT_IFCONFIG_DHCP;
	else if (strcmp (val, "rarp") == 0)
	    bootp->have |= BOOT_IFCONFIG_RARP;
	else if (strcmp (val, "ignore") == 0)
	    bootp->have |= BOOT_IFCONFIG_SKIP;
	else
	    warnx ("bad $ifconfig value for %s: \"%s\"", ifa->ifa_name, val);
    }

    return bootp;
}


static void
free_bootparams (struct bootparams *bootp)
{
    free (bootp->name);
    free (bootp->hostname);
    free (bootp->domain);
    free (bootp->bootfile);
    free (bootp);
}



static struct bootparams *
config (struct ifaddrs *ifa, int numif)
{
    struct bootparams *bootp = NULL;
    struct ifaliasreq addreq;
    int nosave = 0;
    int s = -1;
    /*struct ifreq ifreq;*/

    bootp = create_bootparams (ifa, numif);
    if (!bootp ||
	(bootp->have & BOOT_IFCONFIG) == BOOT_IFCONFIG_SKIP)
	goto skip;

    s = socket (AF_INET, SOCK_DGRAM, 0);
    if (s < 0) {
	warn ("socket()");
	goto skip;
    }

    bzero (&addreq, sizeof(addreq));
    strncpy(addreq.ifra_name, ifa->ifa_name, IFNAMSIZ);

    if (ifa->ifa_flags & IFF_LOOPBACK) {
	/*
	 * Configure the loopback i/f (lo0).
	 */
	setsin ((struct sockaddr_in *)&addreq.ifra_addr, 
		AF_INET, htonl (INADDR_LOOPBACK));
	if (ioctl(s, SIOCAIFADDR, &addreq) < 0)
	    warn ("ioctl (%s, SIOCAIFADDR)", ifa->ifa_name);
	goto skip;
    }

    if (ifa->ifa_flags & IFF_POINTOPOINT)
	/*
	 * Can't do anything about point-to-point interfaces here
	 */
	goto skip;

    /* Then try bootp and dhcp. */
    switch (bootp->have & BOOT_IFCONFIG) {
    case BOOT_IFCONFIG_DHCP:
    case BOOT_IFCONFIG_BOOTP:
	/* try dhcp/bootp first */
	if (check_bootp (s, ifa, bootp))
	    break;
	warnx ("%s: no response from dhcp/bootp", ifa->ifa_name);
	/* fall through */
    case BOOT_IFCONFIG_RARP:
#ifdef notyet
	/* try rarp as fallback */
	if (check_rarp (s, ifa, bootp))
	    break;
	warnx ("%s: no response from rarp", ifa->ifa_name);
#endif
	/* finally fall through to using saved environment */
    case BOOT_IFCONFIG_ENV:
	warnx ("%s: checking stored configuration", ifa->ifa_name);
	nosave = 1;
	check_env (ifa, bootp);
	break;
    }

    if ((bootp->have & bootp->need) != bootp->need) {
	warnx ("%s: network configuration not available",
	       ifa->ifa_name);
	goto done;
    }

    /* Save the interface's DHCP/BOOTP values in the non-volatile
       environment, in case we can't contact the server next time we
       boot. */
    if (!nosave) {
	ia_setenv (ifa, "netaddr", bootp->addr);
	if (bootp->have & BOOT_MASK)
	    ia_setenv (ifa, "netmask", bootp->mask);
	else
	    ia_delenv (ifa, "netmask");
	if (bootp->have & BOOT_BROADADDR)
	    ia_setenv (ifa, "broadcast", bootp->broadaddr);
	else
	    ia_delenv (ifa, "broadcast");
    }

    /* Set the interface address. */
    setsin ((struct sockaddr_in *)&addreq.ifra_addr, 
	    AF_INET, bootp->addr.s_addr);
    if (bootp->have & BOOT_MASK)
	setsin ((struct sockaddr_in *)&addreq.ifra_mask, 
		0, bootp->mask.s_addr);
    if (bootp->have & BOOT_BROADADDR)
	setsin ((struct sockaddr_in *)&addreq.ifra_broadaddr, 
		AF_INET, bootp->broadaddr.s_addr);

    /* set the interface address */
    if (ioctl(s, SIOCAIFADDR, &addreq) < 0) {
	warn ("ioctl(%s, SIOCAIFADDR)", ifa->ifa_name);
	goto done;
    }

 done:
    close (s);
    return bootp;

 skip:
    if (s >= 0) close (s);
    if (bootp) free_bootparams (bootp);
    return NULL;
}



/* Merge parameters from *src into *dst */
static void
host_merge (struct bootparams *dst, struct bootparams *src)
{
    int s;

    if ((src->have & BOOT_HOSTNAME) && !(dst->have & BOOT_HOSTNAME)) {
	dst->hostname = strdup (src->hostname);
	dst->have |= BOOT_HOSTNAME;
    }

    if ((src->have & BOOT_DOMAINNAME) && !(dst->have & BOOT_DOMAINNAME)) {
	dst->domain = strdup (src->domain);
	dst->have |= BOOT_DOMAINNAME;
    }

    if (src->have & BOOT_GATEIP) {
	for (s = 0; s < src->gateip.naddr; s++)
	    bootp_addaddr (&dst->gateip, &src->gateip.addrs[s]);
	dst->have |= BOOT_GATEIP;
    }

    if (src->have & BOOT_DNSIP) {
	for (s = 0; s < src->dnsip.naddr; s++)
	    bootp_addaddr (&dst->dnsip, &src->dnsip.addrs[s]);
	dst->have |= BOOT_DNSIP;
    }

    if (src->have & BOOT_LOGIP) {
	for (s = 0; s < src->logip.naddr; s++)
	    bootp_addaddr (&dst->logip, &src->logip.addrs[s]);
	dst->have |= BOOT_LOGIP;
    }

#if 0
    /* XXX not supported yet */
    if (src->have & BOOT_NTPIP) {
	for (s = 0; s < src->ntp.naddr; s++)
	    bootp_addaddr (&dst->ntp, &src->ntp.addrs[s]);
	dst->have |= BOOT_NTPIP;
    }
#endif
}


/* Program the per-host setup. */
static void
host_setup (struct bootparams *bootp)
{
    int i;

/* XXX save DHCP/BOOTP values permanently in non-volatile environment */

    /* Pass the hostname into the kernel. */
    if (bootp->have & BOOT_HOSTNAME) {
	if (sethostname (bootp->hostname, strlen (bootp->hostname) + 1))
	    warn ("sethostname");
	setenv ("HOSTNAME", bootp->hostname, 3);
    }
    else {
	deleteenv ("HOSTNAME");
	deleteenv ("hostname");
    }

    /* Set NAMESERVER and LOCALDOMAIN environment variables for name
       server lookups. */
    /* XXX This relies on MIPS RTPX having a global environment */

    if (bootp->have & BOOT_DOMAINNAME)
	setenv ("LOCALDOMAIN", bootp->domain, 3);
    else {
	deleteenv ("LOCALDOMAIN");
	deleteenv ("localdomain");
    }

    if (bootp->have & BOOT_DNSIP) {
	char *str = bootp_getaddrstr (&bootp->dnsip);
	setenv ("NAMESERVER", str, 3);
	free (str);
    }
    else {
	deleteenv ("NAMESERVER");
	deleteenv ("nameserver");
    }

    if (bootp->have & BOOT_GATEIP) {
	char *str = bootp_getaddrstr (&bootp->gateip);
	setenv ("GATEWAY", str, 3);
	free (str);
    }
    else {
	deleteenv ("GATEWAY");
	deleteenv ("gateway");
    }

    if (bootp->have & BOOT_LOGIP) {
	char *str = bootp_getaddrstr (&bootp->logip);
	setenv ("LOGSERVER", str, 3);
	free (str);
    }
    else {
	deleteenv ("LOGSERVER");
    }

    /* We can now allow a symbolic $GATEWAY name for the default
       route, if it isn't already set. */

#ifdef KICKSTART_SYMBOLIC_GATEWAY
    if (!(bootp->have & BOOT_GATEIP)) {
	char *gw;
	if ((gw = getenv ("gateway")) || (gw = getenv ("GATEWAY"))) {
	    struct hostent *hp = gethostbyname (gw);
	    if (hp && hp->h_addrtype == AF_INET 
		&& hp->h_length == sizeof (struct in_addr)) {
		/* allow a list of alternate addresses */
		while (hp->h_addr_list[0]) {
		    bootp_addaddr (&bootp->gateip, hp->h_addr_list[0]);
		    hp->h_addr_list++;
		}
		bootp->have |= BOOT_GATEIP;
	    }
	}
    }
#endif

    /*
     * Now setup the routing tables, equivalent to:
     *  route add default $gateway
     */

    if (bootp->have & BOOT_GATEIP) {
	struct sockaddr_in so_dst;
	struct sockaddr_in so_gate;
	struct sockaddr_in so_netmask;
	int rtsock;

	/* install default route via the gateway */
	rtsock = socket(PF_ROUTE, SOCK_RAW, 0);
	if (rtsock < 0) {
	    warn ("socket(PF_ROUTE)");
	    return;
	}
	    
	setsin (&so_dst, AF_INET, htonl (INADDR_ANY));
	setsin (&so_netmask, AF_INET, 0);

#define ROUNDUP(a) \
	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n) (x += ROUNDUP((n)->sa_len))
#define NEXTADDR(w, u) \
	do {\
	    int l = ROUNDUP(u.sin_len); memcpy(cp, &(u), l); cp += l;\
	} while (0)

	/* Try each router in turn... */
	for (i = 0; i < bootp->gateip.naddr; i++) {
	    struct {
		struct	rt_msghdr m_rtm;
		char	m_space[512];
	    } m_rtmsg;
	    char *cp = m_rtmsg.m_space;
	    static int seq;

	    setsin (&so_gate, AF_INET, bootp->gateip.addrs[i].s_addr);

	    memset(&m_rtmsg, 0, sizeof(m_rtmsg));
#define rtm m_rtmsg.m_rtm
	    rtm.rtm_type = RTM_ADD;
	    rtm.rtm_flags = RTF_STATIC | RTF_UP | RTF_GATEWAY;
	    rtm.rtm_version = RTM_VERSION;
	    rtm.rtm_seq = ++seq;
	    rtm.rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;
	    /*rtm.rtm_rmx = rt_metrics;*/
	    /*rtm.rtm_inits = rtm_inits;*/
	    NEXTADDR(RTA_DST, so_dst);
	    NEXTADDR(RTA_GATEWAY, so_gate);
	    NEXTADDR(RTA_NETMASK, so_netmask);
	    rtm.rtm_msglen = cp - (char *)&m_rtmsg;

	    errno = 0;
	    if (write(rtsock, (char *)&m_rtmsg, rtm.rtm_msglen) > 0)
		break;

	    if (errno != ENETUNREACH && errno != ESRCH) {
		warn ("route add default %s", inet_ntoa (so_gate.sin_addr));
		break;
	    }
#ifdef DEBUG
	    warn ("route add default %s", inet_ntoa (so_gate.sin_addr));
#endif
	}
	if (errno != 0)
	    warnx ("no default routes worked");
	close (rtsock);
    }
}


kickstart ()
{
    struct bootparams *bootp, *ob;
    struct ifaddrs *iflist;
    struct ifaddrs *ifa;
    int numif = 0;

    if (getifaddrs (&iflist) < 0) {
	warn ("getifaddrs");
	return 1;
    }

    /* Count real network interfaces */
    for (ifa = iflist; ifa; ifa = ifa->ifa_next)
	if ((ifa->ifa_flags & (IFF_POINTOPOINT | IFF_LOOPBACK)) == 0)
	    ++numif;

    if (numif == 0) {
	fprintf (stderr, "No network interfaces found\n");
	exit (1);
    }

    /* configure each interface */
    /* XXX if we could be bothered, each could be run on a 
       different thread, in parallel. */
    for (ifa = iflist; ifa; ifa = ifa->ifa_next) {
	bootp = config (ifa, numif);
	if (bootp)
	    SLIST_INSERT_HEAD (&bootphead, bootp, list);
    }

    /* Done with interface list */
    freeifaddrs (iflist);

    /* Create a fake bootparams structure */
    bootp = malloc (sizeof (struct bootparams));
    memset (bootp, 0, sizeof (struct bootparams));

    /* Merge all the host-specific data into the new structure */
    SLIST_FOREACH(ob, &bootphead, list) {
	host_merge (bootp, ob);
    }

    /* Program the host-specific configuration */
    host_setup (bootp);

    /* Done with temp */
    free_bootparams (bootp);

    /* Now free all the stored per-interface data. 
       XXX In the future DHCP info will be retained and 
       managed by a forked process. */
    for (bootp = SLIST_FIRST (&bootphead);
	 bootp != SLIST_END (&bootphead);
	 bootp = ob) {
	ob = SLIST_NEXT (bootp, list);
	free_bootparams (bootp);
    }

    return (0);
}



void
usg_kick (FILE *fp)
{
    fprintf (fp, "usage: kick\n");
}


int
com_kick (int argc, char **argv)
{
    exit (kickstart ());
}
