/* micro_httpd - really small HTTP server
** modified for use with MIPS RTPX netmon  Ruth Moulton  June 2001
**
** Copyright  1999 by Jef Poskanzer <jef@acme.com>. All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <syslog.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <getopt.h>

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/file.h>
#include <sys/wait.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#define SERVER_NAME "micro_httpd"
#define SERVER_URL "http://www.acme.com/software/micro_httpd/"
#define PROTOCOL "HTTP/1.0"
#define RFC1123FMT "%a, %d %b %Y %H:%M:%S GMT"

/* local web pages */

/* the pages are in the directory 'info', each has a name and the html 
that makes up the page itself. 

e.g. To access info about MIPS RTPX, the browser should 
get http://<board address>/info
*/

#define PAGEDIR "info"  /* directory */

#define PAGENET "mipsnet" /* re algnet */
#define INFONET "<body bgcolor=\"#FFE4C4\"><center><font size=+3 color=\"#B22222\">MIPSnet TCP/IP software package</font></center></p><font size=+1>MIPSnet is MIPS Technologies' embedded internet communications (\"TCP/IP\") software package. MIPSnet is available as a component of the MIPS Software Toolkit, as an extension to the MIPS SDE toolchain and <a href=\"rtpx\"> MIPS RTPX </a> real-time extension, providing an environment for the development of network-capable embedded applications.</p><p> <a href=mailto:software@mips.com>Contact MIPS Technologies</a> for more information.</p><p> <a href=\"netmon\">Netmon</a> is a sample program written using the MIPSnet environment</p></body>"

#define PAGERTPX "rtpx"
#define INFORTPX "<body bgcolor=\"#DAA520\"> <center><font size=+3 color=\"#FFDEAD\" face=helvetica> MIPS Technologies\' Real-Time POSIX eXtensions</font></center><font size=+1> <p>MIPS RTPX is a component of the MIPS Software Toolkit, and forms a multi-threaded, real-time extension to the single-threaded base MIPS SDE environment. It implements the POSIX Threads and POSIX.4 real-time APIs for CPUs based on the MIPS architecture. Together with the <a href=\"mipsnet\">MIPSnet</a> TCP/IP extensions, a tiny layered OS for embedded networked applications is provided by the MIPS Software Toolkit.</p><p>  <A HREF=mailto:software@mips.com>Get in touch with MIPS Technologies</a> or visit our <a href=http://www.mips.com> web site </a> </p> </body> "

#define PAGENETMON "netmon"
#define INFONETMON "<body bgcolor=\"#D2691E\"><center><font size=+3 color=\"#FFD700\" face=helvetica> MIPS Technologies Network Monitor</font></center><font size=+1 color=\"white\"> <p>Netmon is a program which builds on and demonstrates the <A HREF=\"mipsnet\">MIPSnet environment</A>. It provides a framework within which to implement an Internet servers and clients, and includes Ping, a BOOTP client, a Telnet server, an HTTP server (you are talking to it now!), and other utilities. It has a monitor that can be accessed through the serial port on the board, or over the network via Telnet. It also has an implementation of the Inetd protocol server.</p><p> It is implemented using features of MIPS Technologies\' real-time package <a HREF=\"rtpx\"> MIPS RTPX </a></p></body> " 

/* hard-wired directory name  */
static const char * const our_directories[]={ PAGEDIR, NULL};

/* hard-wired web pages */
static const struct info_file {
  const char *name;
  const char *contents;
} info_files [] = {
  {PAGENET, INFONET},
  {PAGERTPX, INFORTPX},
  {PAGENETMON, INFONETMON},
  {NULL,NULL}
};

#define MAXBUFFS 10

/* per-server parameters */
struct httpopts {
    int			debug;
    int			portnbr;
};

/* per-connection details */
struct connection {
    struct httpopts *	hopt;
    int			fd;
    FILE *		outfp;
    FILE *		infp;
    char *		buff;
};

/* Forwards. */
static void send_error(int status, const char * title, 
		       const char * extra_header, const char * text, 
		       struct connection *conn);
static void send_headers(int status, const char * title,
			 const char * extra_header,
			 const char * mime_type, 
			 int length, time_t mod, FILE *fp);
static const char * get_mime_type (const char * name);
static int httpd_process_call (void *arg);

/* SIGCHLD handler */
static void
reaper (int signum)
{
    int save_errno = errno, status;
    pid_t pid;

    /* XXX would need to use thread-specific data to 
       retrieve the current process' httpopts root. */
    for (;;) {
	pid = waitpid (-1, &status, WNOHANG);
	if (pid <= 0)
	    break;
	syslog (LOG_DEBUG, "child pid %d exited with status 0x%x\n", 
		pid, status);
    }
    errno = save_errno;
}


static int httpd (void *arg)
{
    struct httpopts *hopt = arg;
    struct connection *conn;
    struct sockaddr_in sin;
    int sinlen;
    int s, n, pid, err;
    fd_set readable;
    int maxsock = 0;
    int on = 1;
    fd_set allsock;

    FD_ZERO(&allsock);
  
    if (!hopt->debug) {    
	/* background daemon - detach from terminal */
#ifndef unix
	/* set own process/thread name */
	pthread_setname_np (pthread_self (), "httpd");
	/* move to safe directory */
	chdir ("/");
	/* detach from tty */
	setpgrp ();
	ioctl (0, TIOCNOTTY, 0);
#else
	daemon (1, 1);
#endif
    }

    /* Open our syslog interface */
    openlog ("httpd", LOG_PID | LOG_ODELAY | LOG_CONS, LOG_DAEMON);
    setlogmask (hopt->debug ? LOG_UPTO (LOG_DEBUG) : LOG_UPTO (LOG_INFO));

    /* Grab a TCP socket.  */
    s = socket (AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	syslog (LOG_ERR, "socket(): %m");
	free (hopt);
	exit (1);
    }

    (void)setsockopt (s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

    /* Bind to the specified port */
    memset ((void *)&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = htonl (INADDR_ANY);
    sin.sin_port = htons (hopt->portnbr);
    sin.sin_len = sizeof (sin);

    if (bind (s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	syslog (LOG_ERR, "bind(): %m");
	free (hopt);
	exit (1);
    }

    /* allow ten pending connections at a time */
    if (listen (s, 10) < 0) {
	syslog (LOG_ERR, "listen(): %m");
	free (hopt);
	exit (1);
    }

    /* install catcher to reap zombie children */
    signal (SIGCHLD, reaper);

    syslog (LOG_INFO, "httpd starting on port %d", hopt->portnbr);

    FD_SET(s, &allsock);
    if (maxsock < s)
	maxsock = s;

    while (1) {

	/* select on the  socket */
	
	readable = allsock;

	n = select (maxsock+1,   &readable, NULL, NULL, NULL);

	if (n < 0) {
	    if (errno != EINTR) {
		syslog (LOG_WARNING, "select(): %m");
		sleep(1);
	    }
	    continue;
	}

	if (n == 0)
	    continue;

	if (FD_ISSET(s, &readable)) {
	    int fd;

	    fd = accept (s, (struct sockaddr *)&sin, &sinlen);
	    if (fd < 0) {
		if (errno != EINTR)
		    syslog (LOG_WARNING, "accept(): %m");
		continue;
	    }
	    if (hopt->debug)
		fprintf(stderr,"fd is %d\n",fd);
	  
	    /* create connection data */
	    conn = malloc (sizeof (struct connection));
	    if (!conn) {
		syslog (LOG_ERR, "malloc: %m");
		exit(1);
	    }
	    memset (conn, 0, sizeof (struct connection));
	    conn->hopt = hopt;
	    conn->fd = fd;

	    /* fork process to handle connection  */
#ifdef unix
	    pid = fork ();
	    if (pid < 0) {
		syslog (LOG_ERR, "fork: %m");
		free (conn);
	    }
	    if (pid == 0)
		httpd_process_call(conn);
	    if (pid)
		/* parent */
		close(fd);  
#else /* ! unix */
	    pid = 0;
	    if (err = pthread_proc_create_np (&pid, NULL,
					      httpd_process_call, conn))
	    {
		errno = err;
		syslog (LOG_ERR, "proc_create: %m");
		free (conn);
	    }
	    close(fd);  
#endif /* ! unix */
	}
    }
}

static void
h_exit (int evalue, struct connection *conn)
{
    int i; 
    char *buff;
    
    if (conn->buff)
	free(conn->buff);
    
    if (conn->outfp) {
	(void) fflush(conn->outfp);
	fclose(conn->outfp);
    }
  
    if (conn->infp)
	fclose(conn->infp);

    if (conn->hopt->debug)
	fprintf(stderr,"exit with value %d\n",evalue);

    exit(evalue);
}

static int
h_stat (const char *file, struct stat *sb)
{
    const char *dir, *filen;
    int i, j;

    memset ((void *)sb, 0, sizeof(struct stat));
    sb->st_mtime = -1;

    syslog(LOG_DEBUG,"h_stat: file %s\n",file);

    /* we know about our own 'directory' and 'files' */

    for (i = 0 ; dir = our_directories[i]; i++){

	if (!strcmp(dir,file) ) {
	    sb->st_mode = S_IFDIR;
	    return(0);
	}

	if (file[strlen(file)-1] == '/'){
	    if (strncmp(dir,file,strlen(file)-1) == 0) {
		sb->st_mode = S_IFDIR;
		sb->st_size = -1;
		return(0);
	    }
	}

	/* it's not one of our directories - we are assuming flat
           dir. structure */

	/* check out files */

	if (strncmp (dir, file, strlen(dir)) == 0
	    && file[strlen(dir)] == '/'){
	    const char *fname = &file[strlen(dir)+1];

	    syslog(LOG_DEBUG,"file part is %s\n",fname);

	    for (j=0; (filen = info_files[j].name); j++){
		if (!strcmp(fname, filen) ){
		    sb->st_mode = S_IFREG;
		    sb->st_size = strlen(info_files[j].contents);
		    return(0);
		}
	    }
	}
    }

    /* not one of ours */
    return ( -1);
}


static const char *
h_fopen (const char *file, char *mode)
{
    int i,j;
    const char *filen, *dir, *filename;

    syslog (LOG_DEBUG,"open file  %s\n",file);

    /* assume the file path is <dir>/name */
    for (i = 0; dir = our_directories[i]; i++){
	if (strncmp (file, dir, strlen (dir)) == 0) {

	    filename = &file[strlen (dir) + 1];
          
	    for (i = 0; filen = info_files[i].name; i++) {
		if (strcmp (filename,filen) == 0) {
		    syslog (LOG_DEBUG,"open file at %p",
			    &info_files[i].contents);
		    return (info_files[i].contents);
		}      
	    }
	}
    }

    return (0);
}


static int 
httpd_process_call (void *arg)
{
    struct connection *conn = arg;
    char *file, *input;
    int len, ich, i;
    struct stat sb;
    int fd = conn->fd;
    const char *fp;    /* pointer to internal web page, i.e. internal 'file' */
    FILE *stdoutfp, *stdinfp;  /* fp to use for output instead of stdout */
    int socketfp;    /* socket number */
    int linesize = 10000;
    /* limited stack size, so malloc the space instead */
    /*    char line[10000], method[10000], path[10000], protocol[10000], 
	  idx[20000], location[20000], command[20000]; */
    char *line, *method, *path, *protocol, *idx, *location, *command;

#ifndef unix
    /* Open a new syslog interface */
    openlog ("httpd", LOG_PID | LOG_ODELAY | LOG_CONS, LOG_DAEMON);
    setlogmask (conn->hopt->debug ? LOG_UPTO (LOG_DEBUG)
		: LOG_UPTO (LOG_INFO));
#endif

    syslog (LOG_DEBUG, "accepted call from http client on fd %d", fd);

    /* get some space to store text, client requests etc */
    if (!conn->buff)
	conn->buff = malloc (10000+10000+10000+10000
			     +20000+20000+20000);
    if (!conn->buff) {
	syslog(LOG_ERR,"malloc failed %m");
	h_exit(2, conn);
    }	

    line = conn->buff;
    method = line + 10000;
    path = method + 10000;
    protocol = path + 10000;
    idx = protocol + 10000;
    location = idx + 20000;
    command = location + 20000;

    if (!(stdinfp = conn->infp = fdopen(fd, "r")) ||
	!(stdoutfp = conn->outfp = fdopen(dup(fd), "w"))) {
	syslog (LOG_ERR, "cannot fdopen  on %d", fd);
	h_exit(2,conn);
    }

    input = fgets( line, linesize, stdinfp );
    if ( input == (char*) 0 ) {
	syslog(LOG_DEBUG, "client sent no input");
	send_error( 400, "Bad Request", (char*) 0, "No request found.", conn);
    }
    else
	syslog(LOG_DEBUG, "client sent %s",input);


    if ( sscanf( line, "%[^ ] %[^ ] %[^ ]", method, path, protocol ) != 3 ) {
	syslog(LOG_INFO,"bad request");
	send_error( 400, "Bad Request", (char*) 0, "Can't parse request.",conn);
    }

    while ( fgets( line, linesize, stdinfp ) != (char*) 0 ) {
	if ( strcmp( line, "\n" ) == 0 || strcmp( line, "\r\n" ) == 0 )
	    break;
    }

    if ( strcasecmp( method, "get" ) != 0 ) {
	syslog(LOG_INFO,"%s: method not implemented", method);
	send_error( 501, "Not Implemented", (char*) 0, "That method is not implemented.", conn);
    }

    if ( path[0] != '/' ) {
	syslog(LOG_INFO,"%s: bad filename", path);
	send_error( 400, "Bad Request", (char*) 0, "Bad filename.", conn);
    }

    file = &(path[1]);
    if ( file[0] == '\0' )
	file = "./";
    len = strlen( file );

    if ( file[0] == '/' || strcmp( file, ".." ) == 0 || strncmp( file, "../", 3 ) == 0 || strstr( file, "/../" ) != (char*) 0 || strcmp( &(file[len-3]), "/.." ) == 0 ) {
	syslog(LOG_INFO,"%s: illegal filename", file);
	send_error( 400, "Bad Request", (char*) 0, "Illegal filename.",conn);
    }
   
    if ( h_stat( file, &sb ) < 0 ) {
	syslog(LOG_INFO,"%s: file not found", file);
	send_error( 404, "Not Found", (char*) 0, "File not found.", conn);
    }

    if ( S_ISDIR( sb.st_mode ) ) {
	const char *dir;

	if ( file[len-1] != '/' ) {
	    syslog(LOG_DEBUG,"%s: is a directory", file);
	    (void) snprintf(
		location, linesize, "Location: %s/", path );
	    send_error( 302, "Found", location, "Directories must end with a slash.", conn);
	}
	syslog(LOG_DEBUG,"generating %s/index.html", file);
	(void) snprintf( idx, linesize, "%sindex.html", file );
	if ( stat( idx, &sb ) >= 0 ) {
	    file = idx;
	    goto do_file;
	}
	    
	send_headers( 200, "Ok", (char*) 0, "text/html", -1, sb.st_mtime ,stdoutfp);
	(void) fprintf(stdoutfp, "<HTML><HEAD><TITLE>Index of %s</TITLE></HEAD>\n<BODY BGCOLOR=\"#99cc99\"><H4>Index of %s</H4>\n<PRE>\n", file, file );

	/* catch our own directories */
	for (i = 0; dir = our_directories[i]; i++){
	    if (strncmp (dir, file, strlen(dir)) == 0) {
		const char *filen; 
		int j;

		for (j=0; filen = info_files[j].name; j++) {
		    fprintf(stdoutfp, "<A HREF=\"%s\">%s</A>\n",filen,filen);
		}
		(void) fprintf(stdoutfp, "</PRE>\n<HR>\n<ADDRESS><A HREF=\"%s\">%s</A></ADDRESS>\n</BODY></HTML>\n", SERVER_URL, SERVER_NAME );
		h_exit(0, conn);
	    }
	}
	      
#ifdef unix
	if ( strchr( file, '\'' ) == (char*) 0 )
	{
	    (void) snprintf( command, linesize,
			     "ls -lgF '%s' | tail +2 | sed -e 's/^\\([^ ][^ ]*\\)\\(  *[^ ][^ ]*  *[^ ][^ ]*  *[^ ][^ ]*\\)\\(  *[^ ][^ ]*\\)  *\\([^ ][^ ]*  *[^ ][^ ]*  *[^ ][^ ]*\\)  *\\(.*\\)$/\\1 \\3  \\4  |\\5/' -e '/ -> /!s,|\\([^*]*\\)$,|<A HREF=\"\\1\">\\1</A>,' -e '/ -> /!s,|\\(.*\\)\\([*]\\)$,|<A HREF=\"\\1\">\\1</A>\\2,' -e '/ -> /s,|\\([^@]*\\)\\(@* -> \\),|<A HREF=\"\\1\">\\1</A>\\2,' -e 's/|//'",
			     file );
	    (void) fflush( stdoutfp );
	    system( command );
	}
#else
	syslog(LOG_DEBUG,"%s: unknown directory", file);
	fprintf (stdoutfp, "Unknown directory");
#endif
	(void) fprintf(stdoutfp, "</PRE>\n<HR>\n<ADDRESS><A HREF=\"%s\">%s</A></ADDRESS>\n</BODY></HTML>\n", SERVER_URL, SERVER_NAME );
    }
    else
    {
    do_file:
	fp = h_fopen ( file, "r" );
	if ( fp ==  0 )
	    send_error( 403, "Forbidden", (char*) 0, "File is protected.",conn);
	
	send_headers( 200, "Ok", (char*) 0, get_mime_type( file ),
		      sb.st_size, sb.st_mtime, stdoutfp );
	/*	while ( ( ich = getc( fp ) ) != EOF )
		putchar( ich ); */
	fprintf(stdoutfp,"%s",fp);
    }
    h_exit( 0, conn);
}

static void
send_error( int status, const char* title, const char* extra_header,
	    const char* text, struct connection *conn)
{
    FILE *stdoutfp = conn->outfp;

    send_headers( status, title, extra_header, "text/html", -1, -1,stdoutfp );
    (void) fprintf(stdoutfp, "<HTML><HEAD><TITLE>%d %s</TITLE></HEAD>\n<BODY BGCOLOR=\"#cc9999\"><H4>%d %s</H4>\n", status, title, status, title );
    (void) fprintf(stdoutfp, "%s\n", text );
    (void) fprintf(stdoutfp, "<HR>\n<ADDRESS><A HREF=\"%s\">%s</A></ADDRESS>\n</BODY></HTML>\n", SERVER_URL, SERVER_NAME );

    h_exit(1, conn);
}

static void
send_headers (int status, const char* title, const char* extra_header, 
	      const char* mime_type, int length, time_t mod, FILE *stdoutfp)
{
    time_t now;
    char timebuf[100];

    (void) fprintf(stdoutfp,  "%s %d %s\r\n", PROTOCOL, status, title );
    (void) fprintf(stdoutfp, "Server: %s\r\n", SERVER_NAME );
    now = time( (time_t*) 0 );
    (void) strftime( timebuf, sizeof(timebuf), RFC1123FMT, gmtime( &now ) );
    (void) fprintf(stdoutfp, "Date: %s\r\n", timebuf );
    if ( extra_header != (char*) 0 )
	(void) fprintf(stdoutfp, "%s\r\n", extra_header );
    if ( mime_type != (char*) 0 )
	(void) fprintf(stdoutfp, "Content-Type: %s\r\n", mime_type );
    if ( length >= 0 )
	(void) fprintf(stdoutfp, "Content-Length: %d\r\n", length );
    if ( mod != (time_t) -1 )
    {
	(void) strftime( timebuf, sizeof(timebuf), RFC1123FMT, gmtime( &mod ) );
	(void) fprintf(stdoutfp, "Last-Modified: %s\r\n", timebuf );
    }
    (void) fprintf(stdoutfp, "Connection: close\r\n" );
    (void) fprintf(stdoutfp, "\r\n" );
}

static const char* get_mime_type (const char* name)
{
    const char* dot;

    dot = strrchr( name, '.' );
    if ( dot == (char*) 0 )
	return "text/html; charset=iso-8859-1";
    if ( strcmp( dot, ".html" ) == 0 || strcmp( dot, ".htm" ) == 0 )
	return "text/html; charset=iso-8859-1";
    if ( strcmp( dot, ".jpg" ) == 0 || strcmp( dot, ".jpeg" ) == 0 )
	return "image/jpeg";
    if ( strcmp( dot, ".gif" ) == 0 )
	return "image/gif";
    if ( strcmp( dot, ".png" ) == 0 )
	return "image/png";
    if ( strcmp( dot, ".css" ) == 0 )
	return "text/css";
    if ( strcmp( dot, ".au" ) == 0 )
	return "audio/basic";
    if ( strcmp( dot, ".wav" ) == 0 )
	return "audio/wav";
    if ( strcmp( dot, ".avi" ) == 0 )
	return "video/x-msvideo";
    if ( strcmp( dot, ".mov" ) == 0 || strcmp( dot, ".qt" ) == 0 )
	return "video/quicktime";
    if ( strcmp( dot, ".mpeg" ) == 0 || strcmp( dot, ".mpe" ) == 0 )
	return "video/mpeg";
    if ( strcmp( dot, ".vrml" ) == 0 || strcmp( dot, ".wrl" ) == 0 )
	return "model/vrml";
    if ( strcmp( dot, ".midi" ) == 0 || strcmp( dot, ".mid" ) == 0 )
	return "audio/midi";
    if ( strcmp( dot, ".mp3" ) == 0 )
	return "audio/mpeg";
    if ( strcmp( dot, ".pac" ) == 0 )
	return "application/x-ns-proxy-autoconfig";
    return "text/plain; charset=iso-8859-1";
}

void usg_httpd(FILE *fp)
{
    fprintf (fp, "httpd [-d] \n");
}

int
#ifdef unix
main( int argc, char** argv )
#else
com_httpd(int argc, char **argv)
#endif
{
    struct httpopts *hopt;
    int ch;

    hopt = malloc (sizeof (struct httpopts));
    if (!hopt) {
	perror ("malloc");
	exit (1);
    }

    /* reset variables */
    memset (hopt, 0, sizeof (struct httpopts));
    hopt->portnbr = 80;		/* default http port */

    while ((ch = getopt (argc, argv, "d")) != -1)
	switch (ch) {
	case 'd':
	    hopt->debug = 1;
	    break;
	case '?':
	default:
	    usg_httpd (stderr);
	    exit (1);
	}
    argv += optind;
    argc -= optind;

    if (argc != 0) {
	usg_httpd (stderr);
	free(hopt);
	exit (1);
    }

    if (!hopt->debug) {
	/* start in background process */
	int err;
	if (err = pthread_proc_create_np (NULL, NULL, httpd, hopt)) {
	    perror ("fork");
	    free (hopt);
	}
    }
    else {
	httpd (hopt);
    }

    exit (0);

}
