/* 
 * netmon.c: Network test command-line shell
 *
 * Copyright (c) 2001 Algorithmics Ltd - all rights reserved.
 * 
 * This program is NOT free software, it is supplied under the terms
 * of the AlgRTX License Agreement, a copy of which is available at:
 *
 *  http://www.algor.co.uk/algor/info/algrtx-license.pdf
 *
 * Any company which has obtained and signed a valid AlgRTX license
 * may use and modify this software internally and use (without
 * restrictions) any derived binary.  You may not, however,
 * redistribute this in whole or in part as source code, nor may you
 * modify or remove any part of this copyright message.  
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <signal.h>
#include <setjmp.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>
#include <err.h>

#ifndef __OPTIMIZE__
/* Enable MIPS RTPX debugging: */

/* Log all events */
unsigned int _rtpx_evlog_mask = ~0;	

/* Do error checking on all mutexes */
int _rtpx_mutex_default_type = PTHREAD_MUTEX_ERRORCHECK_NP;
int _rtpx_deadlock_fatal = 1;
#endif

/* Builtin startup script */
static int scriptidx = 0;
static const char * const script[] = {
#if 1
    /* setup interface using BOOTP */
    "kick",
#else
    /* setup interface manually */
    "setenv NAMESERVER 192.168.8.237",
    "hostname lt7.algor.co.uk",
    "ifconfig lo0 inet 127.0.0.1",
    "ifconfig de0 inet 192.168.8.231 netmask 255.255.255.0",
    "sleep 5",
    "route add default ltrouter",
#endif
    "hostname",
#ifndef MINIMALIST
    "ifconfig -a",
    "route show",
    "inetd",
#endif
#if 0
    "debug",		/* start gdb connection */
#endif
    "telnetd",
    "httpd",
    0
};

extern char   **environ;

struct comtab {
    const char     *name;
    const char     *help;
    void	  (*usage) (FILE *);
    int           (*command) (int argc, char **argv);
    int		    builtin;
};

extern const struct comtab comtab[];

static int
Usage (void (*usg)(FILE *))
{
    (*usg) (stderr);
    return EXIT_FAILURE;
}


static const char hlp_help[] = "give brief help on commands";

static void
usg_help (FILE *fp)
{
    fprintf (fp, "usage: help [COMMAND]\n");
}

static int
com_help (int argc, char **argv)
{
    const struct comtab *ct;
    static int maxname = 0;

    if (argc == 2) {
	for (ct = comtab; ct->name; ct++)
	    if (strcmp (argv[1], ct->name) == 0)
		break;
	if (ct->name)
	    (ct->usage) (stdout);
	else
	    fprintf (stderr, "%s: not found\n", argv[1]);
	return 0;
    }

    if (maxname == 0) {
	size_t len;
	for (ct = comtab; ct->name; ct++) {
	    len = strlen (ct->name);
	    if (len > maxname)
		maxname = len;
	}
    }

    printf ("Available commands:\n");
    for (ct = comtab; ct->name; ct++)
	printf ("  %-*s  %s\n", maxname, ct->name, ct->help);

    return 0;
}


static const char hlp_exit[] = "terminate this shell";

static void
usg_exit (FILE *fp)
{
    fprintf (fp, "usage: exit\n");
}

static int
com_exit (int argc, char **argv)
{
    /* actually, an exit from here would normally do nothing, since ww
       would normally be running in a sub process, and the exit would
       just terminate that. But the command line parser will treat
       exit as a special builtin, and doesn't fork a new process. */
    if (argc != 1) {
	usg_exit (stderr);
	return EXIT_FAILURE;
    }
    exit (0);
}


static const char hlp_setenv[] = "display/set environment variable";

static void usg_setenv (FILE *fp)
{
    fprintf (fp, "usage: setenv [-p] [NAME [VALUE]]\n");
    fprintf (fp, "  With no arguments all variables are displayed.\n");
    fprintf (fp, "  With no VALUE, just the value of NAME is displayed.\n");
    fprintf (fp, "  With NAME and VALUE, variable NAME is set to VALUE,\n");
    fprintf (fp, "    -p stores it permanently in NVRAM\n");
}

static int
com_setenv (int argc, char **argv)
{
    int pflag = 0;
    int c;

    while ((c = getopt (argc, argv, "p")) != -1) {
	switch (c) {
	case 'p':
	    pflag = 1;
	    break;
	default:
	    return Usage (usg_setenv);
	}
    }
    argc -= optind;
    argv += optind;

    if (argc == 0) {
	char **envp;
	getenv ("foo");		/* may be needed to init environ */
	for (envp = environ; *envp; envp++)
	    puts (*envp);
	return EXIT_SUCCESS;
    } else if (argc == 1) {
	char *s = getenv (argv[0]);
	if (s)
	    printf ("%s=%s\n", argv[0], s);
	else
	    printf ("%s: undefined\n", argv[0]);
	return EXIT_SUCCESS;
    } else if (argc == 2) {
	if (setenv (argv[0], argv[1], pflag ? 3 : 1) < 0) {
	    perror ("setenv");
	    return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
    } else
	return Usage (usg_setenv);
}


static const char hlp_unsetenv[] = "delete environment variable(s)";

static void usg_unsetenv (FILE *fp)
{
    fprintf (fp, "usage: unsetenv [-p] NAME ...\n");
    fprintf (fp, "    -p erases permanently from NVRAM\n");
}

static int
com_unsetenv (int argc, char **argv)
{
    int pflag = 0;
    int c, i;

    while ((c = getopt (argc, argv, "p")) != -1) {
	switch (c) {
	case 'p':
	    pflag = 1;
	    break;
	default:
	    return Usage (usg_unsetenv);
	}
    }
    argc -= optind;
    argv += optind;

    for (i = 0; i < argc; i++) {
	if (pflag)
	    deleteenv (argv[0]);
	else
	    unsetenv (argv[0]);
    }
    return EXIT_SUCCESS;
}


static const char hlp_hostname[] = "display/set hostname";

static void usg_hostname (FILE *fp)
{
    fprintf (fp, "usage: hostname [NAME]\n");
}

static int
com_hostname (int argc, char **argv)
{
    if (argc == 1) {
	char buf[MAXHOSTNAMELEN];
	buf[0] = '\0';
	if (gethostname (buf, sizeof(buf)-1) != 0) {
	    perror ("gethostname");
	    return EXIT_FAILURE;
	}
	else if (buf[0]) {
	    buf[sizeof(buf)-1] = 0;
	    printf ("%s\n", buf);
	}
	else {
	    fprintf (stderr, "Hostname unknown\n");
	}
	return EXIT_SUCCESS;
    } else if (argc == 2) {
	if (sethostname (argv[1], strlen (argv[1]) + 1) != 0) {
	    perror ("sethostname");
	    return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
    } else 
	return Usage (usg_hostname);
}


static const char hlp_sleep[] = "delay for some time";

static void usg_sleep (FILE *fp)
{
    fprintf (fp, "usage: sleep SECS\n");
}

static int
com_sleep (int argc, char **argv)
{
    if (argc == 2) {
	int secs = atoi (argv[1]);
	if (secs > 0) {
	    sleep (secs);
	    return EXIT_SUCCESS;
	}
    }
    return Usage (usg_sleep);
}


static const char hlp_ps[] = "display thread status";

static void usg_ps (FILE *fp)
{
    fprintf (fp, "usage: ps\n");
}

static int
com_ps (int argc, char **argv)
{
    if (argc != 1)
	return Usage (usg_ps);
    _rtpx_threadstat_user (printf);
    return EXIT_SUCCESS;
}


static const char hlp_kill[] = "send signal to a process";

static void usg_kill (FILE *fp)
{
    fprintf (fp, "usage: kill [-SIG] PID\n");
}

static int
com_kill (int argc, char **argv)
{
    int signo = SIGTERM;	/* default signal */
    char *cp;

    ++argv;
    --argc;

    /* look for signal number */
    if (argc > 0 && argv[0][0] == '-') {
	char *name = &argv[0][1];
	/* first check for a numeric value */
	signo = strtol (name, &cp, 10);
	if (*cp) {
	    /* nope, try a symbolic one */
	    extern const char * const sys_signame[NSIG];
	    /* remove optional initial "SIG" */
	    if (strncmp (name, "SIG", 3) == 0)
		name += 3;
	    /* look for match in signame table */
	    for (signo = 1; signo < NSIG; signo++)
		if (strcmp (sys_signame[signo], name) == 0)
		    break;
	} 
	if (signo <= 0 || signo >= NSIG) {
	    fprintf (stderr, "bad signal number: %s\n", &argv[0][1]);
	    return EXIT_FAILURE;
	}
	++argv;
	--argc;
    }

    if (argc <= 0)
	return Usage (usg_kill);

    while (argc != 0) {
	int pid = strtol (argv[0], &cp, 10);
	/* negative values are allowed, for process group */
	if (*cp || pid == 0 || pid == 1 || pid == -1 || pid == getpid()) {
	    fprintf (stderr, "bad process id: %s\n", argv[0]);
	    return EXIT_FAILURE;
	}
	if (kill (pid, signo) < 0) {
	    perror ("kill");
	    return EXIT_FAILURE;
	}	    
	++argv;
	--argc;
    }

    return EXIT_SUCCESS;
}


static const char hlp_debug[] = "cause a debugger breakpoint";

static void usg_debug (FILE *fp)
{
    fprintf (fp, "usage: debug\n");
}

static int
com_debug (int argc, char **argv)
{
    if (argc != 1)
	return Usage (usg_debug);
    _debug_breakpoint ();
    return EXIT_SUCCESS;
}


static const char hlp_kick[] = "kickstart network interface(s)";
extern int com_kick (int, char *[]);
extern void usg_kick (FILE *);

static const char hlp_netstat[] = "display network tables and statistics";
extern int com_netstat (int, char *[]);
extern void usg_netstat (FILE *);

static const char hlp_ping[] = "send echo packets to remote host";
extern int com_ping (int, char *[]);
extern void usg_ping (FILE *);

#ifndef MINIMALIST
static const char hlp_ifconfig[] = "setup/display network interfaces";
extern int com_ifconfig (int, char *[]);
extern void usg_ifconfig (FILE *);

static const char hlp_sysctl[] = "get or set network kernel state";
extern int com_sysctl (int, char *[]);
extern void usg_sysctl (FILE *);

static const char hlp_arp[] = "manipulate the system ARP cache";
extern int com_arp (int, char *[]);
extern void usg_arp (FILE *);

static const char hlp_route[] = "manipulate the system routing tables";
extern int com_route (int, char *[]);
extern void usg_route (FILE *);

static const char hlp_inetd[] = "internet test server";
extern int com_inetd (int, char *[]);
extern void usg_inetd (FILE *);
#endif

static const char hlp_telnetd[] = "telnet terminal server";
extern int com_telnetd (int, char *[]);
extern void usg_telnetd (FILE *);

static const char hlp_httpd[] = "http (web) server";
extern int com_httpd (int, char *[]);
extern void usg_httpd (FILE *);

static const char hlp_sh[] = "command line monitor";
extern int com_sh (int, char *[]);
extern void usg_sh (FILE *);

#define COMMAND(x)	{#x, hlp_##x, usg_##x, com_##x, 0}
#define BUILTIN(x)	{#x, hlp_##x, usg_##x, com_##x, 1}

static const struct comtab comtab[] =
{
    BUILTIN(help),
    COMMAND(sh),
    COMMAND(kick),
    COMMAND(netstat),
    COMMAND(ping),
#ifndef MINIMALIST
    COMMAND(ifconfig),
    COMMAND(sysctl),
    COMMAND(arp),
    COMMAND(route),
    COMMAND(inetd),
#endif
    COMMAND(httpd),
    COMMAND(telnetd),
    COMMAND(ps),
    BUILTIN(kill),
    BUILTIN(sleep),
    BUILTIN(debug),
    BUILTIN(hostname),
    BUILTIN(setenv),
    BUILTIN(unsetenv),
    BUILTIN(exit),
    {0},
};

#define IFS 	" \t\n"
#define MAXARG	32

static char    *
envrep (char *in)
{
    char           *envval;

    if (in && in[0] == '$' && (envval = getenv (in + 1))) {
	return envval;
    } else
	return in;
}


struct procargs {
    const struct comtab	*com;
    int			argc;
    char		**argv;
};


int
launchproc (void *arg)
{
    struct procargs pa;

    /* get copy of startup data and free */
    pa = *(struct procargs *)arg;
    free (arg);

    /* set program name for error messages */
    pthread_setname_np (pthread_self(), pa.com->name);

    /* create new process group and attach to terminal */
    setpgrp ();
    if (tcsetpgrp (STDOUT_FILENO, getpgrp()) < 0)
	err (EXIT_FAILURE, "tcsetpgrp");

    /* run the process */
    return (pa.com->command) (pa.argc, pa.argv);
}


static void
runcommand (const struct comtab *com, int argc, char *argv[])
{
    struct procargs *pa;
    int pid, status, errnum;
    pid_t res;

    if (com->builtin) {
	optreset = 1;	/* reset getopt() */
	(com->command) (argc, argv);
	return;
    }

    pa = malloc (sizeof (struct procargs));
    pa->com = com;
    pa->argc = argc;
    pa->argv = argv;

    errnum = pthread_proc_create_np (&pid, NULL, launchproc, pa);
    if (errnum) {
	free (pa);
	errno = errnum;
	warn ("proccreate");
	return;
    }

    /* wait for process to terminate */
    res = waitpid (pid, &status, 0);

    /* reset tty to our own process context (KLUDGE) */
    if (tcsetpgrp (STDOUT_FILENO, getpgrp()) < 0)
	warn ("tcsetpgrp");

    /* display any unusual status */
    if (res != pid) {
	if (res < 0)
	    warn ("waitpid");
	else
	    fprintf (stderr, "waitpid: strange result %d\n");
    }
    else {
	if (WIFSIGNALED (status)) {
	    int sig = WTERMSIG (status);
	    if (sig != SIGINT) {
		const char *sn;
		if (sig < NSIG)
		    sn = sys_siglist[sig];
		else
		    sn = "Unknown signal";
		fprintf (stderr, "%s\n", sn);
	    }
	}
	else {
	    assert (WIFEXITED (status));
	    if (WEXITSTATUS (status) != 0) {
		/* abandon script */
		if (scriptidx >= 0)
		    fprintf (stderr, "Exit %d - script abandoned\n",
			     WEXITSTATUS (status));
#if 0
		else
		    /* XXX only print the result for background jobs */
		    fprintf (stderr, "Exit %d\n", WEXITSTATUS (status));
#endif
		scriptidx = -1;
	    }
	}
    }
}


static int
cmdloop (char *prompt)
{
    while (1) {
	char            combuf[1024];
	char           *xargv[MAXARG];
	int             xargc = 0;
	int             i;
	char           *bp, *cp;

	if (prompt) {
	    fputs (prompt, stdout);
	    fflush (stdout);
	}

	if (scriptidx >= 0 && script[scriptidx]) {
	    fprintf (stdout, "%s\n", script[scriptidx]);
	    strcpy (combuf, script[scriptidx++]);
	    goto docmd;
	}
	scriptidx = -1;

	if (fgets (combuf, sizeof (combuf), stdin) == NULL)
	    return 0;

    docmd:
	/* break up to argc,argv */
	xargc = 0;
	bp = combuf;
	bp += strspn (bp, IFS);
	while ((cp = strsep (&bp, IFS)) && *cp)
	    xargv[xargc++] = envrep (cp);

	if (xargc == 0)
	    continue;

	xargv[xargc] = 0;

	/* XXX should copy argv[] to a malloced array if
	   we were to supoprt background jobs */

	for (i = 0; comtab[i].name; i++)
	    if (strcmp (xargv[0], comtab[i].name) == 0)
		break;

	if (comtab[i].name)
	    runcommand (&comtab[i], xargc, xargv);
	else
	    fprintf (stderr, "%s: command not found\n", xargv[0]);
    }
}


static pthread_key_t	jmpkey;
static pthread_once_t	jmponce = PTHREAD_ONCE_INIT;

static void
jmpinit (void)
{
    if (pthread_key_create (&jmpkey, free) != 0)
	errx (1, "cannot create jmpbuf key");
}


static void
intrhandler (int sig)
{
    sigjmp_buf *intrjb;
    scriptidx = -1;
    intrjb = pthread_getspecific (jmpkey);
    if (intrjb)
	siglongjmp (*intrjb, 1);
}


void
usg_sh (FILE *fp)
{
    fprintf (fp, "usage: sh\n");
}


/*
 * The "shell"
 */

int
com_sh (int argc, char **argv)
{
    struct termios tio;
    sigjmp_buf *intrjb;
    char *s;

    fprintf (stdout, "\n\nAlgNET monitor built %s %s\n", __DATE__, __TIME__);
    fprintf (stdout, "\nType \"help\" for command list; \"help COMMAND\" for usage.\n\n");

    /* little termios example */
    tcgetattr (STDOUT_FILENO, &tio);
    tio.c_cc[VERASE]	= 'h' & 037;	/* I prefer ^H for backspace */
    tio.c_cc[VKILL]	= 'x' & 037;	/* and ^X for line erase */
    /* disable job control */
    tio.c_cc[VSUSP]	= _POSIX_VDISABLE;
    tio.c_cc[VDSUSP]	= _POSIX_VDISABLE;
    if (tcsetattr (STDOUT_FILENO, TCSAFLUSH, &tio) < 0)
	err (EXIT_FAILURE, "tcsetattr");

    /* create thread-specific jmpbuf */
    pthread_once (&jmponce, jmpinit);
    intrjb = malloc (sizeof (sigjmp_buf));
    if (!intrjb)
	err (EXIT_FAILURE, "cannot malloc jmpbuf");
    if (pthread_setspecific (jmpkey, intrjb) != 0)
	errx (EXIT_FAILURE, "cannot set jmpbuf key");
    
    signal (SIGINT, intrhandler);
    if (sigsetjmp (*intrjb, 1)) 
	fprintf (stderr, "\nInterrupt\n");

#ifdef MINIMALIST
    return cmdloop ("netmin> ");
#else
    return cmdloop ("netmon> ");
#endif
}


int
main ()
{
    /* Run the shell */
    exit (com_sh (0, 0));
}
