/*-
 * netstat.c
 * Copyright (c) 2001 Algorithmicd Ltc
 *
 * Based on code from...
 * Copyright (c) 1998 Andrzej Bialecki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: netstat.c,v 1.3 2004/10/04 13:52:42 nigel Exp $
 */


/*
 * Small replacement for netstat. Uses only sysctl(3) to get the info.
 */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
/*#include <osreldate.h>*/
#include <err.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/icmp_var.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <net/route.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <ifaddrs.h>
#include <getopt.h>

static int iflag=0;
static int rflag=0;
static int sflag=0;

#ifndef optarg
extern char *optarg;
extern int optind;
#endif

#define plural(n)	((n) != 1 ? "s" : "")
#define plurales(n)	((n) != 1 ? "es" : "")

void
usg_netstat(FILE *fp)
{
	fprintf(fp,"usage: netstat [-r | -i | -s [-p proto] ]\n");
	fprintf(fp,"  proto: {ip|tcp|udp|icmp}\n");
}


static void
usage ()
{
    usg_netstat (stderr);
    exit (1);
}


/*
 * The following parts related to retrieving the routing table and
 * interface information, were borrowed from R. Stevens' code examples
 * accompanying his excellent book. Thanks!
 */

static char *
sock_ntop(const struct sockaddr *sa, size_t salen)
{
    char portstr[7];
    static char str[128];		/* Unix domain is largest */

	switch (sa->sa_family) {
	case 255: {
		struct sockaddr_in	*sin = (struct sockaddr_in *) sa;
		in_addr_t mask = ntohl (sin->sin_addr.s_addr);
		sprintf(str,"/%d", mask ? 33 - ffs(mask) : 0);
		return(str);
	}
        case AF_UNSPEC:
	case AF_INET: {
		struct sockaddr_in	*sin = (struct sockaddr_in *) sa;

		if (inet_ntop(AF_INET, &sin->sin_addr, str, sizeof(str)) == NULL)
			return(NULL);
		if (ntohs(sin->sin_port)!=0) {
			snprintf(portstr, sizeof(portstr), ".%d", ntohs(sin->sin_port));
			strlcat(str, portstr, sizeof(str));
		}
		if(strcmp(str,"0.0.0.0")==0) strcpy(str,"default");
		return(str);
	}
	case AF_UNIX: {
		struct sockaddr_un *unp = (struct sockaddr_un *) sa;

		/* OK to have no pathname bound to the socket: happens on
		   every connect() unless client calls bind() first. */
		if (unp->sun_path[0] == 0)
			strcpy(str, "(no pathname bound)");
		else
			snprintf(str, sizeof(str), "%s", unp->sun_path);
		return(str);
	}
	case AF_LINK: {
		struct sockaddr_dl *sdl = (struct sockaddr_dl *) sa;
		if (sdl->sdl_nlen > 0)
			snprintf(str, sizeof(str), "%*s",
				 sdl->sdl_nlen, &sdl->sdl_data[0]);
		else
			snprintf(str, sizeof(str), "link#%d", sdl->sdl_index);
		return(str);
	}
	default:
		snprintf(str, sizeof(str), "sock_ntop: unknown family: %d, len %d",
				 sa->sa_family, salen);
		return(str);
	}
    return (NULL);
}

static char *
Sock_ntop(const struct sockaddr *sa, size_t salen)
{
	char	*ptr;

	if ( (ptr = sock_ntop(sa, salen)) == NULL)
		err(1,"sock_ntop error");	/* inet_ntop() sets errno */
	return(ptr);
}


#define ROUNDUP(a,size) (((a) & ((size)-1))?(1+((a)|((size)-1))):(a))

#define NEXT_SA(ap)	ap=(struct sockaddr *) \
	((caddr_t)ap+(ap->sa_len?ROUNDUP(ap->sa_len,sizeof(u_long)):\
	sizeof(u_long)))

static void
get_rtaddrs(int addrs, struct sockaddr *sa, struct sockaddr **rti_info)
{
	int i;
	
	for(i=0;i<RTAX_MAX;i++) {
		if(addrs &(1<<i)) {
			rti_info[i]=sa;
			NEXT_SA(sa);
		} else
			rti_info[i]=NULL;
	}
}

/*
 * Definitions for showing gateway flags.
 */
const struct bits {
	short	b_mask;
	char	b_val;
} bits[] = {
	{ RTF_UP,	'U' },
	{ RTF_GATEWAY,	'G' },
	{ RTF_HOST,	'H' },
	{ RTF_REJECT,	'R' },
	{ RTF_BLACKHOLE, 'B' },
	{ RTF_DYNAMIC,	'D' },
	{ RTF_MODIFIED,	'M' },
	{ RTF_DONE,	'd' }, /* Completed -- for routing messages only */
	{ RTF_MASK,	'm' }, /* Mask Present -- for routing messages only */
	{ RTF_CLONING,	'C' },
	{ RTF_XRESOLVE,	'X' },
	{ RTF_LLINFO,	'L' },
	{ RTF_STATIC,	'S' },
	{ RTF_PROTO1,	'1' },
	{ RTF_PROTO2,	'2' },
	{ 0 }
};

static void
get_flags(char *buf, int flags)
{
    const struct bits *b;
    for (b = bits; b->b_mask; b++)
	if (flags & b->b_mask)
	    *buf++ = b->b_val;
    *buf = '\0';
}

static void
print_interfaces ()
{
    struct ifaddrs *iflist, *ifa;
    char addr[21];

    if (getifaddrs (&iflist) < 0)
	err (1, "getifaddrs");

    printf("\nInterface table:\n");
    printf("----------------\n");
    printf("%-4s %-20s ", 
	   "Name", "Address");
    if (sflag) {
	printf("%5s%5s%9s%6s%9s%6s%6s%6s\n",
	       "MTU", "Bps", "Ipkts", "Ierrs", "Opkts", "Oerrs", "Coll", "Drop");
    }
    else {
	printf("Network\n");
    }

    
    for (ifa = iflist; ifa != NULL; ifa = ifa->ifa_next) {
	if (!ifa->ifa_addr)
	    continue;
	if (ifa->ifa_addr->sa_family == AF_LINK) {
	    struct sockaddr_dl *sdl = (struct sockaddr_dl *) ifa->ifa_addr;
	    printf("%-4s ", ifa->ifa_name);
	    addr[0] = '\0';
	    /* XXX Ethernet specific */
	    if (sdl->sdl_type == IFT_ETHER && sdl->sdl_alen == 6) {
		unsigned char *p = LLADDR(sdl);
		snprintf(addr, sizeof(addr), 
			 "%02x:%02x:%02x:%02x:%02x:%02x",
			 p[0], p[1], p[2], p[3], p[4], p[5] );
	    }
	    /* XXX handle other link types */
	    printf("%-20s ", addr);
	    if (sflag && ifa->ifa_data) {
		struct if_data *ifi = ifa->ifa_data;
		u_long speed = ifi->ifi_baudrate;
		char c;
		if (speed >= IF_Gbps(1)) {
		    speed /= IF_Gbps(1);
		    c = 'G';
		} else if (speed >= IF_Mbps(1)) {
		    speed /= IF_Mbps(1);
		    c = 'M';
		} else if (speed >= IF_Kbps(1)) {
		    speed /= IF_Kbps(1);
		    c = 'K';
		} else
		    c = ' ';
		printf("%5lu%4lu%c%9lu%6lu%9lu%6lu%6lu%6lu",
		       ifi->ifi_mtu,
		       speed, c,
		       ifi->ifi_ipackets,
		       ifi->ifi_ierrors,
		       ifi->ifi_opackets,
		       ifi->ifi_oerrors,
		       ifi->ifi_collisions,
		       ifi->ifi_iqdrops);
	    }
	    printf("\n");

	}
	else if (ifa->ifa_addr->sa_family == AF_INET && !sflag) {
	    struct sockaddr_in *sin = (struct sockaddr_in *)ifa->ifa_addr;
	    printf("%-4s ", ifa->ifa_name);
	    addr[0] = '\0';
	    (void) inet_ntop(AF_INET, &sin->sin_addr, addr, sizeof(addr));
	    printf("%-20s ", addr);
	    if (ifa->ifa_netmask) {
		struct sockaddr_in *snm = (struct sockaddr_in *)ifa->ifa_netmask;
		in_addr_t mask = ntohl (snm->sin_addr.s_addr);
		int mbits = mask ? 33 - ffs(mask) : 0;
		if (inet_net_ntop(AF_INET, &sin->sin_addr, mbits,
				  addr, sizeof(addr)))
		    printf("%s", addr);
	    }
	    printf("\n");
	}
	/* XXX handle other protocols */
    }

    freeifaddrs(iflist);
}



static void
print_routing(char *proto)
{
	int mib[6],i=0;
	char *rt_buf = NULL,*if_buf = NULL;
	char *next,*lim;
	size_t rt_len, if_len;
	struct rt_msghdr *rtm;
	struct if_msghdr *ifm,**ifm_table;
	struct ifa_msghdr *ifam;
	struct sockaddr *sa,*sa1,*rti_info[RTAX_MAX],**if_table;
	struct rt_metrics rm;
	int if_num;
	char fbuf[50];

	/* Get the routing table */
	mib[0]=CTL_NET;
	mib[1]=PF_ROUTE;
	mib[2]=0;
	mib[3]=0;
	mib[4]=NET_RT_DUMP;
	mib[5]=0;
	/* Estimate the size of table */
	if(sysctl(mib,6,NULL,&rt_len,NULL,0)==-1) {
	    warn("sysctl size");
	    goto done;
	}
	if((rt_buf=(char *)malloc(rt_len))==NULL) {
	    warn("malloc");
	    goto done;
	}
	/* Now get it. */
	if(sysctl(mib,6,rt_buf,&rt_len,NULL,0)==-1) {
	    warn("sysctl get");
	    goto done;
	}

	/* Get the interfaces table */
	mib[0]=CTL_NET;
	mib[1]=PF_ROUTE;
	mib[2]=0;
	mib[3]=0;
	mib[4]=NET_RT_IFLIST;
	mib[5]=0;
	/* Estimate the size of table */
	if(sysctl(mib,6,NULL,&if_len,NULL,0)==-1) {
	    warn("sysctl size");
	    goto done;
	}
	if((if_buf=(char *)malloc(if_len))==NULL) {
	    warn("malloc");
	    goto done;
	}
	/* Now get it. */
	if(sysctl(mib,6,if_buf,&if_len,NULL,0)==-1) {
	    warn("sysctl get");
	    goto done;
	}
	lim=if_buf+if_len;
	i=0;
	for(next=if_buf, i = 0;next<lim;next+=ifm->ifm_msglen, i++) {
		ifm=(struct if_msghdr *)next;
		i++;
	}
	if_num=i;
	if_table=(struct sockaddr **)calloc(i,sizeof(struct sockaddr));
	ifm_table=(struct if_msghdr **)calloc(i,sizeof(struct if_msghdr));

	i=0;
	for(next=if_buf, i = 0;next<lim;next+=ifm->ifm_msglen, i++) {
		ifm=(struct if_msghdr *)next;
		if_table[i]=(struct sockaddr *)(ifm+1);
		ifm_table[i]=ifm;
	}

	/* Now dump the routing table */
	printf("\nRouting table:\n");
	printf("--------------\n");
	printf("Destination        Gateway            Flags       Netif  Use\n");
	lim=rt_buf+rt_len;
	for(next=rt_buf;next<lim;next+=rtm->rtm_msglen) {
		rtm=(struct rt_msghdr *)next;
		sa=(struct sockaddr *)(rtm+1);
		get_rtaddrs(rtm->rtm_addrs,sa,rti_info);
#ifdef RTF_WASCLONED
		if(rtm->rtm_flags & RTF_WASCLONED
		   && (rtm->rtm_flags & RTF_LLINFO)==0)
		    continue;
#endif
		if((sa=rti_info[RTAX_DST])!=NULL) {
			sprintf(fbuf,"%s",sock_ntop(sa,sa->sa_len));
			if(((sa1=rti_info[RTAX_NETMASK])!=NULL) && sa1->sa_family==255) {
				strcat(fbuf,sock_ntop(sa1,sa1->sa_len));
			}
			printf("%-19s",fbuf);
		}
		if((sa=rti_info[RTAX_GATEWAY])!=NULL) {
			printf("%-19s",sock_ntop(sa,sa->sa_len));
		}
		memset(fbuf,0,sizeof(fbuf));
		get_flags(fbuf,rtm->rtm_flags);
		printf("%-10s",fbuf);
		for(i=0;i<if_num;i++) {
			ifm=ifm_table[i];
			if((ifm->ifm_index==rtm->rtm_index) &&
				(ifm->ifm_data.ifi_type>0)) {
				sa=if_table[i];
				break;
			}
		}
		if(ifm->ifm_type==RTM_IFINFO) {
			get_rtaddrs(ifm->ifm_addrs,sa,rti_info);
			printf("  %s",Sock_ntop(sa,sa->sa_len));
		} else if(ifm->ifm_type==RTM_NEWADDR) {
			ifam=(struct ifa_msghdr *)ifm_table[rtm->rtm_index-1];
			sa=(struct sockaddr *)(ifam+1);
			get_rtaddrs(ifam->ifam_addrs,sa,rti_info);
			printf("  %s",Sock_ntop(sa,sa->sa_len));
		}
		printf("    %u",rtm->rtm_use);
		printf("\n");
	}

done:	
	free (rt_buf);
	free (if_buf);
}



static int
print_ip_stats()
{
	int mib[4];
	size_t len;
	struct ipstat ipstat;

	mib[0]=CTL_NET;
	mib[1]=PF_INET;
	mib[2]=IPPROTO_IP;
	mib[3]=IPCTL_STATS;
	len=sizeof(struct ipstat);
	if(sysctl(mib,4,&ipstat,&len,NULL,0)<0) {
		perror("sysctl");
		return(-1);
	}
	printf("\nIP statistics:\n");
	printf("--------------\n");
#define	p(f, m) if (ipstat.f || sflag <= 1) \
    printf(m, ipstat.f, plural(ipstat.f))
#define	p1(f, m) if (ipstat.f || sflag <= 1) \
    printf(m, ipstat.f)

	p(ips_total, "\t%lu total packet%s received\n");
	p(ips_badsum, "\t%lu bad header checksum%s\n");
	p1(ips_toosmall, "\t%lu with size smaller than minimum\n");
	p1(ips_tooshort, "\t%lu with data size < data length\n");
	p1(ips_badhlen, "\t%lu with header length < data size\n");
	p1(ips_badlen, "\t%lu with data length < header length\n");
	p1(ips_badoptions, "\t%lu with bad options\n");
	p1(ips_badvers, "\t%lu with incorrect version number\n");
	p(ips_fragments, "\t%lu fragment%s received\n");
	p(ips_fragdropped, "\t%lu fragment%s dropped (dup or out of space)\n");
	p(ips_badfrags, "\t%lu malformed fragment%s dropped\n");
	p(ips_fragtimeout, "\t%lu fragment%s dropped after timeout\n");
	p(ips_reassembled, "\t%lu packet%s reassembled ok\n");
	p(ips_delivered, "\t%lu packet%s for this host\n");
	p(ips_noproto, "\t%lu packet%s for unknown/unsupported protocol\n");
	p(ips_forward, "\t%lu packet%s forwarded\n");
	p(ips_cantforward, "\t%lu packet%s not forwardable\n");
	p(ips_redirectsent, "\t%lu redirect%s sent\n");
	p(ips_localout, "\t%lu packet%s sent from this host\n");
	p(ips_rawout, "\t%lu packet%s sent with fabricated ip header\n");
	p(ips_odropped, "\t%lu output packet%s dropped due to no bufs, etc.\n");
	p(ips_noroute, "\t%lu output packet%s discarded due to no route\n");
	p(ips_fragmented, "\t%lu output datagram%s fragmented\n");
	p(ips_ofragments, "\t%lu fragment%s created\n");
	p(ips_cantfrag, "\t%lu datagram%s that can't be fragmented\n");
	p1(ips_rcvmemdrop, "\t%lu fragment floods\n");
	p(ips_toolong, "\t%lu packet%s with ip length > max ip packet size\n");
	p(ips_nogif, "\t%lu tunneling packet%s that can't find gif\n");
#undef p
#undef p1
}

static int
print_tcp_stats()
{
	int mib[4];
	size_t len;
	struct tcpstat tcpstat;

	mib[0]=CTL_NET;
	mib[1]=PF_INET;
	mib[2]=IPPROTO_TCP;
	mib[3]=TCPCTL_STATS;
	len=sizeof(struct tcpstat);
	if(sysctl(mib,4,&tcpstat,&len,NULL,0)<0) {
		perror("sysctl");
		return(-1);
	}
	printf("\nTCP statistics:\n");
	printf("---------------\n");
#define	p(f, m) if (tcpstat.f || sflag <= 1) \
    printf(m, tcpstat.f, plural(tcpstat.f))
#define	p1(f, m) if (tcpstat.f || sflag <= 1) \
    printf(m, tcpstat.f)
#define	p2(f1, f2, m) if (tcpstat.f1 || tcpstat.f2 || sflag <= 1) \
    printf(m, tcpstat.f1, plural(tcpstat.f1), tcpstat.f2, plural(tcpstat.f2))
#define	p2a(f1, f2, m) if (tcpstat.f1 || tcpstat.f2 || sflag <= 1) \
    printf(m, tcpstat.f1, plural(tcpstat.f1), tcpstat.f2)
#define	p3(f, m) if (tcpstat.f || sflag <= 1) \
    printf(m, tcpstat.f, plurales(tcpstat.f))

	p(tcps_sndtotal, "\t%u packet%s sent\n");
	p2(tcps_sndpack,tcps_sndbyte,
		"\t\t%u data packet%s (%qd byte%s)\n");
	p2(tcps_sndrexmitpack, tcps_sndrexmitbyte,
		"\t\t%u data packet%s (%qd byte%s) retransmitted\n");
	p(tcps_sndrexmitfast, "\t\t%qd fast retransmitted packet%s\n");
	p2a(tcps_sndacks, tcps_delack,
		"\t\t%u ack-only packet%s (%u delayed)\n");
	p(tcps_sndurg, "\t\t%u URG only packet%s\n");
	p(tcps_sndprobe, "\t\t%u window probe packet%s\n");
	p(tcps_sndwinup, "\t\t%u window update packet%s\n");
	p(tcps_sndctrl, "\t\t%u control packet%s\n");
	p(tcps_rcvtotal, "\t%u packet%s received\n");
	p2(tcps_rcvackpack, tcps_rcvackbyte, "\t\t%u ack%s (for %qd byte%s)\n");
	p(tcps_rcvdupack, "\t\t%u duplicate ack%s\n");
	p(tcps_rcvacktoomuch, "\t\t%u ack%s for unsent data\n");
	p2(tcps_rcvpack, tcps_rcvbyte,
		"\t\t%u packet%s (%qu byte%s) received in-sequence\n");
	p2(tcps_rcvduppack, tcps_rcvdupbyte,
		"\t\t%u completely duplicate packet%s (%qd byte%s)\n");
	p(tcps_pawsdrop, "\t\t%u old duplicate packet%s\n");
	p2(tcps_rcvpartduppack, tcps_rcvpartdupbyte,
		"\t\t%u packet%s with some dup. data (%qd byte%s duped)\n");
	p2(tcps_rcvoopack, tcps_rcvoobyte,
		"\t\t%u out-of-order packet%s (%qd byte%s)\n");
	p2(tcps_rcvpackafterwin, tcps_rcvbyteafterwin,
		"\t\t%u packet%s (%qd byte%s) of data after window\n");
	p(tcps_rcvwinprobe, "\t\t%u window probe%s\n");
	p(tcps_rcvwinupd, "\t\t%u window update packet%s\n");
	p(tcps_rcvafterclose, "\t\t%u packet%s received after close\n");
	p(tcps_rcvbadsum, "\t\t%u discarded for bad checksum%s\n");
	p(tcps_rcvbadoff, "\t\t%u discarded for bad header offset field%s\n");
	p1(tcps_rcvshort, "\t\t%u discarded because packet too short\n");
	p1(tcps_rcvnosec, "\t\t%u discarded for missing IPSec protection\n");
	p(tcps_connattempt, "\t%u connection request%s\n");
	p(tcps_accepts, "\t%u connection accept%s\n");
	p(tcps_connects, "\t%u connection%s established (including accepts)\n");
	p2(tcps_closed, tcps_drops,
		"\t%u connection%s closed (including %u drop%s)\n");
	p(tcps_conndrops, "\t%u embryonic connection%s dropped\n");
	p2(tcps_rttupdated, tcps_segstimed,
		"\t%u segment%s updated rtt (of %u attempt%s)\n");
	p(tcps_rexmttimeo, "\t%u retransmit timeout%s\n");
	p(tcps_timeoutdrop, "\t\t%u connection%s dropped by rexmit timeout\n");
	p(tcps_persisttimeo, "\t%u persist timeout%s\n");
	p(tcps_keeptimeo, "\t%u keepalive timeout%s\n");
	p(tcps_keepprobe, "\t\t%u keepalive probe%s sent\n");
	p(tcps_keepdrops, "\t\t%u connection%s dropped by keepalive\n");
	p(tcps_predack, "\t%u correct ACK header prediction%s\n");
	p(tcps_preddat, "\t%u correct data packet header prediction%s\n");
	p3(tcps_pcbhashmiss, "\t%u PCB cache miss%s\n");
	p(tcps_badsyn, "\t%u SYN packet%s received with same src/dst address/port\n");
#undef p
#undef p1
#undef p2
#undef p2a
#undef p3
}

static int
print_udp_stats()
{
	int mib[4];
	struct udpstat udpstat;
	u_int32_t delivered;
	size_t len;

	mib[0]=CTL_NET;
	mib[1]=PF_INET;
	mib[2]=IPPROTO_UDP;
	mib[3]=UDPCTL_STATS;
	len=sizeof(struct udpstat);
	if(sysctl(mib,4,&udpstat,&len,NULL,0)<0) {
		perror("sysctl");
		return(-1);
	}
	printf("\nUDP statistics:\n");
	printf("---------------\n");
#define	p(f, m) if (udpstat.f || sflag <= 1) \
    printf(m, udpstat.f, plural(udpstat.f))
#define	p1(f, m) if (udpstat.f || sflag <= 1) \
    printf(m, udpstat.f)
	p(udps_ipackets, "\t%lu datagram%s received\n");
	p1(udps_hdrops, "\t%lu with incomplete header\n");
	p1(udps_badlen, "\t%lu with bad data length field\n");
	p1(udps_badsum, "\t%lu with bad checksum\n");
	p1(udps_nosum, "\t%lu with no checksum\n");
	p1(udps_noport, "\t%lu dropped due to no socket\n");
	p(udps_noportbcast, "\t%lu broadcast/multicast datagram%s dropped due to no socket\n");
	p1(udps_nosec, "\t%lu dropped due to missing IPSec protection\n");
	p1(udps_fullsock, "\t%lu dropped due to full socket buffers\n");
	delivered = udpstat.udps_ipackets -
		    udpstat.udps_hdrops -
		    udpstat.udps_badlen -
		    udpstat.udps_badsum -
		    udpstat.udps_noport -
		    udpstat.udps_noportbcast -
		    udpstat.udps_fullsock;
	if (delivered || sflag <= 1)
		printf("\t%lu delivered\n", delivered);
	p(udps_opackets, "\t%lu datagram%s output\n");
	p1(udps_pcbhashmiss, "\t%lu missed PCB cache\n");
#undef p
#undef p1
}

static const char * const icmpnames[]={
	"echo reply",
	"#1",
	"#2",
	"destination unreachable",
	"source quench",
	"routing redirect",
	"#6",
	"#7",
	"echo",
	"router advertisement",
	"router solicitation",
	"time exceeded",
	"parameter problem",
	"time stamp",
	"time stamp reply",
	"information request",
	"information request reply",
	"address mask request",
	"address mask reply",
};

static int
print_icmp_stats()
{
	int mib[4],first, i;
	size_t len;
	struct icmpstat icmpstat;

	mib[0]=CTL_NET;
	mib[1]=PF_INET;
	mib[2]=IPPROTO_ICMP;
	mib[3]=ICMPCTL_STATS;
	len=sizeof(struct icmpstat);
	if(sysctl(mib,4,&icmpstat,&len,NULL,0)<0) {
		perror("sysctl");
		return(-1);
	}
	printf("\nICMP statistics:\n");
	printf("----------------\n");
#define	p(f, m) if (icmpstat.f || sflag <= 1) \
    printf(m, icmpstat.f, plural(icmpstat.f))

	p(icps_error, "\t%lu call%s to icmp_error\n");
	p(icps_oldicmp,
	    "\t%lu error%s not generated 'cuz old message was icmp\n");
	for (first = 1, i = 0; i < ICMP_MAXTYPE + 1; i++)
		if (icmpstat.icps_outhist[i] != 0) {
			if (first) {
				printf("\tOutput packet histogram:\n");
				first = 0;
			}
			printf("\t\t%s: %lu\n", icmpnames[i],
				icmpstat.icps_outhist[i]);
		}
	p(icps_badcode, "\t%lu message%s with bad code fields\n");
	p(icps_tooshort, "\t%lu message%s < minimum length\n");
	p(icps_checksum, "\t%lu bad checksum%s\n");
	p(icps_badlen, "\t%lu message%s with bad length\n");
	for (first = 1, i = 0; i < ICMP_MAXTYPE + 1; i++)
		if (icmpstat.icps_inhist[i] != 0) {
			if (first) {
				printf("\tInput packet histogram:\n");
				first = 0;
			}
			printf("\t\t%s: %lu\n", icmpnames[i],
				icmpstat.icps_inhist[i]);
		}
	p(icps_reflect, "\t%lu message response%s generated\n");
#undef p
}

int
stats(char *proto)
{
	if (!sflag)
		return 0;
	if (proto) {
		if(strcmp(proto,"ip")==0) print_ip_stats();
		if(strcmp(proto,"icmp")==0) print_icmp_stats();
		if(strcmp(proto,"udp")==0) print_udp_stats();
		if(strcmp(proto,"tcp")==0) print_tcp_stats();
		return(0);
	}
	print_ip_stats();
	print_icmp_stats();
	print_udp_stats();
	print_tcp_stats();
	return(0);
}

int
#ifdef unix
main (int argc, char *argv[])
#else
com_netstat (int argc, char *argv[])
#endif
{
	int c;
	char *proto=NULL;

#ifndef unix
	iflag = rflag = sflag = 0;
#endif
	while((c=getopt(argc,argv,"irsp:"))!=-1) {
		switch(c) {
		case 'r':
			rflag++;
			break;
		case 'i':
			iflag++;
			break;
		case 's':
			sflag++;
			rflag=0;
			break;
		case 'p':
			sflag++;
			proto=optarg;
			break;
		case '?':
		default:
			usage();
			break;
		}
	}
	argc-=optind;

	if(argc>0)
		usage();

	if (rflag == 0 && sflag == 0 && iflag == 0)
		rflag = 1 ;
	if (iflag)
	    print_interfaces(proto);
	if (rflag)
	    print_routing(proto);
	if (sflag && !iflag)
	    stats(proto);
	exit(0);
}

