/*
 * telnetd.c: simple multi-threaded TELNET daemon for AlgNET
 *
 * Copyright (c) 2001 Algorithmics Ltd - all rights reserved.
 * 
 * Loosely based on 4.3BSD telnet daemon which bears the following
 * copyright message... 
 
 * Copyright (c) 1983, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * TODO
 *
 * o Ask user for password before starting shell
 * o Handle urgent socket data transmission (for flush etc)
 * o Catch child death (SIGCHLD) and shutdown socket
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/queue.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>
#include <syslog.h>
#include <termios.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <paths.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <getopt.h>

/*#define DEBUG*/

extern int com_sh (int, char *[]);
static char * argv_init[] = { "sh", NULL };


/* command line argments for telnet instance */
struct telnetopts {
    int			debug;		/* debug flag */
    int			portnbr;	/* port to listen on */
    char *		password;	/* selected password */
    LIST_HEAD(,connection) connections;	/* connection list */
};



#define BUFSIZE		2048		/* io buffer size */
#define SUBSIZE		128		/* sub option size */


/* circular buffer with separate in/out pointers */
struct circbuf {
    unsigned char *	buf;		/* data buffer */
    unsigned int	size;		/* buffer size */
    unsigned int 	rfirst;		/* read index */
    unsigned int	wfirst;		/* write index */
    unsigned int	count;		/* count of bytes */
    int			urg;		/* urgent index */
};

/* option states */
#define	OPT_NO			0		/* won't do this option */
#define	OPT_YES			1		/* will do this option */
#define	OPT_YES_BUT_ALWAYS_LOOK	2
#define	OPT_NO_BUT_ALWAYS_LOOK	3

/* per connection structure */
struct connection {
    LIST_ENTRY(connection) link;
    
    struct telnetopts	*topt;		/* current options */

    char		tty_name[TTY_NAME_MAX];
    int			netfd;
    int			ptyfd;
    int			pid;
    int			not42;

    struct circbuf	ptybuf;		/* data to be sent to pty */
    struct circbuf	netbuf;		/* data to be sent to net */
    
    /* telnet protocol parser state */
    int			state;
    struct circbuf	subopt;		/* suboptions */
    char		hisopts[256];
    char		myopts[256];

#ifdef TELNET_ENVIRON
    /*
     * The following are some clocks used to decide how to interpret
     * the relationship between various variables. We only need
     * it if we want to retrieve environment variables from the
     * remote client.
     */
    struct {
	int system,		/* what the current time is */
	    echotoggle,		/* last time user entered echo character */
	    modenegotiated,	/* last time operating mode negotiated */
	    didnetreceive,	/* last time we read data from network */
	    ttypeopt,		/* ttype will/won't received */
	    ttypesubopt,	/* ttype subopt is received */
	    getterminal,	/* time started to get terminal information */
	    gotDM;		/* when did we last see a data mark */
    } clocks;
#define	settimer(c,x)		(c->clocks.x = ++c->clocks.system)
#define	sequenceIs(c,x,y)	(c->clocks.x < c->clocks.y)
#else
#define	settimer(c,x)
#endif
};


/* 
 * Circular buffer manipulation
 */
static void
circ_init (struct circbuf *cb, unsigned int size)
{
    cb->buf = malloc (size);
    if (!cb->buf) {
	syslog (LOG_ERR, "malloc: %m");
	exit (1);
    }
    cb->rfirst = cb->wfirst = cb->count = 0;
    cb->size = size;
    cb->urg = -1;
}


static void
circ_free (struct circbuf *cb)
{
    if (cb->buf)
	free (cb->buf);
}


static void
circ_reset (struct circbuf *cb)
{
    cb->rfirst = cb->wfirst = cb->count = 0;
    cb->urg = -1;
}


static void
circ_putbuf (struct circbuf *cb, const char *bp, unsigned int n)
{
    assert (cb->count + n <= cb->size);
    while (n != 0) {
	unsigned int r = cb->size - cb->wfirst;
	if (r > n)
	    r = n;
	memcpy (&cb->buf[cb->wfirst], bp, r);
	if ((cb->wfirst += r) == cb->size)
	    cb->wfirst = 0;
	cb->count += r;
	bp += r;
	n -= r;
    }
}


static void
circ_putc (struct circbuf *cb, unsigned int b)
{
    assert (cb->count < cb->size);
    cb->buf[cb->wfirst] = b;
    if (++cb->wfirst == cb->size)
	cb->wfirst = 0;
    ++cb->count;
}


static void
circ_seturg (struct circbuf *cb)
{
    assert (cb->count > 0);
    if (cb->wfirst == 0)
	cb->urg = cb->size - 1;
    else
	cb->urg = cb->wfirst - 1;
}

static int
circ_empty (struct circbuf *cb)
{
    return cb->count == 0;
}

static int
circ_full (struct circbuf *cb)
{
    return cb->size == cb->count;
}

static unsigned
circ_space (struct circbuf *cb)
{
    return cb->size - cb->count;
}


static int
circ_getbuf (struct circbuf *cb, unsigned char **bp, unsigned int *np)
{
    unsigned int n = cb->size - cb->rfirst;

    if (n > cb->count)
	n = cb->count;

    *bp = &cb->buf[cb->rfirst];
    *np = n;
    return n != 0;
}

static void
circ_remove (struct circbuf *cb, unsigned int n)
{
    if ((cb->count -= n) == 0)
	/* empty, reset pointers */
	cb->rfirst = cb->wfirst = 0;
    else {
	cb->rfirst += n;
	if (cb->rfirst == cb->size)
	    cb->rfirst = 0;
    }
    assert (cb->rfirst < cb->size);
}


static int
circ_getc (struct circbuf *cb)
{
    int c;

    if (cb->count == 0)
	return -1;
    c = cb->buf[cb->rfirst];

    if (--cb->count == 0)
	cb->rfirst = cb->wfirst = 0;
    else {
	if (++cb->rfirst == cb->size)
	    cb->rfirst = 0;
    }
    return c;
}



static void destroy_connection(struct connection *c);

static void
doflush (struct connection *c, struct circbuf *cb, int fd, const char *name)
{
    unsigned char *bp;
    unsigned int nb;
    ssize_t nw;

/* XXX Need to handle urgent network data here! */

    while (circ_getbuf (cb, &bp, &nb)) {
	nw = write (fd, bp, nb);
	if (nw < 0) {
	    if (errno == EINTR)
		continue;
	    if (errno == EWOULDBLOCK)
		return;
	    syslog (LOG_WARNING, "write(%s): %m", name);
	    destroy_connection (c);
	    return;
	}
	circ_remove (cb, nw);
    }
}


static void
netflush (struct connection *c)
{
    doflush (c, &c->netbuf, c->netfd, "net");
}


static void
ptyflush (struct connection *c)
{
    doflush (c, &c->ptybuf, c->ptyfd, "pty");
}



/*
 * nextitem()
 *
 *	Return the index of the next "item" in the TELNET data
 * stream.  This will be the indext of the next character if
 * the current index is a user data character, or it will
 * be the index of the character following the TELNET command
 * if the current index is a TELNET IAC ("I Am a Command")
 * character.
 */

static unsigned int 
nextitem (struct circbuf *cb, unsigned int peek)
{
    unsigned char *bp = &cb->buf[peek];

    if (bp[0] != IAC)
	return peek + 1;

    switch (bp[1]) {
    case DO:
    case DONT:
    case WILL:
    case WONT:
	return peek + 3;
    case SB:		/* loop forever looking for the SE */
	{
	    unsigned int i = 2;
	    for (;;) {
		if (bp[i++] == IAC) {
		    if (bp[i++] == SE) {
			return peek + i;
		    }
		}
	    }
	}
    default:
	return peek + 2;
    }
}


/*
 * net_clear()
 *
 *	We are about to do a TELNET SYNCH operation.  Clear
 * the path to the network.
 *
 *	Things are a bit tricky since we may have sent the first
 * byte or so of a previous TELNET command into the network.
 * So, we have to scan the network buffer from the beginning
 * until we are up to where we want to be.
 *
 *	A side effect of what we do, just to keep things
 * simple, is to clear the urgent data pointer.  The principal
 * caller should be setting the urgent data pointer AFTER calling
 * us in any case.
 */

static void
net_clear (struct connection *c)
{
    struct circbuf *cb = &c->netbuf;
    unsigned int peek, next, lim;

    if (cb->wfirst <= cb->rfirst) {
	/* XXX write pointer has wrapped around, or all sent */
	return;
    }

    /* scan from beginning */
    peek = 0;		

    /* ... to read pointer */
    lim = cb->rfirst;
    while (lim > (next = nextitem (cb, peek))) {
	peek = next;
    }

    /* now copy up to write pointer */
    lim = cb->wfirst;

    /* into a clean buffer */
    circ_reset (cb);

#define	wewant(i) \
    (cb->buf[i] == IAC && (cb->buf[i+1] != EC && cb->buf[i+1] != EL))

    while (lim > peek) {
	if (wewant (peek)) {
	    unsigned int length;
	    next = peek;
	    do {
		next = nextitem (cb, next);
	    } while (lim > next && wewant (next));
	    length = next - peek;
	    circ_putbuf (cb, &cb->buf[peek], length);
	    peek = next;
	} else {
	    peek = nextitem (cb, peek);
	}
    }
#undef wewant
}



static void 
pty_setbinaryin (struct connection *c, int on)
{
    struct termios termbuf;

    ptyflush (c);
    tcgetattr (c->ptyfd, &termbuf);
    if (on) {
	termbuf.c_iflag &= ~ISTRIP;
    }
    else {
	termbuf.c_iflag |= ISTRIP;
    }
    tcsetattr (c->ptyfd, TCSANOW, &termbuf);
}


static void 
pty_setbinaryout (struct connection *c, int on)
{
    struct termios termbuf;

    ptyflush (c);
    tcgetattr (c->ptyfd, &termbuf);
    if (on) {
	termbuf.c_cflag &= ~(CSIZE|PARENB);
	termbuf.c_cflag |= CS8;
	termbuf.c_oflag &= ~OPOST;
    }
    else {
	termbuf.c_cflag &= ~CSIZE;
	termbuf.c_cflag |= CS7|PARENB;
	termbuf.c_oflag |= OPOST;
    }
    tcsetattr (c->ptyfd, TCSANOW, &termbuf);
}

static void
pty_setecho (struct connection *c, int on)
{
    struct termios termbuf;

    ptyflush (c);
    tcgetattr (c->ptyfd, &termbuf);
    if (on)
	termbuf.c_lflag |= ECHO;
    else
	termbuf.c_lflag &= ~ECHO;
    tcsetattr (c->ptyfd, TCSANOW, &termbuf);
}


static void
pty_sendinterrupt (struct connection *c)
{
    static const int sig = SIGINT;
    ptyflush (c);	/* half-hearted */
    (void) ioctl (c->ptyfd, TIOCSIG, &sig);
}


static void
pty_sendbreak (struct connection *c)
{
    static const int sig = SIGQUIT;
    ptyflush (c);	/* half-hearted */
    (void) ioctl (c->ptyfd, TIOCSIG, &sig);
}


static void
pty_sendcc (struct connection *c, int cc)
{
    struct termios termbuf;

    ptyflush (c);	/* half-hearted */
    tcgetattr (c->ptyfd, &termbuf);
    if (termbuf.c_cc[cc] != _POSIX_VDISABLE)
	circ_putc (&c->ptybuf, termbuf.c_cc[cc]);
}



static void
send_escape (struct connection *c, int opt1, int opt2)
{
    circ_putc (&c->netbuf, IAC);
    circ_putc (&c->netbuf, opt1);
    circ_putc (&c->netbuf, opt2);
}

static void
send_dont (struct connection *c, int option)
{
    send_escape (c, DONT, option);
}

static void
send_do (struct connection *c, int option)
{
    send_escape (c, DO, option);
}

static void
send_will (struct connection *c, int option)
{
    send_escape (c, WILL, option);
}


static void
send_wont (struct connection *c, int option)
{
    send_escape (c, WONT, option);
}



static void
willoption (struct connection *c, int option)
{
    int doit = OPT_NO;

    switch (option) {

    case TELOPT_BINARY:
	pty_setbinaryin (c, 1);
	doit = OPT_YES;
	break;
	
    case TELOPT_ECHO:
	c->not42 = 0;		/* looks like a 4.2 system */
	/*
	 * Now, in a 4.2 system, to break them out of ECHOing
	 * (to the terminal) mode, we need to send a "WILL ECHO".
	 * Kludge upon kludge!
	 */
	if (c->myopts[TELOPT_ECHO] == OPT_YES)
	    send_do (c, TELOPT_ECHO);
	doit = OPT_NO;
	break;
	
    case TELOPT_TTYPE:
	settimer (c, ttypeopt);
	if (c->hisopts[TELOPT_TTYPE] == OPT_YES_BUT_ALWAYS_LOOK) {
	    c->hisopts[TELOPT_TTYPE] = OPT_YES;
	    return;
	}
	doit = OPT_YES;
	break;
	
    case TELOPT_SGA:
	doit = OPT_YES;
	break;
	
    case TELOPT_TM:
	doit = OPT_NO;
	break;
	
    default:
	doit = OPT_NO;
	break;
    }

    c->hisopts[option] = doit;
    if (doit == OPT_YES)
	send_do (c, option);
    else
	send_dont (c, option);
}


static void
wontoption (struct connection *c, int option)
{
    switch (option) {
    case TELOPT_ECHO:
	c->not42 = 1;		/* doesn't seem to be a 4.2 system */
	break;
	
    case TELOPT_BINARY:
	pty_setbinaryin (c, 0);
	break;
	
    case TELOPT_TTYPE:
	settimer (c, ttypeopt);
	break;
    }

    c->hisopts[option] = OPT_NO;
    send_dont (c, option);
}

static void
dooption (struct connection *c, int option)
{
    char will = OPT_NO;

    switch (option) {
    case TELOPT_TM:
	will = OPT_NO;
	break;

    case TELOPT_ECHO:
	pty_setecho (c, 1);
	will = OPT_YES;
	break;

    case TELOPT_BINARY:
	pty_setbinaryout (c, 1);
	will = OPT_YES;
	break;

    case TELOPT_SGA:
	will = OPT_YES;
	break;
	
    default:
	will = OPT_NO;
	break;
    }

    c->myopts[option] = will;
    if (will)
	send_will (c, option);
    else
	send_wont (c, option);
}


static void
dontoption (struct connection *c, int option)
{

    switch (option) {
    case TELOPT_ECHO:
	pty_setecho (c, 0);
	break;
    case TELOPT_BINARY:
	pty_setbinaryout (c, 0);
	break;
    }

    c->myopts[option] = OPT_NO;
    send_wont (c, option);
}


/*
 * suboption()
 *
 *	Look at the sub-option buffer, and try to be helpful to the other
 * side.
 */

static void
suboption (struct connection *c)
{
    /* We don't use any of this information currently */
    switch (circ_getc (&c->subopt)) {
    case TELOPT_TTYPE:
	settimer (c, ttypesubopt);
	/* drop through */
    default:
	/* XXX print out for debugging? */
	circ_reset (&c->subopt);
	break;
    }
}



/*
 * State for recv fsm
 */
#define	TS_DATA		0	/* base state */
#define	TS_IAC		1	/* look for double IAC's */
#define	TS_CR		2	/* CR-LF ->'s CR */
#define	TS_SB		3	/* throw away begin's... */
#define	TS_SE		4	/* ...end's (suboption negotiation) */
#define	TS_WILL		5	/* will option negotiation */
#define	TS_WONT		6	/* wont " */
#define	TS_DO		7	/* do " */
#define	TS_DONT		8	/* dont " */
 
static int
parse_telnet (struct connection *c, unsigned char *bp, int len)
{
    unsigned char d;
    
    while (len > 0) {
	d = *bp++, len--;
	switch (c->state) {
	    
	case TS_CR:
	    c->state = TS_DATA;
	    /* Strip off \n or \0 after a \r */
	    if (d == 0 || d == '\n') {
		break;
	    }
	    /* FALL THROUGH */

	case TS_DATA:
	    if (d == IAC) {
		c->state = TS_IAC;
		break;
	    }
	    /*
	     * We now map \r\n ==> \r for pragmatic reasons.
	     * Many client implementations send \r\n when
	     * the user hits the CarriageReturn key.
	     *
	     * We USED to map \r\n ==> \n, since \r\n says
	     * that we want to be in column 1 of the next
	     * printable line, and \n is the standard
	     * unix way of saying that (\r is only good
	     * if CRMOD is set, which it normally is).
	     */
	    if (d == '\r' && c->hisopts[TELOPT_BINARY] == OPT_NO) {
		c->state = TS_CR;
	    }

	    /* stuff into pty buffer */
	    circ_putc (&c->ptybuf, d);
	    break;

	case TS_IAC:
	    switch (d) {
		
		/*
		 * Send the process on the pty side an
		 * interrupt.  Do this with a NULL or
		 * interrupt char; depending on the tty mode.
		 */
	    case IP:
		pty_sendinterrupt (c);
		break;
		
	    case BREAK:
		pty_sendbreak (c);
		break;
		
		/*
		 * Are You There?
		 */
	    case AYT:
		circ_putbuf (&c->netbuf, "\r\n[Yes]\r\n", 
			     sizeof "\r\n[Yes]\r\n" - 1);
		break;

		/*
		 * Abort Output
		 */
	    case AO:
		pty_sendcc (c, VDISCARD);
		net_clear (c);	/* clear buffer back */
		circ_putc (&c->netbuf, IAC);
		circ_putc (&c->netbuf, DM);
		circ_seturg (&c->netbuf);
		break;

	    /*
	     * Erase Character and
	     * Erase Line
	     */
	    case EC:
		pty_sendcc (c, VERASE);
		break;

	    case EL:
		pty_sendcc (c, VKILL);
		break;
	    
	    /*
	     * Check for urgent data...
	     */
	    case DM:
		settimer (c, gotDM);
		break;

		/*
		 * Begin option subnegotiation...
		 */
	    case SB:
		c->state = TS_SB;
		continue;
		
	    case WILL:
		c->state = TS_WILL;
		continue;

	    case WONT:
		c->state = TS_WONT;
		continue;
		
	    case DO:
		c->state = TS_DO;
		continue;
		
	    case DONT:
		c->state = TS_DONT;
		continue;

	    case IAC:
		circ_putc (&c->ptybuf, IAC);
		break;
	    }
	    c->state = TS_DATA;
	    break;

	case TS_SB:
	    if (d == IAC)
		c->state = TS_SE;
	    else
		circ_putc (&c->subopt, d);
	    break;

	case TS_SE:
	    if (d != SE) {
		if (d != IAC)
		    circ_putc (&c->subopt, IAC);
		circ_putc (&c->subopt, d);
		c->state = TS_SB;
	    } else {
		suboption (c);	/* handle sub-option */
		c->state = TS_DATA;
	    }
	    break;

	case TS_WILL:
	    if (c->hisopts[d] != OPT_YES)
		willoption (c, d);
	    c->state = TS_DATA;
	    continue;

	case TS_WONT:
	    if (c->hisopts[d] != OPT_NO)
		wontoption (c, d);
	    c->state = TS_DATA;
	    continue;

	case TS_DO:
	    if (c->myopts[d] != OPT_YES)
		dooption (c, d);
	    c->state = TS_DATA;
	    continue;

	case TS_DONT:
	    if (c->myopts[d] != OPT_NO)
		dontoption (c, d);
	    c->state = TS_DATA;
	    continue;
	    
	default:
	    syslog (LOG_ERR, "panic state=%d data=%d\n", c->state, d);
	    exit (1);
	}
    }
}



static int
start_child (void *arg)
{
    struct connection *c = arg;
    struct termios termbuf;
    int t, i, n;

    /* in child, open the child's side of the tty.  */
    t = open (c->tty_name, O_RDWR);
    if (t < 0)
	err (1, c->tty_name);

    /* the kernel does the initial terminal setup */

#ifdef unix
    /* create new session */
    if (setsid () < 0) 
	err (1, "setsid");
#else
    /* create a new process group and attach tty to it */
    setpgrp ();
    if (tcsetpgrp (t, getpgrp ()) < 0)
	err (1, "tcsetpgrp");
#endif

    /* make it stdin, stdout & stderr */
    dup2 (t, 0);
    dup2 (t, 1);
    dup2 (t, 2);

    /* close all other descriptors in child */
    n = getdtablesize ();
    for (i = 3; i < n; i++)
	close (i);

#ifdef unix
    /* start a login */
    execve (_PATH_LOGIN, argv_init, envp_init);
    err (1, "execl(%s)", _PATH_LOGIN);
#else
    /* XXX need to hendle topt->password, if set */
    pthread_setname_np (pthread_self (), "sh");
    exit (com_sh (1, argv_init));
#endif
}



static int
getpty (char *line)
{
    struct stat stb;
    const char *cp;
    int p;
    int j;

    strcpy (line, "/dev/ptyXX");

#ifdef linux
    cp = "pqrstuvwxyzabcde";
#else
    cp = "pqrstuvwxyzPQRST";
#endif
    for ( ; *cp; cp++) {

	line[8] = *cp;
	line[9] = '0';
	if (stat (line, &stb) < 0) {
	    /* give up as soon as one pty is not found */
	    syslog (LOG_DEBUG,  "out of ptys at %s: %m", line);
	    break;
	}
	for (j = 0; j < 16; j++) {
	    line[9] = "0123456789abcdef"[j];
	    if ((p = open (line, O_RDWR | O_NOCTTY)) >= 0) {
		syslog (LOG_DEBUG, "allocated pty %s", line);
		line[5] = 't';
		return p;
	    }
	}
    }
    return -1;
}


static int
create_connection (struct telnetopts *topt, int net)
{
    struct connection *c;
    struct sockaddr_in peer;
    int peerlen = sizeof (peer);
    int on = 1;
    int flags;
    char *s;
    int pty, pid, err;

    syslog (LOG_DEBUG, "creating new connection on fd %d", net);

    if (getpeername (net, (struct sockaddr *)&peer, &peerlen) == 0)
	s = inet_ntoa (peer.sin_addr);
    else
	s = "<unknown>";
    syslog (LOG_INFO, "new connection from %s", s);
	    
    /* request that OOB data stays inline, otherwise it is very
       difficult to keep a BSD 4.2 telnet client in synch! */
    if (setsockopt (net, SOL_SOCKET, SO_OOBINLINE, &on, sizeof on) < 0)
	syslog (LOG_WARNING, "setsockopt (SO_OOBINLINE): %m");

    /* switch keepalive on */
    if (setsockopt (net, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof on) < 0)
	syslog (LOG_WARNING, "setsockopt (SO_KEEPALIVE): %m");

    c = malloc (sizeof(struct connection));
    if (!c) {
	syslog (LOG_ERR, "malloc: %m");
	return 0;
    }

    memset (c, 0, sizeof (struct connection));
	    
    /* Got a new connection, set up a tty and spawn a shell.  */
    pty = getpty (c->tty_name);
    if (pty < 0) {
	syslog (LOG_ERR, "all pseudo ttys in use!");
	goto error;
    }

    syslog (LOG_DEBUG, "new connection pty %s = fd %d", c->tty_name, pty);

    /* set both streams to non-blocking mode */
    if ((flags = fcntl (net, F_GETFL, 0)) < 0) {
	syslog (LOG_ERR, "fcntl(net,F_GETFL): %m");
	goto error;
    }

    if (fcntl (net, F_SETFL, flags | O_NONBLOCK) < 0) {
	syslog (LOG_ERR, "fcntl(net,F_SETFL): %m");
	goto error;
    }
    
    if ((flags = fcntl (pty, F_GETFL, 0)) < 0) {
	syslog (LOG_ERR, "fcntl(pty,F_GETFL): %m");
	goto error;
    }

    if (fcntl (pty, F_SETFL, flags | O_NONBLOCK) < 0) {
	syslog (LOG_ERR, "fcntl(pty,F_SETFL): %m");
	goto error;
    }

    /* switch pty to packet mode */
    if (ioctl (pty, TIOCPKT, &on) < 0) {
	syslog (LOG_ERR, "ioctl(pty,TIOCPKT): %m");
	return 0;
    }

    c->netfd = net;
    c->ptyfd = pty;
    c->not42 = 1;

    circ_init (&c->netbuf, BUFSIZE);
    circ_init (&c->ptybuf, BUFSIZE);
    circ_init (&c->subopt, SUBSIZE);

    /* initial telnet options */
    dooption (c, TELOPT_ECHO);
    dooption (c, TELOPT_SGA);

    /*
     * Is the client side a 4.2 (NOT 4.3) system?  We need to know this
     * because 4.2 clients are unable to deal with TCP urgent data.
     *
     * To find out, we send out a "DO ECHO".  If the remote system
     * answers "WILL ECHO" it is probably a 4.2 client, and we note
     * that fact ("WILL ECHO" ==> that the client will echo what
     * WE, the server, sends it; it does NOT mean that the client will
     * echo the terminal input).
     */
    send_do (c, TELOPT_ECHO);
    c->hisopts[TELOPT_ECHO] = OPT_YES_BUT_ALWAYS_LOOK;

#ifdef unix
#ifdef HAVE_VFORK
    pid = vfork ();
#else
    pid = fork ();
#endif
    if (pid < 0) {
	syslog (LOG_ERR, "fork: %m");
	goto error;
    }
    if (pid == 0)
	start_child (c);
#else /* ! unix */
    if (err = pthread_proc_create_np (&pid, NULL, start_child, c)) {
	errno = err;
	syslog (LOG_ERR, "proc_create: %m");
	goto error;
    }
#endif /* ! unix */

    syslog (LOG_DEBUG, "new connection pid %d", pid);

    /* success, add to connection list */
    LIST_INSERT_HEAD(&topt->connections, c, link);
    c->pid = pid;
    return 1;

 error:
    destroy_connection (c);
    return 0;
}


static void
destroy_connection (struct connection *c)
{
    syslog (LOG_DEBUG, "freeing connection ptyfd %d, netfd %d, pid %d",
	    c->ptyfd, c->netfd, c->pid);

    if (c->pid != 0) {
	LIST_REMOVE(c, link);
	/* Closing the master pty below will send a HUP to the
	   child process group. */
#if 0
	kill (SIGHUP, -getpgid (c->pid));
#endif
    }

    close (c->ptyfd);
    close (c->netfd);

    circ_free (&c->ptybuf);
    circ_free (&c->netbuf);
    circ_free (&c->subopt);
    free ( c);
}



/* This code is not used yet! */

#if 0
static void
cleanup (struct connection *c)
{
#if unix
    char *p = c->tty_name + sizeof("/dev/") - 1;
    if (logout (p))
	logwtmp (p, "", "");
    (void)chmod (c->tty_name, 0666);
    (void)chown (c->tty_name, 0, 0);
    *p = 'p';
    (void)chmod (c->tty_name, 0666);
    (void)chown (c->tty_name, 0, 0);
#endif
    shutdown (c->netfd, 2);
    destroy_connection (c);
}
#endif


/* SIGCHLD handler */
static void
reapchild (int signum)
{
    int save_errno = errno, status;
    struct telnetopts *topt;
    struct connection *c;
    pid_t pid;

    /* XXX would need to use thread-specific data to 
       retrieve the current process' telnetopts root. */
    for (;;) {
	pid = waitpid (-1, &status, WNOHANG);
	if (pid <= 0)
	    break;
#if 0
	LIST_FOREACH(c, &topt->connections, link) {
	    if (c->pid == pid) {
		cleanup (c);
		break;
	    }
	}
#endif
    }
    errno = save_errno;
}



static int
telnetd (void *arg)
{
    struct telnetopts *topt = arg;
    char databuf[BUFSIZE];
    struct sockaddr_in sin;
    int sinlen;
    int s;
    fd_set ibits, obits, xbits;
    int on = 1;
    int maxfd, n;

    if (!topt->debug) {
#ifndef unix
	/* set own process/thread name */
	pthread_setname_np (pthread_self (), "telnetd");
	/* move to safe directory */
	chdir ("/");
	/* detach from tty */
	setpgrp ();
	ioctl (0, TIOCNOTTY, 0);
#else
	daemon (1, 1);
#endif
    }

    /* Open our syslog interface */
    openlog ("telnetd", 
	     LOG_PID | LOG_ODELAY | LOG_CONS | (topt->debug ? LOG_PERROR : 0),
	     LOG_DAEMON);
    setlogmask (topt->debug ? LOG_UPTO (LOG_DEBUG) : LOG_UPTO (LOG_INFO));
                      
    /* Grab a TCP socket.  */
    s = socket (AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	syslog (LOG_ERR, "socket(): %m");
	exit (1);
    }

    (void)setsockopt (s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

    /* Bind to the specified port */
    memset ((void *)&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = htonl (INADDR_ANY);
    sin.sin_port = htons (topt->portnbr);
    sin.sin_len = sizeof (sin);

    if (bind (s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	syslog (LOG_ERR, "bind(): %m");
	exit (1);
    }

    /* allow four pending connections */
    if (listen (s, 4) < 0) {
	syslog (LOG_ERR, "listen(): %m");
	exit (1);
    }

    /* install catcher to reap zombie children */
    signal (SIGCHLD, reapchild);

    syslog (LOG_INFO, "telnetd starting on port %d", topt->portnbr);

    LIST_INIT(&topt->connections);

    while (1) {
	struct connection *c, *next;

	FD_ZERO(&ibits);
	FD_ZERO(&obits);
	FD_ZERO(&xbits);

	/* select on the master socket, all telnet sockets and their
	 * ptys if there is room in their respective connection buffers.
	 */
	FD_SET(s, &ibits);
	maxfd = s;
	
	LIST_FOREACH(c, &topt->connections, link) {
	    if (!circ_empty (&c->ptybuf)) {
		/* can write to pty */
		FD_SET(c->ptyfd, &obits);
	    }
	    if (circ_space (&c->netbuf) >= 2) {
		/* can read from pty */
		FD_SET(c->ptyfd, &ibits);
	    }
	    else {
		/* else just catch errors from pty */
		FD_SET(c->ptyfd, &xbits);
	    }
	    if (c->ptyfd > maxfd)
		maxfd = c->ptyfd;

	    if (!circ_full (&c->ptybuf)) {
		/* can read from socket */
		FD_SET(c->netfd, &ibits);
		if (c->netfd > maxfd)
		    maxfd = c->netfd;
	    }
	    if (!circ_empty (&c->netbuf)) {
		/* can write to socket */
		FD_SET(c->netfd, &obits);
		if (c->netfd > maxfd)
		    maxfd = c->netfd;
	    }
	}

	n = select (maxfd + 1, &ibits, &obits, &xbits, 0);

	if (n < 0) {
	    if (errno == EINTR)
		continue;
	    syslog (LOG_ERR, "select(): %m");
	    exit (1);
	}

	if (n == 0)
	    continue;

	/* check for new connection */
	if (FD_ISSET(s, &ibits)) {
	    int fd;

	    fd = accept (s, (struct sockaddr *)&sin, &sinlen);
	    if (fd < 0) {
		syslog (LOG_WARNING, "accept(): %m");
		continue;
	    }

	    /* create a new telnet connection */
	    if (! create_connection (topt, fd))
		close (fd);
	}

	/* scan for i/o */
	for (c = LIST_FIRST(&topt->connections);
	     c != LIST_END(&topt->connections); 
	     c = next) {

	    int maxlen, w, r;
	    
	    /* get next pointer now, in case we free it */
	    next = LIST_NEXT(c, link);

	    if (FD_ISSET(c->ptyfd, &obits))
		/* write to pty */
		ptyflush (c);

	    if (FD_ISSET(c->netfd, &obits))
		/* write to socket */
		netflush (c);

	    if (FD_ISSET(c->netfd, &ibits)) {
                /* read from socket and send to pty buffer */
		maxlen = circ_space (&c->ptybuf);
		r = read (c->netfd, databuf, maxlen);
		if (r < 0 && (errno == EINTR || errno == EWOULDBLOCK))
		    continue;
		if (r <= 0) {
		    if (r == 0)
			syslog (LOG_INFO, "peer died");
		    else
			syslog (LOG_WARNING, "read(net): %m");
		    destroy_connection (c);
		    c = next;
		    continue;
		}
		r = parse_telnet (c, databuf, r);
	    }

	    if (FD_ISSET(c->ptyfd, &xbits)) {
		/* exception condition on pty */
		if (read (c->ptyfd, databuf, 1) != 1) {
		    if (errno != EINTR) {
			destroy_connection (c);
			c = next;
		    }
		    continue;
		}
	    }
	    if (FD_ISSET(c->ptyfd, &ibits)) {
                /* read from pty and send to socket buffer */
		unsigned char *bp;

		/* max is half remaining size, because of escape
		   code expansion, */
		maxlen = circ_space (&c->netbuf) / 2;

		r = read (c->ptyfd, databuf, maxlen);
		if (r < 0 && (errno == EINTR || errno == EWOULDBLOCK))
		    continue;
		if (r <= 0) {
		    if (r == 0)
			syslog (LOG_DEBUG, "pty closed");
		    else
			syslog (LOG_WARNING, "read(pty): %m");
		    destroy_connection (c);
		    c = next;
		    continue;
		}
		if (databuf[0] & TIOCPKT_FLUSHWRITE) {
		    net_clear (c);	/* clear buffer back */
		    circ_putc (&c->netbuf, IAC);
		    circ_putc (&c->netbuf, DM);
		    circ_seturg (&c->netbuf);
		    databuf[0] = 0;
		}
		r--;

		/* copy from input buffer to socket buffer */
		for (bp = &databuf[1]; r != 0; bp++, r--) {
		    /* escape protocol characters */
		    if (*bp == IAC)
			circ_putc (&c->netbuf, IAC);
		    circ_putc (&c->netbuf, *bp);
		    /* flag lone carriage returns if in text mode */
		    if (*bp == '\r' && c->myopts[TELOPT_BINARY] == OPT_NO) {
			if (r > 1 && bp[1] == '\n') {
			    circ_putc (&c->netbuf, *++bp);
			    r--;
			}
			else
			    circ_putc (&c->netbuf, '\0');
		    }
		}
	    }
	}

    } 

    return 0;
}


void
usg_telnetd (FILE *fp)
{
    fprintf (fp, "telnetd [-d] [-P password] [portnum]\n");
}


int
#ifdef unix
main (int argc, char **argv)
#else
com_telnetd (int argc, char **argv)
#endif
{
    struct telnetopts *topt;
    int ch;

    topt = malloc (sizeof (struct telnetopts));
    if (!topt) {
	perror ("malloc");
	exit (1);
    }

    memset (topt, 0, sizeof (struct telnetopts));
    topt->portnbr = 23;		/* default telnet port */

    while ((ch = getopt (argc, argv, "dP:")) != -1)
	switch (ch) {
	case 'd':
	    topt->debug = 1;
	    break;
	case 'P':
	    topt->password = strdup (optarg);
	    break;
	case '?':
	default:
	    usg_telnetd (stderr);
	    exit (1);
	}
    argv += optind;
    argc -= optind;

    /* allow port number override */
    if (argc == 1) {
	char *ep;
	topt->portnbr = strtol (argv[0], &ep, 10);
	if (topt->portnbr <= 0 || *ep) {
	    usg_telnetd (stderr);
	    exit (1);
	}
    }
    else if (argc != 0) {
	usg_telnetd (stderr);
	exit (1);
    }

#ifdef unix
    /* check that shell program is available */
    if (access (shellpath, X_OK) < 0) {
	perror (shellpath);
	exit (1);
    }
#endif

#ifndef unix
    if (!topt->debug) {
	/* create background daemon process */
	int err;
	if (err = pthread_proc_create_np (NULL, NULL, telnetd, topt))
	    perror ("fork");
    }
    else {
	/* debugging: run daemon in foreground */
	telnetd (topt);
    }
#else
    telnetd (topt);
#endif

    exit (0);
}
