#ident "sde/rtpx/conc.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * conc.c: Transputer emulation over Posix threads
 */


#define _POSIX_C_SOURCE 199309L
#define _POSIX_SOURCE

#undef NDEBUG			/* enable assertions */

#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#if !defined(__i386__)
#include <alloca.h>
#endif

#include <conc.h>

#ifndef _POSIX_THREAD_PRIORITY_SCHEDULING
#error _POSIX_THREAD_PRIORITY_SCHEDULING
#endif

#ifndef _POSIX_THREAD_ATTR_STACKSIZE
#error _POSIX_THREAD_ATTR_STACKSIZE
#endif

#define TPUTER_STACK_SIZE_MIN		(8*1024)
#define TPUTER_STACK_SIZE_DEFAULT	(16*1024)

#define TPUTERHI_PRIO		0
#define TPUTERLO_PRIO		1

#define TPUTERHI_THREAD_PRIO	96
#define TPUTERLO_THREAD_PRIO	32

static void
tpuClean (void *arg)
{
    Process        *p = arg;
    p->running = 0;
}

static void    *
tpuRun (void *arg)
{
    Process        *p = arg;

    pthread_cleanup_push (tpuClean, p);
    p->running = 1;

    /* Wait until run asked for */
    switch (p->nParam) {
    case 0:
	p->pfun (p);
	break;
    case 1:
	p->pfun (p, p->params[0]);
	break;
    case 2:
	p->pfun (p, p->params[0], p->params[1]);
	break;
    case 3:
	p->pfun (p, p->params[0], p->params[1], p->params[2]);
	break;
    case 4:
	p->pfun (p, p->params[0], p->params[1], p->params[2],
		 p->params[3]);
	break;
    case 5:
	p->pfun (p, p->params[0], p->params[1], p->params[2], p->params[3],
		 p->params[4]);
	break;
    case 6:
	p->pfun (p, p->params[0], p->params[1], p->params[2], p->params[3],
		 p->params[4], p->params[5]);
	break;
    case 7:
	p->pfun (p, p->params[0], p->params[1], p->params[2], p->params[3],
		 p->params[4], p->params[5], p->params[6]);
	break;
    case 8:
	p->pfun (p, p->params[0], p->params[1], p->params[2], p->params[3],
		 p->params[4], p->params[5], p->params[6], p->params[7]);
	break;
    default:
	abort ();
    }
    pthread_cleanup_pop (1);
    return 0;
}

static void
ProcParamList (Process * p, va_list ap)
{
    int             pn;

    for (pn = 0; pn < p->nParam; pn++)
	p->params[pn] = va_arg (ap, int);
}

void
ProcParam (Process * p, ...)
{
    va_list         ap;

    va_start (ap, p);
    ProcParamList (p, ap);
    va_end (ap);
}

static void
ProcInitList (Process * p, void (*f) (), void *sp, int s,
	      int np, va_list ap)
{
    unsigned long   stktop;

    assert (!p->running);

    /* snarf param vector from base of stack */
    assert (s >= TPUTER_STACK_SIZE_MIN);
    assert (sp);

    stktop = ((unsigned long) sp + s) & ~(sizeof (int) - 1);

    p->pfun = f;
    p->nParam = np;
    p->params = (int *) (stktop -= np * sizeof (int));

    assert (stktop > (unsigned long) sp);

    pthread_attr_init (&p->tha);
    pthread_attr_setstackaddr (&p->tha, sp);
    pthread_attr_setstacksize (&p->tha, stktop - (unsigned long) sp);
    pthread_attr_setinheritsched (&p->tha, PTHREAD_EXPLICIT_SCHED);
    pthread_attr_setschedpolicy (&p->tha, SCHED_RR);
    {
	struct sched_param shp;
	pthread_attr_getschedparam (&p->tha, &shp);
	shp.sched_priority = TPUTERLO_THREAD_PRIO;
	pthread_attr_setschedparam (&p->tha, &shp);
    }
    pthread_attr_setcontentionscope (&p->tha, PTHREAD_SCOPE_PROCESS);

    ProcParamList (p, ap);
}

void
ProcInit (Process * p, void (*f) (), void *sp, int s, int np, ...)
{
    va_list         ap;

    va_start (ap, np);

    ProcInitList (p, f, sp, s, np, ap);

    va_end (ap);
}


Process        *
ProcAlloc (void (*f) (), int s, int np, ...)
{
    Process        *p = malloc (sizeof (*p));
    va_list         ap;

    if (p != NULL) {
	void           *sp;
	if (s == 0)
	    s = TPUTER_STACK_SIZE_DEFAULT;

	if ((sp = malloc (s)) == NULL) {
	    free (p);
	    errno = ENOMEM;
	    return NULL;
	}

	va_start (ap, np);

	p->running = 0;
	ProcInitList (p, f, sp, s, np, ap);

	va_end (ap);
    }
    return p;
}

void
ProcFree (Process * p)
{
    void           *sp;

    assert (!p->running);
    pthread_attr_getstackaddr (&p->tha, &sp);
    pthread_attr_destroy (&p->tha);
    free (sp);
    free (p);
}


static void
ProcStart (Process * p, int pri)
{
    assert (!p->running);	/* shouldn't be running */
    assert (p->pfun);		/* must have been initialised */

    if (pri != TPUTERLO_THREAD_PRIO) {
	struct sched_param shp;
	pthread_attr_getschedparam (&p->tha, &shp);
	shp.sched_priority = pri;
	pthread_attr_setschedparam (&p->tha, &shp);
	pthread_attr_setschedpolicy (&p->tha, SCHED_FIFO);
    }
    pthread_create (&p->thid, &p->tha, tpuRun, p);
}

void
ProcParList (Process * pv[])
{
    Process       **pp;
    Process        *p;

    for (pp = pv; (p = *pp); pp++) {
	pthread_attr_setdetachstate (&p->tha, PTHREAD_CREATE_JOINABLE);
	ProcStart (p, (ProcGetPriority () == TPUTERLO_PRIO) ?
		   TPUTERLO_THREAD_PRIO : TPUTERHI_THREAD_PRIO);
    }

    for (pp = pv; (p = *pp); pp++) {
	pthread_join (p->thid, NULL);
    }
}

void
ProcPar (Process * p0, ...)
{
    va_list         ap;
    Process        *p;

    va_start (ap, p0);
    p = p0;

    while (p) {
	pthread_attr_setdetachstate (&p->tha, PTHREAD_CREATE_JOINABLE);
	ProcStart (p, (ProcGetPriority () == TPUTERLO_PRIO) ?
		   TPUTERLO_THREAD_PRIO : TPUTERHI_THREAD_PRIO);
	p = va_arg (ap, Process *);
    }
    va_end (ap);

    va_start (ap, p0);
    p = p0;

    while (p) {
	pthread_join (p->thid, NULL);
	p = va_arg (ap, Process *);
    }
    va_end (ap);
}


void
ProcRun (Process * p)
{
    pthread_attr_setdetachstate (&p->tha, PTHREAD_CREATE_DETACHED);
    ProcStart (p, (ProcGetPriority () == TPUTERLO_PRIO) ?
	       TPUTERLO_THREAD_PRIO : TPUTERHI_THREAD_PRIO);
}

void
ProcRunHigh (Process * phi)
{
    pthread_attr_setdetachstate (&phi->tha, PTHREAD_CREATE_DETACHED);
    ProcStart (phi, TPUTERHI_THREAD_PRIO);
}

void
ProcRunLow (Process * plo)
{
    pthread_attr_setdetachstate (&plo->tha, PTHREAD_CREATE_DETACHED);
    ProcStart (plo, TPUTERLO_THREAD_PRIO);
}

void
ProcPriPar (Process * phi, Process * plo)
{
    pthread_attr_setdetachstate (&phi->tha, PTHREAD_CREATE_JOINABLE);
    ProcStart (phi, TPUTERHI_THREAD_PRIO);
    pthread_attr_setdetachstate (&plo->tha, PTHREAD_CREATE_JOINABLE);
    ProcStart (plo, TPUTERLO_THREAD_PRIO);

    pthread_join (phi->thid, NULL);
    pthread_join (plo->thid, NULL);
}

int
ProcGetPriority (void)
{
    struct sched_param sp;
    pthread_getschedparam (pthread_self (), NULL, &sp);
    return (sp.sched_priority == TPUTERLO_THREAD_PRIO)
	? TPUTERLO_PRIO : TPUTERHI_PRIO;
}

int
ProcToHigh (void)
{
    struct sched_param osp,
                    sp;

    pthread_getschedparam (pthread_self (), NULL, &osp);
    sp.sched_priority = TPUTERHI_THREAD_PRIO;
    pthread_setschedparam (pthread_self (), SCHED_FIFO, &sp);
    return (osp.sched_priority == TPUTERLO_THREAD_PRIO)
	? TPUTERLO_PRIO : TPUTERHI_PRIO;
}

int
ProcToLow (void)
{
    struct sched_param osp,
                    sp;
    pthread_getschedparam (pthread_self (), NULL, &osp);
    sp.sched_priority = TPUTERLO_THREAD_PRIO;
    pthread_setschedparam (pthread_self (), SCHED_RR, &sp);
    return (osp.sched_priority == TPUTERLO_THREAD_PRIO)
	? TPUTERLO_PRIO : TPUTERHI_PRIO;
}

void
ProcReschedule (void)
{
    sched_yield ();
}

void
ProcStop (void)
{
    pthread_exit (NULL);
}

#define NSPERSEC (1*1000*1000*1000)
/* set up a timespec from counter and units per sec */
#define TS_SETUP(ts,cnt,ups) \
	((ts).tv_sec = (cnt)/(ups), \
	 (ts).tv_nsec = (NSPERSEC/(ups)) * ((cnt)%(ups)))

void
ProcWait (int time)
{
    struct timespec ts;

    if (time > 0) {
	if (ProcGetPriority () == TPUTERLO_PRIO)
	    /* lo priority time is 64us ticks */
	    TS_SETUP (ts, time, (1000 * 1000 / 64));
	else
	    /* hi priority time is 1us ticks */
	    TS_SETUP (ts, time, 1000 * 1000);
	while (nanosleep (&ts, &ts) != 0)
	    continue;
    }
}


void
ProcWaitMs (int ms)
{
    struct timespec ts;

    if (ms > 0) {
	TS_SETUP (ts, ms, 1000);
	while (nanosleep (&ts, &ts) != 0)
	    continue;
    }
}


void
ProcWaitUs (int us)
{
    struct timespec ts;

    if (us > 0) {
	TS_SETUP (ts, us, 1000 * 1000);
	while (nanosleep (&ts, &ts) != 0)
	    continue;
    }
}

void
ProcWaitNs (int ns)
{
    struct timespec ts;

    if (ns > 0) {
	ts.tv_sec = ns / NSPERSEC;
	ts.tv_nsec = ns % NSPERSEC;
	while (nanosleep (&ts, &ts) != 0)
	    continue;
    }
}


Semaphore      *
SemAlloc (void)
{
    Semaphore      *sem = malloc (sizeof (*sem));

    if (sem)
	memset (sem, 0, sizeof (*sem));
    return sem;
}

void
SemFree (Semaphore * s)
{
    free (s);
}


#ifdef _POSIX_THREAD_DOORBELL_NP
Channel         _tpuHwChannelTable[32];

static void
chanInit (Channel * cp)
{
    int             evn = cp - _tpuHwChannelTable;
    pthread_doorbellattr_t dba;

    pthread_doorbellattr_init (&dba);
    pthread_doorbellattr_setevn (&dba, evn);

    pthread_doorbell_init (&cp->chan_db, &dba);

    pthread_doorbellattr_destroy (&dba);

    cp->chan_state = CHAN_EMPTY;
}

int
ChanInInt (Channel * cp)
{
    switch (cp->chan_state) {
    case CHAN_DEAD:
	chanInit (cp);
	/* fall thorugh */
    case CHAN_FULL:
	cp->chan_state = CHAN_EMPTY;
	break;
    case CHAN_EMPTY:
	if (pthread_doorbell_wait (&cp->chan_db, NULL) != 0) {
	    /* XXX hmm */
	    return -1;
	}
	cp->chan_value = 42;
	break;
    }
    return cp->chan_value;
}

int
ProcTimerAlt (int time, Channel * cp, ...)
{
    struct timespec ts,
                    to;
    va_list         ap;

    va_start (ap, cp);
    /* cant cope */
    assert (va_arg (ap, Process *) == NULL);

    va_end (ap);

    switch (cp->chan_state) {
    case CHAN_FULL:
	return 0;
    case CHAN_DEAD:
	chanInit (cp);
	/* fall though */
    case CHAN_EMPTY:
	if (time > 0) {
	    if (ProcGetPriority () == TPUTERLO_PRIO)
		/* lo priority time is 64us ticks */
		TS_SETUP (to, time, (1000 * 1000 / 64));
	    else
		/* hi priority time is 1us ticks */
		TS_SETUP (to, time, 1000 * 1000);
	    clock_gettime (CLOCK_REALTIME, &ts);
	    ts.tv_sec += to.tv_sec;
	    if ((ts.tv_nsec += to.tv_nsec) >= NSPERSEC)
		ts.tv_sec++, ts.tv_nsec -= NSPERSEC;
	    if (pthread_doorbell_timedwait (&cp->chan_db, NULL, &ts) != 0)
		return -1;
	}
	else {
	    if (pthread_doorbell_trywait (&cp->chan_db) != 0)
		return -1;
	}
	cp->chan_value = 42;
	cp->chan_state = CHAN_FULL;
    }
    return 0;
}
#endif
