#ident "sde/rtpx/rtpx_flock.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_flock.c: generic recursive locks for SDE C library
 */


#define _POSIX_C_SOURCE 199506L
#include <pthread.h>
#include <sdethread.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include "rtpx_impl.h"

/* REENTRANT FILE LOCK */
struct __sdethread_flock {
    pthread_mutex_t	mx;
    pthread_t		owner;
    unsigned int	nlock;
    pthread_cond_t	notbusy;
    char		name[8];
};


static pthread_mutex_t flock_mx = PTHREAD_MUTEX_INITIALIZER;

static void
flock_cleanup (void *arg)
{
    pthread_flock_np_t *fpl = arg;
    pthread_mutex_unlock (&fpl->mx);
}

__sdethread_flock_t 
__sdethread_flock_create (__sdethread_flock_t *flockp, int fd)
{
    pthread_mutexattr_t mxa;
    pthread_condattr_t cda;
    __sdethread_flock_t fpl;
    const char *name;

    /* Check with mutex protection, in case we're
       in a race with another thread. */
    pthread_mutex_lock (&flock_mx);
    if (!(fpl = *flockp)) {
	/* We really are the first user, so allocate a new lock */
	fpl = malloc (sizeof (struct __sdethread_flock));
	if (!fpl)
	    abort ();

	memset (fpl, 0, sizeof (pthread_flock_np_t));

	switch (fd) {
	case 0:
	    name = "stdin";
	    break;
	case 1:
	    name = "stdout";
	    break;
	case 2:
	    name = "stderr";
	    break;
	default:
	    name = strcpy (fpl->name, "FILEXX");
	    fpl->name[4] = '0' + (fd / 10);
	    fpl->name[5] = '0' + (fd % 10);
	    break;
	}

	/* Initialise the lock */
	pthread_mutexattr_init (&mxa);
	pthread_mutexattr_setname_np (&mxa, name);
	pthread_mutexattr_setprotocol (&mxa, PTHREAD_PRIO_INHERIT);
	pthread_mutexattr_setpshared (&mxa, PTHREAD_PROCESS_PRIVATE);
	pthread_mutex_init (&fpl->mx, &mxa);
	pthread_mutexattr_destroy (&mxa);

	pthread_condattr_init (&cda);
	pthread_condattr_setname_np (&cda, name);
	pthread_condattr_setinterruptible_np (&cda, 0);
	pthread_cond_init (&fpl->notbusy, &cda);
	pthread_condattr_destroy (&cda);

	*flockp = fpl;
    }
    pthread_mutex_unlock (&flock_mx);

    return fpl;
}

/* Lock file lock - recursive locking is allowed. */
void 
__sdethread_flock_lock (__sdethread_flock_t fpl)
{
    pthread_cleanup_push_defer_np (flock_cleanup, fpl);

    if (pthread_mutex_lock (&fpl->mx) != 0)
	abort ();

    if (fpl->nlock != 0) {
	/* lock is busy, is this thread the owner? */
	if (!pthread_equal (fpl->owner, pthread_self ())) {
	    /* not owner, wait until not busy */
	    do {
		pthread_cond_wait (&fpl->notbusy, &fpl->mx);
	    } while (fpl->nlock != 0);
	}
    }

    /* record lock */
    fpl->owner = pthread_self ();
    fpl->nlock++;

    pthread_mutex_unlock (&fpl->mx);
    pthread_cleanup_pop_restore_np (0);
}

/* Try to lock file lock - recursive locking is allowed. */
int
__sdethread_flock_trylock (__sdethread_flock_t fpl)
{
    int 	rv;

    pthread_cleanup_push_defer_np (flock_cleanup, fpl);

    rv = pthread_mutex_trylock (&fpl->mx);
    if (rv != 0)
	/* couldn't lock mutex */
	goto ret;

    if (fpl->nlock != 0) {
	/* lock is busy, is this thread the owner? */
	if (!pthread_equal (fpl->owner, pthread_self ())) {
	    /* not owner, return error */
	    rv = EBUSY;
	    goto fail;
	}
    }

    fpl->owner = pthread_self ();
    fpl->nlock++;

  fail:
    pthread_mutex_unlock (&fpl->mx);

  ret:
    pthread_cleanup_pop_restore_np (0);
    return rv;
}

/* Unlock a file lock. */
void
__sdethread_flock_unlock (__sdethread_flock_t fpl)
{
    unsigned int    nlock;
    pthread_mutex_lock (&fpl->mx);
    nlock = --fpl->nlock;
    pthread_mutex_unlock (&fpl->mx);
    if (nlock == 0)
	pthread_cond_signal (&fpl->notbusy);
}

/* Release all recursive locks when file is closed, or a thread exits
   with locks still held. */
void
__sdethread_flock_release (__sdethread_flock_t fpl)
{
    if (pthread_mutex_lock (&fpl->mx) == 0) {
	if (fpl->nlock != 0 && pthread_equal (fpl->owner, pthread_self ())) {
	    fpl->nlock = 0;
	    pthread_mutex_unlock (&fpl->mx);
	    pthread_cond_signal (&fpl->notbusy);
	}
	else {
	    pthread_mutex_unlock (&fpl->mx);
	}
    }
}

/* Release and free lock when file is closed */
void
__sdethread_flock_free (__sdethread_flock_t fpl)
{
    __sdethread_flock_release (fpl);
    free (fpl);
}

/* C library file lock destructor */
static void (*libc_flock_dtor) (void);

/* Register C library file lock destructor */
void
__sdethread_flock_init (void (*dtor) (void))
{
    libc_flock_dtor = dtor;
}

/* Callback on thread destruction */
void 
_rtpx_flockfilecleanup ()
{
    if (libc_flock_dtor)
	libc_flock_dtor ();
}


/* Called from _pthread_init */
void
_rtpx_flockfileinit (void)
{
}

