#ident "sde/rtpx/rtpx_key.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_key.c: thread-specific data implementation
 */


#define _POSIX_C_SOURCE 199506L
#include <stddef.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "pthread_impl.h"

#ifdef _RTPX_TSD

static pthread_mutex_t tsdMx = PTHREAD_MUTEX_INITIALIZER;
static rtpx_tsd_t tsdTab[PTHREAD_KEYS_MAX];


int
pthread_key_create (pthread_key_t * key, void (*destructor) (void *))
{
    rtpx_tsd_t     *tsdp;

    pthread_mutex_lock (&tsdMx);
    for (tsdp = tsdTab; tsdp < &tsdTab[PTHREAD_KEYS_MAX]; tsdp++)
	if (tsdp->tsd_count == 0) {
	    tsdp->tsd_count = 1;	/* exists but no value */
	    tsdp->tsd_destructor = destructor;
	    *key = tsdp - tsdTab;
	    break;
	}
    pthread_mutex_unlock (&tsdMx);
    return (tsdp < &tsdTab[PTHREAD_KEYS_MAX]) ? 0 : EAGAIN;
}


int
pthread_key_delete (pthread_key_t key)
{
    rtpx_tsd_t     *tsdp;

    if (key >= PTHREAD_KEYS_MAX)
	return EINVAL;

    tsdp = &tsdTab[key];
    pthread_mutex_lock (&tsdp->tsd_mx);

    if (tsdp->tsd_count == 0) {
	/* no longer valid */
	pthread_mutex_unlock (&tsdp->tsd_mx);
	return EINVAL;
    }

    if (tsdp->tsd_count > 1) {
	/* some values still exist */
	pthread_mutex_unlock (&tsdp->tsd_mx);
	return EBUSY;
    }

    tsdp->tsd_destructor = NULL;
    tsdp->tsd_count = 0;
    pthread_mutex_unlock (&tsdp->tsd_mx);

    return 0;
}


int
pthread_setspecific (pthread_key_t key, const void *value)
{
    rtpx_thread_t  *cth = rtpx_self ();
    rtpx_tsd_t     *tsdp = &tsdTab[key];
    int             ret;

    if (key >= PTHREAD_KEYS_MAX)
	return EINVAL;

    /* no need to lock own thread data - no other thread will access */
    cth = rtpx_self ();
    if (!cth->th_tsd) {
	cth->th_tsd = calloc (PTHREAD_KEYS_MAX, sizeof (void *));
	if (!cth->th_tsd)
	    return ENOMEM;
    }
    tsdp = &tsdTab[key];
    pthread_mutex_lock (&tsdp->tsd_mx);
    if (tsdp->tsd_count != 0) {
	if (cth->th_tsd[key] == NULL) {
	    if (value != NULL) {
		cth->th_tsdcount++;
		tsdp->tsd_count++;
	    }
	}
	else {
	    if (value == NULL) {
		cth->th_tsdcount--;
		tsdp->tsd_count--;
		assert (tsdp->tsd_count >= 1);
	    }
	}
	cth->th_tsd[key] = value;
	ret = 0;
    }
    else {
	ret = EINVAL;
    }
    pthread_mutex_unlock (&tsdp->tsd_mx);
    return ret;
}


void           *
(pthread_getspecific) (pthread_key_t key)
{
    rtpx_thread_t  *cth = rtpx_self ();

    /* minimal checking - designed to be fast! */
    if (key >= PTHREAD_KEYS_MAX || !cth->th_tsd)
	return NULL;
    return (void *)cth->th_tsd[key];
}


/* 
 * Clean up all thread-specific data set by thread when it
 * exits.
 */
void
_rtpx_tsdcleanup ()
{
    rtpx_thread_t  *cth = rtpx_self ();
    unsigned int    i;

    for (i = 0; i < PTHREAD_DESTRUCTOR_ITERATIONS && cth->th_tsdcount; i++) {
	pthread_key_t   key;

	for (key = 0; key < PTHREAD_KEYS_MAX && cth->th_tsdcount; key++) {
	    const void *data;

	    if (data = cth->th_tsd[key]) {
		rtpx_tsd_t     *tsdp;
		void            (*destructor) (void *);

		cth->th_tsd[key] = NULL;
		cth->th_tsdcount--;

		tsdp = &tsdTab[key];
		pthread_mutex_lock (&tsdp->tsd_mx);
		tsdp->tsd_count--;
		assert (tsdp->tsd_count >= 1);
		destructor = tsdp->tsd_destructor;
		pthread_mutex_unlock (&tsdp->tsd_mx);

		if (destructor)
		    (destructor) ((void *)data);
	    }
	}
    }

    if (cth->th_tsd) {
	free (cth->th_tsd);
	cth->th_tsd = NULL;
	cth->th_tsdcount = 0;
    }
}

#include <sdethread.h>

/* 
 * One time initialisation
 */
void
_rtpx_tsdinit (void)
{
    rtpx_tsd_t     *tsdp;

    if (sizeof (__sdethread_key_t) != sizeof (pthread_key_t))
	abort ();
    for (tsdp = tsdTab; tsdp < &tsdTab[PTHREAD_KEYS_MAX]; tsdp++)
	pthread_mutex_init (&tsdp->tsd_mx, NULL);
}


/* SDETHREAD API aliases */
int __sdethread_key_create (__sdethread_key_t *, void (*)(void *))
	__attribute__((alias("pthread_key_create")));
int __sdethread_key_delete (__sdethread_key_t)
	__attribute__((alias("pthread_key_delete")));
int __sdethread_setspecific (__sdethread_key_t, const void *)
	__attribute__((alias("pthread_setspecific")));
void * __sdethread_getspecific (__sdethread_key_t)
	__attribute__((alias("pthread_getspecific")));


/* C library reentrancy support: doesn't really belong here ! */

#include <libc_thread.h>

extern int __isthreaded;

void *
_libc_private_storage(key, init, initsz, error)
    	struct _thread_private_key_struct * key;
	void *	init;
	size_t	initsz;
	void *	error;
{
    	struct _thread_private_key_struct * volkey = key;
	void *result;
	void (*cleanfn) (void *);
	sigset_t nm, om;

	/* Use static storage while not threaded: */
	if (!__isthreaded)
		return init;

	/* Create the key once: */
	if (!volkey->inited) {
		if (__sdethread_mutex_lock(&key->mutex) != 0)
			return error;
		if (!volkey->inited) {
			if (key->cleanfn == NULL)
				cleanfn = free;
			else
				cleanfn = key->cleanfn;
			if (__sdethread_key_create(&key->key, cleanfn) != 0) {
				__sdethread_mutex_unlock(&key->mutex);
				return error;
			}
			key->inited = 1;
		}
		__sdethread_mutex_unlock(&key->mutex);
	}

	/* Prevent signals from causing reentrancy */
	sigfillset (&nm);
	pthread_sigmask (SIG_SETMASK, &nm, &om);

	/* Acquire this key's thread-specific storage: */
	result = __sdethread_getspecific(key->key);

	/* Allocate and initialise storage if unallocated: */
	if (result == NULL) {
		result = malloc(initsz);
		if (result == NULL) {
		    	result = error;	
			goto done;
		}
		if (__sdethread_setspecific(key->key, result) != 0) {
			free(result);
		    	result = error;	
			goto done;
		}
		memcpy(result, init, initsz);
	}

 done:
	pthread_sigmask (SIG_SETMASK, &om, NULL);
	return result;
}

#endif
