#ident "sde/rtpx/rtpx_mem.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2001-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_mem.c: low-level memory allocation for AlgRTX
 */


#define _POSIX_C_SOURCE 199506L

#include "pthread_impl.h"
#include <stdarg.h>
#include <string.h>

#define PAGESIZE	4096
#define PAGEMASK	(PAGESIZE - 1)

#ifndef MAXSTACK
#define MAXSTACK	(32 * 1024)
#endif

/* One day we may keep a list of pages attached to each process */
struct page_list {
    struct page_list	*next;		/* next page in list */
    void		*page;		/* page base */
};


struct core_region {
    unsigned int	npg;		/* total pages in region */
    unsigned int	nfree;		/* # free pages in region */
    void		*base;		/* base address of region */
#if 1
    unsigned int	firstfree;	/* first free page in region */
#else
    struct page_list	*pages;		/* page list strucure */
#endif
};


/* Free memory table and protecting mutex */
static struct core_region	*core_tab;
static pthread_mutex_t		core_mx = PTHREAD_MUTEX_INITIALIZER;

/* Memory layout variables */
extern char     _end[];

/* _minbrk and _maxbrk can be set by startup code, or by a linker
   script, so we don't want them in bss where they'll get cleared, so
   they can't be common, but they must be capable of being
   overridden. */
void *		_minbrk __attribute__((weak)) = 0;
void *		_maxbrk __attribute__((weak)) = 0;

/* Return the page size. */
int
getpagesize ()
{
    return PAGESIZE;
}


/* Dummy version of _sbd_memlayout(), which should return a pointer to
   a table of memory regions. */
#pragma weak _sbd_memlayout=_stub_sbd_memlayout
const struct sbd_region *_stub_sbd_memlayout (void);

const struct sbd_region *
_stub_sbd_memlayout (void)
{
    static struct sbd_region mem[2];
    extern paddr_t _mem_top;

    mem[0].type = SBD_MEM_RAM;
    mem[0].base = 0x00000000;
    mem[0].size = _mem_top;
    if (_mem_top)
	/* _mem_top is usually set by monitor-specific startup code */
	mem[0].size = _mem_top;
    else if (_maxbrk)
	/* if we don't know the memory size, use the end of the heap */
	mem[0].size = KVA_TO_PA (_maxbrk);
    else{
	/* final resort, assume stack points to top of memory */
	void * sp;
	__asm__ ("move %0,$sp" : "=d" (sp));
	mem[0].size = KVA_TO_PA (sp);
    }
    return mem;
}

static pthread_once_t sbrkonce = PTHREAD_ONCE_INIT;

/* Initialise the free page pool. */
static void
sbrkfirst (void)
{
    const struct sbd_region * layout, * mr;
    struct core_region *cr;
    void * minva, * maxva;
    paddr_t min, max;
    unsigned int nr;

    if (core_tab)
	/* already initialised */
	return;

    if (_minbrk)
	/* user specified heap start */
	minva = _minbrk;
    else
	/* default heap starts after data & bss segment */
	minva = _end;

    if (IS_KVA01 (minva))
	min = KVA_TO_PA (minva);
    else
	/* XXX assumes one-to-one mapping in KUSEG */
	min = (paddr_t) minva;

    if (_maxbrk)
	/* user specified heap top */
	maxva = _maxbrk;
    else {
	/* default heap top is just below stack */
	char * sp;
	__asm__ ("move %0,$sp" : "=d" (sp));
	maxva = sp - MAXSTACK;
    }

    if (IS_KVA01 (maxva))
	max = KVA_TO_PA (maxva);
    else
	max = (paddr_t) maxva;

    /* determine physical memory layout */
    layout = _sbd_memlayout ();

    /* count disjoint regions */
    nr = 1;
    for (mr = layout; mr->type != SBD_MEM_END; mr++) 
	if (mr->type != SBD_MEM_RAM)
	    ++nr;
    
    /* allocate free memory region header table */
    if (IS_KVA1 (_end))
	core_tab = PA_TO_KVA1 (min);
    else if (IS_KVA0 (_end))
	core_tab = PA_TO_KVA0 (min);
    else
	/* kuseg: assume one-to-one mapping */
	core_tab = (void *)min;

    min += nr * sizeof (struct core_region);

    /* fill the region headers */
    cr = core_tab;
    for (mr = layout; mr->type != SBD_MEM_END; mr++) {
	paddr_t lo, hi;
 	unsigned int npg;

	if (mr->type != SBD_MEM_RAM)
	    continue;

	lo = mr->base;
	hi = lo + mr->size;
	if (lo < max && hi >= min) {
	    /* region overlaps with initial heap area, don't
	       overwrite program data or stack. */
	    if (lo < min)
		lo = min;
	    if (hi > max)
		hi = max;
	}

	/* round high address down to page boundary */
	hi = hi & ~PAGEMASK;

#if 0
	/* In the future we may keep a list of pages allocated
	   to each process, to be freed when it exits. In which
	   case we'll grab the first part of this region as a
	   page list array. */
	npg = (hi - lo) / PAGESIZE;
	cr->pages = IS_KVA0 (_end) ? PA_TO_KVA0 (lo) : PA_TO_KVA1 (lo);
	lo += npg * sizeof (struct page_list);
	memset (cr->pages, 0, sizeof (struct page_list) * npg);
#endif
	
	/* round low address up to page boundary */
	lo = (lo + PAGESIZE - 1) & ~PAGEMASK;
	npg = (hi - lo) / PAGESIZE;

	if (IS_KVA1 (_end))
	    cr->base = PA_TO_KVA1 (lo);
	else if (IS_KVA0 (_end))
	    cr->base = PA_TO_KVA0 (lo);
	else
	    /* kuseg: assume one-to-one mapping */
	    cr->base = (void *)lo;

	cr->nfree = cr->npg = npg;
	cr->firstfree = 0;
	cr++;
    }

    /* terminate core region list */
    cr->npg = 0;
}

void
_sbrk_init (void)
{
    (void) pthread_once (&sbrkonce, sbrkfirst);
}


void *
pthread_getpages_np (size_t n)
{
    unsigned int npg;
    void *p = NULL;
    struct core_region *cr;

    if (n & PAGEMASK)
	abort ();
    npg = n / PAGESIZE;

    _sbrk_init ();

    pthread_mutex_lock (&core_mx);
    for (cr = core_tab; cr->npg != 0; cr++) {
	if (cr->nfree < npg)
	    continue;
	/* In the current scheme we never free pages,
	   so we just take from the first region 
	   with enough space. */
	p = cr->base + (cr->firstfree * PAGESIZE);
	cr->firstfree += npg;
	cr->nfree -= npg;
	break;
    }
    pthread_mutex_unlock (&core_mx);
    return p;
}


/*
 * Old interface to memory allocation (deprecated)
 */
void *
_sbrk (int nb) 
{
    void *ret;

    _sbrk_init ();

    /* return lowest address for malloc */
    if (nb == 0) {
	struct core_region *cr;
	void *base = NULL;
	for (cr = core_tab; cr->npg != 0; cr++)
	    if (! base || cr->base < base)
		base = cr->base;
	return base;
    }

    if (nb < 0)
	/* can't return pages */
	return (void *)-1;
	
    /* round up to page multiple (malloc always asks for page multiples) */
    nb = (nb + PAGESIZE - 1) & ~PAGEMASK;

    ret = pthread_getpages_np (nb);
    if (!ret)
	return (void *)-1;
    return ret;
}

void *
sbrk (int nb) 
{
    void *p;

    p = _sbrk(nb);

    /* sbrk defined to return zeroed pages */
    if (nb > 0)
	memset (p, 0, nb);

   return p;
}
