#ident "sde/rtpx/rtpx_mips.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1997-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_mips.c: cpu-specific RTPX code for MIPS cpus (with MIPS SDE)
 */


#include "pthread_impl.h"
#include <sdethread.h>
#include <libc_thread.h>
#include <string.h>
#include <errno.h>
#include <mips/cpu.h>
#include <mips/fpa.h>
#include <mips/xcpt.h>
#define _KERNEL
#include <sys/filedesc.h>

#define SR_VALID	(SR_IMASK|SR_IE)

#ifdef SR_CEE
#define SR_ENABLES	(SR_IMASK | SR_CU1 | SR_CEE)
#else
#define SR_ENABLES	(SR_IMASK | SR_CU1)
#endif

/* per-cpu state */
struct cpu      _cpuTab[_SBD_CPU_MAX];

/* switch off floating-point support if there's definitely no fpu */
#if #fpu(no)
#undef _RTPX_FLOATING_POINT
#endif

#ifdef _RTPX_FLOATING_POINT
/* who's got the fpu registers? */
static rtpx_thread_t *fpaOwner = NULL;

/* default initial fpa control & status register */
static unsigned int rtpx_fpa_csr;

/* enable fpu "performance" mode - flush denormals */
int _rtpx_fpa_flush __attribute__((weak)) = 0; 
#endif

#ifndef SR_CEE
#undef _RTPX_COREXTEND
#endif
#ifndef CFG3_DSPP
#undef _RTPX_DSPASE
#endif

#ifdef _RTPX_COP2
/* who's got the COP2 registers? */
static rtpx_thread_t *cop2Owner = NULL;
#endif

#ifdef _RTPX_COREXTEND
/* who's got the CorExtend registers? */
static rtpx_thread_t *cexOwner = NULL;
#endif

#ifdef _RTPX_DSPASE
/* who's got the DSP registers? */
static rtpx_thread_t *dspOwner = NULL;
#endif


/* return current time-of-day from hardware */
time_t _cpu_time (time_t * tp)
{
    extern time_t   _sbd_gettod ();
    time_t          t = _sbd_gettod ();
    if (t < 0)		
	/* handle error from _sbd_gettod */
	t = 0;
    if (tp)
	*tp = t;
    return t;
}


/* enter kernel mode by switching to IPL1 and returning executing thread */
rtpx_thread_t  *
_cpu_enter (void)
{
    mips_setsr (_sbd_ipl1_sr);
    _sbd_intrenter ();
    return _cpuTab[0].runThread;
}


/* leave kernel mode by switching from IPL1 to IPL0 */
int
_cpu_leave (rtpx_thread_t *cth, int v)
{
    assert (((mips_getsr () ^ _sbd_ipl1_sr) & SR_VALID) == 0);
    _sbd_intrleave ();
    mips_setsr (_sbd_ipl0_sr | cth->th_cpu. cpu_sr | SR_IE);
    return v;
}


/* start a new thread */
static void
cpu_start (int a0, int a1, int a2, int a3,
	   void *(*fun) (void *), void *arg)
{
    rtpx_thread_t *cth = rtpx_self ();

    /* Initial value of errno */
    cth->th_errno = 0;

    /* We may have been signalled or cancelled before we first get
       scheduled, so we must use this function to handled that
       possibility. */
    _rtpx_leaveWithValueLong (cth, 0);
    
    /* Now at user level. Call thread entrypoint function and exit if
       it returns */
    pthread_exit ((*fun) (arg));
}


/* initialise a new thread's stack and other cpu-specific state */
void
_cpu_thinit (rtpx_thread_t *ath, void *(*fun) (), void *arg)
{
    ath->th_cpu.cpu_xcp = NULL;
    ath->th_cpu.cpu_sr = 0;

    {
	void          **sbase =
	    (void **) &ath->th_stack[ath->th_tha.tha_stacksize];

	/* align sbase on 8 byte boundary */
	sbase = (void **) ((unsigned long) sbase & ~0x7);

	/* 2 args  */
	*--sbase = arg;
	*--sbase = (void *) fun;

	/* reg arg space */
	*--sbase = 0;
	*--sbase = 0;
	*--sbase = 0;
	*--sbase = 0;

	assert (((unsigned) sbase & 0x7) == 0);

	ath->th_cpu.cpu_gprCtx[XJB_RA] = (int) cpu_start;
	ath->th_cpu.cpu_gprCtx[XJB_SP] = (int) sbase;
	ath->th_cpu.cpu_gprCtx[XJB_SR] = _sbd_ipl1_sr;
	ath->th_cpu.cpu_dbgnotified = 0;
    }

#ifdef _RTPX_FLOATING_POINT
    ath->th_cpu.cpu_fpaCtx.sr = rtpx_fpa_csr;
#endif
}


/* hook called when thread is destroyed */
void
_cpu_thdestroy (rtpx_thread_t *ath)
{
#ifdef _RTPX_FLOATING_POINT
    if (ath == fpaOwner) {
	ath->th_cpu.cpu_sr &= ~SR_CU1;
	fpaOwner = NULL;
    }
#endif
#ifdef _RTPX_COP2
    if (ath == cop2Owner) {
	ath->th_cpu.cpu_sr &= ~SR_CU2;
	cop2Owner = NULL;
    }
#endif
#ifdef _RTPX_COREXTEND
    if (ath == cexOwner) {
	ath->th_cpu.cpu_sr &= ~SR_CEU;
	cexOwner = NULL;
    }
#endif
#ifdef _RTPX_DSPASE
    if (ath == dspOwner) {
	ath->th_cpu.cpu_sr &= ~SR_MX;
	dspOwner = NULL;
    }
#endif
}


/* cpu-specific handling of context switch from CTH to NTH */
rtpx_thread_t  *
_cpu_swtch (rtpx_thread_t *cth, rtpx_thread_t *nth)
{
    struct xcptcontext *xcp;

    /* check in normal scheduler mode (IPL1) */
    assert (((mips_getsr () ^ _sbd_ipl1_sr) & SR_VALID) == 0);

    /* if we've not got an exception stack frame then save the register
       state now using xcptsetjmp. */
    if ((xcp = cth->th_cpu.cpu_xcp) == NULL) {
	if (xcptsetjmp (cth->th_cpu.cpu_gprCtx) != 0) {
	    /* if we get here, we've been rescheduled by a zcptlongjmp */
	    return _cpuTab[0].runThread;
	}
    }

    /* new thread is executing on this cpu */
    _cpuTab[0].runThread = nth;

#if 0
    /* context switch errno */
    cth->th_errno = errno;
    errno = nth->th_errno;
#endif

    /* install new process' libc data (no need to swap) */
    _libc_data = nth->th_proc->pr_cpu.cpu_libc;

    if ((xcp = nth->th_cpu.cpu_xcp) == NULL) {
	/* if it new thread has no exception stackframe then we switch to it 
	   with a xcptlongjmp (see xcptsetjmp above) */

	/* check new thread was at IPL1 */
	assert (((nth->th_cpu.cpu_gprCtx[XJB_SR] ^ _sbd_ipl1_sr)
		 & SR_VALID) == 0);

	if (REG_TO_VA (nth->th_cpu.cpu_gprCtx[XJB_SP])
	    <= (void *) nth->th_stack)
	    _rtpx_log (RTPXLOG_PANIC, "Stack overrun in thread '%s'",
		       nth->th_tha.tha_name);

	xcptlongjmp (nth->th_cpu.cpu_gprCtx, 1);

	/* NO REACH */
	abort ();

    }
    else {
	/* new thread was stopped by an interrupt */
	_sbd_intrleave ();
	nth->th_cpu.cpu_xcp = NULL;

	/* must be returning to ipl0, since thread can't be suspended by
	   interrupts when at ipl > 0 */
	xcp->sr = ((xcp->sr & ~SR_ENABLES)
		   | _sbd_ipl0_sr
		   | nth->th_cpu.cpu_sr);
	assert ((xcp->sr & SR_IMASK) != 0);

	if (REG_TO_VA (xcp->regs[XCP_SP]) <= (void *) nth->th_stack)
	    _rtpx_log (RTPXLOG_PANIC, "Stack overrun in thread '%s'",
		       nth->th_tha.tha_name);

	/* switch via exception return */
	xcptrestore (xcp);

	/* NO REACH */
	abort ();
    }
}

/* print a thread's cpu-specific state */
void            _cpu_printctx (rtpx_thread_t *);

void
_cpu_printctx (rtpx_thread_t *th)
{
    struct xcptcontext *xcp;

    if ((xcp = th->th_cpu.cpu_xcp) == NULL) {
	_rtpx_log (RTPXLOG_PRINTF, "pc=%p sp=%p ",
		   REG_TO_VA (th->th_cpu.cpu_gprCtx[XJB_RA]),
		   REG_TO_VA (th->th_cpu.cpu_gprCtx[XJB_SP]));
    }
    else {
	_rtpx_log (RTPXLOG_PRINTF, "pc=%p sp=%p ",
		   REG_TO_VA (xcp->epc), REG_TO_VA (xcp->regs[29]));

    }
}


/* handle high priority interupts, and dispatch low priority */
static int
xcptintHandler (int xcpt, struct xcptcontext *xcp)
{
    unsigned        pend = mips_getcr () & XCP_STATUS(xcp) & SR_IMASK;

    if (pend == 0)
	/* spurious interrupt on changing to ipl1 or ipl2 */
	return 0;

    if (_sbd_intrhandler (pend, xcp))
	/* board-specific code handled it, but may be at ipl1, so return
	   after handling it. */
	return 0;

    /* must have been at ipl0 */

    /* reduce interupt priority to scheduler priority (ipl1) * and handle
       interupt in sbd_evproc */
    assert ((pend & _sbd_ipl1_sr) == 0);
    mips_setsr (_sbd_ipl1_sr);
    {
	rtpx_thread_t  *cth = _cpuTab[0].runThread;
	assert (cth->th_cpu.cpu_xcp == NULL);
	cth->th_cpu.cpu_xcp = xcp;

	_rtpx_eoi (cth);
	mips_setsr (_sbd_ipl1_sr);	/* eoi returned us to ipl0 */

	_sbd_intrleave ();

	/* must be returning to ipl0 */
	assert (cth->th_cpu.cpu_xcp == NULL);
	xcp->sr = ((xcp->sr & ~SR_ENABLES)
		   | _sbd_ipl0_sr
		   | cth->th_cpu.cpu_sr);
	assert ((xcp->sr & SR_IMASK) != 0);
	return 0;
    }
}


#ifndef _mips_intdisable

/* 
 * "Safe" interrupt disable
 */
int
_mips_intdisable (void)
{
    /* note: not atomic, so may return out of date status register */
    return mips_xchsr (_sbd_iplhi_sr);
}


void
_mips_intrestore (int oldsr)
{
    unsigned int    newsr;

    switch (_sbd_probable_ipl (oldsr)) {
    case 2:
	return;			/* no changed */
    case 1:
    default:
	newsr = oldsr;
	break;
    case 0:
	{
	    /* recompute i/u mask for current thread. */
	    rtpx_thread_t  *cth = rtpx_self ();
	    newsr = _sbd_ipl0_sr | cth->th_cpu.cpu_sr | SR_IE;
	}
	break;
    }

    mips_setsr (newsr);
}
#endif




/* 
 * MIPS SDE compatibility interrupt support
 */

/* XXX not implemented yet, but avoid dragging in MIPS SDE code */
int
intraction (unsigned int intrno, const struct intraction *new,
	    struct intraction *old)
{
    _rtpx_log (RTPXLOG_PANIC, "intraction() not implemented");
    return 0;
}

xcpt_t
intrupt (unsigned int intrno, xcpt_t handler, int arg)
{
    struct intraction new,
                    old;
    int             err;

    if (handler == XCPT_DFL) {
	/* remove handler */
	new.ia_handler = NULL;	/* XXX */
	new.ia_arg = 0;
	new.ia_ipl = 0;
    }
    else {
	/* add handler */
	new.ia_handler = handler;
	new.ia_arg = arg;
	new.ia_ipl = intrno + 1;
    }

    if (err = intraction (intrno, &new, &old)) {
	extern int      errno;
	errno = err;
	return XCPT_ERR;
    }

    return old.ia_handler;
}



#ifdef _RTPX_FLOATING_POINT

#if __mipsfp64 || __mips_fpr == 64
#define SR_FPU	(SR_CU1 | SR_FR)
#else
#define SR_FPU	(SR_CU1)
#endif

static int 
fpa_avail (void)
{
    int avail = fpa_enable (_rtpx_fpa_flush);
    if (avail)
	rtpx_fpa_csr = fpa_getsr();
    return avail;
}

/* 
 * This is with interupts enabled, but imask=0
 */
static int
xcptcop1Handler (struct xcptcontext *xcp)
{
    rtpx_thread_t  *cth = rtpx_self ();
    rtpx_thread_t  *oth = fpaOwner;

    assert ((cth->th_cpu.cpu_sr & SR_CU1) == 0);
    cth->th_cpu.cpu_sr |= SR_FPU;
    mips_bissr (SR_FPU);

    if (cth != oth) {
	if (oth) {
	    /* take away fpa from owner */
	    assert (oth->th_cpu.cpu_sr & SR_CU1);
	    fpa_save (&oth->th_cpu.cpu_fpaCtx);
	    oth->th_cpu.cpu_sr &= ~SR_CU1;
	}
	/* give it to current thread */
	fpa_restore (&cth->th_cpu.cpu_fpaCtx);
	cth->th_cpu.cpu_sr |= SR_FPU;
	xcp->sr |= SR_FPU;
	fpaOwner = cth;
    }
    return 0;
}
#endif

#ifdef _RTPX_COP2
/* 
 * This is with interupts enabled, but imask=0
 */
static int
xcptcop2Handler (struct xcptcontext *xcp)
{
    rtpx_thread_t  *cth = rtpx_self ();
    rtpx_thread_t  *oth = cop2Owner;

    assert ((cth->th_cpu.cpu_sr & SR_CU2) == 0);
    cth->th_cpu.cpu_sr |= SR_CU2;
    mips_bissr (SR_CU2);

    if (cth != oth) {
	if (oth) {
	    /* take away fpa from owner */
	    assert (oth->th_cpu.cpu_sr & SR_CU2);
	    cop2_save (&oth->th_cpu.cpu_cop2Ctx);
	    oth->th_cpu.cpu_sr &= ~SR_CU2;
	}
	/* give it to current thread */
	cop2_restore (&cth->th_cpu.cpu_cop2Ctx);
	cth->th_cpu.cpu_sr |= SR_CU2;
	xcp->sr |= SR_CU2;
	cop2Owner = cth;
    }
    return 0;
}
#endif

#ifdef _RTPX_COREXTEND
/* 
 * This is with interupts enabled, but imask=0
 */
static int
xcptceuHandler (int ino, struct xcptcontext *xcp)
{
    rtpx_thread_t  *cth = rtpx_self ();
    rtpx_thread_t  *oth = cexOwner;

    assert ((cth->th_cpu.cpu_sr & SR_CEU) == 0);
    cth->th_cpu.cpu_sr |= SR_CEU;
    mips_bissr (SR_CEU);

    if (cth != oth) {
	if (oth) {
	    /* take away fpa from owner */
	    assert (oth->th_cpu.cpu_sr & SR_CEU);
	    _corextend_save (&oth->th_cpu.cpu_udiCtx);
	    oth->th_cpu.cpu_sr &= ~SR_CEU;
	}
	/* give it to cth */
	_corextend_restore (&cth->th_cpu.cpu_udiCtx);
	cth->th_cpu.cpu_sr |= SR_CEU;
	xcp->sr |= SR_CEU;
	cexOwner = cth;
    }
    return 0;
}
#endif

#ifdef _RTPX_DSPASE

/* 
 * This is with interupts enabled, but imask=0
 */
static int
xcptdspuHandler (int ino, struct xcptcontext *xcp)
{
    rtpx_thread_t  *cth = rtpx_self ();
    rtpx_thread_t  *oth = cexOwner;

    assert ((cth->th_cpu.cpu_sr & SR_MX) == 0);
    cth->th_cpu.cpu_sr |= SR_MX;
    mips_bissr (SR_MX);

    if (cth != oth) {
	if (oth) {
	    /* take away fpa from owner */
	    assert (oth->th_cpu.cpu_sr & SR_MX);
	    _dsp_save (&oth->th_cpu.cpu_dspCtx);
	    oth->th_cpu.cpu_sr &= ~SR_MX;
	}
	/* give it to cth */
	_dsp_restore (&cth->th_cpu.cpu_dspCtx);
	cth->th_cpu.cpu_sr |= SR_MX;
	xcp->sr |= SR_MX;
	dspOwner = cth;
    }
    return 0;
}
#endif

#if defined(_RTPX_FLOATING_POINT) || defined(_RTPX_COP2)
static int
xcptcopHandler (int xno, struct xcptcontext *xcp)
{
#ifdef _RTPX_FLOATING_POINT
    if ((xcp->cr & CR_CEMASK) == 1 << CR_CESHIFT)
	return xcptcop1Handler (xcp);
#endif
#ifdef _RTPX_COP2
    if ((xcp->cr & CR_CEMASK) == 2 << CR_CESHIFT)
	return xcptcop2Handler (xcp);
#endif
    return SIGILL;
}
#endif

static pthread_mutex_t cpuMx = PTHREAD_MUTEX_INITIALIZER;

/* prevent/allow PROM monitor handling SIGNO related exceptions */
void
_cpu_sigcatch (int signo, int enable)
{
    pthread_mutex_lock (&cpuMx);
    _xcpt_sigcatch (signo, enable);
    pthread_mutex_unlock (&cpuMx);
}


/* hook for MIPS SDE xcpt.c */
void
_sig_raise (pid_t pgrp, int signo, struct xcptcontext *xcp)
{
    rtpx_thread_t  *cth;
    struct sigevent sev;
    unsigned int    osr;

    /* give up now if not threading yet */
    if (!__isthreaded)
	_cpu_sigabort (NULL, signo, xcp);

    if (!xcp) {
	/* a normal async s/w generated signal (e.g. ^C) */
	if (pgrp > 0) {
	    union sigval    sival;
	    sival.sival_int = 0;
	    _rtpx_kill (-pgrp, signo, sival, SI_KERNEL);
	}
	return;
    }

    cth = rtpx_self ();

    /* this is a synchronous machine fault, where did it occur  */
    osr = xcp->sr & SR_VALID;
    if (!cth
	|| osr == (_sbd_ipl1_sr & SR_VALID)
	|| osr == (_sbd_iplhi_sr & SR_VALID)) {
	/* was in "kernel mode", abort now */
	if (cth && cth->th_cpu.cpu_xcp == NULL)
	    cth->th_cpu.cpu_xcp = xcp;
	_cpu_sigabort (cth, signo, xcp);
    }

    /* was in "user mode", raise an *immediate* signal */

    /* switch to ipl1 (kernel mode) */
    mips_setsr (_sbd_ipl1_sr);
    assert (cth->th_cpu.cpu_xcp == NULL);
    cth->th_cpu.cpu_xcp = xcp;

    /* send the signal - must be to this thread */
    sev.sigev_notify = SIGEV_SIGNAL;
    sev.sigev_signo = signo;
    sev.sigev_value.sival_ptr = xcp;	/* pass pointer to exception info */
    _rtpx_sigqueue (cth->th_proc, SI_FAULT, &sev);

    /* handle the signal now */
    _rtpx_sigcheck (cth);

    /* return to task if signal handled, ensuring correct i/u mask */
    assert (cth->th_cpu.cpu_xcp == xcp);
    xcp->sr = ((xcp->sr & ~SR_ENABLES)
	       | _sbd_ipl0_sr
	       | cth->th_cpu.cpu_sr);
    assert ((xcp->sr & SR_IMASK) != 0);
}


/* opportunity to report signal via debugger, and choose to ignore */
int
_cpu_sighook (rtpx_thread_t *cth, int signo, int fatal)
{
    /* only pass to debugger if debugger will be fatal, or if
       the debugger is already connected */
    if (fatal || _debug_connected ()) {
	mips_setsr (_sbd_iplhi_sr);
	signo = _debug_sighook (signo);
	mips_setsr (_sbd_ipl1_sr);
    }
    return signo;
}


/* give up on a fatal signal */
void
_cpu_sigabort (rtpx_thread_t *th, int signo, struct xcptcontext *excp)
{
    extern const char *const sys_siglist[];
    struct xcptcontext *xcp;

    mips_setsr (_sbd_iplhi_sr);

    /* tell the debugger */
    _debug_exit (signo << 8);

    /* now tell the poor user */
    _rtpx_log (RTPXLOG_PRINTF, "\n");
    if (signo > 0 && signo < SIGRTMIN)
	_rtpx_log (RTPXLOG_PRINTF, sys_siglist[signo]);
    else if (signo >= SIGRTMIN && signo <= SIGRTMAX)
	_rtpx_log (RTPXLOG_PRINTF, "Realtime %d", signo - SIGRTMIN);
    else
	_rtpx_log (RTPXLOG_PRINTF, "Unknown %d", signo);
    if (th) {
	_rtpx_log (RTPXLOG_PRINTF, " signal in thread \"%s\"",
		   th->th_tha.tha_name);
	if ((xcp = excp) || (xcp = th->th_cpu.cpu_xcp)) {
	    _rtpx_log (RTPXLOG_PRINTF, "\n");
	    xcpt_show (xcp);
	}
	else if (th != rtpx_self () && th->th_waitq) {
	    _rtpx_log (RTPXLOG_PRINTF, " at pc=0x%lx sp=0x%lx\n",
		       (unsigned long) th->th_cpu.cpu_gprCtx[XJB_RA],
		       (unsigned long) th->th_cpu.cpu_gprCtx[XJB_SP]);
	}
	else {
	    _rtpx_log (RTPXLOG_PRINTF, "\n");
	}
    }
    else {
	_rtpx_log (RTPXLOG_PRINTF, " signal, before RTOS startup\n");
	if (excp)
	    xcpt_show (excp);
    }

    {
	extern void     __exit (int) __attribute__ ((noreturn));
	mips_setsr (_sbd_iplhi_sr);
	__exit (signo << 8);
    }
}

/* cpu-specific first process intialisation */
void
_cpu_procfirst (rtpx_proc_t *pr)
{
    pr->pr_cpu.cpu_fds = &_fileio_fd0.fd_fd;
    pr->pr_cpu.cpu_fds->fd_refcnt++;
    pr->pr_cpu.cpu_libc = _libc_data;
    pr->pr_cpu.cpu_first = 1;
}

/* cpu-specific process creation */
int
_cpu_procnew (rtpx_proc_t *ppr, rtpx_proc_t *pr)
{
    pr->pr_cpu.cpu_first = 0;

    /* clone file descriptors */
    pr->pr_cpu.cpu_fds = _fileio_fdcopy (ppr->pr_cpu.cpu_fds);
    if (!pr->pr_cpu.cpu_fds) 
	return ENOMEM;

    /* get new libc state */
    pr->pr_cpu.cpu_libc = _libc_new ();
    if (!pr->pr_cpu.cpu_libc) {
	_fileio_fdfree (pr->pr_cpu.cpu_fds);
	return ENOMEM;
    }

    /* we are the new process, start using the new libc state */
    _libc_data = pr->pr_cpu.cpu_libc;

    return 0;
}

/* cpu-specific process destruction */
void
_cpu_procdestroy (rtpx_proc_t *pr)
{
    if (pr->pr_cpu.cpu_libc && !pr->pr_cpu.cpu_first) {
    	_libc_delete (pr->pr_cpu.cpu_libc);
	pr->pr_cpu.cpu_libc = NULL;
    }

    if (pr->pr_cpu.cpu_fds) {
	_fileio_fdfree (pr->pr_cpu.cpu_fds);
	pr->pr_cpu.cpu_fds = NULL;
    }
}


/* Return current process' file descriptor table for file i/o system. */
struct filedesc *
_fileio_getfdp ()
{
    rtpx_proc_t    *pr = _rtpx_self_proc ();
    return pr->pr_cpu.cpu_fds;
}


/* final termination when all threads finished */
void
_cpu_exit (int rv)
{
    extern void _quit (int) __attribute__ ((noreturn));
    _quit (rv);		/* see kit/share/crt0.sx */
}


/* return base and limit of main program stack, which with the sbrk
   implementation in MIPS SDE is the gap between the _maxbrk (top of heap)
   and the current stack-pointer. */
void
_cpu_stacklimits (void **addr, unsigned int *size)
{
    void	   *sp;
    extern void    *_minbrk;
    extern void    *_maxbrk;
    extern char    _end[];
    void	   *base;
    long            nb;

    /* make sure sbrk has set up the heap limits */
    _sbrk_init ();

    __asm__ ("move %0,$sp" : "=d" (sp));
    if (sp >= _maxbrk) {
	/* stack above _maxbrk */
	base = _maxbrk;
    }
    else if (sp < _minbrk && sp > (void *)_end) {
	/* stack between end and _minbrk */
	base = _end;
    }
    else {
	/* unknown */
	base = sp;
    }

    nb = sp - base;
    nb -= 32;			/* safety gap */
    if (nb <= 32) {
	*addr = 0;
	*size = 0;
    }
    else {
	*addr = _maxbrk;
	*size = nb;
    }
}

extern int	_cop2_avail (void);
extern int	_corextend_avail (void);
extern int	_dspase_avail (void);

struct coproc {
    int		(*avail) (void);
    xcpt_t	handler;
    unsigned int xcptno;
    unsigned int srbit;
};

static const struct coproc coproc[] = {
#ifdef _RTPX_FLOATING_POINT
    {fpa_avail,		xcptcopHandler,		XCPTCPU,	SR_CU1},
#endif
#ifdef _RTPX_COP2
    {_cop2_avail, 	xcptcopHandler,		XCPTCPU,	SR_CU2},
#endif
#ifdef _RTPX_COREXTEND
    {_corextend_avail,	xcptceuHandler, 	XCPTCEU,	SR_CEE},
#endif
#ifdef _RTPX_DSPASE
    {_dspase_avail, 	xcptdspuHandler,	XSCPTDSPU,	SR_MX},
#endif
    {0, 0, 0, 0}
};

/* 
 * CPU-specific initialisation
 */
void
_cpu_rtpxinit (void)
{
    unsigned int clr = SR_CU1;
    const struct coproc *cop;

    /* install our own interrupt exception handler */
    xcption (XCPTINTR, xcptintHandler);

    /* install coprocessor handlers */
    for (cop = coproc; cop->avail; cop++) {
	if ((cop->avail) ())
	    /* install a trap handler to do lazy context switch, if
	       we've got a hardware coprocessor. */
	    xcption (cop->xcptno, cop->handler);
	clr |= cop->srbit;
    }

    /* disable all coprocessors initially */
    mips_bicsr (clr);

    /* make sure that our copy of /dev/tty* gets loaded */
    {
	extern void     _ttydev_load (void);
	_ttydev_load ();
    }
}



/* 
 * Interface to threads debugger 
 */

#include "kit_impl.h"

static struct xcptcontext gxcpt;
static struct xcptcontext *gx;
static rtpx_thread_t *gth;

static struct xcptcontext cxcpt;
static struct xcptcontext *cx;
static rtpx_thread_t *cth;

static struct xcptcontext *ox;
static rtpx_thread_t *oth;

/* ask dbg.c to disable interrupts during single-step */
int _dbg_threadsafe_sstep = 1;

/* flag set when thread kernel is initialised */
extern int __isthreaded;

int
_dbg_getthread (struct xcptcontext *xcp)
{
    /* return cleanly if not threading yet */
    if (!__isthreaded)
	return -1;

    gth = cth = oth = rtpx_self ();
    /* XXX assumes xcp points to self */
    gx = cx = ox = xcp;
    if (!cth)
	return -1;
    cth->th_cpu.cpu_dbgnotified = 1;
    return cth->th_id;
}


struct xcptcontext *
_dbg_setthread (struct xcptcontext *xcp, int general, int thread)
{
    rtpx_thread_t  *th;
    struct xcptcontext *x;

    /* return cleanly if not threading yet */
    if (!__isthreaded)
	return xcp;

    if (thread < 0)
	th = rtpx_self ();
    else
	th = _rtpx_findthread (thread);

    if (!th)
	return NULL;

    if (th == rtpx_self ()) {
	/* it's us - use original debug context */
	x = ox;
    }
    else if ((x = th->th_cpu.cpu_xcp) == NULL) {
	/* blocked in rtpx kernel, invent a dummy xcption context */
	reg_t          *xjb = th->th_cpu.cpu_gprCtx;
	x = general ? &gxcpt : &cxcpt;
	memset (x, 0, sizeof (*x));
	x->regs[16] = xjb[XJB_S0];
	x->regs[17] = xjb[XJB_S1];
	x->regs[18] = xjb[XJB_S2];
	x->regs[19] = xjb[XJB_S3];
	x->regs[20] = xjb[XJB_S4];
	x->regs[21] = xjb[XJB_S5];
	x->regs[22] = xjb[XJB_S6];
	x->regs[23] = xjb[XJB_S7];
	x->regs[29] = xjb[XJB_SP];
	x->regs[30] = xjb[XJB_S8];
	x->regs[31] = xjb[XJB_RA];
	x->sr = xjb[XJB_SR];
	x->epc = xjb[XJB_RA];
	x->signo = 0;
    }

    if (general) {
	gx = x;
	gth = th;
    }
    else {
	cx = x;
	cth = th;
    }

    th->th_cpu.cpu_dbgnotified = 1;

    return x;
}

struct xcptcontext *
_dbg_contthread (struct xcptcontext *x)
{
    /* can only continue original thread */
    return ox;
}

int
_dbg_querythread (int thread)
{
    /* return cleanly if not threading yet */
    if (!__isthreaded)
	return 0;
    return (_rtpx_findthread (thread) != NULL);
}


int
_dbg_newthread (void)
{
    rtpx_thread_t  *th;

    /* return cleanly if not threading yet */
    if (!__isthreaded)
	return -1;
    for (th = _rtpx_firstthread (); th != NULL; th = _rtpx_nextthread (th))
	if (!th->th_cpu.cpu_dbgnotified)
	    return th->th_id;
    return -1;
}


#if 0
/* now handled in sig_raise above */
int
_dbg_restart (struct xcptcontext *xcp, int signo)
{
    if (signo) {
	if (xcp == ox) {
	    /* send signal to currently running thread */
	    /* cannot do this yet! */ ;
	    signo = 0;
	}
	else {
	    /* send signal to currently selected execution thread */
	    _rtpx_resumeWithSig (cth, signo, NULL);
	    signo = 0;
	}
    }

    /* always return to thread which interrupted us */
    return signo;
}
#endif

/* SDETHREAD per-thread errno hooks */

/* helper function for assembler routines to set errno */
__attribute__((__nomips16__))
int
__sdethread_set_errno (int err, int rv)
{
    errno = err;
    return rv;
}

#undef errno

int *
__sdethread_errno_pointer (void)
{
    if (__isthreaded) {
	rtpx_thread_t *cth = rtpx_self ();
	return &cth->th_errno;
    }
    return &errno;
}

#include <sys/init.h>

#ifndef __OPTIMIZE__
/* When optimizing the call to sdethread_too_small should be deleted,
   and if not will call a link error.  When not optimizing we need to
   map to a function which does exist and generates a run-time. */
#define __sdethread_mismatch abort
#else
extern void __sdethread_mismatch (void);
#endif

static void
sde_pthread_init ()
{
    extern void     _pthread_init (void);

    /* Make sure that we can store the underlying pthread data
       structures inside the sdethread data structures */
    if (sizeof (__sdethread_mutex_t) < sizeof (pthread_mutex_t)
	|| sizeof (__sdethread_once_t) < sizeof (pthread_once_t)
	|| sizeof (__sdethread_sigstate_t) < sizeof (pthread_sigstate_np_t))
	__sdethread_mismatch ();

    _pthread_init ();
}
DECL_INIT_FUNC (IF_RTOS, sde_pthread_init)

/* effectively catch the final call to _exit() */
static void
sde_pthread_exit ()
{
    struct sched_param sp;
    sigset_t        nm;

    /* prevent any signals while exiting */
    sigfillset (&nm);
    pthread_sigmask (SIG_SETMASK, &nm, NULL);

    /* cancellation too */
    pthread_setcancelstate (PTHREAD_CANCEL_DISABLE, NULL);

    /* set to highest prio - keep other threads outta the way */
    pthread_getschedparam (pthread_self (), NULL, &sp);
    sp.sched_priority = sched_get_priority_max (SCHED_FIFO);
    pthread_setschedparam (pthread_self (), SCHED_FIFO, &sp);
}
DECL_FINI_FUNC (IF_PREMAIN, sde_pthread_exit)

/* final cleanup before return to monitor */
static void
sde_pthread_fini ()
{
    /* stop all scheduling activity */
    mips_setsr (_sbd_iplhi_sr);
}
DECL_FINI_FUNC (IF_RTOS, sde_pthread_fini)
