#ident "sde/rtpx/rtpx_mqueue.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1997-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_mqueue.c: POSIX.4 message queues 
 */


/* 
 * XXX This implementation works only within a shared address space, 
 * using thread synchronisation.
 */

#include "pthread_impl.h"
#include <mqueue.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <sys/time.h>
#include <sys/fcntl.h>
#include <sys/queue.h>

struct mq_msg {
    struct mq_msg  *msg_next;
    int             msg_prio;
    char           *msg_data;
    size_t          msg_len;
};

struct _mq_opaque {
    LIST_ENTRY(_mq_opaque) mq_link;
    char            *mq_name;
    struct mq_attr  mq_attr;
    struct sigevent mq_sigev;
    pid_t           mq_pid;
    int		    mq_refcnt;
    int		    mq_unlinked;
    pthread_mutex_t mq_mx;
    pthread_cond_t  mq_rxcd;
    pthread_cond_t  mq_txcd;
    int             mq_rxwait;
    int             mq_txwait;
    struct mq_msg  *mq_first;
    struct mq_msg  *mq_last;
    struct mq_msg  *mq_free;
    struct mq_msg   mq_msgs[0];
};

static LIST_HEAD(, _mq_opaque) mqhead = LIST_HEAD_INITIALIZER(mqhead);
static pthread_mutex_t mqh_mx = PTHREAD_MUTEX_INITIALIZER;

static void
mqh_cleanup (void *arg)
{
    pthread_mutex_unlock (&mqh_mx);
}


mqd_t mq_open (const char *name, int oflags, ...)
{
    mqd_t           mq;
    struct mq_attr *attr;
    struct mq_msg  *msg;
    va_list         ap;
    mode_t	    mode;
    int             nmsg,
                    ocancel;
    char           *data;
    sigset_t        sm;
    size_t          namlen, n;

    namlen = strlen (name);
    if (namlen > PATH_MAX) {
	errno = ENAMETOOLONG;
	return (mqd_t) - 1;
    }

    sigfillset (&sm);
    pthread_sigmask (SIG_BLOCK, &sm, &sm);
    pthread_setcanceltype (PTHREAD_CANCEL_DEFERRED, &ocancel);
    pthread_cleanup_push (mqh_cleanup, NULL);
    pthread_mutex_lock (&mqh_mx);

    LIST_FOREACH(mq, &mqhead, mq_link)
	if (strcmp (name, mq->mq_name) == 0)
	    break;

    if (mq != LIST_END(&mqhead)) {
	/* got existing mqueue */
	if ((oflags & (O_CREAT | O_EXCL)) == (O_CREAT | O_EXCL)) {
	    /* can't exclusive-create if it already exists */
	    errno = EEXIST;
	    mq = (mqd_t) (-1);
	}
	else {
	    /* OK, bump ref count */
	    pthread_mutex_lock (&mq->mq_mx);
	    mq->mq_refcnt++;
	    pthread_mutex_unlock (&mq->mq_mx);
	}
	goto ret;
    }

    /* not found - make a new mqueue ? */

    if (!(oflags & O_CREAT)) {
	/* doesn't exists, and not asked to create it */
	errno = ENOENT;
	mq = (mqd_t) (-1);
	goto ret;
    }

    va_start (ap, oflags);
    mode = va_arg (ap, int);
    attr = va_arg (ap, struct mq_attr *);
    va_end (ap);

    nmsg = attr->mq_maxmsg;

    n = sizeof *mq + + namlen + 1 
	+ (sizeof (struct mq_msg) + attr->mq_msgsize) * nmsg;
    mq = (mqd_t) malloc (n);
    if (!mq) {
	errno = ENFILE;
	mq = (mqd_t) (-1);
	goto ret;
    }
    memset (mq, 0, n);

    /* store name */
    mq->mq_name = (char *) &mq->mq_msgs[nmsg];
    memcpy (mq->mq_name, name, namlen + 1);
    mq->mq_refcnt = 1;

    /* initialise attributes */
    mq->mq_attr.mq_maxmsg = attr->mq_maxmsg;
    mq->mq_attr.mq_msgsize = attr->mq_msgsize;
    mq->mq_attr.mq_flags = (oflags & O_NONBLOCK) ? MQ_NONBLOCK : 0;
    mq->mq_attr.mq_curmsgs = 0;

    /* set up message free list */
    data = (char *)mq->mq_name + namlen + 1;
    for (msg = mq->mq_msgs; msg < &mq->mq_msgs[nmsg]; msg++) {
	msg->msg_next = mq->mq_free;
	mq->mq_free = msg;
	msg->msg_data = data;
	data += attr->mq_msgsize;
    }

    /* set up mutex etc */
    pthread_mutex_init (&mq->mq_mx, NULL);
    pthread_cond_init (&mq->mq_rxcd, NULL);
    pthread_cond_init (&mq->mq_txcd, NULL);

    /* add to namespace */
    LIST_INSERT_HEAD (&mqhead, mq, mq_link);

  ret:
    pthread_mutex_unlock (&mqh_mx);
    pthread_cleanup_pop (0);
    pthread_setcanceltype (ocancel, 0);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);
    return mq;
}


int
mq_close (mqd_t mq)
{
    sigset_t        sm;
    int             ret = 0;

    sigfillset (&sm);
    pthread_sigmask (SIG_BLOCK, &sm, &sm);
    pthread_mutex_lock (&mq->mq_mx);

    if (mq->mq_refcnt > 0) {
	/* clear any signal association */
	if (mq->mq_pid == _rtpx_self_proc ()->pr_pid)
	    mq->mq_pid = 0;
	/* if closing unlinked mq, now's the time to free it */
	if (--mq->mq_refcnt == 0 && mq->mq_unlinked) {
	    pthread_mutex_unlock (&mq->mq_mx);
	    pthread_mutex_destroy (&mq->mq_mx);
	    pthread_cond_destroy (&mq->mq_rxcd);
	    pthread_cond_destroy (&mq->mq_txcd);
	    free (mq);
	    mq = NULL;
	}
    }

    if (mq)
	pthread_mutex_unlock (&mq->mq_mx);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);
    return ret;
}


int
mq_unlink (const char *name)
{
    mqd_t           mq;
    sigset_t        sm;
    int		    ret = 0;

    sigfillset (&sm);
    pthread_sigmask (SIG_BLOCK, &sm, &sm);
    pthread_mutex_lock (&mqh_mx);

    LIST_FOREACH(mq, &mqhead, mq_link)
	if (strcmp (name, mq->mq_name) == 0)
	    break;

    if (mq == LIST_END(&mqhead)) {
	errno = ENOENT;
	ret = -1;
	goto done;
    }
	
    /* remove from namespace - no-one can then open it any more */
    LIST_REMOVE (mq, mq_link);

    /* either delete now, or mark for future deletion */
    pthread_mutex_lock (&mq->mq_mx);
    if (mq->mq_refcnt == 0) {
	/* no-one has a (valid) pointer to it any more */
	pthread_mutex_unlock (&mq->mq_mx);
	pthread_mutex_destroy (&mq->mq_mx);
	pthread_cond_destroy (&mq->mq_rxcd);
	pthread_cond_destroy (&mq->mq_txcd);
	free (mq);
    }
    else {
	/* mark as unlinked, so that close will free it */
	mq->mq_unlinked = 1;
	pthread_mutex_unlock (&mq->mq_mx);
    }

done:
    pthread_mutex_unlock (&mqh_mx);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);
    return ret;
}


static void
mq_txcleanup (void *arg)
{
    mqd_t           mq = arg;
    mq->mq_txwait--;
    pthread_mutex_unlock (&mq->mq_mx);
}


int
mq_send (mqd_t mq, const char *buffer, size_t buflen, unsigned int prio)
{
    long            msgsize = mq->mq_attr.mq_msgsize;
    struct mq_msg  *msg;
    int             ocancel,
                    rv = 0;

    if (buflen > msgsize || prio > MQ_PRIO_MAX) {
	errno = EINVAL;
	return -1;
    }

    pthread_setcanceltype (PTHREAD_CANCEL_DEFERRED, &ocancel);
    pthread_cleanup_push (mq_txcleanup, mq);
    pthread_mutex_lock (&mq->mq_mx);

    while (mq->mq_attr.mq_curmsgs == mq->mq_attr.mq_maxmsg) {
	/* queue is full */
	if (mq->mq_attr.mq_flags & MQ_NONBLOCK) {
	    errno = EAGAIN;
	    rv = -1;
	    goto ret;
	}
	/* wait for room in queue */
	mq->mq_txwait++;
	pthread_cond_wait (&mq->mq_txcd, &mq->mq_mx);
	mq->mq_txwait--;
    }

    /* remove first message off free list */
    msg = mq->mq_free;
    mq->mq_free = msg->msg_next;

    /* copy user data into message */
    memcpy (msg->msg_data, buffer, buflen);
    msg->msg_prio = prio;
    msg->msg_len = buflen;

    if (!mq->mq_first) {
	/* an empty queue */
	mq->mq_first = mq->mq_last = msg;
	msg->msg_next = NULL;
    }
    else if (prio <= mq->mq_last->msg_prio) {
	/* usual case: append to end */
	mq->mq_last->msg_next = msg;
	mq->mq_last = msg;
	msg->msg_next = NULL;
    }
    else {
	/* priority insertion */
	struct mq_msg  *p = mq->mq_first;
	if (prio > p->msg_prio) {
	    msg->msg_next = p;
	    mq->mq_first = msg;
	}
	else {
	    while (prio <= p->msg_prio)
		p = p->msg_next;
	    msg->msg_next = p->msg_next;
	    p->msg_next = msg;
	}
    }

    if (mq->mq_attr.mq_curmsgs++ == 0) {
	/* wake up anyone waiting */
	if (mq->mq_rxwait != 0)
	    pthread_cond_signal (&mq->mq_rxcd);
	else if (mq->mq_sigev.sigev_notify == SIGEV_SIGNAL)
	    _rtpx_sigraise (mq->mq_pid, SI_MESGQ, &mq->mq_sigev);
    }

  ret:
    pthread_mutex_unlock (&mq->mq_mx);
    pthread_cleanup_pop (0);
    pthread_setcanceltype (ocancel, 0);
    return rv;
}


static void
mq_rxcleanup (void *arg)
{
    mqd_t           mq = arg;
    mq->mq_rxwait--;
    pthread_mutex_unlock (&mq->mq_mx);
}


int
mq_timedreceive (mqd_t mq, char *buffer, size_t buflen, unsigned int *prio,
		 struct timespec *ts)
{
    long            msgsize = mq->mq_attr.mq_msgsize;
    struct mq_msg  *msg;
    int             ocancel,
                    rv = 0;

    if (buflen > msgsize) {
	errno = EINVAL;
	return -1;
    }

    pthread_setcanceltype (PTHREAD_CANCEL_DEFERRED, &ocancel);
    pthread_cleanup_push (mq_rxcleanup, mq);
    pthread_mutex_lock (&mq->mq_mx);

    while (mq->mq_attr.mq_curmsgs == 0) {
	int             res;
	if (mq->mq_attr.mq_flags & MQ_NONBLOCK) {
	    pthread_mutex_unlock (&mq->mq_mx);
	    errno = EAGAIN;
	    rv = -1;
	    goto ret;
	}
	mq->mq_rxwait++;
	res = pthread_cond_timedwait (&mq->mq_rxcd, &mq->mq_mx, ts);
	mq->mq_rxwait--;
	if (res != 0) {
	    errno = res;
	    rv = -1;
	    goto ret;
	}
    }

    /* remove first message off queue */
    msg = mq->mq_first;
    mq->mq_first = msg->msg_next;
    mq->mq_attr.mq_curmsgs--;

    /* copy message into user buffer */
    memcpy (buffer, msg->msg_data,
	    (buflen < msg->msg_len) ? buflen : msg->msg_len);
    if (prio)
	*prio = msg->msg_prio;

    /* put used message onto free list */
    msg->msg_next = mq->mq_free;
    mq->mq_free = msg;

    /* wake up anyone waiting for a free message buffer */
    if (mq->mq_txwait != 0)
	pthread_cond_signal (&mq->mq_txcd);

  ret:
    pthread_mutex_unlock (&mq->mq_mx);
    pthread_cleanup_pop (0);
    pthread_setcanceltype (ocancel, 0);
    return rv;
}


int
mq_receive (mqd_t mq, char *buffer, size_t buflen, unsigned int *prio)
{
    return mq_timedreceive (mq, buffer, buflen, prio, NULL);
}


int
mq_notify (mqd_t mq, const struct sigevent *sigev)
{
    sigset_t        sm;
    int             ret = 0;

    sigfillset (&sm);
    pthread_sigmask (SIG_BLOCK, &sm, &sm);
    pthread_mutex_lock (&mq->mq_mx);
    if (sigev) {
	if (mq->mq_sigev.sigev_notify == SIGEV_NONE) {
	    mq->mq_sigev = *sigev;
	    mq->mq_pid = _rtpx_self_proc ()->pr_pid;
	}
	else
	    ret = EBUSY;
    }
    else {
	if (mq->mq_pid != 0 && mq->mq_pid != _rtpx_self_proc ()->pr_pid)
	    ret = EBUSY;
	else {
	    mq->mq_sigev.sigev_notify = SIGEV_NONE;
	    mq->mq_pid = 0;
	}
    }
    pthread_mutex_unlock (&mq->mq_mx);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);
    return ret;
}


int
mq_setattr (mqd_t mq, const struct mq_attr *new, struct mq_attr *old)
{
    struct mq_attr  oattr;
    pthread_mutex_lock (&mq->mq_mx);
    if (old)
	oattr = mq->mq_attr;
    mq->mq_attr.mq_flags = new->mq_flags;
    if (old)
	*old = oattr;
    pthread_mutex_unlock (&mq->mq_mx);
    return 0;
}


int
mq_getattr (mqd_t mq, struct mq_attr *attr)
{
    pthread_mutex_lock (&mq->mq_mx);
    *attr = mq->mq_attr;
    pthread_mutex_unlock (&mq->mq_mx);
    return 0;
}
