#ident "sde/rtpx/rtpx_ns16550.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_ns16550.c: interrupt-driven serial i/o for NS16550 uart
 */


#define _KERNEL
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pthread.h>
#include <errno.h>
#include <assert.h>
#include <sys/fcntl.h>
#include <sys/termios.h>
#include <sys/signal.h>
#include <sys/time.h>

#include <mips/cpu.h>
#include <mips/xcpt.h>

#include "tty.h"
#include "sbd.h"
#include "sbdrtpx.h"
#include "sbdser.h"
#include "ns16550.h"

#ifndef NS16550_DEBUG
#define NS16550_DEBUG 0
#endif

/* private data */
static struct serial {
    void           *dev;
    unsigned char   unit;
    unsigned char   evn;	/* doorbell event number */
    unsigned char   fifolen;	/* length of tx/rx fifo */
    unsigned char   mcr;	/* soft copy of mcr register */
    /* the tty structure */
    struct tty      tty;
    /* thread control */
    pthread_cond_t  cd;		/* wait point */
    char            cdname[16];
    pthread_doorbell_t db;	/* interrupt doorbell */
    char            dbname[16];
    pthread_t       tid;	/* interrupt task id */
    char            thname[16];
    char            mxname[16];
    /* statistics */
    unsigned int    perrs;	/* parity errors */
    unsigned int    ferrs;	/* framing errors */
    unsigned int    oerrs;	/* overflow errors */
} serial[] = {
    {0, 0, INTR_UART0},
#ifdef UART1_BASE
    {0, 1, INTR_UART1},
#endif
#ifdef UART2_BASE
    {0, 2, INTR_UART2},
#endif
#ifdef UART3_BASE
    {0, 3, INTR_UART3},
#endif
};

#define dimof(arr) (sizeof(arr) / sizeof(arr[0]))


/* 
 * Public entrpoint table 
 */
static int      nsinit (int ch, int dev);
static int      nsopen (int ch, struct tty **);

struct sbdserint ns16550int = {
    "tty", dimof (serial),
    nsinit,
    nsopen,
};

/* global debug flags */
int             _rtpx_break_debug;


static void
nscleanup (void *arg)
{
    struct tty     *tp = arg;
    pthread_mutex_unlock ((pthread_mutex_t *)&tp->t_mx);
}


/* wait for a tx or rx event to occur */
/* XXX we should use different condition variables for each event */

static int
nswait (struct tty *tp, int waitflags, int tenths)
{
    struct serial  *ser = tp->t_addr;
    struct timespec ts;
    int             res;

    pthread_cleanup_push (nscleanup, tp);
    tp->t_state |= waitflags;
    if (tenths >= 0) {
	clock_gettime (CLOCK_REALTIME, &ts);
	ts.tv_sec += tenths / 10;
	ts.tv_nsec += (tenths % 10) * 100 * 1000 * 1000;
	if (ts.tv_nsec >= 1000000000)
	    ts.tv_sec++, ts.tv_nsec -= 1000000000;
	res = pthread_cond_timedwait (&ser->cd, (pthread_mutex_t *)&tp->t_mx, &ts);
    }
    else {
	res = pthread_cond_wait (&ser->cd, (pthread_mutex_t *)&tp->t_mx);
    }
    pthread_cleanup_pop (0);

    if (res == ETIME)
	_RTPX_ELOG ("nsuart%d: %d.%ds wait timeout", ser->unit,
		    tenths / 10, tenths % 10);
    else if (res != 0)
	_RTPX_ELOG ("nsuart%d: wait error: %s", ser->unit, strerror (res));

    return res;
}


static void
nswakeup (struct tty *tp, int flags)
{
    int             wakeup = 0;
    if (flags & FWRITE) {
	/* let buffer drain to lowat before restarting writing thread */
	if (tp->t_state & TS_TXWAIT && tp->t_outq.c_cc <= tp->t_lowat) {
	    tp->t_state &= ~TS_TXWAIT;
	    wakeup = 1;
	}
	if (tp->t_state & TS_TXDRAIN && tp->t_outq.c_cc == 0) {
	    tp->t_state &= ~TS_TXDRAIN;
	    wakeup = 1;
	}
#ifdef notyet
	/* wakeup on carrier detect */
	if (tp->t_state & TS_WOPEN
	    && (tp->t_state & TS_CARR_ON || tp->t_cflag & CLOCAL)) {
	    tp->t_state &= ~TS_WOPEN;
	    wakeup = 1;
	}
#endif
    }
    if (flags & FREAD) {
	if (tp->t_state & TS_RXWAIT) {
	    tp->t_state &= ~TS_RXWAIT;
	    wakeup = 1;
	}
    }
    if (wakeup) {
	struct serial  *ser = tp->t_addr;
	pthread_cond_broadcast (&ser->cd);
    }
}

/* 
 * Start transmitting to uart
 */

static void
nsstart (struct tty *tp)
{
    struct serial  *ser = tp->t_addr;
    void           *dp = ser->dev;
    int             len, out;

    if (tp->t_state & TS_BUSY)
	/* already transmitting */
	return;

    len = ser->fifolen;
    out = 0;
    do {
	if (tp->t_state & TS_XON_PEND) {
	    tp->t_state &= ~TS_XON_PEND;
	    NS16550_OUTB (dp, NS16550_DATA, tp->t_cc[VSTART]);
#if NS16550_DEBUG
	    _RTPX_ELOG ("xon>nsuart%d", ser->unit);
#endif
	    --len;
	}
	else if (tp->t_state & TS_XOFF_PEND) {
	    tp->t_state &= ~TS_XOFF_PEND;
	    NS16550_OUTB (dp, NS16550_DATA, tp->t_cc[VSTOP]);
#if NS16550_DEBUG
	    _RTPX_ELOG ("xoff>nsuart%d", ser->unit);
#endif
	    --len;
	}
	else if (!(tp->t_state & TS_TTSTOP)) {
	    char           *bp;
	    int             nb, i;

	    nb = _ttygetcblk (&tp->t_outq, &bp);
	    if (nb <= 0)
		break;
	    if (nb > len)
		nb = len;
#if NS16550_DEBUG
	    _RTPX_ELOG ("%d bytes>nsuart%d", nb, ser->unit);
#endif
	    for (i = nb; i != 0; i--) {
#if NS16550_DEBUG > 1
		char            c = *bp & 0x7f;
		_RTPX_ELOG ("%02x(%c)>nsuart%d",
			    c, (c < 0x20 || c >= 0x7f) ? '?' : c,
			    ser->unit);
#endif
		NS16550_OUTB (dp, NS16550_DATA, *bp++);
	    }

	    _ttydelcblk (&tp->t_outq, nb);
	    out += nb;
	    len -= nb;
	}
    } while (len > 0 && !(tp->t_state & TS_TTSTOP));

    if (len < ser->fifolen)
	tp->t_state |= TS_BUSY;
    if (out > 0)
	nswakeup (tp, FWRITE);
}


/* 
 * Main "interrupt" task
 */

static void *
nstask (void *arg)
{
    struct serial  *ser = arg;
    void           *dp = ser->dev;
    struct tty     *tp = &ser->tty;
    unsigned char   iir, lsr, msr;
    int             c;
    sigset_t        sm;

    /* disable all signals in this task */
    sigfillset (&sm);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);

    /* get the lock */
    pthread_mutex_lock ((pthread_mutex_t *)&tp->t_mx);

    /* enable uart interrupts */
    NS16550_OUTB (dp, NS16550_IER, IER_ETXRDY | IER_ERXRDY |
		  IER_ERLS | IER_EMSC);

    while (1) {

	/* wait for an interrupt */
	pthread_doorbell_wait (&ser->db, (pthread_mutex_t *)&tp->t_mx);

	while ((iir =
		NS16550_INB (dp, NS16550_IIR) & IIR_IMASK) != IIR_NOPEND) {
	    switch (iir) {
	    case IIR_RLS:
		lsr = NS16550_INB (dp, NS16550_LSR);
		goto rx;

	    case IIR_RXTOUT:
	    case IIR_RXRDY:
		lsr = 0;
	      rx:
		do {
		    c = NS16550_INB (dp, NS16550_DATA) & 0xff;
		    if (lsr & (LSR_BI | LSR_FE | LSR_PE | LSR_OE)) {
			if (lsr & LSR_BI) {
			    _RTPX_ELOG ("[brk]");
			    if (_rtpx_break_debug) {
				/* immediate debug interrupt */
				extern void     _debug_breakpoint (void);
				_debug_breakpoint ();
				continue;
			    }
			    c = TTY_FE | 0;
			}
			else if (lsr & LSR_FE) {
			    _RTPX_ELOG ("[fram]");
			    ser->ferrs++;
			    c |= TTY_FE;
			}
			else if (lsr & LSR_PE) {
			    _RTPX_ELOG ("[ptye]");
			    ser->perrs++;
			    c |= TTY_PE;
			}
			if (lsr & LSR_OE) {
			    _RTPX_ELOG ("[oflo]");
			    ser->oerrs++;
			}
		    }

#if NS16550_DEBUG
		    {
			char            cm = c & 0x7f;
			_RTPX_ELOG ("%02x(%c)<nsuart%d",
				    cm, (cm < 0x20
					 || cm >= 0x7f) ? '?' : cm,
				    ser->unit);
		    }
#endif
		    if (tp->t_state & TS_ISOPEN)
			_ttyinput (c, tp);
		} while ((lsr =
			  NS16550_INB (dp, NS16550_LSR)) & LSR_RXRDY);
		break;

	    case IIR_TXRDY:
#if NS16550_DEBUG > 1
		_RTPX_ELOG ("[txrdy]");
#endif
		tp->t_state &= ~TS_BUSY;
		if (tp->t_state & TS_ISOPEN)
		    nsstart (tp);
		break;

	    case IIR_MLSC:
		_RTPX_ELOG ("[mlsc]");
		msr = NS16550_INB (dp, NS16550_MSR);
#if 0
		/* pass carrier-detect change to tty.c */
		if (msr & MSR_DDCD)
		    _ttymodem (tp, msr & MSR_DCD);
#endif
		break;

	    default:
		_RTPX_ELOG ("[baduint]");
		/* unexpected interrupt, ignore */
		break;
	    }
	}
    }
    return 0;
}


static void
nstaskstart (struct serial *ser)
{
    /* create interrupt thread on first open */
    pthread_attr_t  tha;
    struct sched_param sp;

    pthread_attr_init (&tha);

    /* set FIFO scheduling */
    pthread_attr_setdetachstate (&tha, PTHREAD_CREATE_DAEMON_NP);
    pthread_attr_setinheritsched (&tha, PTHREAD_EXPLICIT_SCHED);
    pthread_attr_setschedpolicy (&tha, SCHED_FIFO);

    pthread_attr_getschedparam (&tha, &sp);
#ifdef NS16550_PRIO
    sp.sched_priority = NS16550_PRIO;
#else
    /* set maximum priority to catch break debug interrupts */
    sp.sched_priority = sched_get_priority_max (SCHED_FIFO);
#endif
    pthread_attr_setschedparam (&tha, &sp);

    pthread_attr_setname_np (&tha, ser->thname);

    if (pthread_create (&ser->tid, &tha, nstask, ser) != 0)
	abort ();

    pthread_attr_destroy (&tha);
}



static void
nsrxflow (struct tty *tp, int enb)
{
    struct serial  *ser = tp->t_addr;
    void           *dp = ser->dev;

    if (enb)
	ser->mcr |= MCR_RTS;
    else
	ser->mcr &= ~MCR_RTS;
    NS16550_OUTB (dp, NS16550_MCR, ser->mcr);
}


static void
nstxstop (struct tty *tp, int flush)
{
    /* nothing can be done with already output characters */
}


static void
nssendbrk (struct tty *tp, int duration)
{
    struct serial  *ser = tp->t_addr;
    void           *dp = ser->dev;
    struct timespec ts;

    if (duration <= 0)
	duration = 1;
    else if (duration > 40)
	duration = 40;
    duration *= 250;
    ts.tv_sec = duration / 1000;
    ts.tv_nsec = (duration % 1000) * 1000000;

    NS16550_OUTB (dp, NS16550_CFCR,
		  NS16550_INB (dp, NS16550_CFCR) | CFCR_SBREAK);
    while (nanosleep (&ts, &ts) == -1 && errno == EINTR)
	continue;
    NS16550_OUTB (dp, NS16550_CFCR,
		  NS16550_INB (dp, NS16550_CFCR) & ~CFCR_SBREAK);
}



/* 
 * nsconfig
 *
 * Abstract:
 *  Set up the serial port according to specified parameters.  
 *
 * Arguments:
 *   ch:        channel number
 *   tio:       pointer to termios structure
 *
 * Returns:
 *   true on success
 */

static int
nsparams (struct tty *tp, int set, struct termios *tio)
{
    struct serial  *ser = tp->t_addr;
    void           *dp = ser->dev;
    long            brtc;
    unsigned char   cfcr;

#define divrnd(n, q)	(((n) * 2 / (q) + 1) / 2)	/* divide and
							   round off */
    if (set) {
	/* set h/w settings from termios */
	long            brtc, err;

	if (tio->c_ospeed != tio->c_ispeed)
	    return EINVAL;

	if (tio->c_ospeed == 0)
	    return 0;

	brtc = divrnd (NS16550_TICK, tio->c_ospeed);
	if (brtc <= 0)
	    return EINVAL;

#if NS16550_TICK < INT_MAX / 500
	err = divrnd (1000 * NS16550_TICK, brtc * tio->c_ospeed) - 1000;
#else
	err = 1000 * divrnd (NS16550_TICK, brtc * tio->c_ospeed) - 1000;
#endif
	if (err < 0)
	    err = -err;
	if (err > SPEED_TOLERANCE)
	    return EINVAL;

	cfcr = 0;
	if (tio->c_cflag & PARENB) {
	    if (tio->c_cflag & PARODD)
		cfcr |= CFCR_PENAB | CFCR_PODD;
	    else
		cfcr |= CFCR_PENAB | CFCR_PEVEN;
	}

	switch (tio->c_cflag & CSIZE) {
	case CS5:
	    cfcr |= CFCR_5BITS;
	    break;
	case CS6:
	    cfcr |= CFCR_6BITS;
	    break;
	case CS7:
	    cfcr |= CFCR_7BITS;
	    break;
	case CS8:
	    cfcr |= CFCR_8BITS;
	    break;
	}

	if (tio->c_cflag & CSTOPB)
	    cfcr |= CFCR_STOPB;

	/* wait for fully empty fifo */
	while (!(NS16550_INB (dp, NS16550_LSR) & LSR_TSRE))
	    continue;

	NS16550_OUTB (dp, NS16550_CFCR, CFCR_DLAB | cfcr);
	NS16550_OUTB (dp, NS16550_DLL, brtc & 0xff);
	NS16550_OUTB (dp, NS16550_DLM, brtc >> 8);
	NS16550_OUTB (dp, NS16550_CFCR, cfcr);

	/* restart transmitter */
	nsstart (tp);
    }
    else {
	/* convert from h/w settings to termios */

	cfcr = NS16550_INB (dp, NS16550_CFCR);
	NS16550_OUTB (dp, NS16550_CFCR, CFCR_DLAB | cfcr);
	brtc = ((NS16550_INB (dp, NS16550_DLM) & 0xff) << 8) |
	    (NS16550_INB (dp, NS16550_DLL) & 0xff);
	NS16550_OUTB (dp, NS16550_CFCR, cfcr);

	if (brtc == 0)
	    tio->c_ispeed = 9600;	/* default */
	else
	    tio->c_ispeed = divrnd (NS16550_TICK, brtc);
	tio->c_ospeed = tio->c_ispeed;

	tio->c_cflag = CREAD | CLOCAL | HUPCL;

	if (cfcr & CFCR_PENAB) {
	    tio->c_cflag |= PARENB;
	    if (!(cfcr & CFCR_PEVEN))
		tio->c_cflag |= PARODD;
	}

	switch (cfcr & CFCR_CSIZE) {
	case CFCR_5BITS:
	    tio->c_cflag |= CS5;
	    break;
	case CFCR_6BITS:
	    tio->c_cflag |= CS6;
	    break;
	case CFCR_7BITS:
	    tio->c_cflag |= CS7;
	    break;
	case CFCR_8BITS:
	    tio->c_cflag |= CS8;
	    break;
	}

	if (cfcr & CFCR_STOPB)
	    tio->c_cflag |= CSTOPB;

    }
#undef divrnd

    return 0;
}



/* 
 * nsclose
 *
 * Abstract:
 *   Epilogue for the serial port BIOS code
 *
 * Arguments:
 *   ch:  channel number to close
 *
 * Returns:
 *   nothing
 */

static void
nsclose (struct tty *tp)
{
    struct serial  *ser = tp->t_addr;
    void           *dp = ser->dev;
#if 0
    /* leave thread running to catch debug interrupts */
    if (ser->tid) {
	pthread_cancel (ser->tid);
	ser->tid = 0;
	NS16550_OUTB (dp, NS16550_IER, 0);
    }
#endif
    /* disable DTR */
    NS16550_OUTB (dp, NS16550_MCR, (ser->mcr &= ~MCR_DTR));
}


/* 
 * nsopen
 *
 * Abstract:
 *   Initialise a serial port and its low-level buffers.
 *
 * Arguments:
 *   ch:      Channel to open
 *
 * Returns:
 *   0 if hunky-dory, error code if the channel is inaccessible
 */

static int
nsopen (int ch, struct tty **tpp)
{
    struct serial  *ser;
    void           *dp;
    int             ret;

    if (ch < 0 || ch >= dimof (serial))
	return ENXIO;
    ser = &serial[ch];

    /* have we initialised OK */
    if (ser->fifolen <= 0)
	return ENODEV;

    /* start interrupt task if not running */
    if (!ser->tid)
	nstaskstart (ser);

    /* return tty pointer and allocate buffers */
    *tpp = &ser->tty;
    if (ret = _ttyclalloc (&ser->tty))
	return ret;

    /* enable DTR */
    dp = ser->dev;
    NS16550_OUTB (dp, NS16550_MCR, (ser->mcr |= MCR_DTR));
    return 0;
}


/* guaranteed called at init time and safe from multiple threads */
static int
nsinit (int ch, int dev)
{
    struct serial  *ser;
    struct tty     *tp;
    pthread_mutexattr_t mxa;
    pthread_condattr_t cda;
    pthread_doorbellattr_t dba;
    struct timespec ts;
    void           *dp;

    if (ch < 0 || ch >= dimof (serial))
	return ENODEV;

    ser = &serial[ch];

    switch (ch) {
    case 0:
	ser->dev = PA_TO_KVA1 (UART0_BASE);
	break;
#ifdef UART1_BASE
    case 1:
	ser->dev = PA_TO_KVA1 (UART1_BASE);
	break;
#endif
#ifdef UART2_BASE
    case 2:
	ser->dev = PA_TO_KVA1 (UART2_BASE);
	break;
#endif
#ifdef UART3_BASE
    case 3:
	ser->dev = PA_TO_KVA1 (UART3_BASE);
	break;
#endif
    default:
	abort ();
    }

    dp = ser->dev;
    tp = &ser->tty;
    tp->t_addr = ser;		/* back link */

    /* force access to id reg */
    NS16550_OUTB (dp, NS16550_CFCR,
		  NS16550_INB (dp, NS16550_CFCR) & ~CFCR_DLAB);
    NS16550_OUTB (dp, NS16550_IIR, 0);
    if ((NS16550_INB (dp, NS16550_IIR) & 0x38) != 0) {
	/* not a 16550 */
	ser->fifolen = 0;
	return ENODEV;
    }

    /* in case it really is a 16550, enable the fifos */
    NS16550_OUTB (dp, NS16550_IIR,
		  FIFO_ENABLE | FIFO_RCV_RST | FIFO_XMT_RST |
		  FIFO_TRIGGER_4);

    /* wait 100us */
    ts.tv_sec = 0;
    ts.tv_nsec = 100 * 1000;
    nanosleep (&ts, NULL);

    /* determine fifo size */
    ser->fifolen =
	((NS16550_INB (dp, NS16550_IIR) & IIR_FIFO_MASK) ==
	 IIR_FIFO_MASK) ? 16 : 1;

    /* clear pending status reports. */
    (void) NS16550_INB (dp, NS16550_LSR);
    (void) NS16550_INB (dp, NS16550_MSR);

    /* 
     * Enable external interrupt gate (to avoid floating IRQ), enable RTS,
     * but don't enable DTR yet. 
     */
    NS16550_OUTB (dp, NS16550_MCR, ser->mcr = MCR_IENABLE | MCR_RTS);

    /* initialise doorbell */
    pthread_doorbellattr_init (&dba);
    pthread_doorbellattr_setevn (&dba, ser->evn);
    strcpy (ser->dbname, "ttyxxdb");
    ser->dbname[3] = (ser->unit / 10) + '0';
    ser->dbname[4] = (ser->unit % 10) + '0';
    pthread_doorbellattr_setname (&dba, ser->dbname);
    pthread_doorbell_init (&ser->db, &dba);
    pthread_doorbellattr_destroy (&dba);

    /* initialise condition */
    pthread_condattr_init (&cda);
    strcpy (ser->cdname, ser->dbname);
    ser->cdname[5] = 'c';
    ser->cdname[6] = 'd';
    pthread_condattr_setname_np (&cda, ser->cdname);
    pthread_condattr_setinterruptible_np (&cda, 1);
    pthread_cond_init (&ser->cd, &cda);
    pthread_condattr_destroy (&cda);

    /* initialise shared mutex */
    pthread_mutexattr_init (&mxa);
    strcpy (ser->mxname, ser->dbname);
    ser->mxname[5] = 'm';
    ser->mxname[6] = 'x';
    pthread_mutexattr_setname_np (&mxa, ser->mxname);
#ifdef _POSIX_THREAD_PRIO_INHERIT
    /* use priority inheritance so serial port interrupts are not locked
       out by a low-priority thread in tty.c */
    pthread_mutexattr_setprotocol (&mxa, PTHREAD_PRIO_INHERIT);
#endif
    pthread_mutex_init ((pthread_mutex_t *)&tp->t_mx, &mxa);
    pthread_mutexattr_destroy (&mxa);

    /* create links for generic tty driver */
    tp->t_state |= TS_INTRUPT;
    tp->t_dev = dev;
    tp->t_params = nsparams;
    tp->t_start = nsstart;
    tp->t_wait = nswait;
    tp->t_wakeup = nswakeup;
    tp->t_rxflow = nsrxflow;
    tp->t_txstop = nstxstop;
    tp->t_sendbrk = nssendbrk;
    tp->t_close = nsclose;

    /* set interrupt thread name */
    strcpy (ser->thname, ser->dbname);
    ser->thname[5] = 0;
    ser->thname[6] = 0;

    /* start the interrupt thread now (to catch debug interrupts) */
    nstaskstart (ser);

    return 0;
}
