#ident "sde/rtpx/rtpx_once.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_once.c: pthread_once implementation
 */


#include <stddef.h>
#include <pthread.h>

static pthread_mutex_t once_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t once_cond = PTHREAD_COND_INITIALIZER;


/* unlock mutex if wait is cancelled */
static void
once_unlock (void *arg)
{
    pthread_mutex_unlock (&once_lock);
}


/* reset once state if init routine is cancelled */
static void
once_reset (void *arg)
{
    pthread_once_t *once = arg;
    once->state = PTHREAD_ONCE_NEVER;
    pthread_cond_broadcast (&once_cond);
}


int
_pthread_once (pthread_once_t *once, void (*init) (void))
{
    extern int __isthreaded;

    /* fast path */
    if (once->state == PTHREAD_ONCE_DONE)
	return 0;

    /* Safe to continue if thread mechanism not yet started */
    if (! __isthreaded) {
	once->state = PTHREAD_ONCE_DONE;
	(*init) ();
	return 0;
    }

    /* slow path */
    pthread_mutex_lock (&once_lock);
    while (once->state == PTHREAD_ONCE_INPROGRESS) {
	pthread_cleanup_push (once_unlock, once);
	pthread_cond_wait (&once_cond, &once_lock);
	pthread_cleanup_pop (0);
    }

    if (once->state == PTHREAD_ONCE_NEVER) {
	once->state = PTHREAD_ONCE_INPROGRESS;
	pthread_mutex_unlock (&once_lock);
	pthread_cleanup_push (once_reset, once);
	(*init) ();
	pthread_cleanup_pop (0);
	pthread_mutex_lock (&once_lock);
	once->state = PTHREAD_ONCE_DONE;
	pthread_cond_broadcast (&once_cond);
    }

    pthread_mutex_unlock (&once_lock);
    return 0;
}


int (pthread_once) (pthread_once_t *, void (*) (void))
	__attribute__((alias("_pthread_once")));

#include <sdethread.h>

int (___sdethread_once) (__sdethread_once_t *, void (*) (void))
	__attribute__((alias("_pthread_once")));

#undef __sdethread_once
int (__sdethread_once) (__sdethread_once_t *, void (*) (void))
	__attribute__((alias("_pthread_once")));

