#ident "sde/rtpx/rtpx_proc.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_proc.c: basic process emulation
 */


#include "pthread_impl.h"
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

rtpx_proc_t _rtpx_proctab[_RTPX_PROC_MAX];

static pthread_mutex_t	procMx = PTHREAD_MUTEX_INITIALIZER;
static int      	procID;
static int		nproc;

struct rtpx_proc *
_rtpx_findproc (int pid)
{
    rtpx_proc_t    *pr;

    if (pid == 0 || pid == getpid ())
	return _rtpx_self_proc ();

    pthread_mutex_lock (&procMx);
    for (pr = _rtpx_proctab; pr < &_rtpx_proctab[_RTPX_PROC_MAX]; pr++)
	if (pr->pr_state != _RTPX_PROC_FREE && pr->pr_pid == pid)
	    break;
    pthread_mutex_unlock (&procMx);

    if (pr >= &_rtpx_proctab[_RTPX_PROC_MAX])
	return NULL;

    return pr;
}


int
_rtpx_newproc (int *result)
{
    rtpx_proc_t    *pr,
                   *opr,
                   *ppr;
    rtpx_thread_t  *cth;
    int		    err;

    pthread_mutex_lock (&procMx);

    /* find a free process table entry */
    for (pr = _rtpx_proctab; pr < &_rtpx_proctab[_RTPX_PROC_MAX]; pr++)
	if (pr->pr_state == _RTPX_PROC_FREE)
	    break;

    /* oh dear, out of processes */
    if (pr >= &_rtpx_proctab[_RTPX_PROC_MAX]) {
	pthread_mutex_unlock (&procMx);
	return EAGAIN;
    }

    /* bump active process count */
    ++nproc;

    /* allocate a unique process id */
    do {
	procID = (procID + 1) % 32768;
	for (opr = _rtpx_proctab; opr < &_rtpx_proctab[_RTPX_PROC_MAX]; opr++)
	    if (opr->pr_state != _RTPX_PROC_FREE && opr->pr_pid == procID)
		break;
    } while (opr < &_rtpx_proctab[_RTPX_PROC_MAX]);

    pr->pr_pid = procID;
    pr->pr_state = _RTPX_PROC_ZOMBIE;	/* partially allocated */

    pthread_mutex_unlock (&procMx);


    if (ppr = _rtpx_self_proc ()) {
	/* child process */
	pr->pr_ppid = ppr->pr_pid;
	pr->pr_pgrp = ppr->pr_pgrp;

	err = _cpu_procnew (ppr, pr);
	if (err) {
	    pr->pr_state = _RTPX_PROC_FREE;
	    pthread_mutex_lock (&procMx);
	    --nproc;
	    pthread_mutex_unlock (&procMx);
	    return err;
	}

	/* copy signal state to new process */
	_rtpx_sigcopy (pr, ppr);

	/* reset process timers */
	memset (pr->pr_timer, 0, sizeof(pr->pr_timer));

	/* remove this thread from parent process */
	pthread_mutex_lock (&ppr->pr_mx);
	ppr->pr_nthread--;
	pthread_mutex_unlock (&ppr->pr_mx);
    }
    else {
	rtpx_proc_t *ipr;
	pthread_mutexattr_t mxa;
	pthread_condattr_t cda;

#if 0
	/* initialise process table */
	memset (_rtpx_proctab, 0, sizeof (_rtpx_proctab));
#endif

	pthread_mutexattr_init (&mxa);
	pthread_mutexattr_setname_np (&mxa, "proc");
	pthread_condattr_init (&cda);
	pthread_condattr_setinterruptible_np (&cda, 1);
	pthread_condattr_setname_np (&cda, "waitpid");
	for (ipr = _rtpx_proctab; ipr < &_rtpx_proctab[_RTPX_PROC_MAX]; ipr++) {
	    pthread_mutex_init (&pr->pr_mx, NULL);
	    pthread_cond_init (&pr->pr_waitcd, &cda);
	}
	pthread_condattr_destroy (&cda);
	pthread_mutexattr_destroy (&mxa);
	    
	/* initialise first process */
	pr->pr_ppid = pr->pr_pgrp = pr->pr_pid;

	_rtpx_siginit (pr);

	_cpu_procfirst (pr);
    }

    pr->pr_exitval = 0;
    pr->pr_waiting = 0;
    pr->pr_nthread = 1;
    pr->pr_state = _RTPX_PROC_RUN;

    cth = rtpx_self ();
    cth->th_proc = pr;

    if (result)
	*result = pr->pr_pid;

    return 0;
}


void
_rtpx_joinproc (rtpx_thread_t *nth)
{
    rtpx_proc_t    *pr = nth->th_proc;

    pthread_mutex_lock (&pr->pr_mx);
    pr->pr_nthread++;
    pthread_mutex_unlock (&pr->pr_mx);
}


/* 
 * Called from pthread_exit() to detach a thread from its
 * process. Returns true if there are other processes still
 * running.
 */

int
_rtpx_leaveproc ()
{
    rtpx_proc_t    *pr = _rtpx_self_proc ();
    int             stillgoing = 1;
    timer_t	    tid;
    int		    i;

    pthread_mutex_lock (&pr->pr_mx);
    assert (pr->pr_nthread > 0);
    if (--pr->pr_nthread == 0) {

	/* delete timers */
	for (i = 0; i < NITIMER; i++)
	    if (tid = pr->pr_timer[i])
		timer_delete (tid);

	/* remove our signal handlers */
	_rtpx_sigexit (pr);

	/* destroy per-cpu data */
	_cpu_procdestroy (pr);

	/* XXX wake up parent process waiting for our status */
	if (pr->pr_ppid == pr->pr_pid) {
	    /* first "init" process exiting */
	    /* XXX should this be fatal? */
	    stillgoing = 0;
	}
	else {
	    rtpx_proc_t *ppr = _rtpx_findproc (pr->pr_ppid);
	    if (ppr) {
		pthread_mutex_lock (&ppr->pr_mx);
		if (ppr->pr_pid != pr->pr_ppid) {
		    /* parent has gone, nobody to wait for us */
		    pr->pr_state = _RTPX_PROC_FREE;
		}
		else if (ppr->pr_waiting) {
		    /* parent is waiting for a child to die, wake it */
		    pr->pr_state = _RTPX_PROC_ZOMBIE;
		    pthread_cond_broadcast (&ppr->pr_waitcd);
		}
		else {
		    /* try sending a signal (may be ignored) */
		    union sigval    sival;
		    pr->pr_state = _RTPX_PROC_ZOMBIE;
		    sival.sival_int = 0;
		    (void) _rtpx_kill (pr->pr_ppid, SIGCHLD, sival, SI_KERNEL);
		}
		pthread_mutex_unlock (&ppr->pr_mx);
	    }

	}

	pthread_mutex_lock (&procMx);
	if (--nproc == 0)
	    stillgoing = 0;
	pthread_mutex_unlock (&procMx);
    }
    pthread_mutex_unlock (&pr->pr_mx);
    return stillgoing;
}



/* 
 * Internal RTPX signal raise function, but callable only from
 * "user" mode. 
 */

int
_rtpx_sigraise (pid_t pid, int code, const struct sigevent *sigev)
{
    rtpx_thread_t  *cth;
    rtpx_proc_t    *pr;
    int             err;
    int		    found;

    if (pid >= 0) {
	/* send to single process */
	if ((pr = _rtpx_findproc (pid))
	    && pr->pr_state == _RTPX_PROC_RUN) {
	    cth = _rtpx_enter ();
	    err = _rtpx_sigqueue (pr, code, sigev);
	    return _rtpx_leaveWithValueLong (cth, err);
	}
	return ESRCH;
    }

    /* send to all processes in process group */
    pid = -pid;
    found = 0;

    cth = _rtpx_enter ();	/* XXX should we lock procMx instead */
    for (pr = _rtpx_proctab; pr < &_rtpx_proctab[_RTPX_PROC_MAX]; pr++)
	if (pr->pr_state == _RTPX_PROC_RUN && pr->pr_pgrp == pid) {
	    ++found;
	    if (err = _rtpx_sigqueue (pr, code, sigev))
		break;
	}
    if (found == 0)
	err = ESRCH;
    return _rtpx_leaveWithValueLong (cth, err);
}



/* 
 * Simple POSIX process interface
 */

pid_t
getpid (void)
{
    rtpx_proc_t    *pr = _rtpx_self_proc ();
    return pr->pr_pid;
}


pid_t
getppid (void)
{
    rtpx_proc_t    *pr = _rtpx_self_proc ();
    return pr->pr_ppid;
}


pid_t
getpgid (pid_t pid)
{
    rtpx_proc_t    *pr = _rtpx_findproc (pid);

    if (!pr) {
	errno = ESRCH;
	return -1;
    }

    return pr->pr_pgrp;
}


int
setpgid (pid_t pid, pid_t pgid)
{
    rtpx_proc_t    *pr = _rtpx_findproc (pid);

    if (!pr) {
	errno = ESRCH;
	return -1;
    }

    pthread_mutex_lock (&pr->pr_mx);

    /* check process still exists and is running */
    if ((pid != 0 && pr->pr_pid != pid) || pr->pr_state != _RTPX_PROC_RUN) {
	pthread_mutex_unlock (&pr->pr_mx);
	errno = ESRCH;
	return -1;
    }

    if (pgid == 0)
	pr->pr_pgrp = pr->pr_pid;
    else 
	pr->pr_pgrp = pgid;

    pthread_mutex_unlock (&pr->pr_mx);
    return 0;
}


pid_t
getpgrp (void)
{
    return getpgid (0);
}


int
setpgrp ()
{
    return setpgid (0, 0);
}


uid_t
getuid (void)
{
    return 0;
}

gid_t
getgid ()
{
    return 0;
}

uid_t
geteuid ()
{
    return 0;
}

gid_t
getegid ()
{
    return 0;
}


int issetugid (void);

int
issetugid ()
{
    return 0;
}


int
sched_get_priority_min (int policy)
{
    return _RTPX_PRIO_MIN;
}

int
sched_get_priority_max (int policy)
{
    if (policy == SCHED_RR)
	return _RTPX_RR_PRIO_MAX;
    else
	return _RTPX_FIFO_PRIO_MAX;
}

int
sched_rr_get_interval (pid_t pid, struct timespec *ts)
{
    ts->tv_sec = 0;
    ts->tv_nsec = (1000 * 1000 * 1000 + _rtpx_rrfreq / 2) / _rtpx_rrfreq;
    return 0;
}


int
sched_rr_set_interval (pid_t pid, const struct timespec *ts)
{
    int             err = _rtpx_rr_set_interval (pid, ts);
    if (err != 0) {
	errno = err;
	return -1;
    }
    return 0;
}


/* process-wide scheduling */
int
sched_getscheduler (pid_t pid, struct sched_param *sp)
{
    errno = ENOSYS;
    return -1;
}

int
sched_setscheduler (pid_t pid, const struct sched_param *sp)
{
    errno = ENOSYS;
    return -1;
}

int
sched_getparam (pid_t pid, struct sched_param *sp)
{
    errno = ENOSYS;
    return -1;
}

int
sched_setparam (pid_t pid, const struct sched_param *sp)
{
    errno = ENOSYS;
    return -1;
}
