#ident "sde/rtpx/rtpx_rwlock.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2001-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_rwlock.c: UNIX98 read/write locks
 */


#define _POSIX_C_SOURCE 199506L

#include "pthread_impl.h"
#include <stdarg.h>
#include <string.h>

/*
 * Safe but slow implementation of read/write locks.
 * We should think about whether this can be implemented
 * using low-level atomic primitives, as for mutexes and
 * semaphores, at least for the common rdlock/unlock case.
 */

int
pthread_rwlock_init (pthread_rwlock_t *rw,
		     const pthread_rwlockattr_t *rwa)
{
    int ret;

    ret = pthread_mutex_init (&rw->rw_mx, NULL);
    if (ret != 0)
	return ret;

    ret = pthread_cond_init (&rw->rw_rsig, NULL);
    if (ret != 0) {
	pthread_mutex_destroy (&rw->rw_mx);
	return ret;
    }

    ret = pthread_cond_init (&rw->rw_wsig, NULL);
    if (ret != 0) {
	pthread_cond_destroy (&rw->rw_rsig);
	pthread_mutex_destroy (&rw->rw_mx);
	return ret;
    }

    if (rwa)
	rw->rw_rwa = *rwa;
    rw->rw_locks = 0;
    rw->rw_wrwait = 0;

    return 0;
}


int
pthread_rwlock_destroy (pthread_rwlock_t *rw)
{
    int ret;

    if ((ret = pthread_mutex_lock (&rw->rw_mx)) != 0)
	return ret;

    if (rw->rw_locks != 0 || rw->rw_wrwait != 0) {
        pthread_mutex_unlock (&rw->rw_mx);
        return EBUSY;
    }
    
    pthread_mutex_unlock (&rw->rw_mx);

    pthread_mutex_destroy (&rw->rw_mx);
    pthread_cond_destroy (&rw->rw_rsig);
    pthread_cond_destroy (&rw->rw_wsig);

    return 0;
}


/* Cleanup if a read wait is cancelled */
static void 
rdcleanup (void *arg)
{
    pthread_rwlock_t *rw = (pthread_rwlock_t *)arg;

    --rw->rw_rdwait;
    pthread_mutex_unlock (&rw->rw_mx);
}


int
pthread_rwlock_rdlock (pthread_rwlock_t *rw)
{
    int ret;
    
    if ((ret = pthread_mutex_lock (&rw->rw_mx)) != 0)
	return ret;
    
    /* writers have priority over readers */
    if (rw->rw_wrwait || rw->rw_locks < 0) {
	/* simple deadlock check */
	if (rw->rw_locks < 0 && rw->rw_owner == pthread_self ()) {
	    pthread_mutex_unlock (&rw->rw_mx);
	    return EDEADLK;
	}

	rw->rw_rdwait++;
        pthread_cleanup_push (rdcleanup, (void*)rw);
	do {
	    ret = pthread_cond_wait (&rw->rw_rsig, &rw->rw_mx);
	    if (ret != 0)
		break;
	} while (rw->rw_wrwait || rw->rw_locks < 0);
        pthread_cleanup_pop (0);
	--rw->rw_rdwait;
    }

    
    if (ret == 0) {
	if (rw->rw_locks == INT_MAX - 1)
	    ret = EAGAIN;
	else
	    rw->rw_locks++;		/* read lock */
    }

    pthread_mutex_unlock (&rw->rw_mx);

    return ret;
}


int
pthread_rwlock_tryrdlock (pthread_rwlock_t *rw)
{
    int ret;

    if ((ret = pthread_mutex_lock (&rw->rw_mx)) != 0)
	return ret;

    /* writers have priority over readers */
    if (rw->rw_wrwait || rw->rw_locks < 0)
	ret = EBUSY;
    else if (rw->rw_locks == INT_MAX - 1)
	ret = EAGAIN;
    else
	rw->rw_locks++;		/* read lock */

    pthread_mutex_unlock (&rw->rw_mx);

    return ret;
}


/* Cleanup if a write wait is cancelled */
static void 
wrcleanup (void *arg)
{
    pthread_rwlock_t *rw = (pthread_rwlock_t *)arg;

    --rw->rw_wrwait;
    pthread_mutex_unlock (&rw->rw_mx);
}


int
pthread_rwlock_wrlock (pthread_rwlock_t *rw)
{
    int ret;

    if ((ret = pthread_mutex_lock (&rw->rw_mx)) != 0)
	return ret;

    if (rw->rw_locks != 0) {
	/* simple deadlock check */
	if (rw->rw_locks < 0 && rw->rw_owner == pthread_self ()) {
	    pthread_mutex_unlock (&rw->rw_mx);
	    return EDEADLK;
	}
	rw->rw_wrwait++;
        pthread_cleanup_push (wrcleanup, (void*)rw);
	do {
	    ret = pthread_cond_wait (&rw->rw_wsig, &rw->rw_mx);
	    if (ret != 0) 
		break;
	} while (rw->rw_locks != 0);
        pthread_cleanup_pop (0);
	--rw->rw_wrwait;
    }

    if (ret == 0) {
	rw->rw_locks = -1;		/* write lock */
	rw->rw_owner = pthread_self ();
    }

    pthread_mutex_unlock (&rw->rw_mx);
    
    return ret;
}


int
pthread_rwlock_trywrlock (pthread_rwlock_t *rw)
{
    int ret;

    if ((ret = pthread_mutex_lock (&rw->rw_mx)) != 0)
	return ret;

    if (rw->rw_locks != 0)
	ret = EBUSY;
    else
	rw->rw_locks = -1;		/* write lock */

    pthread_mutex_unlock (&rw->rw_mx);

    return ret;
}


int
pthread_rwlock_unlock (pthread_rwlock_t *rw)
{
    int ret;

    if ((ret = pthread_mutex_lock (&rw->rw_mx)) != 0)
	return ret;

    if (rw->rw_locks > 0) {
	/* wake up a waiting writer when no more read locks held */
	if (--rw->rw_locks == 0 && rw->rw_wrwait)
	    ret = pthread_cond_signal (&rw->rw_wsig);
    } 
    else if (rw->rw_locks < 0) {
	rw->rw_locks = 0;
	rw->rw_owner = 0;
	if (rw->rw_wrwait)
	    /* wake up one waiting writer */
	    ret = pthread_cond_signal (&rw->rw_wsig);
	else if (rw->rw_rdwait)
	    /* wake up all waiting readers */
	    ret = pthread_cond_broadcast (&rw->rw_rsig);
    } 
    else
	ret = EINVAL;

    pthread_mutex_unlock (&rw->rw_mx);

    return ret;
}
