#ident "sde/rtpx/rtpx_sdetty.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2006 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_sdetty.c: Posix style terminal i/o for MIPS SDE under i/u control
 */


#include "pthread_impl.h"

#define _KERNEL
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/init.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <assert.h>

#include "file.h"
#include "tty.h"
#include "kit_impl.h"
#include "sbdser.h"

#include "sbdrtpx.h"

#ifdef _RTPX_SERINT_LIST

static int  	ttyread (struct file *, struct uio *);
static int	ttywrite (struct file *, struct uio *);
static int      ttyioctl (struct file *, int, void *);
static int      ttyselect (struct file *, int);
static int      ttyopen (struct file *, dev_t, int);
static int      ttyclose (struct file *);

/* POSIX file i/o ops for /dev/ttyx */
static const struct fileops _tty_ops = {
    ttyread,
    ttywrite,
    ttyioctl,
    ttyselect,
    ttyopen,
    ttyclose,
};


/* private per-tty data structure */
struct ttydev {
    const struct sbdserint *td_si;	/* driver interface */
    int             td_chan;		/* port channel number */
    struct tty     *td_tty;		/* generic tty structure */
    char            td_name[10];	/* device name */
};

pthread_mutex_t tty_mx = PTHREAD_MUTEX_INITIALIZER;
static struct ttydev *ttys;
static int      ntty;

static int      dbg_active;



static int
ttywrite (struct file *fp, struct uio *uio)
{
    struct tty *tp = (struct tty *) fp->f_data;

#if 0
    if (dbg_active && _sbddbgport == 0 && tp == ttys->td_tty) {
	/* write to console with debug active on same port */
	/* XXX omits tty.c output processing, and not safe
	   to do polled input from UART in parallel with
	   interrupt driver. Need to do this via interrupt
	   driver. */
	_dbg_write (buf, n);
	return n;
    }
#endif
    return _ttywrite (tp, uio, fp->f_flag);
}

static int
ttyread (struct file *fp, struct uio *uio)
{
    return _ttyread ((struct tty *) fp->f_data, uio, fp->f_flag);
}

static int
ttyioctl (struct file *fp, int cmd, void *arg)
{
    return _ttyioctl ((struct tty *) fp->f_data, cmd, arg);
}

static int
ttyselect (struct file *fp, int rw)
{
    return _ttyselect ((struct tty *) fp->f_data, rw);
}

static int
ttyclose (struct file *fp)
{
    int error = 0;
    if (fp->f_data)
	error = _ttyclose ((struct tty *) fp->f_data);
    return error;
}



static int
ttyopen (struct file *fp, dev_t dev, int fflags)
{
    struct ttydev  *td = &ttys[dev];
    struct tty     *tp;
    int error = 0;

    if (dev >= ntty)
	return ENXIO;

    if (!(tp = td->td_tty)) {
	error =
	    (*td->td_si->si_open) (td->td_chan, &td->td_tty);
	if (error != 0)
	    return error;
	tp = td->td_tty;
    }

    /* generic tty open */
    error = _ttyopen (tp);

    /* point file at private data */
    if (error == 0)
	fp->f_data = tp;

    return error;
}



/* 
 * Very low-level output to console, avoiding termio etc.
 */

void
_dbg_attachhook (void)
{
    dbg_active = 1;
}


void
_dbg_detachhook (void)
{
    dbg_active = 0;
}


#pragma weak _dbg_write
void
_dbg_write (const char *b, size_t n)
{
}


void
_mon_putc (char c)
{
    if (dbg_active && _sbddbgport == 0) {
	/* write to console with debug active on same port */
	char cb = c;
	_dbg_write (&cb, 1);
    }
    else {
	/* use old polled i/o code XXX what about PMON */
	const struct sbdser *ss = &_sbdsertab[0];
	(*ss->ss_tx) (ss->ss_data, c, 0);
    }
}


void
_mon_puts (const char *s)
{
    if (dbg_active && _sbddbgport == 0) {
	/* write to console with debug active on same port */
	const char *e;
	for (e = s; *e; e++)
	    continue;
	_dbg_write (s, e - s);
    }
    else {
	const struct sbdser *ss = &_sbdsertab[0];
	char c;
	while ((c = *s++))
	    (*ss->ss_tx) (ss->ss_data, c, 0);
    }
}


void 
_mon_write (const char *buf, unsigned int nb)
{
    if (dbg_active && _sbddbgport == 0) {
	/* write to console with debug active on same port */
	_dbg_write (buf, nb);
    }
    else {
	const struct sbdser *ss = &_sbdsertab[0];
	while (nb-- > 0)
	    (*ss->ss_tx) (ss->ss_data, *buf++, 0);
    }
}


int
_mon_getc (int canblock)
{
    const struct sbdser *ss = &_sbdsertab[0];

    do {
	int c = (*ss->ss_rxpoll) (ss->ss_data, canblock, 0);
	if (c)
	    return (c & 0xff);
    } while (canblock);
    return -1;
}



static const struct sbdserint *const serinttab[] = {
    _RTPX_SERINT_LIST,
    0
};


/* /dev/tty*:   crw-rw-r-- */
#define TTYMODE	(S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IWOTH)

static void
ttydev_init (void)
{
    const struct sbdserint *const *sip;
    const struct sbdserint *si;
    static int      inited = 0;
    dir_t           ttynode;
    struct ttydev  *td;

    if (inited++ != 0)
	return;

    ntty = 0;
    for (sip = serinttab; *sip; sip++)
	ntty += (*sip)->si_nchan;

    if (ntty == 0)
	_rtpx_log (RTPXLOG_PANIC, "no serial ports found");

    ttys = malloc (ntty * sizeof (*ttys));
    if (!ttys)
	abort ();
    memset (ttys, 0, ntty * sizeof (*ttys));

    memset (&ttynode, 0, sizeof (ttynode));
    ttynode.d_mode = TTYMODE;
    ttynode.d_ops = &_tty_ops;
    ttynode.d_dev = 0;

    td = ttys;
    for (sip = serinttab; si = *sip; sip++) {
	int             ch;
	for (ch = 0; ch < si->si_nchan; ch++, td++, ttynode.d_dev++) {
	    char           *cp;

	    if (si->si_init (ch, (int) ttynode.d_dev) != 0)
		continue;

	    cp = strcpy (td->td_name, si->si_devname);
	    cp += strlen (cp);
	    if (ch >= 10)
		*cp++ = (ch / 10) + '0';
	    *cp++ = (ch % 10) + '0';
	    *cp = '\0';

	    td->td_si = si;
	    td->td_chan = ch;

	    /* add to /dev directory */
	    ttynode.d_name = td->td_name;
	    if (_mknod ("/dev", &ttynode))
		_rtpx_log (RTPXLOG_PANIC, "mknod(/dev/%s) failed\n",
			   td->td_name);
	}
    }

    /* links for /dev/console and /dev/tty to /dev/tty0 */
    ttynode.d_dev = 0;
    ttynode.d_name = "console";
    if (_mknod ("/dev", &ttynode))
	_rtpx_log (RTPXLOG_PANIC, "mknod(/dev/console) failed\n");

    /* XXX change to support per-process controlling tty? */
    ttynode.d_name = "tty";
    if (_mknod ("/dev", &ttynode))
	_rtpx_log (RTPXLOG_PANIC, "mknod(/dev/tty) failed\n");
}


static void
ttydev_fini (void)
{
    struct ttydev  *td;

    (void) pthread_mutex_trylock (&tty_mx);
    for (td = ttys; td < &ttys[ntty]; td++) {
	struct tty     *tp;
	if (tp = td->td_tty) {
	    int             err;
	    _RTPX_ELOG ("%s: flushing at exit", td->td_name);
	    if (err = __sdethread_mutex_lock (&tp->t_mx)) {
		_RTPX_ELOG ("%s: lock error: %s", td->td_name,
			    strerror (err));
		continue;
	    }
	    /* XXX should we use a timeout on this */
	    if (tp->t_state & TS_OPEN)
		(void) _ttywait (tp);
	    __sdethread_mutex_unlock (&tp->t_mx);
	    _RTPX_ELOG ("%s: flushing done", td->td_name);
	}
    }
    pthread_mutex_unlock (&tty_mx);
}


DECL_INIT_FUNC (IF_MOUNT, ttydev_init)
DECL_FINI_FUNC (IF_MOUNT, ttydev_fini)

/* call this dummy function to load this module */
void _ttydev_load (void);
void _ttydev_load (void)
{
}

#endif				/* _RTPX_SERINT_LIST */
