#ident "sde/rtpx/rtpx_sem.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2001-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_sem.c: Posix semaphores
 */


#include "pthread_impl.h"
#include <string.h>

int
sem_init (sem_t *sem, int pshared, unsigned int val)
{
    if (val > SEM_VALUE_MAX) {
	errno = EINVAL;
	return -1;
    }
    sem->sem_waitq.wq_first = NULL;
    sem->sem_value = val;
    return 0;
}

int
sem_destroy (sem_t *sem)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    if (sem->sem_waitq.wq_first)
	return rtpx_leaveWithError (cth, EBUSY);
    sem->sem_value = SEM_VALUE_MAX;
    return rtpx_leaveWithValue (cth, 0);
}


int
sem_trywait (sem_t *sem)
{
    /* fast path */
    int             oval = _cpu_atomic_add (&sem->sem_value, -1);
    if (oval > 0)
	return 0;		/* we've got it */

    /* cleanup on failure */
    (void) _cpu_atomic_add (&sem->sem_value, 1);
    errno = EAGAIN;
    return -1;
}


int
sem_wait (sem_t *sem)
{
    /* fast path */
    int             oval = _cpu_atomic_add (&sem->sem_value, -1);
    if (oval > 0)
	return 0;		/* we've got it */

    /* slow path */
    {
	rtpx_thread_t  *cth = rtpx_enter ();

	/* don't block if someone has sem_posted since the atomic op */
	if (sem->sem_value >= 0)
	    return rtpx_leaveWithValue (cth, 0);

	return rtpx_leaveWhenResumed (cth, THSTATE_SEMWAIT,
				      &sem->sem_waitq, 0);
    }
}


int
sem_post (sem_t *sem)
{
    /* fast path */
    int             oval = _cpu_atomic_add (&sem->sem_value, 1);
    if (oval >= 0) {
	/* no-one to wake up */
	if (oval == SEM_VALUE_MAX) {
	    _cpu_atomic_add (&sem->sem_value, -1);
	    return EINVAL;
	}
	return 0;
    }

    /* slow path */
    {
	rtpx_thread_t  *cth = rtpx_enter ();

	if (sem->sem_waitq.wq_first) {
	    /* thread is waiting - wake one up */
	    rtpx_thread_t  *nth = _rtpx_popq (&sem->sem_waitq);
	    assert (sem->sem_value <= 0);
	    if (nth && rtpx_resumeWithValue (nth, 0))
		return rtpx_leaveWithValueAndDispatch (cth, 0);
	}

	return rtpx_leaveWithValue (cth, 0);
    }
}


int
sem_getvalue (sem_t *sem, int *valp)
{
    *valp = sem->sem_value;
    return 0;
}

int
sem_setname_np (sem_t *sem, const char *name)
{
    sem->sem_name = name;
    return 0;
}

int
sem_getname_np (sem_t *sem, const char **namep)
{
    *namep = sem->sem_name;
    return 0;
}
