#ident "sde/rtpx/rtpx_semopen.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2001-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_senopen.c: POSIX.4 named semaphores
 */


/* 
 * XXX This implementation works only within a shared address space, 
 * using thread synchronisation.
 */

#include "pthread_impl.h"
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <semaphore.h>
#include <sys/time.h>
#include <sys/fcntl.h>
#include <sys/queue.h>

struct sem_name {
    sem_t	    sn_sem;		/* must be first */
    LIST_ENTRY(sem_name) sn_link;
    char            *sn_name;
    int		    sn_refcnt;
    int		    sn_unlinked;
};

static LIST_HEAD(, sem_name) snhead = LIST_HEAD_INITIALIZER(snhead);
static pthread_mutex_t snh_mx = PTHREAD_MUTEX_INITIALIZER;

sem_t * 
sem_open (const char *name, int oflags, ...)
{
    struct sem_name *sn;
    sem_t           *sem;
    va_list         ap;
    mode_t	    mode;
    sigset_t        sm;
    size_t          namlen;
    unsigned int    value;

    namlen = strlen (name);
    if (namlen > PATH_MAX) {
	errno = ENAMETOOLONG;
	return (sem_t *) - 1;
    }

    sigfillset (&sm);
    pthread_sigmask (SIG_BLOCK, &sm, &sm);
    pthread_mutex_lock (&snh_mx);

    LIST_FOREACH(sn, &snhead, sn_link)
	if (strcmp (name, sn->sn_name) == 0)
	    break;

    if (sn != LIST_END(&snhead)) {
	/* got existing mqueue */
	if ((oflags & (O_CREAT | O_EXCL)) == (O_CREAT | O_EXCL)) {
	    /* can't exclusive-create if it already exists */
	    errno = EEXIST;
	    sem = (sem_t *) (-1);
	}
	else
	    sem = &sn->sn_sem;
	goto ret;
    }

    /* not found - make a new mqueue ? */

    if (!(oflags & O_CREAT)) {
	/* doesn't exists, and not asked to create it */
	errno = ENOENT;
	sem = (sem_t *) (-1);
	goto ret;
    }

    va_start (ap, oflags);
    mode = va_arg (ap, int);
    value = va_arg (ap, unsigned int);
    va_end (ap);

    sn = (struct sem_name *) malloc (sizeof (struct sem_name) + namlen + 1);
    if (!sn) {
	errno = ENFILE;
	sem = (sem_t *) (-1);
	goto ret;
    }
    memset (sn, 0, sizeof (struct sem_name));

    /* store name */
    sn->sn_name = (char *)(sn + 1);
    memcpy (sn->sn_name, name, namlen + 1);
    sn->sn_refcnt = 1;

    /* initiliase new semaphore */
    sem = &sn->sn_sem;
    sem_init (sem, 1, value);
    sem_setname_np (sem, sn->sn_name);		/* set debug name too */

    /* add to namespace */
    LIST_INSERT_HEAD (&snhead, sn, sn_link);

  ret:
    pthread_mutex_unlock (&snh_mx);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);
    return sem;
}


int
sem_close (sem_t *sem)
{
    struct sem_name *sn = (struct sem_name *) sem;
    sigset_t        sm;
    int             ret = 0;

    sigfillset (&sm);
    pthread_sigmask (SIG_BLOCK, &sm, &sm);
    pthread_mutex_lock (&snh_mx);

    if (sn->sn_refcnt > 0) {
	/* if closing unlinked sem, now's the time to free it */
	if (--sn->sn_refcnt == 0 && sn->sn_unlinked) {
	    sem_destroy (&sn->sn_sem);
	    free (sn);
	}
    }

    pthread_mutex_unlock (&snh_mx);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);
    return ret;
}


int
sem_unlink (const char *name)
{
    struct sem_name *sn;
    sigset_t        sm;
    int		    ret = 0;

    sigfillset (&sm);
    pthread_sigmask (SIG_BLOCK, &sm, &sm);
    pthread_mutex_lock (&snh_mx);

    LIST_FOREACH(sn, &snhead, sn_link)
	if (strcmp (name, sn->sn_name) == 0)
	    break;

    if (sn == LIST_END(&snhead)) {
	errno = ENOENT;
	ret = -1;
	goto done;
    }
	
    /* remove from namespace - no-one can then open it any more */
    LIST_REMOVE (sn, sn_link);

    /* either delete now, or mark for future deletion */
    if (sn->sn_refcnt == 0) {
	/* no-one has a (valid) pointer to it any more */
	sem_destroy (&sn->sn_sem);
	free (sn);
    }
    else {
	/* mark as unlinked, so that last close will free it */
	sn->sn_unlinked = 1;
    }

done:
    pthread_mutex_unlock (&snh_mx);
    pthread_sigmask (SIG_SETMASK, &sm, NULL);
    return ret;
}
