#ident "sde/rtpx/rtpx_sig.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_sig.c: process-level signal emulation 
 */


#include "pthread_impl.h"

/* 
 * Signal properties and actions.
 * The array below categorises the signals and their default actions
 * according to the following properties:
 */
#define	SA_KILL		0x01	/* terminates process by default */
#define	SA_IGNORE	0x02	/* ignore by default */
#define	SA_CANTMASK	0x04	/* non-maskable, catchable */
#define	SA_ABORT	0x08	/* trap into monitor */

static const unsigned char sigprop[NSIG + 1] = {
    0,				/* unused */
    SA_KILL,			/* SIGHUP */
    SA_KILL,			/* SIGINT */
    SA_KILL,			/* SIGQUIT */
    SA_KILL | SA_ABORT,		/* SIGILL */
    SA_KILL | SA_ABORT,		/* SIGTRAP */
    SA_KILL | SA_ABORT,		/* SIGABRT */
    SA_KILL | SA_ABORT,		/* SIGEMT */
    SA_KILL | SA_ABORT,		/* SIGFPE */
    SA_KILL | SA_CANTMASK,	/* SIGKILL */
    SA_KILL | SA_ABORT,		/* SIGBUS */
    SA_KILL | SA_ABORT,		/* SIGSEGV */
    SA_KILL | SA_ABORT,		/* SIGSYS */
    SA_KILL,			/* SIGPIPE */
    SA_KILL,			/* SIGALRM */
    SA_KILL,			/* SIGTERM */
    SA_IGNORE,			/* SIGURG */
    SA_CANTMASK,		/* SIGSTOP */
    0,				/* SIGTSTP */
    SA_IGNORE,			/* SIGCONT */
    SA_IGNORE,			/* SIGCHLD */
    0,				/* SIGTTIN */
    0,				/* SIGTTOU */
    SA_IGNORE,			/* SIGIO */
    SA_KILL,			/* SIGXCPU */
    SA_KILL,			/* SIGXFSZ */
    SA_KILL,			/* SIGVTALRM */
    SA_KILL,			/* SIGPROF */
    SA_IGNORE,			/* SIGWINCH  */
    SA_IGNORE,			/* SIGINFO */
    SA_KILL,			/* SIGUSR1 */
    SA_KILL,			/* SIGUSR2 */
};

#define VALID_SIGNO(signo)	(signo > 0 && signo <= _RTPX_SIGNO_MAX)


static pthread_mutex_t sigmx = PTHREAD_MUTEX_INITIALIZER;
static unsigned int sigcount[NSIG];
static sigset_t sigcantmask;
static sigset_t sigignore;


static void
sigcleanup (void *arg)
{
    pthread_mutex_t *mx = arg;
    pthread_mutex_unlock (mx);
}


int
sigaction (int signo, const struct sigaction *pna, struct sigaction *poa)
{
    rtpx_proc_t    *pr = _rtpx_self_proc ();

    if (!VALID_SIGNO (signo) || signo == SIGKILL || signo == SIGSTOP) {
	errno = EINVAL;
	return -1;
    }

    pthread_cleanup_push (sigcleanup, &pr->pr_mx);
    pthread_mutex_lock (&pr->pr_mx);

    {
	struct sigaction *sa = &pr->pr_sigaction[signo - 1];
	struct sigaction oa = *sa;

	if (pna) {
	    sigset_t        ocatch = pr->pr_sigcatch;

	    *sa = *pna;
	    sigaddset (&sa->sa_mask, signo);
	    sa->sa_mask &= ~sigcantmask;
	    if (pna->sa_handler == SIG_IGN ||
		(sigismember (&sigignore, signo)
		 && pna->sa_handler == SIG_DFL)) {
		sigdelset (&pr->pr_sigpend, signo);	/* never to be seen
							   again */
		sigdelset (&pr->pr_sigcatch, signo);
		sigaddset (&pr->pr_sigignore, signo);
	    }
	    else {
		sigdelset (&pr->pr_sigignore, signo);
		if (pna->sa_handler == SIG_DFL)
		    sigdelset (&pr->pr_sigcatch, signo);
		else
		    sigaddset (&pr->pr_sigcatch, signo);
	    }

	    /* do we need to start catching this exception */
	    if (pr->pr_sigcatch != ocatch) {
		if (sigismember (&pr->pr_sigcatch, signo)) {
		    pthread_mutex_lock (&sigmx);
		    if (sigcount[signo]++ == 0)
			_cpu_sigcatch (signo, 1);
		    pthread_mutex_unlock (&sigmx);
		}
		else {
		    pthread_mutex_lock (&sigmx);
		    assert (sigcount[signo] > 0);
		    if (--sigcount[signo] == 0)
			_cpu_sigcatch (signo, 0);
		    pthread_mutex_unlock (&sigmx);
		}
	    }
	}

	if (poa)
	    *poa = oa;
    }
    pthread_cleanup_pop (1);
    return 0;
}


/* 
 * Install a signal handler (old-fashioned style).
 */
void (*signal (int sig, void (*handler) (int))) (int) 
{
    struct sigaction nsa,
                    osa;
    nsa.sa_handler = handler;
    sigemptyset (&nsa.sa_mask);
    nsa.sa_flags = SA_RESTART;
    if (sigaction (sig, &nsa, &osa) < 0)
	return SIG_ERR;
    return osa.sa_handler;
}


int
_rtpx_kill (pid_t pid, int sig, union sigval sival, int code)
{
    struct sigevent sigev;

    sigev.sigev_notify = SIGEV_SIGNAL;
    sigev.sigev_signo = sig;
    sigev.sigev_value = sival;
    return _rtpx_sigraise (pid, code, &sigev);
}


int
sigqueue (pid_t pid, int sig, const union sigval sival)
{
    int             err = _rtpx_kill (pid, sig, sival, SI_QUEUE);
    if (err != 0) {
	errno = err;
	return -1;
    }
    return 0;
}


int
kill (pid_t pid, int sig)
{
    union sigval    sival;
    int             err;

    sival.sival_int = 0;
    err = _rtpx_kill (pid, sig, sival, SI_USER);
    if (err != 0) {
	errno = err;
	return -1;
    }
    return 0;

}

int
killpg (pid_t pgrp, int sig)
{
    return kill (-pgrp, sig);
}


int
raise (int sig)
{
    return kill (0, sig);
}


/* 
 * Change a threads signal mask.
 * Called in kernel mode.
 */

int
_rtpx_sigmask (rtpx_thread_t *cth, int how,
	       const sigset_t * np, sigset_t * op)
{
    sigset_t        om = cth->th_sigmask,
                    nm = 0;
    sigset_t        pend;

    if (np)
	nm = *np;
    if (op)
	*op = om;
    if (!np)
	return 0;

    switch (how) {
    case SIG_BLOCK:
	cth->th_sigmask |= (nm & ~sigcantmask);
	return 0;
    case SIG_SETMASK:
	cth->th_sigmask = (nm & ~sigcantmask);
	break;
    case SIG_UNBLOCK:
	cth->th_sigmask &= ~nm;
	break;
    default:
	return EINVAL;
    }

    /* we may have unmasked some signals that were pending */

    while (pend = (~cth->th_sigmask & cth->th_proc->pr_sigpend)) {
	/* turn process signals into thread signals */
	int             signo = _rtpx_sigfirst (pend);
	sigaddset (&cth->th_sigpend, signo);
	sigdelset (&cth->th_proc->pr_sigpend, signo);
	cth->th_siginfo[signo - 1] = cth->th_proc->pr_siginfo[signo - 1];
    }

    return (~cth->th_sigmask & cth->th_sigpend) ? -1 : 0;
}


int
sigprocmask (int op, const sigset_t * nmask, sigset_t * omask)
{
    /* XXX applied to current thread only - wrong */
    return pthread_sigmask (op, nmask, omask);
}



static const unsigned char setbit[256] = {
    /* 0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f */
    8, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* 0x */
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* 1x */
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* 2x */
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* 3x */
    6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* 4x */
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* 5x */
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* 6x */
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* 7x */
    7, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* 8x */
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* 9x */
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* ax */
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* bx */
    6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* cx */
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* dx */
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* ex */
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,	/* fx */
};


/* find first signal in set, return 0 if none set */
int
_rtpx_sigfirst (sigset_t pend)
{
    unsigned int    byte,
                    bit;
    for (bit = 1; pend != 0; pend >>= 8, bit += 8)
	if (byte = (pend & 0xff))
	    return bit + setbit[byte];
    return 0;
}


/* return true if a signal, uncaught, should terminate a process */
int
_rtpx_sigfatal (int signo)
{
    if (signo >= 0 & signo <= NSIG)
	return sigprop[signo] & SA_KILL;
    return 0;
}


/* 
 * Process pending signals for current thread CTH.
 * Called in kernel mode.
 */
static void
rtpx_sigproc (rtpx_thread_t *cth, sigset_t pend)
{
    struct sigaction sa;
    siginfo_t       si;
    sigset_t        omask;
    int             signo;
    int		    fatal;

    signo = _rtpx_sigfirst (pend);
    assert (signo > 0);
    sigdelset (&cth->th_sigpend, signo);

    sa = cth->th_proc->pr_sigaction[signo - 1];
    si = cth->th_siginfo[signo - 1];

    /* note whether signal will be fatal */
    fatal = (!sa.sa_handler 
	     && (si.si_code >= SI_FAULT
		 || (sigprop[signo] & SA_ABORT)));

    /* give debuggers etc a change to handle the signal */
    signo = _cpu_sighook (cth, signo, fatal);
    if (signo <= 0)
	return;

    omask = cth->th_sigmask;
    sigaddset (&cth->th_sigmask, signo);
    cth->th_sigmask |= sa.sa_mask;

    EVLOGSIGDELIVER (cth, signo);

    if (fatal)
	/* fatal signal aborts kernel with an error message */
	/* XXX should this just terminate the current process? */
	_cpu_sigabort (cth, signo, NULL);

    (void) _cpu_leave (cth, 0);

    /* now at "user" level */

    if (!sa.sa_handler)
	/* unignored, uncaught async signal kills this thread 
	   (XXX should it kill whole process?) */
	pthread_exit (PTHREAD_CANCELED);

    /* process the async signal (may not return if longjmps) */
    if (sa.sa_flags & SA_SIGINFO) {
	(sa.sa_sigaction) (signo, &si, NULL);
    }
    else {
	(sa.sa_handler) (signo);
    }

    /* return to kernel mode and restore old signal mask */
    cth = _cpu_enter ();
    cth->th_sigmask = omask;

#if 1
    EVLOGSIGRETURNED (cth, signo);
#endif
}


void
_rtpx_sigcheck (rtpx_thread_t *cth)
{
    sigset_t        pend;
    while ((pend = cth->th_sigpend & ~cth->th_sigmask) != 0)
	rtpx_sigproc (cth, pend);
}


/* 
 * Reset signal handling for this process by removing any signal
 * handlers. Called when creating a new process.
 */

void
_rtpx_sigcopy (rtpx_proc_t *pr, const rtpx_proc_t *ppr)
{
    /* copy signal state */
    pr->pr_sigpend = ppr->pr_sigpend;
    pr->pr_sigcatch = ppr->pr_sigcatch;
    pr->pr_sigignore = ppr->pr_sigignore;
    /* pr->pr_sigmask = ppr->pr_sigmask */

    while (pr->pr_sigcatch) {
	int             signo = _rtpx_sigfirst (pr->pr_sigcatch);
	sigdelset (&pr->pr_sigcatch, signo);
	if (sigismember (&sigignore, signo)) {
	    if (signo != SIGCONT)
		sigaddset (&pr->pr_sigignore, signo);
	    sigdelset (&pr->pr_sigpend, signo);
	}
	pr->pr_sigaction[signo - 1].sa_handler = SIG_DFL;
    }
}


void
_rtpx_sigexit (rtpx_proc_t *pr)
{
    sigset_t        catch = pr->pr_sigcatch & ~sigcantmask;

    while (catch) {
	int             signo = _rtpx_sigfirst (catch);
	pthread_mutex_lock (&sigmx);
	assert (sigcount[signo] > 0);
	if (--sigcount[signo] == 0)
	    _cpu_sigcatch (signo, 0);
	pthread_mutex_unlock (&sigmx);
	sigdelset (&catch, signo);
    }
}


/* 
 * Initialise first process
 */
void
_rtpx_siginit (rtpx_proc_t *pr)
{
    int             signo;

    sigemptyset (&sigcantmask);
    sigemptyset (&sigignore);

    for (signo = 1; signo < NSIG; signo++) {
	if (sigprop[signo] & SA_CANTMASK)
	    sigaddset (&sigcantmask, signo);
	if (sigprop[signo] & SA_IGNORE)
	    sigaddset (&sigignore, signo);
    }
    pr->pr_sigcatch = sigcantmask;
    pr->pr_sigignore = sigignore;
    pr->pr_sigpend = 0;
    /* pr->pr_sigmask = 0; */
}
