#ident "sde/rtpx/rtpx_sigwait.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2001-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_sigwait.c: Posix sigwait etc.
 */


#include "pthread_impl.h"

static int
rtpx_sigtimedwait (const sigset_t * set, siginfo_t * info,
		   const struct timespec *timeout)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    sigset_t        pend;
    int             rv = 0;

    if (timeout && !timespecok (timeout))
	return rtpx_leaveWithValue (cth, EINVAL);

    pend = *set & cth->th_sigpend;

    if (pend == 0) {
	/* nothing pending - shall we wait? */
	if (timeout) {
	    clock64_t       when = _rtpx_tsToClock64 (timeout);
	    if (! clock64isset (&when)) {
		/* just polling */
		rv = EAGAIN;
		goto done;
	    }
	    _rtpx_itimerstart (&cth->th_itimer, IT_TYPE_TIMEOUT, when);
	}

	/* wait for any of the named signals to occur */
	cth->th_sigwait = *set;
	rv = rtpx_leaveWhenResumed (cth, THSTATE_SIGWAIT, NULL, 0);

	/* reenter kernel and recompute pending signals */
	(void) rtpx_enter ();
	pend = *set & cth->th_sigpend;
    }

    if (pend != 0) {
	/* unmasked signal pending - return it to caller */
	int             signo;
	signo = _rtpx_sigfirst (pend);
	assert (signo > 0);
	*info = cth->th_siginfo[signo - 1];
	sigdelset (&cth->th_sigpend, signo);
	rv = 0;
    }

  done:
    return rtpx_leaveWithValueLong (cth, rv);
}


int
sigwait (const sigset_t * set, int *signalp)
{
    int             rv;
    siginfo_t	    si;

    rv = rtpx_sigtimedwait (set, &si, NULL);
    if (rv == 0 && signalp)
	*signalp = si.si_signo;

    /* Pthreads error return */
    return rv;
}


int
sigtimedwait (const sigset_t * set, siginfo_t * sip,
	      const struct timespec *timeout)
{
    int		    rv;
    siginfo_t	    si;

    if (!sip)
	sip = &si;

    rv = rtpx_sigtimedwait (set, sip, timeout);
    if (rv == 0)
	return sip->si_signo;

    /* POSIX.1b error return */
    errno = rv;
    return -1;
}


int
sigwaitinfo (const sigset_t * set, siginfo_t * si)
{
    return sigtimedwait (set, si, NULL);
}


int
sigsuspend (const sigset_t * mask)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    sigset_t 	    omask;
    
    if (_rtpx_sigmask (cth, SIG_SETMASK, mask, &omask) == 0) {
	/* no pending signals, block until an unmasked one arrives */
	sigemptyset (&cth->th_sigwait);
	rtpx_leaveWhenResumed (cth, THSTATE_SIGWAIT, NULL, 0);
    }
    else {
	/* exit kernel mode and run pending handlers */
	rtpx_leaveWithValueLong (cth, 0);
    }

    /* reenter kernel and restore old mask */
    (void) rtpx_enter ();

    /* restore old mask */
    _rtpx_sigmask (cth, SIG_SETMASK, &omask, NULL);

    /* always return EINTR! */
    errno = EINTR;
    return rtpx_leaveWithValueLong (cth, -1);
}


int
sigpending (sigset_t * set)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    *set = cth->th_sigpend | cth->th_proc->pr_sigpend;
    return rtpx_leaveWithValue (cth, 0);
}
