#ident "sde/rtpx/rtpx_spawn.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2001-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_spawn.c: heavy-weight process emulation
 */


#include "pthread_impl.h"
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>


/* Thread communication structure */
struct procinit {
    pthread_mutex_t	pi_mx;
    pthread_cond_t	pi_sig;
    int			(*pi_fun) (void *);
    void		*pi_arg;
    int			pi_pid;
    int			pi_err;
    int			pi_done;
};


/* New thread which upgrades itself to a new process, and passes
   the results back to its parent thread, and then runs the new
   process code. */

static void *
launchproc (void *arg)
{
    volatile struct procinit *vpi = arg;
    struct procinit *pi = arg;
    int (*nfun) (void *) = vpi->pi_fun;
    void *narg = vpi->pi_arg;
    int err;

    /* wait until parent blocks */
    pthread_mutex_lock (&pi->pi_mx);

    /* upgrade self to a process */
    vpi->pi_err = err = _rtpx_newproc (&pi->pi_pid);

    /* signal parent to collect results */
    pi->pi_done = 1;
    pthread_cond_signal (&pi->pi_sig);
    pthread_mutex_unlock (&pi->pi_mx);

    /* the "pi* structure may now have disappeared (it was 
       on our parent's stack), so don't use it again */

    /* run the chosen process' "main" function */
    if (err == 0)
	exit ((*nfun) (narg));

    return NULL;
}


/*
 * Create a new process, returning its process id in *pidp. 
 * The new process will start off running the function FUN, 
 * and be passed argument ARG. 
 */
int
pthread_proc_create_np (int *pidp, const pthread_attr_t *attrp,
			int (*fun)(void *), void *arg)
{
    pthread_attr_t tha;
    struct procinit pi;
    int err;

    /* initialise thread communication structure */
    pthread_mutex_init (&pi.pi_mx, NULL);
    pthread_cond_init (&pi.pi_sig, NULL);
    pi.pi_fun = fun;
    pi.pi_arg = arg;
    pi.pi_done = 0;

    /* get thread attributes, but force it to be "detached" */
    if (attrp)
	tha = *attrp;
    else
	pthread_attr_init (&tha);
    pthread_attr_setdetachstate (&tha, PTHREAD_CREATE_DETACHED);

    /* prevent child from signalling us until we're ready */
    pthread_mutex_lock (&pi.pi_mx);

    /* create process launcher thread */
    err = pthread_create (NULL, &tha, launchproc, &pi);
    if (err) {
	pthread_mutex_unlock (&pi.pi_mx);
	return err;
    }

    /* destroy thread attributes */
    pthread_attr_destroy (&tha);

    /* wait for child to signal us */

    /* XXX Note that pthread_cond_wait could be interrupted and return
       EINTR after unlocking pi.pi_mx, but before the child locks
       pi.pi_mx and fills in pi.pi_pid. So we must loop until we know
       it has completed. An alternative would be to block signals and
       cancellations while starting the child. */

    do {
	(void) pthread_cond_wait (&pi.pi_sig, &pi.pi_mx);
    } while (! pi.pi_done);

    pthread_mutex_unlock (&pi.pi_mx);

    /* destroy mutex and cond var */
    pthread_mutex_destroy (&pi.pi_mx);
    pthread_cond_destroy (&pi.pi_sig);
    
    /* return status to user */
    if (pidp)
	*pidp = pi.pi_pid;
    return pi.pi_err;
}


/*
 * Process exit.
 * Kill all threads in the current process, and pass exit status
 * back to the parent process (if it still exists).
 */
void
_exit (int rv)
{
    /* cancel all threads own process */
    rtpx_thread_t  *cth = rtpx_enter ();
    rtpx_proc_t    *pr = _rtpx_self_proc ();
    rtpx_thread_t  *oth;

    /* store exit value for parent */
    pr->pr_exitval = W_EXITCODE (rv, 0);

    /* enable own cancellation */
    cth->th_flags &= ~(RTPX_THF_CANCEL_DISABLE | RTPX_THF_CANCEL_DEFERRED);

    /* cancel all threads in process */
    for (oth = _rtpx_firstthread (); oth; oth = _rtpx_nextthread (oth)) {
	if (oth->th_proc == pr && oth->th_state != THSTATE_ZOMBIE
	    && oth->th_tha.tha_detachstate != PTHREAD_CREATE_DAEMON_NP)
	    _rtpx_sendcancel (cth, oth);
    }

    /* leave, never to return... */
    (void) _rtpx_leaveWithValueAndDispatch (cth, 0);
    abort ();
}



/* XXX non-Posix BSD variants */
pid_t	wait3 __P((int *, int, struct rusage *));
pid_t	wait4 __P((pid_t, int *, int, struct rusage *));

static void
wait_cleanup (void *arg)
{
    rtpx_proc_t *pr = arg;
    pr->pr_waiting--;
    pthread_mutex_unlock (&pr->pr_mx);
}

/*
 * BSD wait4() 
 * Wait for a child process to terminate and return its exit status
 */
pid_t
wait4 (pid_t wantpid, int *status, int options, struct rusage *rusage)
{
    rtpx_proc_t *pr = _rtpx_self_proc ();
    rtpx_proc_t *cpr;
    int err = 0;
    pid_t res = 0;
    int foundany;

    pthread_mutex_lock (&pr->pr_mx);

    do {
	/* scan for any dead children */
	foundany = 0;
	for (cpr = _rtpx_proctab; cpr < &_rtpx_proctab[_RTPX_PROC_MAX]; cpr++) {
	    if (cpr != pr && cpr->pr_state != _RTPX_PROC_FREE
		&& cpr->pr_ppid == pr->pr_pid) {
		/* found a child process, see if it matches */
		if (wantpid == -1 ? 1
		    : wantpid < 0 ? cpr->pr_pgrp == -wantpid
		    : wantpid == 0 ? cpr->pr_pgrp == pr->pr_pgrp
		    : cpr->pr_pid == wantpid) {
		    /* no need to lock child, it has to lock us before 
		       changing its state. */
		    ++foundany;	
		    if (cpr->pr_state == _RTPX_PROC_ZOMBIE) {
			/* got a zombie child, return its result */
			res = cpr->pr_pid;
			if (status)
			    *status = cpr->pr_exitval;
			if (rusage)		/* XXX */
			    memset (rusage, 0, sizeof (struct rusage));
			cpr->pr_state = _RTPX_PROC_FREE;
			goto done;
		    }
		}
	    }
	}
	
	if (foundany == 0) {
	    err = ECHILD;
	    break;
	}

	if (options & WNOHANG)
	    break;

	pthread_cleanup_push (wait_cleanup, pr);
	pr->pr_waiting++;
	err = pthread_cond_wait (&pr->pr_waitcd, &pr->pr_mx);
	pr->pr_waiting--;
	pthread_cleanup_pop (0);
	
    } while (err == 0);

 done:
    pthread_mutex_unlock (&pr->pr_mx);
    if (err) {
	errno = err;
	res = -1;
    }
    return res;
}

/*
 * POSIX waitpid() 
 */
pid_t
waitpid (pid_t pid, int *status, int options)
{
    return wait4 (pid, status, options, NULL);
}

/*
 * BSD wait3() 
 */
pid_t
wait3 (int *status, int options, struct rusage *rusage)
{
    return wait4 (-1, status, options, rusage);
}

/*
 * POSIX wait3() 
 */
pid_t
wait (int *status)
{
    return wait4 (-1, status, 0, NULL);
}
