#ident "sde/rtpx/rtpx_stats.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 2000-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_stats.c: real-time statistics and logging
 */


#include "pthread_impl.h"
#include <string.h>
#include <ctype.h>

/* global debug control */
extern int      _rtpx_mutex_default_type;
extern int      _rtpx_deadlock_fatal;
extern int      _rtpx_break_debug;
extern int      _rtpx_sync_tty;

#ifndef _RTPX_NO_STATS

/* global circular logging control */
#pragma weak _rtpx_evlog_mask
unsigned int    _rtpx_evlog_mask = 0;
#pragma weak _rtpx_evlog_size
unsigned int    _rtpx_evlog_size = 0;

/* global miscellaneous stats */
struct rtpx_stats _rtpx_stats;

/* circular event log */
static struct {
    unsigned int    first;
    unsigned int    cnt;
    unsigned int    size;
    struct rtpx_evrec *buf;
} evlog;


static const char *const statename[] = {
    "ready",
    "timew",
    "semw",
    "mutexw",
    "condw",
    "sigw",
    "dbellw",
    "zombie",
    "dead"
};


static const char *const timername[] = {
    "dead",
    "base",
    "sleep",
    "wait",
    "sig",
};


void
_rtpx_elog_add (rtpx_thread_t *th, unsigned short what, void *arg)
{
    struct rtpx_evrec *ev;
    unsigned int emax;

    if (evlog.buf == 0)
	return;

    emax = evlog.size - 1;
    ev = &evlog.buf[(evlog.first + evlog.cnt) & emax];
    if (evlog.cnt < emax)
	/* append to end */
	++evlog.cnt;
    else
	/* wrap around */
	evlog.first = (evlog.first + 1) & emax;

    ev->ev_th = th;
    ev->ev_what = what;
    ev->ev_arg = arg;
    ev->ev_pri = th->th_prio;
    ev->ev_state = th->th_state;
    ev->ev_time = _rtpx_relclock ();
}

/* generic log function fdr user code */
void
_rtpx_elog (struct _rtpx_eloginfo *ei, ...)
{
    rtpx_thread_t  *cth;
    struct rtpx_evrec *ev;
    unsigned int    emax;
    va_list         ap;
    int             i,
                    n;

    /* quick cop out */
    if ((_rtpx_evlog_mask & (1 << (ELOG_USER >> 12))) == 0)
	return;

    if (evlog.buf == 0)
	return;

    /* work out how many args to save */
    if ((n = ei->nargs) < 0) {
	const char     *s = ei->fmt;
	char            c;
	n = 0;
	/* count the number of singleton '%' signs */
	while ((c = *s++) != '\0') {
	    if (c == '%') {
		if ((c = *s++) == '%')
		    continue;
		if (c == '\0')
		    break;
		++n;
	    }
	}
	if (n > ELOG_NARGS)
	    n = ELOG_NARGS;
	ei->nargs = n;
    }

    if (ei->intern)
	cth = rtpx_self ();
    else
	cth = _cpu_enter ();

    emax = evlog.size - 1;
    ev = &evlog.buf[(evlog.first + evlog.cnt) & emax];
    if (evlog.cnt < emax)
	++evlog.cnt;
    else
	evlog.first = (evlog.first + 1) & emax;

    ev->ev_th = cth;
    ev->ev_what = ELOG_USER | n;
    ev->ev_time = _rtpx_relclock ();
    ev->ev_arg = ei;

    va_start (ap, ei);
    for (i = 1; i <= n; i++)
	ev->ev_arglist[i] = va_arg (ap, void *);
    va_end (ap);

    if (!ei->intern)
	(void) _cpu_leave (cth, 0);
}



static const char *
mxname (const pthread_mutex_t *mx)
{
    if (!_cpu_badaddr (mx)
	&& VALID_TYPE (mx->mx_mxa.mxa_type)
	&& !_cpu_badaddr (mx->mx_mxa.mxa_name))
	return mx->mx_mxa.mxa_name;
    /* XXX use sprintf to create errmsg in static buffer */
    return NULL;
}


static const char *
cdname (const pthread_cond_t *cd)
{
    if (!_cpu_badaddr (cd)
	&& VALID_PSHARED (cd->cd_cda.cda_pshared)
	&& !_cpu_badaddr (cd->cd_cda.cda_name))
	return cd->cd_cda.cda_name;
    /* XXX use sprintf to create errmsg in static buffer */
    return NULL;
}

static const char *
dbname (const pthread_doorbell_t *db)
{
    if (!_cpu_badaddr (db)
	&& db->db_dba.dba_evn > 0
	&& db->db_dba.dba_evn < _SBD_INTR_MAX
	&& !_cpu_badaddr (db->db_dba.dba_name))
	return db->db_dba.dba_name;
    /* XXX use sprintf to create errmsg in static buffer */
    return NULL;
}

static const char *
semname (const sem_t *sem)
{
    if (!_cpu_badaddr (sem)
	&& sem->sem_value < SEM_VALUE_MAX
	&& !_cpu_badaddr (sem->sem_name))
	return sem->sem_name;
    /* XXX use sprintf to create errmsg in static buffer */
    return NULL;
}



void
_rtpx_elog_dump ()
{
    int             s = _cpu_intdisable ();
    struct rtpx_evrec *ev;
    struct timespec ts;
    clock64_t       dumptime = _rtpx_relclock ();
    clock64_t       first, last, delta;

    clock64clear (&first);
    clock64clear (&last);

    if (!evlog.buf) {
	_rtpx_log (RTPXLOG_PRINTF, "No event log\n");
	return;
    }

    while (evlog.cnt != 0) {
	ev = &evlog.buf[evlog.first];
	evlog.first = (evlog.first + 1) & (evlog.size - 1);
	evlog.cnt--;

	if (! clock64isset (&first)) {
	    first = ev->ev_time;
	    _rtpx_tsFromClock64 (&ts, first);
	    _rtpx_log (RTPXLOG_PRINTF, "%lu.%09lu: log start time\n",
		       ts.tv_sec, ts.tv_nsec);
	}
	last = ev->ev_time;

	clock64sub (&ev->ev_time, &first, &delta);
	_rtpx_tsFromClock64 (&ts, delta);
	_rtpx_log (RTPXLOG_PRINTF, " %lu.%09lu: %s ", 
		   ts.tv_sec, ts.tv_nsec, ev->ev_th->th_tha.tha_name);

	switch (ev->ev_what & 0xf000) {
	case ELOG_MUTEX:
	    {
		pthread_mutex_t *mx = ev->ev_arg;
		const char *name = mxname (mx);
		if (name)
		    _rtpx_log (RTPXLOG_PRINTF, "%s ", name);
		else
		    _rtpx_log (RTPXLOG_PRINTF, "mx=%p ", mx);
		_rtpx_log (RTPXLOG_PRINTF, "%s\n",
			   (ev->ev_what & 0x1) ? "acquired" : "released");
		break;
	    }
	case ELOG_STATE:
	    {
		const char     *name = NULL;
		_rtpx_log (RTPXLOG_PRINTF, "newstate=%s",
			   statename[ev->ev_what & 0xff]);
		switch (ev->ev_what & 0xff) {
		case THSTATE_MUTEXWAIT:
		    name = mxname ((pthread_mutex_t *) ev->ev_arg);
		    break;
		case THSTATE_CONDWAIT:
		    name = cdname ((pthread_cond_t *) ev->ev_arg);
		    break;
		case THSTATE_DOORBELLWAIT:
		    name = dbname ((pthread_doorbell_t *) ev->ev_arg);
		    break;
		case THSTATE_SEMWAIT:
		    name = semname ((sem_t *) ev->ev_arg);
		    break;
		}
		if (name)
		    _rtpx_log (RTPXLOG_PRINTF, " %s\n", name);
		else if (ev->ev_arg)
		    _rtpx_log (RTPXLOG_PRINTF, " %p\n", ev->ev_arg);
		else
		    _rtpx_log (RTPXLOG_PRINTF, "\n");
		break;
	    }
	case ELOG_TIMER:
	    _rtpx_log (RTPXLOG_PRINTF, "%s timer expired\n",
		       timername[ev->ev_what & 0xff]);
	    break;
	case ELOG_PRIO:
	    _rtpx_log (RTPXLOG_PRINTF, "prio changed from %d to %d\n", 
		       ev->ev_pri, ev->ev_what & 0xfff);
	    break;
	case ELOG_EVENT:
	    {
		if (ev->ev_arg) {
		    const char *name;
		    name = dbname ((pthread_doorbell_t *)ev->ev_arg);
		    if (name)
			_rtpx_log (RTPXLOG_PRINTF, "%s ", name);
		    else
			_rtpx_log (RTPXLOG_PRINTF, "db=%p ", ev->ev_arg);
		}
		_rtpx_log (RTPXLOG_PRINTF, "event=%d\n", ev->ev_what & 0xfff);
		break;
	    }
	case ELOG_INTERUPT:
	    _rtpx_log (RTPXLOG_PRINTF, "interrupt cr=%x sr=%x\n",
		       ev->ev_what & 0xfff, (unsigned int) ev->ev_arg);
	    break;
	case ELOG_SWTCH:
	    {
		rtpx_thread_t  *th = ev->ev_arg;
		_rtpx_log (RTPXLOG_PRINTF, "swtch to %s, prio %d\n",
			   th->th_tha.tha_name, ev->ev_what & 0xfff);
		break;
	    }
	case ELOG_SCHED:
	    {
		rtpx_thread_t  *th = ev->ev_arg;
		_rtpx_log (RTPXLOG_PRINTF, "round-robin to %s\n",
			   th->th_tha.tha_name);
	    }
	    break;
	case ELOG_CREATE:
	    {
		rtpx_thread_t  *th = ev->ev_arg;
		_rtpx_log (RTPXLOG_PRINTF, "creates %s, prio %d\n",
			   th->th_tha.tha_name, ev->ev_what & 0xfff);
		break;
	    }
	case ELOG_SIGQUEUE:
	    _rtpx_log (RTPXLOG_PRINTF, "signal=%d queued for ",
		       ev->ev_what & 0xff);
	    if (ev->ev_what & 0x100) {
		rtpx_proc_t    *pr = ev->ev_arg;
		_rtpx_log (RTPXLOG_PRINTF, "proc %d (%p)\n", pr->pr_pid, pr);
	    }
	    else {
		rtpx_thread_t  *th = ev->ev_arg;
		_rtpx_log (RTPXLOG_PRINTF, "thread %s\n", th->th_tha.tha_name);
	    }
	    break;
	case ELOG_SIGDELIVER:
	    _rtpx_log (RTPXLOG_PRINTF, "signal %d %s\n",
		       ev->ev_what & 0xff, ev->ev_arg ? "returning" : "delivered");
	    break;
	case ELOG_CANCEL:
	    {
		rtpx_thread_t  *th = ev->ev_arg;
		switch (ev->ev_what & 0xff) {
		case 0:
		    _rtpx_log (RTPXLOG_PRINTF, "cancel queued for %s\n",
			       th->th_tha.tha_name);
		    break;
		case 1:
		    _rtpx_log (RTPXLOG_PRINTF, "cancel sent to %s\n",
			       th->th_tha.tha_name);
		    break;
		case 2:
		    _rtpx_log (RTPXLOG_PRINTF, "cancelling\n");
		    break;
		}
		break;
	    }
	case ELOG_USER:
	    {
		struct _rtpx_eloginfo *ei = ev->ev_arg;
		_rtpx_vlog (RTPXLOG_PRINTF, ei->fmt, &ev->ev_arglist[1]);
		_rtpx_log (RTPXLOG_PRINTF, "\n");
		break;
	    }
	default:
	    _rtpx_log (RTPXLOG_PRINTF, "what=0x%x???\n", ev->ev_what);
	    break;
	}
    }
    if (clock64cmp (&dumptime, &last, >)) {
	clock64sub (&dumptime, &last, &delta);
	_rtpx_tsFromClock64 (&ts, delta);
	_rtpx_log (RTPXLOG_PRINTF, "%lu.%09lu: stopped\n", ts.tv_sec,
		   ts.tv_nsec);
    }
    _cpu_intrestore (s);
}



static int
tsprintf (const char *fmt, ...)
{
    va_list ap;
    va_start (ap, fmt);
    _mon_vprintf (fmt, ap);
    va_end (ap);
    return 0;
}


/* callable from gdb to dump thread table to console */
void
_rtpx_threadstat ()
{
    int s = _cpu_intdisable ();
    _mon_printf ("\n");
    _rtpx_threadstat_user (tsprintf);
    _cpu_intrestore (s);
}


/* callable by user */
void
_rtpx_threadstat_user (int (*prf)(const char *, ...))
{
    extern void     _cpu_printctx (rtpx_thread_t *);
    rtpx_thread_t  *cth = rtpx_self ();
    rtpx_thread_t  *th;
    const char     *waitfor;
    clock64_t       dumptime = _rtpx_clock ();

    prf ("  PID   TID     NAME THADDR   PRI FLG STATE     WAITNAME WAITADDR   TIMER  STKSZ\n");

    for (th = _rtpx_firstthread (); th; th = _rtpx_nextthread (th)) {
	prf ("%5d ", th->th_proc->pr_pid);

	prf ("%5d ", th->th_id);

	prf ("%8.8s %08lx ", th->th_tha.tha_name, (unsigned long) th);

	prf ("%3u %c%c%c ", th->th_prio,
	     (th->th_tha.tha_schedpolicy == SCHED_FIFO) ? 'F' : 'R',
	     th->th_sigpend ? 'S' : ' ',
	     (th->th_flags & RTPX_THF_CANCEL_PENDING) ? 'C' : ' ');

	if (th->th_state == THSTATE_CPUWAIT && th == cth)
	    prf ("running ");
	else
	    prf ("%-7s ", statename[th->th_state]);

	waitfor = NULL;
	switch (th->th_state) {
	case THSTATE_CPUWAIT:
	case THSTATE_ZOMBIE:
	    waitfor = (void *) -1;
	    break;
	case THSTATE_MUTEXWAIT:
	    waitfor = mxname ((pthread_mutex_t *) th->th_waitq);
	    break;
	case THSTATE_CONDWAIT:
	    waitfor = cdname ((pthread_cond_t *) th->th_waitq);
	    break;
	case THSTATE_DOORBELLWAIT:
	    waitfor = dbname ((pthread_doorbell_t *) th->th_waitq);
	    break;
	case THSTATE_SEMWAIT:
	    waitfor = semname ((sem_t *) th->th_waitq);
	    break;
	default:
	    break;
	}

	if (waitfor == (void *) -1)
	    prf ("%20s", "");
	else {
	    prf ("%10.10s ", waitfor ? waitfor : "");
	    if (th->th_waitq)
		prf ("%08lx ", (unsigned long) th->th_waitq);
	    else
		prf ("%9s", "");
	}

	if (th->th_itimer.it_type != IT_TYPE_DEAD) {
	    struct timespec ts;
	    if (clock64cmp (&th->th_itimer.it_value, &dumptime, <=)) {
		prf ("%7s ", "pending");
	    }
	    else {
		clock64_t delta;
		clock64sub (&th->th_itimer.it_value, &dumptime, &delta);
		_rtpx_tsFromClock64 (&ts, delta);
		prf ("%3lu.%03lu ", ts.tv_sec, 
		     (ts.tv_nsec + 500000) / 1000000);
	    }
	}
	else {
	    prf ("%7s ", "");
	}

	prf ("%6u", _rtpx_stackused (th));

	/*_cpu_printctx (th);*/
	prf ("\n");
    }
}


void
_rtpx_mutexstat (const pthread_mutex_t *mx)
{
    int             s = _cpu_intdisable ();
    const pthread_mutexattr_t *mxa = &mx->mx_mxa;
    const char *name;

    if (_cpu_badaddr (mx))
	_rtpx_log (RTPXLOG_PRINTF, "%p: bad address\n", mx);

    name = mxname (mx);
    if (name)
	_rtpx_log (RTPXLOG_PRINTF, "Name:\t%s\n", name);

    _rtpx_log (RTPXLOG_PRINTF, "Type:\t%s\n",
	       (mxa->mxa_type == PTHREAD_MUTEX_FAST_NP) ? "fast/normal" :
	       (mxa->mxa_type == PTHREAD_MUTEX_RECURSIVE_NP) ? "recursive" :
	       (mxa->mxa_type == PTHREAD_MUTEX_ERRORCHECK_NP) ? "errcheck" :
	       (mxa->mxa_type == PTHREAD_MUTEX_DEFAULT_NP) ? "default?" :
	       "???");

    _rtpx_log (RTPXLOG_PRINTF, "Proto:\t%s\n",
	       (mxa->mxa_protocol == PTHREAD_PRIO_NONE) ? "none" :
	       (mxa->mxa_protocol == PTHREAD_PRIO_INHERIT) ? "inherit" :
	       (mxa->mxa_protocol == PTHREAD_PRIO_PROTECT) ? "protect" :
	       "???");

    _rtpx_log (RTPXLOG_PRINTF, "State:\t%slocked\n",
	       (mx->mx_flags & _RTPX_MXF_SLOWLOCK) ? "slow " :
	       (mx->mx_flags & _RTPX_MXF_FASTLOCK) ? "fast " : "un");

    if (mx->mx_owner) {
	if (_cpu_badaddr (mx->mx_owner))
	    _rtpx_log (RTPXLOG_PRINTF, "Owner:\tbad(%p)\n", mx->mx_owner);
	else
	    _rtpx_log (RTPXLOG_PRINTF, "Owner:\t%s\n",
		       mx->mx_owner->th_tha.tha_name);
    }

    if (mxa->mxa_type == PTHREAD_MUTEX_RECURSIVE_NP)
	_rtpx_log (RTPXLOG_PRINTF, "Count:\t%d\n", mx->mx_count);

    if (mx->mx_waitq.wq_first) {
	rtpx_thread_t  *th = mx->mx_waitq.wq_first;
	_rtpx_log (RTPXLOG_PRINTF, "Waiters:\n");
	if (_cpu_badaddr (th))
	    _rtpx_log (RTPXLOG_PRINTF, "  bad(%p)\n", th);
	else
	    for (; th; th = th->th_next)
		_rtpx_log (RTPXLOG_PRINTF, "  %c%s\n",
			   th == mx->mx_owner ? '!' : ' ',
			   th->th_tha.tha_name);
    }

    _cpu_intrestore (s);
}
#endif /* _RTPX_NO_STATS */


struct rmask {
    const char *	name;
    unsigned long	val;
};

static const struct rmask rmaskvals[] = {
    {"all",		~0},
    {"sem",		1 << (ELOG_SEM >> 12)},
    {"cancel",		1 << (ELOG_CANCEL >> 12)},
    {"sigdeliver",	1 << (ELOG_SIGDELIVER >> 12)},
    {"sigqueue",	1 << (ELOG_SIGQUEUE >> 12)},
    {"signal",		(1 << (ELOG_SIGDELIVER >> 12))
     			| (1 << (ELOG_SIGQUEUE >> 12))},
    {"timer",		1 << (ELOG_TIMER >> 12)},
    {"mutex",		1 << (ELOG_MUTEX >> 12)},
    {"mx",		1 << (ELOG_MUTEX >> 12)},
    {"state",		1 << (ELOG_STATE >> 12)},
    {"prio",		1 << (ELOG_PRIO >> 12)},
    {"event",		1 << (ELOG_EVENT >> 12)},
    {"db",		1 << (ELOG_EVENT >> 12)},
    {"intrupt",		1 << (ELOG_INTERUPT >> 12)},
    {"swtch",		1 << (ELOG_SWTCH >> 12)},
    {"create",		1 << (ELOG_CREATE >> 12)},
    {"sched",		1 << (ELOG_SCHED >> 12)},
    {"user",		1 << (ELOG_USER >> 12)},
    {0, 0}
};


void
_rtpx_stats_init ()
{
    char           *s;
    unsigned long   val;
    extern int      errno;

#ifndef _RTPX_NO_STATS
    memset (&_rtpx_stats, 0, sizeof (_rtpx_stats));

    memset (&evlog, 0, sizeof (evlog));

    if (_rtpx_evlog_size == 0) {
	_rtpx_evlog_size = _RTPX_EVLOGSIZE;
	if (s = getenv ("RTPX_EVLOG_SIZE")) {
	    errno = 0;
	    val = strtoul (s, NULL, 0);
	    if (errno != ERANGE)
		_rtpx_evlog_size = val;
	}
    }

    if (_rtpx_evlog_mask == 0) {
	if (s = getenv ("RTPX_EVLOG_MASK")) {
	    char *e;
	    while (*s) {
		unsigned long val = 0;
		int off = 0;
		e = strpbrk (s, ",: ");
		if (!e) 
		    e = s + strlen (s);
		if (*s == '-') {
		    off = 1;
		    s++;
		}
		if (isdigit (*s)) {
		    char *ed;
		    errno = 0;
		    val = strtoul (s, &ed, 0);
		    if (ed != e || errno == ERANGE) {
			*e = '\0';
			_rtpx_log (RTPXLOG_PANIC, "RTPX_EVLOG_MASK bad value '%s'",
				   s);
		    }
		}
		else {
		    const struct rmask *rm;
		    int len = e - s;
		    for (rm = rmaskvals; rm->name; rm++)
			if (strlen (rm->name) == len
			    && strncmp (rm->name, s, len) == 0) {
			    val = rm->val;
			    break;
			}
		    if (!rm->name) {
			*e = '\0';
			_rtpx_log (RTPXLOG_PANIC, "RTPX_EVLOG_MASK bad name '%s'",
				   s);
		    }
		}
		if (off)
		    _rtpx_evlog_mask &= ~val;
		else
		    _rtpx_evlog_mask |= val;
		s = e + strspn (e, ",: ");
	    }
	}
    }
#endif /* _RTPX_NO_STATS */

    if (_rtpx_mutex_default_type == PTHREAD_MUTEX_DEFAULT_NP) {
	if (getenv ("RTPX_MUTEX_CHECK"))
	    _rtpx_mutex_default_type = PTHREAD_MUTEX_ERRORCHECK_NP;
	else
	    _rtpx_mutex_default_type = PTHREAD_MUTEX_FAST_NP;
    }


    if (getenv ("RTPX_DEADLOCK_FATAL")) {
	_rtpx_deadlock_fatal = 1;
    }

    if (!getenv ("RTPX_NO_BREAK")) {
	_rtpx_break_debug = 1;
    }

    if (getenv ("RTPX_SYNC_TTY")) {
	_rtpx_sync_tty = 1;
    }

    if (_rtpx_evlog_size > 0) {
	/* reduce size to power of two */
	evlog.size = _rtpx_evlog_size & ~(_rtpx_evlog_size - 1);
	evlog.buf = malloc (evlog.size * sizeof (struct rtpx_evrec));
    }

    if (!evlog.buf) {
	/* no buffer, so disable tracing */
	evlog.size = 0;
	_rtpx_evlog_mask = 0;
    }
}
