#ident "sde/rtpx/rtpx_th.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2007 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_th.c: Posix Threads kernel 
 */


/* 
 * XXX This file is a bit monolithic - could it be split up some more?
 */

#include "pthread_impl.h"
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/time.h>

/* various global variables which can control behaviour */
#pragma weak _rtpx_mutex_default_type
int             _rtpx_mutex_default_type = 0;	/* default mutex type */
#pragma weak _rtpx_deadlock_fatal
int             _rtpx_deadlock_fatal = 0;	/* abort on mutex deadlock */
#pragma weak _rtpx_break_debug
int             _rtpx_break_debug = 0;		/* break signal trapped */
#pragma weak _rtpx_sync_tty
int             _rtpx_sync_tty = 0;		/* autoflush on write */

/* timer parameters */
long		_rtpx_tickspersec;	/* timer resolution */
int      	_rtpx_rrclock;		/* round-robin quantum (ticks) */
int      	_rtpx_rrfreq;		/* round-robin freq (Hz) */

/* concurrency level (currently unused) */
int             _rtpx_concurrency;

/* tells various C libraries that they are running with more than one thread */
int		__isthreaded = 0;
int 		__sdethread_threading = 0;

/* thread table */
static rtpx_thread_t rtpx_threads[_RTPX_THREAD_MAX];
static rtpx_thread_t *freeThreadList;
static unsigned int nThread;
static unsigned int threadID;

/* run queue */
static rtpx_waitq_t cpuWaitq;

/* the current "time of day" */
static struct rtpx_clock wallClock;

/* start time */
static clock64_t bootClock;

/* clock period in ns (fixed point) */
static unsigned long _rtpx_nspertick;

/* fixed point: 20 bit integer, 12 bit fraction */
#define NSSHIFT		12
#define NSSCALE		(1<<NSSHIFT)

/* idle timer to keep clock ticking over every 1/2 sec */
static rtpx_itimer_t baseTimer = {
    NULL, NULL, 
    IT_TYPE_BASE,
};

/* timer queue */
static rtpx_itimer_t *itCurrent = &baseTimer;

/* internal interval timer */
static unsigned rtpx_itproc (unsigned now);

/* resume a blocked thread */
static int      rtpx_resumeWithCleanups (rtpx_thread_t *ath, int rv);

/* internal mutex handling */
static void     mx_assign (pthread_mutex_t *, rtpx_thread_t *);
static void     mx_unassign (pthread_mutex_t *, rtpx_thread_t *);
static int      mx_wait (pthread_mutex_t *, rtpx_thread_t *, int);

/* thread cancellation */
static int      rtpx_cancel (rtpx_thread_t *oth);
#define rtpx_cancel_pending(ath) \
  (((ath)->th_flags & (RTPX_THF_CANCELLING | RTPX_THF_CANCEL_DISABLE \
		       | RTPX_THF_CANCEL_PENDING)) == RTPX_THF_CANCEL_PENDING)

/* change state of thread */
#define RTPX_SETSTATE(th,nstate,arg)	\
  do { \
    EVLOGSTATE(th,nstate,arg); \
    th->th_state = nstate; \
  } while (0);

/* local semaphores */
static pthread_mutex_t thMx = PTHREAD_MUTEX_INITIALIZER;

/* doorbell interrupts */
static pthread_doorbell_t *dbTable[_SBD_INTR_MAX];

/* default attributes */
static pthread_attr_t thaDefault;
static pthread_mutexattr_t mxaDefault;
static pthread_condattr_t cdaDefault;

/* stack "checking" */
#define STACKCOLOUR		0x55
#define REDSCARE		0x55AAEEDD

/* return old-fashioned errno result */
static __inline__ int
pthread_error (int error)
{
    errno = error;
    return -1;
}



/* 
 * Add thread to wait queue in priority order 
 */
static void
enqTh (rtpx_waitq_t *wq, rtpx_thread_t *th)
{
    rtpx_thread_t  *p = wq->wq_first;

    assert (th->th_waitq == NULL);
    if (p == NULL || p->th_prio < th->th_prio) {
	th->th_prev = NULL;
	th->th_next = p;
	if (p != NULL)
	    p->th_prev = th;
	wq->wq_first = th;
    }
    else {
	rtpx_thread_t  *n;
	while (n = p->th_next) {
	    if (n->th_prio < th->th_prio)
		break;
	    p = n;
	}
	th->th_prev = p;
	th->th_next = n;
	p->th_next = th;
	if (n)
	    n->th_prev = th;
    }
    th->th_waitq = wq;
}


/* 
 * Remove thread ftom wait queue 
 */
static void
deqTh (rtpx_waitq_t *wq, rtpx_thread_t *th)
{
    rtpx_thread_t  *p = th->th_prev;
    rtpx_thread_t  *n = th->th_next;

    assert (th->th_waitq == wq);
    if (p)
	p->th_next = n;
    else
	wq->wq_first = n;

    if (n)
	n->th_prev = p;
    th->th_waitq = NULL;
}


/* 
 * Return first thread from head of wait queue 
 */
rtpx_thread_t *
_rtpx_popq (rtpx_waitq_t *wq)
{
    rtpx_thread_t  *th = wq->wq_first;
    if (th) {
	rtpx_thread_t  *n = th->th_next;

	wq->wq_first = n;

	if (n)
	    n->th_prev = NULL;

	th->th_waitq = NULL;
    }
    return th;
}
#define popqTh	_rtpx_popq

/* 
 * Reposition thread in wait queue when priority changes 
 */
static void
reqTh (rtpx_waitq_t *wq, rtpx_thread_t *th, int newpri)
{
    deqTh (wq, th);
    th->th_prio = newpri;
    enqTh (wq, th);
}


/* 
 * Change thread priority 
 */
static void
rtpx_setepri (rtpx_thread_t *th, int newpri)
{
    EVLOGPRIO (th, newpri);
    switch (th->th_state) {
    case THSTATE_CPUWAIT:
	reqTh (&cpuWaitq, th, newpri);
	break;
    case THSTATE_SEMWAIT:
    case THSTATE_MUTEXWAIT:
    case THSTATE_CONDWAIT:
    case THSTATE_SIGWAIT:
    case THSTATE_DOORBELLWAIT:
	reqTh (th->th_waitq, th, newpri);
	break;
    case THSTATE_TIMEWAIT:
	th->th_prio = newpri;
	break;
    case THSTATE_ZOMBIE:
    case THSTATE_DEAD:
    default:
	abort ();
    }
}



/* 
 * Enter "kernel mode" where we cannot be interrupted by other threads
 * (although higher-prio interrupts are still possible), and return the
 * currently executing thread. 
 */
rtpx_thread_t *
_rtpx_enter ()
{
    rtpx_thread_t  *cth = _cpu_enter ();

    assert (cth->th_next == NULL || cth->th_prio >= cth->th_next->th_prio);

    return cth;
}

/* 
 * Leave kernel mode, returning a value to the caller 
 */
int
_rtpx_leaveWithValue (rtpx_thread_t *cth, int rv)
{
    return _cpu_leave (cth, rv);
}


/* 
 * Leave kernel mode, setting errno and returning -1 to the caller 
 */
int
_rtpx_leaveWithError (rtpx_thread_t *cth, int error)
{
    extern int (errno);
    (errno) = cth->th_errno = error;
    return rtpx_leaveWithValue (cth, -1);
}


/* 
 * Leave kernel mode, checking for cancellation and signals on the way 
 */
int
_rtpx_leaveWithValueLong (rtpx_thread_t *cth, int rv)
{
    /* handle pending cancellation */
    if (cth->th_flags & RTPX_THF_CANCELLED) {
	EVLOGCANCEL (cth, cth, 2);
	cth->th_flags &= ~RTPX_THF_CANCELLED;
	(void) rtpx_leaveWithValue (cth, rv);
	pthread_exit (PTHREAD_CANCELED);
    }

    /* check for pending signals */
    /* must be last because sighandler may longjmp */
    _rtpx_sigcheck (cth);

    return rtpx_leaveWithValue (cth, rv);
}




/* 
 * This is it - the main scheduler loop.
 * Exit kernel mode but choose the highest prio thread to run.
 * Also handles timer events and round-robin scheduling.
 */
int
_rtpx_leaveWithValueAndDispatch (rtpx_thread_t *cth, int rv)
{
    rtpx_thread_t  *nth;
    unsigned        now = _sbd_gettickval ();
    int             loopcount = 0;

    cth->th_wval = rv;

    if (cth->th_tha.tha_schedpolicy != SCHED_FIFO) {
	unsigned        runticks = now - _cpuTab[0].startTicks;
	if (runticks > cth->th_ticksToRr)
	    cth->th_ticksToRr = 0;
	else
	    cth->th_ticksToRr -= runticks;
    }

    while (1) {
	/* ticks until timer event */
	int             ticksToEvent = _cpuTab[0].timerTicks - now;

	if (ticksToEvent < -_rtpx_tickspersec
	    || ticksToEvent > _rtpx_tickspersec) {
	    /* we'll assume that this is because we've been stopped by a
	       debugger. */
	    _rtpx_log (RTPXLOG_DEBUG, "tick overrun fixup %d/%ld",
		       ticksToEvent, _rtpx_tickspersec);
	    ticksToEvent = 0;
	    if (cth->th_tha.tha_schedpolicy != SCHED_FIFO)
		cth->th_ticksToRr = 0;
	}

	if (ticksToEvent <= 0 || _cpuTab[0].timerReload) {
	    _cpuTab[0].timerTicks = rtpx_itproc (now);
	    {
		unsigned        when = _cpuTab[0].timerTicks;
		_cpuTab[0].timerReload = 0;
		_sbd_settickcmp (when);
	    }
	    now = _sbd_gettickval ();
	    if (++loopcount > 10)
		abort ();
	    continue;
	}

	loopcount = 0;

	{
	    /* This is where we may block waiting for an interrupt event if
	       there is nothing on the CPU run queue */
	    int             evn = _sbd_intrnext (cpuWaitq.wq_first == 0);
	    if (evn != 0) {
		if (evn == -1) {
		    /* clock interrupt always event -1, clear the clock
		       interrupt */
		    unsigned        when = _cpuTab[0].timerTicks;
		    _sbd_settickcmp (when);
		}
		else {
		    /* some other hardware interrupt, ring the doorbell */
		    assert (evn > 0 && evn < _SBD_INTR_MAX);
		    {
			pthread_doorbell_t *db = dbTable[evn];
			EVLOGEVN (cth, evn, db);
			if (db != NULL) {
			    /* wakeup up all waiting threads */
			    rtpx_thread_t  *ath;
			    while (ath = popqTh (&db->db_waitq))
				rtpx_resumeWithCleanups (ath, 0);
			}
		    }
		}
		now = _sbd_gettickval ();
		/* if (++loopcount > 10) abort (); */
		continue;
	    }
	}

	/* get the highest-priority runnable thread */
	nth = cpuWaitq.wq_first;
	assert (nth);

	/* check for stack overrun */
	if (nth->th_stack) {
	    long           *red = (long *) nth->th_stack;
	    if (red[0] != REDSCARE || red[1] != REDSCARE ||
		red[2] != REDSCARE || red[3] != REDSCARE)
		_rtpx_log (RTPXLOG_PANIC, "Stack overrun in thread '%s'",
			   nth->th_tha.tha_name);
	}

	/* check for end of round-robin timeslice */
	if (nth->th_tha.tha_schedpolicy != SCHED_FIFO) {
	    unsigned        runticks = nth->th_ticksToRr;

	    if (runticks > _rtpx_rrclock)
		/* _rtpx_rrclock may have been reduced */
		runticks = _rtpx_rrclock;

	    /* EVLOGSCHED(nth,runticks); */

	    if (runticks == 0) {
		/* round robin and try again */
		deqTh (&cpuWaitq, nth);
		nth->th_ticksToRr = _rtpx_rrclock;
		enqTh (&cpuWaitq, nth);

		/* log if round-robin has caused a reschedule */
		if (nth != cpuWaitq.wq_first)
		    EVLOGSCHED (nth, cpuWaitq.wq_first);

		now = _sbd_gettickval ();	/* reasonable */
		/* if (++loopcount > 10) abort (); */
		continue;
	    }

	    if (runticks < ticksToEvent)
		ticksToEvent = runticks;
	}

	assert ((unsigned) ticksToEvent < (unsigned) _rtpx_tickspersec);

	/* decide when the next timer interrupt should fire */
	{
	    unsigned        when = now + ticksToEvent;

	    _cpuTab[0].startTicks = now;
	    _sbd_settickcmp (when);
	    now = _sbd_gettickval ();
	    ticksToEvent = when - now;

	    if (ticksToEvent <= 0) {
		/* we've missed the clock interrupt, go around again */
		if (nth->th_tha.tha_schedpolicy != SCHED_FIFO) {
		    /* we may not have enough ticks to get out of this loop * 
		       make sure we charge so we get a RRCLOCK full some
		       time! */
		    unsigned int    runticks = now - _cpuTab[0].startTicks;
		    if (runticks > nth->th_ticksToRr)
			nth->th_ticksToRr = 0;
		    else
			nth->th_ticksToRr -= runticks;
		}
		/* if (++loopcount > 10) abort (); */
		continue;
	    }
	}
	break;
    }

    /* 
     * Now committed to scheduling nth 
     */

    if (nth != cth) {
	/* context switch to new thread */
	EVLOGSWTCH (cth, nth, nth->th_prio);
	RTPX_INCSTAT (ncswitch);
	cth = _cpu_swtch (cth, nth);
    }

    /* We are not the new thread */

    /* cpu-specific reschedule hook */
    _cpu_schedhook (cth);

    /* return to scheduled thread */
    return rtpx_leaveWithValueLong (cth, cth->th_wval);
}


/* 
 * Called from board-specific interrupt code to force a reschedule.
 * Should only be called from "user mode", never from kernel mode.
 */
int
_rtpx_eoi (rtpx_thread_t *cth)
{
    return rtpx_leaveWithValueAndDispatch (cth, 0);
}



/* 
 * Put thread on a waitq; it will leave kernel mode and return
 * when the thread is "resumed" and return value RV.
 */
int
_rtpx_leaveWhenResumed (rtpx_thread_t *cth, int nstate,
			rtpx_waitq_t *wq, int rv)
{
    deqTh (&cpuWaitq, cth);
    RTPX_SETSTATE (cth, nstate, wq);
    if (wq)
	enqTh (wq, cth);

    if (rtpx_cancel_pending (cth) && nstate != THSTATE_MUTEXWAIT)
	/* about to block with a pending cancellation - cancel now */
	(void) rtpx_cancel (cth);

    return rtpx_leaveWithValueAndDispatch (cth, rv);
}


/* 
 * Put thread on a waitq; this will leave kernel mode and return
 * when the thread is "resumed". The mutex MX will be unlocked 
 * before waiting, but will have been relocked before returning.
 * If ABSTIME is non-null then the wait will be timed out. 
 */
static int
rtpx_leaveWhenResumedWithMx (rtpx_thread_t *cth, int nstate,
			      rtpx_waitq_t *wq,
			      pthread_mutex_t *mx,
			      const struct timespec *abstime)
{
    rtpx_thread_t  *oth;

    assert (nstate == THSTATE_CONDWAIT || nstate == THSTATE_DOORBELLWAIT);

    /* mutex must be currently locked by this thread and not a recursive type 
     */
    if (!mx || !(mx->mx_flags & _RTPX_MXF_LOCKED)
	|| ((mx->mx_flags & _RTPX_MXF_SLOWLOCK)
	    && mx->mx_owner != cth)
	|| (mx->mx_mxa.mxa_type == PTHREAD_MUTEX_RECURSIVE_NP))
	return rtpx_leaveWithValue (cth, EINVAL);

    /* and must be same as other waiters on this condition */
    for (oth = wq->wq_first; oth; oth = oth->th_next)
	if (oth->th_waitMx != mx)
	    return rtpx_leaveWithValue (cth, EINVAL);

    if (abstime && !timespecok (abstime))
	return rtpx_leaveWithValue (cth, EINVAL);

    /* release mutex (required by spec, may schedule another thread) */
    if (mx->mx_flags == _RTPX_MXF_FASTLOCK) {
	mx->mx_flags = 0;
#ifndef NDEBUG
	mx->mx_owner = NULL;
#endif
    }
    else
	mx_unassign (mx, cth);

    if (abstime) {
	/* start a timeout timer */
	clock64_t       when = _rtpx_tsToClock64 (abstime);
	_rtpx_itimerstartabs (&cth->th_itimer, IT_TYPE_TIMEOUT, when);
    }

    /* store mutex to be reacquired then block */
    cth->th_waitMx = mx;
    return rtpx_leaveWhenResumed (cth, nstate, wq, 0);
}



/* 
 * Wake up a blocked thread ATH by posting on cpuWaitq, and cause it
 * to return with value RV.
 * Returns true if we need to reschedule.
 */
int
_rtpx_resumeWithValue (rtpx_thread_t *ath, int rv)
{
    assert (ath->th_itimer.it_type == IT_TYPE_DEAD);
    assert (ath->th_waitq == NULL);
    assert (ath->th_waitMx == NULL);

    ath->th_wval = rv;
    ath->th_ticksToRr = _rtpx_rrclock;	/* restart time slice */
    RTPX_SETSTATE (ath, THSTATE_CPUWAIT, NULL);
    enqTh (&cpuWaitq, ath);
    return 1;
}


/* 
 * Wake up a blocked thread ATH by posting on cpuWaitq, and cause it
 * to return with value RV.
 * Cancels any timeouts and reacquires mutex etc.
 * Return true if we need to reschedule.
 */
static int
rtpx_resumeWithCleanups (rtpx_thread_t *ath, int rv)
{
    if (ath->th_itimer.it_type != IT_TYPE_DEAD) {
	/* clear timedwait timer */
	_rtpx_itimerstop (&ath->th_itimer);
	assert (ath->th_itimer.it_type == IT_TYPE_DEAD);
    }

    {
	pthread_mutex_t *mx = ath->th_waitMx;
	if (mx) {
	    ath->th_waitMx = NULL;
	    if (mx->mx_flags == 0) {
		/* fast lock */
		mx->mx_flags = _RTPX_MXF_FASTLOCK;
#ifndef NDEBUG
		mx->mx_owner = ath;
#endif
	    }
	    else {
		/* slow path */
		if (mx->mx_flags & (_RTPX_MXF_LOCKED | _RTPX_MXF_WAITERS)) {
		    rtpx_thread_t  *oth;
		    if (mx->mx_mxa.mxa_protocol == PTHREAD_PRIO_INHERIT
			&& (oth = mx->mx_owner)
			&& ath->th_prio > oth->th_prio) {
			/* boost prio of owning thread to our prio? */
			rtpx_setepri (oth, ath->th_prio);
		    }
		    /* put thread back into waitq for the mutex */
		    /* XXX this thread can't own this mutex, but could a more
		       complex deadlock have arisen since we released it? */
		    assert (ath->th_waitq == NULL);
		    RTPX_INCSTAT (nmutexwait);
		    RTPX_SETSTATE (ath, THSTATE_MUTEXWAIT, mx);
		    mx->mx_flags |= _RTPX_MXF_WAITERS;
		    enqTh (&mx->mx_waitq, ath);
		    return 0;
		}
		/* resumed thread gets mutex */
		mx_assign (mx, ath);
	    }
	}
    }

    return rtpx_resumeWithValue (ath, rv);
}


/* 
 * Wake up a blocked thread ATH by posting on cpuWaitq, and set
 * its errno to ERROR, and make it return with value -1.
 */
int
_rtpx_resumeWithError (rtpx_thread_t *ath, int error)
{
    ath->th_errno = error;
    return rtpx_resumeWithCleanups (ath, -1);
}




/* 
 * 64-bit clock handling.
 *
 * The clock holds a number of ticks since boot time; each tick is
 * of a resolution determined by the hardware, as defined by
 * _rtpx_tickspersec, which could be anything from 50Hz to 1GHz.
 *
 * The higher the tick frequency the more accurate scheduling will be, 
 * but note that we don't take an interrupt every tick, it's just the
 * resolution of the timer hardware.
 */

/* convert timespec to clock64 */
/* XXX rewrite so that clock64_t is {secs; ticks;} */
clock64_t
_rtpx_tsToClock64 (const struct timespec *ts)
{
    clock64_t       clk64;
#if _RTPX_CLOCK64_64BIT
    if (ts->tv_sec != 0)
	/* 32z32->64 multiply - fast on MIPS */
	clk64 = (clock64_t)_rtpx_tickspersec * ts->tv_sec;
    else
	clk64 = 0;
    if (ts->tv_nsec != 0) {
	/* 64-bit divide */
	clk64 += ((clock64_t) _rtpx_tickspersec * ts->tv_nsec) / NSPERSEC;
    }
#else
    clk64.c_sec = ts->tv_sec;
    if (ts->tv_nsec == 0)
	clk64.c_tick = 0;
    else
	/* 32x32->64 mult - OK */
	/* XXX 64-bit divide? How to avoid? */
	clk64.c_tick = ((unsigned long long)_rtpx_tickspersec * ts->tv_nsec)
	    / NSPERSEC;
#endif
    return clk64;
}


/* convert clock64 to timespec */
/* XXX rewrite so that clock64_t is {secs; ticks;} */
void
_rtpx_tsFromClock64 (struct timespec *ts, clock64_t ticks)
{
    register tick_t hz = _rtpx_tickspersec;
#if _RTPX_CLOCK64_64BIT
    if (ticks == 0 || hz == 0) {
	ts->tv_sec = ts->tv_nsec = 0;
    }
    else if (ticks == hz) {
	ts->tv_sec = 1;
	ts->tv_nsec = 0;
    }
    else {
	register tick_t sticks;
	if (ticks >= hz) {
	    ts->tv_sec = ticks / hz;
	    sticks = ticks % hz;
	}
	else {
	    ts->tv_sec = 0;
	    sticks = ticks;
	}
	if (sticks == 0)
	    ts->tv_nsec = 0;
	else if (sticks <= ULONG_MAX / NSPERSEC)
	    ts->tv_nsec = (sticks * NSPERSEC + hz / 2) / hz;
	else
	    ts->tv_nsec = (sticks * (clock64_t)NSPERSEC + hz / 2) / hz;
    }
#else
    if (hz == 0) {
	timespecclear (ts);
    }
    else {
	ts->tv_sec = ticks.c_sec;
	if (ticks.c_tick == 0)
	    ts->tv_nsec = 0;
	else if (ticks.c_tick <= ULONG_MAX / NSPERSEC)
	    /* good accuracy for small times (<= 4 ticks!) */
	    ts->tv_nsec = (ticks.c_tick * NSPERSEC + hz / 2) / hz;
	else
	    /* XXX accurate, but another 64-bit division XXX use nspertick */
	    ts->tv_nsec = (ticks.c_tick * (unsigned long long)NSPERSEC
			   + hz / 2) / hz;
    }
#endif
}


/* return the clock64 time at last clock tick */
clock64_t
_rtpx_clock ()
{
    return wallClock.clk_t64;
}


/* return the clock64 time now (by querying hardware for precision) */
clock64_t
_rtpx_clockNow ()
{
    clock64_t       value = wallClock.clk_t64;
    unsigned        inc = _sbd_gettickval () - wallClock.clk_when;
    clock64inc (&value, inc);
    return value;
}


/* return the clock64 time now relative to our start time */
clock64_t
_rtpx_relclock ()
{
    clock64_t	     value = _rtpx_clockNow ();
    clock64sub (&value, &bootClock, &value);
    return value;
}



/* 
 * INTERVAL TIMER 
 */

/* start an absolute timer */
void
_rtpx_itimerstartabs (rtpx_itimer_t *it, int type, clock64_t when)
{
    /* when = time when timer should fire (in ticks since boot) */
    it->it_value = when;
    it->it_type = type;

    {
	rtpx_itimer_t  *nit = itCurrent;

	if (clock64cmp (&when, &nit->it_value, <)) {
	    /* XXX cpu_timer needs reloading */
	    itCurrent = it;
	    it->it_next = nit;
	    it->it_prev = NULL;
	    nit->it_prev = it;
	    /* force reload of new timer now */
	    /* XXX must call leaveWithDispatch after this */
	    _cpuTab[0].timerReload = 1;
	}
	else {
	    rtpx_itimer_t  *pit = nit;

	    while (1) {
		if ((nit = nit->it_next) == NULL) {
		    break;
		}
		if (clock64cmp (&when, &nit->it_value, <)) {
		    nit->it_prev = it;
		    break;
		}
		pit = nit;
	    }
	    pit->it_next = it;
	    it->it_next = nit;
	    it->it_prev = pit;
	}
    }
}


/* start a relative timer */
void
_rtpx_itimerstart (rtpx_itimer_t *it, int type, clock64_t when)
{
    clock64_t now = _rtpx_clockNow ();
    clock64add (&when, &now, &now);
    _rtpx_itimerstartabs (it, type, now);
}


/* stop a timer (remove from queue) */
void
_rtpx_itimerstop (rtpx_itimer_t *it)
{
    rtpx_itimer_t  *pit = it->it_prev;
    rtpx_itimer_t  *nit = it->it_next;

    assert (it->it_type != IT_TYPE_DEAD);
    assert (pit != it);
    assert (nit != it);

    if (pit == NULL) {
	assert (it == itCurrent);
	assert (nit);
	itCurrent = nit;
	nit->it_prev = NULL;
	/* Note don't bother reloading longer value to sbd_ticktimer, we'll
	   reset it on the next tick. */
    }
    else {
	assert (it != itCurrent);
	pit->it_next = nit;
	if (nit)
	    nit->it_prev = pit;
    }

    /* store the number of clock ticks remaining */
    {
	clock64_t       now = _rtpx_clockNow ();
	if (clock64cmp(&now, &it->it_value, <))
	    clock64sub (&it->it_value, &now, &it->it_value);
	else
	    clock64clear (&it->it_value);
    }
    it->it_next = it->it_prev = (void *) 0xbad;
    it->it_type = IT_TYPE_DEAD;
}


/* process pending timers (called in scheduling loop) */
static unsigned
rtpx_itproc (unsigned now)
{
    clock64_t       aTime = wallClock.clk_t64;
    rtpx_itimer_t  *it = itCurrent;

    {
	unsigned        ntick = (now - wallClock.clk_when);

	/* Would normally <= tickspersec/2, but for breakpoints and
	   potentially (very) long interrupt handlers may cause
	   failure. so assume that it was a breakpoint (or broken
	   interrupt handler) and limit to 0.5 seconds. */
	if (ntick > _rtpx_tickspersec)
	    ntick = _rtpx_tickspersec / 2;

	clock64inc (&aTime, ntick);

	wallClock.clk_when = now;
	wallClock.clk_t64 = aTime;
    }

    while (clock64cmp (&aTime, &it->it_value, >=)) {
	rtpx_itimer_t  *nit = it->it_next;
	if (clock64isset (&it->it_interval)) {
	    clock64_t       when;
	    clock64add (&it->it_value, &it->it_interval, &when);
	    if (nit && clock64cmp (&when, &nit->it_value, >=)) {
		/* requeue expired timer */
		itCurrent = nit;
		nit->it_prev = NULL;
		{
		    rtpx_itimer_t  *pit = nit;
		    while (1) {
			if ((nit = nit->it_next) == NULL) {
			    break;
			}
			if (clock64cmp (&when, &nit->it_value, <)) {
			    nit->it_prev = it;
			    break;
			}
			pit = nit;
		    }
		    pit->it_next = it;
		    it->it_next = nit;
		    it->it_prev = pit;
		}
	    }
	    it->it_value = when;
	}
	else {
	    /* drop expired timer */
	    itCurrent = nit;
	    if (nit)
		nit->it_prev = NULL;
	    it->it_next = it->it_prev = (void *) 0xdead;
	}

	/* should still have at least base timer */
	assert (itCurrent);


	/* fire the expired timer */
	switch (it->it_type) {
	case IT_TYPE_BASE:
	    break;
	case IT_TYPE_SLEEP:
	    it->it_type = IT_TYPE_DEAD;
	    clock64clear (&it->it_value);
	    {
		rtpx_thread_t  *ath = it->it_thread;
		if (ath->th_state == THSTATE_TIMEWAIT) {
		    EVLOGTIMER (ath, IT_TYPE_SLEEP);
		    rtpx_resumeWithValue (ath, 0);
		}
	    }
	    break;
	case IT_TYPE_TIMEOUT:
	    it->it_type = IT_TYPE_DEAD;
	    {
		rtpx_thread_t  *ath = it->it_thread;
		EVLOGTIMER (ath, IT_TYPE_TIMEOUT);
		if (ath->th_state == THSTATE_CONDWAIT ||
		    ath->th_state == THSTATE_DOORBELLWAIT) {
		    deqTh (ath->th_waitq, ath);
		    rtpx_resumeWithCleanups (ath, ETIME);
		}
		else if (ath->th_state == THSTATE_SIGWAIT) {
		    rtpx_resumeWithValue (ath, EAGAIN);
		}
	    }
	    break;
	case IT_TYPE_TIMER:
	    if (! clock64isset (&it->it_interval))
		it->it_type = IT_TYPE_DEAD;
	    {
		rtpx_timer_t   *ti = (rtpx_timer_t *) it;
		switch (ti->ti_sigev.sigev_notify) {
		case SIGEV_NONE:
		    break;
		case SIGEV_SIGNAL:
		case SIGEV_THREAD:
		    ti->ti_overrun++;	/* XXX should be reset when sig
					   delivered */
		    EVLOGTIMER (rtpx_self (), IT_TYPE_TIMER);
		    (void) _rtpx_sigqueue (ti->ti_pr, SI_TIMER, &ti->ti_sigev);
		    break;
		default:
		    abort ();
		}
	    }
	    break;
	default:
	    abort ();
	}

	it = itCurrent;
    }

    {
	clock64_t       delta;
	clock64sub (&it->it_value, &aTime, &delta);
	assert (clock64subsec (&delta));
	return now + clock64ticks (&delta);
    }
}




/* search for thread id */
rtpx_thread_t  *
_rtpx_findthread (int id)
{
    rtpx_thread_t  *th;

    for (th = rtpx_threads; th < &rtpx_threads[_RTPX_THREAD_MAX]; th++)
	if (th->th_state != THSTATE_DEAD && th->th_id == id)
	    return th;
    return NULL;
}


/* return next non-dead thread after TH */
rtpx_thread_t  *
_rtpx_nextthread (rtpx_thread_t *th)
{
    do {
	if (++th >= &rtpx_threads[_RTPX_THREAD_MAX])
	    return NULL;
    } while (th->th_state == THSTATE_DEAD);
    return th;
}


/* return first non-dead thread */
rtpx_thread_t  *
_rtpx_firstthread (void)
{
    return _rtpx_nextthread (rtpx_threads - 1);
}


/* return number of bytes of stack used by thread ATH */
int
_rtpx_stackused (rtpx_thread_t *ath)
{
    unsigned char  *sp,
                   *limit;

    if (!(sp = ath->th_stack))
	return 0;

    limit = sp + ath->th_tha.tha_stacksize;
    for (sp += sizeof (long) * 4; sp < limit; sp++)
	if (*sp != STACKCOLOUR)
	    break;
    return limit - sp;
}



/* 
 * SLEEP ETC.
 */

int
nanosleep (const struct timespec *req, struct timespec *res)
{
    rtpx_thread_t  *cth;
    int             rv;

    if (!timespecok (req))
	return pthread_error (EINVAL);

    cth = rtpx_enter ();
    _rtpx_itimerstart (&cth->th_itimer, IT_TYPE_SLEEP,
		       _rtpx_tsToClock64 (req));
    rv = rtpx_leaveWhenResumed (cth, THSTATE_TIMEWAIT, NULL, 0);

    if (res) {
	if (rv == EINTR && clock64isset (&cth->th_itimer.it_value))
	    _rtpx_tsFromClock64 (res, cth->th_itimer.it_value);
	else
	    timespecclear (res);
    }

    return rv ? pthread_error (rv) : 0;
}


#if 0
/* Some Unices have this */
unsigned int
usleep (unsigned int usecs)
{
    struct timespec ts;
    int             rv;

    ts.tv_sec = usecs / 1000000;
    ts.tv_nsec = (usecs % 1000000) * 1000;
    rv = nanosleep (&ts, &ts);
    if (rv < 0 && errno == EINTR)
	return (ts.tv_sec * 1000000) + ((ts.tv_nsec + 500) / 1000);
    return rv;
}
#endif


unsigned int
sleep (unsigned int secs)
{
    struct timespec ts;
    int             rv;

    ts.tv_sec = secs;
    ts.tv_nsec = 0;
    rv = nanosleep (&ts, &ts);
    if (rv < 0 && errno == EINTR)
	return ts.tv_sec + (ts.tv_nsec >= NSPERSEC / 2);
    return 0;
}


int
sched_yield (void)
{
    rtpx_thread_t  *cth = rtpx_enter ();

    if (cth->th_next != NULL) {
	assert (cth->th_prio >= cth->th_next->th_prio);
	if (cth->th_prio == cth->th_next->th_prio) {
	    reqTh (&cpuWaitq, cth, cth->th_prio);
	    return rtpx_leaveWithValueAndDispatch (cth, 0);
	}
    }
    return rtpx_leaveWithValue (cth, 0);
}


/* 
 * MUTEXES
 */

/* 
 * Assign ownership of a mutex (slow and careful)
 */
static void
mx_assign (pthread_mutex_t *mx, rtpx_thread_t *ath)
{
    /* ath must be in transition, or the highest priority run thread */
    assert (ath->th_waitq == NULL ||
	    (ath->th_waitq == &cpuWaitq && cpuWaitq.wq_first == ath));

    mx->mx_flags |= _RTPX_MXF_LOCKED;
    mx->mx_owner = ath;

    if (mx->mx_mxa.mxa_type == PTHREAD_MUTEX_RECURSIVE_NP)
	mx->mx_count = 1;

    EVLOGMUTEX (ath, 1, mx);

    if (mx->mx_mxa.mxa_protocol != PTHREAD_PRIO_NONE) {
	/* calculate new priority */
	int             mxprio = mx->mx_mxa.mxa_prioceiling;

	/* add to list of owned mutexes */
	mx->mx_olink = ath->th_mxlist;
	ath->th_mxlist = mx;

	if (mx->mx_mxa.mxa_protocol == PTHREAD_PRIO_INHERIT) {
	    /* highest priority waiting thread */
	    rtpx_thread_t  *wth = mx->mx_waitq.wq_first;
	    if (wth != NULL)
		mxprio = wth->th_prio;
	}

	if (mxprio > ath->th_prio) {
	    /* head of run list or not waiting, 
	       so don't call rtpx_setepti() */
	    EVLOGPRIO (ath, mxprio);
	    ath->th_prio = mxprio;
	}
    }
    else if (mx->mx_mxa.mxa_pshared == PTHREAD_PROCESS_SHARED) {
	/* add to list of owned mutexes */
	mx->mx_olink = ath->th_mxlist;
	ath->th_mxlist = mx;
    }
}


/* 
 * Release a mutex, resuming any thread which is waiting for it.
 */
static void
mx_unassign (pthread_mutex_t *mx, rtpx_thread_t *ath)
{
    assert (mx->mx_owner == ath || ! (mx->mx_flags & _RTPX_MXF_COMPLEX));

    EVLOGMUTEX (ath, 0, mx);
    mx->mx_flags &= ~_RTPX_MXF_LOCKED;
    mx->mx_owner = NULL;

    if (mx->mx_mxa.mxa_type == PTHREAD_MUTEX_RECURSIVE_NP)
	mx->mx_count = 0;

    if (ath->th_mxlist == NULL) {
	/* no complex priority recalculation - do a fast reassign */
	rtpx_thread_t  *oth = popqTh (&mx->mx_waitq);
	if (oth != NULL) {
	    if (!mx->mx_waitq.wq_first)
		mx->mx_flags &= ~_RTPX_MXF_WAITERS;
	    if (mx->mx_flags == 0) {
		/* fast mutex with no waiters - do a VERY fast reassign */
		mx->mx_flags = _RTPX_MXF_FASTLOCK;
#ifndef NDEBUG
		mx->mx_owner = oth;
#endif
	    }
	    else {
		mx->mx_flags |= _RTPX_MXF_LOCKED;
		mx->mx_owner = oth;
		if (mx->mx_mxa.mxa_type == PTHREAD_MUTEX_RECURSIVE_NP)
		    mx->mx_count = 1;
		EVLOGMUTEX (oth, 1, mx);
	    }
	    RTPX_INCSTAT (nmutexwake);
	    rtpx_resumeWithValue (oth, oth->th_wval);
	}
	return;
    }

    /* remove mutex from list of owned muticies */
    if (mx->mx_mxa.mxa_protocol != PTHREAD_PRIO_NONE
	|| mx->mx_mxa.mxa_pshared == PTHREAD_PROCESS_SHARED) {
	pthread_mutex_t **omx = &ath->th_mxlist;
	while (*omx != mx)
	    omx = &(*omx)->mx_olink;
	*omx = mx->mx_olink;
	mx->mx_olink = NULL;
    }

    /* recalculate priority */
    {
	unsigned char   nprio = ath->th_tha.tha_schedparam.sched_priority;
	pthread_mutex_t *omx;

	for (omx = ath->th_mxlist; omx; omx = omx->mx_olink) {
	    unsigned char   mxprio;
	    switch (omx->mx_mxa.mxa_protocol) {
	    case PTHREAD_PRIO_INHERIT:
		if (omx->mx_waitq.wq_first == NULL)
		    continue;
		mxprio = omx->mx_waitq.wq_first->th_prio;
		break;
	    case PTHREAD_PRIO_PROTECT:
		mxprio = omx->mx_mxa.mxa_prioceiling;
		break;
	    case PTHREAD_PRIO_NONE:
		continue;
	    default:
		abort ();
	    }
	    if (mxprio > nprio)
		nprio = mxprio;
	}

	if (ath->th_prio != nprio) {
	    assert (ath->th_prio > nprio);
	    rtpx_setepri (ath, nprio);
	}
    }


    {
	rtpx_thread_t  *oth = popqTh (&mx->mx_waitq);
	if (oth != NULL) {
	    if (!mx->mx_waitq.wq_first)
		mx->mx_flags &= ~_RTPX_MXF_WAITERS;
	    mx_assign (mx, oth);
	    RTPX_INCSTAT (nmutexwake);
	    (void) rtpx_resumeWithValue (oth, oth->th_wval);
	}
    }
}


/* try to lock a thread (internal version) */
static int
mx_trylock (pthread_mutex_t *mx, rtpx_thread_t *cth, unsigned int oflags)
{
    if (mx->mx_flags & _RTPX_MXF_FIRSTTIME) {
	/* first lock attempt */

	/* set default mutex type - may be changed by user for debugging */
	if (mx->mx_mxa.mxa_type == PTHREAD_MUTEX_DEFAULT_NP)
	    mx->mx_mxa.mxa_type = _rtpx_mutex_default_type;

	/* decide whether this is a simple mutex, for which we can use
	   fast atomic operations, or a complex one which is handled
	   slowly. */
	if (mx->mx_mxa.mxa_type == PTHREAD_MUTEX_FAST_NP
	    && mx->mx_mxa.mxa_protocol == PTHREAD_PRIO_NONE
	    && mx->mx_mxa.mxa_pshared == PTHREAD_PROCESS_PRIVATE)
	    /* simple and fast */
	    mx->mx_flags = 0;
	else {
	    /* complex and slow */
	    mx->mx_flags = _RTPX_MXF_COMPLEX;
	    if (mx->mx_mxa.mxa_type == PTHREAD_MUTEX_RECURSIVE_NP)
		mx->mx_count = 0;
	}
    }

    if (mx->mx_flags == 0) {
	/* the original fastlock owner has since unlocked the mutex, so we
	   can now take it over. */
	mx->mx_flags = _RTPX_MXF_FASTLOCK;
#ifndef NDEBUG
	mx->mx_owner = cth;
#endif
	return 0;
    }

    if (!(oflags & _RTPX_MXF_LOCKED)
	|| !(mx->mx_flags & _RTPX_MXF_LOCKED)) {
	/* this is "complex" or slow mutex but 1) no-one had the lock before, 
	   or 2) it has since been unlocked let's take it. */
	mx_assign (mx, cth);
	return 0;
    }


    if (mx->mx_mxa.mxa_type == PTHREAD_MUTEX_RECURSIVE_NP
	&& mx->mx_owner == cth) {
	mx->mx_count++;
	return 0;
    }

    return EBUSY;
}


/* wait for a mutex to be unlocked */
static int
mx_wait (pthread_mutex_t *mx, rtpx_thread_t *cth, int rv)
{
    /* mutex is currently locked by mx_owner (NULL for fast mutexes) */
    rtpx_thread_t  *mth = mx->mx_owner;

    /* check for deadlock */
    if (mx->mx_mxa.mxa_type != PTHREAD_MUTEX_FAST_NP
	&& (mx->mx_flags & _RTPX_MXF_SLOWLOCK)) {
	rtpx_thread_t  *oth = mth;
	pthread_mutex_t *omx = mx;
	
	while (1) {

	    if (oth == cth) {
		int             type =
		    _rtpx_deadlock_fatal ? RTPXLOG_PANIC : RTPXLOG_WARNING;
		RTPX_INCSTAT (nmutexdead);
		if (mx->mx_mxa.mxa_name && omx->mx_mxa.mxa_name
		    && cth->th_tha.tha_name && mth->th_tha.tha_name)
		    _rtpx_log (type,
			       "Deadlock: mutex %s, thread %s, owner %s, other mutex %s",
			       mx->mx_mxa.mxa_name, cth->th_tha.tha_name,
			       mth->th_tha.tha_name, oth->th_tha.tha_name);
		else
		    _rtpx_log (type,
			       "Deadlock: mutex %p, thread %p, owner %p, other mutex %p",
			       mx, cth, mth, oth);
		return rtpx_leaveWithValue (cth, EDEADLK);
	    }

	    assert (oth);
	    if (oth->th_state == THSTATE_MUTEXWAIT) {
		/* find the mutex which the owner is waiting for */
		omx = (pthread_mutex_t *) (oth->th_waitq);
	    }
	    else if ((oth->th_state == THSTATE_CONDWAIT
		     || oth->th_state == THSTATE_DOORBELLWAIT)
		     && oth->th_waitMx) {
		/* Thread oth must be runnable in order to unlock the mutex
		   mx it cannot run if we hold the mutex which oth requires
		   to get out of cond_wait. */
		omx = oth->th_waitMx;
	    }
	    else
		break;

	    assert (omx);
	    if (!(omx->mx_flags & _RTPX_MXF_SLOWLOCK))
		break;		/* can't check fast locked mutexes */

	    assert (omx->mx_owner && omx->mx_owner != oth);
	    oth = omx->mx_owner;
	}
    }

    if (mx->mx_mxa.mxa_protocol == PTHREAD_PRIO_INHERIT) {
	/* boost prio of owning thread to our prio? */
	if (mth && cth->th_prio > mth->th_prio)
	    rtpx_setepri (mth, cth->th_prio);
    }
    else if (mx->mx_mxa.mxa_protocol == PTHREAD_PRIO_PROTECT) {
	/* if this thread has a higher prio than the mutexes ceiling, then
	   that's an error */
#if 1
	if (mx->mx_mxa.mxa_prioceiling <
	    cth->th_tha.tha_schedparam.sched_priority) return EINVAL;
#else
	if (mx->mx_mxa.mxa_prioceiling < cth->th_prio)
	    return EINVAL;
#endif
    }

    /* wait for mutex to be unlocked */
    RTPX_INCSTAT (nmutexwait);
    mx->mx_flags |= _RTPX_MXF_WAITERS;
    return rtpx_leaveWhenResumed (cth, THSTATE_MUTEXWAIT, &mx->mx_waitq,
				  rv);
}



int
pthread_mutex_init (pthread_mutex_t *mx, const pthread_mutexattr_t *mxa)
{
    if (mxa == NULL)
	mxa = &mxaDefault;

    mx->mx_waitq.wq_first = NULL;
    mx->mx_flags = ~0;
    mx->mx_mxa = *mxa;
    mx->mx_olink = NULL;
    mx->mx_owner = NULL;

    /* for backwards compatibility with old libraries we don't write mx_count 
       unless we know that this is a recursive mutex */
    /* mx->mx_count = 0; */

    return 0;
}


int
pthread_mutex_setname_np (pthread_mutex_t *mx, const char *name)
{
    mx->mx_mxa.mxa_name = name;
    return 0;
}


int
pthread_mutex_getname_np (const pthread_mutex_t *mx, const char **namep)
{
    *namep = mx->mx_mxa.mxa_name;
    return 0;
}


int
pthread_mutex_getprioceiling (const pthread_mutex_t *mx, int *ceiling)
{
    *ceiling = mx->mx_mxa.mxa_prioceiling;
    return 0;
}


int
pthread_mutex_setprioceiling (pthread_mutex_t *mx, int newceiling,
			      int *oldceiling)
{
    int             rv;
    if (newceiling < _RTPX_PRIO_MIN || newceiling > _RTPX_PRIO_MAX)
	return EINVAL;
    rv = pthread_mutex_lock (mx);
    if (rv != 0)
	return rv;
    if (oldceiling)
	*oldceiling = mx->mx_mxa.mxa_prioceiling;
    mx->mx_mxa.mxa_prioceiling = newceiling;
    return pthread_mutex_unlock (mx);
}


int
pthread_mutex_destroy (pthread_mutex_t *mx)
{
    /* XXX need to add a magic number checkword */
    if (mx->mx_flags & _RTPX_MXF_LOCKED)
	return EBUSY;
    mx->mx_mxa.mxa_type = ~0;
    return 0;
}


int
pthread_mutex_lock (pthread_mutex_t *mx)
{
    rtpx_thread_t  *cth;
    unsigned int    oflags;

    /* fast path for simple mutexes */
    oflags = _cpu_atomic_bis (&mx->mx_flags, _RTPX_MXF_FASTLOCK);
    if (oflags == 0) {
	/* we got in first, and it's not a complex mutex, so we're done */
#ifndef NDEBUG
	mx->mx_owner = rtpx_self ();
#endif
	return 0;
    }

    /* Safe to continue if thread mechanism not yet started */
    if (nThread == 0)
	return 0;

    /* XXX there is a window here where another thread could try to lock a
       complex mutex and find that we've fast-locked it so it has to block;
       but we've not set the owner field yet, so the deadlock test below may
       not detect a deadlock. */

    cth = rtpx_enter ();
    if (mx_trylock (mx, cth, oflags) == 0)
	return rtpx_leaveWithValue (cth, 0);

    return mx_wait (mx, cth, 0);
}


int
pthread_mutex_trylock (pthread_mutex_t *mx)
{
    rtpx_thread_t  *cth;
    unsigned int    oflags;

    /* fast path for simple mutexes */
    oflags = _cpu_atomic_bis (&mx->mx_flags, _RTPX_MXF_FASTLOCK);
    if (oflags == 0) {
#ifndef NDEBUG
	mx->mx_owner = rtpx_self ();
#endif
	return 0;
    }

    /* Safe to continue if thread mechanism not yet started */
    if (nThread == 0)
	return 0;

    cth = rtpx_enter ();
    return rtpx_leaveWithValue (cth, mx_trylock (mx, cth, oflags));
}


int
pthread_mutex_unlock (pthread_mutex_t *mx)
{
    rtpx_thread_t  *cth;
    unsigned int    oflags;

    /* fast path for simple mutexes */
    oflags = _cpu_atomic_bic (&mx->mx_flags, _RTPX_MXF_FASTLOCK);
    if (oflags == _RTPX_MXF_FASTLOCK) {
#if !defined(NDEBUG) && 0
	/* not safe, we don't own it anymore */
	mx->mx_owner = NULL;
#endif
	return 0;
    }

    /* Safe to continue if thread mechanism not yet started */
    if (nThread == 0)
	return 0;

    cth = rtpx_enter ();

    if ((mx->mx_flags & _RTPX_MXF_COMPLEX)
	&& mx->mx_owner != cth)
	return rtpx_leaveWithValue (cth, EINVAL);

    if (mx->mx_mxa.mxa_type == PTHREAD_MUTEX_RECURSIVE_NP
	&& --mx->mx_count > 0)
	return rtpx_leaveWithValue (cth, 0);

    mx_unassign (mx, cth);

    return rtpx_leaveWithValueAndDispatch (cth, 0);
}



/* 
 * CONDITION VARIABLES
 */

int
pthread_cond_init (pthread_cond_t *cd, const pthread_condattr_t *cda)
{
    if (cda == 0)
	cda = &cdaDefault;

    cd->cd_waitq.wq_first = NULL;
    cd->cd_cda = *cda;
    return 0;
}


int
pthread_cond_setname_np (pthread_cond_t *cd, const char *name)
{
    cd->cd_cda.cda_name = name;
    return 0;
}


int
pthread_cond_getname_np (const pthread_cond_t *cd, const char **namep)
{
    *namep = cd->cd_cda.cda_name;
    return 0;
}


int
pthread_cond_destroy (pthread_cond_t *cd)
{
    /* XXX need to add a magic number checkword */
    cd->cd_cda.cda_pshared = 0xff;
    return 0;
}


int
pthread_cond_timedwait (pthread_cond_t *cd, pthread_mutex_t *mx,
			const struct timespec *abstime)
{
    rtpx_thread_t  *cth = rtpx_enter ();

    if (cd->cd_cda.cda_interruptible) {
	if ((cth->th_sigpend | cth->th_proc->pr_sigpend) 
	    & ~cth->th_sigdefermask) 
	    return rtpx_leaveWithValue (cth, EINTR);
	cth->th_flags |= RTPX_THF_CONDINTRUPT;
    }
    else {
	cth->th_flags &= ~RTPX_THF_CONDINTRUPT;
    }
    return rtpx_leaveWhenResumedWithMx (cth, THSTATE_CONDWAIT,
					&cd->cd_waitq, mx, abstime);
}


int
pthread_cond_wait (pthread_cond_t *cd, pthread_mutex_t *mx)
{
    return pthread_cond_timedwait (cd, mx, NULL);
}


int
pthread_cond_signal (pthread_cond_t *cd)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    rtpx_thread_t  *ath = popqTh (&cd->cd_waitq);

    if (ath == NULL)
	return rtpx_leaveWithValue (cth, 0);

    rtpx_resumeWithCleanups (ath, 0);

    return rtpx_leaveWithValueAndDispatch (cth, 0);
}


int
pthread_cond_broadcast (pthread_cond_t *cd)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    rtpx_thread_t  *ath = popqTh (&cd->cd_waitq);

    if (ath == NULL)
	return rtpx_leaveWithValue (cth, 0);

    do {
	rtpx_resumeWithCleanups (ath, 0);
    } while ((ath = popqTh (&cd->cd_waitq)) != NULL);

    return rtpx_leaveWithValueAndDispatch (cth, 0);
}



/* 
 * DOORBELLS (DEVICE INTERRUPTS)
 */

int
pthread_doorbell_init (pthread_doorbell_t *db,
		       const pthread_doorbellattr_t *dba)
{
    if (!dba || dba->dba_evn == 0 || dba->dba_evn >= _SBD_INTR_MAX
	|| dbTable[dba->dba_evn] != NULL)
	return EBUSY;

    dbTable[dba->dba_evn] = db;
    db->db_waitq.wq_first = NULL;
    db->db_dba = *dba;
    return 0;
}


int
pthread_doorbell_destroy (pthread_doorbell_t *db)
{
    if (db->db_waitq.wq_first)
	return EBUSY;
    if (db->db_dba.dba_evn > 0 && db->db_dba.dba_evn < _SBD_INTR_MAX)
	dbTable[db->db_dba.dba_evn] = NULL;
    db->db_dba.dba_evn = 0;
    return 0;
}


int
pthread_doorbell_trywait (pthread_doorbell_t *db)
{
    int             evn = db->db_dba.dba_evn;
    rtpx_thread_t  *cth;

    if (evn <= 0 || evn >= _SBD_INTR_MAX)
	return EINVAL;

    cth = rtpx_enter ();

    if (!db->db_waitq.wq_first && _sbd_intrpoll (evn))
	return rtpx_leaveWithValue (cth, 0);

    return rtpx_leaveWithValue (cth, EAGAIN);
}


int
pthread_doorbell_timedwait (pthread_doorbell_t *db,
			    pthread_mutex_t *mx, struct timespec *abstime)
{
    int             evn = db->db_dba.dba_evn;
    rtpx_thread_t  *cth;

    if (evn <= 0 || evn >= _SBD_INTR_MAX)
	return EINVAL;

    cth = rtpx_enter ();

    /* check for interrupt active and no-other thread waiting */
    if (!db->db_waitq.wq_first && _sbd_intrpoll (evn))
	return rtpx_leaveWithValue (cth, 0);

    /* enable the interrupt */
    _sbd_intrenable (evn);

    return rtpx_leaveWhenResumedWithMx (cth, THSTATE_DOORBELLWAIT,
					&db->db_waitq, mx, abstime);
}


int
pthread_doorbell_wait (pthread_doorbell_t *db, pthread_mutex_t *mx)
{
    return pthread_doorbell_timedwait (db, mx, NULL);
}


#if 0
/* NOTE only for soft doorbells - needs to be fast for low-level interrupts */
int
pthread_doorbell_post (pthread_doorbell_t *db)
{
    rtpx_thread_t  *cth = rtpx_enter ();

    if (db->db_evn != 0)
	return rtpx_leaveWithValue (cth, EINVAL);

    if (db->db_value >= 0) {
	db->db_value++;
	return rtpx_leaveWithValue (cth, 0);
    }

    /* wake one thread from queue to mutex/cpuWaitq * like cond_signal */
    {
	rtpx_thread_t  *ath = popqTh (&db->db_waitq);

	if (ath == NULL)
	    return rtpx_leaveWithValue (cth, 0);

	rtpx_resumeWithCleanups (ath, 0);

	return rtpx_leaveWithValueAndDispatch (cth, 0);
    }
}
#endif



pthread_t
pthread_self (void)
{
    return rtpx_self ();
}


void
pthread_exit (void *v)
{
    rtpx_thread_t  *cth = rtpx_self ();
    unsigned int oflags;
    sigset_t sm;

    /* prevent recursion */
    oflags = _cpu_atomic_bis (&cth->th_flags, RTPX_THF_CANCELLING
			      | RTPX_THF_EXITING);

    sigfillset (&sm);

    if (! (oflags & RTPX_THF_EXITING)) {
	/* mask all maskable signals */
	pthread_sigmask (SIG_SETMASK, &sm, NULL);

	/* XXX if an unmaskable signal happens now, the 
	   remaining cleanups will be skipped! */

	/* run the thread-specific cleanup functions at user level */
	{
	    struct _pthread_handler_rec *rec;
	    
	    while ((rec = cth->th_cleanups) != NULL) {
		_pthread_cleanup_pop (rec);
		(*rec->rtn) (rec->arg);
	    }
	}

#if defined(_RTPX_FLOCKFILE)
	/* cleanup flocks */
	_rtpx_flockfilecleanup ();
#endif

#if defined(_RTPX_TSD)
	/* cleanup thread-specific data */
	/* XXX should this be delayed until a joinable thread is joined! */
	_rtpx_tsdcleanup ();
#endif
    }

    /* really block all signals, including KILL! */
    cth->th_sigmask = sm;

    pthread_mutex_lock (&thMx);
    assert (nThread > 0);
    nThread--;

    /* detach from process */
    if (cth->th_tha.tha_detachstate != PTHREAD_CREATE_DAEMON_NP) {
	if (!_rtpx_leaveproc ()) {
	    /* last thread in last process has expired,
	       terminate and return to monitor. */
	    pthread_mutex_unlock (&thMx);
	    _cpu_exit (0);
	}
    }

    /* wakeup any threads trying to join with us */
    if (cth->th_tha.tha_detachstate == PTHREAD_CREATE_JOINABLE)
	pthread_cond_broadcast (&cth->th_joinCd);

    (void) rtpx_enter ();

    /* Now committed to thread destruction, no more user-level
       code may be run. */

    deqTh (&cpuWaitq, cth);
    
    _cpu_thdestroy (cth);
    
    if (cth->th_tha.tha_detachstate == PTHREAD_CREATE_JOINABLE) {
	cth->th_zval = v;
	RTPX_SETSTATE (cth, THSTATE_ZOMBIE, NULL);
    }
    else {
	RTPX_SETSTATE (cth, THSTATE_DEAD, NULL);
	/* XXX we don't free the stack here, it will be freed when/if the 
	   slot in the thread table is reallocated for a new thread. */
	cth->th_next = freeThreadList;
	freeThreadList = cth;
    }

    mx_unassign (&thMx, cth);
    
    /* leave, never to return... */
    (void) rtpx_leaveWithValueAndDispatch (cth, 0);
    abort ();
}



int
pthread_create (pthread_t *rv, const pthread_attr_t *tha,
		void *(*fun) (void *), void *arg)
{
    rtpx_thread_t  *cth = pthread_self ();
    rtpx_thread_t  *ath = NULL;
    unsigned char  *sp = NULL;
    const char	   *cdname;

    if (tha == 0)
	tha = &thaDefault;
    else {
	/* check tha for sanity */
	assert (VALID_STACKSIZE (tha->tha_stacksize));
	assert (VALID_INHERITSCHED (tha->tha_inheritsched));
	assert (VALID_SCHEDPOLICY (tha->tha_schedpolicy));
	assert (VALID_PRIORITY (tha->tha_schedparam.sched_priority));
    }

    if ((sp = tha->tha_stackaddr) == NULL) {
	if ((sp = malloc (tha->tha_stacksize)) == NULL)
	    return ENOMEM;
    }
    else {
	sp = tha->tha_stackaddr;
    }

    /* Initialise stack and insert trivial guard region */
    {
	long           *red = (long *) sp;
	memset (sp, STACKCOLOUR, tha->tha_stacksize);
	red[0] = red[1] = red[2] = red[3] = REDSCARE;
    }

    pthread_mutex_lock (&thMx);

    ath = freeThreadList;
    if (ath == NULL) {
	pthread_mutex_unlock (&thMx);
	if (tha->tha_stackaddr == NULL)
	    free (sp);
	return EAGAIN;
    }
    freeThreadList = ath->th_next;

    /* allocate a unique thread id */
    do {
	threadID = (threadID + 1) % 32768;
    } while (threadID == 0 || _rtpx_findthread (threadID));
    ath->th_id = threadID;

    ++nThread;
    RTPX_INCSTAT (ncreate);

    /* free old thread stack if this is a recycled thread */
    /* XXXX could recycle if stack size matches */
    if (ath->th_stack && ath->th_tha.tha_stackaddr == NULL)
	free (ath->th_stack);

    /* point to new stack */
    ath->th_stack = sp;

    /* initialise thread attributes */
    ath->th_tha = *tha;
    if (ath->th_tha.tha_inheritsched == PTHREAD_INHERIT_SCHED) {
	ath->th_tha.tha_schedpolicy = cth->th_tha.tha_schedpolicy;
	ath->th_tha.tha_schedparam = cth->th_tha.tha_schedparam;
    }
    
    /* priority and round robin scheduling params */
    ath->th_prio = ath->th_tha.tha_schedparam.sched_priority;
    ath->th_ticksToRr = _rtpx_rrclock;
    
    /* default cancel state */
    ath->th_flags = RTPX_THF_CANCEL_DEFERRED;
    
    /* not on any queue yet */
    ath->th_mxlist = NULL;
    ath->th_waitq = NULL;
    ath->th_waitMx = NULL;
    
    /* no errno initially */
    ath->th_errno = 0;
    
    /* initial signal state */
    ath->th_sigmask = cth->th_sigmask;	/* sigmask inherited */
    ath->th_sigpend = ath->th_sigwait = ath->th_sigdefermask = 0;
    
    /* no cleanups pending */
    ath->th_cleanups = NULL;
    
#ifdef _RTPX_TSD
    /* enpty thread-specific data */
    ath->th_tsd = NULL;
    ath->th_tsdcount = 0;
#endif

    /* invent a name for this thread, if we haven't been given one */
    if (!(cdname = ath->th_tha.tha_name)) {
        const char     *s = "join-th";
	char           *d = ath->th_name;
	unsigned int    id = ath->th_id;
	char            nbuf[8], *n = nbuf;

	while (*s)
	  *d++ = *s++;
	do {
	    *n++ = (id % 10) + '0';
	    id /= 10;
	} while (id != 0);
	do {
	    *d++ = *--n;
	} while (n != nbuf);
	*d++ = '\0';
	ath->th_tha.tha_name = &ath->th_name[5];
	cdname = ath->th_name;
    }

    /* init pthread_join cond variable */
    pthread_cond_init (&ath->th_joinCd, NULL);
    pthread_cond_setname_np (&ath->th_joinCd, cdname);

    /* set owning process */
    ath->th_proc = cth->th_proc;
    if (ath->th_tha.tha_detachstate != PTHREAD_CREATE_DAEMON_NP)
	_rtpx_joinproc (ath);

    /* now a valid thread - let others see it */
    pthread_mutex_unlock (&thMx);

    /* return thread id */
    if (rv)
	*rv = (pthread_t) ath;

    /* initialise stack to call function */
    _cpu_thinit (ath, fun, arg);

    /* add new thread to run queue */
    {
	rtpx_thread_t  *cth = rtpx_enter ();

	EVLOGCREATE (cth, ath, ath->th_prio);

	rtpx_resumeWithValue (ath, 0);

	return rtpx_leaveWithValueAndDispatch (cth, 0);
    }
}



void
_pthread_init (void)
{
    static int      pthread_inited = 0;
    rtpx_thread_t  *cth;

    if (pthread_inited)
	return;
    pthread_inited = 1;

    pthread_attr_init (&thaDefault);
    pthread_mutexattr_init (&mxaDefault);
    pthread_condattr_init (&cdaDefault);

    pthread_mutex_setname_np (&thMx, "thrdtab");

    /* initialise free threads */
    {
	rtpx_thread_t  *th;

	for (th = rtpx_threads; th < &rtpx_threads[_RTPX_THREAD_MAX]; th++) {

	    th->th_itimer.it_type = IT_TYPE_DEAD;
	    th->th_itimer.it_thread = th;

	    th->th_state = THSTATE_DEAD;
	    th->th_next = freeThreadList;
	    freeThreadList = th;
	}
    }

    {
	cth = freeThreadList;
	freeThreadList = cth->th_next;

	cth->th_id = threadID = 2;
	strcpy (cth->th_name, "main");

	/* initialise thread attributes */
	cth->th_tha = thaDefault;	/* nearly */
	cth->th_tha.tha_detachstate = PTHREAD_CREATE_DETACHED;
	cth->th_tha.tha_inheritsched = PTHREAD_EXPLICIT_SCHED;
	cth->th_tha.tha_schedpolicy = SCHED_RR;
	cth->th_tha.tha_schedparam.sched_priority = _RTPX_PRIO_MIN;
	cth->th_tha.tha_name = cth->th_name;

	/* priority and round robin scheduling params */
	cth->th_ticksToRr = _rtpx_rrclock;
	cth->th_prio = cth->th_tha.tha_schedparam.sched_priority;

	/* join cond variable */
	pthread_cond_init (&cth->th_joinCd, NULL);
	pthread_cond_setname_np (&cth->th_joinCd, cth->th_name);

	/* default cancel state */
	cth->th_flags = RTPX_THF_MAIN | RTPX_THF_CANCEL_DEFERRED;

#ifdef _RTPX_TSD
	/* enpty thread-specific data */
	cth->th_tsd = NULL;
	cth->th_tsdcount = 0;
#endif

	/* start it */
	RTPX_SETSTATE (cth, THSTATE_CPUWAIT, NULL);
	enqTh (&cpuWaitq, cth);
	_cpuTab[0].runThread = cth;
    }

    /* default round robin at 100 Hz (10ms) */
    if (_rtpx_rrfreq == 0)
	_rtpx_rrfreq = _RTPX_RRFREQ;

    /* initialise board-specific interrupts and timer */
    _sbd_intrinit ();

    /* get timer frequency */
    _rtpx_tickspersec = _sbd_tickspersec ();

    /* calculate clock tick period in ns */
    _rtpx_nspertick = ((NSPERSEC + (unsigned long)_rtpx_tickspersec / 2) /
		       (unsigned long)_rtpx_tickspersec);

    /* round robin rescheduling timer in ticks */
    _rtpx_rrclock = (_rtpx_tickspersec + _rtpx_rrfreq / 2) / _rtpx_rrfreq;

    /* start up the timer */
    {
	time_t          tstart = _cpu_time (NULL);
	unsigned int    now = _sbd_gettickval ();
#if _RTPX_CLOCK64_64BIT
	bootClock = (clock64_t)_rtpx_tickspersec * tstart;
	baseTimer.it_interval = _rtpx_tickspersec / 2;
#else
	bootClock.c_sec = tstart;
	bootClock.c_tick = 0;
	baseTimer.it_interval.c_sec = 0;
	baseTimer.it_interval.c_tick = _rtpx_tickspersec / 2;
#endif
	clock64add (&bootClock, &baseTimer.it_interval, &baseTimer.it_value);
	wallClock.clk_t64 = bootClock;
	wallClock.clk_when = now;
	_cpuTab[0].startTicks = now;
	_cpuTab[0].timerTicks = now + clock64ticks (&baseTimer.it_interval);
	_sbd_settickcmp (_cpuTab[0].timerTicks);
    }

    /* initialise cpu-specific code */
    _cpu_rtpxinit ();

    /* current stack limits */
    _cpu_stacklimits (&cth->th_tha.tha_stackaddr,
		      &cth->th_tha.tha_stacksize);

    /* find stack limits and insert trivial stack guard region */
    if (cth->th_stack = cth->th_tha.tha_stackaddr) {
	long           *red = (long *) cth->th_stack;
	memset (cth->th_stack, STACKCOLOUR, cth->th_tha.tha_stacksize);
	red[0] = red[1] = red[2] = red[3] = REDSCARE;
    }

    /* initialise counters and logs */
    _rtpx_stats_init ();

    /* create the initial thread */
    (void) rtpx_leaveWithValueAndDispatch (rtpx_enter (), 0);
    nThread = 1;
    __isthreaded = 1;
    __sdethread_threading = 1;

    /* create initial process containing just this thread */
    _rtpx_newproc (NULL);
    assert (cth->th_proc);

#if defined(_RTPX_FLOCKFILE)
    /* initialise stdio file locking */
    _rtpx_flockfileinit ();
#endif

#ifdef _RTPX_TSD
    /* initialise thread-specific data */
    _rtpx_tsdinit ();
#endif
}

void
pthread_init (void)
{
#if defined(__i386__)
    _pthread_init ();
#endif
}



/* 
 * Convert pthread_t to a rtpx_thread_t pointer.
 * This should only be called when in "kernel mode" (rtpx_enter), or
 * with the thMx mutex locked.
 */
static rtpx_thread_t *
rtpx_threadLookup (pthread_t thid)
{
    rtpx_thread_t  *ath = (rtpx_thread_t *) thid;
    if (ath < rtpx_threads || ath >= &rtpx_threads[_RTPX_THREAD_MAX]
	|| ath->th_state == THSTATE_DEAD)
	return NULL;
    return ath;
}


int
pthread_setname_np (pthread_t thid, const char *name)
{
    rtpx_thread_t  *ath;
    int err = 0;

    pthread_mutex_lock (&thMx);
    ath = rtpx_threadLookup (thid);
    if (ath)
      ath->th_tha.tha_name = name;
    else
      err = ESRCH;
    pthread_mutex_unlock (&thMx);
    return err;
}


int
pthread_getname_np (pthread_t thid, const char **namep)
{
    rtpx_thread_t  *ath;
    int err = 0;

    pthread_mutex_lock (&thMx);
    ath = rtpx_threadLookup (thid);
    if (ath)
      *namep = ath->th_tha.tha_name;
    else
      err = ESRCH;
    pthread_mutex_unlock (&thMx);
    return err;
}


int
pthread_getsequence_np (pthread_t thid, int *seqp)
{
    rtpx_thread_t  *ath;
    int err = 0;

    pthread_mutex_lock (&thMx);
    ath = rtpx_threadLookup (thid);
    if (ath)
      *seqp = ath->th_id;
    else
      err = ESRCH;
    pthread_mutex_unlock (&thMx);
    return err;
}


int
pthread_stackused_np (pthread_t thid, int *usedp)
{
    rtpx_thread_t  *ath;
    int err = 0;

    pthread_mutex_lock (&thMx);
    ath = rtpx_threadLookup (thid);
    if (ath)
      *usedp = _rtpx_stackused (ath);
    else
      err = ESRCH;
    pthread_mutex_unlock (&thMx);
    return err;
}


int
pthread_setconcurrency_np (int new)
{
    if (new < 0)
	return EINVAL;
    _rtpx_concurrency = new;
    return 0;
}


int
pthread_getconcurrency_np ()
{
    return _rtpx_concurrency;
}



static void
cleanup_join (void *arg)
{
    pthread_mutex_unlock (&thMx);
}


int
pthread_join (pthread_t thid, void **rvp)
{
    if (thid == pthread_self ())
	return EDEADLK;

    pthread_mutex_lock (&thMx);

    while (1) {
	rtpx_thread_t  *oth;

	oth = rtpx_threadLookup (thid);
	if (!oth || oth->th_tha.tha_detachstate != PTHREAD_CREATE_JOINABLE) {
	    pthread_mutex_unlock (&thMx);
	    return ESRCH;
	}

	if (oth->th_state != THSTATE_ZOMBIE) {
	    pthread_cleanup_push (cleanup_join, 0);
	    pthread_cond_wait (&oth->th_joinCd, &thMx);
	    pthread_cleanup_pop (0);
	    continue;
	}

	/* free thread data of zombie thread */
	oth->th_state = THSTATE_DEAD;
	if (oth->th_tha.tha_stackaddr == NULL && oth->th_stack != NULL)
	    free (oth->th_stack);
	oth->th_stack = NULL;
	oth->th_next = freeThreadList;
	freeThreadList = oth;

	if (rvp)
	    *rvp = oth->th_zval;
	break;
    }

    pthread_mutex_unlock (&thMx);
    return 0;
}


int
pthread_detach (pthread_t thid)
{
    rtpx_thread_t  *oth;
    int             rv = 0;

    pthread_mutex_lock (&thMx);

    oth = rtpx_threadLookup (thid);
    if (oth && oth->th_tha.tha_detachstate == PTHREAD_CREATE_JOINABLE) {
	if (oth->th_state == THSTATE_ZOMBIE) {
	    /* release zombie thread storage */
	    oth->th_state = THSTATE_DEAD;
	    if (oth->th_tha.tha_stackaddr == NULL && oth->th_stack != NULL)
		free (oth->th_stack);
	    oth->th_stack = NULL;
	    oth->th_next = freeThreadList;
	    freeThreadList = oth;
	}
	else {
	    /* set to detached state */
	    oth->th_tha.tha_detachstate = PTHREAD_CREATE_DETACHED;
	}
    }
    else if (oth)
	rv = EINVAL;
    else
	rv = ESRCH;

    pthread_mutex_unlock (&thMx);
    return rv;
}


void
_pthread_cleanup_push (struct _pthread_handler_rec *rec)
{
    /* should be no need for protection, since it only effect current thread */
    rtpx_thread_t  *cth = rtpx_self ();
    rec->next = cth->th_cleanups;
    cth->th_cleanups = rec;
}


void
_pthread_cleanup_pop (struct _pthread_handler_rec *rec)
{
    /* should be no need for protection, since it only effect current thread */
    rtpx_thread_t  *cth = rtpx_self ();
    assert (rec == cth->th_cleanups);
    cth->th_cleanups = rec->next;
}


void
_pthread_cleanup_push_defer (struct _pthread_handler_rec *rec)
{
    pthread_setcanceltype (PTHREAD_CANCEL_DEFERRED, &rec->canceltype);
    _pthread_cleanup_push (rec);
}


void
_pthread_cleanup_pop_restore (struct _pthread_handler_rec *rec)
{
    _pthread_cleanup_pop (rec);
    pthread_setcanceltype (rec->canceltype, NULL);
}



int
pthread_getschedparam (pthread_t thid, int *policy, struct sched_param *sp)
{
    rtpx_thread_t  *th;

    pthread_mutex_lock (&thMx);

    th = rtpx_threadLookup (thid);
    if (!th) {
	pthread_mutex_unlock (&thMx);
	return ESRCH;
    }

    if (policy)
	*policy = th->th_tha.tha_schedpolicy;
    if (sp)
	*sp = th->th_tha.tha_schedparam;

    pthread_mutex_unlock (&thMx);
    return 0;
}



int
pthread_setschedparam (pthread_t thid, int policy,
		       const struct sched_param *sp)
{
    rtpx_thread_t  *cth;
    rtpx_thread_t  *th;
    int             nprio = sp->sched_priority;
    int             oprio;

    if (nprio < _RTPX_PRIO_MIN || nprio > _RTPX_PRIO_MAX)
	return EINVAL;
    if (policy != SCHED_FIFO && policy != SCHED_RR)
	return EINVAL;

    cth = rtpx_enter ();

    th = rtpx_threadLookup (thid);
    if (!th)
	return rtpx_leaveWithValue (cth, ESRCH);

    oprio = th->th_tha.tha_schedparam.sched_priority;
    th->th_tha.tha_schedpolicy = policy;
    th->th_tha.tha_schedparam = *sp;
    if (oprio == nprio)
	return rtpx_leaveWithValue (cth, 0);

    /* calculate new effective priority */
    {
	pthread_mutex_t *omx;

	for (omx = th->th_mxlist; omx; omx = omx->mx_olink) {
	    unsigned char   mxprio;
	    switch (omx->mx_mxa.mxa_protocol) {
	    case PTHREAD_PRIO_INHERIT:
		if (omx->mx_waitq.wq_first == NULL)
		    continue;
		mxprio = omx->mx_waitq.wq_first->th_prio;
		break;
	    case PTHREAD_PRIO_PROTECT:
		mxprio = omx->mx_mxa.mxa_prioceiling;
		break;
	    case PTHREAD_PRIO_NONE:
		continue;
	    default:
		abort ();
	    }
	    if (mxprio > nprio)
		nprio = mxprio;
	}
    }

    if (nprio != th->th_prio)
	rtpx_setepri (th, nprio);

    if (cth != cpuWaitq.wq_first)
	/* we're no longer highest priority - give up the cpu */
	return rtpx_leaveWithValueAndDispatch (cth, 0);

    return rtpx_leaveWithValue (cth, 0);
}


/* set round-robin interval (XXX not per-process yet) */
int
_rtpx_rr_set_interval (pid_t pid, const struct timespec *ts)
{
    rtpx_thread_t  *cth;
    clock64_t clk;

    /* arbitrary limits: shoulc be less than 100ms (10Hz) and more than 1ms
       (1000Hz) */
    if (!timespecok (ts) || ts->tv_sec != 0
	|| ts->tv_nsec > NSPERSEC / 10
	|| ts->tv_nsec < NSPERSEC / 1000) 
	return EINVAL;

    clk = _rtpx_tsToClock64 (ts);

    cth = rtpx_enter ();
    _rtpx_rrclock = clock64ticks (&clk);
    return rtpx_leaveWithValue (cth, 0);
}



/* 
 * THREAD SIGNALS
 */

int
_rtpx_resumeWithSig (rtpx_thread_t *oth, int signo,
		     const struct siginfo *si)
{
    int		masked;

    EVLOGSIGQUEUE (rtpx_self (), signo, oth);
    sigaddset (&oth->th_sigpend, signo);
    if (si)
	oth->th_siginfo[signo - 1] = *si;
    else {
	oth->th_siginfo[signo - 1].si_signo = signo;
	oth->th_siginfo[signo - 1].si_code = SI_USER;
	oth->th_siginfo[signo - 1].si_value.sival_ptr = NULL;
    }

    masked = sigismember (&oth->th_sigmask, signo);

    switch (oth->th_state) {
    case THSTATE_CPUWAIT:	/* will find it when rescheduled */
	if (!masked && oth == rtpx_self ())
	    return 1;
	break;

    case THSTATE_MUTEXWAIT:	/* can't break a mutex wait */
	break;

    case THSTATE_SEMWAIT:
	if (!masked) {
	    sem_t          *sem = (sem_t *) (oth->th_waitq);
	    assert (sem->sem_value < 0);
	    sem->sem_value++;
	    deqTh (oth->th_waitq, oth);
	    return rtpx_resumeWithError (oth, EINTR);
	}
	break;

    case THSTATE_CONDWAIT:
    case THSTATE_DOORBELLWAIT:
	if (!masked) {
	    deqTh (oth->th_waitq, oth);
	    return rtpx_resumeWithCleanups (oth, EINTR);
	}
	break;

    case THSTATE_TIMEWAIT:
	if (!masked)
	    return rtpx_resumeWithCleanups (oth, EINTR);
	break;

    case THSTATE_SIGWAIT:
	if (!masked || sigismember (&oth->th_sigwait, signo))
	    return rtpx_resumeWithCleanups (oth, EINTR);
	break;

    case THSTATE_ZOMBIE:	/* can't happen (should be masked) */
    case THSTATE_DEAD:		/* can't happen (lookup failure) */
    default:
	abort ();
    }

    return 0;
}


int
pthread_sigmask (int how, const sigset_t * np, sigset_t * op)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    int             rv;

    rv = _rtpx_sigmask (cth, how, np, op);
    if (rv > 0)
	return rtpx_leaveWithValue (cth, rv);
    else if (rv < 0)
	return rtpx_leaveWithValueLong (cth, 0);
    else
	return rtpx_leaveWithValue (cth, 0);
}


int
pthread_kill (pthread_t thid, int signo)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    rtpx_thread_t  *oth;

    if (signo < 0 || signo > _RTPX_SIGNO_MAX)
	return rtpx_leaveWithValue (cth, EINVAL);

    oth = rtpx_threadLookup (thid);
    if (oth == NULL || oth->th_state == THSTATE_ZOMBIE)
	return rtpx_leaveWithValue (cth, ESRCH);

    if (signo == 0)
	return rtpx_leaveWithValue (cth, 0);

    if (_rtpx_resumeWithSig (oth, signo, NULL))
	return rtpx_leaveWithValueAndDispatch (cth, 0);

    return rtpx_leaveWithValue (cth, 0);
}


/* 
 * POSIX "kernel" support: disable all asynchronous signals and
 * cancellations while handling a "system call"
 */

void
pthread_sigdisable_np (pthread_sigstate_np_t * state)
{
    rtpx_thread_t  *cth = rtpx_self ();
    sigset_t        nm;

    /* cancellation point */
    pthread_testcancel ();

    /* block all signals and save old mask */
    sigfillset (&nm);
    pthread_sigmask (SIG_SETMASK, &nm, &state->omask);

    /* disable cancellation */
    pthread_setcancelstate (PTHREAD_CANCEL_DISABLE, &state->cancel);

    /* use old signal mask for to select which signals will restart
       interruptible condition waits */
    state->odefer = cth->th_sigdefermask;
    cth->th_sigdefermask = state->omask;
}


void
pthread_sigrestore_np (const pthread_sigstate_np_t * state)
{
    rtpx_thread_t  *cth = rtpx_self ();
    cth->th_sigdefermask = state->odefer;
    pthread_setcancelstate (state->cancel, NULL);
    pthread_testcancel ();
    pthread_sigmask (SIG_SETMASK, &state->omask, NULL);
}


int
pthread_sigdeferred_np (void)
{
    rtpx_thread_t  *cth = rtpx_self ();
    sigset_t        sigpend;
    sigpending (&sigpend);
    return (sigpend & ~cth->th_sigdefermask) != 0;
}

/* SDETHREAD API aliases */
void	__sdethread_sigdisable (__sdethread_sigstate_t *)
	__attribute__((alias("pthread_sigdisable_np")));
void	__sdethread_sigrestore (const __sdethread_sigstate_t *)
	__attribute__((alias("pthread_sigrestore_np")));
int	__sdethread_sigdeferred (void)
	__attribute__((alias("pthread_sigdeferred_np")));



/* 
 * GLOBAL (PROCESS) SIGNALS
 */

int
_rtpx_sigqueue (rtpx_proc_t *pr, int code, const struct sigevent *sev)
{
    int             signo = sev->sigev_signo;
    rtpx_thread_t  *cth;
    rtpx_thread_t  *oth;
    siginfo_t       si;
    int		    fatal;

    if (signo <= 0 || signo >= _RTPX_SIGNO_MAX)
	return EINVAL;

    if (sigismember (&pr->pr_sigignore, signo))
	return 0;

    si.si_code = code;
    si.si_signo = signo;
    si.si_value = sev->sigev_value;

    cth = rtpx_self ();

    /* detect an uncaught, unignored fatal signal */
    fatal = ! sigismember (&pr->pr_sigcatch, signo) && _rtpx_sigfatal (signo);

    if (fatal){
	/* uncaught, unignored fatal signal: send to all threads in process 
	   (XXX the signal could be masked though, check for that). */
	int all = 1;
	pr->pr_exitval = signo;
	for (oth = rtpx_threads; oth < &rtpx_threads[_RTPX_THREAD_MAX]; oth++)
	    if (oth->th_proc == pr
		&& oth->th_state != THSTATE_DEAD
		&& oth->th_state != THSTATE_ZOMBIE
		&& ! _rtpx_resumeWithSig (oth, signo, &si))
		all = 0;
	/* if all threads were resumed, then we're done */
	if (all)
	    return 0;
	/* check for any that are in an interruptible cond_wait */
	goto intrwait;
    }

    /* can we send to current thread? */
    oth = cth;
    if (oth->th_proc == pr && !sigismember (&oth->th_sigmask, signo))
	goto gotone;

    /* cpu fault must go to current thread */
    if (code >= SI_FAULT)
	return EAGAIN;

    /* XXX should keep threads on a list per process to speed up these
       searches, or just search table once and compute best thread. */

    /* any thread in THSTATE_SIGWAIT and waiting for this sig */
    for (oth = rtpx_threads; oth < &rtpx_threads[_RTPX_THREAD_MAX]; oth++)
	if (oth->th_proc == pr
	    && oth->th_state == THSTATE_SIGWAIT
	    && (sigismember (&oth->th_sigwait, signo)
		|| !sigismember (&oth->th_sigmask, signo)))
	    goto gotone;

    /* any runnable thread */
    for (oth = cpuWaitq.wq_first; oth; oth = oth->th_next)
	if (oth->th_proc == pr && !sigismember (&oth->th_sigmask, signo))
	    goto gotone;

    /* any blocked thread (but not mutex!) */
    for (oth = rtpx_threads; oth < &rtpx_threads[_RTPX_THREAD_MAX]; oth++)
	if (oth->th_proc == pr
	    && (oth->th_state == THSTATE_CONDWAIT
		|| oth->th_state == THSTATE_SEMWAIT
		|| oth->th_state == THSTATE_DOORBELLWAIT
		|| oth->th_state == THSTATE_TIMEWAIT)
	    && !sigismember (&oth->th_sigmask, signo))
	    goto gotone;

    /* any mutex waiting thread (won't get sig until wait completes) */
    for (oth = rtpx_threads; oth < &rtpx_threads[_RTPX_THREAD_MAX]; oth++)
	if (oth->th_proc == pr
	    && oth->th_state == THSTATE_MUTEXWAIT
	    && !sigismember (&oth->th_sigmask, signo))
	    goto gotone;

    /* wake up any interruptible cond_waits */
 intrwait:
    for (oth = rtpx_threads; oth < &rtpx_threads[_RTPX_THREAD_MAX]; oth++)
	if (oth->th_proc == pr
	    && oth->th_state == THSTATE_CONDWAIT
	    && (oth->th_flags & RTPX_THF_CONDINTRUPT)
	    && !sigismember (&oth->th_sigdefermask, signo)) {
	    /* wake up, but don't send signal (yet) */
	    deqTh (oth->th_waitq, oth);
	    rtpx_resumeWithCleanups (oth, EINTR);
	}

    /* fatal signals have already been sent to all threads in process */
    if (!fatal) {
	/* queue the signal for the process, it should be dequeued when/if the
	   signal is unmasked by any thread. */
	EVLOGSIGQUEUE (cth, 0x100 | signo, pr);
	sigaddset (&pr->pr_sigpend, signo);
	pr->pr_siginfo[signo - 1] = si;
    }
    return 0;

  gotone:
    _rtpx_resumeWithSig (oth, signo, &si);
    return 0;
}



/* 
 * CANCELLATION
 */

static int
rtpx_cancel (rtpx_thread_t *oth)
{
    EVLOGCANCEL (rtpx_self (), oth, 1);
    oth->th_flags |= RTPX_THF_CANCELLED | RTPX_THF_CANCELLING;
    oth->th_flags &= ~RTPX_THF_CANCEL_PENDING;

    switch (oth->th_state) {
    case THSTATE_CPUWAIT:	/* will find it when rescheduled */
	if (oth == rtpx_self ())
	    return 1;
	break;

    case THSTATE_MUTEXWAIT:	/* ditto (we can't break a mutex wait) */
	break;

    case THSTATE_SEMWAIT:
	{
	    sem_t          *sem = (sem_t *) (oth->th_waitq);
	    assert (sem->sem_value < 0);
	    sem->sem_value++;
	    deqTh (oth->th_waitq, oth);
	    return rtpx_resumeWithError (oth, EINTR);
	}

    case THSTATE_CONDWAIT:
    case THSTATE_DOORBELLWAIT:
	deqTh (oth->th_waitq, oth);
	/* fall through */

    case THSTATE_TIMEWAIT:
    case THSTATE_SIGWAIT:
	return rtpx_resumeWithCleanups (oth, EINTR);

    case THSTATE_ZOMBIE:
    case THSTATE_DEAD:		/* cant happen lookup failure */
    default:
	abort ();
    }
    return 0;
}


int
pthread_setcancelstate (int new, int *oldp)
{
    rtpx_thread_t  *cth = rtpx_self ();
    unsigned int    oflags;

    switch (new) {
    case PTHREAD_CANCEL_ENABLE:
	oflags = _cpu_atomic_bic (&cth->th_flags, RTPX_THF_CANCEL_DISABLE);
	break;
    case PTHREAD_CANCEL_DISABLE:
	oflags = _cpu_atomic_bis (&cth->th_flags, RTPX_THF_CANCEL_DISABLE);
	break;
    default:
	return EINVAL;
    }

#if RTPX_THF_CANCEL_DISABLE==PTHREAD_CANCEL_DISABLE \
    && PTHREAD_CANCEL_ENABLE==0
    if (oldp)
	*oldp = oflags & RTPX_THF_CANCEL_DISABLE;
#else
    if (oldp)
	*oldp = (oflags & RTPX_THF_CANCEL_DISABLE) ?
	    PTHREAD_CANCEL_DISABLE : PTHREAD_CANCEL_ENABLE;
#endif

    return 0;
}


int
pthread_setcanceltype (int new, int *oldp)
{
    rtpx_thread_t  *cth = rtpx_self ();
    unsigned int    oflags;

    switch (new) {
    case PTHREAD_CANCEL_ASYNCHRONOUS:
	oflags =
	    _cpu_atomic_bic (&cth->th_flags, RTPX_THF_CANCEL_DEFERRED);
	break;
    case PTHREAD_CANCEL_DEFERRED:
	oflags =
	    _cpu_atomic_bis (&cth->th_flags, RTPX_THF_CANCEL_DEFERRED);
	break;
    default:
	return EINVAL;
    }

#if RTPX_THF_CANCEL_DEFERRED==PTHREAD_CANCEL_DEFERRED \
    && PTHREAD_CANCEL_ASYNCHRONOUS==0
    if (oldp)
	*oldp = oflags & RTPX_THF_CANCEL_DEFERRED;
#else
    if (oldp)
	*oldp = (oflags & RTPX_THF_CANCEL_DEFERRED) ?
	    PTHREAD_CANCEL_DEFERRED : PTHREAD_CANCEL_ASYNCHRONOUS;
#endif

    return 0;
}


int
_rtpx_sendcancel (rtpx_thread_t *cth, rtpx_thread_t *oth)
{
    if ((oth->th_flags & (RTPX_THF_CANCEL_DISABLE | RTPX_THF_CANCELLING))
	|| ((oth->th_flags & RTPX_THF_CANCEL_DEFERRED)
	    && oth != cth
	    && (oth->th_state == THSTATE_CPUWAIT
		|| oth->th_state == THSTATE_MUTEXWAIT))) {
	/* disabled or deferred and not cancellable */
	oth->th_flags |= RTPX_THF_CANCEL_PENDING;
	EVLOGCANCEL (cth, oth, 0);
    }
    else {
	/* immediate async cancel */
	if (rtpx_cancel (oth))
	    return 1;
    }

    return 0;
}


int
pthread_cancel (pthread_t thid)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    rtpx_thread_t  *oth;

    oth = rtpx_threadLookup (thid);
    if (oth == NULL || oth->th_state == THSTATE_ZOMBIE)
	return rtpx_leaveWithValue (cth, ESRCH);

    if (_rtpx_sendcancel (cth, oth))
	/* force reschedule */
	return rtpx_leaveWithValueAndDispatch (cth, 0);

    return rtpx_leaveWithValue (cth, 0);
}


void
pthread_testcancel (void)
{
    if (rtpx_cancel_pending (rtpx_self ())) {
	rtpx_thread_t  *cth = rtpx_enter ();
	EVLOGCANCEL (cth, cth, 1);
	cth->th_flags |= RTPX_THF_CANCELLED | RTPX_THF_CANCELLING;
	cth->th_flags &= ~RTPX_THF_CANCEL_PENDING;
	rtpx_leaveWithValueLong (cth, 0);
    }
}



/* 
 * POSIX CLOCK SUPPOET
 */

int
clock_getres (clockid_t which, struct timespec *ts)
{
    switch (which) {
    case CLOCK_REALTIME:
    case CLOCK_MONOTONIC:
    case CLOCK_PROCTIME:
    case CLOCK_IDLETIME:
	ts->tv_sec = 0;
	ts->tv_nsec = (NSPERSEC + _rtpx_tickspersec / 2)
	    / _rtpx_tickspersec;
	break;
    default:
	return pthread_error (EINVAL);
    }
    return 0;
}


int
clock_gettime (clockid_t which, struct timespec *ts)
{
    rtpx_thread_t  *cth = rtpx_enter ();

    switch (which) {
    case CLOCK_REALTIME:
    case CLOCK_MONOTONIC:
	_rtpx_tsFromClock64 (ts, _rtpx_clockNow ());
	assert (timespecok (ts));
	return rtpx_leaveWithValue (cth, 0);
    case CLOCK_PROCTIME:
    case CLOCK_IDLETIME:
    default:
	return rtpx_leaveWithError (cth, EINVAL);
    }
}


int
clock_settime (clockid_t which, const struct timespec *ts)
{
    rtpx_thread_t  *cth = rtpx_enter ();

    switch (which) {
    case CLOCK_REALTIME:
	/* XXX set local s/w clock and h/w clock - to be implemented */
	return rtpx_leaveWithError (cth, EPERM);
    case CLOCK_MONOTONIC:
    case CLOCK_PROCTIME:
    case CLOCK_IDLETIME:
    default:
	return rtpx_leaveWithError (cth, EINVAL);
    }
    return rtpx_leaveWithError (cth, 0);
}



#include <sdethread.h>

/* Apart from __sdethread_mutex_init, most sdethread functions map
   one-to-one with the similarly named pthread functions. The
   sdethread objects are designed to be large enough to hold RTPX's
   equivalent pthread object implementations directly. This may not be
   true for other thread libraries/kernels, of course; in which case
   some indirection will be required. */

static pthread_once_t st_mxa_once = PTHREAD_ONCE_INIT;
static pthread_mutexattr_t st_mxa;

static void
st_mxa_init ()
{
    /* Enable priority inheritance */
    if (pthread_mutexattr_init (&st_mxa) == 0)
	pthread_mutexattr_setprotocol (&st_mxa, PTHREAD_PRIO_INHERIT);
}


int __sdethread_mutex_init (__sdethread_mutex_t *mx, const char *name)
{
    pthread_mutex_t *pmx = (pthread_mutex_t *)mx;
    int ret;
    
    ret = pthread_once (&st_mxa_once, st_mxa_init);
    if (ret == 0)
	ret = pthread_mutex_init (pmx, &st_mxa);
    if (ret == 0 && name)
	ret = pthread_mutex_setname_np (pmx, name);
    return ret;
}

int __sdethread_mutex_lock (__sdethread_mutex_t *)
	__attribute__((alias("pthread_mutex_lock")));

int __sdethread_mutex_trylock (__sdethread_mutex_t *)
	__attribute__((alias("pthread_mutex_trylock")));

int __sdethread_mutex_unlock (__sdethread_mutex_t *)
	__attribute__((alias("pthread_mutex_unlock")));

int __sdethread_mutex_destroy (__sdethread_mutex_t *)
	__attribute__((alias("pthread_mutex_destroy")));

__sdethread_t __sdethread_self (void)
	__attribute__((alias("pthread_self")));
