#ident "sde/rtpx/rtpx_timer.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2004 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * rtpx_timer.c: Posix real-time interval timers
 */


#include "pthread_impl.h"

int
timer_gettime (timer_t tid, struct itimerspec *oit)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    rtpx_timer_t   *ti = (rtpx_timer_t *) tid;

    clock64_t       n = _rtpx_clockNow ();
    clock64_t       v = ti->ti_it.it_value;
    clock64_t       i = ti->ti_it.it_interval;

    (void) rtpx_leaveWithValue (cth, 0);

    if (clock64cmp (&v, &n, <)) {
	/* already expired */
	timespecclear (&oit->it_value);
    }
    else {
	clock64_t d;
	clock64sub (&v, &n, &d);
	_rtpx_tsFromClock64 (&oit->it_value, d);
    }
    _rtpx_tsFromClock64 (&oit->it_interval, i);

    return 0;
}


/* XXX this is not strictly as defined, it is meant to return the overrun
   count at the point where the signal "was delivered" */
int
timer_getoverrun (timer_t tid)
{
    rtpx_thread_t  *cth = rtpx_enter ();
    rtpx_timer_t   *ti = (rtpx_timer_t *) tid;
    int             val = ti->ti_overrun;
    ti->ti_overrun = -1;
    return rtpx_leaveWithValue (cth, val);
}


int
timer_settime (timer_t tid, int flags,
	       const struct itimerspec *nit, struct itimerspec *oit)
{
    if (oit == nit) {
	struct itimerspec ait;
	int             rv;

	if ((rv = timer_settime (tid, flags, nit, &ait)) == 0)
	    *oit = ait;

	return rv;
    }

    switch (flags) {
    case 0:
    case TIMER_ABSTIME:
	break;
    default:
	errno = EINVAL;
	return -1;
    }

    assert (nit);

    if (!timespecok (&nit->it_value) || !timespecok (&nit->it_interval)) {
	errno = EINVAL;
	return -1;
    }

    {
	rtpx_thread_t  *cth = rtpx_enter ();
	rtpx_timer_t   *ti = (rtpx_timer_t *) tid;
	clock64_t       n = _rtpx_clockNow ();

	if (ti->ti_it.it_type != IT_TYPE_DEAD)
	    _rtpx_itimerstop (&ti->ti_it);

	if (oit) {
	    if (clock64cmp (&ti->ti_it.it_value, &n, <=))
		timespecclear (&oit->it_value);	/* XXX yuk */
	    else {
		clock64_t delta;
		clock64sub (&ti->ti_it.it_value, &n, &delta);
		_rtpx_tsFromClock64 (&oit->it_value, delta);
	    }
	    _rtpx_tsFromClock64 (&oit->it_interval, ti->ti_it.it_interval);
	}

	if (timespecisset (&nit->it_value)) {
	    clock64_t       v = _rtpx_tsToClock64 (&nit->it_value);
	    clock64_t       i = _rtpx_tsToClock64 (&nit->it_interval);

	    ti->ti_it.it_interval = i;
	    if (flags == TIMER_ABSTIME) {
		if (clock64cmp (&v, &n, <)) {
		    /* XXX should fire already !! */
		    return rtpx_leaveWithError (cth, EINVAL);
		}
	    }
	    else {
		clock64add (&v, &n, &v);
	    }
	    _rtpx_itimerstartabs (&ti->ti_it, IT_TYPE_TIMER, v);
	}

	return rtpx_leaveWithValueAndDispatch (cth, 0);
    }
}


int
timer_delete (timer_t tid)
{
    int             ret = 0;
    if (tid) {
	struct itimerspec zit;
	timespecclear (&zit.it_value);
	timespecclear (&zit.it_interval);
	ret = timer_settime (tid, 0, &zit, NULL);
	free (tid);
    }
    return ret;
}


int
timer_create (clockid_t which, struct sigevent *sev, timer_t * ptid)
{
    sigset_t        sm;
    int             error = 0;

    sigfillset (&sm);
    pthread_sigmask (SIG_BLOCK, &sm, &sm);
    {
	rtpx_timer_t   *ti;

	if (which != CLOCK_REALTIME && which != CLOCK_MONOTONIC) {
	    error = EINVAL;
	}
	else if (sev) {
	    switch (sev->sigev_notify) {
	    default:
		error = EINVAL;
		break;
	    case SIGEV_NONE:
	    case SIGEV_SIGNAL:
	 /* case SIGEV_THREAD: */
		break;
	    }
	}
	if (error == 0 && (ti = malloc (sizeof (*ti))) != NULL) {
	    ti->ti_it.it_next = NULL;
	    ti->ti_it.it_prev = NULL;
	    clock64clear (&ti->ti_it.it_value);
	    clock64clear (&ti->ti_it.it_interval);
	    ti->ti_it.it_type = IT_TYPE_DEAD;
	    ti->ti_pr = _rtpx_self_proc ();
	    ti->ti_overrun = -1;
	    if (sev) {
		ti->ti_sigev = *sev;
	    }
	    else {
		ti->ti_sigev.sigev_notify = SIGEV_SIGNAL;
		ti->ti_sigev.sigev_signo = SIGALRM;
		ti->ti_sigev.sigev_value.sival_ptr = ti;
	    }
	    if (ptid)
		*ptid = (timer_t) ti;
	}
	else
	    error = ENOMEM;
    }
    pthread_sigmask (SIG_SETMASK, &sm, NULL);
    if (error) {
	errno = error;
	return -1;
    }
    return 0;
}
