#define _POSIX_C_SOURCE 199309L

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include "conc.h"

#include <mips/cpu.h>
#if SB3000
#include <algvme/cpu_board.h>
#else
#include <p4000/cpu_board.h>
#endif

struct pstate {
    struct timespec lastrun;
    int	hasrun;
};

struct pstate pstate[4];


static void pokeDisplay(size_t idx,char c)
{ 
  volatile unsigned long *display;
#ifdef SB3000
  *(volatile unsigned int *)PA_TO_KVA1(BCRR) 
    = BCRR_AUBLK|BCRR_AUCLR; /* |BCRR_SCSI|BCRR_ETH; */
  display = (volatile unsigned long  *)PA_TO_KVA1(ALPHA_DISP(idx));
#else
  *(volatile unsigned long *) PA_TO_KVA1 (ALPHN_BLNK_) = RESET_ONE;
  display = (unsigned long *) PA_TO_KVA1 (ALPHN_CHAR (idx));
#endif
  *display = c << 24;
}


void hello (Process *p, int idx);
void sonic (Process *p, int idx, Channel *cp);
void checker (Process *p, int idx);

int main(int argc,char **argv)
{
  Process *p;
  int pcount = 0;
  int i;

  printf("Channel test program\n");
  printf("To find long gaps with no process being scheduled\n");

  /* force initial priority low */
  ProcToLow();
  
  /* run a sonic interrupt monitor */
  p = ProcAlloc(sonic,0,2,pcount++,LINK5IN);
  ProcRunHigh(p);
  
  /* run two background processes */
  for (i = 1; i <= 2; i++) {
    p = ProcAlloc(hello,0,pcount++,i);
    ProcRunLow(p);
  }

  printf ("Processes started...\n");

  /* the final background process is us! */
  checker(NULL, pcount++);

  exit(0);
}


void hello (Process *p, int idx )
{
    struct pstate *ps = &pstate[idx];
    int dcount = 0;
    
    while (1) {
	pokeDisplay(idx,"0123456789abcdef"[++dcount & 0xf]);
	ps->hasrun = 1;
	clock_gettime (CLOCK_REALTIME, &ps->lastrun);
    }
}


#include <algvme/sonic.h>
  
#define SONIC_FREQ	5000000		/* clock is 5 MHz (half ethernet) */

/*
 * Load Sonic timer with timeout in uSecs 
 */ 
void sonic_load(unsigned long us)
{
  volatile struct sonic_reg *snr = PA_TO_KVA1 (SONIC_BASE);
  unsigned long cnt = us * (SONIC_FREQ / 1000000);
  
  /* ack any existing timer */
  snr->s_isr = ISR_TC; mips_wbflush();
  
  /* load timeout */
  snr->s_wt1 = cnt>>16; snr->s_wt0 = cnt; mips_wbflush();
  
  /* enable timer */
  snr->s_imr = IMR_TCEN;
  
  /* start timer comand */
  snr->s_cr = CR_ST; mips_wbflush();
}


/*
 * Acknowledge and clear Sonic timer interrupt.
 */ 
void sonic_ack()
{
  volatile struct sonic_reg *snr = PA_TO_KVA1 (SONIC_BASE);

  /* stop timer */
  snr->s_cr = CR_STP; mips_wbflush();

  /* acknowledge the timer interupt (if any) */
  snr->s_isr = ISR_TC; mips_wbflush();

  /* acknowledge all interupts (if any) */
  snr->s_isr = snr->s_isr; mips_wbflush();
  snr->s_imr = 0;		/* and mask */
}


void sonic(Process *p, int idx, Channel *cp)
{
    struct pstate *ps = &pstate[idx];
    int dcount = 0;

    while(1) {
	/* 1000 us clock */
	sonic_load(1000);
	/* wait for sonic timer interrupt to fire */
	(void) ChanInInt(cp);
	/* acknowledge interrupt */
	sonic_ack();
	/* advance display */
	pokeDisplay(idx,"0123456789abcdef"[++dcount & 0xf]);
	/* advance process state */
	ps->hasrun = 1;
	clock_gettime (CLOCK_REALTIME, &ps->lastrun);
    }
}


static char *
myctime (const time_t *tp)
{
    static char stime[32];
    strftime (stime, sizeof(stime), "%a %d/%m/%y %H:%M:%S", localtime(tp));
    return stime;
}


void checker (Process *p, int idx)
{
    struct pstate *ps;
    int dcount = 0;
    struct timespec thents, nowts;
    unsigned long diffms, elapsed;
    int i;

#define MSTICK	50			/* 50ms */
#define LOG	(1*60*1000/MSTICK)	/* 1 min */

    clock_gettime (CLOCK_REALTIME, &nowts);
    elapsed = 0;

    while (1) {

	thents = nowts;

	ProcReschedule ();

	clock_gettime (CLOCK_REALTIME, &nowts);

	diffms = ((nowts.tv_sec - thents.tv_sec) * 1000)
	    + ((nowts.tv_nsec - thents.tv_nsec) / 1000000);

	if (diffms >= 25) {
	    printf ("%s: ERROR: checker delayed for %d msecs\n", 
		    myctime (&nowts.tv_sec), diffms);
	    diffms = MSTICK;
	}

	elapsed += diffms;
	if (elapsed >= MSTICK) {

	    pokeDisplay(idx,"0123456789abcdef"[++dcount & 0xf]);

	    ProcToHigh ();
	    for (i = 0, ps = pstate; i < 4; i++, ps++) {
		if (i != idx && !ps->hasrun) {
		    unsigned long diffus;
		    diffus = ((nowts.tv_sec - ps->lastrun.tv_sec) * 1000000)
			+ ((nowts.tv_nsec - ps->lastrun.tv_nsec) / 1000);
		    printf ("%s: ERROR: process %d last ran %ld usecs ago\n", 
			    myctime (&nowts.tv_sec), i, diffus);
		}
		ps->hasrun = 0;
	    }
	    ProcToLow ();

	    /* restart timer */
	    clock_gettime (CLOCK_REALTIME, &nowts);
	    elapsed = 0;
	}
    }
}
