#define _POSIX_C_SOURCE 199506L

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <errno.h>
#include <assert.h>
#include <conc.h>

#include <mips/cpu.h>
#if defined(SB3000)
#include <algvme/cpu_board.h>
#elif defined(P4000)
#include <p4000/cpu_board.h>
#elif defined(P4032)
#include <sbd.h>
#include <icu.h>
#endif

struct pstate {
    pthread_mutex_t	mx;
    Process		*proc;
    unsigned int	rseed;
    struct timespec	lastrun;
    int			hasrun;
    int			stopped;
    long		maxdelay;
};

#define NPROC	11
struct pstate pstate[NPROC];

#define NSPERSEC	(1000*1000*1000)
#define NSPERMSEC	(1000*1000)
#define NSPERUSEC	(1000)

static inline void
ps_note (struct pstate *ps)
{
    pthread_mutex_lock (&ps->mx);
    clock_gettime (CLOCK_REALTIME, &ps->lastrun);
    ps->hasrun = 1;
    pthread_mutex_unlock (&ps->mx);
}

int rand_r(unsigned int *seed)
{
    *seed = *seed * 1103515245 + 12345;
#if 0
    return (*seed % ((unsigned int)RAND_MAX + 1));
#else
    return (*seed % ((unsigned int)LONG_MAX + 1));
#endif
}

static void pokeDisplay(size_t idx,char c)
{ 
  volatile unsigned long *display;
#if defined(SB3000)
  *(volatile unsigned int *)PA_TO_KVA1(BCRR) 
    = BCRR_AUBLK|BCRR_AUCLR; /* |BCRR_SCSI|BCRR_ETH; */
  display = (volatile unsigned long  *)PA_TO_KVA1(ALPHA_DISP(idx));
  *display = c << 24;
#elif defined(P4000)
  *(volatile unsigned long *) PA_TO_KVA1 (ALPHN_BLNK_) = RESET_ONE;
  display = (unsigned long *) PA_TO_KVA1 (ALPHN_CHAR (idx));
  *display = c << 24;
#elif defined(P4032)
  _bcr_bis (BCR_LED_ON);
  display = PA_TO_KVA1 (LED_BASE + LED(idx));
  *display = c;
#endif
}


void hello (Process *p, int idx);
void soak (Process *p, int idx);
void checker (Process *p, int idx);

#if defined(SB3000) || defined(P4000)
void sonic (Process *p, int idx, Channel *cp);
#elif defined(P4032)
void rtc (Process *p, int idx, Channel *cp);
#endif

int main(int argc,char **argv)
{
  Process *p;
  struct pstate *ps;
  int pcount = 0;
  int i;

  printf("Channel test program\n");
  printf("To find long gaps with no process being scheduled\n");
  printf("Includes multiple floating-point tasks\n");

  /* force initial priority low */
  ProcToLow();
  
  for (i = 0, ps = pstate; i < NPROC; i++, ps++) {
      pthread_mutex_init (&ps->mx, NULL);
      ps_note (ps);
      ps->rseed = ps->lastrun.tv_sec + ps->lastrun.tv_nsec;

      if (i == 0) {
#if defined(SB3000) || defined(P4000)
	  /* run a sonic interrupt monitor */
	  ps->proc = ProcAlloc(sonic,0,2,i,LINK5IN);
	  ProcRunHigh(ps->proc);
#elif defined(P4032)
	  /* run a real-time clock interrupt monitor */
	  ps->proc = ProcAlloc(rtc,0,2,i,LINKIN(INTR_RTC));
	  ProcRunHigh(ps->proc);
#endif
      }
      else if (i < 3) {
	  /* run two background processes writing to display */
	  ps->proc = ProcAlloc(hello,0,1,i);
	  ProcRunLow(ps->proc);
      }
      else if (i == 3) {
	  /* skip proc #3, which will be us */
      }
      else {
	  ps->proc = ProcAlloc(soak,0,1,i);
	  ProcRunLow(ps->proc);
      }
  }

  /*printf ("Processes started...\n");*/

  /* the final background process is us! */
  ProcToHigh ();
  checker(NULL, 3);

  exit(0);
}


void soak (Process *p, int idx)
{
    struct pstate *ps = &pstate[idx];
    int rcount = 0;

    ps->maxdelay = NSPERMSEC * 250;

    while (1) {
	extern double sqrt (double);
	int i = rand_r (&ps->rseed);
	double v1;
	volatile auto double v2;

	ps_note (ps);

	v2 = v1 = i;
	assert (sqrt (v1 * v1) == v1);

	/* 1 million loops ~ 1 sec */
	if (++rcount == 10000000) {
	    int delay = i % 256;
	    if (delay >= 250) {
		ps->stopped = 1;
		ProcStop ();
	    }
	    else if (delay >= 128)
		ProcReschedule ();
	    else
		ProcWaitMs (delay);
	    rcount = 0;
	}

	assert (v2 == v1);
	assert (sqrt (v2 * v2) == v1);

    }
}

void hello (Process *p, int idx)
{
    struct pstate *ps = &pstate[idx];
    int dcount = 0;
    
    ps->maxdelay = NSPERMSEC * 5;	/* 5ms */
    while (1) {
	pokeDisplay(idx,"0123456789abcdef"[++dcount & 0xf]);
	ps_note (ps);
    }
}


#if defined(SB3000) || defined(P4000)

#include <algvme/sonic.h>
  
#define SONIC_FREQ	5000000		/* clock is 5 MHz (half ethernet) */

/*
 * Load Sonic timer with timeout in uSecs 
 */ 
void sonic_load(unsigned long us)
{
  volatile struct sonic_reg *snr = PA_TO_KVA1 (SONIC_BASE);
  unsigned long cnt = us * (SONIC_FREQ / 1000000);
  
  /* ack any existing timer */
  snr->s_isr = ISR_TC; mips_wbflush();
  
  /* load timeout */
  snr->s_wt1 = cnt>>16; snr->s_wt0 = cnt; mips_wbflush();
  
  /* enable timer */
  snr->s_imr = IMR_TCEN;
  
  /* start timer comand */
  snr->s_cr = CR_ST; mips_wbflush();
}


/*
 * Acknowledge and clear Sonic timer interrupt.
 */ 
void sonic_ack()
{
  volatile struct sonic_reg *snr = PA_TO_KVA1 (SONIC_BASE);

  /* stop timer */
  snr->s_cr = CR_STP; mips_wbflush();

  /* acknowledge the timer interupt (if any) */
  snr->s_isr = ISR_TC; mips_wbflush();

  /* acknowledge all interupts (if any) */
  snr->s_isr = snr->s_isr; mips_wbflush();
  snr->s_imr = 0;		/* and mask */
}


void sonic(Process *p, int idx, Channel *cp)
{
    struct pstate *ps = &pstate[idx];
    int dcount = 0;

#define SONICUSECS	201

    ps->maxdelay = NSPERUSEC * SONICUSECS;
    while(1) {
	/* 200 us clock */
	sonic_load(SONICUSECS);
	/* wait for sonic timer interrupt to fire */
	(void) ChanInInt(cp);
	/* acknowledge interrupt */
	sonic_ack();
	/* update process state */
	ps_note (ps);
	/* advance display */
	pokeDisplay(idx,"0123456789abcdef"[++dcount & 0xf]);
    }
}

#elif defined(P4032)

#include <rtc.h>

void rtc (Process *p, int idx, Channel *cp)
{
    struct pstate *ps = &pstate[idx];
    int dcount = 0;

    _rtc_bic (RTC_STATUSA, RTC_RATE_MASK);
    _rtc_bis (RTC_STATUSA, RTC_RATE_1024Hz);
    _rtc_bis (RTC_STATUSB, RTCSB_PIE);
    _rtc_get (RTC_INTR);

    ps->maxdelay = NSPERMSEC; /* 1024Hz = 1ms */

    while(1) {
	/* wait for sonic timer interrupt to fire */
	(void) ChanInInt(cp);

	/* acknowledge i/u */
	_rtc_get (RTC_INTR);

	/* update process state */
	ps_note (ps);

	/* advance display */
	pokeDisplay(idx,"0123456789abcdef"[++dcount & 0xf]);
    }
}

#endif

static char *
myctime (const time_t *tp)
{
    static char stime[32];
    strftime (stime, sizeof(stime), "%a %d/%m/%y %H:%M:%S", localtime(tp));
    return stime;
}


void checker (Process *p, int idx)
{
    struct pstate *ps;
    int dcount = 0;
    struct timespec thents, nowts;
    unsigned long diffms, elapsed;
    int i;

#define MSTICK	10			/* 10ms */
#define LOG	(1*60*1000/MSTICK)	/* 1 min */

    clock_gettime (CLOCK_REALTIME, &nowts);
    elapsed = 0;

    while (1) {

	thents = nowts;

#if 0
	ProcReschedule ();
#else
	ProcWaitMs (MSTICK);
#endif

	clock_gettime (CLOCK_REALTIME, &nowts);

	diffms = ((nowts.tv_sec - thents.tv_sec) * 1000)
	    + ((nowts.tv_nsec - thents.tv_nsec) / 1000000);

	if (diffms >= MSTICK + 2) {
	    printf ("%s: ERROR: checker delayed for %d msecs\n", 
		    myctime (&nowts.tv_sec), diffms);
	    diffms = MSTICK;
	}

	elapsed += diffms;
	if (elapsed >= MSTICK) {

	    pokeDisplay(idx,"0123456789abcdef"[++dcount & 0xf]);

	    /*ProcToHigh ();*/
	    for (i = 0, ps = pstate; i < NPROC; i++, ps++) {
	        unsigned long diffns;
	        int hasrun;

		if (i == idx)
		    continue;

		pthread_mutex_lock (&ps->mx);
		hasrun = ps->hasrun;
		ps->hasrun = 0;
		diffns = ((nowts.tv_sec - ps->lastrun.tv_sec) * NSPERSEC)
		  + ((nowts.tv_nsec - ps->lastrun.tv_nsec));
		pthread_mutex_unlock (&ps->mx);

		if (!hasrun) {
		    if (diffns > ps->maxdelay + NSPERSEC/200*(NPROC+1)) {
			if (ps->stopped) {
			    /* manually stopped, restart it at random */
			    if (rand_r (&ps->rseed) % 100 == 0) {
				ps->stopped = 0;
				ProcRunLow (ps->proc);
			    }
			}
			else {
			    printf ("%s: ERROR: process %d last ran %ld usecs ago\n", 
				    myctime (&nowts.tv_sec), i, diffns/1000);
			    clock_gettime (CLOCK_REALTIME, &ps->lastrun);
			}
		    }
		}
	    }
	    /*ProcToLow ();*/

	    /* restart timer */
	    clock_gettime (CLOCK_REALTIME, &nowts);
	    elapsed = 0;
	}
    }
}
