#define _POSIX_C_SOURCE 199506L

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <limits.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <assert.h>
#include <pthread.h>
#include <sys/ioctl.h>
#include <sys/wait.h>

static char *__progname = "proc1";

void
verr(eval, fmt, ap)
	int eval;
	const char *fmt;
	va_list ap;
{
	int sverrno;

	sverrno = errno;
	(void)fprintf(stderr, "%s: ", __progname);
	if (fmt != NULL) {
		(void)vfprintf(stderr, fmt, ap);
		(void)fprintf(stderr, ": ");
	}
	(void)fprintf(stderr, "%s\n", strerror(sverrno));
	pthread_proc_exit_np (eval);
}


void
err(int eval, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	verr(eval, fmt, ap);
	va_end(ap);
}


#define MAXPROC 10
pid_t 	reaped[MAXPROC];
int	reapcnt;

struct procinfo {
    int 		(*main)(int, char **);
    int			argc;
    char		**argv;
};


void
setttypgrp ()
{
    int pgrp = getpgrp ();
    if (ioctl (STDIN_FILENO, TIOCSPGRP, &pgrp) < 0)
	err (1, "ioctl(TIOCSPGRP)");
}


int
startproc (void *arg)
{
    struct procinfo proc;

    /* get copy of startup data and free */
    proc = *(struct procinfo *)arg;
    free (arg);

    /* create new process group and attach to terminal */
    setpgrp ();
    setttypgrp ();

    /* run thr process */
    return (proc.main (proc.argc, proc.argv));
}


int
procexecv (const char *name, int (*main)(int, char **), int argc, char **argv)
{
    struct procinfo *proc;
    int pid;
    int errnum;

    proc = malloc (sizeof (struct procinfo));
    proc->main = main;
    proc->argc = argc;
    proc->argv = argv;

    errnum = pthread_proc_create_np (&pid, startproc, proc);
    if (errnum) {
	errno = errnum;
	err (1, "procexecv(%s)", name);
    }
    
    return pid;
}

pid_t
waitfor (const char *name, pid_t pid, int options, int wantsig)
{
    pid_t res;
    int status;

    res = waitpid (pid, &status, options);

    if (res < 0) {
	char buf[64];
	fprintf (stderr, "waitfor(%s): %s\n", name, 
		 strerror_r (errno, buf, sizeof(buf)));
	return -1;
    }

    if (res == 0) {
	if (options & WNOHANG)
	    return 0;
	fprintf (stderr, "waitfor(%s): returned %d, expected %d\n", 
		 name, res, pid);
	return -1;
    }

    if (res != pid && pid > 0) {
	fprintf (stderr, "waitfor(%s): returned %d, expected %d\n", 
		 name, res, pid);
	return -1;
    }

    if (! WIFEXITED (status)) {
	int sig = WTERMSIG (status);
	const char *c;
	if (sig < NSIG)
	    c = sys_siglist[sig];
	else
	    c = "Unknown signal";
	fprintf (stderr, "waitfor(%s): pid %d, %s\n", name, pid, c);
	return (sig == wantsig) ? res : -1;
    }

    if (WEXITSTATUS (status) != 0) {
	fprintf (stderr, "waitfor(%s): pid %d, exit status %d, expected\n",
		 name, pid, WEXITSTATUS (status));
	return -1;
    }

    return (wantsig < 0) ? res : -1;

}

void
reaper (int signo)
{
    pid_t pid;

    while ((pid = waitfor ("reaper", WAIT_ANY, WNOHANG, -1)) > 0) {
	if (reapcnt == MAXPROC)
	    err (1, "too many dead processes");
	reaped[reapcnt++] = pid;
    }
}



int
hello (int argc, char **argv)
{
    sleep (1);
    printf ("pid %d: hello\n", getpid());
    fflush (stdout);
    return 0;
}


int
suicide (int argc, char **argv)
{
    printf ("pid %d: suicide initiated\n", getpid());
    alarm (2);
    while (1)
	sleep (10);
    return 0;
}

int
sigint (int argc, char **argv)
{
    printf ("pid %d: please press ^C\n", getpid());
    while (1)
	sleep (10);
    return 0;
}



main ()
{
    struct sigaction sa;
    pid_t pid;
    pid_t wpid;

    /* Wait for child to exit */
    setttypgrp ();
    printf ("Test 1 start\n");
    pid = procexecv ("hello", hello, 0, 0);
    wpid = waitfor ("hello", pid, 0, -1);
    printf ("Test 1 (pid %d) %s\n",  pid, 
	    wpid == pid ? "succeeded" : "failed");

    /* Wait for zombie (already dead) child */
    setttypgrp ();
    printf ("Test 2 start\n");
    pid = procexecv ("hello", hello, 0, 0);
    sleep (2);
    wpid = waitfor ("hello", pid, 0, -1);
    printf ("Test 2 (pid %d) %s\n",  pid, 
	    wpid == pid ? "succeeded" : "failed");

    sa.sa_handler = reaper;
    sigemptyset (&sa.sa_mask);
    sa.sa_flags = 0;
    if (sigaction (SIGCHLD, &sa, NULL) != 0)
	err (1, "sigaction(SIGCHLD, reaper)");

    /* Catch SIGCHLD */
    setttypgrp ();
    printf ("Test 3 start\n");
    reapcnt = 0;
    pid = procexecv ("hello", hello, 0, 0);
    sleep (2);
    printf ("Test 3 (pid %d) %s\n",  pid,
	    reapcnt == 1 && reaped[0] == pid ? "succeeded" : "failed");
	
    /* disable SIGCHLD */
    sa.sa_handler = SIG_DFL;
    if (sigaction (SIGCHLD, &sa, NULL) != 0)
	err (1, "sigaction(SIGCHLD, SIG_DFL)");

    /* Let signal kill process */
    setttypgrp ();
    printf ("Test 4 start\n");
    reapcnt = 0;
    pid = procexecv ("suicide", suicide, 0, 0);
    wpid = waitfor ("suicide", pid, 0, SIGALRM);
    printf ("Test 4 (pid %d) %s\n",  pid,
	    wpid == pid ? "succeeded" : "failed");

    /* Let SIGINT kill process */
    setttypgrp ();
    printf ("Test 5 start\n");
    reapcnt = 0;
    pid = procexecv ("sigint", sigint, 0, 0);
    wpid = waitfor ("sigint", pid, 0, SIGINT);
    printf ("Test 5 (pid %d) %s\n",  pid,
	    wpid == pid ? "succeeded" : "failed");
    
    return (0);
}
