#define _POSIX_C_SOURCE 199309L

#include <limits.h>
#include <unistd.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <stdio.h>
#include <signal.h>
#include <pthread.h>
#include <sched.h>
#include <semaphore.h>

#define NSPERSEC (1000*1000*1000)

#include <mips/cpu.h>
/*#include <mips/xcpt.h>*/

static int 	actionFired = -1;
siginfo_t	actionInfo;


#define  err_check(op) \
  { \
	int _err = (op); \
        if (_err != 0) { \
      		fprintf (stderr, "%s: %d: `%s': %s\n", \
			 __FUNCTION__, __LINE__, \
			#op, strerror (_err)); \
	      	abort (); \
        } \
  }

#define  errno_check(op) \
  ({ \
	int _ret = (op); \
        if (_ret == -1) { \
      		fprintf (stderr, "%s: %d: `%s': %s\n", \
			 __FUNCTION__, __LINE__, \
			#op, strerror (errno)); \
 	      	abort (); \
        } \
  	_ret; \
  })


static void noteaction(int signo, siginfo_t *si, void *arg)
{
  extern const char * const sys_signame[];

  /* XXX calling stdion is normally illegal in a signal handler! */
  if (signo >= SIGRTMIN)
      fprintf(stderr,"Noteaction SIGRTMIN+%d\n", signo - SIGRTMIN);
  else
      fprintf(stderr,"Noteaction SIG%s\n", sys_signame[signo]);
  
  actionFired = signo;

  assert (si);
  actionInfo = *si;
}

static int noteExpect;

static void *note(void *arg)
{
  assert( (int)arg == noteExpect);
  noteExpect++;
  return arg;
}

static sem_t notesem = SEM_INITIALIZER(0);

static void *noteWithSem(void *arg)
{
    errno_check (sem_wait(&notesem));
    return note(arg);
}

static void clock_getelapsed (struct timespec *ts)
{
  static struct timespec tsfirst;
  static int first = 1;

  errno_check (clock_gettime(CLOCK_REALTIME, ts));

  if (first) {
    tsfirst = *ts;
    first = 0;
  }

  ts->tv_sec -= tsfirst.tv_sec;
  if ((ts->tv_nsec -= tsfirst.tv_nsec)  < 0) {
    ts->tv_nsec += NSPERSEC;
    ts->tv_sec--;
  }
}

static void pass(int tnum)
{
  struct timespec ats;
  clock_getelapsed (&ats);
  printf("Time %2ld.%09ld Test %2d pass\n",
	 ats.tv_sec, ats.tv_nsec,
	 tnum);
}

int test9() {

  sigset_t sm;
  struct sigaction sa1;
  struct sigaction sa2;
    
  sa1.sa_flags = SA_SIGINFO;
  sa1.sa_sigaction = noteaction;
  sigemptyset(&sa1.sa_mask);
  
  sa2 = sa1;
    
  /* install signal handlers for SIGRTMIN + 1 & SIGRTMIN + 2 */
  errno_check (sigaction(SIGRTMIN + 2,&sa2,&sa2));
  errno_check (sigaction(SIGRTMIN + 1,&sa1,&sa1));
    
  actionFired = -1;
    
  /* block SIGRTMIN + 2 */
  sigemptyset(&sm);
  sigaddset(&sm,SIGRTMIN + 2);
  err_check (pthread_sigmask(SIG_BLOCK,&sm,&sm));
    
  /* send SIGRTMIN + 2 - should be queued, not handled yet */
  err_check (pthread_kill(pthread_self(),SIGRTMIN + 2));
    
  assert(actionFired == -1);
    
  /* send SIGRTMIN + 1 - should be handled */
  err_check (pthread_kill(pthread_self(),SIGRTMIN + 1));
    
  assert(actionFired == SIGRTMIN + 1);
  actionFired = -1;
    
  /* unblock SIGRTMIN + 2 - queued signal should be handled */
  err_check (pthread_sigmask(SIG_SETMASK,&sm,NULL));
    
  assert(actionFired == SIGRTMIN + 2);

  /* reset signal handlers */
  errno_check (sigaction(SIGRTMIN + 2,&sa2,NULL));
  errno_check (sigaction(SIGRTMIN + 1,&sa1,NULL));

  return 0;
}

int test10() {
  struct timespec 	startTime;
  struct timespec 	waitTime;
  struct timespec 	stopTime;

  waitTime.tv_sec = 4;
  waitTime.tv_nsec = NSPERSEC/5;

  errno_check (clock_gettime(CLOCK_REALTIME,&startTime));
  errno_check (nanosleep(&waitTime,NULL));
  errno_check (clock_gettime(CLOCK_REALTIME,&stopTime));
  
  assert(startTime.tv_sec <= stopTime.tv_sec);
    
  stopTime.tv_sec -= startTime.tv_sec;
  if( (stopTime.tv_nsec -= startTime.tv_nsec)  < 0) {
    stopTime.tv_nsec += NSPERSEC;
    stopTime.tv_sec--;
  }
  
  printf("sleep %ld.%ld took %ld.%ld\n", 
	 waitTime.tv_sec, waitTime.tv_nsec,
	 stopTime.tv_sec, stopTime.tv_nsec);
  return 0;
}

int test11() {
  timer_t tid;
  struct itimerspec 	its;
  struct timespec 	now;
  struct sigaction sa1;
    
  sa1.sa_flags = SA_SIGINFO;
  sa1.sa_sigaction = noteaction;
  sigemptyset(&sa1.sa_mask);
    
  actionFired = 0;
  
  errno_check (sigaction(SIGALRM,&sa1,&sa1));
  errno_check (clock_gettime(CLOCK_REALTIME,&now));
  errno_check (timer_create(CLOCK_REALTIME,NULL,&tid));
    
  its.it_value.tv_sec = 3;
  its.it_value.tv_nsec = 0;
  its.it_interval.tv_sec = 0;
  its.it_interval.tv_nsec = 0;
    
  errno_check (timer_settime(tid,0,&its,NULL));
    
  /* spin for 10 seconds */
  {
    int abit = now.tv_sec + 10;
    do {
	errno_check (clock_gettime(CLOCK_REALTIME,&now));
    } while(now.tv_sec < abit);
  }
  
  assert(actionFired);
  assert(actionInfo.si_signo == SIGALRM);
  assert(actionInfo.si_code == SI_TIMER);
  assert(actionInfo.si_value.sival_ptr == tid);
  
  errno_check (timer_delete(tid));
  errno_check (sigaction(SIGALRM,&sa1,NULL));

  return 0;
}

#define NTHREAD	5
static void *crunch(void *arg)
{
  unsigned i = 50000;
  
  while(i--)
    rand();
  
  clock_getelapsed ((struct timespec *)arg);
  
  return NULL;
}

static int test12(void)
{
  int i;

  /* fire off 5 round robin threads */
  pthread_t		ttab[NTHREAD];
  struct timespec	stime[NTHREAD];
  struct timespec	etime[NTHREAD];
  
  pthread_attr_t	tha;
  
  /* set to minimum priority, round-robin */
  err_check (pthread_attr_init(&tha));
  err_check (pthread_attr_setinheritsched(&tha,PTHREAD_EXPLICIT_SCHED));
  err_check (pthread_attr_setschedpolicy(&tha,SCHED_RR));
  {
    struct sched_param sp;
    err_check (pthread_attr_getschedparam(&tha,&sp));
    sp.sched_priority = errno_check (sched_get_priority_min(SCHED_RR));
    err_check (pthread_attr_setschedparam(&tha,&sp));
  }
  
  
  /* start each thread */
  for(i=0;i<NTHREAD;i++) {
    clock_getelapsed (&stime[i]);
    err_check (pthread_create(&ttab[i],&tha,crunch,&etime[i]));
  }
  
  /* wait for each thread to exit */
  for(i=0;i<NTHREAD;i++) {
      err_check (pthread_join(ttab[i],NULL));
  }

  /* now print the end times */
  for(i=0;i<NTHREAD;i++) {
    printf("Thread[%d] start %2ld.%09ld end %2ld.%09ld\n",
	   i, 
	   stime[i].tv_sec, 
	   stime[i].tv_nsec,
	   etime[i].tv_sec,
	   etime[i].tv_nsec);
  }
  return 0;
}


volatile unsigned int ticks;

static void ticker (int signo)
{
    /* strictly calling stdio from a signal handler is illegal!! */
    fprintf (stderr, "tick ");
    ++ticks;
}

int test13 () {
  timer_t tid;
  struct itimerspec 	its;
  struct timespec 	now;
  struct sigaction sa1;
  unsigned int wakeups;
    
  sa1.sa_flags = 0;
  sa1.sa_handler = ticker;
  sigemptyset(&sa1.sa_mask);
    
  ticks = 0;
  
  errno_check (sigaction(SIGALRM,&sa1,&sa1));
  
  errno_check (clock_gettime(CLOCK_REALTIME,&now));

  errno_check (timer_create(CLOCK_REALTIME,NULL,&tid));
    
  its.it_value.tv_sec = 1;
  its.it_value.tv_nsec = 0;
  its.it_interval.tv_sec = 1;
  its.it_interval.tv_nsec = 0;
    
  errno_check (timer_settime(tid,0,&its,NULL));
    
  /* sleep for 5 seconds */
#define SLEEP_SECS	5
  {
    struct timespec delay;
    int rv;
    delay.tv_sec = SLEEP_SECS;
    delay.tv_nsec = 0;
    wakeups = 0;
    do {
      struct timespec rem;
      rv = nanosleep (&delay, &rem);
      assert (rv == 0 || errno == EINTR);
      if (rv != 0) {
	assert (rv == -1);
	assert (rem.tv_sec < delay.tv_sec ||
		(rem.tv_sec == delay.tv_sec && rem.tv_nsec < delay.tv_nsec));
	delay = rem;
	++wakeups;
      }
    } while (rv != 0);
  }
  fprintf (stderr, "\n");

  errno_check (timer_delete(tid));
  assert(ticks == SLEEP_SECS);
  assert(wakeups == ticks);

  {
    unsigned int oticks = ticks;
    assert (sleep (2) == 0);
    assert (ticks == oticks);
  }

  errno_check (sigaction(SIGALRM,&sa1,NULL));

  return 0;
}


int main()
{

  pthread_attr_t hitha;
  
  printf("Main thread running\n");
  
  /* set main thread to highest priority */
  err_check (pthread_attr_init(&hitha));
  err_check (pthread_attr_setinheritsched(&hitha,PTHREAD_EXPLICIT_SCHED));
  err_check (pthread_attr_setschedpolicy(&hitha,SCHED_FIFO));
  {
    struct sched_param sp;
    err_check (pthread_attr_getschedparam(&hitha,&sp));
    sp.sched_priority = errno_check (sched_get_priority_max(SCHED_FIFO));
    err_check (pthread_attr_setschedparam(&hitha,&sp));
  }
  
  /* check we can yield */
  errno_check (sched_yield());

  /* check that we can find ourselves */
  err_check (pthread_kill(pthread_self(),0));
  
  assert( test12() == 0);
  pass(12);

  /* test 1 check pthread_create same prio does not yield until told to
   */
  {
    pthread_t 		id;
    void		*rv;
    
    noteExpect = 10;
    
    note((void *)10);
    
    err_check (pthread_create(&id,NULL,note,(void *)(12)));
    
    note((void *)(11));
    sched_yield(); 
    note((void *)(13));
    
    err_check (pthread_join(id,&rv));

    assert((int)rv == 12);
  }
  pass(1);
  
  /* test 2 check pthread_create same prio does yield when told to
   */
  {
    pthread_t 		id;
    void		*rv;

    noteExpect = 20;
    note((void *)(20));
    err_check (pthread_create(&id,NULL,note,(void *)(21)));

    sched_yield(); 
    note((void *)(22));
    
    err_check (pthread_join(id,&rv));

    assert((int)rv == 21);
  }
  pass(2);
  
  /* test 3 check pthread_create hi prio does yield when not told
   */
  {
    pthread_t 		id;
    void		*rv;

    noteExpect = 30;
    
    note((void *)(30));
    
    err_check (pthread_create(&id,&hitha,note,(void *)31));

    note((void *)(32));
    
    err_check (pthread_join(id,&rv));

    assert((int)rv == 31);
  }
  pass(3);
  
  /* test 4 check sem does block 
   */
  {
    pthread_t 		id;
    void		*rv;

    noteExpect = 40;
    note((void *)(40));

    err_check (pthread_create(&id,&hitha,noteWithSem,(void *)43));

    note((void *)41);
    sched_yield(); 
    note((void *)42);
    errno_check (sem_post(&notesem));
    note((void *)44);
    
    err_check (pthread_join(id,&rv));

    assert((int)rv == 43);
    
  }
  pass(4);
  
  /* test 5 check sem does queue 
   */
  {
    pthread_t 		id1;
    pthread_t 		id2;
    pthread_t 		id3;
    void		*rv;

    noteExpect = 50;
    
    note((void *)(50));

    err_check (pthread_create(&id1,&hitha,noteWithSem,(void *)(52)));
    err_check (pthread_create(&id2,&hitha,noteWithSem,(void *)(54)));
    err_check (pthread_create(&id3,&hitha,noteWithSem,(void *)(55)));

    note((void *)(51));
    errno_check (sem_post(&notesem));
    note((void *)(53));
    errno_check (sem_post(&notesem));
    errno_check (sem_post(&notesem));
    note((void *)(56));
    
    err_check (pthread_join(id1,&rv));
    assert((int)rv == 52);

    err_check (pthread_join(id2,&rv));
    assert((int)rv == 54);

    err_check(pthread_join(id3,&rv));
    assert((int)rv == 55);
    
  }
  pass(5);

  /* test 6 check can/cannot kill(0) a created thread
   */
  {
    pthread_t 	id;
    void	*rv;
    int		v;

    noteExpect = 60;
    
    note((void *)(60));
    
    err_check (pthread_create(&id,&hitha,noteWithSem,(void *)(61)));
    
    err_check (pthread_kill(id,0));
    
    errno_check (sem_post(&notesem));
    
    v = pthread_kill(id,0);
    assert (v == 0 || v == ESRCH);
    
    note((void *)(62));
    
    err_check (pthread_join(id,&rv));
    assert((int)rv == 61);
  }
  pass(6);

  /* test 7 check can kill a thread 
   */
  {
    pthread_t id;
    
    noteExpect = 70;
    
    note((void *)(70));
    
    err_check (pthread_create(&id,&hitha,noteWithSem,(void *)(70)));
    
    err_check (pthread_cancel(id));
    
    errno = 0;
    if (sem_trywait(&notesem) != -1 && errno != EAGAIN) {
	perror("sem_trywait");
	abort ();
    }

    errno_check (sem_post(&notesem));
    
    errno_check (sem_trywait(&notesem));
    
    note((void *)(71));
    
  }
  pass(7);

  /* test8 check sigaction/pthread_kill */
  {
    struct sigaction sa;
    
    sa.sa_flags = SA_SIGINFO;
    sa.sa_sigaction = noteaction;
    sigemptyset(&sa.sa_mask);

    errno_check (sigaction(SIGRTMIN + 2,&sa,&sa));
    
    actionFired = -1;
    
    err_check (pthread_kill(pthread_self(),SIGRTMIN + 2));
    
    assert(actionFired == SIGRTMIN + 2);
    
    errno_check (sigaction(SIGRTMIN + 2,&sa,NULL));
  }
  pass(8);

  assert( test9() == 0);
  pass(9);

  assert( test10() == 0);
  pass(10);
  
  assert( test11() == 0);
  pass(11);

  assert( test12() == 0);
  pass(12);
  
  assert( test13() == 0);
  pass(13);

  printf("All tests done\n");
  pthread_exit((void *)0x22);

  return (0);
}
